/*
 * GCPolicy.java
 * Base class for objects that implement a specific policy for the GC.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * @JANOSVM_JVMLIB_LICENSE@
 */

package edu.utah.janosvm.resources;

/**
 * Base class for objects that implement a specific policy for the GC.
 *
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class GCPolicy
{
    /**
     * The maximum amount of memory available to this GC.
     */
    protected final long max;

    private GCPolicy()
    {
	this.max = 0;
    }

    public GCPolicy(long max)
    {
	this.max = max;
    }

    /**
     * @return true if the GC was successful in starting all of its threads.
     */
    public native boolean startThreads();

    /**
     * The enable function is called just before the team is finished being
     * initialized, so you have a chance to do something, like start the GC
     * and finalizer threads.
     *
     * @return True if the enable was successful, false otherwise.
     */
    public boolean enable()
    {
	return this.startThreads();
    }

    /**
     * @return The initial number that should be used when trying to decide
     * whether to do a GC or allocate more memory from the OS.
     */
    public long initialTrigger()
    {
	return 250 * 1024;
    }

    /**
     * This function is called after a GC happened to determine the next time
     * to do a GC.
     *
     * @param used The current amount of memory used.
     * @param freed (Bogus at this time I think).
     * @return The new trigger for the GC.
     */
    public long newTrigger(long used, long freed)
    {
	return used + used / 3;
    }

    /**
     * @return The GCPolicy for the current team.
     */
    public static native GCPolicy currentPolicy();
}
