/*
 * KernelReaper.java
 * The class used to represent a team
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * @JANOSVM_JVMLIB_LICENSE@
 */

package edu.utah.janosvm.sys;

import java.util.Vector;
import edu.utah.janosvm.resources.ResourceSpec;
import edu.utah.janosvm.resources.ResourceSpecException;

/**
 * Kernel class used to represent a team, this is then extended internally
 * with a nativeClass.  The class is an Exportable since all instances
 * of RemoteTeam are allocated in the kernel team.  The team they represent
 * has a LocalTeam instance that points to its RemoteTeam.
 *
 * @see LocalTeam
 * @see Team
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class RemoteTeam
    extends Exportable
    implements PersistentExport
{
    /**
     * Reference to the kernel team.
     */
    private static RemoteTeam kernelTeam;

    /**
     * @return The kernel RemoteTeam object.
     */
    static RemoteTeam getKernelTeam()
    {
	return RemoteTeam.kernelTeam;
    }

    /**
     * Indicates whether or not the team is being terminated.
     */
    private boolean dying = false;

    /**
     * The object local to the team's heap that contains
     * per-team state specific to that team.  Specifically,
     * it contains the tables of items exported from the team.
     */
    private LocalTeam peer;

    /**
     * Queue of teams to be reaped.
     *
     * XXX move this stuff into a separate class.
     */
    private static final Object[] reaperQueue = new Object[REAPERQUEUE_SIZE];
    private static final int[] reaperQueueOp = new int[REAPERQUEUE_SIZE];
    private static int reaperQueueHead = 0;
    private static int reaperQueueTail = 0;
    private static final int REAPERQUEUE_SIZE = 10;
    private static final int REAPERQOP_TEAMTERMINATE = 1;
    private static final int REAPERQOP_TEAMCOLLECT   = 2;

    /**
     * Gets the LocalTeam peer associated with this RemoteTeam.
     *
     * @return The peer object.
     */
    LocalTeam getPeer()
    {
	return this.peer;
    }

    /**
     * Constructs the native part of the class.
     *
     * @param name The name of this team.
     * @param specs The resources specifications for the team.
     */
    private native boolean nativeConstructor(String name,
					     ResourceSpec specs[],
					     kaffe.util.Ptr nativeSpecs)
	throws ResourceSpecException;

    /**
     * Finalizes the native part of the class.
     */
    private native void nativeFinalize();

    /**
     * Tests to see if an object is owned by this team.
     * (Ownership is defined as paying for the memory used by
     * the object instance.)
     *
     * @param obj The object to test for ownership.
     * @return True if this RemoteTeam owns the given object.
     */
    native boolean owns(Object obj);

    /**
     * Start a visit.  A visit is completed with a matching call
     * to <code>endVisit()</code>.
     *
     * @param wrapCount Indicates the number of wrapper functions around this
     * one.  This is required so that we can record where visit take place and
     * can properly back out of them if an exception is thrown.
     * @return Whether the visit was successful or not.
     * @see #endVisit()
     */
    native boolean visit(int wrapCount) throws DeadTeamException;

    /**
     * Basic visit wrapper provided for code that operates directly on
     * RemoteTeams.
     *
     * @return Whether the visit was successful or not.
     */
    boolean visit()
	throws DeadTeamException
    {
	return this.visit(1);
    }
    
    /**
     * End a visit and return to the previous team.
     * <p>
     * XXX this shouldn't be static.  Just to make the symmetry a bit
     * more apparent, if nothing else.  Also would be a bit easier
     * to add sanity checking code with a this pointer.
     */
    static native void endVisit();

    /**
     * Crash into the team boundary, this is basically a violent exit from
     * a visit and results in a DeadTeamException being thrown in the
     * previous team context.  NOTE: This is only used for testing, its
     * possible for the VM to do something similar so we needed a way to
     * explicitly cause it.
     */
    public static native void crash();
    
    /**
     * Get the RemoteTeam associated with the current thread.
     * This returns whatever team the thread is currently executing
     * in (i.e., during a visit), and is not necessairly the home
     * team of the thread.
     *
     * @return The team that this thread is currently running in.
     */
    static native RemoteTeam getCurrentTeam();

    /**
     * Get the unique integer identifier for this team.
     * Identifiers are uniquie among currently executing
     * teams, but may be re-used when a team is destroyed.
     * (Analogous to a "PID" in unix.)
     *
     * @return The unique integer identifier for this team.
     */
    native int getID();

    /**
     * Anchor an object so the collector doesn't free it.  The
     * object becomes a "root" in the collector so the collector
     * will never free it.  The object must be explicitly
     * un-anchored before it will be freed.
     *
     * @see #floatObject(Object)
     */
    private native void anchorObject(Object obj);

    /**
     * Float a previously anchored object.  The object
     * will be reclaimed when it is unreachable.
     *
     * @see #anchorObject(Object)
     */
    private native void floatObject(Object obj);
    
    /**
     * Register an object suspected to be a leak.
     */
    public static native void registerLeak(Object obj);

    /**
     * Unregister an object suspected to be a leak.
     */
    public static native void unregisterLeak(Object obj);

    /**
     * Constructor used for the kernel team.  Starts the
     * kernel reaper thread, too.
     */
    private RemoteTeam()
    {
	/* Need to use the special Exportable constructor */
	super(0);
	RemoteTeam.kernelTeam = this;
	this.exportKernel("kernel");
	/*
	 * The native structure will setup the resources and migrate us
	 * to this team
	 */
	try
	{
	    this.nativeConstructor("kernel", null, null);
	    this.peer = new LocalTeam(this, this);
	}
	catch(Throwable th)
	{
		// XXX
	}

	/*
	 * Start the kernel reaper thread.
	 *
	 * XXX prolly shoudn't invoke java.lang stuff here...
	 */
	new Thread(new Runnable() {
	    public void run() {
		reaper();
	    }
	}, "Kernel Reaper Thread");
    }

    /**
     * Create a user team with the given name and resources
     *
     * @param name The name of the team.
     * @param specs The specifications required by the team.
     */
    RemoteTeam(String name, ResourceSpec specs[])
	throws ResourceSpecException, DuplicateExportException
    {
	    this(name, specs, null);
    }
    
    /**
     * Create a user team with the given name and resources.  This is the
     * internal version that can be used by the native code to construct a
     * team with native resource specifications.
     *
     * @param name The name of the team.
     * @param specs The specifications required by the team.
     * @param nativeSpecs Native array of resource specifications.
     */
    private RemoteTeam(String name,
		       ResourceSpec specs[],
		       kaffe.util.Ptr nativeSpecs)
	throws ResourceSpecException, DuplicateExportException
    {
	if( this.nativeConstructor(name, specs, nativeSpecs) )
	{
	    String localName = null;
	    
	    try
	    {
		ExportTable et = null;
		Vector local = null;
		Team team = null;
		    
		if( this.visit() )
		{
		    /*
		     * NOTE: We are now in the new team, but the
		     * team object is held in kernel space, so any
		     * memory allocated and referenced by the
		     * RemoteTeam object will not be traced by the
		     * new team's collector.  Therefore, any
		     * objects created must be anchored or they
		     * might be freed.
		     */
		    if( (name != null) && !name.equals("") )
		    {
			localName = (String)name.clone();
			this.anchorObject(localName);
		    }
		    
		    /* Create the peer object */
		    this.peer = new LocalTeam(kernelTeam, this);
		    this.anchorObject(this.peer);
		    
		    RemoteTeam.endVisit();
		}
		/* Register the team in the export table */
		if( localName != null )
		    this.exportObject(localName);
	    }
	    catch(DeadTeamException e)
	    {
		/* Shouldn't happen */
		throw new InternalError();
	    }
	    catch(DuplicateExportException e)
	    {
		this.terminate();
		throw e;
	    }
	}
	else
	{
	    ExportTable et = LocalTeam.fetch().getExports();

	    synchronized(et)
	    {
		this.remove();
	    }
	    throw new ResourceSpecException();
	}
    }

    /**
     * Get the name of the team.
     */
    public String getName()
    {
	return (String)this.getIdentifier();
    }

    /**
     * Native level terminate code
     */
    private native void nativeTerminate();

    /**
     * Terminate this team, revoke all its exports, kill its threads and
     * return its resources to the system.
     * <p>
     * If the calling thread is part of the team being terminated
     * then this function will not return.
     * <p>
     * Note that termination is delayed by any active visits from
     * other teams, so to provide quick termination all visits must be
     * "short".
     */
    void terminate()
    {
	boolean responsible = false;
	
	/* Need to be in the kernel to do this */
	try
	{
	    if( RemoteTeam.kernelTeam.visit() )
	    {
		ExportTable et = null;

		try
		{
		    /* Visit the team and revoke its exports */
		    if( this.visit() )
		    {
			try
			{
			    /*
			     * Only the first time the Team is
			     * terminated is anyone responsible
			     * for cleaning up the external
			     * references.
			     */
			    synchronized(this)
			    {
				if( this.dying == false )
				{
				    this.dying = true;
				    responsible = true;
				}
			    }
			    if( responsible )
			    {
				/* Cut external refs */
				this.peer.getExports().revokeAll();
				Importable.revokeAll();
				/*
				 * Null the references since our memory
				 * will disappear soon.  And we need to
				 * stop anyone from exporting more
				 * stuff.
				 */
				this.peer = null;
			    }
			}
			catch(Throwable th)
			{
			    System.debug("Error/Exception during terminate");
			    System.debugE(th);
			}
			RemoteTeam.endVisit();
		    }

		    /* Let the kernel's reaper thread do the rest */
		    enqueueForReaper();

		    if( responsible )
		    {
			kernelTeam.anchorObject(this);
			/* Revoke any references to us. */
			this.revoke();
			try
			{
			    this.nativeTerminate();
			}
			catch(Throwable th)
			{
			    System.debug("Error/Exception during nativeTerminate");
			    System.debugE(th);
			}
		    }
		}
		catch(DeadTeamException dte)
		{
		}
		/*
		 * If a team is being killed by one of its own threads
		 * then this is the actual spot where it will exit and
		 * wait to be garbage collected by the kernel gc.
		 */
		RemoteTeam.endVisit();
	    }
	}
	catch(DeadTeamException dte)
	{
	}
    }
    
    /**
     * Enqueue this thread on the kernel reaper's queue for
     * slaughter.  Doing this ensures a couple things: First, that the
     * kernel is charged for cleaning up the process state (and not
     * whatever thread happens to leave the process last).  Second,
     * that the team is always terminated by a thread from another
     * team, avoiding race conditions in resources.
     * <p>
     * This function
     * returns immediately after enqueuing this team.
     * <p>
     * Only called by terminate().
     *
     * @see #terminate()
     */
    private void enqueueForReaper()
    {
	synchronized(reaperQueue)
	{
	    if ((reaperQueueHead+1 % REAPERQUEUE_SIZE) == reaperQueueTail)
		throw new InternalError("Reaper Queue has filled up.  Replace fixed-size queue with a linked list, please.");

	    reaperQueueHead++;
	    reaperQueue[reaperQueueHead] = this;
	    reaperQueueOps[reaperQueueHead] = REAPERQOP_TEAMTERMINATE;

	    reaperQueue.notify();
	}
    }

    /**
     * The kernel reaper work function.  Never returns.
     */
    private void reaper()
    {
	while (true)
	{
	    synchronized(reaperQueue)
	    {
		while(reaperQueue.isEmpty())
		    reaperQueue.wait();

		reaperQueue.remove
	    }
	}
    }

    public final void finalize()
    {
	/* Call the native finalizer to clean up an internals */
	this.nativeFinalize();
    }

    /**
     * @param wrapCount Indicates the number of wrapper functions around this
     * one.  This is required so that we can record where pushes take place and
     * can properly back out of them if an exception is thrown.
     * @return A TeamActivation object that reserves this team.
     */
    native TeamActivation pushActivation(int wrapCount);

    /**
     * Pop the last TeamActivation off the stack.
     */
    static native void popActivation();

    /**
     * Store a per-team class local object.
     *
     * @param cl The class the object should be stored with.
     * @param obj The object to store with the class.
     */
    public static native void setClassLocal(Class cl, Object obj);
    
    /**
     * @param cl The class to retrieve the local object for.
     * @return The Object associated with the class or null.
     */
    public static native Object getClassLocal(Class cl);
}
