/*
 * ExportTable.java
 * Manages any objects exported from a team
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * @JANOSVM_JVMLIB_LICENSE@
 */

package edu.utah.janosvm.sys;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/**
 * Class that manages any objects exported from a team.  Each team has
 * exactly one instance of this table.
 * <p>
 * This table tracks two kinds of Exportables, "identified" ones and
 * anonymous ones.  Identified Exportables need a unique identifier.
 *
 * @see LocalTeam#exports
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
class ExportTable
{
    /**
     * Hashtable for tracking exports that have some associated identifier.
     * Most likely these are persistent exports that can have many clients.
     *
     * @see Exportable#Exportable(Object)
     */
    private Hashtable table = new Hashtable();
	
    /**
     * Vector for tracking anonymous exports, these are most likely
     * temporary exports that have some peer in another team.  So we need
     * to have a reference to them in the current team, otherwise they'll
     * be GC'ed.
     *
     * @see Exportable#Exportable(Object)
     */
    private List anonymous = new List();

    /**
     * Construct an empty ExportTable.
     * <p>
     * XXX change to take parameters of Hashtable and pass on to this.table's Hashtable.
     */
    public ExportTable()
    {
    }

    /**
     * Get the exported object with the given identifier.
     *
     * @param id Unique identifier associated with the export.
     * @return Null or the exported object.
     */
    public synchronized Exportable get(Object id)
    {
	return (Exportable)this.table.get(id);
    }

    /**
     * Put an object in the ExportTable which will effectively make it
     * available for other teams to import.
     *
     * @param id Unique(?) identifier to associate with the export.
     * @param ex The object to export.
     * @exception Throws DuplicateExportException if there is already an export
     * with the given Identifier.
     */
    public synchronized void put(Object id, Exportable ex)
	throws DuplicateExportException
    {
	if( this.table.get(id) == null )
	    this.table.put(id, ex);
	else
	    throw new DuplicateExportException(id.toString());
    }

    /**
     * Remove the export associated with the given identifier.
     *
     * @param id The identifier associated with the exported object.
     * @return the object removed or null if none was in the table.
     */
    public Object remove(Object id)
    {
	return this.table.remove(id);
    }

    /**
     * Add an anonymous export to the table.
     *
     * @param ex The object to export.
     */
    public void addAnonymous(Exportable ex)
    {
	this.anonymous.addTail(ex);
    }

    /**
     * Remove an anonymous export.
     *
     * @param ex The object to remove from the table.
     */
    public void remAnonymous(Exportable ex)
    {
	ex.remove();
    }

    /**
     * Revoke all exported objects.  Iterates through
     * all of the objects being tracked and invokes
     * the revoke method on each.
     *
     * @see Exportable#revoke()
     */
    public synchronized void revokeAll()
    {
	Exportable ex;
	Enumeration e;
	int lpc, size;

	e = this.table.elements();
	size = this.table.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    ex = (Exportable)e.nextElement();
	    ex.revoke();
	}
	while( (ex = (Exportable)this.anonymous.remHead()) != null )
	{
	    ex.revoke();
	}

	// this is why the fields aren't final:
	this.table = null;
	this.anonymous = null;
    }
}
