/*
 * Exportable.java
 * Base class for objects exported to other teams.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * @JANOSVM_JVMLIB_LICENSE@
 */

package edu.utah.janosvm.sys;

/**
 * All objects that are to be safely referenced and manipulated
 * by teams other than the allocating team must be "Exportable".
 * This is required so that the JanosVM can know to track all
 * the inter-team references, so that when it terminates a Team
 * it can safely clean up all the references that point to
 * objects exported by that team.
 * <p>
 * See the description of Importable for some details on how
 * to use Importable and Exportable.
 * <p>
 * XXX Should this be abstract?  (Would one ever want to create an
 * instance of 'Exportable' without subclassing it?)
 *
 * @see ObjectExporter
 * @see Importable
 * @see TemporaryExport
 * @see PersistentExport
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */ 
public class Exportable
    extends ListNode
{
    /**
     * The list of objects that import this object.
     */
    private List subscribers = new List();

    /**
     * Indicates whether this export is still valid.
     */
    private boolean valid = true;

    /**
     * The team that is exporting this object.
     */
    private RemoteTeam owner;

    /**
     * The object that uniquely identifies this export.
     */
    private Object id;

    /**
     * Special constructor for the kernel RemoteTeam object.
     */
    public Exportable(int dummy)
    {
    }

    /**
     * Construct an anonymous object that exports `this'.
     */
    public Exportable()
    {
	this.owner = RemoteTeam.getCurrentTeam();
	try
	{
	    this.exportObject(null);
	}
	catch(DuplicateExportException e)
	{
	    throw new InternalError("It's anonymous this shouldn't happen...");
	}
    }

    /**
     * Construct an object that exports `this' under the given identifier.
     *
     * @param id Unique identifier the export should be identified by.
     * @exception DuplicateExportException if there is already an export
     * with the given Identifier.
     */
    public Exportable(Object id)
	throws DuplicateExportException
    {
	this.owner = RemoteTeam.getCurrentTeam();
	this.exportObject(id);
    }

    /**
     * Special function for exporting the kernel RemoteTeam object.
     */
    void exportKernel(Object id)
    {
	this.valid = true;
	this.id = id;
	this.owner = (RemoteTeam)this;
    }

    /**
     * Export `this' under the given identifier.
     *
     * @param id Unique identifier the export should be identified by.
     * @exception Throws DuplicateExportException if there is already an export
     * with the given Identifier.
     */
    public void exportObject(Object id)
	throws DuplicateExportException
    {
	ExportTable et = LocalTeam.fetch().getExports();

	if( et != null )
	{
	    synchronized(et)
	    {
		this.remove();
		if( id == null )
		{
		    et.addAnonymous(this);
		}
		else
		{
		    et.put(id, this);
		}
		this.id = id;
	    }
	}
	else
	{
	    /* We're dying */
	    this.valid = false;
	}
    }
    
    /**
     * @return The unique identifier for this export or null if its anonymous.
     */
    public Object getIdentifier()
    {
	return this.id;
    }

    /**
     * @return The owner of this export.
     */
    RemoteTeam getOwner()
    {
	return this.owner;
    }

    /**
     * Visit the team that owns this Exportable.
     *
     * @exception Throws DeadTeamException if you or the other team is dead.
     */
    public void visitOwner()
	throws DeadTeamException
    {
	if( !this.owner.visit(1) )
	    throw new DeadTeamException("other team dead in visitOwner()");
    }

    /**
     * End a visit to an owner team.
     */
    public void endOwnerVisit()
    {
	RemoteTeam.endVisit();
    }
    
    /**
     * @return The TeamActivation for the owner or null if you or the other
     * team is dead.
     */
    public TeamActivation pushOwnerActivation()
    {
	TeamActivation retval;

	retval = this.owner.pushActivation(1);
	return retval;
    }

    /**
     * Pop the owner activation off the stack.
     */
    public void popOwnerActivation()
    {
	RemoteTeam.popActivation();
    }
    
    /**
     * Subscribe an importable to a team.  NOTE: Should be called from within a
     * visit (XXX Still applicable with activations?).
     * <p>
     * XXX perhaps a better name for this would be 'addSubscriber'
     *
     * @param im The object that wants to import this export.
     * @return True if the subscription succeeded, false otherwise.
     */
    public boolean subscribe(Importable im)
    {
	boolean retval = false;

	synchronized(this.subscribers)
	{
	    if( this.valid )
	    {
		this.subscribers.addTail(im);
		im.bind(this);
		retval = true;
	    }
	}
	return retval;
    }

    /**
     * Prepare to transfer the import from one Importable to another.  This is
     * different from subscribe since the Exportable won't revoke itself if its
     * a temporary.
     *
     * @param im The Importable that had ownership and is transfering it to
     * someone else.
     */
    void transfer(Importable im)
    {
	synchronized(this.subscribers)
	{
	    im.remove();
	}
	im.unbind();
    }

    /**
     * Unsubscribe the given import from the export.  Should be
     * called from within a visit.  If this exportable implements the
     * TemporaryExport interface and the given importable was the
     * last, this exportable will be revoked.
     *
     * @param im The importable to unsubscribe.
     */
    public void unsubscribe(Importable im)
    {
	synchronized(this.subscribers)
	{
	    im.remove();
	    if( this.subscribers.isEmpty() )
	    {
		if( this instanceof TemporaryExport )
		    this.revoke();
	    }
	}
	im.unbind();
    }

    /**
     * Drop any references to objects.
     */
    public void unbind()
    {
	this.id = null;
    }
    
    /**
     * Revoke this export.  The owner of the exportable is visited,
     * and the exportable is removed from the owner's export table.
     * All the importables subscribed to this exportable are poked
     * so that they drop their references to this exportable.  Any
     * importables which implement the ManagedExportable interface have
     * their 'release' method invoked.
     * <p>
     * XXX isn't this always invoked in the context of the owner of the
     * exportable?
     *
     * @see Importable#unbind()
     * @see ManagedExportable
     */
    public final void revoke()
    {
	try
	{
	    if( this.owner.visit() )
	    {
	        LocalTeam lt = LocalTeam.fetch();
		ExportTable et = lt.getExports();
		boolean responsible = false; // XXX replace with 'subs != null'
		Importable subs[] = null;
		int lpc = 0;
		
		/* Remove this from the owner's export table */
		if( et != null )
		{
		    synchronized(et)
		    {
			if( this.id == null )
			    et.remAnonymous(this);
			else
			    et.remove(this.id);
		    }
		}

		/*
		 * Copy the importables out of the linked list they're
		 * in, and into a linear array. XXX Why?
		 */
		synchronized(this.subscribers)
		{
		    if( this.valid )
		    {
			Importable im;
			
			responsible = true;
			this.valid = false;
			this.unbind();
			subs = new Importable[this.subscribers.count()];  // XXX allocation
			while( (im = (Importable)this.subscribers.remHead())
			       != null )
			{
			    subs[lpc] = im;
			    lpc++;
			}
		    }
		}
		if( responsible )
		{
		    for( lpc = 0; lpc < subs.length; lpc++ )
		    {
			subs[lpc].unbind();
		    }
		    if( this instanceof ManagedExportable )
		    {
			((ManagedExportable)this).release();
		    }
		}
		RemoteTeam.endVisit(); // XXX this.owner.endVisit()
	    }
	}
	catch(DeadTeamException dte)
	{
	}
    }

    protected void finalize()
    {
	this.revoke();
    }
}
