/*
 * Importable.java
 * Base class for objects imported into teams.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * @JANOSVM_JVMLIB_LICENSE@
 */
package edu.utah.janosvm.sys;

/**
 * An Importable is a safe, cross-process reference to a remote
 * "Exportable" object.
 * <p>
 * Methods on Importable and subclasses of Importable may store
 * pointers in a remote process.  Beware that getting these operations
 * wrong can create bogus pointers into remote heaps.  Such problems
 * generally manifest as the VM crashing.
 * <p>
 * Example of a 'Foo' and 'FooRef'.
 * <pre>
 * // The "server" object that other Teams will have references to
 * class Foo extends Exportable {
 *   ...
 * }
 *
 * // The "client" object that Teams have locally as a reference to the
 * // "server"s Foo.
 * class FooRef extends Importable {
 *   ...
 *   private FooRef() { } 
 *
 *   // Lookup a Foo in the given team.
 *   public static FooRef lookupFoo(Team team, int id, ...) {
 *     Foo remoteFoo;
 *     FooRef refFoo = new FooRef();
 *     
 *     team.visit();
 *     remoteFoo = Foo.lookup(id, ...);
 *     team.endVisit();
 *    
 *     refFoo.bind(remoteFoo);
 *
 *     return refFoo;
 *   }
 *
 *   // perform an operation via the foo ref on the foo
 *   public void performOp(...) {
 *     synchronized(this) {
 *       Foo remoteFoo = super.get();
 *
 *       // do some stuff in the context of current team
 *       remoteFoo.doSomethingHere(...);
 *   
 *       // do some stuff in the context of the owner
 *       remoteFoo.owner().visit();
 *       remoteFoo.doSomethingThere(...);
 *       remoteFoo.owner().endVisit();
 *     }
 *   }
 *   ...
 * }
 * </pre>
 *
 * In the above example, a <code>FooRef</code> is bound to a
 * particular Team's <code>Foo</code> object by looking the object up
 * with some sort of key (an <code>int</code> in this example).
 * Operations on the remote foo are perfomed by invoking
 * <code>performOp()</code> on the FooRef.  Internally this function
 * visits the owner of the actual Foo, and performs the operation.
 * <p>
 * Note that all of the ops are done in a synchronized block on the
 * importable.  This will prevent the remote team from destroying the
 * Foo while the operation is being performed on it (unless the operation
 * in question is to destroy the Foo... be careful).
 *
 * @see Team#visit()
 * @see Exportable
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class Importable
    extends ListNode
{
    /**
     * The thing we're importing.
     */
    private Exportable export;
    
    /**
     * Register the object internally so that we're revoked when the
     * team dies.
     *
     * @return True if the registration was successful, false otherwise.
     */
    private native boolean register();
    
    /**
     * Unregister the object internally.
     */
    private native void unregister();

    /**
     * Revoke all Importables belonging to the current team.
     * <p>
     * XXX This sucks.
     */
    static native void revokeAll();

    /**
     * Create a bare importable, waiting to be used to import something.
     */
    public Importable()
    {
	if( !this.register() )
	    throw new Error(); // XXX new error, better message
    }

    /**
     * Bind this Importable to the given Exportable.
     *
     * @param export The exportable that manages the object.
     */
    protected synchronized void bind(Exportable export)
    {
	this.export = export;
    }

    /**
     * Fetch the exportable object referenced from this Importable.  This
     * should only be used within a synchronized block on this object,
     * otherwise the object may go dead while it is being used.
     *
     * @return A direct pointer to the imported object.
     */
    protected Exportable get()
    {
	// XXX Assert(this.locked());
	return this.export;
    }

    /**
     * @param im The source to copy the imported object from.
     */
    protected void copy(Importable im)
    {
	Exportable ex = null;

	if( this.export != null )
		this.revoke();
	synchronized(im)
	{
	    if( (im.export != null) &&
		(im.export.pushOwnerActivation() != null) )
	    {
		ex = im.export;
	    }
	}
	if( ex != null )
	{
	    ex.subscribe(this);
	    Team.popActivation();
	}
    }

    /**
     * @param im The Importable to swap imports with.
     */
    protected void swap(Importable im)
    {
	Exportable thisExport = null, imExport = null;

	synchronized(im)
	{
	    if( (im.export != null) &&
		(im.export.pushOwnerActivation() != null) )
	    {
		imExport = im.export;
	    }
	}
	synchronized(this)
	{
	    if( (this.export != null) &&
		(this.export.pushOwnerActivation() != null) )
	    {
		thisExport = this.export;
	    }
	    if( thisExport != null )
	    {
		thisExport.transfer(this);
	    }
	    if( imExport != null )
	    {
		imExport.transfer(im);
		imExport.subscribe(this);
	    }
	}
	synchronized(im)
	{
	    if( thisExport != null )
	    {
		thisExport.subscribe(im);
		Team.popActivation();
	    }
	}
	if( imExport != null )
	{
	    Team.popActivation();
	}
    }

    /**
     * Test logical equality of this Importable and the given
     * Importable.  Two Importables are equal if they point to
     * the exact same (i.e., ==) object.
     *
     * @return Whether or not the imported objects are the same.
     */
    public boolean equals(Object obj)
    {
	boolean retval = false;
	
	if( obj instanceof Importable )
	{
	    Importable imp = (Importable)obj;
	    Object us, them;
	    
	    synchronized(this)
	    {
		us = this.export;
	    }
	    synchronized(imp)
	    {
		them = imp.export;
	    }
	    if( us == them )
		retval = true;
	}
	return retval;
    }
	
    /**
     * Test to see if this Importable points to the given Exportable
     *
     * @param export Exportable to check against
     * @return true if this importable points to the given exportable.
     */
    public boolean pointsTo(Exportable export)
    {
	synchronized(this)
	{
	    return this.export == export;
	}
    }

    /**
     * Clears reference to the remote object.
     * <p>
     * Subclasses of Importable must override this to clear out any
     * cross-process object references which they have cached.
     */
    protected synchronized void unbind()
    {
	this.export = null;
    }

    /**
     * Revoke access to the imported object.
     */
    protected void revoke()
    {
	Exportable ex;

	ex = this.export;
	try
	{
	    if( (ex != null) && ex.getOwner().visit() )
	    {
		ex.unsubscribe(this);
		RemoteTeam.endVisit();
	    }
	}
	catch(DeadTeamException dte)
	{
	}
    }

    protected void finalize()
    {
	try
	{
	    this.revoke();
	}
	catch(Throwable th)
	{
	}
	this.unregister();
    }
    
    /**
     * Clone an Importable, basically just creates another object of this
     * type and copies the imported object.
     */
    public Object clone()
    {
	Class cl = this.getClass();
	Object retval = null;
	
	try
	{
	    Importable imp;
	    
	    retval = cl.newInstance();
	    imp = (Importable)retval;
	    imp.copy(this);
	}
	catch(InstantiationException ie)
	{
	}
	catch(IllegalAccessException iae)
	{
	}
	return retval;
    }
}
