/*
 * LocalTeam.java
 * Peer object for a team.
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * @JANOSVM_JVMLIB_LICENSE@
 */

package edu.utah.janosvm.sys;

import java.util.Vector;

import edu.utah.janosvm.resources.ResourceSpecException;

/**
 * Peer object for a RemoteTeam.  An instance of a LocalTeam
 * is created inside each Team to contain that team's local
 * data.  Specifically, this class keeps track of the Exportables 
 * coming from this team.  LocalTeam objects are never created
 * directly, they are created implicitly as part of 'new Team()'.
 * <p>
 * The association with the RemoteTeam in the kernel team is implicit
 * (through RemoteTeam.getCurrentTeam()).
 * <p>
 * Because this class extends Team, it is importable.
 *
 * @see RemoteTeam
 * @see Team
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public final class LocalTeam
    extends Team
{
    /**
     * The table of objects exported from this team.
     */
    private final ExportTable exports = new ExportTable();

    /**
     * The vector of data that is local to this team.  This
     * vector is not used by the JanosVM.  The protocol for
     * using this vector is up to the system hosted by the JanosVM.
     */
    private final Vector localData = new Vector();

    /**
     * The cached kernel team handle.
     */
    private final Team kernel;

    /**
     * XXX What a stupid name.
     *
     * @return The LocalTeam object for the current team.
     */
    public static LocalTeam fetch()
    {
	return RemoteTeam.getCurrentTeam().getPeer();
    }

    /**
     * Construct a LocalTeam object.
     * <p>
     * NOTE: For kernel use only; there is only one of these objects
     * per team.
     *
     * @param kern The kernel RemoteTeam object.
     * @param rt The RemoteTeam object to associate with
     */
    LocalTeam(RemoteTeam kern, RemoteTeam rt)
	throws ResourceSpecException
    {
	super(rt);
	this.kernel = new Team(kern);
    }

    /**
     * Get a handle on the kernel team.
     *
     * @return The cached kernel Team handle.
     */
    public Team getKernelTeam()
    {
	return this.kernel;
    }

    /**
     * Look up an exported object in the export table for this team.
     * 
     * @param id The identifier to search for in the export table.
     * @return The exported object or null if it wasn't found.
     * @see ExportTable
     */
    public Exportable getExport(Object id)
    {
	Exportable retval = null;
	
	synchronized(this.exports)
	{
	    retval = this.exports.get(id);
	}
	return retval;
    }

    /**
     * Get the ExportTable for the current team.
     *
     * @return The ExportTable for the current team.
     */
    ExportTable getExports()
    {
	return this.exports;
    }

    /**
     * @return The local data vector for the current team.
     */
    public Vector getData()
    {
	return this.localData;
    }

    /**
     * Store a per-team class local object.
     *
     * @param cl The class the object should be stored with.
     * @param obj The object to store with the class.
     */
    public void setClassLocal(Class cl, Object obj)
    {
	RemoteTeam.setClassLocal(cl, obj);
    }

    /**
     * @param cl The class to retrieve the local object for.
     * @return The Object associated with the class or null.
     */
    public Object getClassLocal(Class cl)
    {
	return RemoteTeam.getClassLocal(cl);
    }
}
