/*
 * Team.java
 * A handle for accessing RemoteTeam objects
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * @JANOSVM_JVMLIB_LICENSE@
 */

package edu.utah.janosvm.sys;

import java.util.Vector;
import edu.utah.janosvm.resources.ResourceSpec;
import edu.utah.janosvm.resources.ResourceSpecException;

/**
 * Provides a safe handle to a team.  Technically, this points to the
 * kernel team's RemoteTeam object that represents that team in
 * question.  This is also the public interface to Team creation.  A
 * parent process creates a new Team object which implicitly creates
 * the underlying RemoteTeam and the LocalTeam.
 * <p>
 *
 * @see RemoteTeam
 * @see LocalTeam
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class Team
    extends Importable
{
    /**
     * Does the actual team construction.  All the Team constructors
     * just call this private method.
     * <p>
     * XXX isn't it odd that the first parameter of a static method
     * is just an instance of the class the method is in?
     *
     * @param team the new team object
     * @param name the name for the new team, may be null.
     * @param specs the resource specifications for the new team's
     * resource limits.  May be null.
     * @exception DuplicateExportException if a team with the given name already exists
     * (except null names, duplicates of that are allowed).
     * @exception ResourceSpecException if the given resource requests cannot
     * be satisfied.
     */
    private static void createTeam(final Team team,
				   final String name,
				   final ResourceSpec specs[])
	throws DuplicateExportException, ResourceSpecException
    {
	try
	{
	    RemoteTeam rt = null;
	    ExportTable et;
	    
	    if( RemoteTeam.getKernelTeam().visit() )
	    {
		et = LocalTeam.fetch().getExports();
		/*
		 * Sync on the export table so noone can add/remove any teams.
		 */
		synchronized(et)
		{
		    /* Check for a preexisting team  XXX Redundant? */
		    if( (name == null) ||
			name.equals("") ||
			(et.get(name) == null) )
		    {
			/* Try to make a new one */
			rt = new RemoteTeam(name, specs);
		    }
		    else
		    {
			throw new DuplicateExportException("Team name `"
							   + name
							   + "' is in use");
		    }
		    /*
		     * Set ourselves to import this team.  Since we hold the
		     * lock on the export table the team isn't going to go
		     * away.  However, once we're outside the sync, the team
		     * can be revoked.
		     */
		    if( rt != null )
			rt.subscribe(team);
		}
		RemoteTeam.endVisit();
	    }
	}
	catch(DeadTeamException dte)
	{
	}
    }
    
    /**
     * Create a team with the given resources.
     *
     * @param name The name to give the newly created team.
     * @param specs The resource requirements for the team.
     * @exception Throws DuplicateExportException if there is already a team
     * with the given name.
     * @exception Throws ResourceSpecException if the kernel couldn't satisfy
     * the requested amount of resources.
     */
    public Team(final String name, final ResourceSpec specs[])
	throws DuplicateExportException, ResourceSpecException
    {
	Team.createTeam(this, name, specs);
    }

    /**
     * Same as above, except use the default resource specs
     *
     * @param name The name to give the newly created team.
     * @exception Throws DuplicateExportException if there is already a team
     * with the given name.
     * @exception Throws ResourceSpecException if the kernel couldn't satisfy
     * the requested amount of resources.
     */
    public Team(String name)
	throws DuplicateExportException, ResourceSpecException
    {
	this(name, null);
    }

    /**
     * Create an anonymous team with the given specs
     *
     * @param specs The resource requirements for the team.
     * @exception Throws ResourceSpecException if the kernel couldn't satisfy
     * the requested amount of resources.
     */
    public Team(ResourceSpec specs[])
	throws ResourceSpecException
    {
	try
	{
	    Team.createTeam(this, null, specs);
	}
	catch(DuplicateExportException e)
	{
	    throw new InternalError("Its anonymous, this shouldn't happen...");
	}
    }

    /**
     * Create an anonymous team with the default specs
     *
     * @exception Throws ResourceSpecException if the kernel couldn't satisfy
     * the requested amount of resources.
     */
    public Team()
	throws ResourceSpecException
    {
	try
	{
	    Team.createTeam(this, null, null);
	}
	catch(DuplicateExportException e)
	{
	    throw new InternalError("Its anonymous, this shouldn't happen...");
	}
    }

    /**
     * Create a Team object that imports the given RemoteTeam.
     *
     * @param rt The RemoteTeam to import.
     */
    public Team(RemoteTeam rt)
    {
	rt.subscribe(this);
    }

    /**
     * @param team The team object to copy.
     */
    public Team(Team team)
    {
	this.copy(team);
    }
    
    /**
     * For internal use to just create the team object, we'll set the
     * imported object later on.  See findTeam().
     *
     * @param dummy Used just for overloading purposes.
     * @see #findTeam(String)
     */
    private Team(int dummy)
    {
    }

    /**
     * Get a handle on the team with the given name.  Teams with the null
     * name cannot be found this way.
     *
     * @param name The name of the team to locate.
     * @return Null, or a Team object that imports the RemoteTeam with the given
     * name.
     */
    public Team findTeam(String name)
    {
	Team retval = null;

	if( name != null )
	{
	    Team team = new Team(0);
	    boolean found = false;
	    RemoteTeam kernel;
	    
	    /* Need to visit the kernel and check its export table */
	    kernel = RemoteTeam.getKernelTeam();
	    try
	    {
		if( kernel.visit() )
		{
		    ExportTable et;
		    
		    try
		    {
			et = kernel.getPeer().getExports();
			synchronized(et)
			{
			    RemoteTeam rt;
			    
			    if( (rt = (RemoteTeam)et.get(name)) != null )
			    {
				/*
				 * We have the table lock so the team can't go
				 * away.
				 */
				rt.subscribe(team);
				found = true;
			    }
			}
		    }
		    catch(Throwable th)
		    {
		    }
		    RemoteTeam.endVisit();
		}
	    }
	    catch(Throwable th)
	    {
	    }
	    if( found )
		retval = team;
	}
	return( retval );
    }

    /**
     * Get the name of the RemoteTeam associated with this
     * Team handle.
     *
     * @return A copy of the team's name.
     */
    public synchronized String getName()
    {
	String retval = null;
	RemoteTeam rt;
	
	if( (rt = (RemoteTeam)this.get()) != null )
	{
	    String name;
			
	    if( (name = rt.getName()) != null )
		retval = CloneFactory.clone(name);
	    else
		retval = "(none)";
	}
	return retval;
    }

    /**
     * @return The unique integer identifier for this team.
     */
    public synchronized int getID()
    {
	int retval = 0;
	RemoteTeam rt;

	if( (rt = (RemoteTeam)this.get()) != null )
	{
	    retval = rt.getID();
	}
	return retval;
    }

    /**
     * Test if the given object is owned by this team.
     *
     * @param obj The object to test for ownership.
     * @return True if this team owns the object.
     */
    public boolean owns(Object obj)
    {
	return ((RemoteTeam)this.get()).owns(obj);
    }

    /**
     * Import the exportable mapped to the given id from this team
     * into the current team.  A new ObjectImporter is created and the
     * remote object must be an ObjectExporter instance.
     *
     * @param id The name of the object to import.
     * @param im An Importable that will be set to reference the imported
     * object.
     * @see #importObject(Importable,Object,Class)
     */
    public void importObject(Importable im, Object id)
	throws NoSuchExportException
    {
	if( !this.importObject(im, id, null) )
	    throw new NoSuchExportException(id.toString());
    }

    /**
     * Import a new object from this team into the current team.  A
     * new instance of the given class is created in this team (the
     * remote team) and the given importable is set to point to that
     * new object.  The constructor for the object is run in the
     * context of the new object.  Only a no-arg constructor for the
     * given class can be executed.
     *
     * @param cl The class to create the imported object from.
     * @param im Importable that will be set to reference the imported
     * object.
     * @see #importObject(Importable,Object,Class)
     */
    public void importNewObject(Importable im, Class cl)
	throws NoSuchExportException
    {
	if( !this.importObject(im, null, cl) )
	    throw new NoSuchExportException(cl.getName());
    }

    /**
     * Primary function for importing an object into the current team
     * from this team.
     * <p>
     * This is complicated, but that's why its private.  The simpler
     * interfaces are public.
     * 
     *
     * @param im Importable that will be set to reference the imported
     * object.
     * @param id The name of the object to import.
     * @param cl The class to create the imported object from.
     * @return True if the object was successfully imported, false otherwise.
     */
    private synchronized boolean importObject(Importable im,
					      Object id,
					      Class cl)
    {
	boolean retval = false;
	RemoteTeam rt;

	// XXX if (im == null) throw new... 

	if( (rt = (RemoteTeam)this.get()) != null )
	{
	    try
	    {
		if( rt.visit() )
		{
		    try
		    {
			Exportable ex;
			
			/*
			 * If there is an id, lookup the associated
			 * exportable, otherwise create a new
			 * exportable in the remote team and import that.
			 */
			if( id != null )
			{
			    ExportTable et;
			    
			    et = rt.getPeer().getExports();
			    if( et != null )
			    {
				synchronized(et)
				{
				    if( (ex = et.get(id)) != null )
				    {
					if( (cl == null) ||
					    (ex.getClass() == cl) )
					{
					    if( ex.subscribe(im) )
						retval = true;
					}

					// XXX error if remote is of wrong type?
				    }
				}
			    }
			}
			else
			{
			    // XXX cl must be subclass of Exportable
			    ex = (Exportable)cl.newInstance();
			    if( ex.subscribe(im) )
				retval = true;

			    // XXX add to remote's ExportTable with
			    // the given id?
			}
		    }
		    catch(Throwable th)
		    {
			/* XXX Fix me */
			th.printStackTrace();
		    }
		    RemoteTeam.endVisit();
		}
	    }
	    catch(DeadTeamException dte)
	    {
	    }
	}
	return retval;
    }
    
    /**
     * Terminate this team.  This will clean up all
     * the importables and exportables owned by the
     * team and schedule all of its threads for
     * destruction.
     * <p>
     * If the calling thread is owned by the team being
     * terminated, this function will not return.
     * <p>
     * Active visits to the team being terminated will
     * delay termination.
     *
     * @see RemoteTeam#terminate()
     */
    public synchronized void terminate()
    {
	RemoteTeam rt;

	if( (rt = (RemoteTeam)this.get()) != null )
	{
	    rt.terminate();
	    rt = null;
	}
    }

    /**
     * Create a new activation that reserves the team this object
     * points to.  The activation remains "active" until
     * <code>popActivation()</code> is invoked by the current thread.
     * Unlike a visit(), this call does not implicitly switch context
     * to the other team.  That is done by invoking use on the
     * TeamActivation instance returned by this call.
     * 
     * @return A TeamActivation object that reserves this team.
     * @see #visit()
     * @see RemoteTeam#pushActivation(int)
     * @see TeamActivation#use()
     * @see TeamActivation#end()
     */
    public synchronized TeamActivation pushActivation()
    {
	RemoteTeam rt = (RemoteTeam)this.get();
	TeamActivation retval = null;

	if( rt != null )
	    retval = rt.pushActivation(1);
	return retval;
    }

    /**
     * Start a visit to the team this object points to.  While the
     * visit is active, neither team can be terminated, so object
     * references in either team can be manipulated.  Write-barriers
     * are also disabled during this sequence.  The visit ends
     * when <code>endVisit()</code> is invoked.
     *
     * @see #endVisit()
     * @see RemoteTeam#visit(int)
     */
    public void visit()
	throws DeadTeamException
    {
	RemoteTeam rt = (RemoteTeam)this.get();

	if( rt == null )
	    throw new DeadTeamException("you are dead in visit()");
	if( !rt.visit(1) )
	    throw new DeadTeamException("other team is dead in visit()");
    }

    /**
     * End a visit to this team.
     *
     * @see #visit()
     * @see RemoteTeam#endVisit()
     */
    public static void endVisit()
    {
	RemoteTeam.endVisit();
    }
    
    /**
     * Pop the last TeamActivation off the stack.
     */
    public static void popActivation()
    {
	RemoteTeam.popActivation();
    }
}
