/*
 * TeamActivation.java
 * Manages team reservation/migration.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * @JANOSVM_JVMLIB_LICENSE@
 */

package edu.utah.janosvm.sys;

/**
 * Record a connection to a remote team.  A TeamActivation is a
 * link between two teams that prevents either from being terminated.
 * A TeamActivation is more flexible than a visit, as the
 * termination-delaying time can be expanded to include more than the
 * context switch to the other team.  Here's an example:
 * <p>
 * TeamActivation objects are a bit odd since they are never explicitly
 * created.  They are only implicitly created by the JanosVM due to
 * a call to <code>team.pushActivation()</code>.
 * 
 * <pre>
 * {
 *      TeamActivation act;
 *      ...
 *      act = otherTeam.pushActivation();
 *      if (act != null)
 *	{
 *		Object obj;
 *
 *		act.use();
 *		// Do things in the context of 'otherTeam'
 *		obj = lookupSimpleFoo();
 *		act.end();
 *
 *		// Do things in original context, but guarantee that
 *		// otherTeam will not go away ... activation still active. 
 *  
 *              Object localObj = new Foo(obj);
 *
 *		// End the activation, as we're done with all pointers into otherTeam
 *		Team.popActivation();
 *      }
 * }
 * </pre>
 *
 *
 * @see Team#pushActivation()
 * @see Team#popActivation()
 * @see Team#visit()
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class TeamActivation
{
    /**
     * Hide the constructor.  No one can allocate a TeamActivation
     * the JVM manufactures them in special memory in response to
     * a <code>Team.pushActivation()</code>.
     *
     * @see Team#pushActivation()
     */
    private TeamActivation()
    {
    }

    /**
     * Switch the current thread to the team encoded in this activation.
     */
    public native void use() throws DeadTeamException;

    /**
     * Return to the previous team.
     */
    public static native void end();

    /**
     * Pop the last TeamActivation off the stack.
     */
    public static void pop()
    {
	RemoteTeam.popActivation();
    }
}
