/*
 * Copyright (c) 2000, 2001, 2002 The University of Utah. All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Contributed by the Flux Research Group at the University of Utah.
 * Author: Patrick Tullmann
 */

#define NODEOS_IN_MD_C

#include <nodeos/an_moab_setup.h>
#include <nodeos/an_domain.h>
#include <nodeos/an_cred.h>
#include <nodeos/an_error.h>
#include <nodeos/an_debug.h>

#include "config.h"
#include "gtypes.h"
#include "external.h"
#include "systems/nodeos-threads/thread-impl.h"
#include "md.h"

/*
 * XXX the oskit is a serious pile of crap.
 */
struct oskit_stream;
extern struct oskit_stream *default_console_stream;

/*
 * The classpath defined in this glue code.  This lets
 * kaffe_nodeos read a classpath out of a file.
 *
 * Note, only used if the standard classpath environment
 * variables are empty.  See kaffe/kaffe/main.c
 */
char* kaffe_nodeos_classpath;


/* XXX shouldn't these be static? */
an_mempool_t rootMempool;
an_domain_t rootDomain;
an_cred_t rootCred;
/* an_cred_t eeCred; */

static struct nodeos_initarg 
{
	int argc;
	char **argv;
} initarg;

static void nodeos_bootstrap(void *arg);
static void nodeos_start_root_domain(void (*callback)(void *), void *callbackArg);

/* XXXX --- config file interface */

typedef struct {
	FILE* fp;
} ConfigFile;

int
config_file_open(ConfigFile *cf, const char *filename)
{
	cf->fp = fopen(filename, "r");
	return (cf->fp == NULL);
}

int
config_file_getline(ConfigFile *cf, char *buf, int bufsize)
{
	char *bufchk;
	bufchk = fgets(buf, bufsize, cf->fp);
	return (bufchk != buf);
}

int
config_file_close(ConfigFile *cf)
{
	return fclose(cf->fp);
}

/* XXXX --- config file interface */
	
void
kaffe_nodeos_setup_libpath(void)
{
	/*
	 * We need to be able to find .la files.  Provide a default
	 * LIBRARYPATH.  LIBRARYPATH is Kaffe #define that is the name
	 * of the library search path environment variable.
	 *
	 * The nodeos mkimage.sh stuff puts all the libraries in the
	 * the /lib directory in the bmod.
	 */
	setenv(LIBRARYPATH, "/lib", 0);
}

void
kaffe_nodeos_setup_classpath(void)
{
#define BUFSIZE 2048
	char *cpName;
	char buf[BUFSIZE];
	ConfigFile cpFile;
	int rc;

	/*
	 * If the env var CLASSPATH is defined, just let that
	 * get used by the kaffe's main.
	 *
	 * Otherwise, try and find something to stuff into
	 * kaffe_nodoes_classpath as a first-fallback.
	 */
	if (getenv("CLASSPATH") != NULL)
	{
		kaffe_nodeos_classpath = NULL;
		return;
	}

	/*
	 * Typing a full CLASSPATH on a command line is a bit too much.
	 * Instead, read it out of /etc/kaffe_classpath (or, whatever
	 * CLASSPATHFILE names).
	 */
        if ((cpName = getenv("CLASSPATHFILE")) == NULL)
		cpName = "/etc/kaffe_classpath";
	
	rc = config_file_open(&cpFile, cpName);
	if (!rc)
	{
		int c;

		config_file_getline(&cpFile, buf, sizeof(buf));
		config_file_close(&cpFile);

		c = strlen(buf);
		if (buf[c - 1] == '\n')
			buf[c - 1] = NULL;

		if ((kaffe_nodeos_classpath = (char *) malloc(c + 1)) == NULL)
			panic(__FUNCTION__ ": No memory for kaffe_nodeos_classpath.\n");

		strcpy(kaffe_nodeos_classpath, buf);
	}
	else
	{
		osenv_log(OSENV_LOG_CRIT,
			  "JanosVM has no CLASSPATH to search for classes in.\n");
	}

}

/*
 * The real entrypoint for NodeOS-based version of Kaffe.  We
 * "override" the Kaffe main through CPP magic, defining main to be
 * 'originalKaffeMain' in md.h (except when md.h is included in this
 * file, of course).
 *
 * originalKaffeMain is invoked by nodeos_bootstrap() (which is
 * run in the context of the root domain).
 */
int /* invokedonce */
main(int argc, char** argv)
{
	if (default_console_stream == NULL)
	{
		/* Nothing it initialized.  So DON'T CALL assert() or printf()! */
		console_putchar('C');
		console_putchar('R');
		console_putchar('A');
		console_putchar('P');
		console_putchar('!');
		exit(111);
	}

#ifdef DEBUG
	{
		const char* err;
		
		/* override the default for Moab debugging */
		err = an_debug_setmask("OSKIT_LOG_ERR");

		if (err != NULL)
		{
			/* Again, not much is initialized at this point... */
			console_putchar('O');
			console_putchar('o');
			console_putchar('p');
			console_putchar('s');
			console_putchar('.');
			exit(111);
		}
	}
#endif
	an_moab_setup();

	initarg.argc = argc;
	initarg.argv = argv;

	kaffe_nodeos_setup_libpath();

	/* Does not return unless the root domain shutsdown. */
	nodeos_start_root_domain(nodeos_bootstrap, &initarg);

	return 0;
}

static void /* invokedonce */
nodeos_bootstrap(void *arg)
{
	struct nodeos_initarg *initarg = (struct nodeos_initarg*)arg;
	
	kaffe_nodeos_setup_classpath();

	/* pre-init the root thread */
	anjthread_preinitroot(jthread_current());

	originalKaffeMain(initarg->argc, initarg->argv);
}


/*
 * Invoked by INIT_MD() macro in Kaffe's main() (i.e., by originalKaffeMain()).
 */
void /* invokedonce */
nodeos_main_md(void)
{

}

int
nodeos_option_parsefunc(char **argv, int *iptr)
{
	int i;
	int rc = 0;
	
	assert(argv);
	assert(iptr);
	assert(argv[*iptr]);
	
	i = *iptr;

#if defined(DEBUG)
	if (strcmp("-moabdebug", argv[i]) == 0) {
		i++;
		if (argv[i] == NULL) {
			fprintf(stderr, "-moabdebug requires an argument.\n");
			exit(1);
		}
		
		/* Exit if returns non-zero, e.g., with 'list' option */
		if (an_debug_setmask(argv[i]))
		{
			fprintf(stderr, "moab debug option message: %s\n",
				an_error_message());
			exit(0);
		}

		rc = 1;
	}
#endif	
	
	*iptr = i;
	
	return rc;
}


struct _jthread *an_jthreadlist[ROOT_DOMAIN_THREAD_CT];

/*
 * Start the single root thread on its way.
 * This function does not return unless the root domain has shut down
 */
static void /* invokedonce */
nodeos_start_root_domain(void (*callback)(void *), void *callbackArg)
{
#define ROOT_DOMAIN_THREAD_STACKSIZE (32 * 1024) // XXX a bit big
	static void* threadlist[ROOT_DOMAIN_THREAD_CT];
	static void* stacklist[ROOT_DOMAIN_THREAD_CT];
	int i;

	for (i = 0; i < ROOT_DOMAIN_THREAD_CT; i++)
	{
		struct _jthread *jth;

		/* The jthread code assumes its allocator zeros memory. */
		an_jthreadlist[i] = jth = malloc(sizeof(struct _jthread));
		memset(jth, 0, sizeof(struct _jthread)); 

		threadlist[i] = &(jth->native_thread);
		stacklist[i] = malloc(ROOT_DOMAIN_THREAD_STACKSIZE);
		jth->stackptr = stacklist[i];
	}

	/*
	 * NOTE: we don't have to register the single root thread with
	 * the GC or the VM, it will do a jthread_createfirst() call
	 * which will register the root thread with the threading
	 * system and the GC.  Additionally, the root thread exits
	 * almost immediately after starting the first EE.
	 *
	 * Also, the current "thread" never runs in a NodeOS context
	 * it just blocks in an_moab_start() until the root
	 * domain dies, and then it returns from an_moab_start().
	 */

	an_moab_start(&rootDomain,
		      &rootCred,
		      &rootMempool,
		      ROOT_DOMAIN_THREAD_CT, ROOT_DOMAIN_THREAD_STACKSIZE,
		      threadlist, stacklist,
		      callback, callbackArg);

	/*
	 * Only returns when the root domain has died.
	 */
}


