#!/bin/sh
#
# Copyright (c) 1998, 1999, 2000, 2001 The University of Utah. All rights reserved.
#
# This file is licensed under the terms of the GNU Public License.  
# See the file "license.terms" for restrictions on redistribution 
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# Contributed by the Flux Research Group at the University of Utah.
# Authors: Leigh Stoller, Patrick Tullmann
#

# Build a multiboot image from Kaffe plus other stuff. You can either
# copy this script to your object directory and edit the paths below,
# or you can pass it all in as command line args.
#
# By default, you alway get the contents of the Kaffe
# install/share/kaffe directory (which includes the standard kaffe
# classes) and install/lib/kaffe/*.la (the libtool archive description
# files).  The default classes can be skipped by providing the 
# --skipdefaultclasses argument to this script.
#
# 
# LIBTOOL NOTES:
# 
# When using preloaded (static) libraries, libtool still needs to
# examine the .la files (but not the real archives) at runtime,
# KAFFELIBRARYPATH should contain the directories where all the native
# .la library descriptions can be found.
# 
# KAFFELIBRARYPATH's default value in the OSKit is /lib.  This script
# loads all the .la files into that /lib in the boot image.
# 

# Your Kaffe install directory
KAFFEDIR=/opt/kaffe/install

# where your oskit was installed
OSKITDIR=/opt/oskit/install

# Default classpath file (see the README for a description of this file)
CLASSPATHFILE=/tmp/default_kaffe_classpath

# Default boot image construction program.  We use Multiboot images.
#  NOTE that mkmb2 is broken in 990722.
MAKEIMAGE=

# default directory to put the new kernel image into
KERNELDIR=.

## ---

# Should include default classes ($KAFFEDIr/share/kaffe/*.jar) 
# in bmod and default classpath?
INCLUDEDEFAULTCLASSES=yes

# List of class archives to include in bmod and in generated classpath
CLASSARCHIVES=

# If non-null will be passed as kernel command line
KERNELCOMMANDLINE=

# Make mkimage.sh quiet by making this variable non-nil.
QUIET=

# A list of directories
DIRS=

# A list of unrooted directories
URDIRS=

# A list of mapped directories
MAPPEDDIRS=

# A list of explicit file:file pairs to put in the image
EXPLICITFILES=

# Control whether the default_kaffe_classpath is generated by this script or not
CREATE_CLASSPATHFILE=yes

# Parse the command-line options to override above.
until [ $# -eq 0 ]
do
	case "$1" in
		--kerneldir=* )
			KERNELDIR=`echo "$1" | sed 's/[-_a-zA-Z0-9]*=//'` ;
			shift;;
		--kaffedir=* )
			KAFFEDIR=`echo "$1" | sed 's/[-_a-zA-Z0-9]*=//'` ;
			shift;;
		--oskitdir=* )
			OSKITDIR=`echo "$1" | sed 's/[-_a-zA-Z0-9]*=//'` ;
			shift;;
		--makeimage=* )
			MAKEIMAGE=`echo "$1" | sed 's/[-_a-zA-Z0-9]*=//'` ;
			shift;;
		--skipdefaultclasses)
			INCLUDEDEFAULTCLASSES=no
			shift;;
		--classpathfile=* )
			CLASSPATHFILE=`echo "$1" | sed 's/[-_a-zA-Z0-9]*=//'`;
			CREATE_CLASSPATHFILE=no
			shift;;
		--classarchive=* )
			CLASSARCHIVES="$CLASSARCHIVES `echo "$1" | sed 's/[-_a-zA-Z0-9]*=//'`";
			shift;;
		--commandline=* )
			## XXX only works for *arguments to main not env vars*
			KERNELCOMMANDLINE=`echo "$1" | sed 's/[-a-z]*=//'`;
			KERNELCOMMANDLINE="-c \"${KERNELCOMMANDLINE}\""
			shift;;
		--explicitfile=* )
			EXPLICITFILES="$EXPLICITFILES `echo "$1" | sed 's/[-_a-zA-Z0-9]*=//'`" ;
			shift;;
		--help )
			echo "Create a boot image for Kaffe/OSKit"
			echo "Options (and their defaults):"
			echo "  --kerneldir=<dir>       The directory in which to drop the created Image ($IMAGEDIR)"
			echo "  --kaffedir=<dir>        The directory in which kaffe was installed ($KAFFEDIR)"
			echo "  --oskitdir=<dir>        The directory in which the OSKit was installed ($OSKITDIR)"
			echo "  --makeimage=prog        The program to use to create the image ($MAKEIMAGE)"
			echo "  --classarchive=<file>   A file to add to image and add to auto-gen classpath file"
			echo "  --classpathfile=<file>  The file containing the classpath for Kaffe ($CLASSPATHFILE)"
			echo "         (Note that $CLASSPATHFILE will be generated for you if not specified.)"
			echo "  --explicitfile=<file>[:<imagename>]   Add <file> to the boot image.  Make <file> "
			echo "         visibile in the kernel filesystem as <imagename>.  Repeat as needed."
			echo "  --skipdefaultclasses    Do NOT include the default archives in share/kaffe/."
			echo "  -dir <dir>              Add all the files in <dir> to the image.  No name mapping."
			echo "  -unrooteddir <dir>      Add all the files in <dir> to the *root* of the image."
			echo "  --mappeddir <local-dir>:<image-dir>   Add the directory tree rooted at local-dir into the boot image."
			exit 7;;
		-dir )
			DIRS="$DIRS $2" ;
			shift;
			shift;;
		-unrooteddir )
			URDIRS="$URDIRS $2" ;
			shift;
			shift;;
		--mappeddir )
			MAPPEDDIRS="$MAPPEDDIRS $2" ;
			shift;
			shift;;
		* ) echo "Bad arg: $1";
			  exit 8 ;;
	esac
done

if test ! -d $KAFFEDIR; then
    echo "$KAFFEDIR is not a directory.  Specify with --kaffedir=<dir>"
    exit 9
fi
    
if test ! -d $KERNELDIR; then
    echo "ERROR: $KERNELDIR is not a valid directory.  Specify one with --kerneldir=<dir>"
    exit 10
fi


CPF="$CLASSPATHFILE:/etc/kaffe_classpath"

# The Kaffe kernel
KAFFE=$KAFFEDIR/libexec/Kaffe

if test ! -x $KAFFE; then
    echo "ERROR: $KAFFE is not an executable."
    exit 11
fi

# The directory with the minimum necessary class files.
DEFAULTCLASSESDIR=$KAFFEDIR/share/kaffe

# Find the image maker, if none was provided
if test -z "${MAKIMAGE}"; then
    if test -x "${OSKITDIR}/bin/mkmb2"; then
	MAKEIMAGE="${OSKITDIR}/bin/mkmb2";
    elif test -x "${OSKITDIR}/bin/mkmbimage"; then
	MAKEIMAGE="${OSKITDIR}/bin/mkmb2";
    fi
else
    MAKEIMAGE="${OSKITDIR}/bin/${MAKEIMAGE}";
fi

if test ! -x "${MAKEIMAGE}"; then
    echo "ERROR: Image generator script '${MAKEIMAGE}' is not executable"
    exit 11;
fi


### Create the classpath file if just using the default
if test "$CREATE_CLASSPATHFILE" = "yes"; then
    rm -f ${CLASSPATHFILE}
    touch ${CLASSPATHFILE}
    if test "$INCLUDEDEFAULTCLASSES" = "yes"; then
	for i in $DEFAULTCLASSESDIR/*.jar; do
	    echo -n "${i}:" >> ${CLASSPATHFILE};
	done
    fi
    if test -n "$CLASSARCHIVES"; then
	for i in $CLASSARCHIVES; do
	    echo -n "${i}:" >> ${CLASSPATHFILE};
	done
    fi
    echo "." >> ${CLASSPATHFILE};
else 
    if test ! -f $CLASSPATHFILE; then
	echo "$CLASSPATHFILE doesn't exist.  Please create or point elsewhere with --classpathfile=<file>"
	exit 12
    fi
fi

### Print out some 
if test -z "$QUIET"; then
    if test "$INCLUDEDEFAULTCLASSES" = "yes"; then
	echo "  Sucking (full-path) default class archives from: $DEFAULTCLASSESDIR"
    fi
    echo "  Sucking (full-path) all files in: $DIRS"
    echo "  Sucking (rel. path) all files in: $URDIRS"
    echo "  Plus all the *.la files in $KAFFEDIR/lib/kaffe"
    test ! -z "$EXPLICITFILES" && echo "  Plus these files: $EXPLICITFILES"
    echo "  Plus these class archives: $CLASSARCHIVES"
    echo "  Plus the classpath: $CPF"
fi

rm -f ${KERNELDIR}/Image

# Add DEFAULTCLASSESDIR to DIRS if they should be included
# in the image.
if test "$INCLUDEDEFAULTCLASSES" = "yes"; then
    DIRS="$DEFAULTCLASSESDIR $DIRS"
fi

# Add all the files in the DIRS and all of the .la files.
{	
	for DIR in $DIRS; do
		for FILE in `find -L $DIR -type f -o -type l`
		do
		    echo "$FILE:$FILE"
		done
	done
	test -n "$CLASSARCHIVES" && for FILE in $CLASSARCHIVES; do
		if test ! -f "$FILE"; then
		    echo "ERROR: $FILE does not exist!" >&2
		    exit 1;
		fi
		echo "$FILE"
	done
	for DIR in $URDIRS; do
		for FILE in `find -L $DIR -type f -o -type l`
		do
		    echo "$FILE:"`echo $FILE | sed -e "s,^$DIR,/,"`
		done
	done
	for DIR in $MAPPEDDIRS; do
		SRCDIR=`echo "$DIR" | sed 's/:[\/-_a-zA-Z0-9]*//'`;
		DSTDIR=`echo "$DIR" | sed 's/[\/-_a-zA-Z0-9]*://'`;
		echo "Mapping: $SRCDIR to $DSTDIR" >&2;
		for FILE in `find -L $SRCDIR \( -type f -o -type l \) \! -path "*CVS*" \! -path "*~"`
		do
		    echo "$FILE:"`echo $FILE | sed -e "s,^$SRCDIR,$DSTDIR,"`
		    echo "Adding:  $FILE:"`echo $FILE | sed -e "s,^$SRCDIR,$DSTDIR,"` >&2;
		done
	done

	for FILE in $KAFFEDIR/lib/kaffe/*.la; do
		echo "$FILE:/lib/"$(basename $FILE)
	done
	for FILE in $KAFFEDIR/lib/*.la;	do
		echo "$FILE:/lib/"$(basename $FILE)
	done
} | eval $MAKEIMAGE -o $KERNELDIR/Image $KERNELCOMMANDLINE -stdin \
    $KAFFE $CPF $EXPLICITFILES \
    && echo Created $KERNELDIR/Image

exit 0

#eof
