#
# inliner.awk
# Generates a .c file for inlined functions
#
# Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this file
# for any purpose with or without restriction is hereby granted.
#

#
# usage: awk -f inliner.awk <name of generated file> <non-inline .c file>
#
# example: awk -f inliner.awk foo-inline.c foo.c > foo-inline.c
#
# The first argument should be the name of the file where the output is
# stored.  The second argument is the source file for any non-inlined
# functions, we use this to grab any necessary `#include's and add them
# to the generated file.
#

BEGIN {

#
#       ifile - The name of the file we're generating
#       srcfile - The name of the file whose `#include's we're gonna steal.
#                 Should be fully qualified.
#       basename - The base names for the files (i.e. no .h, .c, etc...)
#       ihdrfile - The name of the header file with the inline functions
#       foundit - Indicates whether or not the srcfile includes its own inline
#                 header file.
#
	if( ARGC < 3 )
	{
		printf("Usage: awk -f inliner.awk <name of generated file> <non-inline .c file>.\n") > "/dev/stderr";
		exit -1;
	}
	ifile = ARGV[1];
	srcfile = ARGV[2];
	basename = gensub(/\-inline\.c/, "", "g", ifile);
	ihdrfile = sprintf("%s-inline.h", basename);
	foundit = 0;
	
	ARGV[1] = "";

#
#       Print out a header
#
	
	printf("/*\n");
	printf(" * %s\n", ifile);
	printf(" *\n");
	printf(" * Holds definitions for inlined functions in file %s\n",
	       ihdrfile);
	printf(" *\n");
	printf(" * Automatically generated by developers/inliner.awk\n");
	printf(" * on %s\n", strftime());
	printf(" */\n");
	printf("\n");
}

# Match all #include statements and just add them, if we hit the inline
# file add `#undef KAFFE_INLINE'.
/[[:blank:]]*\#include/ {
	if( index($2, ihdrfile) )
	{
		printf("#undef KAFFE_INLINE\n");
		printf("#define KAFFE_INLINE\n");
		foundit = 1;
	}
	printf("%s\n", $0);
}

# If the source file didn't include its inline header we'll need to do it here
END {
	if( ARGC < 3 )
	{
		exit -1;
	}
	if( !foundit )
	{
		printf("#undef KAFFE_INLINE\n");
		printf("#define KAFFE_INLINE\n");
		printf("#include \"%s\"\n", ihdrfile);
	}
	printf("\n");
	printf("/* END OF FILE */\n");
}
