;;; jvmconfig.el --- major mode for editing jvmconfig source in Emacs

;; Copyright (C) 1993, 1994, 1995, 1996, 1997, 2000 Free Software Foundation, Inc.

;; Author: Espen Skoglund <espensk@stud.cs.uit.no>
;; Keywords: languages

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; USAGE
;; =====

;; Emacs should enter Jvmconfig mode when you find a Jvmconfig source file.
;; When you have entered Jvmconfig mode, you may get more info by pressing
;; C-h m. You may also get online help describing various functions by:
;; C-h f <Name of function you want described>

;; If you want to customize Jvmconfig mode to fit you better, you may add
;; these lines (the values of the variables presented here are the defaults):
;;
;; ;; User customization for Jvmconfig mode
 (setq jvmconfig-indent-level       2
       jvmconfig-auto-newline       nil
       jvmconfig-tab-always-indent  t
       jvmconfig-auto-endcomments   t
       jvmconfig-auto-lineup        '(all)
       jvmconfig-toggle-completions nil
       jvmconfig-start-keywords     '("begin" "end" "include" "text")
       jvmconfig-separator-keywords '())

;; KNOWN BUGS / BUGREPORTS
;; =======================
;; As far as I know, there are no bugs in the current version of this
;; package.  This may not be true however, since I never use this mode
;; myself and therefore would never notice them anyway.   If you do
;; find any bugs, you may submit them to: espensk@stud.cs.uit.no
;; as well as to bug-gnu-emacs@prep.ai.mit.edu.

;;; Code:

(defgroup jvmconfig nil
  "Major mode for editing Jvmconfig source in Emacs"
  :group 'languages)

(defvar jvmconfig-mode-abbrev-table nil
  "Abbrev table in use in Jvmconfig-mode buffers.")
(define-abbrev-table 'jvmconfig-mode-abbrev-table ())

(defvar jvmconfig-mode-map ()
  "Keymap used in Jvmconfig mode.")
(if jvmconfig-mode-map
    ()
  (setq jvmconfig-mode-map (make-sparse-keymap))
  (define-key jvmconfig-mode-map ";"        'electric-jvmconfig-semi-or-dot)
  (define-key jvmconfig-mode-map "."        'electric-jvmconfig-semi-or-dot)
  (define-key jvmconfig-mode-map ":"        'electric-jvmconfig-colon)
  (define-key jvmconfig-mode-map "="        'electric-jvmconfig-equal)
  ;(define-key jvmconfig-mode-map "#"        'electric-jvmconfig-hash)
  (define-key jvmconfig-mode-map "\r"       'electric-jvmconfig-terminate-line)
  (define-key jvmconfig-mode-map "\t"       'electric-jvmconfig-tab)
  (define-key jvmconfig-mode-map "\M-\t"    'jvmconfig-complete-word)
  (define-key jvmconfig-mode-map "\M-?"     'jvmconfig-show-completions)
  (define-key jvmconfig-mode-map "\177"     'backward-delete-char-untabify)
  (define-key jvmconfig-mode-map "\M-\C-h"  'jvmconfig-mark-defun)
  (define-key jvmconfig-mode-map "\C-c\C-b" 'jvmconfig-insert-block)
  (define-key jvmconfig-mode-map "\M-*"     'jvmconfig-star-comment)
  (define-key jvmconfig-mode-map "\C-c\C-c" 'jvmconfig-comment-area)
  (define-key jvmconfig-mode-map "\C-c\C-u" 'jvmconfig-uncomment-area)
  (define-key jvmconfig-mode-map "\M-\C-a"  'jvmconfig-beg-of-defun)
  (define-key jvmconfig-mode-map "\M-\C-e"  'jvmconfig-end-of-defun)
  (define-key jvmconfig-mode-map "\C-c\C-d" 'jvmconfig-goto-defun)
  (define-key jvmconfig-mode-map "\C-c\C-o" 'jvmconfig-outline)
;;; A command to change the whole buffer won't be used terribly
;;; often, so no need for a key binding.
;  (define-key jvmconfig-mode-map "\C-cd"    'jvmconfig-downcase-keywords)
;  (define-key jvmconfig-mode-map "\C-cu"    'jvmconfig-upcase-keywords)
;  (define-key jvmconfig-mode-map "\C-cc"    'jvmconfig-capitalize-keywords)
  )

(defvar jvmconfig-imenu-generic-expression
  '("^[ \t]*\\(function\\|procedure\\|begin\\|include\\|text\\)[ \t\n]+\\([a-zA-Z0-9_.:]+\\)" . (2))
  "Imenu expression for Jvmconfig-mode.  See `imenu-generic-expression'.")

(defvar jvmconfig-keywords
  '("and" "array" "begin" "include" "text" "case" "const" "div" "do" "downto" "else" "end" 
    "file" "for" "function" "if" "in" "label" "mod" "nil" "not" "of" 
    "or" "packed" "procedure" "program" "record" "repeat" "set" "then" "to" 
    "type" "until" "var" "while" "with"
    ;; The following are not standard in jvmconfig, but widely used.
    "get" "put" "input" "output" "read" "readln" "reset" "rewrite" "write"
    "writeln"))

;;;
;;; Regular expressions used to calculate indent, etc.
;;;
(defconst jvmconfig-symbol-re      "\\<[a-zA-Z_][a-zA-Z_0-9.]*\\>")
(defconst jvmconfig-beg-block-re   "\\<\\(script\\|begin\\|text\\)\\>")
(defconst jvmconfig-end-block-re   "\\<\\(end\\)\\>")
(defconst jvmconfig-declaration-re "\\<\\(\\)\\>")
(defconst jvmconfig-defun-re       "\\<\\(\\)\\>")
(defconst jvmconfig-sub-block-re   "\\<\\(\\)\\>")
(defconst jvmconfig-noindent-re    "\\<\\(begin\\|script\\|text\\|end\\)\\>")
(defconst jvmconfig-nosemi-re      "\\<\\(begin\\|\n\\|script\\|text\\)\\>")
(defconst jvmconfig-autoindent-lines-re
  "\\<\\(end\\|begin\\|script\\|text\\)\\>")

;;; Strings used to mark beginning and end of excluded text
(defconst jvmconfig-exclude-str-start "{-----\\/----- EXCLUDED -----\\/-----")
(defconst jvmconfig-exclude-str-end " -----/\\----- EXCLUDED -----/\\-----}")

(defvar jvmconfig-mode-syntax-table nil
  "Syntax table in use in Jvmconfig-mode buffers.")

(if jvmconfig-mode-syntax-table
    ()
  (setq jvmconfig-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?\\ "."   jvmconfig-mode-syntax-table)
  ;(modify-syntax-entry ?( "()1"  jvmconfig-mode-syntax-table)  
  ;(modify-syntax-entry ?) ")(4"  jvmconfig-mode-syntax-table)
  ;(modify-syntax-entry ?* ". 23" jvmconfig-mode-syntax-table)
  ;(modify-syntax-entry ?{ "<"    jvmconfig-mode-syntax-table)
  ;(modify-syntax-entry ?} ">"    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?+ "."    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?- "."    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?= "."    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?% "."    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?# "<"    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?\n ">"    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?< "."    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?> "."    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?& "."    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?| "."    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?_ "_"    jvmconfig-mode-syntax-table)
  (modify-syntax-entry ?\' "\""  jvmconfig-mode-syntax-table))

(defconst jvmconfig-font-lock-keywords (purecopy
  (list
   '("^[ \t%]*\\(begin\\|include\\|text\\|function\\|pro\\(cedure\\|gram\\)\\)\\>[ \t]*\\(.*\\)"
     1 font-lock-keyword-face)
   '("^[ \t%]*\\(begin\\|include\\|text\\|function\\|pro\\(cedure\\|gram\\)\\)\\>[ \t]*\\(.*\\)"
     3 font-lock-function-name-face t)
;   ("type" "const" "real" "integer" "char" "boolean" "var"
;    "record" "array" "file")
   (cons (concat "\\<\\(array\\|boolean\\|c\\(har\\|onst\\)\\|file\\|"
		 "integer\\|re\\(al\\|cord\\)\\|type\\|var\\)\\>")
	 'font-lock-type-face)
   '("\\<\\(label\\|external\\|forward\\)\\>" . font-lock-reference-face)
   '("\\<\\([0-9]+\\)[ \t]*:" 1 font-lock-function-name-face)
;   ("of" "to" "for" "if" "then" "else" "case" "while"
;    "do" "until" "and" "or" "not" "in" "with" "repeat" "begin" "end" "script" "text")
   (concat "\\<\\("
	   "and\\|begin\\|script\\|case\\|do\\|e\\(lse\\|nd\\)\\|for\\|i[fn]\\|"
	   "not\\|o[fr]\\|repeat\\|t\\(hen\\|o\\|ext\\)\\|until\\|w\\(hile\\|ith\\)"
	   "\\)\\>")))
  "Additional expressions to highlight in Jvmconfig mode.")
(put 'jvmconfig-mode 'font-lock-defaults '(jvmconfig-font-lock-keywords nil t))

(defcustom jvmconfig-indent-level 3
  "*Indentation of Jvmconfig statements with respect to containing block."
  :type 'integer
  :group 'jvmconfig)

(defcustom jvmconfig-case-indent 2
  "*Indentation for case statements."
  :type 'integer
  :group 'jvmconfig)

(defcustom jvmconfig-auto-newline nil
  "*Non-nil means automatically insert newlines in certain cases.
These include after semicolons and after the punctuation mark after an `end'."
  :type 'boolean
  :group 'jvmconfig)

(defcustom jvmconfig-tab-always-indent t
  "*Non-nil means TAB in Jvmconfig mode should always reindent the current line.
If this is nil, TAB inserts a tab if it is at the end of the line
and follows non-whitespace text."
  :type 'boolean
  :group 'jvmconfig)

(defcustom jvmconfig-auto-endcomments t
  "*Non-nil means automatically insert comments after certain `end's.
Specifically, this is done after the ends of cases statements and functions.
The name of the function or case is included between the braces."
  :type 'boolean
  :group 'jvmconfig)

(defcustom jvmconfig-auto-lineup '(all)
  "*List of contexts where auto lineup of :'s or ='s should be done.
Elements can be of type: 'paramlist', 'declaration' or 'case', which will
do auto lineup in parameterlist, declarations or case-statements
respectively. The word 'all' will do all lineups. '(case paramlist) for
instance will do lineup in case-statements and parameterlist, while '(all)
will do all lineups."
  :type '(repeat (choice (const all)
			 (const paramlist)
			 (const declaration)
			 (const case)))
  :group 'jvmconfig)

(defcustom jvmconfig-toggle-completions nil
  "*Non-nil means \\<jvmconfig-mode-map>\\[jvmconfig-complete-word] should try all possible completions one by one.
Repeated use of \\[jvmconfig-complete-word] will show you all of them.
Normally, when there is more than one possible completion,
it displays a list of all possible completions."
  :type 'boolean
  :group 'jvmconfig)

(defcustom jvmconfig-type-keywords
  '("array" "file" "packed" "char" "integer" "real" "string" "record")
  "*Keywords for types used when completing a word in a declaration or parmlist.
These include integer, real, char, etc.
The types defined within the Jvmconfig program
are handled in another way, and should not be added to this list."
  :type '(repeat (string :tag "Keyword"))
  :group 'jvmconfig)

(defcustom jvmconfig-start-keywords
  '("begin" "script" "end" "text" "function" "procedure" "repeat" "until" "while"
    "read" "readln" "reset" "rewrite" "write" "writeln")
  "*Keywords to complete when standing at the first word of a statement.
These are keywords such as begin, repeat, until, readln.
The procedures and variables defined within the Jvmconfig program
are handled in another way,  and should not be added to this list."
  :type '(repeat (string :tag "Keyword"))
  :group 'jvmconfig)

(defcustom jvmconfig-separator-keywords
  '("downto" "else" "mod" "div" "then" "\n")
  "*Keywords to complete when NOT standing at the first word of a statement.
These are keywords such as downto, else, mod, then.
Variables and function names defined within the Jvmconfig program
are handled in another way, and should not be added to this list."
  :type '(repeat (string :tag "Keyword"))
  :group 'jvmconfig)


;;;
;;;  Macros
;;;

(defsubst jvmconfig-get-beg-of-line (&optional arg)
  (save-excursion
    (beginning-of-line arg)
    (point)))

(defsubst jvmconfig-get-end-of-line (&optional arg)
  (save-excursion
    (end-of-line arg)
    (point)))

(defun jvmconfig-declaration-end ()
  (let ((nest 1))
    (while (and (> nest 0)
		(re-search-forward 
		 "[:=]\\|\\(\\<record\\>\\)\\|\\(\\<end\\>\\)" 
		 (save-excursion (end-of-line 2) (point)) t))
      (cond ((match-beginning 1) (setq nest (1+ nest)))
	    ((match-beginning 2) (setq nest (1- nest)))
	    ((looking-at "[^(\n]+)") (setq nest 0))))))


(defun jvmconfig-declaration-beg ()
  (let ((nest 1))
    (while (and (> nest 0)
		(re-search-backward "[:=]\\|\\<\\(type\\|var\\|label\\|const\\)\\>\\|\\(\\<record\\>\\)\\|\\(\\<end\\>\\)" (jvmconfig-get-beg-of-line 0) t))
      (cond ((match-beginning 1) (setq nest 0))
	    ((match-beginning 2) (setq nest (1- nest)))
	    ((match-beginning 3) (setq nest (1+ nest)))))
    (= nest 0)))

  
(defsubst jvmconfig-within-string ()
  (save-excursion
    (nth 3 (parse-partial-sexp (jvmconfig-get-beg-of-line) (point)))))


;;;###autoload
(defun jvmconfig-mode ()
  "Major mode for editing Jvmconfig code. \\<jvmconfig-mode-map>
TAB indents for Jvmconfig code.  Delete converts tabs to spaces as it moves back.

\\[jvmconfig-complete-word] completes the word around current point with respect \
to position in code
\\[jvmconfig-show-completions] shows all possible completions at this point.

Other useful functions are:

\\[jvmconfig-mark-defun]\t- Mark function.
\\[jvmconfig-insert-block]\t- insert begin ... end;
\\[jvmconfig-star-comment]\t- insert (* ... *)
\\[jvmconfig-comment-area]\t- Put marked area in a comment, fixing nested comments.
\\[jvmconfig-uncomment-area]\t- Uncomment an area commented with \
\\[jvmconfig-comment-area].
\\[jvmconfig-beg-of-defun]\t- Move to beginning of current function.
\\[jvmconfig-end-of-defun]\t- Move to end of current function.
\\[jvmconfig-goto-defun]\t- Goto function prompted for in the minibuffer.
\\[jvmconfig-outline]\t- Enter jvmconfig-outline-mode (see also jvmconfig-outline).

Variables controlling indentation/edit style:

 jvmconfig-indent-level      (default 3)
    Indentation of Jvmconfig statements with respect to containing block.
 jvmconfig-case-indent       (default 2)
    Indentation for case statements.
 jvmconfig-auto-newline      (default nil)
    Non-nil means automatically newline after semicolons and the punctuation
    mark after an end.
 jvmconfig-tab-always-indent (default t)
    Non-nil means TAB in Jvmconfig mode should always reindent the current line,
    regardless of where in the line point is when the TAB command is used.
 jvmconfig-auto-endcomments  (default t)
    Non-nil means a comment { ... } is set after the ends which ends cases and
    functions. The name of the function or case will be set between the braces.
 jvmconfig-auto-lineup       (default t)
    List of contexts where auto lineup of :'s or ='s should be done.

See also the user variables jvmconfig-type-keywords, jvmconfig-start-keywords and
jvmconfig-separator-keywords.

Turning on Jvmconfig mode calls the value of the variable jvmconfig-mode-hook with
no args, if that value is non-nil."
  (interactive)
  (kill-all-local-variables)
  (use-local-map jvmconfig-mode-map)
  (setq major-mode 'jvmconfig-mode)
  (setq mode-name "Jvmconfig")
  (setq local-abbrev-table jvmconfig-mode-abbrev-table)
  (set-syntax-table jvmconfig-mode-syntax-table)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'jvmconfig-indent-line)
  (make-local-variable 'comment-indent-function)
  (setq comment-indent-function 'jvmconfig-indent-comment)
  (make-local-variable 'parse-sexp-ignore-comments)
  (setq parse-sexp-ignore-comments nil)
  (make-local-variable 'case-fold-search)
  (setq case-fold-search t)
  (make-local-variable 'comment-start)
  (setq comment-start "#")
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "#+[ \t]*")
  (make-local-variable 'comment-end)
  (setq comment-end "")
  ;; Font lock support
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(jvmconfig-font-lock-keywords nil t))
  ;; Imenu support
  (make-local-variable 'imenu-generic-expression)
  (setq imenu-generic-expression jvmconfig-imenu-generic-expression)
  (run-hooks 'jvmconfig-mode-hook))



;;;
;;;  Electric functions
;;;
(defun electric-jvmconfig-terminate-line ()
  "Terminate line and indent next line."
  (interactive)
  ;; First, check if current line should be indented
  (save-excursion
    (beginning-of-line)
    (skip-chars-forward " \t")
    (if (looking-at jvmconfig-autoindent-lines-re)
	(jvmconfig-indent-line)))
  (delete-horizontal-space) ; Removes trailing whitespaces
  (newline)
  ;; Indent next line
  (jvmconfig-indent-line)
  ;; Maybe we should set some endcomments
  (if jvmconfig-auto-endcomments
      (jvmconfig-set-auto-comments))
    ;; If last line was a star comment line then this one shall be too.
    )


(defun electric-jvmconfig-semi-or-dot ()
  "Insert `;' or `.' character and reindent the line."
  (interactive)
  (insert last-command-char)
  (save-excursion
    (beginning-of-line)
    (jvmconfig-indent-line))
  (if jvmconfig-auto-newline
      (electric-jvmconfig-terminate-line)))

(defun electric-jvmconfig-colon ()
  "Insert `:' and do all indentions except line indent on this line."
  (interactive)
  (insert last-command-char)
  ;; Do nothing if within string.
  (if (jvmconfig-within-string)
      ()
    (save-excursion
      (beginning-of-line)
      (jvmconfig-indent-line))
    (let ((jvmconfig-tab-always-indent nil))
      (jvmconfig-indent-command))))

(defun electric-jvmconfig-equal ()
  "Insert `=', and do indention if within type declaration."
  (interactive)
  (insert last-command-char)
  (if (eq (car (jvmconfig-calculate-indent)) 'declaration)
      (let ((jvmconfig-tab-always-indent nil))
	(jvmconfig-indent-command))))

(defun electric-jvmconfig-hash ()
  "Insert `#', and indent to column 0 if this is a CPP directive."
  (interactive)
  (insert last-command-char)
  (if (save-excursion (beginning-of-line) (looking-at "^[ \t]*#"))
      (save-excursion (beginning-of-line)
		      (delete-horizontal-space))))

(defun electric-jvmconfig-tab ()
  "Function called when TAB is pressed in Jvmconfig mode."
  (interactive)
  ;; Do nothing if within a string or in a CPP directive.
  (if (or (jvmconfig-within-string)
	  (and (not (bolp))
	       (save-excursion (beginning-of-line) (eq (following-char) ?#))))
      (insert "\t")
    ;; If jvmconfig-tab-always-indent, indent the beginning of the line.
    (if jvmconfig-tab-always-indent
	(save-excursion
	  (beginning-of-line)
	  (jvmconfig-indent-line))
      (if (save-excursion
	    (skip-chars-backward " \t")
	    (bolp))
	  (jvmconfig-indent-line)
	(insert "\t")))
    (jvmconfig-indent-command)))



;;;
;;; Interactive functions
;;;
(defun jvmconfig-insert-block ()
  "Insert Jvmconfig begin ... end; block in the code with right indentation."
  (interactive)
  (jvmconfig-indent-line)
  (insert "%begin")
  (electric-jvmconfig-terminate-line)
  (save-excursion
    (electric-jvmconfig-terminate-line)
    (insert "end")
    (beginning-of-line)
    (jvmconfig-indent-line)))

(defun jvmconfig-star-comment ()
  "Insert Jvmconfig star comment at point."
  (interactive)
  (jvmconfig-indent-line)
  (insert "#")
  (electric-jvmconfig-terminate-line)
  (save-excursion
    (electric-jvmconfig-terminate-line)
    (delete-horizontal-space)
    (insert ""))
  (insert "  "))

(defun jvmconfig-mark-defun ()
  "Mark the current jvmconfig function (or procedure).
This puts the mark at the end, and point at the beginning."
  (interactive)
  (push-mark (point))
  (jvmconfig-end-of-defun)
  (push-mark (point))
  (jvmconfig-beg-of-defun)
  (if (fboundp 'zmacs-activate-region)
      (zmacs-activate-region)))

(defun jvmconfig-comment-area (start end)
  "Put the region into a Jvmconfig comment.
The comments that are in this area are \"deformed\":
`*)' becomes `!(*' and `}' becomes `!{'.
These deformed comments are returned to normal if you use
\\[jvmconfig-uncomment-area] to undo the commenting.

The commented area starts with `jvmconfig-exclude-str-start', and ends with
`jvmconfig-include-str-end'.  But if you change these variables,
\\[jvmconfig-uncomment-area] won't recognize the comments."
  (interactive "r")
  (save-excursion
    ;; Insert start and endcomments
    (goto-char end)
    (if (and (save-excursion (skip-chars-forward " \t") (eolp))
	     (not (save-excursion (skip-chars-backward " \t") (bolp))))
	(forward-line 1)
      (beginning-of-line))
    (insert jvmconfig-exclude-str-end)
    (setq end (point))
    (newline)
    (goto-char start)
    (beginning-of-line)
    (insert jvmconfig-exclude-str-start)
    (newline)
    ;; Replace end-comments within commented area
    (goto-char end)
    (save-excursion
      (while (re-search-backward "\n" start t)
	(replace-match "!#" t t)))
    (save-excursion
      (while (re-search-backward "}" start t)
	(replace-match "!#" t t)))))

(defun jvmconfig-uncomment-area ()
  "Uncomment a commented area; change deformed comments back to normal.
This command does nothing if the pointer is not in a commented
area.  See also `jvmconfig-comment-area'."
  (interactive)
  (save-excursion
    (let ((start (point))
	  (end (point)))
      ;; Find the boundaries of the comment
      (save-excursion
	(setq start (progn (search-backward jvmconfig-exclude-str-start nil t)
			   (point)))
	(setq end (progn (search-forward jvmconfig-exclude-str-end nil t)
			 (point))))
      ;; Check if we're really inside a comment
      (if (or (equal start (point)) (<= end (point)))
	  (message "Not standing within commented area.")
	(progn
	  ;; Remove endcomment
	  (goto-char end)
	  (beginning-of-line)
	  (let ((pos (point)))
	    (end-of-line)
	    (delete-region pos (1+ (point))))
	  ;; Change comments back to normal
	  (save-excursion
	    (while (re-search-backward "!#" start t)
	      (replace-match "\n" t t)))
	  ;; Remove startcomment
	  (goto-char start)
	  (beginning-of-line)
	  (let ((pos (point)))
	    (end-of-line)
	    (delete-region pos (1+ (point)))))))))

(defun jvmconfig-beg-of-defun ()
  "Move backward to the beginning of the current function or procedure."
  (interactive)
  (catch 'found
    (if (not (looking-at (concat "\\s \\|\\s)\\|" jvmconfig-defun-re)))
	(forward-sexp 1))
    (let ((nest 0) (max -1) (func 0)
	  (reg (concat jvmconfig-beg-block-re "\\|" 
		       jvmconfig-end-block-re "\\|"
		       jvmconfig-defun-re)))
      (while (re-search-backward reg nil 'move)
	(cond ((let ((state (save-excursion
			      (parse-partial-sexp (point-min) (point)))))
		 (or (nth 3 state) (nth 4 state))) ; Inside string or comment
	       ())
	      ((match-end 1)                       ; begin|case|record|repeat
	       (if (and (looking-at "\\<record\\>") (>= max 0))
		   (setq func (1- func)))
	       (setq nest (1+ nest)
		     max (max nest max)))
	      ((match-end 2)                       ; end|until
	       (if (and (= nest max) (>= max 0))
		   (setq func (1+ func)))
	       (setq nest (1- nest)))
	      ((match-end 3)                       ; function|procedure
	       (if (= 0 func)
		   (throw 'found t)
		 (setq func (1- func)))))))
    nil))

(defun jvmconfig-end-of-defun ()
  "Move forward to the end of the current function or procedure."
  (interactive)
  (if (looking-at "\\s ")
      (forward-sexp 1))
  (if (not (looking-at jvmconfig-defun-re))
      (jvmconfig-beg-of-defun))
  (forward-char 1)
  (let ((nest 0) (func 1)
	(reg (concat jvmconfig-beg-block-re "\\|" 
		     jvmconfig-end-block-re "\\|"
		     jvmconfig-defun-re)))
    (while (and (/= func 0)
		(re-search-forward reg nil 'move))
      (cond ((let ((state (save-excursion
			      (parse-partial-sexp (point-min) (point)))))
		 (or (nth 3 state) (nth 4 state))) ; Inside string or comment
	       ())
	    ((match-end 1)
	     (setq nest (1+ nest))
	     (if (save-excursion
		   (goto-char (match-beginning 0))
		   (looking-at "\\<record\\>"))
		 (setq func (1+ func))))
	    ((match-end 2)
	     (setq nest (1- nest))
	     (if (= nest 0)
		 (setq func (1- func))))
	    ((match-end 3)
	     (setq func (1+ func))))))
  (forward-line 1))

(defun jvmconfig-end-of-statement ()
  "Move forward to end of current statement."
  (interactive)
  (let ((parse-sexp-ignore-comments t)
	(nest 0) pos
	(regexp (concat "\\(" jvmconfig-beg-block-re "\\)\\|\\("
			jvmconfig-end-block-re "\\)")))
    (if (not (looking-at "[ \t\n]")) (forward-sexp -1))
    (or (looking-at jvmconfig-beg-block-re)
	;; Skip to end of statement
	(setq pos (catch 'found
		    (while t
		      (forward-sexp 1)
		      (cond ((looking-at "[ \t]*;")
			     (skip-chars-forward "^;")
			     (forward-char 1)
			     (throw 'found (point)))
			    ((save-excursion
			       (forward-sexp -1)
			       (looking-at jvmconfig-beg-block-re))
			     (goto-char (match-beginning 0))
			     (throw 'found nil))
			    ((eobp)
			     (throw 'found (point))))))))
    (if (not pos)
	;; Skip a whole block
	(catch 'found
	  (while t
	    (re-search-forward regexp nil 'move)
	    (setq nest (if (match-end 1) 
			   (1+ nest)
			 (1- nest)))
	    (cond ((eobp)
		   (throw 'found (point)))
		  ((= 0 nest)
		   (throw 'found (jvmconfig-end-of-statement))))))
      pos)))

(defun jvmconfig-downcase-keywords ()
  "Downcase all Jvmconfig keywords in the buffer."
  (interactive)
  (jvmconfig-change-keywords 'downcase-word))

(defun jvmconfig-upcase-keywords ()
  "Upcase all Jvmconfig keywords in the buffer."
  (interactive)
  (jvmconfig-change-keywords 'upcase-word))

(defun jvmconfig-capitalize-keywords ()
  "Capitalize all Jvmconfig keywords in the buffer."
  (interactive)
  (jvmconfig-change-keywords 'capitalize-word))

;; Change the keywords according to argument.
(defun jvmconfig-change-keywords (change-word)
  (save-excursion
    (let ((keyword-re (concat "\\<\\("
			      (mapconcat 'identity jvmconfig-keywords "\\|")
			      "\\)\\>")))
      (goto-char (point-min))
      (while (re-search-forward keyword-re nil t)
	(funcall change-word -1)))))



;;;
;;; Other functions
;;;
(defun jvmconfig-set-auto-comments ()
  "Insert `{ case }' or `{ NAME }' on this line if appropriate.
Insert `{ case }' if there is an `end' on the line which
ends a case block.  Insert `{ NAME }' if there is an `end'
on the line which ends a function or procedure named NAME."
  (save-excursion
    (forward-line -1)
    (skip-chars-forward " \t")
    (if (and (looking-at "\\<end")
	     (not (save-excursion
		    (end-of-line)
		    (search-backward "#" (jvmconfig-get-beg-of-line) t))))
	(let ((type (car (jvmconfig-calculate-indent))))
	  (if (eq type 'declaration)
	      ()
	    (if (eq type 'case)
		;; This is a case block
		(progn
		  (end-of-line)
		  (delete-horizontal-space)
		  (insert " # case"))
	      (let ((nest 1))
		;; Check if this is the end of a function
		(save-excursion
		  (while (not (or (looking-at jvmconfig-defun-re) (bobp)))
		    (backward-sexp 1)
		    (cond ((looking-at jvmconfig-beg-block-re)
			   (setq nest (1- nest)))
			  ((looking-at jvmconfig-end-block-re)
			   (setq nest (1+ nest)))))
		  (if (bobp)
		      (setq nest 1)))
		(if (zerop nest)
		    (progn
		      (end-of-line)
		      (delete-horizontal-space)
		      (insert " # ")
		      (let (b e)
			(save-excursion
			  (setq b (progn (jvmconfig-beg-of-defun)
					 (skip-chars-forward "^ \t")
					 (skip-chars-forward " \t")
					 (point))
				e (progn (skip-chars-forward "a-zA-Z0-9_")
					 (point))))
			(insert-buffer-substring (current-buffer) b e))
		      (insert " "))))))))))



;;;
;;; Indentation
;;;
(defconst jvmconfig-indent-alist
  '((block . (+ ind jvmconfig-indent-level))
    (case . (+ ind jvmconfig-case-indent))
    (caseblock . ind) (cpp . 0)
    (declaration . (+ ind jvmconfig-indent-level))
    (paramlist . (jvmconfig-indent-paramlist t))
    (comment . (jvmconfig-indent-comment t))
    (defun . ind) (contexp . ind)
    (unknown . 0) (string . 0)))

(defun jvmconfig-indent-command ()
  "Indent for special part of code."
  (let* ((indent-str (jvmconfig-calculate-indent))
	 (type (car indent-str))
	 (ind (car (cdr indent-str))))
    (cond ((and (eq type 'paramlist)
		(or (memq 'all jvmconfig-auto-lineup)
		    (memq 'paramlist jvmconfig-auto-lineup)))
	   (jvmconfig-indent-paramlist)
	   (jvmconfig-indent-paramlist))
	  ((and (eq type 'declaration)
		(or (memq 'all jvmconfig-auto-lineup)
		    (memq 'declaration  jvmconfig-auto-lineup)))
	   (jvmconfig-indent-declaration))
	  ((and (eq type 'case) (not (looking-at "^[ \t]*$"))
		(or (memq 'all jvmconfig-auto-lineup)
		    (memq 'case jvmconfig-auto-lineup)))
	   (jvmconfig-indent-case)))
    (if (looking-at "[ \t]+$")
	(skip-chars-forward " \t"))))

(defun jvmconfig-indent-line ()
  "Indent current line as a Jvmconfig statement."
  (let* ((indent-str (jvmconfig-calculate-indent))
	 (type (car indent-str))
	 (ind (car (cdr indent-str))))
    (if (looking-at "^[0-9a-zA-Z]+[ \t]*:[^=]")
	(search-forward ":" nil t))
    (delete-horizontal-space)
    ;; Some things should not be indented
    (if (or (and (eq type 'declaration) (looking-at jvmconfig-declaration-re))
	    (eq type 'cpp)
	    (looking-at jvmconfig-defun-re))
	()
      ;; Other things should have no extra indent
      (if (looking-at jvmconfig-noindent-re)
	  (indent-to ind)
	;; But most lines are treated this way:
	(indent-to (eval (cdr (assoc type jvmconfig-indent-alist))))
	))))

(defun jvmconfig-calculate-indent ()
  "Calculate the indent of the current Jvmconfig line.
Return a list of two elements: (INDENT-TYPE INDENT-LEVEL)."
  (save-excursion
    (let* ((parse-sexp-ignore-comments t)
	   (oldpos (point))
	   (state (save-excursion (parse-partial-sexp (point-min) (point))))
	   (nest 0) (par 0) (complete (looking-at "[ \t]*end\\>"))
	   (elsed (looking-at "[ \t]*else\\>"))
	   (type (catch 'nesting
		   ;; Check if inside a string, comment or parenthesis
		   (cond ((nth 3 state) (throw 'nesting 'string))
			 ((nth 4 state) (throw 'nesting 'comment))
			 ((> (car state) 0)
			  (goto-char (scan-lists (point) -1 (car state)))
			  (setq par (1+ (current-column))))
			 ((save-excursion (beginning-of-line)
					  (eq (following-char) ?#))
			  (throw 'nesting 'cpp)))
		   ;; Loop until correct indent is found
		   (while t
		     (backward-sexp 1)
		     (cond (;--Escape from case statements
			    (and (looking-at "[A-Za-z0-9]+[ \t]*:[^=]")
				 (not complete)
				 (save-excursion (skip-chars-backward " \t")
						 (bolp))
				 (= (save-excursion
				      (end-of-line) (backward-sexp) (point))
				    (point))
				 (> (save-excursion (goto-char oldpos)
						    (beginning-of-line)
						    (point))
				    (point)))
			    (throw 'nesting 'caseblock))
			   (;--Nest block outwards
			    (looking-at jvmconfig-beg-block-re)
			    (if (= nest 0)
				(cond ((looking-at "case\\>")
				       (throw 'nesting 'case))
				      ((looking-at "record\\>")
				       (throw 'nesting 'declaration))
				      (t (throw 'nesting 'block)))
			      (setq nest (1- nest))))
			   (;--Nest block inwards
			    (looking-at jvmconfig-end-block-re)
			    (if (and (looking-at "\\<end\\s ")
				     elsed (not complete))
				(throw 'nesting 'block))
			    (setq complete t
				  nest (1+ nest)))
			   (;--Defun (or parameter list)
			    (looking-at jvmconfig-defun-re)
			    (if (= 0 par)
				(throw 'nesting 'defun)
			      (setq par 0)
			      (let ((n 0))
				(while (re-search-forward
					"\\(\\<record\\>\\)\\|\\<end\\>"
					oldpos t)
				  (if (match-end 1)
				      (setq n (1+ n)) (setq n (1- n))))
				(if (> n 0)
				    (throw 'nesting 'declaration)
				  (throw 'nesting 'paramlist)))))
			   (;--Declaration part
			    (looking-at jvmconfig-declaration-re)
			    (if (save-excursion
				  (goto-char oldpos)
				  (forward-line -1)
				  (looking-at "^[ \t]*$"))
				(throw 'nesting 'unknown)
			      (throw 'nesting 'declaration)))
			   (;--If, else or while statement
			    (and (not complete)
				 (looking-at jvmconfig-sub-block-re))
			    (throw 'nesting 'block))
			   (;--Found complete statement
			    (save-excursion (forward-sexp 1)
					    (= (following-char) ?\;))
			    (setq complete t))
			   (;--No known statements
			    (bobp)
			    (throw 'nesting 'unknown))
			   )))))

      ;; Return type of block and indent level.
      (if (> par 0)                               ; Unclosed Parenthesis 
	  (list 'contexp par)
	(list type (jvmconfig-indent-level))))))

(defun jvmconfig-indent-level ()
  "Return the indent-level the current statement has.
Do not count labels, case-statements or records."
  (save-excursion
    (beginning-of-line)
    (if (looking-at "[ \t]*[0-9a-zA-Z]+[ \t]*:[^=]")
	(search-forward ":" nil t)
      (if (looking-at ".*=[ \t]*record\\>")
	  (search-forward "=" nil t)))
    (skip-chars-forward " \t")
    (current-column)))

(defun jvmconfig-indent-comment (&optional arg)
  "Indent current line as comment.
If optional arg is non-nil, just return the
column number the line should be indented to."
  (let* ((stcol (save-excursion
		  (re-search-backward "#" nil t)
		  (1+ (current-column)))))
    (if arg stcol
      (delete-horizontal-space)
      (indent-to stcol))))

(defun jvmconfig-indent-case ()
  "Indent within case statements."
  (let ((savepos (point-marker))
	(end (prog2
		 (end-of-line)
		 (point-marker)
	       (re-search-backward "\\<case\\>" nil t)))
	(beg (point)) oldpos
	(ind 0))
    ;; Get right indent
    (while (< (point) end)
      (if (re-search-forward 
	   "^[ \t]*[^ \t,:]+[ \t]*\\(,[ \t]*[^ \t,:]+[ \t]*\\)*:"
	   (marker-position end) 'move)
	  (forward-char -1))
      (if (< (point) end)
	  (progn
	    (delete-horizontal-space)
	    (if (> (current-column) ind)
		(setq ind (current-column)))
	    (jvmconfig-end-of-statement))))
    (goto-char beg)
    (setq oldpos (marker-position end))
    ;; Indent all case statements
    (while (< (point) end)
      (if (re-search-forward
	   "^[ \t]*[^][ \t,\\.:]+[ \t]*\\(,[ \t]*[^ \t,:]+[ \t]*\\)*:"
	   (marker-position end) 'move)
	  (forward-char -1))
      (indent-to (1+ ind))
      (if (/= (following-char) ?:)
	  ()
	(forward-char 1)
	(delete-horizontal-space)
	(insert " "))
      (setq oldpos (point))
      (jvmconfig-end-of-statement))
    (goto-char savepos)))

(defun jvmconfig-indent-paramlist (&optional arg)
  "Indent current line in parameterlist.
If optional arg is non-nil, just return the
indent of the current line in parameterlist."
  (save-excursion
    (let* ((oldpos (point))
	   (stpos (progn (goto-char (scan-lists (point) -1 1)) (point)))
	   (stcol (1+ (current-column)))
	   (edpos (progn (jvmconfig-declaration-end) 
			 (search-backward ")" (jvmconfig-get-beg-of-line) t)
			 (point)))
	   (usevar (re-search-backward "\\<var\\>" stpos t)))
      (if arg (progn
		;; If arg, just return indent
		(goto-char oldpos)
		(beginning-of-line)
		(if (or (not usevar) (looking-at "[ \t]*var\\>"))
		    stcol (+ 4 stcol)))
	(goto-char stpos)
	(forward-char 1)
	(delete-horizontal-space)
	(if (and usevar (not (looking-at "var\\>")))
	    (indent-to (+ 4 stcol)))
	(jvmconfig-indent-declaration nil stpos edpos)))))

(defun jvmconfig-indent-declaration (&optional arg start end)
  "Indent current lines as declaration, lining up the `:'s or `='s."
  (let ((pos (point-marker)))
    (if (and (not (or arg start)) (not (jvmconfig-declaration-beg)))
	()
      (let ((lineup (if (or (looking-at "\\<var\\>\\|\\<record\\>") arg start) 
			":" "="))
	    (stpos (if start start
		       (forward-word 2) (backward-word 1) (point)))
	    (edpos (set-marker (make-marker)
			       (if end end
				 (max (progn (jvmconfig-declaration-end)
					     (point))
				      pos))))
	    ind)

	(goto-char stpos)
	;; Indent lines in record block
	(if arg
	    (while (<= (point) edpos)
	      (beginning-of-line)
	      (delete-horizontal-space)
	      (if (looking-at "end\\>")
		  (indent-to arg)
		(indent-to (+ arg jvmconfig-indent-level)))
	      (forward-line 1)))

	;; Do lineup
	(setq ind (jvmconfig-get-lineup-indent stpos edpos lineup))
	(goto-char stpos)
	(while (and (<= (point) edpos) (not (eobp)))
	  (if (search-forward lineup (jvmconfig-get-end-of-line) 'move)
	      (forward-char -1))
	  (delete-horizontal-space)
	  (indent-to ind)
	  (if (not (looking-at lineup))
	      (forward-line 1) ; No more indent if there is no : or =
	    (forward-char 1)
	    (delete-horizontal-space)
	    (insert " ")
	    ;; Indent record block
	    (if (looking-at "record\\>")
		(jvmconfig-indent-declaration (current-column)))
	    (forward-line 1)))))

    ;; If arg - move point
    (if arg (forward-line -1)
      (goto-char pos))))

;  "Return the indent level that will line up several lines within the region
;from b to e nicely. The lineup string is str."
(defun jvmconfig-get-lineup-indent (b e str)
  (save-excursion
    (let ((ind 0)
	  (reg (concat str "\\|\\(\\<record\\>\\)")))
      (goto-char b)
      ;; Get rightmost position
      (while (< (point) e)
	(if (re-search-forward reg (min e (jvmconfig-get-end-of-line 2)) 'move)
	    (progn
	      ;; Skip record blocks
	      (if (match-beginning 1)
		  (jvmconfig-declaration-end)
		(progn
		  (goto-char (match-beginning 0))
		  (skip-chars-backward " \t")
		  (if (> (current-column) ind)
		      (setq ind (current-column)))
		  (goto-char (match-end 0))
		  (end-of-line)
		  )))))
      ;; In case no lineup was found
      (if (> ind 0)
	  (1+ ind)
	;; No lineup-string found
	(goto-char b)
	(end-of-line)
	(skip-chars-backward " \t")
	(1+ (current-column))))))
    


;;;
;;; Completion
;;;
(defvar jvmconfig-str nil)
(defvar jvmconfig-all nil)
(defvar jvmconfig-pred nil)
(defvar jvmconfig-buffer-to-use nil)
(defvar jvmconfig-flag nil)

(defun jvmconfig-string-diff (str1 str2)
  "Return index of first letter where STR1 and STR2 differs."
  (catch 'done
    (let ((diff 0))
      (while t
	(if (or (> (1+ diff) (length str1))
		(> (1+ diff) (length str2)))
	    (throw 'done diff))
	(or (equal (aref str1 diff) (aref str2 diff))
	    (throw 'done diff))
	(setq diff (1+ diff))))))

;; Calculate all possible completions for functions if argument is `function',
;; completions for procedures if argument is `procedure' or both functions and
;; procedures otherwise.

(defun jvmconfig-func-completion (type)
  ;; Build regular expression for function/procedure names
  (if (string= jvmconfig-str "")
      (setq jvmconfig-str "[a-zA-Z_]"))
  (let ((jvmconfig-str (concat (cond
			     ((eq type 'procedure) "\\<\\(procedure\\)\\s +")
			     ((eq type 'function) "\\<\\(function\\)\\s +")
			     (t "\\<\\(function\\|procedure\\|begin\\|script\\|text\\)\\s +"))
			    "\\<\\(" jvmconfig-str "[a-zA-Z0-9_.]*\\)\\>"))
	match)
    
    (if (not (looking-at "\\<\\(function\\|procedure\\)\\>"))
	(re-search-backward "\\<\\(function\\|procedure\\)\\>" nil t))
    (forward-char 1)

    ;; Search through all reachable functions
    (while (jvmconfig-beg-of-defun)
      (if (re-search-forward jvmconfig-str (jvmconfig-get-end-of-line) t)
	  (progn (setq match (buffer-substring (match-beginning 2)
					       (match-end 2)))
		 (if (or (null jvmconfig-pred)
			 (funcall jvmconfig-pred match))
		     (setq jvmconfig-all (cons match jvmconfig-all)))))
      (goto-char (match-beginning 0)))))

(defun jvmconfig-get-completion-decl ()
  ;; Macro for searching through current declaration (var, type or const)
  ;; for matches of `str' and adding the occurrence to `all'
  (let ((end (save-excursion (jvmconfig-declaration-end)
			     (point)))
	match)
    ;; Traverse lines
    (while (< (point) end)
      (if (re-search-forward "[:=]" (jvmconfig-get-end-of-line) t)
	  ;; Traverse current line
	  (while (and (re-search-backward 
		       (concat "\\((\\|\\<\\(var\\|type\\|const\\)\\>\\)\\|" 
			       jvmconfig-symbol-re)
		       (jvmconfig-get-beg-of-line) t)
		      (not (match-end 1)))
	    (setq match (buffer-substring (match-beginning 0) (match-end 0)))
	    (if (string-match (concat "\\<" jvmconfig-str) match)
		(if (or (null jvmconfig-pred)
			(funcall jvmconfig-pred match))
		    (setq jvmconfig-all (cons match jvmconfig-all))))))
      (if (re-search-forward "\\<record\\>" (jvmconfig-get-end-of-line) t)
	  (jvmconfig-declaration-end)
	(forward-line 1)))))

(defun jvmconfig-type-completion ()
  "Calculate all possible completions for types."
  (let ((start (point))
	goon)
    ;; Search for all reachable type declarations
    (while (or (jvmconfig-beg-of-defun)
	       (setq goon (not goon)))
      (save-excursion
	(if (and (< start (prog1 (save-excursion (jvmconfig-end-of-defun)
						 (point))
			    (forward-char 1)))
		 (re-search-forward
		  "\\<type\\>\\|\\<\\(function\\|procedure\\)\\>"
		  start t)
		 (not (match-end 1)))
	    ;; Check current type declaration
	    (jvmconfig-get-completion-decl))))))

(defun jvmconfig-var-completion ()
  "Calculate all possible completions for variables (or constants)."
  (let ((start (point))
	goon twice)
    ;; Search for all reachable var declarations
    (while (or (jvmconfig-beg-of-defun)
	       (setq goon (not goon)))
      (save-excursion
	(if (> start (prog1 (save-excursion (jvmconfig-end-of-defun)
					    (point))))
	    () ; Declarations not reachable
	  (if (search-forward "(" (jvmconfig-get-end-of-line) t)
	      ;; Check parameterlist
		(jvmconfig-get-completion-decl))
	  (setq twice 2)
	  (while (>= (setq twice (1- twice)) 0)
	    (cond ((and (re-search-forward
			 (concat "\\<\\(var\\|const\\)\\>\\|"
				 "\\<\\(function\\|procedure\\)\\>")
			 start t)
			(not (match-end 2)))
		   ;; Check var/const declarations
		   (jvmconfig-get-completion-decl))
		  ((match-end 2)
		   (setq twice 0)))))))))


(defun jvmconfig-keyword-completion (keyword-list)
  "Give list of all possible completions of keywords in KEYWORD-LIST."
  (mapcar '(lambda (s) 
	     (if (string-match (concat "\\<" jvmconfig-str) s)
		 (if (or (null jvmconfig-pred)
			 (funcall jvmconfig-pred s))
		     (setq jvmconfig-all (cons s jvmconfig-all)))))
	  keyword-list))

;; Function passed to completing-read, try-completion or
;; all-completions to get completion on STR. If predicate is non-nil,
;; it must be a function to be called for every match to check if this
;; should really be a match. If flag is t, the function returns a list
;; of all possible completions. If it is nil it returns a string, the
;; longest possible completion, or t if STR is an exact match. If flag
;; is 'lambda, the function returns t if STR is an exact match, nil
;; otherwise.

(defun jvmconfig-completion (jvmconfig-str jvmconfig-pred jvmconfig-flag)
  (save-excursion
    (let ((jvmconfig-all nil))
      ;; Set buffer to use for searching labels. This should be set
      ;; within functions which use jvmconfig-completions
      (set-buffer jvmconfig-buffer-to-use)

      ;; Determine what should be completed
      (let ((state (car (jvmconfig-calculate-indent))))
	(cond (;--Within a declaration or parameterlist
	       (or (eq state 'declaration) (eq state 'paramlist)
		   (and (eq state 'defun)
			(save-excursion
			  (re-search-backward ")[ \t]*:"
					      (jvmconfig-get-beg-of-line) t))))
	       (if (or (eq state 'paramlist) (eq state 'defun))
		   (jvmconfig-beg-of-defun))
	       (jvmconfig-type-completion)
	       (jvmconfig-keyword-completion jvmconfig-type-keywords))
	      (;--Starting a new statement
	       (and (not (eq state 'contexp))
		    (save-excursion
		      (skip-chars-backward "a-zA-Z0-9_.")
		      (backward-sexp 1)
		      (or (looking-at jvmconfig-nosemi-re)
			  (progn
			    (forward-sexp 1)
			    (looking-at "\\s *\\(;\\|:[^=]\\)")))))
	       (save-excursion (jvmconfig-var-completion))
	       (jvmconfig-func-completion 'procedure)
	       (jvmconfig-keyword-completion jvmconfig-start-keywords))
	      (t;--Anywhere else
	       (save-excursion (jvmconfig-var-completion))
	       (jvmconfig-func-completion 'function)
	       (jvmconfig-keyword-completion jvmconfig-separator-keywords))))
      
      ;; Now we have built a list of all matches. Give response to caller
      (jvmconfig-completion-response))))

(defun jvmconfig-completion-response ()
  (cond ((or (equal jvmconfig-flag 'lambda) (null jvmconfig-flag))
	 ;; This was not called by all-completions
	 (if (null jvmconfig-all)
	     ;; Return nil if there was no matching label
	     nil
	   ;; Get longest string common in the labels
	   (let* ((elm (cdr jvmconfig-all))
		  (match (car jvmconfig-all))
		  (min (length match))
		  tmp)
	     (if (string= match jvmconfig-str)
		 ;; Return t if first match was an exact match
		 (setq match t)
	       (while (not (null elm))
		 ;; Find longest common string
		 (if (< (setq tmp (jvmconfig-string-diff match (car elm))) min)
		     (progn
		       (setq min tmp)
		       (setq match (substring match 0 min))))
		 ;; Terminate with match=t if this is an exact match
		 (if (string= (car elm) jvmconfig-str)
		     (progn
		       (setq match t)
		       (setq elm nil))
		   (setq elm (cdr elm)))))
	     ;; If this is a test just for exact match, return nil ot t
	     (if (and (equal jvmconfig-flag 'lambda) (not (equal match 't)))
		 nil
	       match))))
	;; If flag is t, this was called by all-completions. Return
	;; list of all possible completions
	(jvmconfig-flag
	 jvmconfig-all)))

(defvar jvmconfig-last-word-numb 0)
(defvar jvmconfig-last-word-shown nil)
(defvar jvmconfig-last-completions nil)

(defun jvmconfig-complete-word ()
  "Complete word at current point.
\(See also `jvmconfig-toggle-completions', `jvmconfig-type-keywords',
`jvmconfig-start-keywords' and `jvmconfig-separator-keywords'.)"
  (interactive)
  (let* ((b (save-excursion (skip-chars-backward "a-zA-Z0-9_") (point)))
	 (e (save-excursion (skip-chars-forward "a-zA-Z0-9_") (point)))
	 (jvmconfig-str (buffer-substring b e))
	 ;; The following variable is used in jvmconfig-completion
	 (jvmconfig-buffer-to-use (current-buffer))
	 (allcomp (if (and jvmconfig-toggle-completions
			   (string= jvmconfig-last-word-shown jvmconfig-str))
		      jvmconfig-last-completions
		    (all-completions jvmconfig-str 'jvmconfig-completion)))
	 (match (if jvmconfig-toggle-completions
		    "" (try-completion
			jvmconfig-str (mapcar '(lambda (elm)
					      (cons elm 0)) allcomp)))))
    ;; Delete old string
    (delete-region b e)

    ;; Toggle-completions inserts whole labels
    (if jvmconfig-toggle-completions
	(progn
	  ;; Update entry number in list
	  (setq jvmconfig-last-completions allcomp
		jvmconfig-last-word-numb 
		(if (>= jvmconfig-last-word-numb (1- (length allcomp)))
		    0
		  (1+ jvmconfig-last-word-numb)))
	  (setq jvmconfig-last-word-shown (elt allcomp jvmconfig-last-word-numb))
	  ;; Display next match or same string if no match was found
	  (if (not (null allcomp))
	      (insert "" jvmconfig-last-word-shown)
	    (insert "" jvmconfig-str)
	    (message "(No match)")))
      ;; The other form of completion does not necessarily do that.

      ;; Insert match if found, or the original string if no match
      (if (or (null match) (equal match 't))
	  (progn (insert "" jvmconfig-str)
		 (message "(No match)"))
	(insert "" match))
      ;; Give message about current status of completion
      (cond ((equal match 't)
	     (if (not (null (cdr allcomp)))
		 (message "(Complete but not unique)")
	       (message "(Sole completion)")))
	    ;; Display buffer if the current completion didn't help 
	    ;; on completing the label.
	    ((and (not (null (cdr allcomp))) (= (length jvmconfig-str)
						(length match)))
	     (with-output-to-temp-buffer "*Completions*"
	       (display-completion-list allcomp))
	     ;; Wait for a keypress. Then delete *Completion*  window
	     (momentary-string-display "" (point))
	     (delete-window (get-buffer-window (get-buffer "*Completions*")))
	     )))))

(defun jvmconfig-show-completions ()
  "Show all possible completions at current point."
  (interactive)
  (let* ((b (save-excursion (skip-chars-backward "a-zA-Z0-9_") (point)))
	 (e (save-excursion (skip-chars-forward "a-zA-Z0-9_") (point)))
	 (jvmconfig-str (buffer-substring b e))
	 ;; The following variable is used in jvmconfig-completion
	 (jvmconfig-buffer-to-use (current-buffer))
	 (allcomp (if (and jvmconfig-toggle-completions
			   (string= jvmconfig-last-word-shown jvmconfig-str))
		      jvmconfig-last-completions
		    (all-completions jvmconfig-str 'jvmconfig-completion))))
    ;; Show possible completions in a temporary buffer.
    (with-output-to-temp-buffer "*Completions*"
      (display-completion-list allcomp))
    ;; Wait for a keypress. Then delete *Completion*  window
    (momentary-string-display "" (point))
    (delete-window (get-buffer-window (get-buffer "*Completions*")))))


(defun jvmconfig-get-default-symbol ()
  "Return symbol around current point as a string."
  (save-excursion
    (buffer-substring (progn
			(skip-chars-backward " \t")
			(skip-chars-backward "a-zA-Z0-9_")
			(point))
		      (progn
			(skip-chars-forward "a-zA-Z0-9_")
			(point)))))

(defun jvmconfig-build-defun-re (str &optional arg)
  "Return function/procedure starting with STR as regular expression.
With optional second arg non-nil, STR is the complete name of the instruction."
  (if arg
      (concat "^\\(function\\|procedure\\)[ \t]+\\(" str "\\)\\>")
    (concat "^\\(function\\|procedure\\)[ \t]+\\(" str "[a-zA-Z0-9_]*\\)\\>")))

;; Function passed to completing-read, try-completion or
;; all-completions to get completion on any function name. If
;; predicate is non-nil, it must be a function to be called for every
;; match to check if this should really be a match. If flag is t, the
;; function returns a list of all possible completions. If it is nil
;; it returns a string, the longest possible completion, or t if STR
;; is an exact match. If flag is 'lambda, the function returns t if
;; STR is an exact match, nil otherwise.

(defun jvmconfig-comp-defun (jvmconfig-str jvmconfig-pred jvmconfig-flag)
  (save-excursion
    (let ((jvmconfig-all nil)
	  match)

      ;; Set buffer to use for searching labels. This should be set
      ;; within functions which use jvmconfig-completions
      (set-buffer jvmconfig-buffer-to-use)

      (let ((jvmconfig-str jvmconfig-str))
	;; Build regular expression for functions
	(if (string= jvmconfig-str "")
	    (setq jvmconfig-str (jvmconfig-build-defun-re "[a-zA-Z_]"))
	  (setq jvmconfig-str (jvmconfig-build-defun-re jvmconfig-str)))
	(goto-char (point-min))
      
	;; Build a list of all possible completions
	(while (re-search-forward jvmconfig-str nil t)
	  (setq match (buffer-substring (match-beginning 2) (match-end 2)))
	  (if (or (null jvmconfig-pred)
		  (funcall jvmconfig-pred match))
	      (setq jvmconfig-all (cons match jvmconfig-all)))))

      ;; Now we have built a list of all matches. Give response to caller
      (jvmconfig-completion-response))))

(defun jvmconfig-goto-defun ()
  "Move to specified Jvmconfig function/procedure.
The default is a name found in the buffer around point."
  (interactive)
  (let* ((default (jvmconfig-get-default-symbol))
	 ;; The following variable is used in jvmconfig-comp-function
	 (jvmconfig-buffer-to-use (current-buffer))
	 (default (if (jvmconfig-comp-defun default nil 'lambda)
		      default ""))
	 (label (if (not (string= default ""))
		    ;; Do completion with default
		    (completing-read (concat "Label: (default " default ") ")
				     'jvmconfig-comp-defun nil t "")
		  ;; There is no default value. Complete without it
		  (completing-read "Label: "
				   'jvmconfig-comp-defun nil t ""))))
    ;; If there was no response on prompt, use default value
    (if (string= label "")
	(setq label default))
    ;; Goto right place in buffer if label is not an empty string
    (or (string= label "")
	(progn
	  (goto-char (point-min))
	  (re-search-forward (jvmconfig-build-defun-re label t))
	  (beginning-of-line)))))



;;;
;;; Jvmconfig-outline-mode
;;;
(defvar jvmconfig-outline-map nil "Keymap used in Jvmconfig Outline mode.")

(if jvmconfig-outline-map
    nil
  (if (boundp 'set-keymap-name)
      (set-keymap-name jvmconfig-outline-map 'jvmconfig-outline-map))
  (if (not (boundp 'set-keymap-parent))
      (setq jvmconfig-outline-map (copy-keymap jvmconfig-mode-map))
    (setq jvmconfig-outline-map (make-sparse-keymap))
    (set-keymap-parent jvmconfig-outline-map jvmconfig-mode-map))
  (define-key jvmconfig-outline-map "\M-\C-a"  'jvmconfig-outline-prev-defun)
  (define-key jvmconfig-outline-map "\M-\C-e"  'jvmconfig-outline-next-defun)
  (define-key jvmconfig-outline-map "\C-c\C-d" 'jvmconfig-outline-goto-defun)
  (define-key jvmconfig-outline-map "\C-c\C-s" 'jvmconfig-show-all)
  (define-key jvmconfig-outline-map "\C-c\C-h" 'jvmconfig-hide-other-defuns))

(defvar jvmconfig-outline-mode nil "Non-nil while using Jvmconfig Outline mode.")
(make-variable-buffer-local 'jvmconfig-outline-mode)
(set-default 'jvmconfig-outline-mode nil)
(if (not (assoc 'jvmconfig-outline-mode minor-mode-alist))
    (setq minor-mode-alist (append minor-mode-alist
				   (list '(jvmconfig-outline-mode " Outl")))))

(defun jvmconfig-outline (&optional arg)
  "Outline-line minor mode for Jvmconfig mode.
When in Jvmconfig Outline mode, portions
of the text being edited may be made invisible. \\<jvmconfig-outline-map>

Jvmconfig Outline mode provides some additional commands.

\\[jvmconfig-outline-prev-defun]\
\t- Move to previous function/procedure, hiding everything else.
\\[jvmconfig-outline-next-defun]\
\t- Move to next function/procedure, hiding everything else.
\\[jvmconfig-outline-goto-defun]\
\t- Goto function/procedure prompted for in minibuffer,
\t  hide all other functions.
\\[jvmconfig-show-all]\t- Show the whole buffer.
\\[jvmconfig-hide-other-defuns]\
\t- Hide everything but the current function (function under the cursor).
\\[jvmconfig-outline]\t- Leave jvmconfig-outline-mode."
  (interactive "P")
  (setq jvmconfig-outline-mode
	(if (null arg) (not jvmconfig-outline-mode) t))
  (if (boundp 'redraw-mode-line)
      (redraw-mode-line))
  (if jvmconfig-outline-mode
      (progn
	(setq selective-display t)
	(use-local-map jvmconfig-outline-map))
    (progn
      (setq selective-display nil)
      (jvmconfig-show-all)
      (use-local-map jvmconfig-mode-map))))

(defun jvmconfig-outline-change (b e jvmconfig-flag)
  (let ((modp (buffer-modified-p)))
    (unwind-protect
	(subst-char-in-region b e (if (= jvmconfig-flag ?\n) 
				      ?\^M ?\n) jvmconfig-flag)
      (set-buffer-modified-p modp))))

(defun jvmconfig-show-all ()
  "Show all of the text in the buffer."
  (interactive)
  (jvmconfig-outline-change (point-min) (point-max) ?\n))

(defun jvmconfig-hide-other-defuns ()
  "Show only the current defun."
  (interactive)
  (save-excursion
    (let ((beg (progn (if (not (looking-at "\\(function\\|procedure\\)\\>"))
			  (jvmconfig-beg-of-defun))
		      (point)))
	  (end (progn (jvmconfig-end-of-defun)
		      (backward-sexp 1)
		      (search-forward "\n\\|\^M" nil t)
		      (point)))
	  (opoint (point-min)))
      (goto-char (point-min))

      ;; Hide all functions before current function
      (while (re-search-forward "^\\(function\\|procedure\\)\\>" beg 'move)
	(jvmconfig-outline-change opoint (1- (match-beginning 0)) ?\^M)
	(setq opoint (point))
	;; Functions may be nested
	(if (> (progn (jvmconfig-end-of-defun) (point)) beg)
	    (goto-char opoint)))
      (if (> beg opoint)
	  (jvmconfig-outline-change opoint (1- beg) ?\^M))

      ;; Show current function
      (jvmconfig-outline-change beg end ?\n)
      ;; Hide nested functions
      (forward-char 1)
      (while (re-search-forward "^\\(function\\|procedure\\)\\>" end 'move)
	(setq opoint (point))
	(jvmconfig-end-of-defun)
	(jvmconfig-outline-change opoint (point) ?\^M))

      (goto-char end)
      (setq opoint end)

      ;; Hide all function after current function
      (while (re-search-forward "^\\(function\\|procedure\\)\\>" nil 'move)
	(jvmconfig-outline-change opoint (1- (match-beginning 0)) ?\^M)
	(setq opoint (point))
	(jvmconfig-end-of-defun))
      (jvmconfig-outline-change opoint (point-max) ?\^M)

      ;; Hide main program
      (if (< (progn (forward-line -1) (point)) end)
	  (progn
	    (goto-char beg)
	    (jvmconfig-end-of-defun)
	    (backward-sexp 1)
	    (jvmconfig-outline-change (point) (point-max) ?\^M))))))

(defun jvmconfig-outline-next-defun ()
  "Move to next function/procedure, hiding all others."
  (interactive)
  (jvmconfig-end-of-defun)
  (jvmconfig-hide-other-defuns))

(defun jvmconfig-outline-prev-defun ()
  "Move to previous function/procedure, hiding all others."
  (interactive)
  (jvmconfig-beg-of-defun)
  (jvmconfig-hide-other-defuns))

(defun jvmconfig-outline-goto-defun ()
  "Move to specified function/procedure, hiding all others."
  (interactive)
  (jvmconfig-goto-defun)
  (jvmconfig-hide-other-defuns))

(provide 'jvmconfig)

;;; jvmconfig.el ends here
