#
# moduleList.awk
# Processes header files to figure out the order of module initialization
#
# Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this file
# for any purpose with or without restriction is hereby granted.
#

#
# A simple AWK script that generates a C header and source file pair for
# defining the list of modules in the system.
#
# usage: awk -f moduleList.awk <variable prefix> <output header file> *.h
#
# Specifically, the generated header file has definitions for identifier
# symbols for each modules, and the source file defines an ordered array used
# to determine how/when to initialize the modules.  The script actually works
# by parsing a bunch of header files in the system which contain special
# comments of the form:
#
# /*
#  * <Directive name>: <values>
#  */
#
# Where each directive occupies a single line with the available directives
# being:
#
#  AddModule - Adds a module to the system.  This needs to be the first
#    directive used since it sets up the context for the rest of the
#    directives.  The value is a single string which names the module
#    and also corresponds to the global variable which defines the
#    `module_t' structure.
#
#  Conditional - Causes `#if's testing the given value to be placed around the
#    module in the initializer list.  This allows for an easy way to determine
#    whether to include a module.
#
#  ModuleDependency - Adds dependencies to the current module.  The values
#    are a space separated list of other module names.  The total number
#    of dependencies determines the modules place in the initialization list.
#    This directive is additive so you can use it multiple times.
#
#  Transition - Mark the end of an initialization stage, the value is used to
#    name the transition from the previous stage to the next.  This is mostly
#    useful if a module needs to be initialized in stages, where one stage has
#    fewer dependencies than subsequent ones.  The modules that a stage depends
#    on are determined by the placement of `ModuleDependency' directives
#    between the `AddModule' (which adds an implicit stage) and other `Stage'
#    directives.  Stages also introduce new names that can be used for
#    `ModuleDependency' directives.  So, to depend on a certain stage in
#    another module, you simply append `.<stage name>' to that module name, the
#    last stage of a module is simply the module's name.
#
#  Implementation - Marks the end of an initialization stage like transition,
#    except that it will automatically add dependencies to the next stage
#    in the module.  The implicit dependencies added are any stages that
#    depend on this implementation stage.  For example, the classPath module
#    defines a interface for interacting with individual entries in the class
#    path, however, the implementations for these entries (jars, directories,
#    etc...) are in other modules.  So the classPath module uses an
#    implementation stage that the other modules depend on, thus the
#    classPath module will automatically depend on its various implementors
#    without having to enumerate them in the module definition.
#
#  Comment - Adds a comment to the generated output, helpful for making sure
#    the generated file has everything in the correct order.
#
# Example usage:
#
# /*
#  * AddModule: utfModule
#  *   ModuleDependency: gcModule
#  */
#
#
# (Its not the greatest thing in the world, but it works)
#


#
# Module related variables:
#
#  mod_count - Number of modules
#  mod_current - Index of current module
#
#  mod_names[] - Names of the modules
#  mod_mid[] - Symbol and value used to make a `#define' in the header
#  mod_cond[] - Conditionals for the module
#
#  mod_curr_stage[] - Current module stage, only used when printing
#

#
# Stage related variables:
#
#  stage_count - Total number of stages
#  stage_current - Current stage
#
#  stage_names[] - Stage names
#  stage_mod[] - Index of module this stage belongs to
#  stage_depcount[] - Number of dependencies for each stage
#  stage_deps[,] - Dependencies for each stage
#  stage_sorted[] - List of stage indexes ordered from the smallest number
#    of dependencies to the largest
#  stage_comments[] - Any comments for this stage
#  stage_impl[] - Indicates whether its an implementation stage
#

#
# Misc. variables:
#
#  header_file - The name of the header file we should output too.
#


# Generate a header for the output file
BEGIN {
	if( ARGC < 4 )
	{
		printf("Usage: awk -f moduleList.awk <variable prefix> <output header file> <header file 1> ... <header file n>.\n") > "/dev/stderr";
		exit -1;
	}
	FS = "[[:blank:]]+";
	# Grab the header file name and null it out,
	# so we don't try to parse it
	prefix_name = ARGV[1];
	ARGV[1] = "";
	header_file = ARGV[2];
	ARGV[2] = "";
	mod_count = 0;
	stage_count = 0;
	current_line = 0;
}

# Add a module to the mix
/\*[[:blank:]]*AddModule\:/ {
	if( NF == 4 )
	{
		addModule($4);
	}
	else
	{
		parseError("Malformed `AddModule' directive", -1);
	}
}

# Add dependencies to the module (Needs to be other modules)
/\*[[:blank:]]*ModuleDependency\:/ {
	if( NF >= 4 )
	{
		for( lpc = 4; lpc <= NF; lpc++ )
		{
			addStageDependency(stage_current, $lpc);
		}
	}
	else
	{
		parseError("Malformed `ModuleDependency' directive", -1);
	}
}

# Add an #ifdef around the module, so it can be conditionally compiled in
/\*[[:blank:]]*Conditional\:/ {
	cond_list = mod_cond[mod_current];
	for( lpc = 4; lpc <= NF; lpc++ )
	{
		cond_list = sprintf("%s %s", cond_list, $lpc);
	}
	mod_cond[mod_current] = cond_list;
}

# End the previous initialization stage of a module
/\*[[:blank:]]*Transition\:/ {
	if( NF == 4 )
	{
		addStage($4, 1);
	}
	else
	{
		parseError("Malformed `Transition' directive", -1);
	}
}

# End the previous initialization stage and mark it as an implementation
/\*[[:blank:]]*Implementation\:/ {
	if( NF == 4 )
	{
		addImpl($4, 1);
	}
	else
	{
		parseError("Malformed `Implementation' directive", -1);
	}
}

# Add a comment to the stage.
/\*[[:blank:]]*Comment\:/ {
	if( NF >= 4 )
	{
		tmp = $0;
		gsub(/\*[[:blank:]]*Comment\:/, "", tmp);
		addStageComment(stage_current, tmp);
	}
}

# Sort the modules, and actually print something useful out
END {
	if( mod_count == 0 )
	{
		printf("No modules found! (Perhaps you're using an incorrect version of make?)\n") > "/dev/stderr";
		exit -1;
	}
	propagateStages();
	sortStages();
	
	printf("/*\n");
	printf(" * Automatically generated by moduleList.awk\n");
	printf(" * so... DO NOT EDIT THIS!\n");
	printf(" */\n");
	printf("\n");
	printf("#include \"module.h\"\n");
	printf("\n");
	
	for( lpc = 0; lpc < mod_count; lpc++ )
	{
		printf("extern module_t %s;\n", mod_names[lpc]);
	}
	printf("\n");
	printf("/*\n");
	printf(" * Stage dependencies -\n");
	printf(" *\n");
	for( lpc = 0; lpc < stage_count; lpc++ )
	{
		sort_index = stage_sorted[lpc];
		mod_index = stage_mod[sort_index];
		printf(" * %s(count=%d):\n",
		       stage_names[sort_index],
		       stage_depcount[sort_index]);
		for( lpc2 = 0; lpc2 < stage_depcount[sort_index]; lpc2++ )
		{
			printf(" *   %s\n", stage_deps[sort_index, lpc2]);
		}
		printf(" *\n");
	}
	printf(" */\n");
	printf("moduleInitializer_t %sModuleInitList[] = {\n", prefix_name);
	for( lpc = 0; lpc < stage_count; lpc++ )
	{
		sort_index = stage_sorted[lpc];
		mod_index = stage_mod[sort_index];
		if( stage_comments[sort_index] )
		{
			printf("\t/*\n");
			printf("%s", stage_comments[sort_index]);
			printf("\t */\n");
		}
		if( mod_cond[mod_index] != "" )
			printf("#if%s\n", mod_cond[mod_index]);
		printf("\t{ &%s, %d },\n",
		       mod_names[mod_index],
		       mod_curr_stage[mod_index]);
		mod_curr_stage[mod_index]++;
		if( mod_cond[mod_index] != "" )
			printf("#endif /*%s */\n", mod_cond[mod_index]);
	}
	printf("\t{ 0, 0 }\n};\n");
	printf("\n");
	printf("module_t *%sModuleList[] = {\n", prefix_name);
	for( lpc = 0; lpc < mod_count; lpc++ )
	{
		if( mod_cond[lpc] != "" )
			printf("#if%s\n", mod_cond[lpc]);
		printf("\t&%s,\n", mod_names[lpc]);
		if( mod_cond[lpc] != "" )
			printf("#endif /*%s */\n", mod_cond[lpc]);
	}
	printf("\t0\n};\n");
	
	printf("/*\n") > header_file;
	printf(" * Automatically generated by moduleList.awk\n") > header_file;
	printf(" * so... DO NOT EDIT THIS MOTHER FUCKER!\n") > header_file;
	printf(" */\n") > header_file;
	printf("\n") > header_file;
	printf("#ifndef _%s_module_list_h\n", prefix_name) > header_file;
	printf("#define _%s_module_list_h\n", prefix_name) > header_file;
	printf("\n") > header_file;
	printf("enum {\n") > header_file;
	for( lpc = 0; lpc < mod_count; lpc++ )
	{
		if( mod_cond[lpc] != "" )
			printf("#if%s\n", mod_cond[lpc]) > header_file;
		printf("\t%s,\n", mod_mid[lpc]) > header_file;
		if( mod_cond[lpc] != "" )
			printf("#endif /*%s */\n",
			       mod_cond[lpc]) > header_file;
	}
	printf("\tMID_MAX\n") > header_file;
	printf("};\n") > header_file;
	printf("\n") > header_file;
	if( mod_cond[stage_mod[0]] != "" )
		printf("#if%s\n", mod_cond[stage_mod[0]]) > header_file;
	printf("enum {\n") > header_file;
	last_mod = stage_mod[0];
	for( lpc = 0; lpc < stage_count; lpc++ )
	{
		if( last_mod != stage_mod[lpc] )
		{
			printf("};\n") > header_file;
			if( mod_cond[mod_index] != "" )
				printf("#endif /*%s */\n",
				       mod_cond[mod_index]) > header_file;
			if( mod_cond[stage_mod[lpc]] != "" )
				printf("\n#if%s",
				       mod_cond[stage_mod[lpc]]) > header_file;
			printf("\nenum {\n") > header_file;
		}
		mod_index = stage_mod[lpc];
		last_mod = mod_index;
		gsub(/\./, "_", stage_names[lpc]);
		printf("\tSID_%s,\n",
		       toupper(stage_names[lpc])) > header_file;
	}
	printf("};\n") > header_file;
	if( mod_cond[mod_index] != "" )
		printf("#endif /*%s */\n", mod_cond[mod_index]) > header_file;
	printf("\n") > header_file;
	printf("extern moduleInitializer_t %sModuleInitList[];\n",
	       prefix_name) > header_file;
	printf("extern module_t *%sModuleList[];\n",
	       prefix_name) > header_file;
	printf("\n") > header_file;
	printf("#endif /* _%s_module_list_h */\n", prefix_name) > header_file;
}

# Tell the user about a parse error
function parseError(msg, errval)
{
	printf("%s:%d - Parse error: %s\n",
	       ARGV[ARGIND],
	       FNR,
	       msg) > "/dev/stderr";
	if( errval != 0 )
		exit errval;
}

# Tell the user about some dependency/whatever error
function typeError(msg, errval)
{
	printf("Error: %s\n", msg) > "/dev/stderr";
	if( errval != 0 )
		exit errval;
}

# Add a module with the given name
function addModule(name)
{
	mod_current = mod_count;
	mod_count++;
	mod_names[mod_current] = name;
	mod_mid[mod_current] = sprintf("MID_%s", toupper(name));
	mod_cond[mod_current] = "";
	mod_curr_stage[mod_current] = 0;
	addStage("", 0);
}

# Add a stage to the current module with the given name.  If prev != 0 then
# this is a user defined stage, so we have to bubble up the implicit module
# stage, and add a dependency for the previous stage.
function addStage(name, prev,  temp)
{
	stage_current = stage_count;
	stage_count++;
	stage_mod[stage_current] = mod_current;
	if( name != "" )
		stage_names[stage_current] = sprintf("%s.%s",
						     mod_names[mod_current],
						     name);
	else
		stage_names[stage_current] = mod_names[mod_current];
	if( prev == 0 )
	{
		stage_depcount[stage_current] = 0;
	}
	else
	{
		temp = stage_names[stage_current - 1];
		stage_names[stage_current - 1] = stage_names[stage_current];
		stage_names[stage_current] = temp;
		stage_depcount[stage_current] = 1;
		stage_deps[stage_current, 0] = stage_names[stage_current - 1];
	}
}

# Add an implementation stage
function addImpl(name, prev,  temp)
{
	addStage(name, prev);
	stage_impl[stage_current - 1] = 1;
}

# Find a stage with the given name and return its index
function findStage(name,  lpc)
{
	for( lpc = 0; lpc < stage_count; lpc++ )
	{
		if( stage_names[lpc] == name )
			return lpc;
	}
	return -1;
}

# Find a dependency with the given name in the given stage
function findStageDependency(curr, name,  lpc)
{
	for( lpc = 0; lpc < stage_depcount[curr]; lpc++ )
	{
		if( stage_deps[curr, lpc] == name )
			return lpc;
	}
	return -1;
}

# Add a comment to this stage
function addStageComment(curr, comment)
{
	stage_comments[curr] = sprintf("%s\t *%s\n",
				       stage_comments[curr],
				       comment);
}

# Add a dependency with the given name in the given stage.  The dependency
# is only added if it wasn't already in there.  The function returns 1, if
# the dependency was added, otherwise its 0.
function addStageDependency(curr, name)
{
	if( (findStageDependency(curr, name) == -1) &&
	    (stage_names[curr] != name) )
	{
		stage_deps[curr, stage_depcount[curr]] = name;
		stage_depcount[curr]++;
		return 1;
	}
	return 0;
}

# Add any implicit dependencies created by implementation stages.
function addInterStageDependencies(curr,  lpc)
{
	for( lpc = 0; lpc < stage_count; lpc++ )
	{
		if( stage_mod[curr] != stage_mod[lpc] )
		{
			if( findStageDependency(lpc, stage_names[curr]) != -1 )
			{
				if( ((curr + 1) < stage_count) &&
				    (stage_mod[curr] == stage_mod[curr + 1]) )
				{
					addStageDependency(curr + 1,
							   stage_names[lpc]);
				}
			}
		}
	}
}

# Sort the stages by the number of dependencies
function sortStages(  i, j, sort_count, curr_min, min_idx, lpc)
{
	sort_count = 0;
	for( i = 0; i < stage_count; i++ )
	{
		curr_min = 1024;
		for( j = 0; j < stage_count; j++ )
		{
			if( stage_depcount[j] < curr_min )
			{
				curr_min = stage_depcount[j];
				min_idx = j;
			}
		}
		stage_sorted[sort_count] = min_idx;
		stage_depcount[min_idx] += 1025;
		sort_count++;
	}
	for( i = 0; i < stage_count; i++ )
	{
		stage_depcount[i] -= 1025;
	}
}

# Propagate the dependencies of an ancestors to the given stage.
#
# XXX Bleh, this is pretty bad code.
function propagateStageDependencies(curr,  i, j, dep, retval)
{
	retval = 0;
	for( i = 0; i < stage_depcount[curr]; i++ )
	{
		dep = findStage(stage_deps[curr, i]);
		if( dep != -1 )
		{
			for( j = 0; j < stage_depcount[dep]; j++ )
			{
				retval += addStageDependency(curr,
							     stage_deps[dep,
								       j]);
			}
		}
		else
		{
			typeError(sprintf("Module(.stage) %s depends on module(.stage) %s, which doesn't exist!\n",
					  stage_names[curr],
					  stage_deps[curr, i]),
				  -1);
		}
	}
	return retval;
}

# Propagate all dependencies to their descendents until we can't do it no
# more.
#
# XXX This is truly awful
function propagateStages(  lpc, change)
{
	for( lpc = 0; lpc < stage_count; lpc++ )
	{
		if( stage_impl[lpc] )
		{
			addInterStageDependencies(lpc);
		}
	}

	do {
		change = 0;
		for( lpc = 0; lpc < stage_count; lpc++ )
		{
			change += propagateStageDependencies(lpc);
		}
	} while(change);
}
