/*
 * classGroup.c
 * Manages a group of classes for a team.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "assert_pp.h"

#include "gtypes.h"
#include "thread.h"
#include "stringSupport.h"
#include "classMethod.h"
#include "access.h"
#include "classGroup.h"
#include "mungWallProxy.h"
#include "module.h"
#include "logResource.h"

#include <misc/objectStack.h>

/* Stores the string `identity' */
static Utf8Const *utfIdentity;

static viewGroupType identityType;

/** Module declarations **/
static int initClassGroupModule(int stage);
static int testClassGroupModule(void);

module_t classGroupModule = {
	"classgroup",
	0,
	initClassGroupModule,
	0, /* Kill */
	testClassGroupModule
};

/** Module functions **/

int initClassGroupModule(int stage)
{
	int retval = 0;

	if( (utfIdentity = utf8ConstNew("identity", -1)) )
	{
		identityType.name = utfIdentity;
		retval = 1;
	}
	else
	{
		fprintf(stderr, "Unable to allocate UTF string\n");
	}
	return( retval );
}

int testClassGroupModule(void)
{
	int retval = 1;
	
	return( retval );
}

/** classGroup definitions **/

/* Add a view to the viewGroup for the given class */
static
classView *addClassView(viewGroup *vg, struct Hjava_lang_Class *cl);
/* Remove a view from the viewGroup for the given class */
static
void remClassView(viewGroup *vg, struct Hjava_lang_Class *cl);
/* Free the contents of a classView */
static
void collectClassView(classView *cv);

/* Structure invariants */
#if !defined(NO_PRECONDITIONS) || !defined(NO_POSTCONDITIONS)
static
int classGroupTypeInvariant(classGroupType *cgt);
static
int viewGroupTypeInvariant(viewGroupType *vgt);
static
int classGroupInvariant(classGroup *cg);
static
int viewGroupInvariant(viewGroup *vg);
#endif

#define DEFAULT_METHOD_TABLE_SIZE 31

#if !defined(NO_PRECONDITIONS) || !defined(NO_POSTCONDITIONS)
int classGroupTypeInvariant(classGroupType *cgt)
{
	require(cgt->name != 0);
	require(cgt->defs != 0);
	
	return( 1 );
}
#endif

classGroupType *createClassGroupType(Utf8Const *name, tagList *tl)
{
	classGroupType *retval;

	require(name != 0);
	require(tl != 0);

	if( (retval = gc_malloc(sizeof(classGroupType),
				GC_ALLOC_CLASS_GROUP)) )
	{
		utf8ConstAddRef(name);
		retval->name = name;
		retval->defs = tl;

		ensure(classGroupTypeInvariant(retval));
	}
	return( retval );
}

void deleteClassGroupType(classGroupType *cgt)
{
	if( cgt )
	{
		require(classGroupTypeInvariant(cgt));
		
		gc_free(cgt);
	}
}

classGroupType *findClassGroupType(classGroupType *head, Utf8Const *name)
{
	classGroupType *curr, *retval = 0;

	require(head != 0);
	require(name != 0);
	
	curr = head;
	while( curr && !retval )
	{
		if( utf8ConstEqual(curr->name, name) )
		{
			retval = curr;
		}
		curr = curr->next;
	}
	return( retval );
}

#if !defined(NO_PRECONDITIONS) || !defined(NO_POSTCONDITIONS)
int viewGroupTypeInvariant(viewGroupType *vgt)
{
	require(vgt->name != 0);
	require(vgt->map != 0);
	
	return( 1 );
}
#endif

viewGroupType *createViewGroupType(Utf8Const *name, tagList *map)
{
	viewGroupType *retval;

	require(name != 0);
	require(map != 0);
	
	if( (retval = gc_malloc(sizeof(viewGroupType), GC_ALLOC_CLASS_GROUP)) )
	{
		utf8ConstAddRef(name);
		retval->name = name;
		retval->map = map;

		ensure(viewGroupTypeInvariant(retval));
	}
	return( retval );
}

void deleteViewGroupType(viewGroupType *vgt)
{
	if( vgt )
	{
		require(viewGroupTypeInvariant(vgt));
		
		gc_free(vgt);
	}
}

#if !defined(NO_PRECONDITIONS) || !defined(NO_POSTCONDITIONS)
int classGroupInvariant(classGroup *cg)
{
	if( cg->parent )
	{
		require(cg->cachedMethods == cg->parent->cachedMethods);
	}
	require(cg->classCount >= 0);
	require(cg->vGroups.data != 0);
	require_valid_memory(cg->vGroups.data);
	require(cg->vGroups.size > 0);
	{
		int lpc;
		
		for( lpc = 0; lpc < cg->vGroups.size; lpc++ )
		{
			require(viewGroupInvariant(&cg->vGroups.data[lpc]));
		}
	}
	require(cg->cachedMethods != 0);
	require_valid_memory(cg->cachedMethods->data);
	require(findViewGroup(cg, utfIdentity) != 0);
	
	return( 1 );
}
#endif

classGroup *createClassGroup(classGroupType *type, classGroup *parent)
{
	classGroup *retval;

	if( (retval = gc_malloc(sizeof(classGroup), GC_ALLOC_CLASS_GROUP)) )
	{
		viewGroupType *vgt = 0;
		int error = 0;
		
		if( parent )
		{
			/* We have a parent group, use their cache */
			retval->parent = parent;
			retval->cachedMethods = parent->cachedMethods;
		}
		else if( (retval->cachedMethods =
			  gc_malloc(sizeof(methodTable),
				    GC_ALLOC_CLASS_GROUP)) )
		{
			/* Need to make our own cache */
			if( (retval->cachedMethods->data =
			     gc_malloc(sizeof(struct _methods *) *
				       DEFAULT_METHOD_TABLE_SIZE,
				       GC_ALLOC_CLASS_GROUP)) )
			{
				retval->cachedMethods->size =
					DEFAULT_METHOD_TABLE_SIZE;
			}
			else
			{
				gc_free(retval->cachedMethods);
				error = 1;
			}
		}
		else
		{
			error = 1;
		}
		retval->type = type;
		/* Add the primary view */
		error = error || !addViewGroup(retval, &identityType);
		if( type )
			vgt = type->views;
		while( !error && vgt )
		{
			error = !addViewGroup(retval, vgt);
			vgt = vgt->next;
		}
		if( error )
		{
			/* Something went awry */
			gc_free(retval);
			retval = 0;
		}
		else
		{
			ensure(classGroupInvariant(retval));
		}
	}
	return( retval );
}

void deleteClassGroup(classGroup *cg)
{
	if( cg )
	{
		int lpc;

		require(classGroupInvariant(cg));
		
		/* Remove ourself if we were exported */
		if( cg->cgl )
		{
			remClassGroup(cg);
		}
		/* Free any view groups */
		for( lpc = cg->vGroups.size - 1; lpc >= 0; lpc-- )
		{
			remViewGroup(&cg->vGroups.data[lpc]);
		}
		gc_free(cg->vGroups.data);
		if( !cg->parent )
		{
			gc_free(cg->cachedMethods->data);
			gc_free(cg->cachedMethods);
		}
		gc_free(cg);
	}
}

void checkClassGroup(classGroup *cg)
{
	if( cg )
	{
		checkClassGroup(cg->parent);
		if( cg->classCount < 0 )
		{
			fprintf(stderr,
				"Class group %p has negative classCount\n",
				cg);
		}
		if( cg->vGroups.size >= 0 )
		{
			int lpc;

			for( lpc = 0; lpc < cg->vGroups.size; lpc++ )
			{
				checkViewGroup(&cg->vGroups.data[lpc]);
			}
		}
		else
		{
			fprintf(stderr,
				"Class group %p has negative vGroups.size\n",
				cg);
		}
		if( cg->cachedMethods )
		{
		}
		else
		{
			fprintf(stderr,
				"Class group %p has no methods cache\n",
				cg);
		}
	}
}

void setClassGroupParent(classGroup *cg, classGroup *parent)
{
	struct _methods *meth, *next;
	methodTable *old_cache;
	int lpc;

	require(cg != 0);
	require(cg->parent == 0);
	require(parent != 0);

	old_cache = cg->cachedMethods;
	cg->parent = parent;
	cg->cachedMethods = parent->cachedMethods;
	/* Add any cached methods to the parent */
	for( lpc = 0; old_cache->size; lpc++ )
	{
		meth = old_cache->data[lpc];
		while( meth )
		{
			next = meth->next;
			cacheGroupMethod(meth);
			meth = next;

			ensure(findGroupMethod(parent,
					       METHOD_CODE_START(meth)) != 0);
		}
	}
	gc_free(old_cache->data);
	gc_free(old_cache);
}

static int findFreeSlot(classGroup *cg)
{
	int lpc, retval;
	viewGroup *vg;

	require(cg != 0);
	require(viewGroupInvariant(&cg->vGroups.data[0]));

	vg = &cg->vGroups.data[0];
	retval = vg->cViews.size;
	for( lpc = 0; lpc < cg->classCount; lpc++ )
	{
		if( vg->cViews.data[lpc].cl == NULL )
		{
			retval = lpc;
			break;
		}
	}
	return retval;
}

int addGroupClass(classGroup *cg, struct Hjava_lang_Class *cl)
{
	int lpc, retval = 1;

	require(cg != 0);
	require(classGroupInvariant(cg));
	require(cl != 0);
	
	/* Make sure it doesn't belong to another group */
	if( cl->cGroup )
		remGroupClass(cl);
	/* Update the class's binding */
	cl->cGroup = cg;
	cl->viewSlot = findFreeSlot(cg);
	/* Cache any jitted methods */
	for( lpc = 0; lpc < cl->nmethods; lpc++ )
	{
		/* XXX Need a better way to detect when to cache */
		if( METHOD_JITTED(&cl->methods[lpc]) &&
		    METHOD_NATIVECODE(&cl->methods[lpc]) )
		{
			cacheGroupMethod(&cl->methods[lpc]);
		}
	}
	/* Add a new slot in every view for the class */
	for( lpc = 0; retval && (lpc < cg->vGroups.size); lpc++ )
	{
		if( !addClassView(&cg->vGroups.data[lpc], cl) )
			retval = 0;
	}
	if( retval )
	{
		/* We're good, update the count */
		cg->classCount++;
	}
	else
	{
		/* Remove whatever views we added */
		for( lpc-- ; lpc >= 0; lpc-- )
		{
			viewGroup *vg;

			vg = &cg->vGroups.data[lpc];
			if( vg->type->map )
			{
				collectClassView(&vg->cViews.
						 data[cl->viewSlot]);
			}
			remClassView(&cg->vGroups.data[lpc], cl);
		}
		/* Clear the class's binding */
		cl->cGroup = 0;
		cl->viewSlot = -1;
	}

	ensure(classGroupInvariant(cg));
	
	return( retval );
}

void remGroupClass(struct Hjava_lang_Class *cl)
{
	classGroup *cg;
	int lpc;

	require(cl != 0);
	require(cl->cGroup != 0);
	require(classGroupInvariant(cl->cGroup));
	require(cl->viewSlot >= 0);
	require(cl->viewSlot < cl->cGroup->vGroups.data[0].cViews.size);

	cg = cl->cGroup;
	/* Remove each view of the class */
	for( lpc = 0; lpc < cg->vGroups.size; lpc++ )
	{
		remClassView(&cg->vGroups.data[lpc], cl);
	}
	/* Flush any cached methods in the class */
	for( lpc = 0; lpc < cl->nmethods; lpc++ )
	{
		if( METHOD_JITTED(&cl->methods[lpc]) &&
		    METHOD_CODE_START(&cl->methods[lpc]) )
		{
			flushGroupMethod(&cl->methods[lpc]);
		}

		ensure(cl->methods[lpc].next == 0);
	}
	/* Clear the class's bindings */
	cl->cGroup = 0;
	cl->viewSlot = -1;
	cg->classCount--;

	ensure(classGroupInvariant(cg));
}

/* This is a bit homemade.  We need a 7-bit hash from the address here */
KAFFE_INLINE
unsigned int hashCodeAddress(classGroup *cg, void *pc)
{
	unsigned int retval;
	
	require(cg != 0);
	require(pc != 0);
	
	retval = ((((uintp)pc) >> 2) ^ ((uintp)pc) >> 9) %
		cg->cachedMethods->size;

	ensure(retval < cg->cachedMethods->size);
	return( retval );
}

void cacheGroupMethod(struct _methods *meth)
{
	unsigned int hash;
	classGroup *cg;
	int iLockRoot;
	void *pc_key;
	
	require(meth != 0);
	require(meth->class != 0);
	require(meth->class->cGroup != 0);
	require(METHOD_CODE_START(meth) != 0);

	pc_key = METHOD_CODE_START(meth);
	cg = meth->class->cGroup;
	hash = hashCodeAddress(cg, pc_key);
	lockMutex(cg->cachedMethods);
	meth->next = cg->cachedMethods->data[hash];
	cg->cachedMethods->data[hash] = meth;
	unlockMutex(cg->cachedMethods);
}

struct _methods *findGroupMethod(classGroup *cg, void *pc)
{
	struct _methods *retval = 0;
	Collector *c;

	require(cg != 0);
	require(pc != 0);
	
	/*
	 * Need to find the base of the code from the pc
	 * XXX This is redundant wrt findMethodPC.
	 */
	if( (c = gc_get_heap(pc)) )
	{
		void *pc_key;

		if( (pc_key = GC_getObjectBase(c, pc)) )
		{
			struct _methods *curr;
			unsigned int hash;
			int iLockRoot;
			
			/* Try to find a match */
			hash = hashCodeAddress(cg, pc_key);
			lockMutex(cg->cachedMethods);
			curr = cg->cachedMethods->data[hash];
			while( curr && !retval )
			{
				if( METHOD_CODE_START(curr) == pc_key )
					retval = curr;
				curr = curr->next;
			}
			unlockMutex(cg->cachedMethods);
		}
	}
	return( retval );
}

void flushGroupMethod(struct _methods *meth)
{
	struct _methods **prev, *curr;
	unsigned int hash;
	classGroup *cg;
	int iLockRoot;
	void *pc_key;

	require(meth != 0);
	require(meth->class != 0);
	require(meth->class->cGroup != 0);

	pc_key = METHOD_CODE_START(meth);
	cg = meth->class->cGroup;
	hash = hashCodeAddress(cg, pc_key);
	lockMutex(cg->cachedMethods);
	prev = &cg->cachedMethods->data[hash];
	curr = cg->cachedMethods->data[hash];
	while( curr && (curr != meth) )
	{
		prev = &curr->next;
		curr = curr->next;
	}

	/*
	 * When xprofiling is available it will flush the method from the table
	 * but it won't null out the method code pointers.  So we have to check
	 * to make sure that we found the method we're trying to flush.
	 * Otherwise we should always find it.
	 */
#if defined(KAFFE_XPROFILER)
	if( curr )
#endif
	{
		require(curr == meth);
		
		*prev = curr->next;
		curr->next = 0;
	}
	unlockMutex(cg->cachedMethods);
}

#if !defined(NO_PRECONDITIONS) || !defined(NO_POSTCONDITIONS)
int viewGroupInvariant(viewGroup *vg)
{
	int lpc;
	
	require(vg->type != 0);
	require(vg->cGroup != 0);
	require_valid_memory(vg->cViews.data);
	require(vg->cViews.size >= 0);
	require(vg->cViews.max >= 0);
	require(vg->cViews.size <= vg->cViews.max);

	for( lpc = 0; lpc < vg->cViews.size; lpc++ )
	{
		if( vg->cViews.data[lpc].cl != NULL )
		{
			require(vg->cViews.data[lpc].cl->viewSlot <
				vg->cViews.size);
		}
	}
	return( 1 );
}
#endif

int addViewGroup(classGroup *cg, viewGroupType *vgt)
{
	viewGroup *new_data;
	int retval = 0;

	require(cg != 0);
	require(vgt != 0);
	
	/* Add a new slot in the array for the new view group */
	if( (new_data = gc_realloc(cg->vGroups.data,
				   sizeof(viewGroup) * (cg->vGroups.size + 1),
				   GC_ALLOC_CLASS_GROUP)) )
	{
		viewGroup *vg;
		int slot;

		/* Store the new array, and get the viewGroup object */
		cg->vGroups.data = new_data;
		slot = cg->vGroups.size++;
		vg = &cg->vGroups.data[slot];
		/* Fill in the object */
		vg->type = vgt;
		vg->cGroup = cg;
		retval = 1;

		ensure(viewGroupInvariant(vg));
	}
	return( retval );
}

void remViewGroup(viewGroup *vg)
{
	int lpc, slot;

	require(vg != 0);
	require(vg->cGroup != 0);
	
	if( vg->type->map )
	{
		/*
		 * This is a mapped view, we have to free the classViews,
		 * their methodViews, and their signatures.
		 */
		for( lpc = 0; lpc < vg->cViews.size; lpc++ )
		{
			collectClassView(&vg->cViews.data[lpc]);
		}
	}
	else
	{
		/*
		 * This is the primary view, since all of the views are just
		 * copies from whats in the classes, just free the array.
		 */
	}
	gc_free(vg->cViews.data);
	/* Fill the hole left by the removal of this group */
	slot = vg - vg->cGroup->vGroups.data;
	for( lpc = slot; lpc < (vg->cGroup->vGroups.size - 1); lpc++ )
	{
		vg->cGroup->vGroups.data[lpc] =
			vg->cGroup->vGroups.data[lpc + 1];
	}
	vg->cGroup->vGroups.size--;
}

void checkViewGroup(viewGroup *vg)
{
	if( vg )
	{
		if( vg->cGroup == 0 )
		{
			fprintf(stderr,
				"View group %p has no class group\n",
				vg);
		}
		if( (vg->cViews.size >= 0) &&
		    (vg->cViews.size < vg->cViews.max) )
		{
		}
		else
		{
			fprintf(stderr,
				"View group %p has bogus class views size, "
				"%d, max %d\n",
				vg,
				vg->cViews.size,
				vg->cViews.max);
		}
	}
}

viewGroup *findViewGroup(classGroup *cg, Utf8Const *name)
{
	viewGroup *retval = 0;
	int lpc;

	require(cg != 0);
	require(name != 0);
	
	for( lpc = 0; !retval && (lpc < cg->vGroups.size); lpc++ )
	{
		if( utf8ConstEqual(cg->vGroups.data[lpc].type->name, name) )
			retval = &cg->vGroups.data[lpc];
	}
	return( retval );
}

struct Hjava_lang_Class *findViewClass(viewGroup *vg, Utf8Const *name)
{
	struct Hjava_lang_Class *retval = 0;
	int lpc;

	require(vg != 0);
	require(name != 0);
	
	for( lpc = 0; !retval && (lpc < vg->cViews.size); lpc++ )
	{
		if( utf8ConstEqual(vg->cViews.data[lpc].name, name) )
			retval = vg->cViews.data[lpc].cl;
	}
	return( retval );
}

static
int partLen(char *name)
{
	int retval;

	for( retval = 0; name[retval] != ';' && name[retval]; retval++ );
	return( retval );
}

struct Hjava_lang_Class *findViewClassPart(viewGroup *vg, char *name)
{
	struct Hjava_lang_Class *retval = 0;
	int lpc, len = 0;

	require(vg != 0);
	require(name != 0);

	while( name[0] == '[' )
		name++;
	switch( name[0] )
	{
	case 'L':
		name++;
		len = partLen(name);
		break;
	case 'V':
		name = "void";
		len = 4;
		break;
	case 'I': 
		name = "int";
		len = 3;
		break;
	case 'Z':
		name = "boolean";
		len = 7;
		break;
	case 'S':
		name = "short";
		len = 5;
		break;
	case 'B':
		name = "byte";
		len = 4;
		break;
	case 'C':
		name = "char";
		len = 4;
		break;
	case 'F':
		name = "float";
		len = 5;
		break;
	case 'D':
		name = "double";
		len = 6;
		break;
	case 'J':
		name = "long";
		len = 4;
		break;
	}
	for( lpc = 0; !retval && (lpc < vg->cViews.size); lpc++ )
	{
		if( !strncmp(vg->cViews.data[lpc].name->data,
			     name,
			     len) )
		{
			retval = vg->cViews.data[lpc].cl;
		}
	}
	return( retval );
}

void addClassGroup(classGroupList *cgl, classGroup *cg)
{
	classGroup *retval;
	int iLockRoot;

	require(cgl != 0);
	require(cg != 0);
	require(cg->type != 0);

	/*
	 * The name of the class group is managed within this functions and
	 * the next.
	 */
	cg->cgl = cgl;
	lockMutex(cgl);
	if( !(retval = findClassGroup(cgl, cg->type->name)) )
	{
		cg->next = cgl->groups;
		cgl->groups = cg;
		retval = cg;
	}
	unlockMutex(cgl);
}

void remClassGroup(classGroup *cg)
{
	classGroupList *cgl = cg->cgl;
	classGroup *curr, **prev;
	int iLockRoot;

	require(cg != 0);
	require(cg->cgl != 0);
	
	lockMutex(cgl);
	prev = &cgl->groups;
	curr = cgl->groups;
	while( curr != cg )
	{
		prev = &curr->next;
		curr = curr->next;
	}
	*prev = curr->next;
	unlockMutex(cgl);
	cg->cgl = 0;
}

classGroup *findClassGroup(classGroupList *cgl, Utf8Const *name)
{
	classGroup *curr, *retval = 0;
	int iLockRoot;

	require(cgl != 0);
	require(name != 0);
	
	lockMutex(cgl);
	curr = cgl->groups;
	while( curr && !retval )
	{
		if( utf8ConstEqual(curr->type->name, name) )
			retval = curr;
		curr = curr->next;
	}
	unlockMutex(cgl);
	return( retval );
}

static
int translateMethodSig(objectStack *os, viewGroup *vg,
		       methodView *src, methodView *dest)
{
	char *start, *end, *pos, *name;
	int len, retval = 1;
	tagItem *ti;
	
	for( pos = (char *)src->parsed_sig->signature->data;
	     *pos && retval;
	     pos++ )
	{
		switch( *pos )
		{
		case 'L':
			/* Object type, need to map */
			pos++;
			start = pos;
			/* Find the end of the type */
			end = strchr(start, ';');
			len = end - start;
			/* Copy a mapped name into the buf */
			if( !vg->type->map )
			{
				name = "";
				len = 0;
			}
			else if( (ti = findTagStr(vg->type->map,
						  start,
						  len,
						  0)) )
			{
				name = (char *)ti->data.str->data;
				len = strlen(name);
			}
			else
			{
				name = start;
				teamLogEntry(
					"classGroup",
					LEA_String, "Warning", "No mapping for class", -1,
					LEA_String, "name", start, len,
					LEA_TAG_DONE);
			}
			retval = pushStackObjects(os,
						  OSA_Char, 'L',
						  OSA_Bytes, name, len,
						  OSA_Char, ';',
						  OSA_TAG_DONE);
			pos = end;
			break;
		default:
			/* Primitive type or syntax */
			retval = pushStackObject(os, pos, 1);
			break;
		}
	}
	retval = retval && pushStackObject(os, pos, 1);
	if( retval )
	{
		char *c_sig;

		if( (c_sig = topStackObject(os, &len)) )
		{
			Utf8Const *new_sig;
			
			if( (new_sig = utf8ConstNew(c_sig, len - 1)) )
			{
				errorInfo einfo;
				
				if( !(dest->parsed_sig =
				      parseSignature(new_sig, &einfo)) )
				{
					teamLogEntry("classGroup",
						     LEA_String,
						     "Error",
						     "Couldn't parse mapped signature",
						     -1,
						     LEA_String,
						     "signature",
						     new_sig->data,
						     -1,
						     LEA_TAG_DONE);
					retval = 0;
				}
			}
			else
			{
				retval = 0;
			}
			popStackObject(os, c_sig, len);
		}
		else
		{
			retval = 0;
		}
	}
	return( retval );
}

/* Translate the method signatures of a class using the given classView */
static
int translateMethods(viewGroup *vg, classView *cv)
{
	objectStack *os;
	int retval = 0;

	require(vg != 0);
	require(cv != 0);
	
	/* Allocate methodView's for each method in the class */
	if( cv->cl->nmethods == 0 )
	{
		retval = 1;
	}
	else if( (cv->methods = gc_malloc(sizeof(methodView) *
					  cv->cl->nmethods,
					  GC_ALLOC_CLASS_GROUP)) &&
		 (os = createObjectStack(1024)) )
	{
		methodView *src, *dest;
		void *frame;
		int lpc;
		
		retval = 1;
		frame = objectStackFrame(os);
		/* Walk over every method in the class */
		for( lpc = 0; retval && (lpc < cv->cl->nmethods); lpc++ )
		{
			/* Map the method from the primary view into ours */
			src = &cv->cl->primaryView.methods[lpc];
			dest = &cv->methods[lpc];
			retval = translateMethodSig(os, vg, src, dest);
			cutObjectStack(os, frame);
		}
		deleteObjectStack(os);
	}
	else
	{
		gc_free(cv->methods);
	}
	return( retval );
}

#define VG_GROWTH 3

classView *addClassView(viewGroup *vg, struct Hjava_lang_Class *cl)
{
	classView *retval = 0;

	require(vg != 0);
	require(viewGroupInvariant(vg));
	require(cl != 0);
	
	/* Get a classView slot */
	if( cl->viewSlot >= vg->cViews.max )
	{
		classView *new_cv;

		if( (new_cv = gc_realloc(vg->cViews.data,
					 sizeof(classView) *
					 (vg->cViews.max + VG_GROWTH),
					 GC_ALLOC_CLASS_GROUP)) )
		{
			vg->cViews.data = new_cv;
			retval = &vg->cViews.data[cl->viewSlot];
			vg->cViews.max += VG_GROWTH;
		}
	}
	else
	{
		retval = &vg->cViews.data[cl->viewSlot];
	}
	if( retval )
	{
		if( vg->type->map )
		{
			tagItem *ti;
			
			retval->cl = cl;
			/* Map the class name */
			if( (ti = findTag(vg->type->map,
					  cl->primaryView.name,
					  0)) )
			{
				retval->name = ti->data.str;
			}
			else
			{
				retval->name = cl->primaryView.name;
			}
			utf8ConstAddRef(retval->name);
			if( !translateMethods(vg, retval) )
			{
				collectClassView(retval);
				retval = 0;
			}
		}
		else if( vg->type->name == utfIdentity )
		{
			require(cl->primaryView.cl != 0);
			require(cl->primaryView.name != 0);
			
			/*
			 * We're is the primary view, just copy the one in the
			 * class
			 */
			*retval = cl->primaryView;
		}
		else
		{
			ensure(0);
		}
		if( retval && (cl->viewSlot == vg->cViews.size) )
			vg->cViews.size = cl->viewSlot + 1;
	}
	return( retval );
}

void remClassView(viewGroup *vg, struct Hjava_lang_Class *cl)
{
	require(vg != 0);
	require(viewGroupInvariant(vg));
	require(cl != 0);

	vg->cViews.data[cl->viewSlot].cl = NULL;
	if( (cl->viewSlot + 1) == vg->cViews.size )
	{
		int lpc;

		for( lpc = vg->cViews.size - 1; lpc >= 0; lpc-- )
		{
			if( vg->cViews.data[lpc].cl == NULL )
				vg->cViews.size--;
			else
				break;
		}
	}
}

void collectClassView(classView *cv)
{
	int lpc;

	require(cv != 0);
	
	utf8ConstRelease(cv->name);
	for( lpc = 0; lpc < cv->cl->nmethods; lpc++ )
	{
		parsed_signature_t *psig;

		psig = cv->methods[lpc].parsed_sig;
		if( psig )
		{
			utf8ConstRelease(psig->signature);
			gc_free(psig);
		}
	}
	gc_free(cv->methods);
}
