/*
 * classGroup.h
 * Manages a group of classes for a team.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _classgroup_h
#define _classgroup_h

/*
 * Class groups are used to track classes that have some loose association
 * with each other.  Either they all belong to the root scope of a name space,
 * or they are exported together to some other name space.
 *
 * XXX The structures in here can be improved quite a bit, they take up more
 * space than they have to.
 */

#include "tags.h"
#include "locks.h"

struct Hjava_lang_Class;
struct _methods;
struct _classGroup;
struct _classGroupList;

/*
 * A methodView is used to map a single method into multiple name spaces.
 *
 * `parsed_sig' - Contains the mapped signature for this method.  The mapping
 *   is determined by the classView and viewGroup it is contained within.
 */
typedef struct _methodView {
	struct _parsed_signature *parsed_sig;
} methodView;

/*
 * A classView is used to map a single class into multiple name spaces.
 *
 * `cl' - The actual class object this view maps to.
 * `name' - The mapped name of the class.
 * `methods' - The array of mapped methods.
 */
typedef struct _classView {
	struct Hjava_lang_Class *cl;
	Utf8Const *name;
	// XXX make this a [0] array and dynamically size the allocation
	methodView *methods;
} classView;

/*
 * A viewGroupType describes the mapping between actual class names and their
 * names in the view.
 *
 * `next' - Link to next type in list.
 * `name' - The name of this view group type.
 * `map' - The tag list that contains the name mappings.  The tag name should
 *   refer to the class name and the tag value should be the name to map it
 *   to.
 */
typedef struct _viewGroupType {
	struct _viewGroupType *next;
	Utf8Const *name;
	tagList *map;
} viewGroupType;

/*
 * A viewGroup is used to map a group of classes into another name space.  For
 * every classGroup there can be one or more viewGroups that provide mappings
 * for every class in the group.
 *
 * `type' - The type of this view group.
 * `cGroup' - The group of classes we are mapping.
 * `cViews.data' - An array of classViews, one for each class in the classGroup
 * `cViews.size' - The number of viewGroup's in `data' that are used.
 * `cViews.max' - The allocated size of the `data' array.
 */
typedef struct _viewGroup {
	struct _viewGroupType *type;
	struct _classGroup *cGroup;
	struct {
		classView *data;
		int size;
		int max;
	} cViews;
} viewGroup;

/*
 * A hash table for tracking the methods of any classes in a group.
 *
 * `data' - The hash array, collisions are chained off of these roots.
 * `size' - The size of the `data' array.
 */
typedef struct _methodTable {
	struct _iLock *lock;
	struct _methods **data;
	int size;
} methodTable;

/*
 * A classGroupType is used to describe the classes and view mappings that
 * can make up a class group.
 *
 * `next' - Link to next type in a list.
 * `name' - The name of this class group.
 * `defs' - The list of classes that make up this group.  The tag names should
 *   be `/' separated class names.
 * `views' - The list of view group types that can be used with this class
 *   group.
 */
typedef struct _classGroupType {
	struct _classGroupType *next;
	Utf8Const *name;
	tagList *defs;
	struct _viewGroupType *views;
} classGroupType;

/*
 * A classGroup is used to track a group of classes, which can then be exported
 * for use by other name spaces.
 *
 * `next' - Link to the next view group in a classGroupList.
 * `parent' - The parent of this class group.  If non-NULL than the group
 *   uses the parent's method table for caching methods.
 * `cgl' - If non-NULL this is the classGroupList this group belongs to.
 * `type' - The type of this class group.
 * `classCount' - The number of classes within this group.
 * `vGroups.data' - An array of viewGroups coupled to this classGroup.
 * `vGroups.size' - The allocated size of the vGroups.data array.
 * `cachedMethods' - The methodTable for this group, if this is a subgroup
 *   then it points to the parent's methodTable.
 */
typedef struct _classGroup {
	struct _classGroup *next;
	struct _classGroup *parent;
	struct _classGroupList *cgl;
	struct _classGroupType *type;
	int classCount;
	struct {
		viewGroup *data;
		int size;
	} vGroups;
	methodTable *cachedMethods;
} classGroup;

/*
 * A classGroupList is a simple structure for tracking exported classGroups.
 *
 * `lock' - Lock to control access to the list.
 * `groups' - A singly linked list of classGroup's.
 */
typedef struct _classGroupList {
	struct _iLock *lock;
	classGroup *groups;
} classGroupList;

/*
 * Create a class group type with the given name and class list.
 *
 * `name' - The name of the class group.  Note: The utf8's reference count is
 *  incremented.
 * `tl' - A TagList with tag identifiers that correspond to the classes that
 *  should be included in the group.  This object is consumed, no reference
 *  is added.
 * returns - A classGroupType object thats been initialized with the given
 *  values.
 */
classGroupType *createClassGroupType(Utf8Const *name, tagList *tl);
/*
 * Delete a class group type.
 *
 * `cgt' - A classGroupType as returned by createClassGroupType.
 */
void deleteClassGroupType(classGroupType *cgt);
/*
 * Find a classGroupType in a singly linked list.
 *
 * `head' - The head of a list of classGroupType's.
 * `name' - The name of the classGroupType to find.
 * returns - The classGroupType with the given name or NULL if it wasn't
 *  found.
 */
classGroupType *findClassGroupType(classGroupType *head, Utf8Const *name);
/*
 * Create a view group type with the given name and mapping.
 *
 * `name' - The name of the view group.  Note: The utf8's reference count is
 *  incremented.
 * `tl' - A TagList with tag identifier that correspond to classes in the
 *  group and values that are the desired mapping.  This object is consumed,
 *  no reference is added.
 * returns - A viewGroupType object thats been initialized with the given
 *  values.
 */
viewGroupType *createViewGroupType(Utf8Const *name, tagList *map);
/*
 * Delete a view group type.
 *
 * `vgt' - A viewGroupType as returned by createViewGroupType.
 */
void deleteViewGroupType(viewGroupType *vgt);
/*
 * Create and initialize a classGroup object.  The returned group will already
 * contain the `identity' viewGroup for storing the primary views of the
 * objects and any other views as described by `type'.
 *
 * `type' - The list of classes that belong in this class group.
 * `parent' - The parent of this class group or NULL if this is a root group.
 *  If a parent group is given then the child group is expected to be a subset
 *  of a larger name space and any method lookups should be shared with it.
 * returns - An empty class group with the given type and the default
 *  `identity' view group.  Note: Classes still need to be added to the group
 *  they aren't automatically loaded in by this function.
 */
classGroup *createClassGroup(classGroupType *type, classGroup *parent);
/*
 * Finalize and deallocate a classGroup object.
 *
 * `cg' - A classGroup as returned by createClassGroup.
 */
void deleteClassGroup(classGroup *cg);
/*
 * Run a sanity check on the class group.
 *
 * `cg' - A classGroup as returned by createClassGroup.
 */
void checkClassGroup(classGroup *cg);
/*
 * Manually, set the parent of a class group. XXX Currently unused and prolly
 * for a good reason...
 */
void setClassGroupParent(classGroup *cg, classGroup *parent);
/*
 * Add a class to a group and any viewGroups the classGroup contains.  The
 * class name and method signatures will be automatically mapped in each
 * view.
 *
 * `cg' - The classGroup to add the class to.
 * `cl' - The class to add to the group.
 * returns - True if the add succeeded, false otherwise.
 */
int addGroupClass(classGroup *cg, struct Hjava_lang_Class *cl);
/*
 * Remove a class from a group and any viewGroups.  XXX It would be unusual to
 * remove a class without immediately adding it to another group, perhaps this
 * should be transferGroupClass?
 *
 * `cl' - The class to remove from its classGroup.
 */
void remGroupClass(struct Hjava_lang_Class *cl);
/*
 * Cache a method in the class group to speed up searches for methods based on
 * a program counter value.  This is primarily used when walking backwards over
 * the stack during exceptions.
 *
 * `meth' - The method to cache in the class group.  Note: The classGroup is
 *  found from the class pointer held in the method object.
 */
void cacheGroupMethod(struct _methods *meth);
/*
 * Flush a method from the group's cache.
 *
 * `meth' - The method to remove from the class group lookup cache.
 */
void flushGroupMethod(struct _methods *meth);
/*
 * Find a method that contains the given pc.
 *
 * `cg' - The classGroup whose cache should be searched.
 * `pc' - The program counter value to search for in the cache.
 * returns - The method that occupies the segment of memory the `pc' points to
 *  or NULL if a match wasn't found.
 */
struct _methods *findGroupMethod(classGroup *cg, void *pc);
/*
 * Add a viewGroup with the given type to the classGroup.
 *
 * `cg' - An initialized classGroup.
 * `vgt' - A viewGroup that contains mappings for the classes in the classGroup
 */
int addViewGroup(classGroup *cg, viewGroupType *vgt);
/*
 * Remove the given viewGroup from its classGroup.
 *
 * `vg' - The viewGroup to remove from its classGroup.
 */
void remViewGroup(viewGroup *vg);
/*
 * Find a viewGroup with the given name.
 *
 * `cg' - The classGroup in which to search for the viewGroup.
 * `name' - The name of the viewGroup.
 * returns - The viewGroup held in the classGroup or NULL if its wasn't found.
 */
viewGroup *findViewGroup(classGroup *cg, Utf8Const *name);
/*
 * Run a sanity check on the view group.
 *
 * `vg' - A viewGroup as returned by createViewGroup.
 */
void checkViewGroup(viewGroup *vg);
/*
 * Find a class in a class group as seen through the given view.
 *
 * `vg' - The viewGroup to search through for a class name mapping.
 * `name' - The mapped class name to search for in the group.
 * returns - The class that matched the given name as seen through the view or
 *  NULL if there was no match.
 */
struct Hjava_lang_Class *findViewClass(viewGroup *vg, Utf8Const *name);
struct Hjava_lang_Class *findViewClassPart(viewGroup *vg, char *name);

/*
 * Publish a classGroup with the given name in the group list.
 *
 * `cgl' - An initialized classGroupList.
 * `cg' - A previously unpublished classGroup.
 */
void addClassGroup(classGroupList *cgl, classGroup *cg);
/*
 * Remove the given classGroup from its classGroupList.
 *
 * `cg' - A published classGroup.
 */
void remClassGroup(classGroup *cg);
/*
 * Find an exported class group.
 *
 * `cgl' - The classGroupList in which to search for the classGroup.
 * `name' - The name of the classGroup to search for.
 * returns - A classGroup matching the given name or NULL if it wasn't found.
 */
classGroup *findClassGroup(classGroupList *cgl, Utf8Const *name);

/*
 * AddModule: classGroupModule
 *   ModuleDependency: utfModule
 *   Comment: Get the utf string `identity'
 */

#endif /* _classgroup_h */
