/*
 * classPath.c
 * Provides access to the class path for the jvm.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdarg.h>

#include "assert_pp.h"

#include "gtypes.h"
#include "system.h"
#include "jsyscall.h"
#include "files.h"
#include "gc.h"
#include "jteam.h"
#include "jteamReservation.h"
#include "module.h"
#include "moduleList.h"
#include "stringSupport.h"
#include "file.h"
#include "itypes.h"
#include "readClass.h"
#include "mungWallProxy.h"
#include "nativeClass.h"
#include "nativeClass-inline.h"
#include "scripting.h"

#include "logResource.h"

#include "classPath.h"
#include "classPath-inline.h"

#include <jsi/jsi.h>
#include <misc/stringParsing.h>

#include "edu_utah_janosvm_resources_ClassPathSpec.h"

#define JAVA_CLASS_PATH_SPEC "edu/utah/janosvm/resources/ClassPathSpec"

/* Name of the default JAR file */
#define KLASSES_JAR	"Klasses.jar"

/* Singly linked list of fileTypeHandlers */
static fileTypeHandler *fileTypeHandlerList;

/* The class path actually being used */
char *realClassPath;

static struct Hjava_lang_Class *stringClass;

static int appendClassPathEntry(classPathResource *cpr, classPathEntry *cpe);
static classPathEntry *findClassPathEntry(classPathResource *cpr, char *path,
					  int grab, int hidden);

/** Module declarations **/
static int initClassPathModule(int stage);

module_t classPathModule = {
	"classpath",
	0,
	initClassPathModule,
	0,
	0
};

/** ClassPathResource declarations **/
static int classPathConfigure(struct _resource *res,
			      sectionFile *sf,
			      tagSectionData *tsd);
static int classPathFillSpec(struct _resource *res,
			     struct _resourceSpec *rs,
			     sectionFile *sf,
			     tagSectionData *tsd);
static struct _resourceSpec *classPathCreateJSISpec(struct _resource *res,
						    jsiMessage *jm);
static void classPathDeleteJSISpec(struct _resource *res,
				   struct _resourceSpec *rs);
static int classPathSubscribe(struct _resource *res,
			      struct _resourceClient *rc,
			      struct _resourceSpec *rs);
static void classPathUnsubscribe(struct _resource *res,
				 struct _resourceClient *rc);

static resourceFunctions classPathResourceFunctions = {
	0,
	0,
	classPathConfigure,
	classPathFillSpec,
	classPathCreateJSISpec,
	classPathDeleteJSISpec,
	0,
	classPathSubscribe,
	0,
	classPathUnsubscribe
};

/** ClassPathResource variables **/

/* JSI handlers/property info's */
static jsiHandlerFunctions cpAddFunctions;
static jsiNamedHandler cpAddHandler;

static char *cpHandleAddMessage(jsiHandlerContext *hc,
				jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions cpHideFunctions;
static jsiNamedHandler cpHideHandler;

static char *cpHandleHideMessage(jsiHandlerContext *hc,
				 jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions cpRemoveFunctions;
static jsiNamedHandler cpRemoveHandler;

static char *cpHandleRemoveMessage(jsiHandlerContext *hc,
				   jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions cpRootFunctions;
static jsiNamedHandler cpRootHandler;

static char *cpHandleRootMessage(jsiHandlerContext *hc,
				 jsiHandler *jh, jsiMessage *jm);
static int cpGetSuites(jsiHandler *jh, jsiMessage *jm);

static jsiPropertyInfo cpProperties[] = {
	{ "add", (1L << JSI_EXECUTE_PROPERTY), "",
	  "path:string - Add an entry to the class path.\n" },
	{ "hide", (1L << JSI_EXECUTE_PROPERTY), "",
	  "path:string - Hide an entry from the class path.\n" },
	{ "remove", (1L << JSI_EXECUTE_PROPERTY), "",
	  "path:string - Remove an entry from the class path.\n" },
	{ 0 }
};

static jsiPropertyInfo cpRootProperties[] = {
	{ "classpath", (1L << JSI_LIST_PROPERTIES), "",
	  "Get the list of properties contained within the class path.\n" },
	{ "classpath", (1L << JSI_GET_PROPERTY), "",
	  "Get the list of class path entries.\n" },
	{ 0 }
};

classPathResource metalClassPathResource;

static tagDesc classPathResourceTags[] = {
	{ 0, TAG_NONE }
};

static tagData classPathResourceTagData[1];

static tagArray classPathResourceDefault = {
	classPathResourceTags,
	classPathResourceTagData
};

/*
 * These are used to capture a class path and associated principal from a
 * section file.
 */
static tagDesc classPathSpecTags[] = {
	{ "path", TAG_STRING },
	{ 0, TAG_NONE }
};

static tagData classPathSpecTagData[1];

static tagArray classPathSpecDefault = {
	classPathSpecTags,
	classPathSpecTagData
};

/** Helper declarations **/

/*
 * Parse a formatted class path and add the entries to the resource or client,
 * depending on which is non-NULL.
 */
static int parseClassPath(classPathResource *cpr,
			  classPathClient *cpc,
			  char *classpath);
/*
 * Compute a class path by scanning the given directory and then placing any
 * matching entries in the resource.
 */
static int discoverClassPath(classPathResource *cpr, char *home);

/** Module definitions **/

/* Fill in any default tag data */
static int fillTagDefaults(void)
{
	int retval = 0;

	if( (classPathSpecTagData[0].str = utf8ConstNew("", -1)) )
	{
		retval = 1;
	}
	else
	{
		fprintf(stderr, "Unable to allocate utf strings.\n");
	}
	return( retval );
}

/* Get the default class path, either given to us, or compute our own */
static int subscribeDefaultPath(void)
{
	char *defaultPath;
	int retval = 0;

	/* Get the default path */
	defaultPath = (char *)Kaffe_JavaVMArgs[0].classpath;
	if( defaultPath && defaultPath[0] )
	{
		/*
		 * Good, it was given, just parse it and add it to the
		 * resource.
		 */
		retval = parseClassPath(&metalClassPathResource,
					0,
					defaultPath);
		if( !retval )
		{
			fprintf(stderr,
				"The following class path is not valid:\n"
				"  %s\n",
				defaultPath);
		}
	}
	/* No default path or it was empty, try to discover a path */
	if( retval && (metalClassPathResource.entries == 0) )
	{
		/* No path given, try to figure it out ourselves */
		retval = discoverClassPath(
			&metalClassPathResource,
			(char*)Kaffe_JavaVMArgs[0].classhome);
	}
	/* If it worked, subscribe to the resource */
	if( retval )
	{
		classPathSpec cps;
		
		cps.spec.handler = &metalClassPathResource.link;
		cps.classpath = realClassPath;
		retval = attachTeamResource(metalTeam, &cps.spec);
	}
	if( !retval )
	{
		// XXX exception or abort here?
		fprintf(stderr, "Unable to build a valid classpath.\n");
	}
	return( retval );
}

static int initClassPathModule(int stage)
{
	int retval = 0;

	switch( stage )
	{
	case SID_CLASSPATHMODULE_HANDLERS:
		/*
		 * This is a dummy stage so that the fileTypeHandler modules
		 * can depend on it.  And, since our next stage will depend
		 * on them we can be sure that they're all set up for us to
		 * use.
		 */
		retval = 1;

		/* Add the JSI handlers... */
		jsiInitNamedHandler(
			&cpRootHandler,
			JNA_Name, "classpath",
			JNA_Functions, &cpRootFunctions,
			JNA_HandleMessage, cpHandleRootMessage,
			JNA_GetSuites, cpGetSuites,
			JNA_Child, jsiInitNamedHandler(
				&cpAddHandler,
				JNA_Name, "add",
				JNA_Functions, &cpAddFunctions,
				JNA_HandleMessage, cpHandleAddMessage,
				JNA_TAG_DONE),
			JNA_Child, jsiInitNamedHandler(
				&cpHideHandler,
				JNA_Name, "hide",
				JNA_Functions, &cpHideFunctions,
				JNA_HandleMessage, cpHandleHideMessage,
				JNA_TAG_DONE),
			JNA_Child, jsiInitNamedHandler(
				&cpRemoveHandler,
				JNA_Name, "remove",
				JNA_Functions, &cpRemoveFunctions,
				JNA_HandleMessage, cpHandleRemoveMessage,
				JNA_TAG_DONE),
			JNA_TAG_DONE);
		jsiInstallHandler(&jvmHandlerList, &cpRootHandler);
		break;
	case SID_CLASSPATHMODULE_INITMETAL:
		installResource(&metalClassPathResource.link,
				RA_Name, "classpath",
				RA_Flags, RF_INTERNAL,
				RA_Functions, &classPathResourceFunctions,
				RA_ClientSize, sizeof(classPathClient),
				RA_SpecSize, sizeof(classPathSpec),
				RA_TAG_DONE);
		metalClassPathResource.lastEntry =
			&metalClassPathResource.entries;
		retval = subscribeDefaultPath();
		break;
	case SID_CLASSPATHMODULE:
		if( fillTagDefaults() )
		{
			errorInfo einfo;
		    
			installResource(
				&metalClassPathResource.link,
				RA_ResourceSectionName, "classpath-resource",
				RA_SpecSectionName, "classpath-spec",
				RA_ResourceSectionDefault,
				&classPathResourceDefault,
				RA_SpecSectionDefault, &classPathSpecDefault,
				RA_SpecNativeClass, JAVA_CLASS_PATH_SPEC,
				RA_TAG_DONE);
			
			stringClass = lookupClass("java/lang/String",
						  NULL,
						  &einfo);
			retval = 1;
		}
		break;
	}
	return( retval );
}

/** Resource definitions **/

int classPathConfigure(struct _resource *res,
		       sectionFile *sf,
		       tagSectionData *tsd)
{
	classPathResource *cpr = (classPathResource *)res;
	classPathClient *cpc = (classPathClient *)
		getTeamInternalResource(0, &metalClassPathResource.link);
	tagList *tl = tsd->taglist;
	int lpc, retval = 1;
	tagItem *path;

	for( lpc = 0;
	     (path = findTagStr(tl, "path", -1, lpc)) && retval;
	     lpc++ )
	{
		retval = parseClassPath(cpr,
					0,
					(char *)path->data.str->data);
		retval = parseClassPath(0,
					cpc,
					(char *)path->data.str->data);
	}
	return( retval );
}

int classPathFillSpec(struct _resource *res,
		      struct _resourceSpec *rs,
		      sectionFile *sf,
		      tagSectionData *tsd)
{
	classPathSpec *cps = (classPathSpec *)rs;

	cps->classpath = (char *)getTagData(&tsd->tags, "path").str->data;
	return( 1 );
}

resourceSpec *classPathCreateJSISpec(struct _resource *res, jsiMessage *jm)
{
	resourceSpec *retval = 0;
	int count;

	/* Count the number of class path strings that need to be processed */
	count = jsiMessageIdentifierCount(jm, "classpath");
	if( count && (retval = createResourceSpec(res)) )
	{
		classPathSpec *cps = (classPathSpec *)retval;
		int len = 1, lpc, error = 0;
		char *path, *new_path;
		
		for( lpc = 0;
		     !error &&
		     jsiMessageGet(jm,
				   JMA_String, "classpath", lpc, &path,
				   JMA_TAG_DONE);
		     lpc++ )
		{
			/* Append separator and the path */
			len += strlen(path_separator) + strlen(path);
			if( (new_path = gc_realloc(cps->classpath,
						   len,
						   GC_ALLOC_CLASSPATH)) )
			{
				cps->classpath = new_path;
				if( lpc > 0 )
					strcat(cps->classpath, path_separator);
				strcat(cps->classpath, path);
			}
			else
			{
				error = 1;
			}
		}
		if( error )
		{
			gc_free(cps->classpath);
			gc_free(retval);
			retval = 0;
		}
		else
		{
			ensure_valid_memory(cps->classpath);
		}
	}
	return( retval );
}

void classPathDeleteJSISpec(struct _resource *res, struct _resourceSpec *rs)
{
	classPathSpec *cps = (classPathSpec *)rs;

	if( cps->classpath[0] )
	{
		require_valid_memory(cps->classpath);
	}
	
	gc_free(cps->classpath);
	deleteResourceSpec(rs);
}

int classPathSubscribe(struct _resource *res,
		       struct _resourceClient *rc,
		       struct _resourceSpec *rs)
{
	classPathResource *cpr = (classPathResource *)res;
	classPathClient *cpc = (classPathClient *)rc;
	classPathSpec *cps = (classPathSpec *)rs;
	char *clientPath;
	int iLockRoot, retval;
	
	cpc->entries.data = 0;
	cpc->entries.count = 0;
	/* XXX Need to protect realClassPath since we write to the string */
	if( cps == 0 )
		lockMutex(cpr);
	clientPath = cps ? cps->classpath : realClassPath;
	/* Parse the path and add any entries to our client */
	if( !(retval = parseClassPath(0,
				      cpc,
				      clientPath)) )
	{
		teamLogEntry("classpath",
			     LEA_String, "Error", "Couldn't subscribe to class path", -1,
			     LEA_String, "path", clientPath, -1,
			     LEA_TAG_DONE);
	}
	if( cps == 0 )
		unlockMutex(cpr);
	return( retval );
}

void classPathUnsubscribe(struct _resource *res, struct _resourceClient *rc)
{
	classPathResource *cpr = (classPathResource *)res;
	classPathClient *cpc = (classPathClient *)rc;
	int iLockRoot;
	int lpc;

	switchToTeam(kernelTeam, 0);
	lockMutex(cpr);
	/* Drop the references to any class path entries. */
	for( lpc = 0; lpc < cpc->entries.count; lpc++ )
	{
		classPathEntry *cpe = cpc->entries.data[lpc].entry;

		cpe->users--;
		if( (cpe->users == 0) && !(cpe->flags & CPEF_PUBLISHED) )
		{
			/* We're holding the last reference */
			fthDeleteEntry(cpe);
		}
	}
	unlockMutex(cpr);
	returnFromTeam();
}

/** classpath definitions **/

/* Append a new class path entry */
int appendClassPathEntry(classPathResource *cpr, classPathEntry *cpe)
{
	classPathEntry *deadEntry = 0, *dupEntry;
	int retval = 0;
	int iLockRoot;

	require(cpr != 0);
	require(cpe != 0);
	require((currentTeam() == kernelTeam) ||
		(currentTeam() == metalTeam));
	
	lockMutex(cpr);
	if( (dupEntry = findClassPathEntry(cpr, cpe->path, false, true)) )
	{
		/* Already added */
		if( dupEntry->flags & CPEF_HIDDEN )
		{
			require(dupEntry->flags & CPEF_PUBLISHED);
			
			/*
			 * Its hidden, so we need to add the path back to
			 * the realClassPath.
			 */
			deadEntry = cpe;
			cpe = dupEntry;
		}
		else
		{
			fthDeleteEntry(cpe);
			cpe = 0;
			retval = 1;
		}
	}
	if( cpe )
	{
		int real_len, len;
		char *new_cp;

		require_valid_memory(realClassPath);

		/* Append to the class path */
		if( realClassPath )
			real_len = strlen(realClassPath);
		else
			real_len = 0;
		len = real_len +
			strlen(path_separator) +
			strlen(cpe->path) +
			1;
		if( (new_cp = gc_realloc(realClassPath,
					 len,
					 GC_ALLOC_CLASSPATH)) )
		{
			realClassPath = new_cp;
			if( real_len == 0 )
				realClassPath[0] = '\0';
			else
				strcat(realClassPath, path_separator);
			strcat(realClassPath, cpe->path);
			if( cpe->flags & CPEF_HIDDEN )
			{
				require(deadEntry);
				
				/*
				 * Its already on the list, just clear the
				 * hidden flag.
				 */
				cpe->flags &= ~CPEF_HIDDEN;
				fthDeleteEntry(deadEntry);
			}
			else
			{
				/* Add to the list of entries */
				cpe->flags |= CPEF_PUBLISHED;
				*cpr->lastEntry = cpe;
				cpr->lastEntry = &cpe->next;
				cpe->next = 0;
			}
			retval = 1;
		}

		ensure_valid_memory(realClassPath);
	}
	unlockMutex(cpr);
	return( retval );
}

void installFileType(fileTypeHandler *fth)
{
	require(fth != 0);
	
	fth->next = fileTypeHandlerList;
	fileTypeHandlerList = fth;
}

void uninstallFileType(fileTypeHandler *fth)
{
	fileTypeHandler *curr, **prev;

	require(fth != 0);
	
	prev = &fileTypeHandlerList;
	curr = fileTypeHandlerList;
	while( curr != fth )
	{
		prev = &curr->next;
		curr = curr->next;
	}
	if( curr )
	{
		*prev = curr->next;
	}
}

static
int findClassPathHandler(classPathResource *cpr, char *path,
			 struct stat *st, int fd, char *buf_ptr)
{
	fileTypeHandler *fth;
	classPathEntry *cpe;
	int retval = 0;
	int error = 0;

	require(cpr != 0);
	require(path != 0);
	require(st != 0);
	
	/* Walk the list of fileTypeHandlers looking for a match */
	fth = fileTypeHandlerList;
	while( fth && !retval && !error )
	{
		if( fthMatchPath(fth, path, st, fd, buf_ptr) )
		{
			if( (cpe = fthCreateEntry(fth, path)) )
			{
				require(cpe->fileType == fth);
				require(cpe->path != 0);
				require(!strcmp(cpe->path, path));
				
				/* Add the entry to the resource */
				if( appendClassPathEntry(cpr, cpe) )
				{
					retval = 1;
				}
				else
				{
					fthDeleteEntry(cpe);
					error = 1;
				}
			}
			else
			{
				error = 1;
			}
		}
		fth = fth->next;
	}
	return( retval );
}

int addClassPathEntry(classPathResource *cpr, char *path)
{
	classPathEntry *cpe;
	int retval = 0;
	struct stat st;

	require((currentTeam() == kernelTeam) ||
		(currentTeam() == metalTeam));
	require(cpr != 0);
	require(path != 0);

	if( (cpe = grabClassPathEntry(cpr, path)) )
	{
		/* The entry already exists */
		dropClassPathEntry(cpr, cpe);
	}
	else if( !KSTAT(path, &st) )
	{
		char *buf_ptr = 0;
		char buf[FTH_FILE_HEADER_SIZE];
		int fd = -1;
		
		/* We've got the file. */
		if( S_ISDIR(st.st_mode) )
		{
			/* Its a directory */
			retval = 1;
		}
		else if( !KOPEN(path, O_RDONLY, 0, &fd) )
		{
			int bytes_left, bytes_read;

			/*
			 * Its a file, try to read in the first few bytes
			 * so that the handlers can match on it.
			 */
			buf_ptr = buf;
			bytes_left = FTH_FILE_HEADER_SIZE;
			while( bytes_left &&
			       !KREAD(fd,
				      &buf[FTH_FILE_HEADER_SIZE - bytes_left],
				      bytes_left,
				      &bytes_read) &&
			       bytes_read )
			{
				bytes_left -= bytes_read;
			}
			if( bytes_left == 0 )
			{
				retval = 1;
			}
			else
			{
				teamLogEntry("classPath",
					     LEA_String, "Error",
					     "Couldn't read path header", -1,
					     LEA_String, "path", path, -1,
					     LEA_TAG_DONE);
			}
		}
		else
		{
			teamLogEntry("classPath",
				     LEA_String, "Error",
				     "Couldn't open path entry", -1,
				     LEA_String, "path", path, -1,
				     LEA_TAG_DONE);
		}
		if( retval )
		{
			retval = findClassPathHandler(cpr,
						      path,
						      &st,
						      fd,
						      buf_ptr);
		}
		if( fd != -1 )
			KCLOSE(fd);
	}
	return( retval );
}

/* Structure used to pass args to remParsedClassPathEntry */
struct removeClassPath {
	parsedString entry;
	char *path;
};

/*
 * A parseString handler function that will remove the rcp->path string from
 * the realClassPath.
 */
static
int remParsedClassPathEntry(struct removeClassPath *rcp)
{
	int retval = 1;

	if( cmpPStrStr(&rcp->entry, rcp->path) == 0 )
	{
		/* Found the path in the string, remove it */
		if( rcp->entry.data == realClassPath )
		{
			int sep_len;

			/*
			 * Its at the start, so there isn't a colon to copy
			 * over.
			 */
			if( rcp->entry.data[rcp->entry.len] == 0 )
			{
				realClassPath[0] = 0;
			}
			else
			{
				sep_len = strlen(path_separator);
				memmove(rcp->entry.data,
					rcp->entry.data + rcp->entry.len +
					sep_len,
					strlen(rcp->entry.data +
					       rcp->entry.len +
					       sep_len) +
					1);
			}
		}
		else
		{
			memmove(rcp->entry.data - strlen(path_separator),
				rcp->entry.data + rcp->entry.len,
				strlen(rcp->entry.data +
				       rcp->entry.len) +
				1);
		}
		/* Stop parseString from continuing. */
		retval = 0;
	}
	return( retval );
}

/*
 * Remove the given path from the realClassPath.
 */
static
void removeClassPathString(char *path)
{
	struct removeClassPath rcp;

	require(path != 0);
	
	rcp.path = path;
	/*
	 * Just use parseString to walk the class path and use the do/while
	 * handler function to do the work of overwriting the path.
	 */
	parseString(realClassPath,
		    SPO_Do,
		      SPO_String, &rcp.entry,
		      SPO_While, path_separator, "",
		      SPO_Handle, remParsedClassPathEntry, &rcp,
		      SPO_End,
		    SPO_End);

	ensure_valid_memory(realClassPath);
}

int hideClassPathEntry(classPathResource *cpr, char *path)
{
	classPathEntry *curr;
	int retval = 0;
	int iLockRoot;

	require(cpr != 0);
	require(path != 0);
	require((currentTeam() == kernelTeam) ||
		(currentTeam() == metalTeam));
	
	lockMutex(cpr);
	curr = cpr->entries;
	while( curr && (strcmp(curr->path, path) != 0) )
	{
		curr = curr->next;
	}
	if( curr )
	{
		/*
		 * Found it, mark it as hidden and remove it from the default
		 * class path.
		 */
		curr->flags |= CPEF_HIDDEN;
		removeClassPathString(path);
		retval = 1;
	}
	unlockMutex(cpr);
	return( retval );
}

int remClassPathEntry(classPathResource *cpr, char *path)
{
	classPathEntry **prev, *curr;
	int retval = 0;
	int iLockRoot;

	require(cpr != 0);
	require(path != 0);
	require((currentTeam() == kernelTeam) ||
		(currentTeam() == metalTeam));
	
	lockMutex(cpr);
	curr = cpr->entries;
	prev = &cpr->entries;
	while( curr && (strcmp(curr->path, path) != 0) )
	{
		prev = &curr->next;
		curr = curr->next;
	}
	if( curr )
	{
		/*
		 * Found it, remove it from the default class path string and
		 * yank it out of the global list.
		 */
		removeClassPathString(path);
		*prev = curr->next;
		if( *prev == 0 )
			cpr->lastEntry = prev;
		curr->flags &= ~(CPEF_PUBLISHED|CPEF_HIDDEN);
	}
	unlockMutex(cpr);
	if( curr )
	{
		if( curr->users == 0 )
		{
			fthDeleteEntry(curr);
			retval = 1;
		}
		else
		{
			retval = fthSnapshotEntry(curr);
		}
	}
	return( retval );
}

static
classPathEntry *findClassPathEntry(classPathResource *cpr, char *path,
				   int grab, int hidden)
{
	classPathEntry *retval = 0;
	
	require(cpr != 0);
	require(path != 0);

	if( !kernelTeam || (reserveTeam(kernelTeam) == TEAM_ERROR_NONE) )
	{
		classPathEntry *curr;
		int iLockRoot;
		
		lockMutex(cpr);
		curr = cpr->entries;
		while( curr && !retval )
		{
			if( !strcmp(curr->path, path) &&
			    (!(curr->flags & CPEF_HIDDEN) || hidden) )
			{
				retval = curr;
			}
			curr = curr->next;
		}
		if( retval && grab )
			retval->users++;
		unlockMutex(cpr);
		if( kernelTeam )
			unreserveTeam(kernelTeam);
	}
	return( retval );
}

classPathEntry *grabClassPathEntry(classPathResource *cpr, char *path)
{
	return findClassPathEntry(cpr, path, true, false);
}

void dropClassPathEntry(classPathResource *cpr, classPathEntry *cpe)
{
	require(cpr != 0);
	require(cpe != 0);
	require(cpe->users > 0);
	
	if( !kernelTeam || (reserveTeam(kernelTeam) == TEAM_ERROR_NONE) )
	{
		int iLockRoot;
		
		lockMutex(cpr);
		cpe->users--;
		if( (cpe->users == 0) && !(cpe->flags & CPEF_PUBLISHED) )
		{
			fthDeleteEntry(cpe);
		}
		unlockMutex(cpr);
		if( kernelTeam )
			unreserveTeam(kernelTeam);
	}
}

/* Insert an orderedClassPathEntry into a client */
static
int insertClassPath(classPathClient *cpc, classPathEntry *cpe, int pri)
{
	orderedClassPathEntry *new_entries;
	int retval = 0;

	require(cpc != 0);
	require_valid_memory(cpc->entries.data);
	require(cpe != 0);
	
	/* Resize for the new entry */
	if( (new_entries = gc_realloc(cpc->entries.data,
				      sizeof(orderedClassPathEntry) *
				      (cpc->entries.count + 1),
				      GC_ALLOC_CLASSPATH)) )
	{
		int lpc;
		
		cpc->entries.data = new_entries;
		/*
		 * Walk backwards through the array, shifting any lesser
		 * priority entries until we get to the start of the array,
		 * or an entry with a larger priority.
		 */
		for( lpc = cpc->entries.count - 1; lpc >= 0; lpc-- )
		{
			if( cpc->entries.data[lpc].priority >= pri )
			{
				/* The current slot has a larger priority */
				break;
			}
			else
			{
				/* Lesser priority, shift it over */
				cpc->entries.data[lpc + 1] =
					cpc->entries.data[lpc];
			}
		}
		/* The previous slot we traversed is where we need to insert */
		lpc++;
		/* Fill it in */
		cpc->entries.data[lpc].entry = cpe;
		cpc->entries.data[lpc].priority = pri;
		cpc->entries.count++;
		retval = 1;

		ensure((lpc == 0) ||
		       (cpc->entries.data[lpc - 1].priority >= pri));
		ensure((lpc == (cpc->entries.count - 1)) ||
		       (pri > cpc->entries.data[lpc + 1].priority));
	}

	ensure_valid_memory(cpc->entries.data);
	
	return( retval );
}

/*
 * Find a classPathEntry in a classPathClient
 */
static
int findClassPath(classPathClient *cpc, classPathEntry *cpe)
{
	int lpc, retval = -1;

	require(cpc != 0);
	require(cpe != 0);
	
	for( lpc = 0; (retval < 0) && (lpc < cpc->entries.count); lpc++ )
	{
		if( cpc->entries.data[lpc].entry == cpe )
			retval = lpc;
	}
	return( retval );
}

int addClassPath(classPathClient *cpc, char *path, int pri)
{
	jteamReservation *jr;
	classPathEntry *cpe;
	int retval = 0;

	require(cpc != 0);
	require(path != 0);

	/* Push a kernel reservation since we're grabbing a classPathEntry */
	if( (jr = pushTeamReservation(kernelTeam)) )
	{
		/*
		 * Make sure the path exists in the resource, even if its
		 * hidden.
		 */
		if( (cpe = findClassPathEntry((classPathResource *)
					      cpc->client.handler,
					      path,
					      true,
					      true)) )
		{
			/* Its available */
			int iLockRoot;
			
			lockMutex(cpc);
			if( findClassPath(cpc, cpe) < 0 )
			{
				retval = insertClassPath(cpc, cpe, pri);
			}
			else
			{
				retval = 1;
			}
			unlockMutex(cpc);
			if( !retval )
			{
				dropClassPathEntry((classPathResource *)
						   cpc->client.handler,
						   cpe);
			}
		}
		popTeamReservation(jr);
	}
	return( retval );
}

int loadClassPathFile(classPathClient *cpc, char *cname, classFile *out_cf,
		      errorInfo *einfo)
{
	int lpc, done = 0, retval = FTH_LOAD_ERROR;
	classPathEntry *cpe;
	int iLockRoot;
	
	require(cpc != 0);
	require(cname != 0);
	require(out_cf != 0);
	require(einfo != 0);
	
	out_cf->owner = cpc->client.owner;
	/*
	 * The kernel owns the fileTypeHandlers so we go there...  The
	 * handlers should be reentrant so we don't have to lock each entry.
	 */
	/* XXX Is a full switchToTeam required? */
	if( switchToTeam(kernelTeam, 0) == 0 )
	{
		postExceptionMessage(einfo, DEAD_TEAM_EXCEPTION_S,
				     "Can't visit kernel in "
				     __FUNCTION__ "?");
		return FTH_LOAD_ERROR;
	}
	lockMutex(cpc);
	/* Walk over the ordered entries trying to load the file */
	for( lpc = 0; !done && (lpc < cpc->entries.count); lpc++ )
	{
		require(cpc->entries.data[lpc].entry != 0);
		
		cpe = cpc->entries.data[lpc].entry;
		switch( (retval = fthLoadEntryFile(cpe, cname,
						   out_cf, einfo)) )
		{
		case FTH_LOAD_ERROR:
			/* Is stopping on error too aggressive? */
			done = 1;
			teamLogEntry("classPath",
				     LEA_String, "Error",
				     "Couldn't open file in class path "
				     "entry, stopping search...", -1,
				     LEA_String, "file", cname, -1,
				     LEA_String, "path", cpe->path, -1,
				     LEA_TAG_DONE);
			break;
		case FTH_LOAD_NOT_FOUND:
			/* Not found, move on to the next entry */
			break;
		case FTH_LOAD_FOUND:
			done = 1;
			break;
		default:
			assert(0);
			break;
		}
		
		ensure(retval > FTH_LOAD_NONE);
		ensure(retval < FTH_LOAD_MAX);
	}
	unlockMutex(cpc);
	returnFromTeam();
	if( !done )
	{
		static int classLen = sizeof(".class") - 1;
		int clen;

		clen = strlen(cname);
		if( (clen > classLen) &&
		    (strcmp(&cname[clen - classLen], ".class") == 0) )
		{
			/* cut off the ".class" suffix for the exception msg */
			cname[clen - classLen] = '\0';
		}
		/*
		 * NOTE: This is a different exception that what needs to be
		 * thrown when failing to load a class for use, the caller
		 * should take care of throwing the appropriate exception.
		 */
		postExceptionMessage(einfo, JAVA_IO(IOException),
				     "Couldn't find %s",
				     cname);
	}
	return( retval );
}

struct parsedClassPath {
	classPathResource *cpr;
	classPathClient *cpc;
	parsedString pathEntry;
};

static
int addParsedClassPathEntry(struct parsedClassPath *pcp)
{
	int retval;

	if( pcp->pathEntry.len > 0 )
	{
		char save;
		
		save = pcp->pathEntry.data[pcp->pathEntry.len];
		pcp->pathEntry.data[pcp->pathEntry.len] = 0;
		if( pcp->cpr )
		{
			if( !(retval = addClassPathEntry(pcp->cpr,
							 pcp->pathEntry.
							 data)) )
			{
				fprintf(stderr,
					"Couldn't add class path entry %s\n",
					pcp->pathEntry.data);
			}
		}
		else
		{
			retval = addClassPath(pcp->cpc,
					      pcp->pathEntry.data,
					      0);
		}
		pcp->pathEntry.data[pcp->pathEntry.len] = save;
	}
	else
	{
		teamLogEntry("classPath",
			     LEA_String, "Warning",
			     "Empty path in class path", -1,
			     LEA_TAG_DONE);
		retval = 1;
	}
	
	return( retval );
}

static
int parseClassPath(classPathResource *cpr,
		   classPathClient *cpc,
		   char *classpath)
{
	struct parsedClassPath pcp = { cpr, cpc, { 0, 0 } };
	int retval;

	pp_decl(int cp_len_prime);
	
	require((cpr != 0) || (cpc != 0));
	require(classpath != 0);
	
	pp_block({ cp_len_prime = strlen(classpath); });

DBG(INITCLASSPATH,
    dprintf("parseClassPath(%p, %p, `%s')\n",
	    cpr, cpc, classpath));

	retval = parseString(classpath,
			     SPO_NotEmpty,
			       SPO_Do,
			         SPO_String, &pcp.pathEntry,
			         SPO_While, path_separator, "",
			         SPO_Handle, addParsedClassPathEntry, &pcp,
			         SPO_End,
			       SPO_End,
			     SPO_End);

	ensure(strlen(classpath) == cp_len_prime);
	
	return( retval );
}

static
int discoverClassPath(classPathResource *cpr, char *home)
{
	int retval = 1;
#if defined(HAVE_DIRENT_H)
	DIR* dir;
	struct dirent* entry;
	int hlen, sep_len, len;
	char* name;
	char* buf;

	require(cpr != 0);
	require(home != 0);
	
	dir = opendir(home);
	if (dir == 0) {
		return( 0 );
	}

	/* Add '.' and <home>/Klasses.jar at the beginning */
	retval = addClassPathEntry(cpr, ".");
	hlen = strlen(home);
	sep_len = strlen(file_separator);
	if( retval &&
	    (buf = gc_malloc(hlen +			/* $HOME */
			     sep_len +			/* / */
			     strlen(KLASSES_JAR) +	/* Klasses.jar */
			     1,				/* '\0' */
			     GC_ALLOC_CLASSPATH)) )
	{
		sprintf(buf, "%s%s%s", home, file_separator, KLASSES_JAR);
		retval = addClassPathEntry(cpr, buf);
		gc_free(buf);
	}
	else
	{
		retval = 0;
	}

	while( retval && ((entry = readdir(dir)) != 0) )
	{
		name = entry->d_name;
		len = strlen(name);
		if( (buf = gc_malloc(hlen +
				     sep_len +
				     len +
				     1,
				     GC_ALLOC_CLASSPATH)) )
		{
			sprintf(buf, "%s%s%s", home, file_separator, name);
			/* Ask the handlers to match the path */
			retval = addClassPathEntry(cpr, buf);
			gc_free(buf);
		}
		else
		{
			retval = 0;
		}
	}
	closedir(dir);
#else
#ifdef DEBUG
	fprintf(stderr,
		"Warning: Can't use opendir()/readdir()/closedir() to "
		"discover class path.\n");
#endif
#endif
	return( retval );
}

void edu_utah_janosvm_resources_ClassPathSpec_nativeConstructor(
	struct Hedu_utah_janosvm_resources_ClassPathSpec *jcps,
	HArrayOfObject *entries)
{
	classPathSpec *cps = OBJECT2NATIVE(&metalClassPathResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jcps);
	char **tmp_array;
	errorInfo einfo;
	int error = 0;

	cps->spec.handler = &metalClassPathResource.link;
	if( !entries || (ARRAY_SIZE(entries) == 0) )
	{
		/* Need to allocate because the finalizer will free it */
		if( (cps->classpath = gc_malloc(1, GC_ALLOC_CLASSPATH)) )
		{
			cps->classpath[0] = '\0';
		}
		else
		{
			postOutOfMemory(&einfo);
			throwError(&einfo);
		}
		return;
	}
	if( (tmp_array = gc_malloc(sizeof(char *) * ARRAY_SIZE(entries),
				   GC_ALLOC_CLASSPATH)) )
	{
		int lpc, total_len = 1, sep_len;
		struct Hjava_lang_String *curr;

		/* Note: we depend on tmp_array being initialized to zero */
		sep_len = strlen(path_separator);
		for( lpc = 0; !error && (lpc < ARRAY_SIZE(entries)); lpc++ )
		{
			curr = ((struct Hjava_lang_String **)
				ARRAY_DATA(entries))[lpc];
			if( curr == NULL )
			{
			}
			else if( (tmp_array[lpc] = stringJava2C(curr)) )
			{
				total_len += strlen(tmp_array[lpc]) + sep_len;
			}
			else
			{
				postOutOfMemory(&einfo);
				error = 1;
			}
		}
		if( !error &&
		    (cps->classpath = gc_malloc(total_len,
						GC_ALLOC_CLASSPATH)) )
		{
			char *pos = cps->classpath;
			int actual = 0;
			
			for( lpc = 0; lpc < ARRAY_SIZE(entries); lpc++ )
			{
				if( tmp_array[lpc] )
				{
					sprintf(pos,
						"%s%s",
						(actual > 0) ?
						path_separator : "",
						tmp_array[lpc]);
					pos += strlen(pos);
					actual += 1;
				}
			}

			ensure_valid_memory(cps->classpath);
		}
		else
		{
			postOutOfMemory(&einfo);
			error = 1;
		}
		for( lpc = 0; lpc < ARRAY_SIZE(entries); lpc++ )
		{
			/* Free the path strings */
			gc_free(tmp_array[lpc]);
		}
		gc_free(tmp_array);
	}
	else
	{
		postOutOfMemory(&einfo);
		error = 1;
	}
	if( error )
		throwError(&einfo);
}

void edu_utah_janosvm_resources_ClassPathSpec_nativeFinalizer(
	struct Hedu_utah_janosvm_resources_ClassPathSpec *jcps)
{
	struct Hedu_utah_janosvm_resources_ResourceSpec *jrs =
		(struct Hedu_utah_janosvm_resources_ResourceSpec *)jcps;
	classPathSpec *cps = OBJECT2NATIVE(&metalClassPathResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jcps);

	if( !jrs->found )
	{
		if( cps->classpath )
		{
			require_valid_memory(cps->classpath);
		}

		gc_free(cps->classpath);
	}
}

jint edu_utah_janosvm_resources_ClassPathSpec_getPathCount(
	struct Hedu_utah_janosvm_resources_ClassPathSpec *jcps)
{
	classPathSpec *cps = OBJECT2NATIVE(&metalClassPathResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jcps);
	jint retval = 0;

	parseString(cps->classpath,
		    SPO_NotEmpty,
		      SPO_Do,
		        SPO_Count, &retval,
		        SPO_While, path_separator, "",
		        SPO_End,
		      SPO_End,
		    SPO_End);
	return( retval );
}

struct parsedSingleClassPath {
	parsedString entry;
	parsedString extracted;
	int current;
	int requested;
};

static
int extractClassPath(struct parsedSingleClassPath *pcp)
{
	int retval = 1;

	if( pcp->current == pcp->requested )
	{
		pcp->extracted = pcp->entry;
	}
	pcp->current++;
	return( retval );
}

Hjava_lang_String *edu_utah_janosvm_resources_ClassPathSpec_getPath(
	struct Hedu_utah_janosvm_resources_ClassPathSpec *jcps,
	jint index)
{
	classPathSpec *cps = OBJECT2NATIVE(&metalClassPathResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jcps);
	struct Hjava_lang_String *retval = 0;
	struct parsedSingleClassPath pcp;

	pcp.current = 0;
	pcp.requested = index;
	pcp.extracted.data = 0;
	pcp.extracted.len = 0;
	if( parseString(cps->classpath,
			SPO_NotEmpty,
			  SPO_Do,
			    SPO_String, &pcp.entry,
			    SPO_While, path_separator, "",
			    SPO_Handle, extractClassPath, &pcp,
			    SPO_End,
			  SPO_End,
			SPO_End) )
	{
		if( pcp.extracted.data )
		{
			retval = checkPtr(stringCBuf2Java(pcp.extracted.data,
							  pcp.extracted.len));
		}
	}
	return( retval );
}

struct Hedu_utah_janosvm_resources_ClassPathSpec *
edu_utah_janosvm_resources_ClassPathSpec_current(void)
{
	struct Hedu_utah_janosvm_resources_ClassPathSpec *retval = 0;
	classPathClient *cpc = (classPathClient *)
		getTeamInternalResource(0, &metalClassPathResource.link);
	classPathSpec *cps;
	int lpc, len = 1;

	retval = (struct Hedu_utah_janosvm_resources_ClassPathSpec *)
		newObject(metalClassPathResource.link.specNativeClass.peer);
	cps = OBJECT2NATIVE(&metalClassPathResource.link.specNativeClass,
			   (Hjava_lang_Object *)retval);
	cps->spec.handler = &metalClassPathResource.link;
	for( lpc = 0; lpc < cpc->entries.count; lpc++ )
	{
		len += strlen(cpc->entries.data[lpc].entry->path) + 1;
	}
	if( (cps->classpath = gc_malloc(len, GC_ALLOC_CLASSPATH)) )
	{
		cps->classpath[0] = '\0';
		for( lpc = 0; lpc < cpc->entries.count; lpc++ )
		{
			if( lpc > 0 )
				strcat(cps->classpath, ":");
			strcat(cps->classpath,
			       cpc->entries.data[lpc].entry->path);
		}
	}
	else
	{
		errorInfo einfo;
		
		postOutOfMemory(&einfo);
		throwError(&einfo);
	}
	return( retval );
}

Hjava_lang_Object *
edu_utah_janosvm_resources_ClassPathSpec_clone(
	struct Hedu_utah_janosvm_resources_ClassPathSpec *jcps)
{
	struct Hedu_utah_janosvm_resources_ClassPathSpec *retval;
	classPathSpec *cps, *retval_cps;

	retval = (struct Hedu_utah_janosvm_resources_ClassPathSpec *)
		newObject(metalClassPathResource.link.specNativeClass.peer);
	cps = OBJECT2NATIVE(&metalClassPathResource.link.specNativeClass,
			   (Hjava_lang_Object *)jcps);
	retval_cps = OBJECT2NATIVE(&metalClassPathResource.link.
				   specNativeClass,
				   (Hjava_lang_Object *)retval);
	retval_cps->spec.handler = &metalClassPathResource.link;
	if( (retval_cps->classpath = gc_malloc(strlen(cps->classpath) + 1,
					       GC_ALLOC_CLASSPATH)) )
	{
		strcpy(retval_cps->classpath, cps->classpath);
	}
	else
	{
		errorInfo einfo;
		
		postOutOfMemory(&einfo);
		throwError(&einfo);
	}
	return( (Hjava_lang_Object *)retval );
}

HArrayOfByte *
edu_utah_janosvm_resources_ClassPathResource_loadFile(
	struct Hjava_lang_String* str)
{
	HArrayOfByte *retval = NULL;
	errorInfo einfo;

	if( str )
	{
		char *name, *lname;
		classFile cf;
		
		name = checkPtr(stringJava2C(str));
		lname = name;
		while( *lname && (*lname == '/') )
		{
			lname++;
		}
		if( loadClassPathFile((classPathClient *)
				      getTeamInternalResource(
					0,
					&metalClassPathResource.link),
				      lname,
				      &cf,
				      &einfo) == FTH_LOAD_FOUND )
		{
			if( (retval = (HArrayOfByte *)
			     newArrayChecked(TYPE_CLASS(TYPE_Byte),
					     cf.size,
					     &einfo)) )
			{
				memcpy(unhand_array(retval)->body,
				       cf.buf,
				       cf.size);
			}
			if( cf.base != NULL )
				gc_free(cf.base);
		}
		gc_free(name);
	}
	else
	{
		postException(&einfo, JAVA_LANG(NullPointerException));
	}
	if( !retval )
		throwError(&einfo);
	return( retval );
}

void
edu_utah_janosvm_resources_ClassPathResource_addPath(
	struct Hjava_lang_String *jPath)
{
	jteamReservation *jr;
	errorInfo einfo;
	int ok = false;

	if( jPath == 0 )
	{
		postException(&einfo, JAVA_LANG(NullPointerException));
	}
	else if( (jr = pushTeamReservation(kernelTeam)) )
	{
		char *path = 0;
		
		if( visitReservedTeam(jr, 0) )
		{
			path = checkPtr(stringJava2C(jPath));
			ok = addClassPathEntry(&metalClassPathResource,
					       path);
			leaveReservedTeam();
		}
		else
		{
			postExceptionMessage(&einfo, DEAD_TEAM_EXCEPTION_S,
					     "Can't reach kernel team");
		}
		if( !ok )
		{
			postExceptionMessage(&einfo, JAVA_IO(IOException),
					     "Couldn't find %s",
					     path);
		}
		GC_free((Collector *)teamCollector(kernelTeam), path);
		popTeamReservation(jr);
	}
	else
	{
		postExceptionMessage(&einfo, DEAD_TEAM_EXCEPTION_S,
				     "Can't reach kernel team");
	}
	if( !ok )
		throwError(&einfo);
}

void
edu_utah_janosvm_resources_ClassPathResource_hidePath(
	struct Hjava_lang_String *jPath)
{
	errorInfo einfo;
	int ok = false;

	if( jPath == 0 )
	{
		postException(&einfo, JAVA_LANG(NullPointerException));
	}
	else if( switchToTeam(kernelTeam, 0) )
	{
		char *path = 0;

		path = checkPtr(stringJava2C(jPath));
		ok = hideClassPathEntry(&metalClassPathResource, path);
		gc_free(path);
		returnFromTeam();
		if( !ok )
		{
			postExceptionMessage(&einfo, JAVA_IO(IOException),
					     "Couldn't find %s",
					     path);
		}
	}
	else
	{
		postExceptionMessage(&einfo, DEAD_TEAM_EXCEPTION_S,
				     "Can't reach kernel team");
	}
	if( !ok )
		throwError(&einfo);
}

void
edu_utah_janosvm_resources_ClassPathResource_removePath(
	struct Hjava_lang_String *jPath)
{
	errorInfo einfo;
	int ok = false;

	if( jPath == 0 )
	{
		postException(&einfo, JAVA_LANG(NullPointerException));
	}
	else if( switchToTeam(kernelTeam, 0) )
	{
		char *path = 0;

		path = checkPtr(stringJava2C(jPath));
		remClassPathEntry(&metalClassPathResource, path);
		gc_free(path);
		returnFromTeam();
		ok = true;
	}
	else
	{
		postExceptionMessage(&einfo, DEAD_TEAM_EXCEPTION_S,
				     "Can't reach kernel team");
	}
	if( !ok )
		throwError(&einfo);
}

Hjava_lang_Object *
edu_utah_janosvm_resources_ClassPathResource_getPaths(void)
{
	Hjava_lang_Object *retval = 0;
	jteamReservation *jr;
	errorInfo einfo;

	if( (jr = pushTeamReservation(kernelTeam)) )
	{
		classPathEntry *cpe;
		extern JNIEnv *env;
		int iLockRoot;
		jarray paths;
		int count;

		lockMutex(&metalClassPathResource);
		for( count = 0, cpe = metalClassPathResource.entries;
		     cpe;
		     cpe = cpe->next )
		{
			if( !(cpe->flags & CPEF_HIDDEN) )
				count++;
		}
		if( (paths = (*env)->NewObjectArray(
			env,
			count,
			stringClass,
			0)) )
		{
			jobject str;
			int lpc = 0;
			
			cpe = metalClassPathResource.entries;
			while( paths && cpe )
			{
				if( cpe->flags & CPEF_HIDDEN )
				{
					/* Ignore it */
				}
				else if( (str = (*env)->
					  NewStringUTF(env, cpe->path)) )
				{
					(*env)->SetObjectArrayElement(env,
								      paths,
								      lpc,
								      str);
					lpc++;
				}
				else
				{
					postOutOfMemory(&einfo);
					paths = 0;
				}
				cpe = cpe->next;
			}
			retval = paths;
		}
		else
		{
			postOutOfMemory(&einfo);
		}
		unlockMutex(&metalClassPathResource);
		popTeamReservation(jr);
	}
	else
	{
		postExceptionMessage(&einfo, DEAD_TEAM_EXCEPTION_S,
				     "Can't reach kernel team");
	}
	if( !retval )
	{
		throwError(&einfo);
	}
	return( retval );
}

char *cpHandleAddMessage(jsiHandlerContext *hc, jsiHandler *jh, jsiMessage *jm)
{
	char *path, *retval = 0;

	switch( jm->what )
	{
	case JSI_EXECUTE_PROPERTY:
		if( jsiMessageGet(jm,
				  JMA_String, "path", 0, &path,
				  JMA_TAG_DONE) )
		{
			int ok;
			
			ok = addClassPathEntry(&metalClassPathResource,
					       path);
			jsiClearMessage(jm);
			jsiMessageAdd(jm,
				      JMA_What, JSI_REPLY,
				      JMA_Integer, "result", ok,
				      JMA_TAG_DONE);
		}
		else
		{
		    retval = "No path value";
		}
		break;
	default:
		retval = "Unsupported operation";
		break;
	}
	return( retval );
}

char *cpHandleHideMessage(jsiHandlerContext *hc,
			  jsiHandler *jh,
			  jsiMessage *jm)
{
	char *path, *retval = 0;

	switch( jm->what )
	{
	case JSI_EXECUTE_PROPERTY:
		if( jsiMessageGet(jm,
				  JMA_String, "path", 0, &path,
				  JMA_TAG_DONE) )
		{
			int ok;
			
			ok = hideClassPathEntry(&metalClassPathResource,
						path);
			jsiClearMessage(jm);
			jsiMessageAdd(jm,
				      JMA_What, JSI_REPLY,
				      JMA_Integer, "result", ok,
				      JMA_TAG_DONE);
		}
		else
		{
		    retval = "No path value";
		}
		break;
	default:
		retval = "Unsupported operation";
		break;
	}
	return( retval );
}

char *cpHandleRemoveMessage(jsiHandlerContext *hc,
			    jsiHandler *jh,
			    jsiMessage *jm)
{
	char *path, *retval = 0;

	switch( jm->what )
	{
	case JSI_EXECUTE_PROPERTY:
		if( jsiMessageGet(jm,
				  JMA_String, "path", 0, &path,
				  JMA_TAG_DONE) )
		{
			int ok;
			
			ok = remClassPathEntry(&metalClassPathResource,
					       path);
			jsiClearMessage(jm);
			jsiMessageAdd(jm,
				      JMA_What, JSI_REPLY,
				      JMA_Integer, "result", ok,
				      JMA_TAG_DONE);
		}
		else
		{
		    retval = "No path value";
		}
		break;
	default:
		retval = "Unsupported operation";
		break;
	}
	return( retval );
}

char *cpHandleRootMessage(jsiHandlerContext *hc,
			  jsiHandler *jh, jsiMessage *jm)
{
	char *retval = 0;

	switch( jm->what )
	{
	case JSI_LIST_PROPERTIES:
		retval = namedHandleMessage(hc, jh, jm);
		break;
	case JSI_GET_PROPERTY:
		{
			classPathEntry *cpe;
			int iLockRoot;
			
			jsiClearMessage(jm);
			jm->what = JSI_REPLY;
			lockMutex(&metalClassPathResource);
			cpe = metalClassPathResource.entries;
			while( cpe && !retval )
			{
				if( !jsiMessageAdd(
					jm,
					JMA_String, "result", cpe->path,
					JMA_Integer, "resultHidden",
					(cpe->flags & CPEF_HIDDEN) ? 1 : 0,
					JMA_TAG_DONE) )
				{
					retval = "Out of memory";
				}
				cpe = cpe->next;
			}
			unlockMutex(&metalClassPathResource);
		}
		break;
	case JSI_GET_SUITES:
		jsiClearMessage(jm);
		if( !jsiMessageAdd(jm,
				   JMA_What, JSI_REPLY,
				   JMA_PropertyList, "result", cpProperties,
				   JMA_TAG_DONE) )
		{
			retval = "Out of memory";
		}
		break;
	default:
		retval = "Unsupported operation";
		break;
	}
	return( retval );
}

int cpGetSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", cpRootProperties,
			       JMA_TAG_DONE);
	return( retval );
}
