/*
 * classPath.h
 * Provides access to the class path for the jvm.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef __classpath_h
#define __classpath_h

/*
 * This module manages the class path for the virtual machine.  The class path
 * functions are bound to a resource so that we can control access to each
 * entry in the path on a per-team basis.
 */

#include "resource.h"
#include "errors.h"
#include "file.h"
#include "jar.h"
#include "jteam.h"
#include "classMethod.h"

/* Return values for the loadEntryFile functions of a fileTypeHandler */
enum {
	FTH_LOAD_NONE,
	FTH_LOAD_ERROR,		/* There was an error while loading */
	FTH_LOAD_NOT_FOUND,	/* Didn't find the requested file */
	FTH_LOAD_FOUND,		/* Found the requested file */
	FTH_LOAD_MAX
};

struct stat; /* XXX */

/*
 * A fileTypeHandler is used to define a specific type of entry in a class
 * path.  For example, a fileTypeHandler for JAR files will define functions
 * for detecting JAR files in the class path, as well as providing functions
 * for accessing the contents of the JAR.
 *
 * `next' - Link to the next fileTypeHandler in a list.
 * `name' - The symbolic name of this handler.
 * `createEntry' - Create a classPathEntry object of this handlers type.
 * `deleteEntry' - Finalize and free the given classPathEntry.
 * `matchPath' - Ask the handler if the given path can be handled by this
 *   fileTypeHandler.
 * `preloadEntry' - Do any preloading/caching for the classPathEntry.
 *   XXX We actually don't preload at the moment.
 * `loadEntryFile' - Load a file from the given classPathEntry.
 * `snapshotEntry' - Take a snapshot of the path entry's data so that it can
 *   be replaced without disturbing the current set of users.
 */
typedef struct _fileTypeHandler {
	struct _fileTypeHandler *next;
	char *name;
	struct _classPathEntry *(*createEntry)(struct _fileTypeHandler *fth,
					       char *path);
	void (*deleteEntry)(struct _fileTypeHandler *fth,
			    struct _classPathEntry *cpe);
	
	/*
	 * This constant indicates how many bytes of the header of a file will
	 * be read in and placed in the `buf' pointer so that the handler can
	 * not only match on the path name, but the contents as well.
	 */
#define FTH_FILE_HEADER_SIZE 4
	int (*matchPath)(struct _fileTypeHandler *fth,
			 char *path,
			 struct stat *st,
			 int fd,
			 char *buf);
	int (*preloadEntry)(struct _fileTypeHandler *fth,
			    struct _classPathEntry *cpe);
	int (*loadEntryFile)(struct _fileTypeHandler *fth,
			     struct _classPathEntry *cpe,
			     char *fname,
			     classFile *cf,
			     errorInfo *einfo);
	int (*snapshotEntry)(struct _fileTypeHandler *fth,
			     struct _classPathEntry *cpe);
} fileTypeHandler;

enum {
	CPEB_PUBLISHED,
	CPEB_HIDDEN,
};

enum {
	CPEF_PUBLISHED = (1L << CPEB_PUBLISHED),
	CPEF_HIDDEN = (1L << CPEB_HIDDEN),
};

/*
 * A classPathEntry is the base for tracking individual entries in a class
 * path.
 *
 * `next' - Link to the next entry.
 * `access' - Who can access this entry.
 * `fileType' - The type of this entry.
 * `flags' - Storage for the above CPEF_* flags.
 * `users' - The reference count for this entry.
 * `path' - The path that this entry represents.
 */
typedef struct _classPathEntry {
	struct _classPathEntry *next;
	fileTypeHandler *fileType;
	unsigned long flags;
	int users;
	char *path;
} classPathEntry;

/*
 * A classPathResource is the root structure for storing class path data.
 * The resource can be subscribed to by a team to gain access to its entries.
 *
 * `link' - The resource structure header.
 * `entries' - The linked list of classPathEntries managed by this resource.
 */
typedef struct _classPathResource {
	resource link;
	iLock *lock;
	classPathEntry *entries;
	classPathEntry **lastEntry;
} classPathResource;

/*
 * The orderedClassPathEntry structure is used to form an ordered subset of
 * classPathEntry's.  Since a client of a resource might not have total access
 * to every entry, or need every one, we maintain the per-team class path's
 * separately using this structure and classPathClient.
 *
 * `entry' - Pointer to the real classPathEntry in the resource.
 * `priority' - The priority of this entry, this gives us some ability to sort
 *   entries.
 */
typedef struct _orderedClassPathEntry {
	classPathEntry *entry;
	int priority;
} orderedClassPathEntry;

/*
 * A classPathClient is the resourceClient for a classPathResource.  It encodes
 * stores the subset of classPathEntry's that this client requested.
 *
 * `client' - The resourceClient header.
 * `lock' - Lock to protect the entries.data array.
 * `entries.data' - An array of orderedClassPathEntry's that store the class
 *   path in largest to smallest priority.
 * `count' - The allocated size of the `data' array.
 */
typedef struct _classPathClient {
	resourceClient client;
	iLock *lock;
	struct {
		orderedClassPathEntry *data;
		int count;
	} entries;
} classPathClient;

/*
 * A classPathSpec is used to encode the request for class path entries.
 *
 * `spec' - The resourceSpec header.
 * `classpath' - A typical separated list of paths.
 */
typedef struct _classPathSpec {
	resourceSpec spec;
	char *classpath;
} classPathSpec;

/*
 * Install a fileTypeHandler into the global list.
 *
 * `fth' - An initialized fileTypeHandler object.
 */
void installFileType(fileTypeHandler *fth);
/*
 * Uninstall the fileTypeHandler from the global list.
 *
 * `fth' - A previously installed fileTypeHandler object.
 */
void uninstallFileType(fileTypeHandler *fth);

/*
 * Add a classPathEntry to the resource.
 *
 * `cpr' - The classPathResource to add the path to.  Most likely just
 *  &metalClassPathResource.
 * `path' - The path to add to the global class path.
 * returns - True if the addition worked, false otherwise.
 */
int addClassPathEntry(classPathResource *cpr, char *path);
/*
 * Mark a classPathEntry as hidden so that requests for it will fail.  However,
 * this doesn't remove the path from the global list, so if the same path is
 * added later the hide flag of the existing entry will just be cleared.
 * XXX I don't know if this is altogether useful, but I couldn't decide if
 * a full remove might be inconvenient in comparison.
 *
 * `cpr' - The classPathResource that should have the entry to hide.  Most
 *  likely just &metalClassPathResource.
 * `path' - The path to remove from the global class path.
 * returns - True if the path was found and hidden.
 */
int hideClassPathEntry(classPathResource *cpr, char *path);
/*
 * Remove a classPathEntry from the resource.  If the path is currently in use
 * then it will be removed from the list and cleaned up when the last
 * reference has been dropped.
 *
 * `cpr' - The classPathResource to remove the path from.  Most likely just
 *  &metalClassPathResource.
 * `path' - The path to remove from the global class path.
 * returns - True if the path was found and removed.
 */
int remClassPathEntry(classPathResource *cpr, char *path);
/*
 * Grab a reference to a classPathEntry in the resource.
 *
 * `cpr' - The classPathResource in which to search for the path.  Most likely
 *  just &metalClassPathResource.
 * `path' - The path to search for.
 * returns - A classPathEntry object corresponding to the given path or NULL
 *  if the path isn't in the global class path.
 */
classPathEntry *grabClassPathEntry(classPathResource *cpr, char *path);
/*
 * Drop a reference to a classPathEntry in the resource.
 *
 * `cpr' - The classPathResource in which to search for the path.  Most likely
 *  just &metalClassPathResource.
 * `cpe' - A classPathEntry that was gotten through grabClassPathEntry.
 */
void dropClassPathEntry(classPathResource *cpr, classPathEntry *cpe);

/*
 * Add an orderedClassPathEntry to a classPathClient.
 *
 * `rc' - The classPathClient object from the team requesting the addition.
 * `path' - The path to be added to the client.
 * `pri' - The priority of the path, which determines its place in the list of
 *  paths.
 * returns - True if the addition succeeded, false otherwise.
 */
int addClassPath(classPathClient *rc, char *path, int pri);
/*
 * Load a file from the class path.
 *
 * `cpc' - The classPathClient object from the team requesting the load.
 * `cname' - The name of the file to load from the class path.
 * `out_cf' - An uninitialized classFile object which is to be initialized with
 *  the file data, if its found.
 * `einfo' - Holds any error information generated by the function.
 * returns - One of the FTH_* values.
 */
int loadClassPathFile(classPathClient *cpc, char *cname, classFile *out_cf,
		      errorInfo *einfo);

/* The class path that is actually being used, whether computed or given */
extern char *realClassPath;
/* The root classPathResource */
extern classPathResource metalClassPathResource;

/*
 * AddModule: classPathModule
 *   Comment: Let the fileTypeHandlers install themselves
 * Implementation: handlers
 *   ModuleDependency: memResourceModule.initMetal
 *   ModuleDependency: resourceModule.internalResources
 *   ModuleDependency: logResourceModule.initMetal
 *   Comment: Install the resource and subscribe the metal team
 * Transition: initMetal
 *   ModuleDependency: utfModule resourceModule.reinstall
 */

#endif
