/*
 * configFile.h
 * Manages a configuration file for the jvm.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "configFile.h"
#include "registry.h"
#include "errors.h"
#include "stringSupport.h"
#include "nameSpace.h"
#include "tags.h"
#include "external.h"
#include "module.h"
#include "moduleList.h"
#include "scripting.h"
#include "nativeClass.h"
#include "nativeClass-inline.h"

#include "stdioResource.h"
#include "logResource.h"

#include "edu_utah_janosvm_resources_ConfigSpec.h"

#include <misc/sectionFile.h>

#define JAVA_CONFIG_SPEC "edu/utah/janosvm/resources/ConfigSpec"

/** Module declarations **/
static int initConfigFileModule(int stage);

module_t configFileModule = {
	"configfile",
	0,
	initConfigFileModule,
	0,
	0
};

/** Module definitions **/

configFilePart registryConfigPart;
configFilePart *configParts;
Utf8Const *kernelUTF;

/** Resource declarations **/
static struct _resourceSpec *configCreateJSISpec(struct _resource *res,
						 jsiMessage *jm);
static void configDeleteJSISpec(struct _resource *res,
				struct _resourceSpec *rs);
static int configSubscribe(struct _resource *res,
			   struct _resourceClient *rc,
			   struct _resourceSpec *rs);
static void configUnsubscribe(struct _resource *res,
			      struct _resourceClient *rc);

static resourceFunctions configResourceFunctions = {
	0,
	0,
	0,
	0,
	configCreateJSISpec,
	configDeleteJSISpec,
	0,
	configSubscribe,
	0,
	configUnsubscribe
};

configResource metalConfigResource;

configClient metalConfigClient = {
	{
		0,
		0,
		&metalConfigResource.link,
		0
	},
	0
};

/** Section definitions */
static tagSectionType libSectionType;

static
int libSectionWalker(sectionType *st, sectionFileData *sfd, sectionFile *sf);

static sectionTypeFunctions libSectionFunctions = {
	0,
	0,
	0,
	0,
	0,
	0,
	libSectionWalker
};

int libSectionWalker(sectionType *st, sectionFileData *sfd, sectionFile *sf)
{
	tagSectionData *tsd = LOCAL_SECTION_DATA(tagSectionData *, st, sfd);
	int lpc, retval = 1;
	tagList *tl;

	tl = tsd->taglist;
	for( lpc = 0; retval && (lpc < tl->tags.count); lpc++ )
	{
		if( findLibrary((char *)tl->tags.data[lpc].tag->data) < 0 )
		{
			retval = 0;
			fprintf(stderr,
				"Unable to open library %s.\n",
				tl->tags.data[lpc].tag->data);
		}
	}
	return( retval );
}

int initConfigFileModule(int stage)
{
	int retval = 0;

	switch( stage )
	{
	case SID_CONFIGFILEMODULE_INITMETAL:
		installResource(&metalConfigResource.link,
				RA_Name, "config",
				RA_Flags, RF_INTERNAL,
				RA_Functions, &configResourceFunctions,
				RA_ClientSize, sizeof(configClient),
				RA_SpecSize, sizeof(configSpec),
				RA_TAG_DONE);
		metalTeam->internalResources[metalConfigResource.link.
					    internalIndex] =
			&metalConfigClient.client;
		retval = 1;
		break;
	case SID_CONFIGFILEMODULE:
		installResource(&metalConfigResource.link,
				RA_SpecNativeClass, JAVA_CONFIG_SPEC,
				RA_TAG_DONE);
		if( !(kernelUTF = utf8ConstNew("kernel", -1)) )
		{
			fprintf(stderr, "Unable to create UTF string.\n");
			return( 0 );
		}
		retval = 1;
		installSectionType(&libSectionType.type,
				   STA_Name, "libraries",
				   STA_Template, &tagsFileSectionTemplate,
				   STA_Functions, &libSectionFunctions,
				   STA_TAG_DONE);
		/* Get the name of the registry config file */
		if( registryName && exportExists("config") )
		{
			char path[REGISTRY_PATH_LEN];
			
			getRegistryPath(path, "config", 0);
			if( (registryConfigPart.filename =
			     gc_malloc(strlen(path) + 1, GC_ALLOC_FIXED)) )
			{
				strcpy(registryConfigPart.filename, path);
				addConfigFile(&registryConfigPart);
			}
			else
			{
				fprintf(stderr, "Not enough memory\n");
				retval = 0;
			}
		}
		if( retval && configParts )
		{
			sectionFile *configFile;
			
			/* Try to open the file and read in its contents */
			if( (configFile = createSectionFile()) )
			{
				configFilePart *cfp;
				errorInfo einfo;
				
				/*
				 * No name needed, we construct the file from
				 * several parts
				 */
				setSectionFileName(configFile, "");
				cfp = configParts;
				/*
				 * Sync all of the files with our internal data
				 */
				while( cfp && retval )
				{
					if( (retval = syncSectionFileWith(
						configFile,
						cfp->filename,
						0)) )
					{
						DBG(INIT,
						    printf("Loaded config "
							   "file %s\n",
							   cfp->filename));
					}
					else
					{
						fprintf(stderr,
							"Couldn't read config "
							"file %s\n",
							cfp->filename);
					}
					cfp = cfp->next;
				}
				if( retval )
				{
					/* Process the file sections */
					if( walkFileSections(configFile) )
					{
						exportClassGroup(kernelUTF,
								 &einfo);
						metalConfigClient.sf =
							configFile;
					}
					else
					{
						fprintf(stderr,
							"Failure processing "
							"config file "
							"sections\n");
						retval = 0;
					}
				}
			}
			else
			{
				fprintf(stderr, "Not enough memory\n");
				retval = 0;
			}
		}
		else if( kernelTeam )
		{
			/*
			 * We're in JanosVM mode, complain when there isn't a
			 * config
			 */
			fprintf(stderr,
				"Error:  No JanosVM config file given.\n");
			retval = 0;
		}
		break;
	}
	return( retval );
}

void addConfigFile(configFilePart *cfp)
{
	cfp->next = configParts;
	configParts = cfp;
}

/** Resource Functions **/
struct _resourceSpec *configCreateJSISpec(struct _resource *res,
					  jsiMessage *jm)
{
	struct _resourceSpec *retval = 0;
	int count;

	count = jsiMessageIdentifierCount(jm, "config");
	if( count && (retval = createResourceSpec(res)) )
	{
		configSpec *cs = (configSpec *)retval;
		configFilePart *cfp;
		int lpc, error = 0;
		char *path;

		for( lpc = 0; !error && (lpc < count); lpc++ )
		{
			jsiMessageGet(jm,
				      JMA_String, "config", lpc, &path,
				      JMA_TAG_DONE);
			if( path[0] != '\0' )
			{
				if( (cfp = (configFilePart *)
				     gc_malloc(sizeof(configFilePart) +
					       strlen(path) +
					       1,
					       GC_ALLOC_FIXED)) )
				{
					cfp->filename = (char *)(cfp + 1);
					strcpy(cfp->filename, path);
					cfp->next = cs->files;
					cs->files = cfp;
				}
				else
				{
					error = 1;
				}
			}
		}
		if( error )
		{
			configDeleteJSISpec(res, retval);
			retval = 0;
		}
	}
	return( retval );
}

void configDeleteJSISpec(struct _resource *res,
			 struct _resourceSpec *rs)
{
	configSpec *cs = (configSpec *)rs;
	configFilePart *cfp;
	
	while( cs->files )
	{
		cfp = cs->files;
		cs->files = cfp->next;
		gc_free(cfp);
	}
	deleteResourceSpec(rs);
}

int configSubscribe(struct _resource *res,
		    struct _resourceClient *rc,
		    struct _resourceSpec *rs)
{
	// configResource *cr = (configResource *)res;
	configClient *cc = (configClient *)rc;
	configSpec *cs = (configSpec *)rs;
	int retval = 0;

	if( cs )
	{
		if( (cc->sf = createSectionFile()) )
		{
			configFilePart *cfp, *cfp_cp;

			setSectionFileName(cc->sf, "");
			cfp = cs->files;
			retval = 1;
			while( cfp && retval )
			{
				if( (retval = syncSectionFileWith(
					cc->sf, cfp->filename, 0)) )
				{
					if( (cfp_cp = gc_malloc(
						sizeof(configFilePart) +
						strlen(cfp->filename) +
						1,
						GC_ALLOC_FIXED)) )
					{
						cfp_cp->filename = (char *)
							(cfp_cp + 1);
						strcpy(cfp_cp->filename,
						       cfp->filename);
						cfp_cp->next = cc->files;
						cc->files = cfp_cp;
					}
					else
					{
						retval = 0;
					}
				}
				else
				{
					teamLogEntry("configFile",
						     LEA_String, "Error", "Couldn't sync with section file.", -1,
						     LEA_String, "file", cfp->filename, -1,
						     LEA_TAG_DONE);
				}
				cfp = cfp->next;
			}
			if( retval )
			{
				if( walkFileSections(cc->sf) )
				{
				}
				else
				{
					teamLogEntry("configFile",
						     LEA_String, "Error", "Couldn't process file sections.", -1,
						     LEA_TAG_DONE);
					retval = 0;
				}
			}
			if( !retval )
				deleteSectionFile(cc->sf);
		}
		else
		{
			teamLogEntry("configFile",
				     LEA_String, "Error", "Couldn't create section file object", -1,
				     LEA_TAG_DONE);
		}
	}
	else
	{
		/* Just use the kernel's configuration */
		cc->sf = ((configClient *)
			  getTeamInternalResource(kernelTeam, res))->sf;
		retval = 1;
	}
	return( retval );
}

void configUnsubscribe(struct _resource *res, struct _resourceClient *rc)
{
	// configResource *cr = (configResource *)res;
	configClient *cc = (configClient *)rc;
	configClient *kernelCC = (configClient *)
		getTeamInternalResource(kernelTeam, res);
	
	if( (cc->sf != kernelCC->sf) )
	{
		deleteSectionFile(cc->sf);
	}
}

sectionFile *currentConfigFile(void)
{
	configClient *cc = (configClient *)
		getTeamInternalResource(0, &metalConfigResource.link);

	return( cc->sf );
}

struct Hjava_lang_String *
edu_utah_janosvm_sys_ConfigFile_lookup(struct Hjava_lang_String *jSection,
				       struct Hjava_lang_String *jTag,
				       jint instance)
{
	configClient *cc = (configClient *)
		getTeamInternalResource(0, &metalConfigResource.link);
	struct Hjava_lang_String *retval = 0;
	char *section = 0, *tag = 0;
	sectionFile *sf = cc->sf;

	if( !sf )
		return 0;
	/* Convert to C strings */
	if( jSection && jTag &&
	    (section = stringJava2C(jSection)) && (tag = stringJava2C(jTag)) )
	{
		sectionFileData *sfd;

		/* Look up the file section */
		if( (sfd = findSectionInFile(sf,
					     &pairsFileSectionType.type,
					     section)) )
		{
			tagSectionData *tsd =
				LOCAL_SECTION_DATA(tagSectionData *,
						   &pairsFileSectionType,
						   sfd);
			tagItem *ti;

			/* Find the tag */
			if( (ti = findTagStr(tsd->taglist,
					     tag,
					     -1,
					     instance)) )
			{
				retval = utf8Const2Java(ti->data.str);
			}
		}
	}
	gc_free(section);
	gc_free(tag);
	return( retval );
}

void edu_utah_janosvm_resources_ConfigSpec_nativeConstructor(
	struct Hedu_utah_janosvm_resources_ConfigSpec *jcs,
	HArrayOfObject *entries)
{
	configSpec *cs = (configSpec *)OBJECT2NATIVE(&metalConfigResource.link.
						     specNativeClass,
						     (Hjava_lang_Object *)jcs);
	int lpc, error = 0;
	errorInfo einfo;

	cs->spec.handler = &metalConfigResource.link;
	if( !entries || (ARRAY_SIZE(entries) == 0) )
	{
		/* Need to allocate because the finalizer will free it */
		cs->files = 0;
		return;
	}
	for( lpc = 0; !error && (lpc < ARRAY_SIZE(entries)); lpc++ )
	{
		struct Hjava_lang_String *curr;
		configFilePart *cfp;
		char *str;
		
		curr = ((struct Hjava_lang_String **)ARRAY_DATA(entries))[lpc];
		if( curr == NULL )
		{
		}
		else if( (str = stringJava2C(curr)) )
		{
			if( (cfp = gc_malloc(sizeof(configFilePart),
					     GC_ALLOC_FIXED)) )
			{
				cfp->filename = str;
				cfp->next = cs->files;
				cs->files = cfp;
			}
			else
			{
				gc_free(str);
				postOutOfMemory(&einfo);
				error = 1;
			}
		}
		else
		{
			postOutOfMemory(&einfo);
			error = 1;
		}
	}
	if( error )
		throwError(&einfo);
}

jint edu_utah_janosvm_resources_ConfigSpec_getFileCount(
	struct Hedu_utah_janosvm_resources_ConfigSpec *jcs)
{
	configSpec *cs = (configSpec *)OBJECT2NATIVE(&metalConfigResource.link.
						     specNativeClass,
						     (Hjava_lang_Object *)jcs);
	configFilePart *cfp;
	jint retval = 0;

	cfp = cs->files;
	while( cfp )
	{
		retval++;
		cfp = cfp->next;
	}
	return( retval );
}

Hjava_lang_String *edu_utah_janosvm_resources_ConfigSpec_getFile(
	struct Hedu_utah_janosvm_resources_ConfigSpec *jcs,
	jint index)
{
	configSpec *cs = (configSpec *)OBJECT2NATIVE(&metalConfigResource.link.
						     specNativeClass,
						     (Hjava_lang_Object *)jcs);
	Hjava_lang_String *retval = 0;
	configFilePart *cfp;

	cfp = cs->files;
	while( cfp && (index > 0) )
	{
		index--;
		cfp = cfp->next;
	}
	if( (index == 0) && cfp )
	{
		retval = (Hjava_lang_String *)
			checkPtr(stringC2Java(cfp->filename));
	}
	return( retval );
}

Hjava_lang_Object *
edu_utah_janosvm_resources_ConfigSpec_clone(
	struct Hedu_utah_janosvm_resources_ConfigSpec *jcs)
{
	struct Hedu_utah_janosvm_resources_ConfigSpec *retval = 0;
	configFilePart *cfp, *cfp_cp;
	configSpec *cs, *retval_cs;

	retval = (struct Hedu_utah_janosvm_resources_ConfigSpec *)
		newObject(metalConfigResource.link.specNativeClass.peer);
	retval_cs = OBJECT2NATIVE(&metalConfigResource.link.
				  specNativeClass,
				  (Hjava_lang_Object *)retval);
	retval_cs->spec.handler = &metalConfigResource.link;
	cs = (configSpec *)OBJECT2NATIVE(&metalConfigResource.link.
					 specNativeClass,
					 (Hjava_lang_Object *)jcs);
	cfp = cs->files;
	while( cfp )
	{
		if( (cfp_cp = gc_malloc(sizeof(configFilePart) +
					strlen(cfp->filename) +
					1,
					GC_ALLOC_FIXED)) )
		{
			cfp_cp->filename = (char *)(cfp_cp + 1);
			strcpy(cfp_cp->filename, cfp->filename);
			cfp_cp->next = retval_cs->files;
			retval_cs->files = cfp_cp;
		}
		else
		{
			errorInfo einfo;

			postOutOfMemory(&einfo);
			throwError(&einfo);
		}
		cfp = cfp->next;
	}
	return( (Hjava_lang_Object *)retval );
}

void edu_utah_janosvm_resources_ConfigSpec_nativeFinalizer(
	struct Hedu_utah_janosvm_resources_ConfigSpec *jcs)
{
	struct Hedu_utah_janosvm_resources_ResourceSpec *jrs =
		(struct Hedu_utah_janosvm_resources_ResourceSpec *)jcs;
	configSpec *cs = (configSpec *)OBJECT2NATIVE(&metalConfigResource.link.
						     specNativeClass,
						     (Hjava_lang_Object *)jcs);
	configFilePart *cfp;

	if( !jrs->found )
	{
		while( cs->files )
		{
			cfp = cs->files;
			cs->files = cfp->next;
			gc_free(cfp->filename);
			gc_free(cfp);
		}
	}
}

struct Hedu_utah_janosvm_resources_ConfigSpec *
edu_utah_janosvm_resources_ConfigSpec_current(void)
{
	configClient *kernelCC = (configClient *)
		getTeamInternalResource(kernelTeam, &metalConfigResource.link);
	configClient *cc = (configClient *)
		getTeamInternalResource(0, &metalConfigResource.link);
	struct Hedu_utah_janosvm_resources_ResourceSpec *jrs;
	struct Hedu_utah_janosvm_resources_ConfigSpec *retval = 0;

	if( cc->sf == kernelCC->sf )
	{
	}
	else
	{
		configSpec *retval_cs;
		
		retval = (struct Hedu_utah_janosvm_resources_ConfigSpec *)
			newObject(metalConfigResource.link.
				  specNativeClass.peer);
		retval_cs = OBJECT2NATIVE(&metalConfigResource.link.
					  specNativeClass,
					  (Hjava_lang_Object *)retval);
		retval_cs->spec.handler = &metalConfigResource.link;
		jrs = (struct Hedu_utah_janosvm_resources_ResourceSpec *)
			retval;
		jrs->found = 1;
		retval_cs->files = cc->files;
	}
	return( retval );
}
