/*
 * dirFileType.c
 * FileTypeHandler for directories
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdarg.h>

#include "module.h"
#include "system.h"
#include "files.h"
#include "gtypes.h"
#include "jsyscall.h"
#include "gc.h"
#include "mungWallProxy.h"
#include "dirFileType.h"

/** Module declarations **/
static int initDirFileTypeModule(int stage);

module_t dirFileTypeModule = {
	"dirFileType",
	0,
	initDirFileTypeModule,
	0,
	0
};

/* fileTypeHandler declarations */
static classPathEntry *createEntry(fileTypeHandler *fth, char *path);
static void deleteEntry(fileTypeHandler *fth, classPathEntry *cpe);
static int matchPath(fileTypeHandler *fth, char *path, struct stat *st,
		     int fd, char *buf);
static int preloadEntry(fileTypeHandler *fth, classPathEntry *cpe);
static int loadEntryFile(fileTypeHandler *fth, classPathEntry *cpe,
			 char *cname, classFile *cf, errorInfo *einfo);
static int openDirFile(char *filename, classFile *cf, errorInfo *einfo);
static int snapshotEntry(fileTypeHandler *fth, classPathEntry *cpe);

static fileTypeHandler dirFileTypeHandler = {
	0,
	"dir",
	createEntry,
	deleteEntry,
	matchPath,
	preloadEntry,
	loadEntryFile,
	snapshotEntry,
};

/** Module definitions **/

static
int initDirFileTypeModule(int stage)
{
	int retval = 1;

	installFileType(&dirFileTypeHandler);
	return( retval );
}

/** dirFileType Handler definitions **/

classPathEntry *createEntry(fileTypeHandler *fth, char *path)
{
	classPathEntry *retval = 0;

	/*
	 * Allocate space for the entry and the path name.  We only need the
	 * path string to work, so theres no need for an extra structure.
	 */
	if( (retval = gc_malloc(sizeof(classPathEntry) + strlen(path) + 1,
				GC_ALLOC_FILE_TYPE)) )
	{
		retval->next = 0;
		retval->fileType = fth;
		retval->path = (char *)(retval + 1);
		strcpy(retval->path, path);
	}
	return( retval );
}

void deleteEntry(fileTypeHandler *fth, classPathEntry *cpe)
{
	gc_free(cpe);
}

int matchPath(fileTypeHandler *fth, char *path, struct stat *st,
	      int fd, char *buf)
{
	int retval = 0;

	/* Check the state to see if its a directory */
	if( S_ISDIR(st->st_mode) )
	{
		retval = 1;
	}
	return( retval );
}

int preloadEntry(fileTypeHandler *fth, classPathEntry *cpe)
{
	/* We could cache a directory hierarchy here... */
	return( 1 );
}

int loadEntryFile(fileTypeHandler *fth, classPathEntry *cpe, char *fname,
		  classFile *cf, errorInfo *einfo)
{
	int retval = 0;
	char *buf;

	/* Allocate space for the full path */
	if( (buf = gc_malloc(strlen(cpe->path) +
			     strlen(file_separator) +
			     strlen(fname) +
			     1,
			     GC_ALLOC_FILE_TYPE)) )
	{
		sprintf(buf, "%s%s%s", cpe->path, file_separator, fname);
		/* Try to read in the file */
		retval = openDirFile(buf, cf, einfo);
		gc_free(buf);
	}
	else
	{
		postOutOfMemory(einfo);
	}
	return( retval );
}

static
int readDirFile(int fd, struct stat *st, classFile *cf, errorInfo *einfo)
{
	int rc, retval = FTH_LOAD_ERROR;
	
	/* Allocate space from requester */
	if( switchToTeam(cf->owner, 0) )
	{
		cf->size = st->st_size;
		cf->base = NULL;
		/* Allocate space for the file */
		if( (cf->size == 0) ||
		    (cf->base = gc_malloc(cf->size, GC_ALLOC_FILE_TYPE)) )
		{
			int bytes_left, bytes_read;
			
			cf->buf = cf->base;
			bytes_left = cf->size;
			rc = 0;
			/* Read it in */
			while( bytes_left &&
			       !(rc = KREAD(fd,
					    &cf->base[cf->size - bytes_left],
					    bytes_left,
					    &bytes_read)) &&
			       bytes_read )
			{
				bytes_left -= bytes_read;
			}
			if( bytes_left == 0 )
			{
				retval = FTH_LOAD_FOUND;

				ensure_valid_memory(cf->base);
			}
			else
			{
				postExceptionMessage(einfo,
						     JAVA_IO(IOException),
						     "Couldn't read: %s", 
						     SYS_ERROR(rc));
				gc_free(cf->base);
			}
		}
		else
		{
			postOutOfMemory(einfo);
		}
		returnFromTeam();
	}
	else
	{
		postExceptionMessage(einfo, DEAD_TEAM_EXCEPTION_S,
				     "ClassFile owner team is dead in "
				     __FUNCTION__);
	}
	return( retval );
}

static
int openDirFile(char *filename, classFile *cf, errorInfo *einfo)
{
	int fd, rc, retval = FTH_LOAD_ERROR;

	require(filename != 0);
	require(cf != 0);
	require(einfo != 0);
	
	if( !KOPEN(filename, O_RDONLY|O_BINARY, 0, &fd) )
	{
		struct stat st;

		/* Use fstat to get the size of the file */
		if( !(rc = KFSTAT(fd, &st)) )
		{
			if( !(st.st_mode & S_IFDIR) )
			{
				retval = readDirFile(fd, &st, cf, einfo);
			}
			else
			{
				/* XXX Error or Not Found ? */
				retval = FTH_LOAD_NOT_FOUND;
			}
		}
		else
		{
			postExceptionMessage(einfo,
					     JAVA_IO(IOException),
					     "Couldn't fstat %s: %s",
					     filename,
					     SYS_ERROR(rc));
		}
		KCLOSE(fd);
	}
	else
	{
		retval = FTH_LOAD_NOT_FOUND;
	}
	return( retval );
}

int snapshotEntry(fileTypeHandler *fth, classPathEntry *cpe)
{
	return( 0 );	
}
