/*
 * jarFileType.c
 * Class path fileTypeHandler for JAR files
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdarg.h>

#include "module.h"
#include "gtypes.h"
#include "jsyscall.h"
#include "gc.h"
#include "logResource.h"
#include "jarFileType.h"

/** Module declarations **/
static int initJarFileTypeModule(int stage);

module_t jarFileTypeModule = {
	"jarFileType",
	0,
	initJarFileTypeModule,
	0,
	0
};

/** fileTypeHandler declarations **/
static classPathEntry *createEntry(fileTypeHandler *fth, char *path);
static void deleteEntry(fileTypeHandler *fth, classPathEntry *cpe);
static int matchPath(fileTypeHandler *fth, char *path, struct stat *st,
		     int fd, char *buf);
static int preloadEntry(fileTypeHandler *fth, classPathEntry *cpe);
static int loadEntryFile(fileTypeHandler *fth, classPathEntry *cpe,
			 char *cname, classFile *cf, errorInfo *einfo);
static int snapshotEntry(fileTypeHandler *fth, classPathEntry *cpe);

static fileTypeHandler jarFileTypeHandler = {
	0,
	"jar",
	createEntry,
	deleteEntry,
	matchPath,
	preloadEntry,
	loadEntryFile,
	snapshotEntry,
};

/** Module definitions **/

static
int initJarFileTypeModule(int stage)
{
	int retval = 1;

	installFileType(&jarFileTypeHandler);
	return( retval );
}

/** jarFileType handler definitions **/

classPathEntry *createEntry(fileTypeHandler *fth, char *path)
{
	classPathEntry *retval = 0;
	jarClassPathEntry *jcpe;

	/* Allocate the entry and its path string */
	if( (jcpe = gc_malloc(sizeof(jarClassPathEntry) + strlen(path) + 1,
			      GC_ALLOC_FILE_TYPE)) )
	{
		jcpe->entry.next = 0;
		jcpe->entry.fileType = fth;
		jcpe->entry.path = (char *)(jcpe + 1);
		strcpy(jcpe->entry.path, path);
		jcpe->jar = 0;
		retval = &jcpe->entry;
	}
	return( retval );
}

void deleteEntry(fileTypeHandler *fth, classPathEntry *cpe)
{
	jarClassPathEntry *jcpe = (jarClassPathEntry *)cpe;
	
	if( jcpe )
	{
		closeJarFile(jcpe->jar);
		gc_free(jcpe);
		/*
		 * XXX Bugger, OSKit's stat doesn't work right.  So, we need
		 * to flush everything, otherwise it won't notice the change.
		 */
		flushJarCache();
	}
}

int matchPath(fileTypeHandler *fth, char *path, struct stat *st,
	      int fd, char *buf)
{
	int retval = 0;

	/* Check the file header for a match */
	if( buf && (buf[0] == 'P') && (buf[1] == 'K') )
	{
		retval = 1;
	}
	return( retval );
}

int preloadEntry(fileTypeHandler *fth, classPathEntry *cpe)
{
	jarClassPathEntry *jcpe = (jarClassPathEntry *)cpe;

	/* Read in the jar file */
	if( !(jcpe->jar = openJarFile(cpe->path)) )
	{
		teamLogEntry(
			"jarFileType",
			LEA_String, "Warning", "Couldn't preload JAR file", -1,
			LEA_String, "path", cpe->path, -1,
			LEA_TAG_DONE);
	}
	return( 1 );
}

int loadEntryFile(fileTypeHandler *fth, classPathEntry *cpe, char *fname,
		  classFile *cf, errorInfo *einfo)
{
	jarClassPathEntry *jcpe = (jarClassPathEntry *)cpe;
	int retval = FTH_LOAD_ERROR;
	
	if( !jcpe->jar )
	{
		/* Wasn't preloaded, try it now */
		if( !(jcpe->jar = openJarFile(jcpe->entry.path)) )
		{
			postExceptionMessage(einfo,
					     JAVA_IO(IOException),
					     "Couldn't open jar file %s.",
					     jcpe->entry.path);
		}
	}
	if( jcpe->jar )
	{
		jarEntry *je;

		/* Lookup the file in the jar */
		if( (je = lookupJarFile(jcpe->jar, fname)) )
		{
			if( (cf->base = getDataJarFile(cf->owner,
						       jcpe->jar,
						       je)) )
			{
				cf->size = je->uncompressedSize;
				cf->buf = cf->base;
				retval = FTH_LOAD_FOUND;
			}
			else
			{
				postExceptionMessage(einfo,
						     JAVA_IO(IOException),
						     "Couldn't extract "
						     "data from jar: %s", 
						     jcpe->jar->error);
			}
		}
		else
		{
			retval = FTH_LOAD_NOT_FOUND;
		}
	}
	return( retval );
}

int snapshotEntry(fileTypeHandler *fth, classPathEntry *cpe)
{
	jarClassPathEntry *jcpe = (jarClassPathEntry *)cpe;
	int retval;
	
	if( jcpe->jar || (jcpe->jar = openJarFile(jcpe->entry.path)) )
	{
		retval = snapshotJarFile(jcpe->jar);
	}
	else
	{
		retval = 0;
	}
	return( retval );
}
