/*
 * jteam.h
 * Maintains separation of code running in the jvm.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jteam_h
#define _jteam_h

/*
 * This module provides functions for managing jteams, a unix process-like
 * abstraction.
 */

#include <jsi/jsi.h>

#include "thread.h"
#include "locks.h"
#include "resource.h"
#include "gc.h"
#include "classMethod.h"
#include "nativeClass.h"
#include "sparseArray.h"
#include "listNode.h"
#include "edu_utah_janosvm_sys_TeamBackEnd.h"

struct _jteam;

#define JTEAM_DEPENDENCY_POOL 4

/*
 * The jteamDependency describes any dependencies that exist between teams.
 *
 * `next' - The next set of dependency pointers.
 * `deps' - Pool of pointers to teams for which a team has dependencies.
 */
typedef struct _jteamDependency {
	struct _jteamDependency *next;
	struct _jteam *deps[JTEAM_DEPENDENCY_POOL];
} jteamDependency;

/*
 * A jteam structure is used to contain any resources available to the threads
 * of a team.
 *
 * `link' - Link into list of teams.
 * `handler' - jsiHandler for this team.
 * `internalLock' - Lock taken when its probably not a good idea to take the
 *   big team lock.  (Coverage: description).
 * `elem' - Reference to unique identifier.
 * `user' - The user associated with this team.
 * `name' - The name of the team.
 * `description' - A string describing the contents of the team.
 * `deps' - List of teams that this team directly depends on.
 * `ghouls' - A lock for threads waiting for this team to be dead and buried.
 *   NOTE: This is a jmutex lock since the thread might be waiting for a
 *   condition and the threading system might not be able to remove this
 *   thread violently, which means the per-thread iLock is still in use.
 * `zombieCount' - Tracks the number of threads that were unkillable when the
 *   terminator tried to slaughter them.
 * `liveThreads' - The number of live threads belonging to this team,
 *   including daemon threads.
 * `daemonThreads' - The number of daemon threads belonging to this team.
 * `state' - The state of the team (new, alive, dying, dead, or buried).
 *   The state of a team monotonically increases.
 * `guestThreads' - How many threads are visiting our team.
 * `absentThreads' - How many of this teams threads are away visiting other
 *   teams.
 * `internalResources' - An array containing pointers to the internal
 *   resources for a team (see resource.h).
 * `externalResources' - A linked list of externals resources for the 
 *   team (see resource.h)
 */

/* JTeam state values */
typedef enum {
	JTS_NEW = 1,		/* The team has been allocated */
	JTS_ALIVE,		/* The team is alive */
	JTS_DYING,		/* The team is dying (still has visitors/absentees)*/
	JTS_DEAD,		/* The team is dead */
	JTS_BURIED		/*
				 * The team is no longer active and its
				 * resources have been returned to the
				 * system.  The only thing left is the
				 * TeamBackEnd object.
				 */
} jteam_state_t;

enum {
	JTB_NOTIFY_ON_FINISH,
};

enum {
	JTF_NOTIFY_ON_FINISH = (1L << JTB_NOTIFY_ON_FINISH)
};

typedef struct _jteam {
	struct lnMinNode link;
	jsiHandler handler;
	struct {
		iLock *lock;
	} internalLock;
	sparseID_t elemID;
	unsigned long flags;
	char *name;
	char *description;

	int dependents;
	jteamDependency deps;

	struct {
		jmutex lock;
		jcondvar cvar;
	} ghouls;
	
	int zombieCount;
	int liveThreads;
	int daemonThreads;

	jteam_state_t state;
	
	int guestThreads;
	int absentThreads;
	jthread_t threadList;

	resourceClient *externalResources;
	resourceClient *internalResources[0];
} jteam;

/* The first team, contains resources from the bare metal/OS */
extern jteam *metalTeam;
extern struct _metalTeamAndResources {
	jteam team;
	resourceClient *resources[16];
} metalTeamAndResources;
/* Pointer to a resolved `java.lang.TeamBackEnd' class if it exists */
extern Hjava_lang_Class *TeamBackEndClass;
/*
 * nativeClass data structure describing the native data attached to a
 * TeamBackEnd
 */
extern nativeClass jteamNativeClass;
/*
 * If `java.lang.TeamBackEnd' is available then this is a pointer to the
 * kernel team we created out of it.  Otherwise, its NULL.
 */
extern Hedu_utah_janosvm_sys_TeamBackEnd *kernelTeamBackEnd;
/*
 * If kernelTeamBackEnd is non-NULL then this points to the native data section
 * of kernelTeamBackEnd.
 */
extern jteam *kernelTeam;

typedef enum {
	TEAM_ERROR_NONE,
	TEAM_ERROR,
	TEAM_ERROR_NO_MEMORY,
	TEAM_ERROR_NO_SOURCE,
	TEAM_ERROR_NO_DESTINATION,
} team_error_t;

/*
 * Lock a jteam object and disable stopping the current thread.
 *
 * `jt' - The team to lock.
 */
void _lockTeam(jteam *jt, int *iLockRoot);
#define lockTeam(jt) _lockTeam(jt, &iLockRoot)
/*
 * Unlock a jteam object and reenable stopping the current thread.
 *
 * `jt' - The team to unlock.
 */
void _unlockTeam(jteam *jt, int *iLockRoot);
#define unlockTeam(jt) _unlockTeam(jt, &iLockRoot)

/*
 * Set the description string for the given team.
 *
 * `jt' - The jteam object to describe.
 * `str' - The description string which is copied into the team's heap.
 */
void setTeamDescription(jteam *jt, char *str);

/*
 * Terminate the threads of the team and free its resources.  The return of
 * this function doesn't mean that the team is now dead, it could be stalled
 * because of a reservation, queued up on the team reaper thread, or already
 * dead.
 *
 * `jt' - A jteam object that has a java level reference or was grabbed.
 */
void terminateTeam(jteam *jt);
/*
 * Reserve a team for future accesses/visits.  This should never really be
 * called directly, journaled reservations should be used since its safer.
 *
 * `jt' - A jteam object that has a java level reference or was grabbed.
 */
team_error_t reserveTeam(jteam *jt);
/*
 * Drop a reservation for a team.  If all of the reservations for a team have
 * been dropped then it can be terminated.
 *
 * `jt' - A jteam object that has a java level reference or was grabbed.
 */
void unreserveTeam(jteam *jt);
/*
 * Attach a resource to the team.
 *
 * `jt' - A jteam object that has a java level reference or was grabbed.
 * `rs' - A resource specification that describes the desired requirements.
 * returns - True if the attach worked, false otherwise.
 */
int attachTeamResource(jteam *jt, resourceSpec *rs);
/*
 * Detach a resource from the team.  XXX I'm unsure of the whole attach/detach
 * thing...  TSS
 *
 * `jt' - A jteam object that has a java level reference or was grabbed.
 * `res' - The resource type to remove.
 * `id' - The unique identifier of the resource client, if there was one.
 */
void detatchTeamResource(jteam *jt, resource *res, int id);
/*
 * Print the team structure.
 * (For debugging purposes only)
 *
 * `jt' - A jteam object that has a java level reference or was grabbed.
 */
void printTeam(jteam *jt);

/*
 * Make one team depend on another.  Normally, teams are peers and they exist
 * independent of other teams in the system, so adding a dependency can link
 * the fate of one team to another.
 *
 * `jt' - A jteam object that has a java level reference or was grabbed.
 * `dep' - The `jt' should depend on.
 * returns - True if the addition worked, false otherwise.
 */
int addTeamDependency(jteam *jt, jteam *dep);

/*
 * A catastrophe recovery function that does nothing, use this if you want to
 * stop the catastrophe somewhere in C code, but you don't need to do any
 * extra work.
 */
int nullCatastropheRecovery(void *eobj);
/*
 * Find and add a reference to the team with the given unique id.  This is
 * needed since a pointer to the jteam structure is internal to the java object
 * and is therefore not sufficient to keep the java object from being
 * collected.
 *
 * `tid' - The unique identifier for the team.
 * returns - The jteam object which corresponds to the given identifier or
 *  NULL if the specified team wasn't found.
 */
jteam *grabTeamWithID(sparseID_t tid);
/*
 * Add a reference to the team object, otherwise there might be no other
 * references and it could be GC'd.  NOTE: Since a jteam pointer is pointing
 * to the internals of a java object its not enough to just have it on the
 * stack.
 *
 * `jt' - A jteam object that has been previously grabbed or whose java object
 *  exists on the stack.
 */
void grabTeam(jteam *jt);
/*
 * Drop a reference to the team object.
 *
 * `jt' - A previously grabbed jteam object.
 */
void dropTeam(jteam *jt);

/*
 * Reserve and visit the given team.
 *
 * `dest' - The destination jteam.
 * `catastrophe' - A catastrophe handler.
 * returns - True if the switch succeeded, false otherwise.
 */
int switchToTeam(jteam *dest, int (*catastrophe)(void *eobj));
/*
 * Unreserve and return from a switchToTeam operation.
 */
void returnFromTeam(void);

/*
 * Add a thread to a team, if zero is returned then the team is dying/dead, so
 * no more threads can be added to it.
 *
 * `jt' - The team that the thread is trying to spawn in.
 * return - True if the thread was allowed into the team, false otherwise.
 */
int allowTeamThread(jteam *jt);
/*
 * Register a thread in a team.
 *
 * `jt' - The jteam to register the thread with.
 * `tid' - The thread to register.
 * `daemon' - True if the thread is a daemon, false otherwise.  Used to
 *  help determine when all active threads in a team have finished.
 * returns - True...  XXX Why return anything here?
 */
int registerTeamThread(jteam *jt, jthread_t tid, int daemon);
/*
 * Unregister a thread from a team.
 *
 * `jt' - The jteam to register the thread with.
 * `tid' - The thread to register.
 * `daemon' - True if the thread is a daemon, false otherwise.  Used to
 *  help determine when all active threads in a team have finished.
 */
void unregisterTeamThread(jteam *jt, jthread_t tid, int daemon);
/*
 * Change the deamon status of the given jthread.  This will update the
 * team internal daemon vs. regular thread status and can cause a signal
 * to someone doing a waitTeam.
 *
 * `tid' - The thread that changing its daemon status.
 * `daemon' - The True if the thread is now a daemon, false otherwise.
 */
void reregisterTeamThread(jthread_t tid, int daemon);

typedef enum {
	WAIT_FOR_TEAM_THREADS=1,/*
				 * Wait for all the non-daemon threads to die
				 */
	WAIT_FOR_TEAM_ZOMBIES,	/* Wait for all the zombie threads to die */
	WAIT_FOR_TEAM_BURIAL	/* Wait for the team to be buried. */
} teamwaitconditions_t;
/*
 * Wait for a team condition.
 *
 * `jt' - The team to wait on.
 * `condition' - The team condition to wait for, as specified by the above
 *  values.
 */
void waitTeam(jteam *jt, teamwaitconditions_t condition);

enum {
	SIGNAL_THREAD_DEATH,	/*
				 * Signal any threads waiting for the last
				 * non-daemon thread to exit.
				 */
	SIGNAL_TEAM_BURIAL,	/*
				 * Signal any threads waiting for the team to
				 * die.
				 */
};
/*
 * Signal threads waiting on the team object.
 *
 * `jt' - The team to signal.
 * `condition' - The team condition to signal, as specified by the above
 *  values.
 */
void signalTeam(jteam *jt, int condition);

/*
 * The name of the exception to post when you're unable to visit a team.
 */
#define DEAD_TEAM_EXCEPTION_S "edu.utah.janosvm.sys.DeadTeamException"
/*
 * The name of the exception to post when you're violently thrown out of a
 * visit.
 */
#define TEAM_EJECTION_EXCEPTION_S "edu.utah.janosvm.sys.TeamEjectionException"

/*
 * AddModule: jteamModule
 *   ModuleDependency: machineDependentModule
 *   Comment: Put the internal resource array in the metalTeam object.
 * Transition: initMetal
 *   ModuleDependency: baseClassesModule memResourceModule threadModule
 *   ModuleDependency: nameSpaceModule
 *   Comment: Try to get the TeamBackEnd class and create the kernel object.
 */

#endif
