/*
 * jteamReservation.c
 * Manages team reservations and migrations.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "gtypes.h"
#include "gc.h"
#include "classMethod.h"
#include "threadData.h"
#include "threadData-inline.h"
#include "jteam.h"
#include "jteamReservation.h"
#include "stackTrace.h"
#include "access.h"

#include <misc/memoryStack.h>
#include <misc/memoryStack-inline.h>

extern Hjava_lang_Class *TeamReservationClass;

/*
 * XXX Its possible, though not likely now, that the thread journal stack is
 * doing allocations in the wrong team.  I think a quickChange could fix
 * it but I'm not positive and there is prolly a better way to do it.
 *
 * Actually, now that i think about it this is probably the desired behaviour.
 * When we visit a team we'll allocate any extra journal stack within its context,
 * but we'll have deallocated it by the team we leave the team.  So it should
 * just "work".
 */

struct _jteam *quickChangeTeamReservation(jteam *jt)
{
	jthread_t th = jthread_current();
	threadData *td = jthread_get_data(th);
	struct _jteam *retval;

	retval = td->currentTeam;
	if( td->boundary )
		td->boundary->currentReservation->entry.id = jt;
	td->currentTeam = jt;
	return( retval );
}

jteam_reservation_retval_t pushTeamReservation_raw(jteam *jt,
						   void *returnPC,
						   void *framePointer,
						   int shallow,
						   jteamReservation **out_jr)
{
	jteam_reservation_retval_t retval = JRV_OK;
	jthread_t th = jthread_current();
	threadData *td = jthread_get_data(th);
	jteamReservation *jr = 0;

	require(out_jr != 0);
	
	if( (jt == metalTeam) ||	/* Needed to emulate/initialize */
	    (jt == 0) ||		/* "" */
	    (reserveTeam(jt) == TEAM_ERROR_NONE) )
	{
		/* XXX STACK_GROWS_DOWN */
		require((td->boundary == 0) ||
			(framePointer <=
			 td->boundary->currentReservation->framePointer));
			
		if( (shallow &&
		     (jr = (jteamReservation *)allocStackSpace_shallow(
				td->journal, sizeof(jteamReservation)))) ||
		    (!shallow &&
		     (jr = (jteamReservation *)allocStackSpace(
				td->journal, sizeof(jteamReservation)))) )
		{
			/*
			 * Setup object base so java code can use these objects
			 */
			if( TeamReservationClass )
			{
				jr->entry.base.dtable =
					TeamReservationClass->dtable;
			}
			jr->entry.base.lock = 0;
			jr->flags = 0;
			jr->returnPC = returnPC;
			jr->framePointer = framePointer;
			jr->entry.id = jt;
			*out_jr = jr;
		}
		else if( (jt != 0) && (jt != metalTeam) )
		{
			retval = JRV_OVERFLOW;
			unreserveTeam(jt);
		}
	}
	else
	{
		retval = JRV_RESERVE_FAILED;
	}
	return( retval );
}

jteamReservation *pushTeamReservation(jteam *jt)
{
	jteamReservation *retval = 0;
	void *rpc, *fp;

	getCallerFrameAndPC(&fp, &rpc);
	pushTeamReservation_raw(jt, rpc, fp, false, &retval);
	return( retval );
}

void popTeamReservation(jteamReservation *in_jr)
{
	jthread_t th = jthread_current();
	threadData *td = jthread_get_data(th);
	jteamReservation *jr;

	/* Pop the last Reservation */
	jr = (jteamReservation *)topOfMemoryStack(td->journal,
						  sizeof(jteamReservation));
	{
		require((in_jr == 0) || (in_jr == jr));
		require(!(jr->flags & JRF_IN_USE));
		
		/* Check needed for initialization/emulation */
		if( (jr->entry.id != metalTeam) && (jr->entry.id != 0) )
			unreserveTeam(jr->entry.id);
		freeStackSpace(td->journal, sizeof(jteamReservation));
	}
}

int initRootReservation(threadData *td)
{
	jteam *jt = td->ownerTeam;
	jteamReservation *jr;
	
	jr = (jteamReservation *)allocStackSpace(td->journal,
						 sizeof(jteamReservation));
	if( TeamReservationClass )
		jr->entry.base.dtable = TeamReservationClass->dtable;
	jr->entry.base.lock = 0;
	jr->flags = 0;
	jr->returnPC = 0;
	jr->framePointer = (void *)~0;
	jr->entry.id = jt;
	visitReservedTeam_raw(td, jr, 0, (void *)~0, 0);
	return( 1 );
}

int visitReservedTeam_raw(threadData *td,
			  jteamReservation *jr,
			  void *returnPC,
			  void *framePointer,
			  catastrophe_func_t catastrophe)
{
	jteamBoundary *jb;
	int retval = 0;

	require(jr != 0);
	require(!(jr->flags & JRF_IN_USE));

	/* Push a boundary object */
	if( (jb = (jteamBoundary *)allocStackSpace(
		td->journal, sizeof(jteamBoundary))) )
	{
		/* Record the Reservation as in use */
		jr->flags |= JRF_IN_USE;
		jb->prev = td->boundary;
		jb->currentReservation = jr;
		jb->returnPC = returnPC;
		jb->framePointer = framePointer;
		jb->catastrophe = catastrophe;
		td->boundary = jb;
		/* Only increment visits when we actually change teams */
		if( jr->entry.id && (td->currentTeam != jr->entry.id) )
			td->visits++;
		if( jr->entry.id )
			td->currentTeam = jr->entry.id;
		retval = 1;
	}
	else
	{
		/* XXX */
		require(0);
	}
	
	ensure(td->visits >= 0);
	
	return( retval );
}

static
int resCatastropheRecovery(void *eobj)
{
	errorInfo einfo;
	
	leaveReservedTeam();
	postExceptionMessage(&einfo, TEAM_EJECTION_EXCEPTION_S,
			     "Ejection from team during visit()");
	throwError(&einfo);

	ensure(0);
	
	return( 0 );
}

void edu_utah_janosvm_sys_TeamReservation_visit(struct Hjava_lang_Object *obj)
{
	jteamReservation *jr = (jteamReservation *)obj;
	void *rpc, *fp;
	
	getCallerFrameAndPC(&fp, &rpc);
	rpc = getFrameReturnPC(fp);
	fp = getPreviousFramePointer(fp);
	if( !visitReservedTeam_raw(jthread_get_data(jthread_current()),
				   jr,
				   rpc,
				   fp,
				   resCatastropheRecovery) )
	{
		gc_throwOOM();
	}
}

void edu_utah_janosvm_sys_TeamReservation_leave(void)
{
	leaveReservedTeam();
}

int visitReservedTeam(jteamReservation *jr, catastrophe_func_t catastrophe)
{
	void *rpc, *fp;
	int retval;

	require(jr != 0);
	require(!(jr->flags & JRF_IN_USE));

	getCallerFrameAndPC(&fp, &rpc);
	retval = visitReservedTeam_raw(jthread_get_data(jthread_current()),
				       jr,
				       rpc,
				       fp,
				       catastrophe);
	return( retval );
}

void leaveReservedTeam(void)
{
	jthread_t th = jthread_current();
	threadData *td = jthread_get_data(th);
	jteamBoundary *jb = td->boundary;

	require(jb != 0);
	require((td->visits >= 0) || (jb->prev == 0));

	/* Only decrement if we actually changed teams */
	if( jb->prev &&
	    (td->currentTeam != jb->prev->currentReservation->entry.id) &&
	    jb->prev->currentReservation->entry.id )
	{
		td->visits--;
	}
	jb->currentReservation->flags &= ~JRF_IN_USE;
	/* XXX This should only be allowed when called by crash */
	while( topOfMemoryStack(td->journal, sizeof(jteamBoundary)) != jb )
	{
		popTeamReservation(0);
	}
	/* Switch to the previous boundary/team */
	td->boundary = jb->prev;
	if( td->boundary && td->boundary->currentReservation->entry.id )
		td->currentTeam = td->boundary->currentReservation->entry.id;
	else
		td->currentTeam = td->ownerTeam;
	freeStackSpace(td->journal, sizeof(jteamBoundary));

	ensure(td->visits >= 0);
}

void rollbackReservationStack(void *framePointer,
			      void *startPC,
			      void *endPC)
{
	jthread_t th = jthread_current();
	threadData *td = jthread_get_data(th);
	jteamBoundary *jb = td->boundary;
	int done = false;
	
	require(jb != 0);

	while( !done )
	{
		if( (jb == 0) ||
		    (jb->prev == 0) ||
		    (framePointer < jb->framePointer) )
		{
			jb = td->boundary;
			/*
			 * Its still inside the boundary (or its the root
			 * boundary), just get rid of the Reservations.
			 */
			while( (topOfMemoryStack(
				td->journal, sizeof(jteamBoundary)) != jb) &&
			       !done )
			{
				jteamReservation *jr;
				
				jr = (jteamReservation *)topOfMemoryStack(
					td->journal, sizeof(jteamReservation));
				if( framePointer == jr->framePointer )
				{
					if( (startPC <= jr->returnPC) &&
					    (jr->returnPC <= endPC) )
					{
						popTeamReservation(jr);
					}
					else
					{
						done = true;
					}
				}
				else if( framePointer > jr->framePointer )
				{
					popTeamReservation(jr);
				}
				else
				{
					done = true;
				}
			}
			/*
			 * If there aren't any Reservations then we need to
			 * break the loop
			 */
			done = true;
		}
		else if( framePointer == jb->framePointer )
		{
			/* In the same frame */
			if( (startPC <= jb->returnPC) &&
			    (jb->returnPC <= endPC) )
			{
				jb = jb->prev;
				leaveReservedTeam();
			}
			else
			{
				jb = 0;
			}
		}
		else if( framePointer > jb->framePointer )
		{
			/* We've crossed back over the boundary */
			jb = jb->prev;
			leaveReservedTeam();
		}
	}
}

void crashIntoTeamBoundary(void *eobj)
{
	jthread_t th = jthread_current();
	threadData *td = jthread_get_data(th);
	struct _exceptionFrame orig;
	struct _stackTrace trace;
	stackTraceInfo info;

	(void)orig;
	STACKTRACEINIT(trace, 0, 0, orig);
	/*
	 * Back out of each boundary until there are none left or we hit
	 * a handler function.
	 */
	while( td->boundary && (td->boundary->catastrophe == 0) )
	{
		jteamBoundary *jb = td->boundary;
#if defined(STACK_GROWS_UP)
#error "Implement me"
#else
		while( !STACKTRACEEND(trace) &&
		       ((uintp)jb->framePointer >= STACKTRACEFP(trace)) )
#endif
		{
			Hjava_lang_Object *obj;
			
			info.pc = STACKTRACEPC(trace);
			info.fp = STACKTRACEFP(trace);
			info.meth = findMethodFromPC(info.pc);
			
			if( (info.meth == 0) ||
			    (info.meth->accflags & ACC_SYNCHRONISED) == 0 )
			{
				obj = 0;
			}
			else if( info.meth->accflags & ACC_STATIC )
			{
				obj = &info.meth->class->head;
			}
			else
			{
				FRAMEOBJECT(obj, info.fp, 0 /* XXX */);
			}
			if( (obj != 0) &&
			    (info.meth->accflags & ACC_SYNCHRONISED) != 0 )
			{
				_slowUnlockMutexIfHeld(&obj->lock,
						       (void *)info.fp);
			}
			STACKTRACESTEP(trace);
		}
		leaveReservedTeam();
	}
	if( td->boundary )
	{
		jteamBoundary *jb = td->boundary;

#if defined(STACK_GROWS_UP)
#error "Implement me"
#else
		while( !STACKTRACEEND(trace) &&
		       ((uintp)jb->framePointer >= STACKTRACEFP(trace)) )
#endif
		{
			Hjava_lang_Object *obj;
			
			info.pc = STACKTRACEPC(trace);
			info.fp = STACKTRACEFP(trace);
			info.meth = findMethodFromPC(info.pc);
			
			if( (info.meth == 0) ||
			    (info.meth->accflags & ACC_SYNCHRONISED) == 0 )
			{
				obj = 0;
			}
			else if( info.meth->accflags & ACC_STATIC )
			{
				obj = &info.meth->class->head;
			}
			else
			{
				FRAMEOBJECT(obj, info.fp, 0 /* XXX */);
			}
			if( (obj != 0) &&
			    (info.meth->accflags & ACC_SYNCHRONISED) != 0 )
			{
				_slowUnlockMutexIfHeld(&obj->lock,
						       (void *)info.fp);
			}
			STACKTRACESTEP(trace);
		}
		/* Call the handler function */
		CALL_KAFFE_C_EXCEPTION(jb->framePointer,
				       jb->returnPC,
				       jb->catastrophe,
				       eobj);
	}
	else
	{
		/* No handler functions, just kill the thread */
		jthread_slaughter(th);
	}

	/* Shouldn't get to here */
	ensure(0);
}
