/*
 * jteamReservation.h
 * Manages team reservations and migrations.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jteam_reservation_h
#define _jteam_reservation_h

#include <misc/journal.h>

#include "native.h"

struct _jteam;

/*
 * A jteamReservation records a team reservation for a single thread.
 *
 * `base' - A java object base so that these structures can be used at the
 *   the java level.  XXX is this evil?
 * `flags' - Flags for the Reservation.
 * `team' - The team that is reserved by the thread.
 */
enum {
	JRB_IN_USE	/* Marks the object as used for migration */
};

enum {
	JRF_IN_USE = (1L << JRB_IN_USE)
};

typedef struct _jteamReservation {
	journalEntry entry;
	unsigned long flags;
	void *returnPC;
	void *framePointer;
} jteamReservation;

/*
 * A catastrophe function is a special function used to recover from
 * catastrophic situation.  For example, kaffe can get itself into a bad
 * position where the only option would be calling panic().  However, this
 * doesn't work in JanosVM since it can leave the jteam structures in bogus
 * states, so we must back out of each recorded migration/reservation to
 * keep the states sane.  So the catastrophe function provides an option to
 * stop this recovery at a point and run the special handler function.
 */
typedef int (*catastrophe_func_t)(void *eobj);

/*
 * A jteamBoundary records a thread migrating between teams.
 *
 * `prev' - Link to the previous boundary.
 * `currentReservation' - The jteamReservation that was used to migrate.
 * `returnPC' - The location to return to on a catastrophe.
 * `framePointer' - A stack frame pointer that will still be live when a
 *   catastrophe function will be called.
 * `catastrophe' - The function to call when the thread gets into an
 *   unrecoverable position.
 */
typedef struct _jteamBoundary {
	struct _jteamBoundary *prev;
	jteamReservation *currentReservation;
	void *returnPC;
	void *framePointer;
	catastrophe_func_t catastrophe;
} jteamBoundary;

/* XXX bleh */
struct _jteam *quickChangeTeamReservation(jteam *jt);

typedef enum {
	JRV_OK,
	JRV_RESERVE_FAILED,
	JRV_OVERFLOW,
} jteam_reservation_retval_t;
/*
 * Push a reservation onto the stack for the current thread.  On success the
 * given team is reserved and the Reservation is returned.  On failure null is
 * returned and the team is not accessible.
 *
 * `jt' - The team to reserve.
 * `returnPC' - The pc to return to in case of a catastrophe.
 * `framePointer' - The pointer to the frame where returnPC will be running.
 * returns - A jteamReservation that can be used in a visit/leave or NULL if
 *  the reserve failed, or there wasn't enough memory to allocate the journal
 *  entry.
 */
jteam_reservation_retval_t pushTeamReservation_raw(jteam *jt,
						   void *returnPC,
						   void *framePointer,
						   int shallow,
						   jteamReservation **out_jr);
jteamReservation *pushTeamReservation(jteam *jt);
/*
 * Pop a reservation off the stack of the current thread.
 *
 * `jr' - The jteamReservation at the top of the stack, only used to verify.
 *  Or NULL if the top reservation should just be popped off.
 */
void popTeamReservation(jteamReservation *jr);
/*
 * Switch the current thread to the team specified by the given reservation.
 * Normally returns true, but if there was a catastrophe and a function was
 * given then this will return false.
 *
 * `jr' - A reservation for the destination team.
 */
int visitReservedTeam(jteamReservation *jr, catastrophe_func_t catastrophe);
/*
 * Specialized function for the root Reservation, basically it pushes a
 * boundary that refers to the team that the thread was spawned in.
 *
 * `td' - The threadData for the new thread.
 */
int initRootReservation(threadData *td);
/*
 * Generalized function that lets you specify the returnPC and framePointer.
 */
int visitReservedTeam_raw(threadData *td,
			  jteamReservation *act,
			  void *returnPC,
			  void *framePointer,
			  catastrophe_func_t catastrophe);
/*
 * The opposite of a visit, switch from the reserved team back to the previous
 * one.
 */
void leaveReservedTeam(void);
/*
 * Alternative to panic that will back out of each Reservation use and restore
 * the team states.  If any use has a catastrophe function than recovery will
 * stop at that use and call the function with the given argument.
 *
 * `eobj' - The object to pass to the catastrophe handler.  XXX Rarely used
 *  at the moment.
 */
void crashIntoTeamBoundary(void *eobj);
/*
 * Rollback the Reservation stack to the point preceding the given place in the
 * code as described by the stack frame pointer and the range of code within
 * the function.
 *
 * `framePointer' - The frame pointer to roll back to.  Or ~0 to roll back to
 *  the top of the stack.  XXX This is dependent on which direction the stack
 *  grows.
 * `startPC' - The start of the code block within the function that executed
 *  a visit.
 * `endPC' - The end of the code block.
 */
void rollbackReservationStack(void *framePointer,
			      void *startPC,
			      void *endPC);

#define RESERVATION_OVERFLOW_ERROR_S \
	"edu.utah.janosvm.sys.ReservationOverflowError"

#endif
