/*
 * jteamScripting.c
 * Scripting functions for jteams
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <jsi/jsi.h>

#include "assert_pp.h"

#include "gc.h"
#include "classMethod.h"
#include "jteam.h"
#include "nativeClass-inline.h"
#include "java_lang_Throwable.h"
#include "stdioResource.h"
#include "scripting.h"
#include "jteamReservation.h"

#include <misc/stringParsing.h>

/* Defined in jteam.c */
extern sparseArray *teamArray;
extern struct lnMinList teamList;
extern iLock *teamArrayLock;

jclass TeamClass;
jmethodID TeamCreateMethod;
jmethodID TeamGetIDMethod;

/* Information about the root team property. */
jsiPropertyInfo teamRootProperties[] = {
	{ "team", (1L << JSI_CREATE_PROPERTY), "",
	  "name:string - The name of the team to create\n\n"
	  "Create a team\n" },
	{ "team", (1L << JSI_DELETE_PROPERTY), "teamID:int",
	  "Delete the given team\n" },
	{ "teams", (1L << JSI_LIST_PROPERTIES), "",
	  "List the current teams in the system\n" },
	{ 0 }
};

/* Information about properties inside a team property. */
jsiPropertyInfo teamProperties[] = {
	{ "name", (1L << JSI_GET_PROPERTY), "",
	  "The name of the team\n" },
	{ "description", ((1L << JSI_GET_PROPERTY) |
			  (1L << JSI_SET_PROPERTY)), "",
	  "A description of the contents of the team\n" },
	{ "run", (1L << JSI_EXECUTE_PROPERTY), "",
	  "class:string - The name of the class to run\n"
	  "args:string - Space separated list of arguments\n"
	  "arg:string - A single argument\n\n"
	  "Run the main function of a class in the team.\n" },
	{ "addDependency", (1L << JSI_EXECUTE_PROPERTY), "",
	  "team:int - The team to add a dependency for\n\n"
	  "Add a dependency for the given team\n" },
	{ "wait", (1L << JSI_EXECUTE_PROPERTY), "",
	  "Wait for the team to be dead and buried\n" },
	{ "waitForThreads", (1L << JSI_EXECUTE_PROPERTY), "",
	  "Wait for the threads of a team to die\n" },
	{ "gc", (1L << JSI_GET_PROPERTY), "",
	  "Get the current gc stats for the team\n" },
	{ "gc", (1L << JSI_EXECUTE_PROPERTY), "",
	  "Run the garbage collector\n" },
	{ 0 }
};

jsiHandler *teamResolveSpecifier(jsiHandler *jh, jsiMessage *jm,
				 char *name, char *data)
{
	jsiHandler *retval = jh;
	
	return( retval );
}

/*
 * Helper structure used to build the argv for the java class' main function
 *
 * `argc' - The current argument.
 * `argv' - The array to store the java arguments.
 * `arg' - The parsed argument.
 */
struct mainArgs {
	int argc;
	jarray argv;
	parsedString arg;
};

/* Append an argument to the argv array */
static
int appendArg(struct mainArgs *ma)
{
	extern JNIEnv *env;
	int retval = 0;
	jobject str;
	char save;

	save = ma->arg.data[ma->arg.len];
	ma->arg.data[ma->arg.len] = 0;
	if( (str = (*env)->NewStringUTF(env, ma->arg.data)) )
	{
		(*env)->SetObjectArrayElement(env, ma->argv, ma->argc, str);
		ma->argc++;
		retval = 1;
	}
	ma->arg.data[ma->arg.len] = save;
	return( retval );
}

/* Create the argv from the given jsiMessage */
static
jarray createMainArgs(jclass strcls, jsiMessage *jm)
{
	int error = 0, lpc, argc = 0;
	char *id_name, *id_value;
	extern JNIEnv *env;
	jarray retval = 0;

	/* Count the number of args */
	for( lpc = 0; lpc < jm->identifiers.count; lpc++ )
	{
		jsiMessageGetIdentifier(jm, lpc, &id_name, &id_value);
		if( !strcmp(id_name, "arg") )
		{
			argc++;
		}
		else if( !strcmp(id_name, "args") )
		{
			parseString(id_value,
				    SPO_Do,
				      SPO_Count, &argc,
				      SPO_WhileSpace,
				      SPO_End,
				    SPO_End);
		}
	}
	if( (retval = (*env)->NewObjectArray(env, argc, strcls, 0)) )
	{
		struct mainArgs ma;

		ma.argv = retval;
		ma.argc = 0;
		for( lpc = 0; !error && (lpc < jm->identifiers.count); lpc++ )
		{
			jsiMessageGetIdentifier(jm, lpc, &id_name, &id_value);
			/* Add the args to the list */
			if( !strcmp(id_name, "arg") )
			{
				ma.arg.data = id_value;
				ma.arg.len = strlen(id_value);
				error = !appendArg(&ma);
			}
			else if( !strcmp(id_name, "args") )
			{
				error = !parseString(
					id_value,
					SPO_Do,
					  SPO_String, &ma.arg,
					  SPO_WhileSpace,
					  SPO_Handle, appendArg, &ma,
					  SPO_End,
					SPO_End);
			}
		}
		if( error )
		{
			retval = 0;
		}
	}
	return( retval );
}

void
printStackTrace(struct Hjava_lang_Throwable* o,
		struct Hjava_lang_Object* p, int nullOK);

void
printExecuteException(char *className)
{
	Hjava_lang_Throwable *th;
	extern JNIEnv *env;

	if( (th = (Hjava_lang_Throwable *)(*env)->ExceptionOccurred(env)) )
	{
		char *cmsg = 0;
		jclass eiic;
		
		if( th->message )
		{
			cmsg = stringJava2C(th->message);
		}
		stdioPrint(STDIO_ERR,
			   "%s threw exception during run\n"
			   "%s:%s\n",
			   className,
			   CLASS_CNAME(OBJECT_CLASS((Hjava_lang_Object *)th)),
			   cmsg ? cmsg : "");
		printStackTrace(th, 0, 1);
		gc_free(cmsg);
		eiic = (*env)->FindClass(
			env, "java/lang/ExceptionInInitializerError");
		if( eiic && (*env)->IsInstanceOf(env, th, eiic) )
		{
			th = (*env)->CallObjectMethod(
				env, th,
				(*env)->GetMethodID(
					env,
					(*env)->GetObjectClass(env, th),
					"getException",
					"()Ljava/lang/Throwable;"));
			if( th->message )
			{
				cmsg = stringJava2C(th->message);
			}
			stdioPrint(STDIO_ERR,
				   "ExceptionInInitializerError:%s\n",
				   cmsg ? cmsg : "");
			printStackTrace(th, 0, 1);
			gc_free(cmsg);
		}
		(*env)->ExceptionClear(env);
	}
}

/* Run the main function of the given class */
static
int runClassMain(char *className, jsiMessage *jm)
{
	jclass mcls, strcls;
	extern JNIEnv *env;
	int retval = 0;

	if( (mcls = (*env)->FindClass(env, className)) &&
	    (strcls = (*env)->FindClass(env, "java/lang/String")) )
	{
		jmethodID mmth;

		if( (mmth = (*env)->GetStaticMethodID(
			env, mcls, "main", "([Ljava/lang/String;)V")) )
		{

			jarray args;

			if( (args = createMainArgs(strcls, jm)) )
			{
				(*env)->CallStaticVoidMethod(env,
							     mcls,
							     mmth,
							     args);
				if( (*env)->ExceptionOccurred(env) )
				{
					printExecuteException(className);
				}
				else
				{
					retval = 1;
				}
			}
			else
			{
				stdioPrint(STDIO_ERR,
					   "Error parsing arguments for "
					   "main()\n");
			}
		}
		else
		{
			printExecuteException(className);
		}
	}
	else if( !mcls )
	{
		printExecuteException(className);
	}
	return( retval );
}

/* We need a special handler for catching catastrophe's in the other team */
static
int jsiTeamCatastrophe(void *eobj)
{
	stdioDetachJSIChannel();
	returnFromTeam();
	return( 0 );
}

static
char *doAddDependency(jteam *jt, jsiMessage *jm)
{
	char *retval = 0;
	sparseID_t tid;
	
	if( jsiMessageGet(jm,
			  JMA_UnsignedInteger, "team", 0, &tid,
			  JMA_TAG_DONE) )
	{
		if( switchToTeam(kernelTeam, 0) )
		{
			jteam *dep;
			
			if( (dep = grabTeamWithID(tid)) )
			{
				int rc;
				
				rc = addTeamDependency(jt, dep);
				jsiClearMessage(jm);
				jsiMessageAdd(jm,
					      JMA_What, JSI_REPLY,
					      JMA_Integer, "result", rc,
					      JMA_TAG_DONE);
				dropTeam(dep);
			}
			else
			{
				retval = "The `team' ID is invalid";
			}
			returnFromTeam();
		}
		else
		{
			retval = "Unable to switch to kernel";
		}
	}
	else
	{
		retval = "Need `team' and `dependsOn' team ID's";
	}
	return( retval );
}

char *teamHandleMessage_real(jsiHandlerContext *hc,
			     jsiHandler *jh,
			     jsiMessage *jm)
{
	jteam *jt = (jteam *)(((struct lnMinNode *)jh) - 1);
	char *spec_name, *spec_value, *retval = 0;
	
	if( !switchToTeam(jt, jsiTeamCatastrophe) )
	{
		require(currentTeam() == kernelTeam);

		return( "Dead team" );
	}
	stdioAttachJSIChannel(getScriptThreadData()->channel);
	switch( jm->what )
	{
	case JSI_LIST_PROPERTIES:
		jsiClearMessage(jm);
		if( !jsiMessageAdd(jm,
				   JMA_What, JSI_REPLY,
				   JMA_String, "result", "name",
				   JMA_String, "result", "description",
				   JMA_String, "result", "run",
				   JMA_String, "result", "addDependency",
				   JMA_String, "result", "wait",
				   JMA_String, "result", "waitForThreads",
				   JMA_String, "result", "gc",
				   JMA_TAG_DONE) )
		{
			retval = "No memory";
		}
		break;
	case JSI_SET_PROPERTY:
		if( jsiPopMessageSpecifier(jm, &spec_name, &spec_value) )
		{
			if( !strcmp(spec_name, "description") )
			{
				char *data;

				if( jsiMessageGet(jm,
						  JMA_String, "data", 0, &data,
						  JMA_TAG_DONE) )
				{
					setTeamDescription(jt, data);
					jsiClearMessage(jm);
					jsiMessageAdd(jm,
						      JMA_What, JSI_REPLY,
						      JMA_String, "result",
						      "ok",
						      JMA_TAG_DONE);
				}
				else
				{
					retval = "No data to set";
				}
			}
			else
			{
				retval = "Bad specifier for set";
			}
		}
		else
		{
			retval = "No specifier to set";
		}
		break;
	case JSI_GET_PROPERTY:
		if( jsiPopMessageSpecifier(jm, &spec_name, &spec_value) )
		{
			if( !strcmp(spec_name, "gc") )
			{
				jsiClearMessage(jm);
				jm->what = JSI_REPLY;
				if( !gc_report(teamCollector(jt), jm) )
					retval = "No memory";
			}
			else if( !strcmp(spec_name, "name") )
			{
				jsiClearMessage(jm);
				if( !jsiMessageAdd(
					jm,
					JMA_What, JSI_REPLY,
					JMA_String, "result", jt->name,
					JMA_TAG_DONE) )
				{
					retval = "No memory";
				}
			}
			else if( !strcmp(spec_name, "description") )
			{
				int iLockRoot;
				
				jsiClearMessage(jm);
				jm->what = JSI_REPLY;
				lockMutex(&jt->internalLock);
				if( !jsiMessageAdd(
					jm,
					JMA_What, JSI_REPLY,
					JMA_String, "result", jt->description,
					JMA_TAG_DONE) )
				{
					retval = "No memory";
				}
				unlockMutex(&jt->internalLock);
			}
			else
			{
				retval = "Bad specifier for get";
			}
		}
		else
		{
			retval = "No specifier to get";
		}
		break;
	case JSI_DELETE_PROPERTY:
		if( jt != kernelTeam )
		{
			jsiClearMessage(jm);
			terminateTeam(jt);
			/*
			 * The terminate won't happen until we back out of the
			 * team.
			 */
			jsiMessageAdd(jm,
				      JMA_What, JSI_REPLY,
				      JMA_String, "result", "ok",
				      JMA_TAG_DONE);
		}
		else
		{
			retval = "Don't delete the kernel team, use shutdown "
				"if you want to stop the VM.";
		}
		break;
	case JSI_EXECUTE_PROPERTY:
		if( jsiPopMessageSpecifier(jm, &spec_name, &spec_value) )
		{
			if( !strcmp(spec_name, "run") )
			{
				char *cname;
				
				if( jsiMessageGet(jm,
						  JMA_String,
						  "class", 0, &cname,
						  JMA_TAG_DONE) )
				{
					int res;

					res = runClassMain(cname, jm);
					fflush(stdout);
					fflush(stderr);
					jsiClearMessage(jm);
					jsiMessageAdd(jm,
						      JMA_What, JSI_REPLY,
						      JMA_Integer,
						      "result",
						      res,
						      JMA_TAG_DONE);
				}
				else
				{
					retval = "No class to run";
				}
			}
			else if( !strcmp(spec_name, "addDependency") )
			{
				retval = doAddDependency(jt, jm);
			}
			else if( !strcmp(spec_name, "waitForThreads") )
			{
				waitTeam(jt, WAIT_FOR_TEAM_THREADS);
				jsiClearMessage(jm);
				jsiMessageAdd(jm,
					      JMA_What, JSI_REPLY,
					      JMA_Integer, "result", 1,
					      JMA_TAG_DONE);
			}
			else if( !strcmp(spec_name, "gc") )
			{
				invokeGC();
				jsiClearMessage(jm);
				jsiMessageAdd(jm,
					      JMA_What, JSI_REPLY,
					      JMA_String, "result", "ok",
					      JMA_TAG_DONE);
			}
			else
			{
				retval = "Not an executable property";
			}
		}
		else
		{
			retval = "Not an executable property";
		}
		break;
	default:
		retval = "Operation not supported";
		break;
	}
	stdioDetachJSIChannel();
	returnFromTeam();
	return( retval );
}

char *teamHandleMessage(jsiHandlerContext *hc, jsiHandler *jh, jsiMessage *jm)
{
	jteam *jt = (jteam *)(((struct lnMinNode *)jh) - 1);
	char *retval = 0;

	retval = teamHandleMessage_real(hc, jh, jm);
	dropTeam(jt);
	
	ensure(currentTeam() == kernelTeam);
	
	return( retval );
}

jsiHandler *teamResolveRootSpecifier(jsiHandler *jh, jsiMessage *jm,
				     char *name, char *data)
{
	jsiHandler *retval = 0;
	sparseID_t id;
	
	if( data[0] == 0 )
	{
		retval = jh;
	}
	else if( sscanf(data, "%d", &id) )
	{
		sparseElement *se;
		int iLockRoot;

		lockStaticMutex(&teamArrayLock);
		if( (se = findSparseElement(teamArray, id))
		    && (se->data.ptr != NULL) )
		{
			jteam *jt;

			jt = (jteam *)se->data.ptr;
			switch( jm->what )
			{
			case JSI_DELETE_PROPERTY:
				jsiRestoreMessageSpecifier(jm);
				retval = jh;
				break;
			case JSI_EXECUTE_PROPERTY:
				if( jsiPopMessageSpecifier(jm,
							   &name,
							   &data) )
				{
					if( !strcmp(name, "wait") )
					{
						jsiRestoreMessageSpecifier(jm);
						jsiRestoreMessageSpecifier(jm);
						retval = jh;
					}
					else
					{
						jsiRestoreMessageSpecifier(jm);
						grabTeam(jt);
						retval = &jt->handler;
					}
				}
				break;
			case JSI_GET_SUITES:
				retval = jh;
				break;
			default:
				grabTeam(jt);
				retval = &jt->handler;
				break;
			}
		}
		else
		{
			switch( jm->what )
			{
			case JSI_EXECUTE_PROPERTY:
				if( jsiPopMessageSpecifier(jm,
							   &name,
							   &data) )
				{
					if( !strcmp(name, "wait") )
					{
						jsiRestoreMessageSpecifier(jm);
						jsiRestoreMessageSpecifier(jm);
						retval = jh;
					}
				}
				break;
			}
		}
		unlockStaticMutex(&teamArrayLock);
	}
	else
	{
		switch( jm->what )
		{
		case JSI_GET_SUITES:
			retval = jh;
			break;
		}
	}
	return( retval );
}

static
resourceSpec **createNativeSpecs(jsiMessage *jm)
{
	resourceSpec **retval;

	if( (retval = gc_malloc(sizeof(resourceSpec *) *
				internalResourceCount,
				GC_ALLOC_JTEAM)) )
	{
		resource *res;

		res = getResourceList();
		while( res )
		{
			retval[res->internalIndex] =
				createResourceSpecFromJSI(res, jm);
			res = res->next;
		}
	}
	return( retval );
}

static
void deleteNativeSpecs(resourceSpec **nSpecs)
{
	if( nSpecs )
	{
		int lpc;
		
		for( lpc = 0; lpc < internalResourceCount; lpc++ )
		{
			deleteResourceSpecFromJSI(nSpecs[lpc]);
		}
		gc_free(nSpecs);
	}
}

char *teamHandleRootMessage(jsiHandlerContext *hc,
			    jsiHandler *jh, jsiMessage *jm)
{
	char *spec_name, *spec_value;
	Hjava_lang_String *jName = 0;
	Hjava_lang_Object *th;
	char *name, *retval = 0;
	resourceSpec **nSpecs;
	extern JNIEnv *env;
	int iLockRoot;

	require(currentTeam() == kernelTeam);
	require(env != NULL); /* XXX happens if JSI message comes in too early */
	
	switch( jm->what )
	{
	case JSI_LIST_PROPERTIES:
		{
			int error = 0;
			jteam *jt;

			jsiClearMessage(jm);
			jm->what = JSI_REPLY;
			lockStaticMutex(&teamArrayLock);
			jt = (jteam *)teamList.lh_Head;
			while( !error && jt->link.ln_Succ )
			{
				error = !jsiMessageAdd(
					jm,
					JMA_UnsignedInteger, "id", jt->elemID,
					JMA_String, "name", jt->name,
					JMA_String, "description", jt->description,
					JMA_TAG_DONE);
				jt = (jteam *)jt->link.ln_Succ;
			}
			unlockStaticMutex(&teamArrayLock);
			if( error )
				retval = "No memory";
		}
		break;
	case JSI_CREATE_PROPERTY:
		if( !TeamClass )
		{
			TeamClass = (*env)->FindClass(
				env, "edu/utah/janosvm/sys/Team");
			TeamCreateMethod = (*env)->GetStaticMethodID(
				env,
				TeamClass,
				"create",
				"(Ljava/lang/String;"
				"[Ledu/utah/janosvm/resources/ResourceSpec;"
				"Lkaffe/util/Ptr;)"
				"Ledu/utah/janosvm/sys/TeamHandle;");
			TeamGetIDMethod = (*env)->GetMethodID(
				env,
				TeamClass,
				"getID",
				"()I");
			
			ensure(TeamClass != 0);
			ensure(TeamCreateMethod != 0);
			ensure(TeamGetIDMethod != 0);
		}
		if( jsiMessageGet(jm,
				  JMA_String, "name", 0, &name,
				  JMA_TAG_DONE) )
		{
			jName = stringC2Java(name);
		}
		nSpecs = createNativeSpecs(jm);
		/* We can't let anyone kill the team before we get the ID */
		if( (th = (*env)->CallStaticObjectMethod(env,
							 TeamClass,
							 TeamCreateMethod,
							 jName,
							 0,
							 nSpecs)) )
		{
			int id;
			
			jsiClearMessage(jm);
			id = (*env)->CallIntMethod(env,
						   th,
						   TeamGetIDMethod);
			if( id != 0 )
			{
				jsiMessageAdd(
					jm,
					JMA_What, JSI_REPLY,
					JMA_UnsignedInteger, "result", id,
					JMA_TAG_DONE);
			}
			else
			{
				retval = "Team was killed";
			}
		}
		else
		{
			printExecuteException("Team");
			retval = "Unable to create team";
		}
		deleteNativeSpecs(nSpecs);
		break;
	case JSI_DELETE_PROPERTY:
		if( jsiPopMessageSpecifier(jm, &spec_name, &spec_value) )
		{
			sparseID_t tid;
			jteam *jt;
			
			if( (sscanf(spec_value, "%d", &tid) == 1) )
			{
				if( tid != 1 )
				{
					if( (jt = grabTeamWithID(tid)) )
					{
						terminateTeam(jt);
						dropTeam(jt);
					}
					jsiClearMessage(jm);
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_String, "result", "ok",
						JMA_TAG_DONE);
				}
				else
				{
					retval = "Don't delete the kernel "
						"team, use shutdown "
						"if you want to stop the VM.";
				}
			}
			else
			{
				retval = "Bad team ID";
			}
		}
		break;
	case JSI_EXECUTE_PROPERTY:
		if( jsiPopMessageSpecifier(jm, &spec_name, &spec_value) )
		{
			sparseID_t tid;
			jteam *jt;

			sscanf(spec_value, "%d", &tid);
			if( (sscanf(spec_value, "%d", &tid) == 1) &&
			    (jt = grabTeamWithID(tid)) )
			{
				if( jsiPopMessageSpecifier(jm,
							   &spec_name,
							   &spec_value) )
				{
					if( !strcmp(spec_name, "wait") )
					{
						waitTeam(jt,
							 WAIT_FOR_TEAM_BURIAL);
						jsiClearMessage(jm);
						jsiMessageAdd(jm,
							      JMA_What,
							      JSI_REPLY,
							      JMA_String,
							      "result",
							      "All your resources are belong to us!",
							      JMA_TAG_DONE);
					}
					else
					{
						retval = "Not an executable property";
					}
				}
				else
				{
					retval = "No executable property specified";
				}
				dropTeam(jt);
			}
			else
			{
				jsiClearMessage(jm);
				jsiMessageAdd(jm,
					      JMA_What, JSI_REPLY,
					      JMA_String,
					      "result", "All your resources are belong to us!",
					      JMA_TAG_DONE);
			}
		}
		break;
	case JSI_GET_SUITES:
		jsiClearMessage(jm);
		if( !jsiMessageAdd(jm,
				   JMA_What, JSI_REPLY,
				   JMA_PropertyList, "result", teamProperties,
				   JMA_TAG_DONE) )
		{
			retval = "Out of memory";
		}
		break;
	default:
		retval = "Team doesn't support this operation";
		break;
	}
	return( retval );
}

int teamGetRootSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", teamRootProperties,
			       JMA_TAG_DONE);
	return( retval );
}
