/*
 * leak.c
 * A debugging tool for tracking memory leaks.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdio.h>
#include <stdarg.h>

#include "assert_pp.h"
#include "gtypes.h"
#include "gc.h"
#include "registry.h"
#include "leak.h"
#include "logResource.h"

/* Don't bother compiling this if we're not even debugging. */
#if defined(DEBUG) && defined(DEBUG_LEAK)

/*
 * The memoryLeak structure is used to record what memory is being leaked and
 * who originally allocated it.
 *
 * `ptr' - Pointer to the memory that is suspected of leaking.
 * `note' - A note to distinguish this memory from others.
 * `file' - The source file that registered the leak.
 * `line' - The line of the source file that registered the leak.
 */
struct memoryLeak {
	void *ptr;
	char *note;
	char *file;
	int line;
};

/* All leaks are kept in a big array, nothing too sophisticated here... */
#define LEAK_TRACK_MAX 32768
static int memory_leak_count;
static struct memoryLeak memory_leaks[LEAK_TRACK_MAX];

/* The flag indicating whether we're tracking leaks or not */
int tracking_leaks = 1;

int trackingLeaks(void)
{
	return( tracking_leaks );
}

void trackLeaks(int yes_no)
{
	tracking_leaks = yes_no;
}

void *real_registerLeak(void *ptr, char *note, char *file, int line)
{
	require(ptr != 0);
	require(file != 0);
	require(line >= 0);
	
	if( tracking_leaks )
	{
		/* Just put it at the end of the array */
		if( (memory_leak_count + 1) < LEAK_TRACK_MAX )
		{
			struct memoryLeak *ml;

			ml = &memory_leaks[memory_leak_count];
			ml->ptr = ptr;
			ml->note = note;
			ml->file = file;
			ml->line = line;
			memory_leak_count++;
			teamLogEntry("leak",
				     LEA_String, "register-leak", "", -1,
				     LEA_Pointer, "leak-ptr", ptr,
				     LEA_String, "note", note, -1,
				     LEA_String, "file", file, -1,
				     LEA_Integer, "line", line,
				     LEA_TAG_DONE);
		}
	}
	return( ptr );
}

int isLeak(void *ptr)
{
	int retval = 0;

	if( tracking_leaks )
	{
		int lpc;
		
		for( lpc = 0; !retval && (lpc < memory_leak_count); lpc++ )
		{
			if( memory_leaks[lpc].ptr == ptr )
				retval = 1;
		}
	}
	return( retval );
}

void unregisterLeak(void *ptr)
{
	if( tracking_leaks )
	{
		int lpc, found = 0;
		
		for( lpc = 0; lpc < memory_leak_count; lpc++ )
		{
			/* Catch multiple entries of the same pointer */
			if( memory_leaks[lpc].ptr == ptr )
			{
				found = 1;
				memory_leak_count--;
			}
			if( found )
			{
				memory_leaks[lpc] = memory_leaks[lpc + 1];
			}
		}
		if( found )
		{
			teamLogEntry("leak",
				     LEA_String, "unregister-leak", "", -1,
				     LEA_Pointer, "leak-ptr", ptr,
				     LEA_TAG_DONE);
		}
	}
}

void clearLeaks(void)
{
	memory_leak_count = 0;
}

void printLeak(void *ptr, void *from)
{
	if( tracking_leaks )
	{
		int lpc, found = -1;
		
		for( lpc = 0; (found < 0) && (lpc < memory_leak_count); lpc++ )
		{
			if( memory_leaks[lpc].ptr == ptr )
				found = lpc;
		}
		if( found >= 0 )
		{
			struct memoryLeak *ml = &memory_leaks[found];
			char adesc[100];
			
			if( from )
			{
				strncpy(adesc, describeObject(from),
					sizeof(adesc) - 1);
			}
			else
			{
				adesc[0] = 0;
			}
			teamLogEntry("leak",
				     LEA_String, "mark-leak", "leak is still alive", -1,
				     LEA_Pointer, "leak-ptr", ml->ptr,
				     LEA_String, "leak-type", describeObject(ptr), -1,
				     LEA_String, "file", ml->file, -1,
				     LEA_Integer, "line", ml->line,
				     LEA_String, "note", ml->note, -1,
				     LEA_String, "from-type", adesc, -1,
				     LEA_Pointer, "from", from,
				     LEA_TAG_DONE);
		}
	}
}

void printLeaks(void)
{
	if( tracking_leaks )
	{
		int lpc;
		
		fprintf(stderr, "memory leaks:\n");
		for( lpc = 0; lpc < memory_leak_count; lpc++ )
		{
			fprintf(stderr, "  leak %p %s:%d - %s\n",
				memory_leaks[lpc].ptr,
				memory_leaks[lpc].file,
				memory_leaks[lpc].line,
				memory_leaks[lpc].note);
			fflush(stderr);
		}
	}
}

#endif /* defined(DEBUG) */
