/*
 * logResource.c
 * Resource to managed standard I/O for each team.
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "assert_pp.h"

#include "logResource.h"
#include "memResource.h"
#include "jteam.h"
#include "nativeClass.h"
#include "nativeClass-inline.h"
#include "mungWallProxy.h"
#include "scripting.h"
#include "module.h"
#include "moduleList.h"

#include <misc/stringParsing.h>

/** Module declarations **/
static int initLogResourceModule(int stage);

module_t logResourceModule = {
	"logResource",
	0,
	initLogResourceModule,
	0,
	0
};

/** LogResource declarations **/
static int logSubscribe(struct _resource *res,
			struct _resourceClient *rc,
			struct _resourceSpec *rs);
static void logUnsubscribe(struct _resource *res,
			   struct _resourceClient *rc);

static resourceFunctions logResourceFunctions = {
	0,		/* install */
	0,		/* uninstall */
	0,		/* configure */
	0,		/* fillSpec */
	0,		/* createJSISpec */
	0,		/* deleteJSISpec */
	0,		/* defaultSpec */
	logSubscribe,
	0,		/* enable */
	logUnsubscribe
};

logResource metalLogResource;

static logSpec defaultSpec = {
	{ &metalLogResource.link }
};

/** Module Definitions **/
static
void logAtExit(void)
{
	logClient *lc;
	jteam *jt;

	if( kernelTeam )
		jt = kernelTeam;
	else
		jt = metalTeam;
	if( (lc = (logClient *)
	     getTeamInternalResource(jt, &metalLogResource.link)) )
	{
		flushLogger(&lc->log);
	}
}

int initLogResourceModule(int stage)
{
	int retval = 0;

	switch( stage )
	{
	case SID_LOGRESOURCEMODULE_INITMETAL:
		installResource(&metalLogResource.link,
				RA_Name, "log",
				RA_Flags, RF_INTERNAL,
				RA_Functions, &logResourceFunctions,
				RA_ClientSize, sizeof(logClient),
				RA_SpecSize, sizeof(logSpec),
				RA_TAG_DONE);
		retval = attachTeamResource(metalTeam, &defaultSpec.spec);
		atexit(logAtExit);
		break;
	case SID_LOGRESOURCEMODULE:
		retval = 1;
		break;
	}
	return( retval );
}

int logSubscribe(struct _resource *res,
		 struct _resourceClient *rc,
		 struct _resourceSpec *rs)
{
	logClient *lc = (logClient *)rc;
	// logSpec *ls = (logSpec *)rs;
	int retval = 0;

	if( initLogger(&lc->log,
		       LA_Name, rc->owner->name,
		       LA_TAG_DONE) )
	{
		retval = 1;
	}
	return( retval );
}

void logUnsubscribe(struct _resource *res,
		    struct _resourceClient *rc)
{
	logClient *lc = (logClient *)rc;

	killLogger(&lc->log);
}

void teamLog_real(char *srcFile, int srcLine, char *cl, char *msg)
{
	logClient *lc = (logClient *)
		getTeamInternalResource(0, &metalLogResource.link);
	logger *l = &lc->log;
	int iLockRoot;
	
	if( lc )
	{
		lockMutex(lc);
		addLogMessage_real(l, srcFile, srcLine, cl, msg);
		unlockMutex(lc);
	}
}

void teamLogEntry_real(char *srcFile, int srcLine, char *cl, ...)
{
	logClient *lc = (logClient *)
		getTeamInternalResource(0, &metalLogResource.link);
	logger *l = &lc->log;
	va_list args;
	int iLockRoot;

	if( lc )
	{
		lockMutex(lc);
		va_start(args, cl);
		addLogEntryV(l, srcFile, srcLine, cl, args);
		va_end(args);
		unlockMutex(lc);
	}
}

void addTeamLogAnalyzer(logAnalyzerMask *lam)
{
	logClient *lc = (logClient *)
		getTeamInternalResource(0, &metalLogResource.link);
	logger *l = &lc->log;
	int iLockRoot;

	lockMutex(lc);
	addLogAnalyzer(l, lam);
	unlockMutex(lc);
}

void remTeamLogAnalyzer(logAnalyzerMask *lam)
{
	logClient *lc = (logClient *)
		getTeamInternalResource(0, &metalLogResource.link);
	logger *l = &lc->log;
	int iLockRoot;

	lockMutex(lc);
	remLogAnalyzer(l, lam);
	unlockMutex(lc);
}
