/*
 * memResource.c
 * Memory resource for teams, just wraps the collector
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "assert_pp.h"
#include "tags.h"
#include "memResource.h"
#include "jteam.h"
#include "locks.h"
#include "thread.h"
#include "nativeClass-inline.h"
#include "module.h"
#include "moduleList.h"
#include "support.h"
#include "gcProxy.h"
#include "mungWallProxy.h"
#include "jteamReservation.h"
#include "leak.h"
#include "threadData-inline.h"
#include "logResource.h"
#include "exception.h"
#include <misc/memoryStack.h>

#include "edu_utah_janosvm_resources_MemorySpec.h"
#include "edu_utah_janosvm_resources_MemoryResource.h"

#define JAVA_MEMORY_SPEC "edu/utah/janosvm/resources/MemorySpec"

/** Module declarations **/
static int initMemResourceModule(int stage);

module_t memResourceModule = {
	"memory",
	0,
	initMemResourceModule,
	0,
	0,
};

/** Resource declarations **/
static int memConfigure(struct _resource *res,
			sectionFile *sf,
			tagSectionData *tsd);
static int memFillSpec(struct _resource *res,
		       struct _resourceSpec *rs,
		       sectionFile *sf,
		       tagSectionData *tsd);
static struct _resourceSpec *memCreateJSISpec(struct _resource *res,
					      jsiMessage *jm);
static void memDeleteJSISpec(struct _resource *res,
			     struct _resourceSpec *rs);
static struct _resourceSpec *memDefaultSpec(struct _resource *res);
static int memSubscribe(struct _resource *res,
			struct _resourceClient *rc,
			struct _resourceSpec *rs);
static int memEnable(struct _resource *res,
		     struct _resourceClient *rc);
static void memUnsubscribe(struct _resource *res,
			   struct _resourceClient *rc);

static resourceFunctions memResourceFunctions = {
	0,
	0,
	memConfigure,
	memFillSpec,
	memCreateJSISpec,
	memDeleteJSISpec,
	memDefaultSpec,
	memSubscribe,
	memEnable,
	memUnsubscribe
};

/** memResource variables **/

memResource metalMemResource;

/*
 * We statically allocate a client for the metal team so that we can get it set
 * up early on.
 */
memResourceClient metalMemResourceClient = {
	{
		0,
		0,
		&metalMemResource.link,
		0
	},
	0,
	32 * 1024 * 1024,
	0,
	0,
	64 * 1024 * 1024
};

static tagDesc memResourceTags[] = {
	{ "max", TAG_INTEGER },
	{ 0, TAG_NONE }
};

static tagData memResourceTagData[1];

static tagArray memResourceDefault = {
	memResourceTags,
	memResourceTagData
};

static tagDesc memResourceSpecTags[] = {
	{ "max", TAG_INTEGER },
	{ 0, TAG_NONE }
};

static tagData memResourceSpecTagData[1];

static tagArray memResourceSpecDefault = {
	memResourceSpecTags,
	memResourceSpecTagData
};

#define DEFAULT_SPEC_MAX 8 * 1024 * 1024

static memResourceSpec metalMemResourceSpec = {
	{ &metalMemResource.link },
	DEFAULT_SPEC_MAX,
	0
};

/** Module functions **/
int initMemResourceModule(int stage)
{
	switch( stage )
	{
	case SID_MEMRESOURCEMODULE_INITMETAL:
		metalMemResource.min = 512 * 1024; // XXX
		metalMemResource.max = 64 * 1024 * 1024;
		installResource(&metalMemResource.link,
				RA_Name, "memory",
				RA_Flags, RF_INTERNAL,
				RA_Functions, &memResourceFunctions,
				RA_ClientSize, sizeof(memResourceClient),
				RA_SpecSize, sizeof(memResourceSpec),
				RA_TAG_DONE);

		metalTeam->internalResources[metalMemResource.link.
					    internalIndex] =
			&metalMemResourceClient.client;
		metalMemResourceClient.client.owner = metalTeam;
		break;
	case SID_MEMRESOURCEMODULE:
		memResourceTagData[0].i = 64 * 1024 * 1024;
		memResourceSpecTagData[0].i = DEFAULT_SPEC_MAX;
		
		installResource(&metalMemResource.link,
				RA_ResourceSectionName, "memory",
				RA_ResourceSectionDefault, &memResourceDefault,
				RA_SpecSectionName, "memspec",
				RA_SpecSectionDefault, &memResourceSpecDefault,
				RA_SpecNativeClass, JAVA_MEMORY_SPEC,
				RA_TAG_DONE);
		break;
	}
	return( 1 );
}

/** Resource functions **/
int memConfigure(struct _resource *res,
		 sectionFile *sf,
		 tagSectionData *tsd)
{
	memResource *mr = (memResource *)res;
	
	mr->max = getTagData(&tsd->tags, "min").i;
	mr->max = getTagData(&tsd->tags, "max").i;
	return( 1 );
}

int memFillSpec(struct _resource *res,
		struct _resourceSpec *rs,
		sectionFile *sf,
		tagSectionData *tsd)
{
	memResourceSpec *mrs = (memResourceSpec *)rs;
	
	mrs->max = getTagData(&tsd->tags, "max").i;
	return( 1 );
}

resourceSpec *memCreateJSISpec(struct _resource *res, jsiMessage *jm)
{
	resourceSpec *retval = 0;
	char *policy = 0;
	int max = 0;
	
	jsiMessageGet(jm,
		      JMA_String, "memory.policy", 0, &policy,
		      JMA_Integer, "memory.max", 0, &max,
		      JMA_TAG_DONE);
	if( (policy || max) &&
	    (retval = createResourceSpec(res)) )
	{
		memResourceSpec *mrs = (memResourceSpec *)retval;
		
		mrs->max = max;
		mrs->policyClass = policy;
	}
	return( retval );
}

void memDeleteJSISpec(struct _resource *res, struct _resourceSpec *rs)
{
	deleteResourceSpec(rs);
}

resourceSpec *memDefaultSpec(struct _resource *res)
{
	return &metalMemResourceSpec.spec;
}

extern void gcKillMe(void *me);

int memSubscribe(struct _resource *res,
		 struct _resourceClient *rc,
		 struct _resourceSpec *rs)
{
	memResourceClient *mrc = (memResourceClient *)rc;
	memResourceSpec *mrs = (memResourceSpec *)rs;
	memResource *mr = (memResource *)res;
	int retval = 0;
	int iLockRoot;

	if( mrs )
		mrc->max = mrs->max;
	else
		mrc->max = DEFAULT_SPEC_MAX;
	mrc->gcTrigger = mrc->max;
	lockMutex(res);
	/* Make sure the resource has enough to give */
	if( (mrc->max >= mr->min) &&
	    ((mr->used + mrc->max) < mr->max) )
	{
		if( (mrc->collector = newCollector(mrc,
						   mrs ?
						   mrs->policyClass :
						   0)) )
		{
DBG(MUNGWALL,
			addMungWallProxy(mrc->collector));
			mr->used += mrc->max;
			retval = 1;
		}
		else
		{
			gcKillMe(mrc->collector);
		}
	}
	else if( mrc->max < mr->min )
	{
		teamLogEntry("memResource",
			     LEA_String, "Error",
			     "Memory request too small", -1,
			     LEA_Integer, "size", mrc->max,
			     LEA_TAG_DONE);
	}
	else
	{
		teamLogEntry("memResource",
			     LEA_String, "Error",
			     "Memory request too big", -1,
			     LEA_Integer, "size", mrc->max,
			     LEA_TAG_DONE);
	}
	unlockMutex(res);
	if( retval )
	{
		jteamReservation *jr;
		void *rpc, *fp;

		/* assume the worst */
		retval = 0;
		
		getCallerFrameAndPC(&fp, &rpc);
		if( pushTeamReservation_raw(rc->owner, rpc, fp, false, &jr) ==
		    JRV_OK )
		{
			if( visitReservedTeam_raw(
				jthread_get_data(jthread_current()),
				jr, rpc, fp, 0) )
			{
				retval = 1;
			}
			else
			{
				popTeamReservation(jr);
			}
		}
	}
	return( retval );
}

int memEnable(struct _resource *res, struct _resourceClient *rc)
{
	memResourceClient *mrc = (memResourceClient *)rc;
	int retval;

	require(currentTeam() == rc->owner);

	/* Start the collector */
	retval = GC_enable((Collector *)mrc->collector);
	/* XXX Need to prime the GC... */
	gc_free(gc_malloc(1, GC_ALLOC_FIXED));
	returnFromTeam(); /* XXX Is this really evil? */
	return( retval );
}

void memUnsubscribe(struct _resource *res, struct _resourceClient *rc)
{
	memResourceClient *mrc = (memResourceClient *)rc;
	memResource *mr = (memResource *)res;
	int iLockRoot;

	require(jthread_get_data(jthread_current())->currentTeam == rc->owner);
	
	lockMutex(res);
	/* Give the memory back to the resource */
	mr->used -= mrc->max;
	unlockMutex(res);
	/*
	 * Kill the collector, basically just free the pages that this client
	 * was using.
	 */
DBG(MUNGWALL,
	quickChangeTeamReservation(kernelTeam);
	unwrapCollector(mrc->collector);
	quickChangeTeamReservation(rc->owner);
    );
	gcKillMe(mrc->collector);

	ensure(jthread_get_data(jthread_current())->currentTeam == rc->owner);
}

int chargeMemory(memResourceClient *mrc, unsigned int amount)
{
	int retval = 0;

	require(mrc != 0);
	require(mrc->used >= 0);
	require(mrc->used <= mrc->max);
	require((amount % gc_pgsize) == 0);

	if( (mrc->used + amount) <= mrc->max )
	{
		mrc->used += amount;
		mrc->peak = MAX(mrc->peak, mrc->used);
		retval = 1;
	}
	return( retval );
}

void creditMemory(memResourceClient *mrc, unsigned int amount)
{
	require(mrc != 0);
	require(mrc->used >= 0);
	require(mrc->used <= mrc->max);
	require(amount <= mrc->used);
	require((amount % gc_pgsize) == 0);

	mrc->used -= amount;

	ensure(mrc->used >= 0);
	ensure(mrc->used <= mrc->max);
}

void edu_utah_janosvm_resources_MemorySpec_nativeConstructor(
	struct Hedu_utah_janosvm_resources_MemorySpec *jmrs,
	jint max,
	struct Hjava_lang_String *jStr)
{
	memResourceSpec *mrs = OBJECT2NATIVE(&metalMemResource.link.
					     specNativeClass,
					     (Hjava_lang_Object *)jmrs);

	mrs->spec.handler = &metalMemResource.link;
	mrs->max = max;
	mrs->policyClass = jStr ? stringJava2C(jStr) : 0;
}

jint edu_utah_janosvm_resources_MemorySpec_getMax(
	struct Hedu_utah_janosvm_resources_MemorySpec *jmrs)
{
	memResourceSpec *mrs = OBJECT2NATIVE(&metalMemResource.link.
					     specNativeClass,
					     (Hjava_lang_Object *)
					     jmrs);

	return( mrs->max );
}

struct Hjava_lang_String *
edu_utah_janosvm_resources_MemorySpec_getPolicyClass(
	struct Hedu_utah_janosvm_resources_MemorySpec *jmrs)
{
	memResourceSpec *mrs = OBJECT2NATIVE(&metalMemResource.link.
					     specNativeClass,
					     (Hjava_lang_Object *)
					     jmrs);
	
	return( (Hjava_lang_String *)
		checkPtr(stringC2Java(mrs->policyClass)) );
}

void edu_utah_janosvm_resources_MemorySpec_nativeFinalizer(
	struct Hedu_utah_janosvm_resources_MemorySpec *jmrs)
{
	if( !jmrs->found )
	{
		memResourceSpec *mrs = OBJECT2NATIVE(&metalMemResource.link.
						     specNativeClass,
						     (Hjava_lang_Object *)
						     jmrs);

		gc_free(mrs->policyClass);
	}
}

struct Hedu_utah_janosvm_resources_MemorySpec *
edu_utah_janosvm_resources_MemorySpec_currentSpec(void)
{
	memResourceClient *mrc = (memResourceClient *)
		getTeamInternalResource(0, &metalMemResource.link);
	struct Hedu_utah_janosvm_resources_MemorySpec *retval = 0;
	memResourceSpec *mrs;
	
	retval = (struct Hedu_utah_janosvm_resources_MemorySpec *)
		newObject(metalMemResource.link.specNativeClass.peer);
	mrs = OBJECT2NATIVE(&metalMemResource.link.specNativeClass,
			    (Hjava_lang_Object *)retval);
	mrs->spec.handler = &metalMemResource.link;
	mrs->max = mrc->max;
	/* XXX We're not doing the policy class here... */
	return( retval );
}

struct Hedu_utah_janosvm_resources_MemorySpec *
edu_utah_janosvm_resources_MemorySpec_defaultSpec(void)
{
	struct Hedu_utah_janosvm_resources_MemorySpec *retval = 0;
	memResourceSpec *mrs;
	
	retval = (struct Hedu_utah_janosvm_resources_MemorySpec *)
		newObject(metalMemResource.link.specNativeClass.peer);
	mrs = OBJECT2NATIVE(&metalMemResource.link.specNativeClass,
			    (Hjava_lang_Object *)retval);
	*mrs = metalMemResourceSpec;
	return( retval );
}

Hjava_lang_Object *
edu_utah_janosvm_resources_MemorySpec_clone(
	struct Hedu_utah_janosvm_resources_MemorySpec *jcps)
{
	struct Hedu_utah_janosvm_resources_MemorySpec *retval;
	memResourceSpec *mrs, *retval_mrs;

	retval = (struct Hedu_utah_janosvm_resources_MemorySpec *)
		newObject(metalMemResource.link.specNativeClass.peer);
	mrs = OBJECT2NATIVE(&metalMemResource.link.specNativeClass,
			   (Hjava_lang_Object *)jcps);
	retval_mrs = OBJECT2NATIVE(&metalMemResource.link.specNativeClass,
				   (Hjava_lang_Object *)retval);
	retval_mrs->spec.handler = &metalMemResource.link;
	retval_mrs->max = mrs->max;
	retval_mrs->policyClass = (char *)
		checkPtr(gc_malloc(strlen(mrs->policyClass) + 1,
				   GC_ALLOC_FIXED));
	strcpy(retval_mrs->policyClass, mrs->policyClass);
	return( (Hjava_lang_Object *)retval );
}

jlong edu_utah_janosvm_resources_MemoryResource_getUsage(void)
{
	memResourceClient *mrc = (memResourceClient *)
		getTeamInternalResource(0, &metalMemResource.link);
	
	return mrc->used;
}

jlong edu_utah_janosvm_resources_MemoryResource_getPeakUsage(void)
{
	memResourceClient *mrc = (memResourceClient *)
		getTeamInternalResource(0, &metalMemResource.link);
	
	return mrc->peak;
}

jlong edu_utah_janosvm_resources_MemoryResource_getMax(void)
{
	memResourceClient *mrc = (memResourceClient *)
		getTeamInternalResource(0, &metalMemResource.link);
	
	return mrc->max;
}

void edu_utah_janosvm_resources_MemoryResource_registerLeak(struct Hjava_lang_Object *obj)
{
	if( obj != 0 )
		registerLeak(obj, "MemoryResource.registerLeak()");
	else
		throwException(NullPointerException);
}

void edu_utah_janosvm_resources_MemoryResource_unregisterLeak(struct Hjava_lang_Object *obj)
{
	if( obj != 0 )
		unregisterLeak(obj);
	else
		throwException(NullPointerException);
}


void java_lang_System_registerLeak(struct Hjava_lang_Object *obj)
{
	if( obj != 0 )
		registerLeak(obj, "java.lang.System.registerLeak()");
	else
		throwException(NullPointerException);
}

void java_lang_System_unregisterLeak(struct Hjava_lang_Object *obj)
{
	if( obj != 0 )
		unregisterLeak(obj);
	else
		throwException(NullPointerException);
}

