/*
 * memResource.h
 * Memory resource for teams, just wraps the collector
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _mem_resource_h
#define _mem_resource_h

/*
 * This module exports a memory resource for use by jteams.
 */

#include "assert_pp.h"

#include "resource.h"
#include "thread.h"

/*
 * The memResource structure is used to track memory usage for the system as
 * a whole.
 *
 * `link' - The resource header.
 * `used' - The total memory reserved by all teams in the system, except the
 *   kernel.
 * `min' - The minimum required memory for a team.
 * `max' - The maximum amount of memory in the system.
 */
typedef struct _memResource {
	resource link;
	uint32 used;
	uint32 min;
	uint32 max;
} memResource;

/*
 * The memResourceClient is a resource client that tracks memory usage for
 * the owner.
 *
 * `client' - The resourceClient header.
 * `collector' - The garbage collector for this client.
 * `used' - The amount of memory currently used.
 * `peak' - The peak usage for this team.
 * `max' - The maximum allowed memory usage.
 */
typedef struct _memResourceClient {
	resourceClient client;
	void *collector;
	uint32 gcTrigger;
	uint32 used;
	uint32 peak;
	uint32 max;
} memResourceClient;

/*
 * The memResourceSpec is used to specify the parameters for memory usage.
 *
 * `max' - The maximum allowed usage.
 * `policyClass' - The name of the class to use for the GC policy.
 */
typedef struct _memResourceSpec {
	resourceSpec spec;
	uint32 max;
	char *policyClass;
} memResourceSpec;

/*
 * Attempt to charge the client for the given amount of memory.  Returns true
 * if the amount is under the maximum allowed usage.
 *
 * `mrc' - The memory client to charge.
 * `amount' - The number of bytes to charge the client for.  Needs to be a
 *  multiple of the gc page size.
 */
int chargeMemory(memResourceClient *mrc, unsigned int amount);
/*
 * Credit the client for the given amount of memory.
 *
 * `mrc' - The memory client to credit.
 * `amount' - The number of bytes to credit the client.  Needs to be a
 *  multiple of the gc page size.
 */
void creditMemory(memResourceClient *mrc, unsigned int amount);

/*
 * The metal level memory resource.
 */
extern memResource metalMemResource;

#include "jteam.h"
#include "jteam-inline.h"
#include "memResource-inline.h"

/*
 * AddModule: memResourceModule
 *   ModuleDependency: jteamModule.initMetal
 *   Comment: Install the memory resource/collector and subscribe the metal
 *   Comment: team to it
 * Transition: initMetal
 *   ModuleDependency: resourceModule.reinstall
 *   Comment: We need to reinstall to get the section types
 */

#endif
