/*
 * nameSpace.h
 * Manage the root namespace for a team
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "gtypes.h"
#include "support.h"
#include "locks.h"
#include "stringSupport.h"
#include "jteam.h"
#include "jteamReservation.h"
#include "nameSpace.h"
#include "thread.h"
#include "nativeClass.h"
#include "classPath.h"
#include "readClass.h"
#include "itypes.h"
#include "stdioResource.h"
#include "assert_pp.h"
#include "nativeClass-inline.h"
#include "threadData.h"
#include "threadData-inline.h"
#include "mungWallProxy.h"
#include "logResource.h"
#include "module.h"
#include "moduleList.h"
#include "access.h"

#include <misc/stringParsing.h>
#include <jsi/jsi.h>

#include "edu_utah_janosvm_sys_Importable.h"
#include "edu_utah_janosvm_sys_PortAuthority.h"

#include "edu_utah_janosvm_resources_NameSpaceGroup.h"
#include "edu_utah_janosvm_resources_NameSpaceResource.h"
#include "edu_utah_janosvm_resources_NameSpaceSpec.h"

#define JAVA_NAME_SPACE_SPEC "edu/utah/janosvm/resources/NameSpaceSpec"

struct Hedu_utah_janosvm_sys_Importable;
nativeClass importNativeClass;
Hjava_lang_Class *ImportableClass;
Hjava_lang_Class *ExportableClass;
Hjava_lang_Class *ListClass;
Hjava_lang_Class *ListNodeClass;
Hjava_lang_Class *ObjectHolderClass;
Hjava_lang_Class *CrossTeamCloneableClass;

nameSpaceResource metalNameSpaceResource;

Utf8Const *utfKernel;
Utf8Const *utfIdentity;

/** classGroup sectionType **/
typedef struct _classGroupSectionData {
	tagSectionData tsd;
	classGroupType *cgt;
} classGroupSectionData;

static sectionTypeFunctions nsClassGroupSectionFunctions;

static tagSectionType nsClassGroupSection;

int nsClassGroupSectionWalker(struct _sectionType *st,
			      struct _sectionFileData *sfd,
			      struct _sectionFile *sf)
{
	Utf8Const *utfName;
	int retval = 0;

	if( (utfName = utf8ConstNew(childSectionName(sfd), -1)) )
	{
		classGroupSectionData *cgsd =
			LOCAL_SECTION_DATA(classGroupSectionData *, st, sfd);
		
		/* Allocate the group type and add it to the resource */
		if( (cgsd->cgt = createClassGroupType(utfName,
						      cgsd->tsd.taglist)) )
		{
			cgsd->cgt->next = metalNameSpaceResource.types;
			metalNameSpaceResource.types = cgsd->cgt;
			retval = 1;
		}
		else
		{
			utf8ConstRelease(utfName);
		}
	}
	return( retval );
}

/** viewGroup sectionType **/
static sectionTypeFunctions nsViewGroupSectionFunctions;

static tagSectionType nsViewGroupSection;

int nsViewGroupSectionWalker(struct _sectionType *st,
			     struct _sectionFileData *sfd,
			     struct _sectionFile *sf)
{
	Utf8Const *utfName;
	int retval = 0;

	if( (utfName = utf8ConstNew(childSectionName(sfd), -1)) )
	{
		tagSectionData *tsd = LOCAL_SECTION_DATA(tagSectionData *,
							 st,
							 sfd);
		classGroupSectionData *cgsd;
		sectionFileData *par_sfd;

		/* Find the parent section, should be a classGroup */
		par_sfd = getSectionParent(sf, sfd);
		if( par_sfd && (par_sfd->type == &nsClassGroupSection.type) )
		{
			viewGroupType *vgt;
			
			cgsd = LOCAL_SECTION_DATA(classGroupSectionData *,
						  &nsClassGroupSection,
						  par_sfd);

			/* Add the view group to this class group */
			if( (vgt = createViewGroupType(utfName,
						       tsd->taglist)) )
			{
				vgt->next = cgsd->cgt->views;
				cgsd->cgt->views = vgt;
				retval = 1;
			}
			else
			{
				retval = 0;
			}
		}
		else
		{
			retval = 0;
		}
		if( !retval )
			utf8ConstRelease(utfName);
	}
	return( retval );
}

/** nameSpace helper functions **/

/* Add a classGroup/viewGroup pair to a nameSpaceSpec */
static
int subscribeView(nameSpaceSpec *nss,
		  sparseID_t id, Utf8Const *cgName, Utf8Const *vgName)
{
	groupNames *new_data;
	int retval = 0;

	if( (new_data = gc_realloc(nss->viewSubscriptions.data,
				   (nss->viewSubscriptions.size + 1) *
				   sizeof(groupNames),
				   GC_ALLOC_NAME_SPACE)) )
	{
		int slot;
		
		nss->viewSubscriptions.data = new_data;
		slot = nss->viewSubscriptions.size++;
		nss->viewSubscriptions.data[slot].teamID = id;
		nss->viewSubscriptions.data[slot].className = cgName;
		nss->viewSubscriptions.data[slot].viewName = vgName;
		retval = 1;

		ensure_valid_memory(new_data);
	}
	return( retval );
}

/* Add a classGroup/viewGroup pair to a nameSpaceClient */
static
int subscribeViewClient(nameSpaceClient *nsc, classGroup *cg, viewGroup *vg)
{
	viewSubscriberPool **last_pool, *pool;
	int retval = 0;
	int iLockRoot;

	lockMutex(nsc);
	last_pool = &nsc->views;
	pool = nsc->views;
	while( pool && (pool->used == VIEW_SUBSCRIBER_POOL_SIZE) )
	{
		last_pool = &pool->next;
		pool = pool->next;
	}
	if( !pool )
	{
		pool = gc_malloc(sizeof(viewSubscriberPool),
				 GC_ALLOC_NAME_SPACE);
	}
	if( pool )
	{
		viewSubscriber *vs;

		vs = &pool->data[pool->used];
		pool->used++;
		vs->cGroup = cg;
		vs->vGroup = vg;
		/* This must be the last instruction */
		*last_pool = pool;
		retval = 1;
	}
	unlockMutex(nsc);
	return( retval );
}

/** Resource functions **/
static
int nsResourceFillSpec(struct _resource *res,
		       struct _resourceSpec *rs,
		       sectionFile *sf,
		       tagSectionData *tsd)
{
	nameSpaceSpec *nss = (nameSpaceSpec *)rs;
	int lpc, retval = 1;
	tagList *tl = tsd->taglist;

	/*
	 * Walk the tag list interpreting it as each tag is the classGroup
	 * name and the value is the viewGroup name.
	 */
	for( lpc = 0; (lpc < tl->tags.count) && retval; lpc++ )
	{
		retval = subscribeView(nss,
				       0,
				       tl->tags.data[lpc].tag,
				       tl->tags.data[lpc].data.str);
	}
	return( retval );
}

static
resourceSpec *nsCreateJSISpec(struct _resource *res, jsiMessage *jm)
{
	resourceSpec *retval = 0;
	int count;
	
	count = jsiMessageIdentifierCount(jm, "namespace");
	if( count && (retval = createResourceSpec(res)) )
	{
		nameSpaceSpec *nss = (nameSpaceSpec *)retval;
		int error = 0, lpc;

		if( (nss->viewSubscriptions.data =
		     gc_malloc(sizeof(groupNames) * count,
			       GC_ALLOC_NAME_SPACE)) )
		{
			nss->viewSubscriptions.size = count;
			for( lpc = 0; lpc < count; lpc++ )
			{
				parsedString cName, vName;
				groupNames *gn;
				char *val;

				gn = &nss->viewSubscriptions.data[lpc];
				jsiMessageGet(
					jm,
					JMA_String, "namespace", lpc, &val,
					JMA_TAG_DONE);
				if( parseString(val,
						SPO_Integer, &gn->teamID,
						SPO_Expect, ":",
						SPO_String, &cName,
						SPO_Expect, ":",
						SPO_String, &vName,
						SPO_End) )
				{
					gn->className = utf8ConstNew(
						cName.data, cName.len);
					gn->viewName = utf8ConstNew(
						vName.data, vName.len);
					if( !gn->className || !gn->viewName )
					{
						error = 1;
					}
				}
				else
				{
					error = 1;
				}
			}
		}
		else
		{
			error = 1;
		}
		if( error )
		{
			deleteResourceSpecFromJSI(retval);
			retval = 0;
		}
	}
	return( retval );
}

static
void nsDeleteJSISpec(struct _resource *res, struct _resourceSpec *rs)
{
	nameSpaceSpec *nss = (nameSpaceSpec *)rs;
	int lpc;

	for( lpc = 0; lpc < nss->viewSubscriptions.size; lpc++ )
	{
		utf8ConstRelease(nss->viewSubscriptions.data[lpc].className);
		utf8ConstRelease(nss->viewSubscriptions.data[lpc].viewName);
	}
	gc_free(nss->viewSubscriptions.data);
	deleteResourceSpec(rs);
}

static
Hjava_lang_Class *getSubscribedClassView(nameSpaceClient *nsc, char *name)
{
	Hjava_lang_Class *retval = 0;
	viewSubscriberPool *pool;
	int lpc;
	
	pool = nsc->views;
	while( pool && !retval )
	{
		/* Check for an imported class first */
		for( lpc = 0; !retval && (lpc < pool->used); lpc++ )
		{
			retval = findViewClassPart(pool->data[lpc].vGroup,
						   name);
		}
		pool = pool->next;
	}
	return( retval );
}

static
int verifyNameSpace(nameSpaceClient *nsc)
{
	int lpc, lpc2, lpc3, lpc4, retval = 1;
	viewSubscriberPool *vsp;

	vsp = nsc->views;
	while( /* retval && */ vsp )
	{
		for( lpc = 0;
		     // retval &&
		     lpc < vsp->used;
		     lpc++ )
		{
			viewSubscriber *vs = &vsp->data[lpc];

			for( lpc2 = 0;
			     //retval &&
			     lpc2 < vs->vGroup->cViews.size;
			     lpc2++ )
			{
				classView *cv = &vs->vGroup->cViews.data[lpc2];

				if( CLASS_IS_ARRAY(cv->cl) ||
				    CLASS_IS_PRIMITIVE(cv->cl) )
				{
					continue;
				}
				for( lpc3 = 0;
				     //retval &&
				     lpc3 < cv->cl->nmethods;
				     lpc3++ )
				{
					methodView *mv = &cv->methods[lpc3];

					if( (cv->cl->methods[lpc3].accflags &
					     ACC_PRIVATE) ||
					    ((cv->cl->methods[lpc3].accflags &
					      (ACC_PUBLIC|
					       ACC_PRIVATE|
					       ACC_PROTECTED)) == 0) )
					{
						continue;
					}
					for( lpc4 = 0;
					     //retval &&
					     lpc4 < mv->parsed_sig->nargs + 1;
					     lpc4++ )
					{
						if( !getSubscribedClassView(nsc, (char *)&mv->parsed_sig->signature->data[mv->parsed_sig->ret_and_args[lpc4]]) )
						{
							teamLogEntry("namespace",
								     LEA_String, "Error", "Imported class has public reference to non-imported class", -1,
								     LEA_String, "imported-class", cv->name->data, -1,
								     LEA_String, "referenced-class", &mv->parsed_sig->signature->data[mv->parsed_sig->ret_and_args[lpc4]], -1, // XXX partLen()
								     LEA_TAG_DONE);
							retval = 0;
						}
					}
				}
			}
		}
		vsp = vsp->next;
	}
	return( retval );
}

#define DEFAULT_TABLE_SIZE 31

static
int nsResourceSubscribe(struct _resource *res,
			struct _resourceClient *rc,
			struct _resourceSpec *rs)
{
	nameSpaceClient *nsc = (nameSpaceClient *)rc;
	nameSpaceSpec tmp_nss, *nss = (nameSpaceSpec *)rs;
	int lpc, retval = 1;
	groupNames tmp_gn;

	/* Create the root group for this client */
	nsc->cGroup = createClassGroup(0, 0);
	nsc->mapping.table = gc_malloc(sizeof(nameMapping *) *
				       (DEFAULT_TABLE_SIZE + 1),
				       GC_ALLOC_NAME_SPACE);
	assert(nsc->mapping.table != 0); // XXX
	nsc->mapping.size = DEFAULT_TABLE_SIZE;
	nsc->classLocals.table = gc_malloc(sizeof(classLocal *) *
					   (DEFAULT_TABLE_SIZE + 1),
					   GC_ALLOC_NAME_SPACE);
	assert(nsc->classLocals.table != 0); // XXX
	nsc->classLocals.size = DEFAULT_TABLE_SIZE;
	if( nss )
	{
		int foundKernel = 0;
		
		for( lpc = 0;
		     (lpc < nss->viewSubscriptions.size) && !foundKernel;
		     lpc++ )
		{
			if( (nss->viewSubscriptions.data[lpc].teamID ==
			     kernelTeam->elemID) &&
			    !strcmp(nss->viewSubscriptions.data[lpc].
				    className->data,
				    "kernel") )
			{
				foundKernel = 1;
			}
		}
		if( kernelTeam && !foundKernel )
		{
			teamLogEntry("namespace",
				     LEA_String, "Error", "Kernel class group not specified in name space specification", -1,
				     LEA_TAG_DONE);
			retval = 0;
		}
	}
	else
	{
		/* Fill in the default spec arguments. */
		nss = &tmp_nss;
		tmp_gn.teamID = 1;
		tmp_gn.className = utfKernel;
		tmp_gn.viewName = utfIdentity;
		nss->spec.handler = &metalNameSpaceResource.link;
		nss->viewSubscriptions.data = &tmp_gn;
		nss->viewSubscriptions.size = 1;
	}
	/* XXX Preallocate pools since we'll be working in the kernel */
	{
		viewSubscriberPool *pool;
		int count =
			(nss->viewSubscriptions.size +
			 VIEW_SUBSCRIBER_POOL_SIZE -
			 1) /
			VIEW_SUBSCRIBER_POOL_SIZE;

		for( lpc = 0; lpc < count; lpc++ )
		{
			pool = gc_malloc(sizeof(viewSubscriberPool),
					 GC_ALLOC_NAME_SPACE);
			assert(pool != 0); // XXX
			pool->next = nsc->views;
		        nsc->views = pool;
		}
	}
	/* Visit the kernel so we can import the groups from other teams. */
	if( switchToTeam(kernelTeam, 0) )
	{
		for( lpc = 0; lpc < nss->viewSubscriptions.size; lpc++ )
		{
			groupNames *gn = &nss->viewSubscriptions.data[lpc];
			jteam *jt;
			
			if( (jt = grabTeamWithID(nss->viewSubscriptions.
						 data[lpc].teamID)) )
			{
				jteamReservation *jr;
				
				if( (jr = pushTeamReservation(jt)) )
				{
					classGroup *cg;
					viewGroup *vg;
					nameSpaceClient *jt_nsc;

					jt_nsc = (nameSpaceClient *)
						getTeamInternalResource(jt,
									&metalNameSpaceResource.link);
					cg = findClassGroup(&jt_nsc->exportedClasses,
							    gn->className);
					if( cg )
					{
						vg = findViewGroup(cg,
								   gn->viewName);
						if( vg )
						{
							subscribeViewClient(nsc, cg, vg);
							retval = addTeamDependency(nsc->client.owner, jt);
						}
						else
						{
							retval = 0;
							teamLogEntry("namespace",
								     LEA_String, "Error", "Couldn't find view group\n", -1,
								     LEA_String, "viewgroup", gn->viewName->data, -1,
								     LEA_String, "source-team", jt->name, -1,
								     LEA_TAG_DONE);
						}
					}
					else
					{
						retval = 0;
						teamLogEntry("namespace",
							     LEA_String, "Error", "Couldn't find class group\n", -1,
							     LEA_String, "classgroup", gn->className->data, -1,
							     LEA_String, "source-team", jt->name, -1,
							     LEA_TAG_DONE);
					}
					popTeamReservation(jr);
				}
				else
				{
					retval = 0;
				}
				dropTeam(jt);
			}
			else
			{
				retval = 0;
			}
		}
		returnFromTeam();
	}
	else
	{
		retval = 0;
	}
	if( retval )
	{
		retval = verifyNameSpace(nsc);
	}
	return( retval );
}

static
void nsResourceUnsubscribe(struct _resource *res, struct _resourceClient *rc)
{
	nameSpaceClient *nsc = (nameSpaceClient *)rc;
	struct Hjava_lang_Class *cl;
	classGroup *cg;
	classView *cv;
	int lpc;

	require(nsc->flags & NSCF_DEAD);
	require(nsc->imports == 0);

	/*
	 * XXX We need to manually run through and free the interface
	 * implementation indexes for each class.  We should really only
	 * do this for shared interfaces.
	 */
	for( lpc = 0; lpc < nsc->cGroup->vGroups.data[0].cViews.size; lpc++ )
	{
		cv = &nsc->cGroup->vGroups.data[0].cViews.data[lpc];
		cl = cv->cl;
		if( cl && cl->itable2dtable )
		{
			int i;

			for( i = 0; i < cl->total_interface_len; i++ )
			{
				cl->interfaces[i]->
					implementors[cl->impl_index] = -1;
			}
		}
	}
	cg = nsc->exportedClasses.groups;
	while( cg )
	{
		for( lpc = 0; lpc < cg->vGroups.data[0].cViews.size; lpc++ )
		{
			cv = &cg->vGroups.data[0].cViews.data[lpc];
			cl = cv->cl;
			if( cl && cl->itable2dtable )
			{
				int i;
		    
				for( i = 0; i < cl->total_interface_len; i++ )
				{
					cl->interfaces[i]->
						implementors[cl->impl_index] =
						-1;
				}
			}
		}
		cg = cg->next;
	}
}

static resourceFunctions nsResourceFunctions = {
	0,
	0,
	0,
	nsResourceFillSpec,
	nsCreateJSISpec,
	nsDeleteJSISpec,
	0,
	nsResourceSubscribe,
	0,
	nsResourceUnsubscribe
};

/** edu.utah.janosvm.sys.Importable native functions **/
void edu_utah_janosvm_sys_PortAuthority_revokeAllImports(void)
{
	nameSpaceClient *nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	struct Hedu_utah_janosvm_sys_Importable *imp;
	memImportLink *mil;
	int iLockRoot;

	require(currentTeam() == nsc->client.owner);

	/*
	 * Everything is allocated inside the team, so just disconnect any
	 * imported objects.  XXX This should be moved elsewhere.
	 */
	lockMutex(nsc);
	nsc->flags |= NSCF_DEAD;
	mil = nsc->imports;
	unlockMutex(nsc);
	while( mil )
	{
		imp = (struct Hedu_utah_janosvm_sys_Importable *)
			NATIVE2OBJECT(&importNativeClass, mil);
		GC_finalizeObject((Collector *)
				  teamCollector(nsc->client.owner), imp);
		/* XXX ick */
		lockMutex(nsc);
		mil = nsc->imports;
		unlockMutex(nsc);
	}
}

jbool edu_utah_janosvm_sys_PortAuthority_register(struct Hedu_utah_janosvm_sys_Importable *imp)
{
	memImportLink *mil = OBJECT2NATIVE(&importNativeClass,
					   (Hjava_lang_Object *)imp);
	nameSpaceClient *nsc;
	int iLockRoot;
	jbool retval = 0;

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	/* Need to visit the kernel since it manages the links */
	lockMutex(nsc);
	if( !(nsc->flags & NSCF_DEAD) )
	{
		mil->link = nsc->imports;
		nsc->imports = mil;
		retval = 1;
	}
	unlockMutex(nsc);
	return( retval );
}

void edu_utah_janosvm_sys_PortAuthority_unregister(struct Hedu_utah_janosvm_sys_Importable *imp)
{
	memImportLink *mil = OBJECT2NATIVE(&importNativeClass,
					   (Hjava_lang_Object *)imp);
	memImportLink *curr, **prev;
	nameSpaceClient *nsc;
	int iLockRoot;

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(nsc);
	prev = &nsc->imports;
	curr = nsc->imports;
	while( curr && (curr != mil) )
	{
		prev = &curr->link;
		curr = curr->link;
	}
	if( curr )
	{
		*prev = curr->link;
	}
	unlockMutex(nsc);
}

/** Module declarations **/
static int initNameSpaceResourceModule(int stage);

module_t nameSpaceModule = {
	"namespace",
	0,
	initNameSpaceResourceModule,
	0,
	0
};

static int initNameSpaceResourceModule(int stage)
{
	switch( stage )
	{
	case SID_NAMESPACEMODULE_INITMETAL:
		{
			nameSpaceSpec nss;

			utfKernel = utf8ConstNew("kernel", -1);
			utfIdentity = utf8ConstNew("identity", -1);
			
			installResource(
				&metalNameSpaceResource.link,
				RA_Name, "namespace",
				RA_Flags, RF_INTERNAL,
				RA_Functions, &nsResourceFunctions,
				RA_ClientSize, sizeof(nameSpaceClient),
				RA_SpecSize, sizeof(nameSpaceSpec),
				RA_TAG_DONE);
			
			nss.spec.handler = &metalNameSpaceResource.link;
			nss.viewSubscriptions.data = 0;
			nss.viewSubscriptions.size = 0;
			attachTeamResource(metalTeam, &nss.spec);
			break;
		}
	case SID_NAMESPACEMODULE:
		{
			errorInfo einfo;

			installResource(
				&metalNameSpaceResource.link,
				RA_ResourceSectionName, "namespace-resource",
				RA_SpecSectionName, "namespace",
				RA_SpecNativeClass, JAVA_NAME_SPACE_SPEC,
				RA_TAG_DONE);
			installSectionType(
				&nsClassGroupSection.type,
				STA_Name, "classgroup",
				STA_Template, &tagsFileSectionTemplate,
				STA_Functions, &nsClassGroupSectionFunctions,
				STA_Size, (sizeof(classGroupSectionData) -
					   sizeof(tagSectionData)),
				STA_Walk, nsClassGroupSectionWalker,
				STA_TAG_DONE);
			installSectionType(
				&nsViewGroupSection.type,
				STA_Name, "viewgroup",
				STA_Template, &tagsFileSectionTemplate,
				STA_Functions, &nsViewGroupSectionFunctions,
				STA_Walk, nsViewGroupSectionWalker,
				STA_TAG_DONE);
			importNativeClass.size = sizeof(memImportLink);
			/* XXX Load any exempt classes */
			ListClass = lookupClass(
				"edu/utah/janosvm/sys/List",
				NULL,
				&einfo);
			ListNodeClass = lookupClass(
				"edu/utah/janosvm/sys/ListNode",
				NULL,
				&einfo);
			ObjectHolderClass = lookupClass(
				"java/lang/ThreadObjectHolder",
				NULL,
				&einfo);
			CrossTeamCloneableClass = lookupClass(
				"edu/utah/janosvm/sys/CrossTeamCloneable",
				NULL,
				&einfo);
			if( (ImportableClass =
			     lookupClass("edu/utah/janosvm/sys/Importable",
					 NULL,
					 &einfo)) )
			{
				bindNativeClass(&importNativeClass,
						ImportableClass,
						&einfo);
				ExportableClass = lookupClass(
					"edu/utah/janosvm/sys/Exportable",
					NULL,
					&einfo);
			}
		}
		break;
	}
	return( 1 );
}

Hjava_lang_Class *bindClassMapping(nameMapping *nm, errorInfo *einfo)
{
	Hjava_lang_Class *retval = 0;
	viewSubscriberPool *pool;
	nameSpaceClient *nsc;
	char *cname;
	int lpc;
	int iLockRoot;
	
	require(nm != 0);
	require(nm->loader == 0);
	require(einfo != 0);

	cname = (char *)nm->name->data;
	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	pool = nsc->views;
	while( pool && !retval )
	{
		/* Check for an imported class first */
		for( lpc = 0; !retval && (lpc < pool->used); lpc++ )
		{
			struct Hjava_lang_Class *cl;
			
			if( (cl = findViewClass(pool->data[lpc].vGroup,
						nm->name)) )
			{
				if( !CLASS_IS_PRIMITIVE(cl) )
				{
					retval = cl;
				}
			}
		}
		pool = pool->next;
	}
	if( !retval )
	{
		char *buf;

		/* No matching import, try the class path */
		if( (buf = gc_malloc(strlen(cname) +
				     1 + /* `.' */
				     5 + /* `class' */
				     1,
				     GC_ALLOC_NAME_SPACE)) )
		{
			classPathClient *cpc;
			classFile cf;

			{
				sprintf(buf, "%s.class", cname);

				ensure_valid_memory(buf);
			}
			cpc = (classPathClient *)
				getTeamInternalResource(0,
					&metalClassPathResource.link);
			switch( loadClassPathFile(cpc, buf, &cf, einfo) )
			{
			case FTH_LOAD_ERROR:
			case FTH_LOAD_NOT_FOUND:
				break;
			case FTH_LOAD_FOUND:
				/* Found the class try to instantiate it */
				if( (retval = newClass()) )
				{
					utf8ConstAssign(retval->primaryView.
							name,
							nm->name);
					retval = readClass(retval,
							   &cf,
							   NULL,
							   einfo);
					if( retval )
					{
						lockMutex(nsc);
						if(!addGroupClass(nsc->cGroup,
								  retval) ||
						   !gc_add_ref(retval) )
						{
							postOutOfMemory(einfo);
							retval = 0;
						}
						unlockMutex(nsc);
					}
					else
					{
						teamLogEntry("namespace",
							     LEA_String, "Error", "Cannot read in class", -1,
							     LEA_String, "class", nm->name->data, -1,
							     LEA_TAG_DONE);
					}
				}
				else
				{
					postOutOfMemory(einfo);
				}
				gc_free(cf.base);
				break;
			}
			gc_free(buf);
		}
		else
		{
			postOutOfMemory(einfo);
		}
	}
	/*
	 * XXX This is kinda stupid, we should be able to fail more gracefully
	 * than checking each time we try to bind a class
	 */
	if( !retval &&
	    ((strcmp(cname, "java/lang/ClassNotFoundException") == 0) ||
	     (strcmp(cname, "java/lang/Object") == 0)) )
	{
		fprintf(stderr,
			"Cannot find essential class '%s' in class library ..."
			" aborting.\n",
			cname);
		ABORT();
	}
	/* Set the value in the mapping */
	if( retval )
	{
		nm->value = retval;
	}
	/* Post the expected exception if it class wasn't found */
	if( !retval )
	{
		discardErrorInfo(einfo);
		postNoClassDefFoundError(einfo, nm->name->data);
	}
	return( retval );
}

/* Lookup a nameMapping in the given client */
static
nameMapping *internalFindClassMapping(nameSpaceClient *nsc,
				      Utf8Const *name,
				      int table_index,
				      Hjava_lang_ClassLoader *loader)
{
	nameMapping *curr, *retval = 0;

	curr = nsc->mapping.table[table_index];
	while( curr && !retval )
	{
		if( utf8ConstEqual(curr->name, name) &&
		    (curr->loader == loader) )
		{
			retval = curr;
		}
		curr = curr->next;
	}
	return( retval );
}

KAFFE_INLINE
int hashMappingName(nameSpaceClient *nsc, Utf8Const *name)
{
	int retval;

	require(nsc != 0);
	require_valid_memory(nsc);
	require_valid_memory(nsc->mapping.table);
	require(name != 0);
	
	retval = utf8ConstHashValue(name) & nsc->mapping.size;

	ensure(retval >= 0);
	ensure(retval < (nsc->mapping.size + 1));

	return( retval );
}

nameMapping *findClassMapping(Utf8Const *name, Hjava_lang_ClassLoader *loader)
{
	nameMapping *retval = 0;

	nameSpaceClient *nsc;
	int idx;
	int iLockRoot;

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(nsc);
	idx = hashMappingName(nsc, name);
	retval = internalFindClassMapping(nsc, name, idx, loader);
	unlockMutex(nsc);
	return( retval );
}

nameMapping *getClassMapping(Utf8Const *name,
			     Hjava_lang_ClassLoader *loader,
			     errorInfo *einfo)
{
	nameMapping *retval = 0;
	nameSpaceClient *nsc;
	int idx;
	int iLockRoot;

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(nsc);
	idx = hashMappingName(nsc, name);
	if( (retval = internalFindClassMapping(nsc, name, idx, loader)) )
	{
		/* Found a match, return it */
	}
	else if( (retval = gc_malloc(sizeof(nameMapping),
				     GC_ALLOC_NAME_SPACE)) )
	{
		/* Need to make a new one */
		retval->next = nsc->mapping.table[idx];
		nsc->mapping.table[idx] = retval;
		utf8ConstAddRef(name);
		retval->name = name;
		retval->loader = loader;

		ensure_valid_memory(nsc);
		ensure_valid_memory(nsc->mapping.table);
	}
	else
	{
		postOutOfMemory(einfo);
	}
	
	ensure(!retval ||
	       internalFindClassMapping(nsc, name, idx, loader));
	
	unlockMutex(nsc);
	return( retval );
}

void clearClassMapping(Utf8Const *name, Hjava_lang_ClassLoader *loader)
{
	nameSpaceClient *nsc;
	nameMapping *nm = 0;
	int idx;
	int iLockRoot;
	
	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(nsc);
	idx = hashMappingName(nsc, name);
	if( (nm = internalFindClassMapping(nsc, name, idx, loader)) )
	{
		nm->value = 0;
	}
	unlockMutex(nsc);
}

void removeLoaderMappings(Hjava_lang_ClassLoader *loader)
{
	nameMapping *curr, *next, **prev;
	nameSpaceClient *nsc;
	int lpc;
	int iLockRoot;
	
	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	lockMutex(nsc);
	for( lpc = 0; lpc < nsc->mapping.size; lpc++ )
	{
		prev = &nsc->mapping.table[lpc];
		curr = nsc->mapping.table[lpc];
		while( curr )
		{
			next = curr->next;
			if( curr->loader == loader )
			{
				/*
				 * Since we're killing the current node, theres
				 * no need to advance the prev pointer.
				 */
				*prev = next;
				gc_free(curr);
			}
			else
			{
				prev = &curr->next;
			}
			curr = next;
		}
	}
	unlockMutex(nsc);
}

#if defined(KAFFE_STATS) || defined(KAFFE_PROFILER) || defined(DEBUG)
void walkNameSpace(nameSpaceClient *nsc,
		   int (*walker)(Hjava_lang_Class *clazz, void *),
		   void *param)
{
	nameMapping *nm;
	int lpc;
	
	if( !nsc )
	{
		nsc = (nameSpaceClient *)
			getTeamInternalResource(0,
				&metalNameSpaceResource.link);
	}
	for( lpc = 0; lpc < nsc->mapping.size; lpc++ )
	{
		nm = nsc->mapping.table[lpc];
		while( nm )
		{
			if( nm->value )
			{
				walker(nm->value, param);
			}
			nm = nm->next;
		}
	}
}
#endif

classView *getClassView(Hjava_lang_Class *cl)
{
	classView *retval = 0;
	nameSpaceClient *nsc;
	int lpc;

	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	if( (cl->cGroup == 0) || !nsc )
	{
		/* XXX This really shouldn't happen... */
		return( &cl->primaryView );
	}
	if( cl->cGroup == nsc->cGroup )
	{
		retval = &cl->primaryView;
	}
	else
	{
		viewSubscriberPool *pool;

		pool = nsc->views;
		while( pool && !retval )
		{
			for( lpc = 0;
			     !retval && (lpc < VIEW_SUBSCRIBER_POOL_SIZE);
			     lpc++ )
			{
				if( pool->data[lpc].cGroup == cl->cGroup )
				{
					retval = &pool->data[lpc].vGroup->
						cViews.data[cl->viewSlot];
				}
			}
			pool = pool->next;
		}
	}
	if( !retval )
	{
		/*
		 * XXX This means they don't have a view on it.  So,
		 * technically they shouldn't be able to look at it, but not
		 * everything can handle not getting the view so we return
		 * it for now.
		 */
		teamLogEntry("namespace",
			     LEA_String, "file", __FILE__, -1,
			     LEA_String, "function", __PRETTY_FUNCTION__, -1,
			     LEA_String, "Warning",
			     "Team has no view on this class", -1,
			     LEA_String, "class",
			     cl->primaryView.name->data, -1,
			     LEA_TAG_DONE);
		retval = &cl->primaryView;
	}
	return( retval );
}

methodView *getMethodView(Method *meth)
{
	Hjava_lang_Class *cl = meth->class;
	methodView *retval = 0;
	classView *cv;
	int idx;

	idx = meth - cl->methods;
	cv = getClassView(cl);
	retval = &cv->methods[idx];
	return( retval );
}

KAFFE_INLINE
int hashClassLocal(nameSpaceClient *nsc, Utf8Const *name)
{
	int retval;

	require(nsc != 0);
	require(name != 0);
	
	retval = utf8ConstHashValue(name) & nsc->classLocals.size;

	ensure(retval >= 0);
	ensure(retval < (nsc->classLocals.size + 1));

	return( retval );
}

int setClassLocalData(jteam *jt,
		      Hjava_lang_Class *cl,
		      struct Hjava_lang_Object *obj)
{
	int hash, retval = 0;
	nameSpaceClient *nsc;
	classLocal *local;

	nsc = (nameSpaceClient *)
		getTeamInternalResource(jt, &metalNameSpaceResource.link);
	hash = hashClassLocal(nsc, cl->primaryView.name);
	local = nsc->classLocals.table[hash];
	while( local && (local->cl != cl) )
		local = local->next;
	if( !local &&
	    (local = gc_malloc(sizeof(classLocal), GC_ALLOC_NAME_SPACE)) )
	{
		local->next = nsc->classLocals.table[hash];
		nsc->classLocals.table[hash] = local;
		local->cl = cl;
		local->data = 0;

		ensure_valid_memory(nsc);
		ensure_valid_memory(nsc->classLocals.table);
	}
	if( local )
	{
		if( local->data )
			GC_rm_ref((Collector *)teamCollector(jt), local->data);
		local->data = obj;
		if( obj )
			GC_add_ref((Collector *)teamCollector(jt), obj);
		retval = 1;
	}
	return( retval );
}

struct Hjava_lang_Object *getClassLocalData(jteam *jt, Hjava_lang_Class *cl)
{
	struct Hjava_lang_Object *retval = 0;
	nameSpaceClient *nsc;
	classLocal *local;
	int hash;
	
	nsc = (nameSpaceClient *)
		getTeamInternalResource(jt, &metalNameSpaceResource.link);
	hash = hashClassLocal(nsc, cl->primaryView.name);
	local = nsc->classLocals.table[hash];
	while( local && (local->cl != cl) )
		local = local->next;
	if( local )
	{
		retval = local->data;
	}
	return( retval );
}

void checkClass(Hjava_lang_Class *cl, Hjava_lang_ClassLoader *loader)
{
	/* XXX Is this just KaffeOS legacy */
}

void finalizeClassLoader(Collector *collector, void *obj)
{
	Hjava_lang_ClassLoader *loader = (Hjava_lang_ClassLoader *)obj;

	removeLoaderMappings(loader);
}

static
struct Hjava_lang_Class *matchPrimitive(Utf8Const *utfname,
					errorInfo *einfo)
{
	struct Hjava_lang_Class *retval = 0;
	const char *name = utfname->data;

	if( !strcmp(name, "boolean") )
		retval = booleanClass;
	else if( !strcmp(name, "byte") )
		retval = byteClass;
	else if( !strcmp(name, "char") )
		retval = charClass;
	else if( !strcmp(name, "double") )
		retval = doubleClass;
	else if( !strcmp(name, "float") )
		retval = floatClass;
	else if( !strcmp(name, "int") )
		retval = intClass;
	else if( !strcmp(name, "long") )
		retval = longClass;
	else if( !strcmp(name, "short") )
		retval = shortClass;
	else if( !strcmp(name, "void") )
		retval = voidClass;
	if( !retval )
	{
		postExceptionMessage(einfo,
				     "java.lang.ClassNotFoundException",
				     "Can't find primitive class `%s'",
				     utfname->data);
	}
	return( retval );
}

static
int exportClasses(classGroup *cg, errorInfo *einfo)
{
	int retval = 1;
	tagList *tl;
	
	/* Walk the exported classes specified in the config file */
	if( (tl = cg->type->defs) && (tl->tags.count != 0) )
	{
		Hjava_lang_Class *cl;
		int lpc;
		
		for( lpc = 0; retval && (lpc < tl->tags.count); lpc++ )
		{
			Utf8Const *className;
			
			className = tl->tags.data[lpc].tag;
			/* Load each class and add it to the group */
			if( (cl = matchPrimitive(className, einfo)) ||
			    (cl = loadClass(className, 0, einfo)) )
			{
				if( !cl->cGroup->cgl )
				{
					remGroupClass(cl);
					addGroupClass(cg, cl);
				}
				else
				{
					retval = 0;
					postExceptionMessage(
						einfo,
						"java.lang."
						"ClassNotFoundException",
						"Class `%s' already exported",
						className->data);
				}
			}
			else
			{
				postExceptionMessage(
					einfo,
					"java.lang.ClassNotFoundException",
					"Couldn't find exported class `%s'",
					className->data);
				retval = 0;
			}
		}
	}
	else
	{
		postExceptionMessage(einfo,
				     "edu.utah.janosvm.resources."
				     "InvalidClassGroupException",
				     "Class group `%s' is empty.",
				     cg->type->name->data);
		retval = 0;
	}
	return( retval );
}

int exportClassGroup(Utf8Const *cgName, errorInfo *einfo)
{
	classGroupType *type;
	nameSpaceClient *nsc;
	int retval = 0;
	classGroup *cg;
	int iLockRoot;
	
	nsc = (nameSpaceClient *)
		getTeamInternalResource(0, &metalNameSpaceResource.link);
	/* Try to find the class group type to import */
	if( (type = findClassGroupType(metalNameSpaceResource.types, cgName)) )
	{
		lockMutex(nsc);
		if( (cg = createClassGroup(type, nsc->cGroup)) )
		{
			/* Subscribe ourselves to the new group */
			subscribeViewClient(nsc, cg, &cg->vGroups.data[0]);
			/* Export the individual classes */
			if( exportClasses(cg, einfo) )
			{
				/* Add the group to our exports */
				addClassGroup(&nsc->exportedClasses, cg);
				retval = 1;
			}
			else
			{
				/* XXX unsubscribeViewClient */
				teamLogEntry("namespace",
					     LEA_String, "Error", "Unable to export class group", -1,
					     LEA_String, "classGroup",
					     cgName->data, -1,
					     LEA_TAG_DONE);
			}
		}
		else
		{
			postOutOfMemory(einfo);
		}
		unlockMutex(nsc);
	}
	else
	{
		postExceptionMessage(
			einfo,
			"edu.utah.janosvm.resources."
			"ClassGroupNotFoundException",
			"Couldn't find class group `%s'",
			cgName->data);
		teamLogEntry("namespace",
			     LEA_String, "Error", "Couldn't find class group for export.", -1,
			     LEA_String, "classGroup", cgName->data, -1,
			     LEA_TAG_DONE);
	}
	return( retval );
}

Method *findMethodFromPC(uintp pc)
{
	memResourceClient *mrc = gc_get_resource((void *)pc);
	Method *retval = 0;

	/*
	 * Get a reference to the team that owns this code, we shouldn't have
	 * to worry about it going away since we're using the code...  So its
	 * either exported or we're visiting the team.
	 */
	if( mrc )
	{
		nameSpaceClient *nsc;

		/* Get the namespace of the owning team */
		nsc = (nameSpaceClient *)getTeamInternalResource(
			mrc->client.owner,
			&metalNameSpaceResource.link);
		/* Look through the main group */
		/*
		 * XXX The findGroupMethod will do a lock, which might be in
		 * another team, so we have to keep the thread alive.  Should
		 * we do a reserveTeam here?
		 */
		jthread_disable_stop();
		retval = findGroupMethod(nsc->cGroup, (void *)pc);
		jthread_enable_stop();
	}
	return( retval );
}

struct Hjava_lang_Class *getNativeCallerClass(void)
{
	struct Hjava_lang_Class *retval = 0;
	void *fp, *pc;
	Method *meth;

	getCallerFrameAndPC(&fp, &pc);
	pc = getFrameReturnPC(getPreviousFramePointer(fp));
	if( (meth = findMethodFromPC((uintp)pc)) )
	{
		retval = meth->class;
	}
	return( retval );
}

void edu_utah_janosvm_resources_NameSpaceSpec_nativeConstructor(
	struct Hedu_utah_janosvm_resources_NameSpaceSpec *jnss,
	HArrayOfObject *nsg)
{
	nameSpaceSpec *nss = OBJECT2NATIVE(&metalNameSpaceResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jnss);
	struct Hedu_utah_janosvm_resources_NameSpaceGroup *curr;
	Utf8Const *cg, *vg;
	int error = 0, lpc;
	errorInfo einfo;
	sparseID_t id;

	nss->spec.handler = &metalNameSpaceResource.link;
	if( !nsg )
	{
		postException(&einfo, JAVA_LANG(NullPointerException));
		error = 1;
	}
	for( lpc = 0; !error && (lpc < ARRAY_SIZE(nsg)); lpc++ )
	{
		curr = ((struct Hedu_utah_janosvm_resources_NameSpaceGroup **)
			ARRAY_DATA(nsg))[lpc];
		if( curr == NULL )
			continue;
		id = do_execute_java_method(unhand(curr)->team,
					    "getID",
					    "()I",
					    0, 0).i;

		require(curr->classGroup);
		require(curr->viewGroup);
		
		cg = stringJava2Utf8ConstReplace(curr->classGroup, 0, 0);
		vg = stringJava2Utf8ConstReplace(curr->viewGroup, 0, 0);
		if( cg && vg )
		{
			error = !subscribeView(nss, id, cg, vg);
		}
		else
		{
			error = 1;
		}
		if( error )
		{
			postOutOfMemory(&einfo);
			if( cg )
				utf8ConstRelease(cg);
			if( vg )
				utf8ConstRelease(vg);
		}
	}
	if( error )
	{
		throwError(&einfo);
	}
}

void edu_utah_janosvm_resources_NameSpaceSpec_nativeFinalizer(
	struct Hedu_utah_janosvm_resources_NameSpaceSpec *jnss)
{
	struct Hedu_utah_janosvm_resources_ResourceSpec *jrs =
		(struct Hedu_utah_janosvm_resources_ResourceSpec *)jnss;
	nameSpaceSpec *nss = OBJECT2NATIVE(&metalNameSpaceResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jnss);

	if( !jrs->found )
	{
		int lpc;
		
		for( lpc = 0; lpc < nss->viewSubscriptions.size; lpc++ )
		{
			utf8ConstRelease(nss->viewSubscriptions.data[lpc].
					 className);
			utf8ConstRelease(nss->viewSubscriptions.data[lpc].
					 viewName);
		}
		gc_free(nss->viewSubscriptions.data);
	}
}

struct Hedu_utah_janosvm_resources_NameSpaceSpec *
edu_utah_janosvm_resources_NameSpaceSpec_current(void)
{
	struct Hedu_utah_janosvm_resources_NameSpaceSpec *retval;

	if( (retval = (struct Hedu_utah_janosvm_resources_NameSpaceSpec *)
	     newObject(metalNameSpaceResource.link.specNativeClass.peer)) )
	{
		nameSpaceClient *nsc = (nameSpaceClient *)
			getTeamInternalResource(0,
						&metalNameSpaceResource.link);
		viewSubscriberPool *vsp;
		nameSpaceSpec *nss;
		int total = 0;
		
		retval->found = 1;
		nss = OBJECT2NATIVE(&metalNameSpaceResource.link.
				    specNativeClass,
				    (Hjava_lang_Object *)retval);
		nss->spec.handler = &metalNameSpaceResource.link;
		vsp = nsc->views;
		while( vsp )
		{
			total += vsp->used;
			vsp = vsp->next;
		}
		if( (nss->viewSubscriptions.data =
		     gc_malloc(sizeof(groupNames) * total,
			       GC_ALLOC_NAME_SPACE)) )
		{
			groupNames *gn = nss->viewSubscriptions.data;
			int lpc;
			
			nss->viewSubscriptions.size = total;
			vsp = nsc->views;
			while( vsp )
			{
				for( lpc = 0; lpc < vsp->used; lpc++ )
				{
					gn->teamID = gc_get_resource(vsp->data[lpc].cGroup)->client.owner->elemID;
					gn->className = vsp->data[lpc].cGroup->type->name;
					gn->viewName = vsp->data[lpc].vGroup->type->name;
					gn++;
				}
				vsp = vsp->next;
			}
		}
		else
		{
			retval = 0;
		}
	}
	return( retval );
}

void edu_utah_janosvm_resources_NameSpaceResource_exportClassGroup(
	struct Hjava_lang_String *jType)
{
	errorInfo einfo;
	int ok = 0;
	
	if( jType != 0 )
	{
		Utf8Const *type;
		
		if( (type = stringJava2Utf8ConstReplace(jType, 0, 0)) )
		{
			ok = exportClassGroup(type, &einfo);
			utf8ConstRelease(type);
		}
		else
		{
			postOutOfMemory(&einfo);
		}
	}
	else
	{
		postException(&einfo, JAVA_LANG(NullPointerException));
	}
	if( !ok )
	{
		throwError(&einfo);
	}
}
