/*
 * nameSpace.h
 * Manage the root namespace for a team
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _namespace_h
#define _namespace_h

/*
 * The nameSpace module manages the root name space for a jteam.
 */

#include "gtypes.h"
#include "tags.h"
#include "locks.h"
#include "classMethod.h"
#include "resource.h"

struct _jteam;
struct _classView;
struct _classGroup;

/*
 * The nameMapping structure is used to store the mapping of names to classes
 * in a hash table.
 *
 * `next' - Link to the next mapping in the hash table.
 * `name' - The name of the class.
 * `loader' - The loader of this class.
 * `value' - The class itself.
 */
typedef struct _nameMapping {
	struct _nameMapping *next;
	Utf8Const *name;
	struct _iLock *lock;
	struct Hjava_lang_ClassLoader *loader;
	struct Hjava_lang_Class *value;
} nameMapping;

/*
 * The nameSpaceResource structure contains extra data for managing the
 * name space resource.
 *
 * `link' - The resource header.
 * `groups' - The list of exported classGroups.
 */
typedef struct _nameSpaceResource {
	resource link;
	classGroupType *types;
} nameSpaceResource;

/*
 * A viewSubscriber structure is used to track which classGroups and associated
 * viewGroups a nameSpace subscribes to.
 *
 * `cGroup' - An exported classGroup that this nameSpace subscribed to.
 * `vGroup' - The specific view of the above classGroup thats used in this
 *   name space.
 */
typedef struct _viewSubscriber {
	classGroup *cGroup;
	viewGroup *vGroup;
} viewSubscriber;

/*
 * A viewSubscriberPool is used to store the view group subscriptions for a
 * team.
 *
 * `next' - Link to the next pool in the list.
 * `data' - The preallocated pool of viewSubscriber's.
 * `used' - The number of used viewSubscribers in `data'.
 */
typedef struct _viewSubscriberPool {
	struct _viewSubscriberPool *next;
#define VIEW_SUBSCRIBER_POOL_SIZE 4
	viewSubscriber data[VIEW_SUBSCRIBER_POOL_SIZE];
	int used;
} viewSubscriberPool;

/*
 * A memImportLink is used to internally track edu.utah.janosvm.sys.Importable
 * objects for each team.  These are tracked so we can unsubscribe them from
 * their exportable's when the team dies.  XXX Should really be moved elsewhere
 *
 * `link' - Link to the next Importable in the list.
 */
typedef struct _memImportLink {
	struct _memImportLink *link;
} memImportLink;

/*
 * A classLocal is used to store an object associated with a class on a per-
 * team basis.
 *
 * `next' - Link to the next classLocal in the hash table collision list.
 * `cl' - The class the object is paired with.
 * `data' - The object associated with the class.
 */
typedef struct _classLocal {
	struct _classLocal *next;
	struct Hjava_lang_Class *cl;
	struct Hjava_lang_Object *data;
} classLocal;

/*
 * The nameSpaceClient structure is a resourceClient that has extra fields
 * for tracking mapping of names to classes and any external classGroups.
 *
 * `client' - The resourceClient header.
 * `flags' - Bitfield to hold flags.
 * `imports' - Linked list of Importable objects.
 * `cGroup' - The root group for this nameSpace, any classes that are loaded
 *   just for this group are placed here.
 * `exportedClasses' - The classes that the team exports from its namespace.
 * `lock' - Lock for this nameSpace.
 * `mapping.table' - The array used to root the hash table for nameMappings.
 * `mapping.size' - The allocated size of `mapping.table'.
 * `classLocals.table' - Hash table for holding per-team class locals.
 * `classLocals.size' - The size of the `classLocals.table' array.
 * `views' - The list of class view groups this teams subscribes to.
 */

enum {
	NSCB_DEAD
};

enum {
	NSCF_DEAD = (1L << NSCB_DEAD)
};

typedef struct _nameSpaceClient {
	resourceClient client;
	unsigned long flags;
	memImportLink *imports;
	classGroup *cGroup;
	classGroupList exportedClasses;
	struct _iLock *lock;
	struct {
		nameMapping **table;
		int size;
	} mapping;
	struct {
		classLocal **table;
		int size;
	} classLocals;
	viewSubscriberPool *views;
} nameSpaceClient;

/*
 * The groupNames structure is used by the nameSpaceSpec to specify which
 * classGroups and associated view are to be imported.
 *
 * `className' - The name of the classGroup to subscribe to.
 * `viewName' - The name of the viewGroup within the above classGroup to
 *   subscribe to.
 */
typedef struct _groupNames {
	sparseID_t teamID;
	Utf8Const *className;
	Utf8Const *viewName;
} groupNames;

/*
 * The nameSpaceSpec is a resourceSpec used to describe which classGroups are
 * to be imported into the nameSpace.
 *
 * `spec' - The resourceSpec header.
 * `viewSubscriptions.data' - The array of groupNames that we want to
 *   subscribe to.
 * `viewSubscriptions.size' - The allocated size of the viewSubscriptions.data
 *  array.
 */
typedef struct _nameSpaceSpec {
	resourceSpec spec;
	struct {
		groupNames *data;
		int size;
	} viewSubscriptions;
} nameSpaceSpec;

/*
 * Attempt to bind a name to a class object.  If the name matches any classes
 * from imported groups than that class is used, otherwise the function will
 * try to load the class from the class path.
 *
 * `nm' - The nameMapping object to resolve.
 * `einfo' - Stores any error information.
 * returns - The java class that the named resolved to or NULL if name couldn't
 *  be bound to a class.  Note: The class pointer is also put in nm->value.
 */
Hjava_lang_Class *bindClassMapping(nameMapping *nm, errorInfo *einfo);
/*
 * Find a mapping matching the given values.
 *
 * `name' - The class to search for.
 * `loader' - The class loader context in which to search for the class.
 * returns - The nameMapping object matching the given values or NULL if it
 *  wasn't found.
 */
nameMapping *findClassMapping(Utf8Const *name, Hjava_lang_ClassLoader *loader);
/*
 * Get a mapping with the given values, if one has already been created then
 * we return it, otherwise we create a new one.
 *
 * `name' - The class to get a mapping for.
 * `loader' - The class loader context in which to search for the class.
 * `einfo' - Stores any error information.
 * returns - A previously create nameMapping object if one was already created,
 *  a new nameMapping object, or NULL if there wasn't enough memory.
 */
nameMapping *getClassMapping(Utf8Const *name,
			     Hjava_lang_ClassLoader *loader,
			     errorInfo *einfo);
/*
 * Remove the mapping from the table.
 *
 * `name' - The class name.
 * `loader' - The class loader context in which to search for the class.
 */
void clearClassMapping(Utf8Const *name, Hjava_lang_ClassLoader *loader);
/*
 * Remove any mappings matching the given loader.
 *
 * `loader' - The class loader that is being collected.
 */
void removeLoaderMappings(Hjava_lang_ClassLoader *loader);
#if defined(KAFFE_STATS) || defined(KAFFE_PROFILER) || defined(DEBUG)
/*
 * Walk the name space.
 *
 * `walker' - The function to be called with each class in the name space.
 * `param' - The parameter to pass to the walker function.
 */
void walkNameSpace(nameSpaceClient *nsc,
		   int (*walker)(Hjava_lang_Class *clazz, void *),
		   void *param);
#endif
/*
 * Get the class view for the given class within the context of the current
 * team.
 *
 * `cl' - The class to get the view for.
 * returns - The classView from the viewGroup that the team subscribed to.
 */
classView *getClassView(Hjava_lang_Class *cl);
/*
 * Similar to getClassView, except it works on Methods.
 *
 * `meth' - The method to get the view for.
 * returns - The methodView from the viewGroup that the team subscribed to.
 */
methodView *getMethodView(Method *meth);
/*
 * Set the team local data for the given class.
 *
 * `cl' - The class to associate an object with.
 * `obj' - The object to pair with the class in this team.
 */
int setClassLocalData(jteam *jt,
		      Hjava_lang_Class *cl,
		      struct Hjava_lang_Object *obj);
/*
 * Get the team local data for the given class.
 *
 * `cl' - The class to get the team local data for.
 * returns - The object associated with the given class in this team or NULL
 *  if no association was made.
 */
struct Hjava_lang_Object *getClassLocalData(jteam *jt, Hjava_lang_Class *cl);
/*
 * Export a class group from your class.  The `cgName' corresponds to a
 * predefined class group type.
 *
 * `cgName' - The name of the class group in the config file to export.
 * `einfo' - Stores any error information.
 * returns - True if the export succeeded.
 */
int exportClassGroup(Utf8Const *cgName, errorInfo *einfo);

struct Hjava_lang_Class *getNativeCallerClass(void);

extern nameSpaceResource metalNameSpaceResource;

/*
 * AddModule: nameSpaceModule
 *   ModuleDependency: gcModule.heap memResourceModule.initMetal
 *   ModuleDependency: resourceModule.internalResources classPathModule.initMetal
 *   Comment: Install the resource and its section types
 * Transition: initMetal
 *   ModuleDependency: baseClassesModule sectionFileModule.types
 *   ModuleDependency: resourceModule.reinstall
 *   Comment: Lookup the importable class and bind the native data
 */

#endif
