/*
 * nativeResource.c
 * Provides access to the native libraries.
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdarg.h>

#include "assert_pp.h"

#include "gtypes.h"
#include "system.h"
#include "jsyscall.h"
#include "files.h"
#include "gc.h"
#include "jteam.h"
#include "jteamReservation.h"
#include "module.h"
#include "moduleList.h"
#include "stringSupport.h"
#include "file.h"
#include "itypes.h"
#include "readClass.h"
#include "mungWallProxy.h"
#include "nativeClass.h"
#include "nativeClass-inline.h"
#include "scripting.h"
#define LT_NON_POSIX_NAMESPACE
#include "ltdl.h"
#include "jsignal.h"
#include "slib.h"
#include "access.h"
#include "nameSpace.h"
#include "logResource.h"

#include "nativeResource.h"

#include <jsi/jsi.h>

#include <misc/feedback.h>
#include <misc/stringParsing.h>

#include "edu_utah_janosvm_resources_NativeSpec.h"

#define JAVA_NATIVE_SPEC "edu/utah/janosvm/resources/NativeSpec"

#define	MAXSTUBLEN	1024
#define	MAXLIBPATH	1024
#define	MAXLIBS		16

/* Default library info. */
#define	NATIVELIBRARY	"native"
#define	VMLIBRARY	"kaffevm"
#define	MISCLIBRARY	"kaffemisc"

static int initNativeResourceModule(int stage);

module_t nativeResourceModule = {
	"native",
	0,
	initNativeResourceModule,
	0,
	0,
};

static struct _resourceSpec *nativeCreateJSISpec(struct _resource *res,
						 jsiMessage *jm);
static void nativeDeleteJSISpec(struct _resource *res,
				struct _resourceSpec *rs);
static int nativeSubscribe(struct _resource *res,
			   struct _resourceClient *rc,
			   struct _resourceSpec *rs);
static void nativeUnsubscribe(struct _resource *res,
			      struct _resourceClient *rc);

static resourceFunctions nativeResourceFunctions = {
	0,
	0,
	0,
	0,
	nativeCreateJSISpec,
	nativeDeleteJSISpec,
	0,
	nativeSubscribe,
	0,
	nativeUnsubscribe
};

nativeResource metalNativeResource;

static jsiHandlerFunctions nRootFunctions;
static jsiNamedHandler nRootHandler;

static char *nHandleRootMessage(jsiHandlerContext *hc,
				jsiHandler *jh, jsiMessage *jm);
static int nGetSuites(jsiHandler *jh, jsiMessage *jm);

static jsiPropertyInfo nRootProperties[] = {
	{ "nativepath", (1L << JSI_GET_PROPERTY), "",
	  "Get the list of native library path entries.\n" },
	{ 0 }
};

#ifndef STUB_PREFIX
#define STUB_PREFIX ""
#define STUB_PREFIX_LEN 0
#endif

#ifndef STUB_POSTFIX
#define STUB_POSTFIX ""
#endif

#ifndef LIBRARYHANDLE
#define LIBRARYHANDLE lt_dlhandle
#endif

#ifndef LIBRARYINIT
static inline lt_ptr_t kdlmalloc(size_t len)
{
	void *retval;
	
	require((currentTeam() == metalTeam) ||
		(currentTeam() == kernelTeam));

	retval = gc_malloc(len, GC_ALLOC_EXTERNAL);

	return( retval );
}

static inline void kdlfree(lt_ptr_t ptr)
{
	require((currentTeam() == metalTeam) ||
		(currentTeam() == kernelTeam));

	gc_free(ptr); 
}
#define LIBRARYINIT() ((lt_dlmalloc=kdlmalloc),(lt_dlfree=kdlfree),lt_dlinit())
#endif

#ifndef LIBRARYPREFIX
#define LIBRARYPREFIX "lib"
#endif

#ifndef LIBRARYSUFFIX
#define LIBRARYSUFFIX ""
#endif

#ifndef LIBRARYLOAD
#define LIBRARYLOAD(desc,filename)	((desc)=lt_dlopen((filename)))
#endif

#ifndef LIBRARYUNLOAD
#define LIBRARYUNLOAD(desc)		(lt_dlclose(desc))
#endif

#ifndef LIBRARYERROR
#define LIBRARYERROR() lt_dlerror()
#endif

static struct _libHandle {
	LIBRARYHANDLE	desc;
	char*		name;
	int		ref;
} libHandle[MAXLIBS];

#ifdef LIBRARYFUNCTION
#define INITSYMBOL(ptr,name) LIBRARYFUNCTION(ptr,name)
#else
static inline lt_ptr_t findInitSymbol(const char *name) {
	lt_ptr_t retval = 0;
	int lpc;
	
	for( lpc = 0; !retval && (lpc < MAXLIBS); lpc++ )
	{
		if( libHandle[lpc].ref )
			retval = lt_dlsym(libHandle[lpc].desc, name);
	}
	return retval;
}

#define INITSYMBOL(ptr,name) ((ptr)=findInitSymbol(name))
#endif

#ifndef LIBRARYFUNCTION
/*
 * XXX Ugh, what are supposed to do if this isn't used?  We won't be able
 * to restrict access to just the libraries that the team has loaded.
 */
static inline lt_ptr_t findLibraryFunction(const char *name)
{
	nativeClient *nc = (nativeClient *)
		getTeamInternalResource(0, &metalNativeResource.link);
	lt_ptr_t retval = 0;
	nativeLibrary *nl;

	require(name != 0);
	
	nl = nc->libraries;
	while( nl && !retval )
	{
		if( nl->refs )
		{
			retval = lt_dlsym(libHandle[nl->libID].desc, name);
		}
		nl = nl->next;
	}
	return( retval );
}

#define LIBRARYFUNCTION(ptr,name) ((ptr)=findLibraryFunction(name))
#endif

extern jint Kaffe_JNI_native(Method*);

struct libraryName {
	parsedString path;
	parsedString base;
};

/* Get the base name of the library */
static
int parseLibraryName(struct libraryName *dest, char *libname)
{
	int retval;

	require(dest != 0);
	require(libname != 0);

	if( (retval = parseString(libname,
				  SPO_String, &dest->path,
				  SPO_Do,
				    SPO_String, &dest->base,
				    SPO_While, file_separator, "",
				    SPO_End,
				  SPO_End)) )
	{
		int prefixLen = strlen(LIBRARYPREFIX);
		
		if( !prefixLen ||
		    (strncmp(dest->base.data, LIBRARYPREFIX, prefixLen) == 0) )
		{
			dest->base.data += prefixLen;
			dest->base.len -= prefixLen;
		}
		else
		{
			/* No prefix... */
			retval = 0;
		}
	}
	return( retval );
}

/*
 * Error stub function.  Point unresolved link errors here to avoid
 * problems.
 */
static
void*
error_stub(void)
{
	return (0);
}

static
libd_t getLibHandle(char *path)
{
	libd_t lpc, firstFree = -1, retval = -1;
	struct _libHandle *lib;

	require(path != 0);
	
	for( lpc = 0; (lpc < MAXLIBS) && (retval == -1); lpc++ )
	{
		lib = &libHandle[lpc];
		if( lib->desc == 0 )
		{
			/*
			 * Record the first free slot if we don't find a match
			 */
			if( firstFree == -1 )
				firstFree = lpc;
		}
		else if( strcmp(lib->name, path) == 0 )
		{
			/* Found it... */
			lib->ref++;
			retval = lpc;
		}
	}
	if( retval == -1 )
	{
		/* No match, return the first free slot, if any */
		retval = firstFree;
		libHandle[retval].ref = 1;
	}

	ensure((retval == -1) || ((retval >= 0) && (retval < MAXLIBS)));
	
	return( retval );
}

static
void freeLibHandle(libd_t slot)
{
	require(slot >= 0);
	require(slot < MAXLIBS);
	
	if( libHandle[slot].desc != 0 )
	{
		blockAsyncSignals();
		LIBRARYUNLOAD(libHandle[slot].desc);
		unblockAsyncSignals();
		libHandle[slot].desc = 0;
	}
	/* NB: The `name' slot isn't necessarily initialized... */
	gc_free(libHandle[slot].name);
	libHandle[slot].name = 0;
	libHandle[slot].ref = 0;
}

static
void dropLibHandle(libd_t slot)
{
	require(slot >= 0);
	require(slot < MAXLIBS);

	libHandle[slot].ref--;
	if( libHandle[slot].ref == 0 )
	{
		freeLibHandle(slot);
	}
}

struct parsedLibraryPath {
	nativeResource *nr;
	nativeClient *nc;
	parsedString path;
};

static
int addLibraryPath(struct parsedLibraryPath *plp)
{
	nativeLibraryPath *nl;
	int retval = 0;

	require(plp != 0);
	require(plp->nr != 0);

	if( (nl = gc_malloc(sizeof(nativeLibraryPath) + plp->path.len + 1,
			    GC_ALLOC_EXTERNAL)) )
	{
		nl->next = plp->nr->paths;
		plp->nr->paths = nl;
		nl->path = (char *)(nl + 1);
		strncpy(nl->path, plp->path.data, plp->path.len);
		nl->path[plp->path.len] = '\0';
		if( plp->path.len > plp->nr->maxPathLen )
			plp->nr->maxPathLen = plp->path.len;
		retval = 1;

		ensure_valid_memory(nl);
	}
	return( retval );
}

static
int buildLibraryPath(void)
{
	int retval = 0;
	char* lpath;

	/* Get a library path from somewhere */
	lpath = (char*)Kaffe_JavaVMArgs[0].libraryhome;
	if (lpath == 0) {
		lpath = getenv(LIBRARYPATH);
	}

	if( lpath )
	{
		struct parsedLibraryPath plp;
		parseErrorInfo pe;

		plp.nr = &metalNativeResource;
		if( parseString_error(&pe, lpath,
				      SPO_Do,
				        SPO_String, &plp.path,
				        SPO_While, path_separator, "",
				        SPO_Handle, addLibraryPath, &plp,
				        SPO_End,
				      SPO_End) )
		{
			nativeSpec ns;

			LIBRARYINIT(); // XXX Can this fail?

			ns.spec.handler = &metalNativeResource.link;
			ns.libpath = lpath;
			metalNativeResource.libraryPath = lpath;
			retval = attachTeamResource(metalTeam, &ns.spec);
		}
		else
		{
			int lpc, offset;
			
			fprintf(stderr,	"Parse error in library path:");
			fprintf(stderr, "%s\n", lpath);
			offset = pe.position - lpath;
			for( lpc = 0; lpc < offset; lpc++ )
			{
				fprintf(stderr, " ");
			}
			fprintf(stderr, "^\n");
		}
	}
	else
	{
		fprintf(stderr, "Can't proceed without library path.\n");
	}
	return( retval );
}

static
int loadMainLibraries(void)
{
	static char *mainLibraries[] = {
		NATIVELIBRARY,
		VMLIBRARY,
		// MISCLIBRARY,
		0
	};
	
	int lpc, retval = 1;

	for( lpc = 0; mainLibraries[lpc] && retval; lpc++ )
	{
		if( findLibrary(mainLibraries[lpc]) < 0 )
		{
			fprintf(stderr,
				"Failed to locate library: %s\n",
				mainLibraries[lpc]);
			fprintf(stderr,
				"Library path is %s\n",
				metalNativeResource.libraryPath);
			fprintf(stderr, "Aborting.\n");
			fflush(stderr);
			retval = 0;
		}
	}
	return( retval );
}

int initNativeResourceModule(int stage)
{
	int retval = 0;
	
	switch( stage )
	{
	case SID_NATIVERESOURCEMODULE_INITMETAL:
		/* Add the JSI handlers... */
		jsiInitNamedHandler(
			&nRootHandler,
			JNA_Name, "nativepath",
			JNA_Functions, &nRootFunctions,
			JNA_HandleMessage, nHandleRootMessage,
			JNA_GetSuites, nGetSuites,
			JNA_TAG_DONE);
		jsiInstallHandler(&jvmHandlerList, &nRootHandler);
		
		installResource(&metalNativeResource.link,
				RA_Name, "native",
				RA_Flags, RF_INTERNAL,
				RA_Functions, &nativeResourceFunctions,
				RA_ClientSize, sizeof(nativeClient),
				RA_SpecSize, sizeof(nativeSpec),
				RA_TAG_DONE);
		if( buildLibraryPath() )
		{
			retval = loadMainLibraries();
		}
		break;
	case SID_NATIVERESOURCEMODULE:
		installResource(&metalNativeResource.link,
				RA_SpecNativeClass, JAVA_NATIVE_SPEC,
				RA_TAG_DONE);
		retval = 1;
		break;
	}
	return( retval );
}

static
nativeLibraryPath *findLibraryPath(nativeResource *nr, parsedString *path)
{
	nativeLibraryPath *nlp, *retval = 0;

	require(nr != 0);
	require(path != 0);
	
	nlp = nr->paths;
	while( nlp && !retval )
	{
		if( !cmpPStrStr(path, nlp->path) )
		{
			retval = nlp;
		}
		nlp = nlp->next;
	}
	return( retval );
}

static
int nativeClientHasPath(nativeClient *nc, parsedString *path)
{
	nativeLibraryPath *nlp;
	int lpc, retval = 0;

	require(nc != 0);
	require(path != 0);
	
	for( lpc = 0; (lpc < nc->path.count) && !retval; lpc++ )
	{
		nlp = nc->path.data[lpc];
		if( !cmpPStrStr(path, nlp->path) )
		{
			retval = 1;
		}
	}
	return( retval );
}

static
int addClientLibraryPath(struct parsedLibraryPath *plp)
{
	nativeLibraryPath *nlp;
	int retval = 0;

	require(plp != 0);
	require(plp->nr != 0);
	require(plp->nc != 0);
	
	if( nativeClientHasPath(plp->nc, &plp->path) )
	{
		/* Already added... */
		retval = 1;
	}
	else if( (nlp = findLibraryPath(plp->nr, &plp->path)) )
	{
		nativeLibraryPath **new_data;
		
		if( (new_data = (nativeLibraryPath **)
		     gc_realloc(plp->nc->path.data,
				sizeof(nativeLibraryPath *) *
				(plp->nc->path.count + 1),
				GC_ALLOC_EXTERNAL)) )
		{
			int slot;
			
			plp->nc->path.data = new_data;
			slot = plp->nc->path.count++;
			plp->nc->path.data[slot] = nlp;
			retval = 1;
		}
		else
		{
		}
	}
	else
	{
		teamLogEntry("nativeResource",
			     LEA_String, "Error", "Path not available", -1,
			     LEA_String, "path", plp->path.data, plp->path.len,
			     LEA_TAG_DONE);
	}
	return( retval );
}

resourceSpec *nativeCreateJSISpec(struct _resource *res, jsiMessage *jm)
{
	resourceSpec *retval = 0;
	int count;

	/* Count the number of class path strings that need to be processed */
	count = jsiMessageIdentifierCount(jm, "nativepath");
	if( count && (retval = createResourceSpec(res)) )
	{
		nativeSpec *ns = (nativeSpec *)retval;
		int len = 1, lpc, error = 0;
		char *path, *new_path;
		
		for( lpc = 0;
		     !error &&
		     jsiMessageGet(jm,
				   JMA_String, "nativepath", lpc, &path,
				   JMA_TAG_DONE);
		     lpc++ )
		{
			/* Append separator and the path */
			len += strlen(path_separator) + strlen(path);
			if( (new_path = gc_realloc(ns->libpath,
						   len,
						   GC_ALLOC_EXTERNAL)) )
			{
				ns->libpath = new_path;
				if( lpc > 0 )
					strcat(ns->libpath, path_separator);
				strcat(ns->libpath, path);
			}
			else
			{
				error = 1;
			}
		}
		if( error )
		{
			gc_free(ns->libpath);
			ns->libpath = 0;
			deleteResourceSpec(retval);
			retval = 0;
		}
		else
		{
			ensure_valid_memory(ns->libpath);
		}
	}
	return( retval );
}

void nativeDeleteJSISpec(struct _resource *res, struct _resourceSpec *rs)
{
	nativeSpec *ns = (nativeSpec *)rs;

	if( ns->libpath[0] )
	{
		require_valid_memory(ns->libpath);
	}
	
	gc_free(ns->libpath);
	deleteResourceSpec(rs);
}

int nativeSubscribe(struct _resource *res,
		    struct _resourceClient *rc,
		    struct _resourceSpec *rs)
{
	nativeResource *nr = (nativeResource *)res;
	nativeClient *nc = (nativeClient *)rc;
	nativeSpec *ns = (nativeSpec *)rs;
	int retval = 1;

	if( ns && ns->libpath )
	{
		struct parsedLibraryPath plp;

		plp.nr = nr;
		plp.nc = nc;
		if( parseString(ns->libpath,
				SPO_Do,
				  SPO_String, &plp.path,
				  SPO_While, path_separator, "",
				  SPO_Handle, addClientLibraryPath, &plp,
				  SPO_End,
				SPO_End) )
		{
		}
		else
		{
			teamLogEntry("nativeResource",
				     LEA_String, "Error", "Can't parse requested path", -1,
				     LEA_String, "path", ns->libpath, -1,
				     LEA_TAG_DONE);
			retval = 0;
		}
	}
	else
	{
		/* Default to no library path. */
	}
	return( retval );
}

static
void killLibraryModule(libd_t slot)
{
	char miName[MAXLIBPATH]; // XXX
	moduleInitializer_t *mi;
	struct libraryName ln;
	
	parseLibraryName(&ln, libHandle[slot].name);
	sprintf(miName, "%sModuleInitList", ln.base.data);
	blockAsyncSignals();
	INITSYMBOL(mi, miName);
	unblockAsyncSignals();
	if( mi )
	{
		killModules(mi);
	}
}

void nativeUnsubscribe(struct _resource *res, struct _resourceClient *rc)
{
	nativeResource *nr = (nativeResource *)res;
	nativeClient *nc = (nativeClient *)rc;
	nativeLibrary *nl;
	int iLockRoot;

	if( nc->libraries )
	{
		/* Run the module finalizers for each library */
		nl = nc->libraries;
		while( nl )
		{
			killLibraryModule(nl->libID);
			nl = nl->next;
		}
		/* Switch to the kernel and drop the references */
		if( switchToTeam(kernelTeam, 0) )
		{
			lockMutex(nr);
			nl = nc->libraries;
			while( nl )
			{
				dropLibHandle(nl->libID);
				nl = nl->next;
			}
			unlockMutex(nr);
			returnFromTeam();
		}
		else
		{
			/* Really should not happen. */
			assert(0);
		}
	}
}

static
int initializeLibrary(libd_t slot, struct libraryName *ln)
{
	char miName[MAXLIBPATH]; // XXX
	moduleInitializer_t *mi;
	int retval = 1;
	
	/*
	 * The moduleInitList variables will clash in a statically
	 * linked environment so we must use a unique name for each.
	 * So the convention is use the base name of the library to
	 * prefix the variables.
	 */
	
	sprintf(miName, "%sModuleInitList", ln->base.data);
	blockAsyncSignals();
	INITSYMBOL(mi, miName);
	unblockAsyncSignals();
	if( mi )
	{
		if( initModules(mi) )
		{
		}
		else
		{
			teamLogEntry("native",
				     LEA_String, "Error", "Library initializer(s) failed", -1,
				     LEA_String, "library", ln->path.data, ln->path.len,
				     LEA_TAG_DONE);
			retval = 0;
		}
	}
	else
	{
		/* No initializer */
	}
	return( retval );
}

static char *libSuffixes[] = {
	".la",
	".so",
	/* XXX Need a better way to do this! */
	0
};

enum {
	TRY_LOAD_FOUND,
	TRY_LOAD_NOT_FOUND,
	TRY_LOAD_ERROR,
};

/*
 * Try to load the library using a variety of suffixes.
 *
 * `lib' - The libHandle with the name field initialized.
 */
static
int tryLoad(struct _libHandle *lib,
	    struct libraryName *ln,
	    char *errbuf,
	    size_t errsize)
{
	int lpc, retval = TRY_LOAD_NOT_FOUND;

	blockAsyncSignals();
	for( lpc = 0;
	     (retval == TRY_LOAD_NOT_FOUND) && libSuffixes[lpc];
	     lpc++ )
	{
		sprintf(lib->name, "%s%s", ln->path.data, libSuffixes[lpc]);
		LIBRARYLOAD(lib->desc, lib->name);
		if( lib->desc )
		{
			retval = TRY_LOAD_FOUND;
		}
		else
		{
			const char *err = LIBRARYERROR();

			/* XXX Bleh, silly guessing system. */
			if( (strstr(err, "ile not found") ||
			     strstr(err, "annot open")) )
			{
				retval = TRY_LOAD_NOT_FOUND;
			}
			else
			{
				/* We'll assume its a real error. */
				retval = TRY_LOAD_ERROR;
				strncpy(errbuf, err, errsize);
			}
		}
	}
	unblockAsyncSignals();
	return retval;
}

/*
 * Link in a native library. If successful, returns an index >= 0 that
 * can be passed to unloadNativeLibrary(). Otherwise, returns -1 and
 * fills errbuf (if not NULL) with the error message. Assumes synchronization.
 */
libd_t
loadNativeLibrary_kernel(nativeClient *nc, struct libraryName *ln,
			 char *errbuf, size_t errsize)
{
	nativeResource *nr = (nativeResource *)nc->client.handler;
	libd_t slot, retval = -1;
	int iLockRoot;

	require(nc != 0);
	require(ln != 0);
	require((errbuf != 0) || (errsize == 0));
	require((currentTeam() == kernelTeam) ||
		(currentTeam() == metalTeam));

	lockMutex(nr);
	/* Get the handle on this library or a free one */
	if( (slot = getLibHandle(ln->path.data)) >= 0 )
	{
		struct _libHandle *lib = &libHandle[slot];

		if( lib->desc == 0 )
		{
			/* Not loaded yet... */

			/* Allocate some memory for the name. */
			if( (lib->name = gc_malloc(strlen(ln->path.data)
						   + 16 /* extension */
						   + 1,
						   GC_ALLOC_EXTERNAL)) )
			{
				/* Try to load it using various suffixes. */
				switch( tryLoad(lib, ln, errbuf, errsize) )
				{
				case TRY_LOAD_FOUND:
					/* Got it */
#if defined(KAFFE_FEEDBACK)
					feedbackLibrary(lib->name, true);
#endif
					retval = slot;
					break;
				case TRY_LOAD_NOT_FOUND:
					/* No library found */
					strncpy(errbuf,
						"File not found",
						errsize);
					break;
				case TRY_LOAD_ERROR:
					/*
					 * Library found, some other error
					 * occurred though.
					 */
					break;
				}
			}
			else
			{
				strncpy(errbuf, "Out of Memory", errsize);
			}
			if( retval == -1 )
			{
				freeLibHandle(slot);
			}
		}
		else
		{
			/* Already loaded */
			retval = slot;
		}
	}
	else if( errbuf != 0 )
	{
		/* No space left to load libs */
		strncpy(errbuf, "Too many open libraries.", errsize);
	}
	unlockMutex(nr);
	return( retval );
}

static
int loadNativeLibrary_team(nativeClient *nc,
			   libd_t index, struct libraryName *ln,
			   int default_refs,
			   char *errbuf, size_t errsize)
{
	nativeLibrary *nl;
	int retval = 0;

	require(index >= 0);
	require(ln);

	/* Initialize the library in this team context. */
	if( initializeLibrary(index, ln) )
	{
		/* Allocate the tracking structure. */
		if( (nl = (nativeLibrary *)gc_malloc(sizeof(nativeLibrary),
						     GC_ALLOC_EXTERNAL)) )
		{
			/* Add nativeLibrary node */
			nl->next = nc->libraries;
			nc->libraries = nl;
			nl->libID = index;
			nl->refs = default_refs;
			
			retval = 1;
		}
		else
		{
			/* Doh, no memory, deinit the library. */
			killLibraryModule(index);
			strncpy(errbuf, "Out of Memory", errsize);
		}
	}
	else
	{
		strncpy(errbuf, "Library initialized failed.", errsize);
	}
	return( retval );
}

static
libd_t loadNativeLibrary_internal(nativeClient *nc, struct libraryName *ln,
				  int default_refs,
				  char *errbuf, size_t errsize)
{
	nativeResource *nr = (nativeResource *)nc->client.handler;
	jteamReservation *jr;
	int retval = -1;
	int iLockRoot;
	
	/* The kernel manages libraries */
	if( (jr = pushTeamReservation(kernelTeam)) )
	{
		/*
		 * Theres some memory allocation done and it applies globally
		 * so we visit the kernel.
		 */
		if( visitReservedTeam(jr, 0) )
		{
			/* Try to load it */
			retval = loadNativeLibrary_kernel(nc,
							  ln,
							  errbuf,
							  errsize);
			leaveReservedTeam();
			if( retval >= 0 )
			{
				/*
				 * Its rooted in the kernel, now load it into
				 * the local team.
				 */
				if( loadNativeLibrary_team(nc, retval, ln,
						   default_refs,
						   errbuf, errsize) )
				{
					/* success */
				}
				else if( visitReservedTeam(jr, 0) )
				{
					/* Team local malloc/init failed */
					lockMutex(nr);
					dropLibHandle(retval);
					unlockMutex(nr);
					leaveReservedTeam();
					retval = -1;
				}
				else
				{
					/* Err, shouldn't happen... */
					retval = -1;
				}
			}
		}
		else if( errbuf )
		{
			/* Err, shouldn't happen... */
			strncpy(errbuf, "Can't visit kernel", errsize);
		}
		popTeamReservation(jr);
	}
	else if( errbuf )
	{
		strncpy(errbuf, "Can't reserve kernel", errsize);
	}
	return( retval );
}

static
libd_t findClientLibrary(nativeClient *nc, char *path)
{
	libd_t retval = -1;
	nativeLibrary *nl;

	require(nc != 0);
	require(path != 0);

	nl = nc->libraries;
	while( nl && (retval == -1) )
	{
		if( !strcmp(libHandle[nl->libID].name, path) )
		{
			retval = nl->libID;
			nl->refs++;
		}
		nl = nl->next;
	}
	return( retval );
}

libd_t
loadNativeLibrary(char* lib, char *errbuf, size_t errsize)
{
	return( loadNativeLibrary2(lib, 1, errbuf, errsize) );
}

libd_t
loadNativeLibrary2(char* lib, int default_refs, char *errbuf, size_t errsize)
{
	nativeClient *nc = (nativeClient *)
		getTeamInternalResource(0, &metalNativeResource.link);
	struct libraryName ln;
	libd_t retval = -1;
	int iLockRoot;

	require(lib != 0);
	require((errbuf != 0) || (errsize == 0));

	/* Parse the library into its path and base name */
	if( parseLibraryName(&ln, lib) )
	{
		lockMutex(nc);
		/*
		 * Check if we already have it and that this team can actually
		 * load it.
		 */
		if( ((retval = findClientLibrary(nc, lib)) < 0) &&
		    nativeClientHasPath(nc, &ln.path) )
		{
			/* Do the actual loading bits. */
			retval = loadNativeLibrary_internal(nc,
							    &ln,
							    default_refs,
							    errbuf,
							    errsize);
		}
		else if( errbuf )
		{
			/*
			 * Either the team already opened the library or
			 * it doesn't have access to the path.
			 *
			 * Note: Filling out the error doesn't change anything
			 * if retval >= 0, so we might as well do it.
			 */
			strncpy(errbuf,
				"Team doesn't have access to path.",
				errsize);
		}
		unlockMutex(nc);
	}
	else if( errbuf )
	{
		strncpy(errbuf, "Bad library path", errsize);
	}
	if( errbuf )
		errbuf[errsize - 1] = '\0';
	
	ensure((retval == -1) || ((retval >= 0) && (retval < MAXLIBS)));

	return( retval );
}

/*
 * Unlink a native library. Assumes synchronization.
 */
void
unloadNativeLibrary(libd_t slot)
{
	nativeClient *nc = (nativeClient *)
		getTeamInternalResource(0, &metalNativeResource.link);
	nativeLibrary *nl, **nl_prev;
	jteamReservation *jr = 0;
	int done = false;
	int iLockRoot;
	
	require(slot >= 0);
	require(slot < MAXLIBS);

	lockMutex(nc);
	nl = nc->libraries;
	nl_prev = &nc->libraries;
	while( nl && !done )
	{
		if( nl->libID == slot )
		{
			nl->refs--;
			if( nl->refs == 0 )
			{
				/*
				 * We're taking it out of the list so we
				 * don't want to be killed before we drop
				 * the team's reference.
				 */
				if( (jr = pushTeamReservation(kernelTeam)) )
				{
					*nl_prev = nl->next;
				}
				else
				{
					/*
					 * We're dying, the unsubscribe will
					 * take care of it for us.
					 */
					nl = 0;
				}
			}
			done = true;
		}
		else
		{
			nl_prev = &nl->next;
			nl = nl->next;
		}
	}
	unlockMutex(nc);
	if( nl && (nl->refs == 0) )
	{
		char miName[MAXLIBPATH]; // XXX
		moduleInitializer_t *mi;
		struct libraryName ln;

		/* Run the library finalizer, if there is one. */
		parseLibraryName(&ln, libHandle[nl->libID].name);
		sprintf(miName, "%sModuleInitList", ln.base.data);
		blockAsyncSignals();
		INITSYMBOL(mi, miName);
		unblockAsyncSignals();
		if( mi )
			killModules(mi);
		
		/* Drop the team's ref */
		lockMutex(&metalNativeResource);
		dropLibHandle(slot);
		unlockMutex(&metalNativeResource);

		/* We're done, release the kernel reservation. */
		popTeamReservation(jr);
	}
}

/*
 * Get pointer to symbol from symbol name.
 */
void*
loadNativeLibrarySym(char* name)
{
	nativeClient *nc = (nativeClient *)
		getTeamInternalResource(0, &metalNativeResource.link);
	void* func;
	int iLockRoot;

	/* XXX Is this lock necessary? */
	lockMutex(nc);
	blockAsyncSignals();
	LIBRARYFUNCTION(func, name);
	unblockAsyncSignals();
	unlockMutex(nc);
	
	return (func);
}

bool
native(Method *m, errorInfo *einfo)
{
	nativeClient *nc = (nativeClient *)
		getTeamInternalResource(0, &metalNativeResource.link);
	char symbol[MAXSTUBLEN];
	int retval = false;
	const char *name;

	if( nc->path.count == 0 )
	{
		/* XXX No native library access... */
		SET_METHOD_NATIVECODE(m, (void*)error_stub);
		postExceptionMessage(einfo,
				     JAVA_LANG(UnsatisfiedLinkError),
				     "No native library access");
	}
	else if( (name = getClassView(m->class)->name->data) )
	{
		void *function;
		int lpc;

		strcpy(symbol, STUB_PREFIX);
		for( lpc = STUB_PREFIX_LEN; *name; name++, lpc++ )
		{
			if( *name == '/' )
				symbol[lpc] = '_';
			else
				symbol[lpc] = *name;
		}
		symbol[lpc] = '_';
		symbol[lpc + 1] = 0;
		strcat(symbol, m->name->data);
		strcat(symbol, STUB_POSTFIX);
		if( (function = loadNativeLibrarySym(symbol)) )
		{
			/* Fill it in */
			extern void Kaffe_KNI_wrapper(Method* xmeth,
						      void* func);
			Kaffe_KNI_wrapper(m, function);
			retval = true;
		}
		else if( Kaffe_JNI_native(m) )
		{
			retval = true;
		}
		else
		{
			SET_METHOD_NATIVECODE(m, (void*)error_stub);
			postExceptionMessage(einfo,
					     JAVA_LANG(UnsatisfiedLinkError),
					     "Failed to locate native "
					     "function:\t%s.%s%s",
					     getClassView(m->class)->
					     name->data,
					     m->name->data,
					     METHOD_SIGD(m));
		}
	}
	else
	{
		/* No class view */
	}
	return( retval );
}

/*
 * Return the library path.
 */
char*
getLibraryPath(void)
{
	return( metalNativeResource.libraryPath );
}

libd_t
findLibrary(char *baseName)
{
	nativeClient *nc = (nativeClient *)
		getTeamInternalResource(0, &metalNativeResource.link);
	libd_t retval = -1;
	char *libPath;

	/* Allocate some scratch to build path names */
	if( (libPath = (char *)gc_malloc(metalNativeResource.
					 maxPathLen +
					 1 +
					 strlen(LIBRARYPREFIX) +
					 strlen(baseName) +
					 strlen(LIBRARYSUFFIX) +
					 1,
					 GC_ALLOC_EXTERNAL)) )
	{
		nativeLibraryPath *nlp;
		int lpc;

		for( lpc = 0; (lpc < nc->path.count) && (retval == -1); lpc++ )
		{
			nlp = nc->path.data[lpc];
			sprintf(libPath,
				"%s/%s%s%s",
				nlp->path,
				LIBRARYPREFIX,
				baseName,
				LIBRARYSUFFIX);
			/* XXX This is redundant wrt to the path */
			retval = loadNativeLibrary(libPath, NULL, 0);
		}
		gc_free(libPath);
	}
	return( retval );
}

void edu_utah_janosvm_resources_NativeSpec_nativeConstructor(
	struct Hedu_utah_janosvm_resources_NativeSpec *jns,
	HArrayOfObject *entries)
{
	nativeSpec *ns = OBJECT2NATIVE(&metalNativeResource.link.
				       specNativeClass,
				       (Hjava_lang_Object *)jns);
	char **tmp_array;
	errorInfo einfo;
	int error = 0;

	ns->spec.handler = &metalNativeResource.link;
	if( !entries || (ARRAY_SIZE(entries) == 0) )
	{
		/* Need to allocate because the finalizer will free it */
		if( (ns->libpath = gc_malloc(1, GC_ALLOC_EXTERNAL)) )
		{
			ns->libpath[0] = '\0';
		}
		else
		{
			postOutOfMemory(&einfo);
			throwError(&einfo);
		}
		return;
	}
	if( (tmp_array = gc_malloc(sizeof(char *) * ARRAY_SIZE(entries),
				   GC_ALLOC_EXTERNAL)) )
	{
		int lpc, total_len = 1, sep_len;
		struct Hjava_lang_String *curr;

		/* Note: we depend on tmp_array being initialized to zero */
		sep_len = strlen(path_separator);
		for( lpc = 0; !error && (lpc < ARRAY_SIZE(entries)); lpc++ )
		{
			curr = ((struct Hjava_lang_String **)
				ARRAY_DATA(entries))[lpc];
			if( curr == NULL )
			{
			}
			else if( (tmp_array[lpc] = stringJava2C(curr)) )
			{
				total_len += strlen(tmp_array[lpc]) + sep_len;
			}
			else
			{
				postOutOfMemory(&einfo);
				error = 1;
			}
		}
		if( !error &&
		    (ns->libpath = gc_malloc(total_len,
						GC_ALLOC_EXTERNAL)) )
		{
			char *pos = ns->libpath;
			
			for( lpc = 0; lpc < ARRAY_SIZE(entries); lpc++ )
			{
				if( tmp_array[lpc] )
				{
					sprintf(pos,
						"%s%s",
						(lpc > 0) ?
						path_separator : "",
						tmp_array[lpc]);
					pos += strlen(pos);
				}
			}

			ensure_valid_memory(ns->libpath);
		}
		else
		{
			postOutOfMemory(&einfo);
			error = 1;
		}
		for( lpc = 0; lpc < ARRAY_SIZE(entries); lpc++ )
		{
			/* Free the path strings */
			gc_free(tmp_array[lpc]);
		}
		gc_free(tmp_array);
	}
	else
	{
		postOutOfMemory(&einfo);
		error = 1;
	}
	if( error )
		throwError(&einfo);
}

void edu_utah_janosvm_resources_NativeSpec_nativeFinalizer(
	struct Hedu_utah_janosvm_resources_NativeSpec *jns)
{
	struct Hedu_utah_janosvm_resources_ResourceSpec *jrs =
		(struct Hedu_utah_janosvm_resources_ResourceSpec *)jns;
	nativeSpec *ns = OBJECT2NATIVE(&metalNativeResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jns);

	if( !jrs->found )
	{
		if( ns->libpath )
		{
			require_valid_memory(ns->libpath);
		}

		gc_free(ns->libpath);
	}
}

jint edu_utah_janosvm_resources_NativeSpec_getPathCount(
	struct Hedu_utah_janosvm_resources_NativeSpec *jns)
{
	nativeSpec *ns = OBJECT2NATIVE(&metalNativeResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jns);
	jint retval = 0;

	parseString(ns->libpath,
		    SPO_NotEmpty,
		      SPO_Do,
		        SPO_Count, &retval,
		        SPO_While, path_separator, "",
		        SPO_End,
		      SPO_End,
		    SPO_End);
	return( retval );
}

struct parsedSingleNative {
	parsedString entry;
	parsedString extracted;
	int current;
	int requested;
};

static
int extractNative(struct parsedSingleNative *pcp)
{
	int retval = 1;

	if( pcp->current == pcp->requested )
	{
		pcp->extracted = pcp->entry;
	}
	pcp->current++;
	return( retval );
}

Hjava_lang_String *edu_utah_janosvm_resources_NativeSpec_getPath(
	struct Hedu_utah_janosvm_resources_NativeSpec *jns,
	jint index)
{
	nativeSpec *ns = OBJECT2NATIVE(&metalNativeResource.link.
					   specNativeClass,
					   (Hjava_lang_Object *)jns);
	struct Hjava_lang_String *retval = 0;
	struct parsedSingleNative pcp;

	pcp.current = 0;
	pcp.requested = index;
	pcp.extracted.data = 0;
	pcp.extracted.len = 0;
	if( parseString(ns->libpath,
			SPO_NotEmpty,
			  SPO_Do,
			    SPO_String, &pcp.entry,
			    SPO_While, path_separator, "",
			    SPO_Handle, extractNative, &pcp,
			    SPO_End,
			  SPO_End,
			SPO_End) )
	{
		if( pcp.extracted.data )
		{
			retval = checkPtr(stringCBuf2Java(pcp.extracted.data,
							  pcp.extracted.len));
		}
	}
	return( retval );
}

struct Hedu_utah_janosvm_resources_NativeSpec *
edu_utah_janosvm_resources_NativeSpec_current(void)
{
	struct Hedu_utah_janosvm_resources_NativeSpec *retval = 0;
	nativeClient *nc = (nativeClient *)
		getTeamInternalResource(0, &metalNativeResource.link);
	nativeSpec *ns;
	int lpc, len = 1;

	retval = (struct Hedu_utah_janosvm_resources_NativeSpec *)
		newObject(metalNativeResource.link.specNativeClass.peer);
	ns = OBJECT2NATIVE(&metalNativeResource.link.specNativeClass,
			   (Hjava_lang_Object *)retval);
	ns->spec.handler = &metalNativeResource.link;
	for( lpc = 0; lpc < nc->path.count; lpc++ )
	{
		len += strlen(nc->path.data[lpc]->path) + 1;
	}
	if( (ns->libpath = gc_malloc(len, GC_ALLOC_EXTERNAL)) )
	{
		ns->libpath[0] = '\0';
		for( lpc = 0; lpc < nc->path.count; lpc++ )
		{
			if( lpc > 0 )
				strcat(ns->libpath, ":");
			strcat(ns->libpath,
			       nc->path.data[lpc]->path);
		}
	}
	else
	{
		errorInfo einfo;
		
		postOutOfMemory(&einfo);
		throwError(&einfo);
	}
	return( retval );
}

Hjava_lang_Object *
edu_utah_janosvm_resources_NativeSpec_clone(
	struct Hedu_utah_janosvm_resources_NativeSpec *jns)
{
	struct Hedu_utah_janosvm_resources_NativeSpec *retval;
	nativeSpec *ns, *retval_ns;

	retval = (struct Hedu_utah_janosvm_resources_NativeSpec *)
		newObject(metalNativeResource.link.specNativeClass.peer);
	ns = OBJECT2NATIVE(&metalNativeResource.link.specNativeClass,
			   (Hjava_lang_Object *)jns);
	retval_ns = OBJECT2NATIVE(&metalNativeResource.link.
				  specNativeClass,
				  (Hjava_lang_Object *)retval);
	retval_ns->spec.handler = &metalNativeResource.link;
	if( (retval_ns->libpath = gc_malloc(strlen(ns->libpath) + 1,
					      GC_ALLOC_EXTERNAL)) )
	{
		strcpy(retval_ns->libpath, ns->libpath);
	}
	else
	{
		errorInfo einfo;
		
		postOutOfMemory(&einfo);
		throwError(&einfo);
	}
	return( (Hjava_lang_Object *)retval );
}

char *nHandleRootMessage(jsiHandlerContext *hc,
			 jsiHandler *jh, jsiMessage *jm)
{
	char *retval = 0;

	switch( jm->what )
	{
	case JSI_GET_PROPERTY:
		{
			nativeLibraryPath *nlp;
			int iLockRoot;
			
			jsiClearMessage(jm);
			jm->what = JSI_REPLY;
			nlp = metalNativeResource.paths;
			lockMutex(&metalNativeResource);
			while( nlp && !retval )
			{
				if( !jsiMessageAdd(
					jm,
					JMA_String, "result", nlp->path,
					JMA_TAG_DONE) )
				{
					retval = "Out of memory";
				}
				nlp = nlp->next;
			}
			unlockMutex(&metalNativeResource);
		}
		break;
	default:
		retval = "Unsupported operation";
		break;
	}
	return( retval );
}

int nGetSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", nRootProperties,
			       JMA_TAG_DONE);
	return( retval );
}
