/*
 * nativeResource.h
 * Provides access to the native libraries.
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _native_resource_h
#define _native_resource_h

#include "resource.h"

/* Default library info. */
#if !defined(LIBRARYPATH)
#define	LIBRARYPATH	"KAFFELIBRARYPATH"
#endif

/*
 * A nativeLibraryPath is a linked list node that holds an individual library
 * path.
 *
 * `next' - Link to the next path in the list.
 * `path' - The native library path.
 */
typedef struct _nativeLibraryPath {
	struct _nativeLibraryPath *next;
	char *path;
} nativeLibraryPath;

typedef int libd_t;

/*
 * The nativeLibrary structure tracks the references to libraries that a
 * team has opened.
 *
 * `next' - Link to the next node in the list.
 * `libID' - The library descriptor.
 * `refs' - The number of references inside the team.
 */
typedef struct _nativeLibrary {
	struct _nativeLibrary *next;
	libd_t libID;
	int refs;
} nativeLibrary;

/*
 * The nativeResource tracks the globally available library path.
 *
 * `link' - The resource header.
 * `lock' - Lock protecting the library paths.
 * `libraryPath' - The string representation of the global library.
 *  path.
 * `maxPathLen' - The length of the longest path string.
 * `paths' - The linked list of parsed native library paths.
 */
typedef struct _nativeResource {
	resource link;
	iLock *lock;
	char *libraryPath;
	int maxPathLen;
	nativeLibraryPath *paths;
} nativeResource;

/*
 * The nativeClient tracks the library usage for a given team.
 *
 * `client' - The resourceClient header.
 * `lock' - Lock protecting the list of opened libraries.
 * `path.data' - Array of library paths that this team has access too.
 * `path.count' - The size of the `path.data' array.
 * `libraries' - List of libraries opened by this team.
 */
typedef struct _nativeClient {
	resourceClient client;
	iLock *lock;
	struct {
		nativeLibraryPath **data;
		int count;
	} path;
	nativeLibrary *libraries;
} nativeClient;

/**
 * The nativeSpec allows you to specify the requirements for a team's native
 * library needs.
 *
 * `spec' - The resourceSpec header.
 * `libpath' - The list of library paths.
 */
typedef struct _nativeSpec {
	resourceSpec spec;
	char *libpath;
} nativeSpec;

/*
 * Load a native library into the system, if the library has already been
 * loaded then it will increment a reference count and return.  Otherwise,
 * it will try to load the library from the file system and run any module
 * initializers in the library.
 *
 * `path' - The full path for the library to load.  Use the findLibrary
 *   function if you don't know the full path and prefix of the library.
 * `errbuf' - A character array that can hold an error string.
 * `errsiz' - The size of the `errbuf' array.
 * returns - -1, if the load failed, otherwise its an index number that can
 *   be passed to unloadNativeLibrary to unload this library.
 */
libd_t	loadNativeLibrary(char *path, char *errbuf, size_t errsiz);
/*
 * Similar to the above except it allows you specify the reference count to use
 * when the library is first loaded.  This is primarily useful when preloading
 * libraries so that you can give it a zero reference count thus allowing the
 * library to eventually be unloaded without you having to call unloadLibrary.
 * NOTE: This should only be called from within the kernel team.
 */
libd_t
loadNativeLibrary2(char *lib, int default_refs, char *errbuf, size_t errsiz);
/*
 * Unload the native library from the system.  Basically, this will just
 * decrement the reference count of the library and then release it if it
 * reaches zero.
 *
 * `index' - The library index as returned by loadNativeLibrary.
 */
void	unloadNativeLibrary(libd_t index);
/**
 * Locate the given symbol in one of the loaded libraries and return a pointer
 * to its value.
 *
 * `name' - The name of the symbol.
 * returns - A pointer to the symbol or NULL if it wasn't found.
 */
void*	loadNativeLibrarySym(char *name);
/*
 * Locate and construct a stub that will call the native implementation of the
 * given method.  Since Kaffe supports two different native interfaces this
 * function will first try to bind to the KNI function and then try the
 * JNI version.  If both of these fail then it will install a pointer to a
 * predefined error stub that will just returns zero.  (XXX Is this wise?
 * should it throw an exception?)
 *
 * `method' - The native method to construct a stub for.
 * `einfo' - An errorInfo structure to hold any exception info.
 * returns - True if the stub was constructed successfully, false otherwise.
 */
bool	native(struct _methods *method, struct _errorInfo *einfo);
/*
 * returns - The library path that Kaffe is using.
 */
char*	getLibraryPath(void);
/*
 * Find and load a native library with the given name.  The function will
 * search through all of the library directories until it either finds the
 * library or runs out of paths.
 *
 * `baseName' - The name of the library to load, this doesn't include any
 * paths, prefixes, or suffixes.
 * returns - -1, if the load failed, otherwise it returns the index number
 * as returned by loadNativeLibrary.
 */
libd_t	findLibrary(char *baseName);

/*
 * AddModule: nativeResourceModule
 *   Comment: Install the resource and subscribe the metal team.
 *   ModuleDependency: memResourceModule.initMetal gcModule.heap
 *   ModuleDependency: resourceModule.internalResources
 * Transition: initMetal
 *   ModuleDependency: resourceModule.reinstall
 */

#endif
