/*
 * resource.c
 * Manages VM internal resources
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdarg.h>

#include "locks.h"
#include "assert_pp.h"
#include "thread.h"
#include "resource.h"
#include "resource-inline.h"
#include "gc.h"
#include "jteam.h"
#include "object.h"
#include "configFile.h"
#include "mungWallProxy.h"
#include "nativeClass.h"
#include "nativeClass-inline.h"
#include "module.h"

/** Module declarations **/
module_t resourceModule = {
	"resource",
	0,
	0,
	0,
	0
};

/** Resource section file declarations **/
static
int resourceSectionWalker(sectionType *st, sectionFileData *sfd,
			  sectionFile *sf);

static sectionTypeFunctions resourceSectionFunctions = {
	0,
	0,
	0,
	0,
	0,
	0,
	resourceSectionWalker
};

/** Spec section file declarations **/
static
int specSectionWalker(sectionType *st, sectionFileData *sfd,
		      sectionFile *sf);

static sectionTypeFunctions specSectionFunctions = {
	0,
	0,
	0,
	0,
	0,
	0,
	specSectionWalker
};

/** resource variables **/
static resource *resourceList;
static iLock *resourceListLock;

int internalResourceCount;

/** Resource section file functions **/
static
int resourceSectionWalker(sectionType *st, sectionFileData *sfd,
			  sectionFile *sf)
{
	resource *res = (resource *)
		(((uintp)st) -
		 (int)(&((resource *)0)->resourceSection));
	int retval;

	retval = rsrcConfigure(res,
			       sf,
			       LOCAL_SECTION_DATA(tagSectionData *,
						  st,
						  sfd));

	ensure_valid_memory(sfd);
	
	return( retval );
}

/** Spec section file functions **/
static
int specSectionWalker(sectionType *st, sectionFileData *sfd, sectionFile *sf)
{
	resource *res = (resource *)
		(((uintp)st) -
		 (int)(&((resource *)0)->specSection));
	tagSectionData *tsd = LOCAL_SECTION_DATA(tagSectionData *, st, sfd);
	resourceSpec *rs = (resourceSpec *)(((uintp)tsd) +
					    sizeof(tagSectionData));
	int retval;

	rs->handler = res;
	retval = rsrcFillSpec(res, rs, sf, tsd);
	
	ensure_valid_memory(sfd);
	
	return( retval );
}

/** Resource functions **/

void installResource(resource *res, int tag, ...)
{
	tagArray *resDefaultArray = 0, *specDefaultArray = 0;
	char *nativeName = 0, *resName = 0, *specName = 0;
	va_list args;

	va_start(args, tag);
	while( tag != RA_TAG_DONE )
	{
		switch( tag )
		{
		case RA_Name:
			res->name = va_arg(args, char *);
			break;
		case RA_Flags:
			res->flags = va_arg(args, unsigned int);
			break;
		case RA_Functions:
			res->funcs = va_arg(args, resourceFunctions *);
			break;
		case RA_ClientSize:
			res->clientSize = va_arg(args, int);
			break;
		case RA_SpecSize:
			res->specSize = va_arg(args, int);
			break;
		case RA_SpecNativeClass:
			nativeName = va_arg(args, char *);
			break;
		case RA_ResourceSectionName:
			resName = va_arg(args, char *);
			break;
		case RA_SpecSectionName:
			specName = va_arg(args, char *);
			break;
		case RA_ResourceSectionDefault:
			resDefaultArray = va_arg(args, tagArray *);
			break;
		case RA_SpecSectionDefault:
			specDefaultArray = va_arg(args, tagArray *);
			break;
		default:
			assert(0);
			break;
		}
		tag = va_arg(args, int);
	}
	if( res->flags & RF_INSTALLED )
	{
	}
	else
	{
		int iLockRoot;
		
		lockStaticMutex(&resourceListLock);
		res->next = resourceList;
		resourceList = res;
		if( res->flags & RF_INTERNAL )
		{
			res->internalIndex = internalResourceCount;
			internalResourceCount++;
		}
		else
		{
			res->internalIndex = -1;
		}
		res->flags |= RF_INSTALLED;
		unlockStaticMutex(&resourceListLock);
		rsrcInstall(res);
	}
	if( nativeName )
	{
		Hjava_lang_Class *cl;
		errorInfo einfo;

		if( (cl = lookupClass(nativeName, NULL, &einfo)) )
		{
			res->specNativeClass.size = res->specSize;
			bindNativeClass(&res->specNativeClass, cl, &einfo);
		}
	}
	if( resName )
	{
		installSectionType(&res->resourceSection.type,
				   STA_Name, resName,
				   STA_Template, &tagsFileSectionTemplate,
				   STA_Functions, &resourceSectionFunctions,
				   TSA_DefaultArray, resDefaultArray,
				   STA_TAG_DONE);
	}
	if( specName )
	{
		installSectionType(&res->specSection.type,
				   STA_Name, specName,
				   STA_Template, &tagsFileSectionTemplate,
				   STA_Functions, &specSectionFunctions,
				   STA_Size, res->specSize,
				   TSA_DefaultArray, specDefaultArray,
				   STA_TAG_DONE);
	}
	va_end(args);
}

void uninstallResource(resource *res)
{
	if( !(res->flags & RF_INTERNAL) )
	{
		resource *curr, **last;
		int iLockRoot;

		lockStaticMutex(&resourceListLock);
		last = &resourceList;
		curr = resourceList;
		while( curr && (curr != res) )
		{
			last = &curr->next;
			curr = curr->next;
		}
		if( curr == res )
		{
			*last = curr->next;
			res->internalIndex = -1;
		}
		unlockStaticMutex(&resourceListLock);
		rsrcUninstall(res);
	}
}

resource *findResource(char *name)
{
	resource *curr, *retval = 0;

	require(name != 0);
	
	curr = resourceList;
	while( curr && !retval )
	{
		if( !strcmp(curr->name, name) )
			retval = curr;
		curr = curr->next;
	}
	return( retval );
}

resource *getResourceList(void)
{
	return resourceList;
}

resourceSpec *createResourceSpec(resource *res)
{
	resourceSpec *retval;

	require(res != 0);
	
	if( (retval = gc_malloc(res->specSize, GC_ALLOC_RESOURCE)) )
	{
		retval->handler = res;
	}
	return( retval );
}

resourceSpec *createResourceSpecFromJSI(resource *res, jsiMessage *jm)
{
	resourceSpec *retval;

	require(res != 0);
	require(jm != 0);

	retval = rsrcCreateJSISpec(res, jm);
	return( retval );
}

void deleteResourceSpecFromJSI(resourceSpec *rs)
{
	if( rs )
	{
		rsrcDeleteJSISpec(rs);
	}
}

resourceSpec *defaultResourceSpec(resource *res)
{
	resourceSpec *retval;

	require(res != 0);

	retval = rsrcDefaultSpec(res);
	return( retval );
}

void deleteResourceSpec(resourceSpec *rs)
{
	if( rs )
	{
		gc_free(rs);
	}
}

/* Allocates a client and tries to subscribe to the resource */
static
resourceClient *subscribeResourcePrivate(jteam *jt, resource *res,
					 resourceSpec *rs)
{
	resourceClient *retval = 0;

	require(jt != 0);
	require(res != 0);

	if( (retval = (resourceClient *)gc_malloc(res->clientSize,
						  GC_ALLOC_RESOURCE)) )
	{
		retval->next = 0;
		retval->owner = jt;
		retval->handler = res;
		retval->id = -1;
		if( !rsrcSubscribe(retval, rs) )
		{
			gc_free(retval);
			retval = NULL;
		}
	}
	return( retval );
}

resourceClient *subscribeResource(jteam *jt, resourceSpec *rs)
{
	resourceClient *retval;
	resource *res;

	require(jt != 0);
	require(rs != 0);
	require(rs->handler != 0);

	res = rs->handler;
	retval = subscribeResourcePrivate(jt, res, rs);
	return( retval );
}

int enableResource(resourceClient *rc)
{
	int retval;
	
	require(rc != 0);

	retval = rsrcEnable(rc);
	return( retval );
}

resourceClient *subscribeResourceDefault(jteam *jt, resource *res)
{
	resourceClient *retval;

	require(jt != 0);
	require(res != 0);

	retval = subscribeResourcePrivate(jt, res, 0);
	return( retval );
}

void unsubscribeResource(resourceClient *rc)
{
	if( rc )
	{
		rsrcUnsubscribe(rc);
		/*
		 * XXX Needed for freeing memResourceClient.  Really we need
		 * a way to get `bootstrap' memory for a team so we can use
		 * that before the gc has started up.
		 */
		gc_free_in_heap(rc);
	}
}

struct Hedu_utah_janosvm_resources_ResourceSpec *
edu_utah_janosvm_resources_ResourceSpec_nativeFind(
	struct Hjava_lang_String *jName)
{
	struct Hedu_utah_janosvm_resources_ResourceSpec *retval = 0;
	errorInfo einfo;
	char *cName;

	if( (cName = stringJava2C(jName)) )
	{
		sectionFileData *sfd;
		
		if( (sfd = findSectionInFile(currentConfigFile(), 0, cName)) &&
		    (sfd->type->stf == &specSectionFunctions) )
		{
			sectionType *st = sfd->type;
			tagSectionData *tsd =
				LOCAL_SECTION_DATA(tagSectionData *, st, sfd);
			resourceSpec *rs = (resourceSpec *)
				(((uintp)tsd) +
				 tagsFileSectionTemplate.type.dataSize);
			resource *res = rs->handler;
			
			if( res->specNativeClass.peer )
			{
				if( (retval = (struct Hedu_utah_janosvm_resources_ResourceSpec *)
				     newObjectChecked(res->specNativeClass.
						      peer,
						      &einfo)) )
				{
					retval->found = 1;
					memcpy(retval +
					       res->specNativeClass.offset,
					       rs,
					       res->specNativeClass.size);
				}
			}
		}
		gc_free(cName);
	}
	/* XXX Should throwError(&einfo) here */
	return( retval );
}
