/*
 * resource.h
 * Manages VM internal resources
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _resource_h
#define _resource_h

/*
 * The resource module is used to track any VM resources.
 *
 * There are two kinds of resources tracked: "internal" and "external".
 * "internal" resources are more efficient to lookup, but the VM must
 * know about them all very early in the initialization.  "external"
 * resources can be added later (e.g., by external libraries to track
 * per-team resources that the VM doesn't know anything about).
 */

#include <stdarg.h>
#include <misc/sectionFile.h>

#include "tags.h"
#include "nativeClass.h"
#include "edu_utah_janosvm_resources_ResourceSpec.h"

struct _jteam;
struct _resource;
struct _resourceSpec;
struct _resourceClient;
struct _jsiMessage;

/*
 * The resourceFunctions structure contains pointers to functions that
 * implement the various resource management methods.
 *
 * `install' - Called when the resource is installed.
 * `uninstall' - Called when the resource is uninstalled.
 * `configure' - Called to ask the resource to configure itself.
 * `fillSpec' - Fill out a resourceSpec structure with the data specified
 *   in a sectionFile.
 * `createJSISpec' - Create a spec from a jsiMessage.
 * `deleteJSISpec' - Delete a spec created from the createJSISpec method.
 * `subscribe' - Subscribe to the resource.
 * `unsubsribe' - Drop a subscription to a resource.
 */
typedef struct resourceFunctions {
	int (*install)(struct _resource *res);
	void (*uninstall)(struct _resource *res);
	int (*configure)(struct _resource *res,
			 sectionFile *sf,
			 tagSectionData *tsd);
	int (*fillSpec)(struct _resource *res,
			struct _resourceSpec *rs,
			sectionFile *sf,
			tagSectionData *tsd);
	struct _resourceSpec *(*createJSISpec)(struct _resource *res,
					       struct _jsiMessage *jm);
	void (*deleteJSISpec)(struct _resource *res, struct _resourceSpec *rs);
	struct _resourceSpec *(*defaultSpec)(struct _resource *res);
	int (*subscribe)(struct _resource *res,
			 struct _resourceClient *rc,
			 struct _resourceSpec *rs);
	int (*enable)(struct _resource *res, struct _resourceClient *rc);
	void (*unsubscribe)(struct _resource *res, struct _resourceClient *rc);
} resourceFunctions;

/*
 * The resource structure is the base for specifying VM internal
 * resources.  Generally, there is only one instance of a particular
 * resource structure.  Each team has a "resource client" that contains
 * that team's resource-specific limits and usage information.
 *
 * `next' - A link to the next resource in the global list.
 * `funcs' - The implementation for this resource.
 * `flags' - Bitmask for holding flags.
 * `name' - The name of this resource.
 * `internalIndex' - The index used to refer to internal resources in the
 *   jteam internalResources array.
 * `access' - Access rights to this resource.
 * `clientSize' - The size of the resourceClient specific to this resource.
 * `speciSize' - The size of the resourceSpec specific to this resource.
 * `resourceSection' - A section file type used to get configuration data
 *   for the resource.
 * `specSection' - A section file type used to process resourceSpecs for this
 *   resource.
 */

enum {
	RB_INSTALLED,	/* The resource installation is complete */
	RB_INTERNAL,	/* The resource is internal to the VM */
};

enum {
	RF_INSTALLED = (1L << RB_INSTALLED),
	RF_INTERNAL = (1L << RB_INTERNAL),
};

typedef struct _resource {
	struct _resource *next;
	struct _iLock *lock;
	resourceFunctions *funcs;
	unsigned long flags;
	char *name;
	int internalIndex;
	int clientSize;
	int specSize;
	nativeClass specNativeClass;
	tagSectionType resourceSection;
	tagSectionType specSection;
} resource;

/*
 * The resourceClient structure is the base type for all 
 * per-resource, per-client state.
 *
 * `next' - Link to the next client in a list.
 * `owner' - The client team of the subscribed resource.
 * `hander' - The resource that handles the requests of this client.
 * `id' - For differentiating resource clients that have the same
 *   underlying resource (e.g., multiple "network resources" or something).
 *   Not used by any internal resource.
 */
typedef struct _resourceClient {
	struct _resourceClient *next;
	struct _jteam *owner;
	resource *handler;
	int id;
} resourceClient;

/*
 * The resourceSpec structure is the base for specifying the needs of a client.
 *
 * `handler' - The resource that we are going to make a request on.
 */
typedef struct _resourceSpec {
	resource *handler;
} resourceSpec;

/*
 * Function tags for installResource.
 */
enum {
	RA_TAG_DONE,
	RA_Name,		/* (char *) - The resource name */
	RA_Flags,		/* (int) - Resource flags */
	RA_Functions,		/*
				 * (resourceFunctions *) - The function table
				 * for this resource.
				 */
	RA_ClientSize,		/* (int) - The size of the client object */
	RA_SpecSize,		/* (int) - The size of the spec object */
	RA_SpecNativeClass,	/*
				 * (char *) - Name of the java spec class for
				 * this resource
				 */
	RA_ResourceSectionName,	/*
				 * (char *) - The name for the resource section
				 * type.
				 */
	RA_SpecSectionName,	/*
				 * (char *) - The name for the spec section
				 * type.
				 */
	RA_ResourceSectionDefault,	/*
					 * (tagArray *) - The default tagArray
					 * for the resource section type.
					 */
	RA_SpecSectionDefault,	/*
				 * (tagArray *) - The default tagArray for the
				 * spec section type.
				 */
};

/*
 * Install the resource into the system.  Note:  This can be called multiple
 * times on the same resource if its not possible to completely initialize
 * it all at once.
 *
 * `res' - The resource object to initialize.
 * `...' - The tag list used to initialize `res'.
 */
void installResource(resource *res, int tag, ...);
/*
 * Uninstall the resource.
 *
 * `res' - An installed resource object.
 */
void uninstallResource(resource *res);
/*
 * Find a resource with the given name.
 *
 * `name' - The name of the resource object.
 * returns - The resource object with the given name or NULL if it wasn't
 *  found.
 */
resource *findResource(char *name);
/*
 * returns - The complete list of resources.
 */
resource *getResourceList(void);

/*
 * Create a resource spec of the appropriate size for the given resource.
 *
 * `res' - The resource to create a spec for.
 * returns - A partially initialized resource spec object or NULL if the
 *  allocation failed.
 */
resourceSpec *createResourceSpec(resource *res);
/*
 * Create and initialize a resource specification from the given jsiMessage.
 * XXX This really shouldn't be here, resources shouldn't have to have direct
 * knowledge of JSI.  A separate chunk of code should do the translation from
 * user provided data to the actual spec.
 *
 * `res' - The resource to create a spec for.
 * `jm' - The jsiMessage to use when initializing the spec.
 * returns - An initialized resourceSpec or NULL if something didn't work.
 */
resourceSpec *createResourceSpecFromJSI(resource *res, struct _jsiMessage *jm);
resourceSpec *defaultResourceSpec(resource *res);
/*
 * Destruct and free the given resource spec that was allocated with
 * createResourceSpec.
 *
 * `rs' - The resourceSpec to delete.
 */
void deleteResourceSpec(resourceSpec *rs);
/*
 * Delete a resourceSpec that was allocated with createResourceSpecFromJSI.
 *
 * `rs' - A resourceSpec as returned by createResourceSpecFromJSI.
 */
void deleteResourceSpecFromJSI(resourceSpec *rs);

/*
 * Subscribe to a resource with the given specifications.
 *
 * `jt' - The team that wants to subscribe to the resource.
 * `rs' - Specification of the resource requirements of the team.
 * returns - A resourceClient to use when interacting with the resource or NULL
 *  if the requirements couldn't be satisfied.
 */
resourceClient *subscribeResource(struct _jteam *jt, resourceSpec *rs);
/*
 * Prime the resource for general use.
 *
 * `rc' - The resource to prime.
 * returns - True if the resource primed itself correctly, false otherwise.
 */
int enableResource(resourceClient *rc);
/*
 * Subscribe to a resource using their default specification
 *
 * `jt' - The team that wants to subscribe to the resource.
 * `res' - The resource to subscribe too, its up to it to determine what a
 *  decent set of default requirements are for a team.
 * returns - A resourceClient to use when interacting with the resource or NULL
 *  if the default requirements couldn't be satisfied.
 */
resourceClient *subscribeResourceDefault(struct _jteam *jt, resource *res);
/*
 * Unsubcribe from the given resource.
 *
 * `rc' - A resourceClient as returned by subscribeResource*.
 */
void unsubscribeResource(resourceClient *rc);

/*
 * The number of internal resources in the system.  Used to determine the
 * array size to allocate for each jteam.
 */
extern int internalResourceCount;

/*
 * AddModule: resourceModule
 *   ModuleDependency: machineDependentModule
 *   Comment: Make sure installResource is gonna work correctly
 * Implementation: internalResources
 *   ModuleDependency: sectionFileModule.types baseClassesModule
 *   Comment: The internal resources are installed, we're good to go.
 * Implementation: reinstall
 */

#endif
