/*
 * scripting.h
 * An implementation for providing jsi scripting in the vm.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <jsi/jsi.h>
#include <jsi/jsiImpl.h>
#include <jsi/jsiStdChannel.h>
#include <jsi/jsiPlexLink.h>
#include <jsi/jsiPlexUDP.h>
#include <sys/param.h>

#include <jsyscall.h>

#include "assert_pp.h"
#include "nameSpace.h"
#include "module.h"
#include "thread.h"
#include "scripting.h"
#include "registry.h"
#include "logResource.h"
#include "stdioResource.h"

/** scripting variables **/
char *jsiProviderName = 0;

/* The provider that will monitor the tcp port and call our handler */
static jsiProvider *jprovider;

/* Holds the list of files that can be used to send/recv jsiMessages */
static scriptFile *scriptingFiles;

static iLock *proxyListLock;
static proxyProvider *proxyProviders;
static sparseArray *proxies;

jsiNamedHandler *jvmHandlerList;

static struct {
	iLock *lock;
	int stalled;
} stallData;

/** scripting declarations **/

/*
 * Process any messages coming over the channel given in `arg'.
 */
static void janosvmMessageLooper(void *arg);
/*
 * Root looper function that waits for connections from the outside.
 */
static void janosvmLooper(void *arg);
static void remScriptProxy(proxyProvider *pp);
/*
 * The function that the jsiProvider will call when it receives a connection.
 * Currently just creates a daemon thread that runs janosvmMessageLooper to
 * handle messages.
 */
static int janosvmConnectionFunction(jsiProvider *jp, jsiChannel *jc);
static jsiConnectionHandler janosvmConnectionHandler = {
	janosvmConnectionFunction
};
static int janosvmProxyConnectionFunction(jsiProvider *jp, jsiChannel *jc);

/** Module declarations **/
static int initScriptingModule(int stage);

module_t scriptingModule = {
	"scripting",
	0,
	initScriptingModule,
	0,
	0
};

static jsiNamedHandler fileHandler;
static jsiHandlerFunctions fileHandlerFunctions;

/** JSI identity property */
jsiPropertyInfo identityRootProperties[] = {
	{ "identity", (1L << JSI_GET_PROPERTY), "",
	  "Get the identity of the running program\n" },
	{ 0 }
};

static
char *handleIdentityMessage(jsiHandlerContext *hc,
			    jsiHandler *jh,
			    jsiMessage *jm);
static
int getIdentitySuites(jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions identityFunctions;
static jsiNamedHandler identityHandler;

/** JSI continue property **/
jsiPropertyInfo continueRootProperties[] = {
	{ "continue", (1L << JSI_EXECUTE_PROPERTY), "",
	  "Unblock any threads blocked in the stallVM() function\n" },
	{ 0 }
};

static
char *handleContinueMessage(jsiHandlerContext *hc,
			    jsiHandler *jh,
			    jsiMessage *jm);
static
int getContinueSuites(jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions continueFunctions;
static jsiNamedHandler continueHandler;

/** JSI proxy property **/
jsiPropertyInfo proxyRootProperties[] = {
	{ "proxy", (1L << JSI_CREATE_PROPERTY), "",
	  "Create a proxy\n" },
	{ "proxy", (1L << JSI_DELETE_PROPERTY), "proxyID:int",
	  "delete a proxy\n" },
	{ "proxies", (1L << JSI_LIST_PROPERTIES), "",
	  "List the current set of proxies\n" },
	{ 0 }
};

static
char *handleProxyMessage(jsiHandlerContext *hc,
			 jsiHandler *jh,
			 jsiMessage *jm);
static
int getProxySuites(jsiHandler *jh, jsiMessage *jm);

static jsiHandlerFunctions proxyFunctions;
static jsiNamedHandler proxyHandler;

static jsiHandlerFunctions proxiesFunctions;
static jsiNamedHandler proxiesHandler;

/** JSI stdin property **/
jsiPropertyInfo stdinRootProperties[] = {
	{ "in", (1L << JSI_LIST_PROPERTIES), "",
	  "List the set of properties for the standard input.\n" },
	{ 0 }
};

static
int getStdinSuites(jsiHandler *jh, jsiMessage *jm);

static jsiFileHandler jsiStdinHandler;
static jsiHandlerFunctions jsiStdinFunctions;

/** JSI stdout property **/
jsiPropertyInfo stdoutRootProperties[] = {
	{ "out", (1L << JSI_LIST_PROPERTIES), "",
	  "List the set of properties for the standard output.\n" },
	{ 0 }
};

static
int getStdoutSuites(jsiHandler *jh, jsiMessage *jm);

static jsiFileHandler jsiStdoutHandler;
static jsiHandlerFunctions jsiStdoutFunctions;

/** JSI stderr property **/
jsiPropertyInfo stderrRootProperties[] = {
	{ "err", (1L << JSI_LIST_PROPERTIES), "",
	  "List the set of properties for the standard error.\n" },
	{ 0 }
};

static
int getStderrSuites(jsiHandler *jh, jsiMessage *jm);

static jsiFileHandler jsiStderrHandler;
static jsiHandlerFunctions jsiStderrFunctions;

/** Module functions **/
int initScriptingModule(int stage)
{
#if 0
	char *handoffChannel;
#endif
	proxyProvider *pp;
	jsiPlexLink *pl;
	scriptFile *sf;
	int retval = 1;

	/* Install our handlers */
	jsiInitNamedHandler(&fileHandler,
			    JNA_Name, "file",
			    JNA_Functions, &fileHandlerFunctions,
			    JNA_HandleMessage, jsiHandleFileMessage,
			    JNA_ResolveSpecifier, jsiResolveFileSpecifier,
			    JNA_GetSuites, jsiGetFileSuites,
			    JNA_TAG_DONE);
	jsiInstallHandler(&jvmHandlerList, &fileHandler);

	jsiInitNamedHandler(
		&identityHandler,
		JNA_Name, "identity",
		JNA_Functions, &identityFunctions,
		JNA_HandleMessage, handleIdentityMessage,
		JNA_GetSuites, getIdentitySuites,
		JNA_TAG_DONE);
	jsiInstallHandler(&jvmHandlerList, &identityHandler);
	
	jsiInitNamedHandler(
		&continueHandler,
		JNA_Name, "continue",
		JNA_Functions, &continueFunctions,
		JNA_HandleMessage, handleContinueMessage,
		JNA_GetSuites, getContinueSuites,
		JNA_TAG_DONE);
	jsiInstallHandler(&jvmHandlerList, &continueHandler);
	
	jsiInitNamedHandler(
		&proxyHandler,
		JNA_Name, "proxy",
		JNA_Functions, &proxyFunctions,
		JNA_HandleMessage, handleProxyMessage,
		JNA_TAG_DONE);
	jsiInstallHandler(&jvmHandlerList, &proxyHandler);
	
	jsiInitNamedHandler(
		&proxiesHandler,
		JNA_Name, "proxies",
		JNA_Functions, &proxiesFunctions,
		JNA_HandleMessage, handleProxyMessage,
		JNA_GetSuites, getProxySuites,
		JNA_TAG_DONE);
	jsiInstallHandler(&jvmHandlerList, &proxiesHandler);
	
	jsiInitNamedHandler(&jsiStdinHandler.handler,
			    JNA_Name, "in",
			    JNA_Functions, &jsiStdinFunctions,
			    JNA_HandleMessage, jsiHandleFileIOMessage,
			    JNA_ResolveSpecifier, jsiResolveFileIOSpecifier,
			    JNA_GetSuites, getStdinSuites,
			    JNA_TAG_DONE);
	jsiStdinHandler.fd = 0; // stdin
	jsiStdinHandler.flags = FHF_READ;
	jsiInstallHandler(&jvmHandlerList, &jsiStdinHandler.handler);
	jsiInitNamedHandler(&jsiStdoutHandler.handler,
			    JNA_Name, "out",
			    JNA_Functions, &jsiStdoutFunctions,
			    JNA_HandleMessage, jsiHandleFileIOMessage,
			    JNA_ResolveSpecifier, jsiResolveFileIOSpecifier,
			    JNA_GetSuites, getStdoutSuites,
			    JNA_TAG_DONE);
	jsiStdoutHandler.fd = 1; // stdout
	jsiStdoutHandler.flags = FHF_WRITE;
	jsiInstallHandler(&jvmHandlerList, &jsiStdoutHandler.handler);
	jsiInitNamedHandler(&jsiStderrHandler.handler,
			    JNA_Name, "err",
			    JNA_Functions, &jsiStderrFunctions,
			    JNA_HandleMessage, jsiHandleFileIOMessage,
			    JNA_ResolveSpecifier, jsiResolveFileIOSpecifier,
			    JNA_GetSuites, getStderrSuites,
			    JNA_TAG_DONE);
	jsiStderrHandler.fd = 2; // stderr
	jsiStderrHandler.flags = FHF_WRITE;
	jsiInstallHandler(&jvmHandlerList, &jsiStderrHandler.handler);
	
	/* XXX Attach any plex links */
	pl = jsiCreatePlexUDP();
	jsiAddPlexLink(pl);
	/* We were given a provider spec, attach to it */
	if( jsiProviderName )
	{
		if( (jprovider = jsiAttachProviderTo(
			jsiProviderName,
			0,
			&janosvmConnectionHandler)) )
		{
			errorInfo einfo;
			
			if( createDaemon(janosvmLooper,
					 "janosvmLooper",
					 jprovider,
					 5,
					 4096 + 2048,
					 &einfo) )
			{
				/*
				 * Register ourself if this instance has a name
				 */
				if( registryName )
				{
					jsiRegisterExport(registryName,
							  JSI_INTERFACE_EXPORT,
							  jsiProviderName);
				}
			}
			else
			{
				jsiDeleteProvider(jprovider);
				fprintf(stderr,
					"Can't create JSI looper thread.\n");
				retval = 0;
			}
		}
		else
		{
			fprintf(stderr,
				"Unable to attach to scripting provider %s.\n",
				jsiProviderName);
			retval = 0;
		}
	}
	/* Walk the list of scripting files and start up loopers for them. */
	sf = scriptingFiles;
	while( sf && retval )
	{
		jsiChannel *jc;

		if( (jc = jsiCreateStdChannel()) )
		{
			if( jsiConnectStdChannel(jc,
						 JCA_Protocol, JCP_FILE,
						 JCA_FileName, sf->filename,
						 JCA_Create, 1,
						 JCA_TAG_DONE) )
			{
				errorInfo einfo;
				
				if( !createDaemon(janosvmMessageLooper,
						  "janosvmMessageLooper",
						  jc,
						  5,
						  threadStackSize,
						  &einfo) )
				{
					jsiDeleteChannel(jc);
					fprintf(stderr,
						"Can't create JSI message "
						"handler thread.\n");
					retval = 0;
				}
			}
			else
			{
				jsiDeleteChannel(jc);
				retval = 0;
			}
		}
		else
		{
			fprintf(stderr, "Not enough memory\n");
			retval = 0;
		}
		sf = sf->next;
	}

	/* Startup the proxies */
	if( (proxies = createSparseArray(SAA_TAG_DONE)) )
	{
		pp = proxyProviders;
		while( pp && retval )
		{
			jsiProvider *jp;

			if( !pp->elem )
			{
				if( (pp->elem = getSparseElement(proxies)) )
					pp->elem->data.ptr = pp;
			}
			if( (jp = jsiAttachProviderTo(pp->provider,
						      0,
						      &pp->handler)) )
			{
				errorInfo einfo;
				
				if( createDaemon(janosvmLooper,
						 "janosvmLooper",
						 jp,
						 5,
						 4096 + 2048,
						 &einfo) )
				{
				}
				else
				{
					jsiDeleteProvider(jp);
					fprintf(stderr,
						"Can't create JSI proxy "
						"looper thread.\n");
					retval = 0;
				}
			}
			else
			{
				fprintf(stderr,
					"Unable to attach to scripting "
					"provider %s.\n",
					pp->provider);
				retval = 0;
			}
			pp = pp->next;
		}
	}
	
#if 0
	/* Pickup a channel that was handed off to us from netboot */

	/*
	 * XXX A work in progress...  The handoff works, the only problem
	 * is telling NetBoot that we responded to the message, which most
	 * likely means modifying some of netboot's persistent storage.
	 * Unfortunately, I haven't found a way to do that just yet...
	 */
	if( (handoffChannel = getenv("JSI_HANDOFF")) )
	{
		char *encodedChannel;
		jsiChannel *jc;

		sscanf(handoffChannel, "%p", &encodedChannel);
		fprintf(stderr, "%p\n", encodedChannel);
		fprintf(stderr, "%c\n", encodedChannel[0]);
		fprintf(stderr, "%s\n", encodedChannel);
		if( (jc = jsiReconnectChannel(encodedChannel)) )
		{
			jsiMessage *jm;

			if( (jm = jsiBuildMessage(JMA_What, JSI_REPLY,
						  JMA_String, "result", "ok",
						  JMA_TAG_DONE)) )
			{
				int bufferSize;
				char *buffer;

				if( (buffer = jsiFlattenMessage(
					jm, &bufferSize)) )
				{
					if( jsiReplyMessage(jc,
							    buffer,
							    bufferSize) )
					{
						encodedChannel[0] = '\0';
					}
					else
					{
						retval = 0;
						fprintf(stderr,
							"Couldn't send "
							"reply\n");
					}
				}
				else
				{
					retval = 0;
					fprintf(stderr,
						"Couldn't flatten message\n");
				}
				jsiDeleteMessage(jm);
			}
			else
			{
				retval = 0;
				fprintf(stderr,
					"Couldn't allocate reply message\n");
			}
			jsiDeleteChannel(jc);
		}
		else
		{
			retval = 0;
			fprintf(stderr,
				"Couldn't reconnect to handoff channel `%s'\n",
				handoffChannel);
		}
	}
#endif

	return( retval );
}

/** Scripting functions **/

void addScriptFile(scriptFile *sf)
{
	sf->next = scriptingFiles;
	scriptingFiles = sf;
}

int addScriptProxy(proxyProvider *pp)
{
	int retval = 0;
	int iLockRoot;
	
	require(pp != 0);
	
	lockStaticMutex(&proxyListLock);
	pp->handler.func = janosvmProxyConnectionFunction;
	pp->disabled = 0;
	pp->next = proxyProviders;
	proxyProviders = pp;
	if( proxies )
	{
		if( (pp->elem = getSparseElement(proxies)) )
		{
			pp->elem->data.ptr = pp;
			retval = pp->elem->id;
		}
	}
	unlockStaticMutex(&proxyListLock);
	return( retval );
}

void remScriptProxy(proxyProvider *pp)
{
	proxyProvider *curr, **prev;
	int iLockRoot;
	
	require(pp != 0);
	
	lockStaticMutex(&proxyListLock);
	curr = proxyProviders;
	prev = &proxyProviders;
	while( curr && (curr != pp) )
	{
		prev = &curr->next;
		curr = curr->next;
	}
	if( curr == pp )
	{
		*prev = curr->next;
		if( curr->elem )
			dropSparseElement(proxies, curr->elem);
		gc_free(pp);
	}
	unlockStaticMutex(&proxyListLock);
}

scriptThreadData *getScriptThreadData(void)
{
	return (scriptThreadData *)jthread_get_data(jthread_current())->
		userData;
}

void setStallState(int stalled)
{
	int iLockRoot;
	
	lockMutex(&stallData);
	stallData.stalled = stalled;
	if( !stalled )
		signalCond(&stallData);
	unlockMutex(&stallData);
}

void stallVM(void)
{
	int iLockRoot;
	
	lockMutex(&stallData);
	while( stallData.stalled )
	{
		waitCond(&stallData, 0);
	}
	unlockMutex(&stallData);
}

void janosvmMessageLooper(void *arg)
{
	jsiChannel *jc = (jsiChannel *)arg;
	jsiHandlerContext hc = { &jvmHandlerList, jc };
	scriptThreadData std = { jc };
	jsiMessage *jm;
	char *buffer;
	int bufferSize;

	require(arg != 0);

	jthread_get_data(jthread_current())->userData = &std;

	/* Just get the messages and dispatch them */
	while( (buffer = jsiRecvMessage(jc, &bufferSize)) )
	{
		if( (jm = jsiInstantiateMessage(buffer, bufferSize)) )
		{
			if( jm->what == JSI_SHUTDOWN )
			{
				jsiDeleteChannel(jc);
				/* XXX Gotta be a better way... */
				EXIT(0);
			}
			/*
			 * We clear the daemon bit so that any threads
			 * spawned by the message do not end up as daemons.
			 */
			unhand(getCurrentThread())->daemon = 0;
			jsiDispatchMessage(&hc, jm);
			unhand(getCurrentThread())->daemon = 1;
			if( (buffer = jsiFlattenMessage(jm, &bufferSize)) )
			{
				jsiReplyMessage(jc, buffer, bufferSize);
			}
			jsiDeleteMessage(jm);
		}
		else
		{
			/* Instantiate will free the buffer for us... */
			teamLogEntry("scripting",
				     LEA_String, "Error",
				     "Can't instantiate JSI message", -1,
				     LEA_TAG_DONE);
		}
	}
	/* We're done, destruct the channel and exit the thread */
	jsiDeleteChannel(jc);
	jthread_exit();
}

int janosvmConnectionFunction(jsiProvider *jp, jsiChannel *jc)
{
	Hjava_lang_Thread *tid;
	errorInfo einfo;

	require(jp != 0);
	require(jc != 0);

	if( !(tid = createDaemon(janosvmMessageLooper,
				 "janosvmMessageLooper",
				 jc,
				 5,
				 threadStackSize,
				 &einfo)) )
	{
		/* No thread :( */
		jsiDeleteChannel(jc);
		teamLogEntry("scripting",
			     LEA_String, "Error",
			     "Can't create JSI message handler thread", -1,
			     LEA_TAG_DONE);
	}
	/*
	 * XXX We should be able to let the looper thread die and detach the
	 * provider
	 */
	return( 0 );
}

struct proxyThreadArgs {
	proxyProvider *pp;
	jsiChannel *jc[2];
};

void janosvmProxyLooper(struct proxyThreadArgs *pta)
{
	int bufferSize, direction = 0, ok = 1;
	char *buffer;

	while( ok && (buffer = jsiRecvMessage(pta->jc[direction],
					      &bufferSize)) )
	{
		direction = !direction;
		ok = jsiSendMessage(pta->jc[direction], buffer, bufferSize);
		gc_free(buffer);
	}
}

void janosvmProxyConnect(void *arg)
{
	struct proxyThreadArgs *pta = (struct proxyThreadArgs *)arg;

	if( (pta->jc[1] = jsiConnectChannelTo(pta->pp->destination, 0)) )
	{
		janosvmProxyLooper(pta);
		jsiDeleteChannel(pta->jc[1]);
	}
	jsiDeleteChannel(pta->jc[0]);
	gc_free(pta);
}

int janosvmProxyConnectionFunction(jsiProvider *jp, jsiChannel *jc)
{
	proxyProvider *pp = (proxyProvider *)jp->handleConnection;
	struct proxyThreadArgs *pta;
	Hjava_lang_Thread *tid;
	errorInfo einfo;
	int retval = 0;

	require(jp != 0);
	require(jc != 0);

	if( pp->disabled )
	{
		jsiDeleteChannel(jc);
		jsiDeleteProvider(jp);
		remScriptProxy(pp);
		retval = 1;
	}
	else if( (pta = gc_malloc(sizeof(struct proxyThreadArgs),
				  GC_ALLOC_FIXED)) )
	{
		pta->pp = pp;
		pta->jc[0] = jc;
		if( !(tid = createDaemon(janosvmProxyConnect,
					 "janosvmProxyLooper",
					 pta,
					 5,
					 threadStackSize,
					 &einfo)) )
		{
			/* No thread :( */
			gc_free(pta);
			jsiDeleteChannel(jc);
			teamLogEntry("scripting",
				     LEA_String, "Error",
				     "Can't create JSI message handler thread",
				     -1,
				     LEA_TAG_DONE);
		}
	}
	return( retval );
}

void janosvmLooper(void *arg)
{
	jsiProvider *jp = (jsiProvider *)arg;
	
	/* Just let the provider do its work */
	jsiHandleProvider(jp);
	jthread_exit();
}

extern char* engine_name;	/* defined in the engine's library */
extern char* engine_version;	/* defined in the engine's library */

char *handleIdentityMessage(jsiHandlerContext *hc,
			    jsiHandler *jh,
			    jsiMessage *jm)
{
	char *retval = 0;

	switch( jm->what )
	{
	case JSI_GET_PROPERTY:
		jsiClearMessage(jm);
		jsiMessageAdd(jm,
			      JMA_What, JSI_REPLY,
			      JMA_String, "result", "Kaffe",
			      JMA_String, "result", engine_name,
			      JMA_String, "result", engine_version,
			      JMA_TAG_DONE);
		break;
	default:
		retval = "Unsupported identity operation";
		break;
	}
	return( retval );
}

int getIdentitySuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(
		jm,
		JMA_PropertyList, "result", identityRootProperties,
		JMA_TAG_DONE);
	return( retval );
}

char *handleContinueMessage(jsiHandlerContext *hc,
			    jsiHandler *jh,
			    jsiMessage *jm)
{
	char *retval = 0;

	switch( jm->what )
	{
	case JSI_EXECUTE_PROPERTY:
		jsiClearMessage(jm);
		jsiMessageAdd(jm,
			      JMA_What, JSI_REPLY,
			      JMA_String, "result", "ok",
			      JMA_TAG_DONE);
		setStallState(false);
		break;
	default:
		retval = "Unsupported continue operation";
		break;
	}
	return( retval );
}

int getContinueSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(
		jm,
		JMA_PropertyList, "result", continueRootProperties,
		JMA_TAG_DONE);
	return( retval );
}

char *handleProxyMessage(jsiHandlerContext *hc,
			 jsiHandler *jh,
			 jsiMessage *jm)
{
	char *provider, *destination;
	char *retval = 0;
	int iLockRoot;
	
	switch( jm->what )
	{
	case JSI_CREATE_PROPERTY:
		if( jsiMessageGet(jm,
				  JMA_String, "provider", 0, &provider,
				  JMA_String, "destination", 0, &destination,
				  JMA_TAG_DONE) )
		{
			proxyProvider *pp;

			if( (pp = gc_malloc(sizeof(proxyProvider) +
					    strlen(provider) + 1 +
					    strlen(destination) + 1,
					    GC_ALLOC_FIXED)) )
			{
				jsiProvider *jp;
				
				pp->provider = (char *)(pp + 1);
				strcpy(pp->provider, provider);
				pp->destination = ((char *)(pp + 1)) +
					strlen(provider) + 1;
				strcpy(pp->destination, destination);
				if( (jp = jsiAttachProviderTo(pp->provider,
							      0,
							      &pp->handler)) )
				{
					errorInfo einfo;
					sparseID_t id;
					
					id = addScriptProxy(pp);
					if( createDaemon(janosvmLooper,
							 "janosvmLooper",
							 jp,
							 5,
							 4096 + 2048,
							 &einfo) )
					{
						jsiClearMessage(jm);
						jsiMessageAdd(
							jm,
							JMA_What, JSI_REPLY,
							JMA_Integer, "result",
							id,
							JMA_TAG_DONE);
					}
					else
					{
						jsiDeleteProvider(jp);
						remScriptProxy(pp);
						retval = "Can't create daemon "
							"thread";
					}
				}
				else
				{
					retval = "Can't attach provider";
				}
			}
			else
			{
				retval = "Out of memory";
			}
		}
		else
		{
			retval = "No provider and/or destination given";
		}
		break;
	case JSI_DELETE_PROPERTY:
		{
			char *name, *data;
			sparseID_t id;
			
			jsiRestoreMessageSpecifier(jm);
			jsiPopMessageSpecifier(jm, &name, &data);
			if( sscanf(data, "%d", &id) == 1 )
			{
				sparseElement *se;
				
				lockStaticMutex(&proxyListLock);
				if( (se = findSparseElement(proxies, id)) )
				{
					proxyProvider *pp = (proxyProvider *)
						se->data.ptr;
					jsiChannel *jc;
					
					((proxyProvider *)se->data.ptr)->
						disabled = 1;
					if( (jc = jsiConnectChannelTo(
						pp->provider, 0)) )
					{
						jsiDeleteChannel(jc);
					}
					jsiClearMessage(jm);
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_String, "result", "ok",
						JMA_TAG_DONE);
				}
				else
				{
					retval = "Can't find proxy with "
						"given id";
				}
				unlockStaticMutex(&proxyListLock);
			}
			else
			{
				retval = "Invalid proxy id value";
			}
		}
		break;
	case JSI_LIST_PROPERTIES:
		{
			proxyProvider *pp;
			int error = 0;
			
			jsiClearMessage(jm);
			jm->what = JSI_REPLY;
			lockStaticMutex(&proxyListLock);
			pp = proxyProviders;
			while( pp )
			{
				if( pp->elem )
				{
					error = !jsiMessageAdd(
						jm,
						JMA_Integer, "id", pp->elem->id,
						JMA_String, "provider", pp->provider,
						JMA_String, "destination", pp->destination,
						JMA_TAG_DONE);
				}
				pp = pp->next;
			}
			unlockStaticMutex(&proxyListLock);
			if( error )
				retval = "Out of memory";
		}
		break;
	default:
		retval = "Unsupported proxy operation";
		break;
	}
	return( retval );
}

int getProxySuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", proxyRootProperties,
			       JMA_TAG_DONE);
	return( retval );
}

int getStdinSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(jm,
			       JMA_PropertyList, "result", stdinRootProperties,
			       JMA_TAG_DONE);
	return( retval );
}

int getStdoutSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(
		jm,
		JMA_PropertyList, "result", stdoutRootProperties,
		JMA_TAG_DONE);
	return( retval );
}

int getStderrSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(
		jm,
		JMA_PropertyList, "result", stderrRootProperties,
		JMA_TAG_DONE);
	return( retval );
}
