/*
 * sparseArray.c
 * A data structure for storing structures with unique identifiers.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <assert.h>
#include <stdarg.h>

#include "assert_pp.h"
#include "gc.h"
#include "module.h"
#include "mungWallProxy.h"
#include "sparseArray.h"

/** Module declarations **/
static int testSparseArrayModule(void);

module_t sparseArrayModule = {
	"sparse-array",
	0,
	0,
	0,
	testSparseArrayModule
};

/** Module functions **/

int testSparseArrayModule(void)
{
	sparseElement *se, **se_array;
	sparseArray *sa;
	int retval = 1;

	if( (sa = createSparseArray(SAA_TAG_DONE)) &&
	    (se_array = gc_malloc(sizeof(sparseElement *) *
				  sa->xSize * sa->ySize,
				  GC_ALLOC_FIXED)) )
	{
		int lpc, array_size;

		array_size = (sa->xSize * sa->ySize) - 1;
		se = getSparseElement(sa);
		assert(findSparseElement(sa, se->id) == se);
		dropSparseElement(sa, se);
		for( lpc = 0; lpc < array_size; lpc++ )
		{
			se_array[lpc] = getSparseElement(sa);
		}
		for( lpc = array_size - 1; lpc >= 0; lpc-- )
		{
			assert(findSparseElement(sa, se_array[lpc]->id) ==
			       se_array[lpc]);
		}
		for( lpc = array_size - 1; lpc >= 0; lpc-- )
		{
			dropSparseElement(sa, se_array[lpc]);
		}
		for( lpc = 1; lpc < 16384; lpc++ )
		{
			assert(findSparseElement(sa, lpc) == 0);
		}
		gc_free(se_array);
	}
	deleteSparseArray(sa);
	if( (sa = createSparseArray(SAA_Flags, SAF_NO_DELAY,
				    SAA_TAG_DONE)) &&
	    (se_array = gc_malloc(sizeof(sparseElement *) *
				  sa->xSize * sa->ySize,
				  GC_ALLOC_FIXED)) )
	{
		int lpc, array_size;
		
		array_size = (sa->xSize * sa->ySize) - 1;
		for( lpc = 1; lpc < array_size; lpc++ )
		{
			se_array[lpc] = getSpecificSparseElement(sa, lpc);
			assert(se_array[lpc] != 0);
			assert(se_array[lpc]->id == lpc);
		}
		gc_free(se_array);
	}
	deleteSparseArray(sa);
	return( retval );
}

/** sparseArray functions **/

/* Default sizes for the X and Y dimensions. */
#define DEFAULT_X_SIZE 64
#define DEFAULT_Y_SIZE 32

/*
 * The bit field layout for a unique ID is the following:
 *
 *  |            | sa->xBitCount | sa->yBitCount |
 *  | sa->timeID |  x dimension  |  y dimension  |
 */

/* XXX Ick */
static int bitCount(int bits)
{
	int lpc, retval = 0;

	for( lpc = 0; lpc < (sizeof(int) * 8); lpc++ )
	{
		if( (1L << lpc) & bits )
			retval++;
	}
	return( retval );
}

sparseArray *createSparseArray(int tag, ...)
{
	int xSize = DEFAULT_X_SIZE, ySize = DEFAULT_Y_SIZE;
	unsigned long flags = 0;
	sparseArray *retval;
	va_list args;

	va_start(args, tag);
	while( tag != SAA_TAG_DONE )
	{
		switch( tag )
		{
		case SAA_XSize:
			xSize = va_arg(args, int);

			/* Must be a power of 2 */
			ensure(bitCount(xSize) == 1);
			break;
		case SAA_YSize:
			ySize = va_arg(args, int);

			/* Must be a power of 2 */
			ensure(bitCount(ySize) == 1);
			break;
		case SAA_Flags:
			flags = va_arg(args, unsigned long);
			break;
		}
		tag = va_arg(args, int);
	}
	va_end(args);
	if( (retval = gc_malloc(sizeof(sparseArray) +
				(sizeof(sparseDimension *) * (xSize - 1)),
				GC_ALLOC_FIXED)) )
	{
		sparseElement *sentinel;

		retval->flags = flags;
		retval->xSize = xSize;
		retval->xBitCount = bitCount(xSize - 1);
		retval->xUsed = 0;
		retval->ySize = ySize;
		retval->yBitCount = bitCount(ySize - 1);
		if( (sentinel = getSparseElement(retval)) )
		{
			/*
			 * Get the zeroth element so that we can use it as a
			 * sentinel to indicate an unused element.  We force
			 * this to one so that the ID's won't ever match if
			 * someone tries to lookup zero in the array.
			 */
			sentinel->id = 1;
			retval->xDimension[0]->inUse--;

			ensure_valid_memory(retval);
		}
		else
		{
			deleteSparseArray(retval);
		}
	}
	return( retval );
}

void deleteSparseArray(sparseArray *sa)
{
	if( sa )
	{
		int lpc;

		for( lpc = 0; lpc < sa->xUsed; lpc++ )
		{
			gc_free(sa->xDimension[lpc]);
		}
		gc_free(sa);
	}
}

/*
 * Add a Y dimension to the array.
 */
static
sparseDimension *addSparseDimension(sparseArray *sa)
{
	sparseDimension *retval = 0;

	require(sa != 0);
	
	if( ((sa->xUsed + 1) <= sa->xSize) &&
	    (retval = gc_malloc(sizeof(sparseDimension) +
				(sizeof(sparseElement) * (sa->ySize - 1)),
				GC_ALLOC_FIXED)) )
	{
		sparseElement *se;
		int lpc;
		
		sa->xDimension[sa->xUsed] = retval;
		sa->xUsed++;
		/* Put all the elements on the free list */
		se = retval->elements;
		retval->freeElements = se;
		/* Start the delay number at zero */
		retval->timeID = 0;
		for( lpc = 0; lpc < (sa->ySize - 1); lpc++, se++ )
		{
			se->data.next = (se + 1);
		}
		retval->lastElementPointer = &se->data.next;

		ensure_valid_memory(retval);
	}
	return( retval );
}

sparseElement *getSparseElement(sparseArray *sa)
{
	sparseID_t x_index = 0, y_index;
	sparseElement *retval = 0;
	sparseDimension *sd = 0;
	int lpc;

	require(sa != 0);

	/* First, try to find a dimension with free elements */
	for( lpc = (sa->xUsed - 1); !sd && (lpc >= 0); lpc-- )
	{
		if( sa->xDimension[lpc]->freeElements )
		{
			sd = sa->xDimension[lpc];
			x_index = lpc;
		}
	}
	if( !sd )
	{
		/* No free dimensions, allocate a new one */
		sd = addSparseDimension(sa);
		x_index = sa->xUsed - 1;
	}
	if( sd )
	{
		sd->inUse++;
		/* Pull the element off the free list */
		retval = sd->freeElements;
		sd->freeElements = sd->freeElements->data.next;
		/*
		 * If there are no more free elements set the
		 * lastElementPointer to the freeElements pointer
		 */
		if( !sd->freeElements )
			sd->lastElementPointer = &sd->freeElements;
		/* Compute the elements id */
		y_index = retval - sd->elements;
		/*
		 * Use the timeID to check to see if we've cycled through all
		 * of the elements and need to update the dimension's saved
		 * delay number to something else.
		 */
		if( retval->id != sd->timeID )
			sd->timeID = retval->id;
		/* Construct the sparseID */
		retval->id =
			sd->timeID |
			(x_index << sa->yBitCount) |
			y_index;
		
		ensure(retval >= sd->elements);
		ensure(retval < (sd->elements + sa->ySize));
	}
	return( retval );
}

sparseElement *getSpecificSparseElement(sparseArray *sa, sparseID_t id)
{
	sparseElement *se, *retval = 0;
	sparseID_t x_index, y_index;
	sparseDimension *sd;
	int error = 0;
	
	require(sa != 0);
	require(sa->flags & SAF_NO_DELAY);

	/* Compute the index */
	x_index = (id >> sa->yBitCount) & (sa->xSize - 1);
	y_index = id & (sa->ySize - 1);
	/* Make sure theres a dimension for it */
	while( !error && !(sd = sa->xDimension[x_index]) )
	{
		if( !addSparseDimension(sa) )
			error = 1;
	}
	if( sd )
	{
		se = sd->elements + y_index;
		if( se->id != id )
		{
			sparseElement **prev, *curr;

			prev = &sd->freeElements;
			curr = sd->freeElements;
			while( curr != se )
			{
				prev = &curr->data.next;
				curr = curr->data.next;
			}
			assert(curr == se);
			*prev = curr->data.next;
			if( !sd->freeElements )
				sd->lastElementPointer = &sd->freeElements;
			sd->inUse++;
			se->id = id;
			retval = se;
		}
		else
		{
			/* Already allocated */
		}
	}
	return( retval );
}

void dropSparseElement(sparseArray *sa, sparseElement *se)
{
	sparseDimension *sd;
	sparseID_t x_index;

	require(sa != 0);
	require(se != 0);
	require(se->id != 0);

	x_index = (se->id >> sa->yBitCount) & (sa->xSize - 1);
	sd = sa->xDimension[x_index];
	{
		require(se >= sd->elements);
		require(se < (sd->elements + sa->ySize));
		
		sd->inUse--;
		se->data.next = 0;
		/* Add the element to the end of the list */
		*sd->lastElementPointer = se;
		/*
		 * We're the last element on the free list, adjust
		 * lastElementPointer accordingly.
		 */
		sd->lastElementPointer = &se->data.next;
		if( !(sa->flags & SAF_NO_DELAY) )
		{
			/*
			 * Mark this element as unused by setting it to a later
			 * delay number.  Since the x/y bits will be zero it
			 * will indicate the element as unused.  And the
			 * increased delay number can be used to tell when
			 * we've cycled through all the elements in the free
			 * list and need to use a different delay number to
			 * distinguish newer elements from stale ones.
			 */
			 se->id = sd->timeID + (1L << (sa->xBitCount +
						       sa->yBitCount));
		}
	}
}

sparseElement *findSparseElement(sparseArray *sa, sparseID_t id)
{
	sparseElement *se, *retval = 0;
	sparseID_t x_index, y_index;
	sparseDimension *sd;

	require(sa != 0);
	
	/* Compute the index */
	x_index = (id >> sa->yBitCount) & (sa->xSize - 1);
	y_index = id & (sa->ySize - 1);
	/* Make sure theres a dimension for it */
	if( (sd = sa->xDimension[x_index]) )
	{
		se = sd->elements + y_index;
		/* Make sure the id's match */
		if( se->id == id )
			retval = se;
	}

	ensure(!retval || (retval->id >= 0));
	
	return( retval );
}
