/*
 * sparseArray.h
 * A data structure for storing structures with unique identifiers.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _sparse_array_h
#define _sparse_array_h

/*
 * The sparse array module implements a data structure that can quickly
 * create/delete/lookup elements in the array based on a unique identifier.
 * The structure is arranged as a two dimensional array with the X axis
 * possibly containing pointers to structures that hold the elements on the
 * Y axis.  An element's X axis and Y axis indexes are then used to create
 * unique identifier along with some high order bits to delay the time it
 * takes for an identifier to repeat.
 */

/* The sparseID_t is the type of the unique identifier, just an integer */
typedef unsigned int sparseID_t;

/*
 * A sparseElement is used to store/refer to information about some other data
 * structure.
 *
 * `id' - The unique id for this element.  If its zero than it isn't allocated
 *   to anyone yet.
 * `data.nextFree' - If this element is allocated then this points to the
 *   next free element in the list.
 * `data.c/i/f/ptr' - Storage for data.
 */
typedef struct _sparseElement {
	sparseID_t id;
	union {
		char c;
		int i;
		float f;
		void *ptr;
		struct _sparseElement *next;
	} data;
} sparseElement;

/*
 * The sparseDimension structure represents the second dimension in the array
 * and manages/contains the actual elements.
 *
 * `freeElements' - The list of free elements in this dimension.
 * `lastElementPointer' - Pointer to the `nextFree' field in the last free
 *   element in this dimension, or, the freeElements pointer.
 * `timeID' - The current delay number for this dimension.
 */
typedef struct _sparseDimension {
	sparseElement *freeElements;
	sparseElement **lastElementPointer;
	int inUse;
	sparseID_t timeID;
	/* Elements immediately follow */
	sparseElement elements[1];
} sparseDimension;

enum {
	SAB_NO_DELAY,
};

enum {
	SAF_NO_DELAY = (1L << SAB_NO_DELAY),
};

/*
 * The sparseArray structure is the root for the other structures.  All
 * references pass through this structure.
 *
 * `flags' - Holds any bit flags.
 * `xSize' - The size of the X dimension.
 * `xBitCount' - The number of bits used to encode an index in the X dimension
 * `xUsed' - The number of elements used in the `xDimension' array.
 * `ySize' - The number of elements in the Y dimension.
 * `yBitCount' - The number of bits used to encode an index in the Y dimension
 * `xDimension' - The X dimension for this array, contains pointers to all of
 *   the allocated Y dimensions.
 */
typedef struct _sparseArray {
	unsigned long flags;
	int xSize;
	int xBitCount;
	int xUsed;
	int ySize;
	int yBitCount;
	sparseDimension *xDimension[1];
} sparseArray;

enum {
	SAA_TAG_DONE,
	SAA_XSize,
	SAA_YSize,
	SAA_Flags,
};

/*
 * Create an empty sparse array.
 *
 * `tag, ...' - The tag list that should be used to initialize the object.
 * returns - A new sparseArray object thats been initialized with the given
 *  values or NULL if there wasn't enough memory.
 */
sparseArray *createSparseArray(int tag, ...);
/*
 * Delete a sparse array and its elements.
 *
 * `sa' - A sparseArray as returned by createSparseArray.
 */
void deleteSparseArray(sparseArray *sa);
/*
 * Get an element from the sparse array.
 *
 * `sa' - A sparseArray as returned by createSparseArray.
 * returns - A sparseElement from the array with a new unique identifier in
 *  id or NULL if there aren't anymore elements available.
 */
sparseElement *getSparseElement(sparseArray *sa);
/*
 * Get a specific element from the sparse array.
 *
 * `sa' - A sparseArray as returned by createSparseArray.
 * `id' - The identifier of the element to retrieve.
 * returns - A sparseElement from the array with a new unique identifier in
 *  id or NULL if there aren't anymore elements available.
 */
sparseElement *getSpecificSparseElement(sparseArray *sa, sparseID_t id);
/*
 * Return an element to the sparse array.
 *
 * `sa' - A sparseArray as returned by createSparseArray.
 * `se' - A sparseElement that was allocated from `sa'.
 */
void dropSparseElement(sparseArray *sa, sparseElement *se);
/*
 * Try to find an element in the sparse array with the given id.
 *
 * `sa' - A sparseArray as returned by createSparseArray.
 * `id' - The unique identifier to search for in `sa'.
 * returns - The allocated sparseElement object or NULL if there is no element
 *  with that identifier.
 */
sparseElement *findSparseElement(sparseArray *sa, sparseID_t id);

/*
 * AddModule: sparseArrayModule
 *   Comment: Just used to test the module
 */

#endif
