/*
 * stdioResource.c
 * Resource to managed standard I/O for each team.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "assert_pp.h"

#include "stdioResource.h"
#include "memResource.h"
#include "jteam.h"
#include "nativeClass.h"
#include "nativeClass-inline.h"
#include "mungWallProxy.h"
#include "scripting.h"
#include "module.h"
#include "moduleList.h"

#include <misc/stringParsing.h>

#include "edu_utah_janosvm_resources_StdioSpec.h"

#define JAVA_STDIO_SPEC "edu/utah/janosvm/resources/StdioSpec"

/** Module declarations **/
static int initStdioResourceModule(int stage);

module_t stdioResourceModule = {
	"stdio",
	0,
	initStdioResourceModule,
	0,
	0
};

/** StdioResource declarations **/
static int stdioFillSpec(struct _resource *res,
			 struct _resourceSpec *rs,
			 sectionFile *sf,
			 tagSectionData *tsd);
static resourceSpec *stdioCreateJSISpec(struct _resource *res, jsiMessage *jm);
static void stdioDeleteJSISpec(struct _resource *res,
			       struct _resourceSpec *rs);
static int stdioSubscribe(struct _resource *res,
			  struct _resourceClient *rc,
			  struct _resourceSpec *rs);
static void stdioUnsubscribe(struct _resource *res,
			     struct _resourceClient *rc);

static resourceFunctions stdioResourceFunctions = {
	0,
	0,
	0,
	stdioFillSpec,
	stdioCreateJSISpec,
	stdioDeleteJSISpec,
	0,
	stdioSubscribe,
	0,
	stdioUnsubscribe
};

stdioResource metalStdioResource;

static stdioSpec defaultSpec = {
	{ &metalStdioResource.link },
	"builtin://stdin",
	"builtin://stdout",
	"builtin://stderr"
};

/** Module Definitions **/
int initStdioResourceModule(int stage)
{
	int retval = 0;

	switch( stage )
	{
	case SID_STDIORESOURCEMODULE_INITMETAL:
		installResource(&metalStdioResource.link,
				RA_Name, "stdio",
				RA_Flags, RF_INTERNAL,
				RA_Functions, &stdioResourceFunctions,
				RA_ClientSize, sizeof(stdioClient),
				RA_SpecSize, sizeof(stdioSpec),
				RA_TAG_DONE);
		retval = attachTeamResource(metalTeam, &defaultSpec.spec);
		break;
	case SID_STDIORESOURCEMODULE:
		installResource(&metalStdioResource.link,
				RA_SpecNativeClass, JAVA_STDIO_SPEC,
				RA_TAG_DONE);
		retval = 1;
		break;
	}
	return( retval );
}

int stdioFillSpec(struct _resource *res,
		  struct _resourceSpec *rs,
		  sectionFile *sf,
		  tagSectionData *tsd)
{
	return( 1 );
}

resourceSpec *stdioCreateJSISpec(struct _resource *res, jsiMessage *jm)
{
	resourceSpec *retval = 0;

	if( (retval = createResourceSpec(res)) )
	{
		stdioSpec *ss = (stdioSpec *)retval;

		/*
		 * Copy the default values in case they aren't specified in
		 * the message.
		 */
		*ss = defaultSpec;
		/*
		 * The message lives longer than the spec so we don't have to
		 * copy them out.
		 */
		jsiMessageGet(jm,
			      JMA_String, "stdin", 0, &ss->inURI,
			      JMA_TAG_DONE);
		jsiMessageGet(jm,
			      JMA_String, "stdout", 0, &ss->outURI,
			      JMA_TAG_DONE);
		jsiMessageGet(jm,
			      JMA_String, "stderr", 0, &ss->errURI,
			      JMA_TAG_DONE);
	}
	return( retval );
}

void stdioDeleteJSISpec(struct _resource *res, struct _resourceSpec *rs)
{
	deleteResourceSpec(rs);
}

/*
 * The ioStreamOptions structure is a temporary structure used to encode
 * various options for a stream.
 */
struct ioStreamOptions {
	parsedString option;
	parsedString optionValue;
	
	enum {
		IOSOB_APPEND,
		IOSOB_ADD_TID,
	};
	
	enum {
		IOSOF_APPEND = (1L << IOSOB_APPEND),
		IOSOF_ADD_TID = (1L << IOSOB_ADD_TID),
	};
	unsigned long flags;
};

/* ioStream functions */

/* Parse the options for the stream from the options struct */
static
int setStreamOptions(struct ioStreamOptions *so);
/* Fill in an ioStream object that conforms to the given URI */
static
int openIOStream(ioStream *dest, char *spec, int outputStream);
/* Share the src stream with the destination stream */
static
int reopenIOStream(ioStream *dest, ioStream *src);
/* Close the io stream and any of its resources */
static
void closeIOStream(ioStream *ios);

/* Helper function that parses any options given in the ioStream URI */
static
int setStreamOptions(struct ioStreamOptions *so)
{
	int retval = 0;

	if( !cmpPStrStr(&so->option, "append") )
	{
		if( !cmpPStrStr(&so->optionValue, "true") ||
		    !cmpPStrStr(&so->optionValue, "1") )
		{
			so->flags |= IOSOF_APPEND;
		}
		retval = 1;
	}
	else if( !cmpPStrStr(&so->option, "add-tid") )
	{
		if( !cmpPStrStr(&so->optionValue, "true") ||
		    !cmpPStrStr(&so->optionValue, "1") )
		{
			so->flags |= IOSOF_ADD_TID;
		}
		retval = 1;
	}
	return( retval );
}

/* Fill in an ioStream object that conforms to the given URI */
static
int openIOStream(ioStream *dest, char *spec, int outputStream)
{
	parsedString protocol, path;
	struct ioStreamOptions so;
	int retval = 0;

	so.flags = 0;
	if( parseString(spec,
			SPO_Cond, ":",
			  SPO_Do,
			    SPO_String, &so.option,
			    SPO_Expect, "=",
			    SPO_String, &so.optionValue,
			    SPO_While, ",", ":",
			    SPO_Handle, setStreamOptions, &so,
			    SPO_End,
			  SPO_End,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, "://",
			SPO_String, &path,
			SPO_End) )
	{
		if( !cmpPStrStr(&protocol, "builtin") )
		{
			/* Builtin streams */
			dest->kind = SK_SHARED_FILE;
			if( !strcmp(path.data, "none") )
			{
				dest->kind = SK_NONE;
				retval = 1;
			}
			else if( !strcmp(path.data, "stdin") )
			{
				if( !outputStream )
				{
					dest->handle.file = stdin;
					retval = 1;
				}
			}
			else if( !strcmp(path.data, "stdout") )
			{
				if( outputStream )
				{
					dest->handle.file = stdout;
					retval = 1;
				}
			}
			else if( !strcmp(path.data, "stderr") )
			{
				if( outputStream )
				{
					dest->handle.file = stderr;
					retval = 1;
				}
			}
			else
			{
				fprintf(stderr,
					"Unknown builtin %s.\n",
					path.data);
			}
		}
		else if( !cmpPStrStr(&protocol, "jsi") )
		{
			/* JSI stream */
			dest->kind = SK_JSI;
			dest->handle.jsi = 0;
			retval = 1;
		}
		else if( !cmpPStrStr(&protocol, "file") )
		{
			char *myPath = path.data, *mode;

			/* Regular file stream */
			if( outputStream )
			{
				if( so.flags & IOSOF_APPEND )
					mode = "a";
				else
					mode = "w";
			}
			else
			{
				mode = "r";
			}
			if( so.flags & IOSOF_ADD_TID )
			{
				if( (myPath = gc_malloc(path.len +
							32 +
							1,
							GC_ALLOC_FIXED)) )
				{
					sprintf(myPath,
						"%s.%u",
						path.data,
						currentTeam()->elemID);

					ensure_valid_memory(myPath);
				}
			}
			dest->kind = SK_FILE;
			if( myPath &&
			    (dest->handle.file = fopen(myPath, mode)) )
			{
				retval = 1;
			}
			if( myPath && (myPath != path.data) )
				gc_free(myPath);
		}
		else if( !cmpPStrStr(&protocol, "fd") )
		{
			char *mode = outputStream ? "w" : "r";
			int newfd, fd;

			/* FD file stream */
			dest->kind = SK_FILE;
			sscanf(path.data, "%d", &fd);
			if( ((newfd = dup(fd)) >= 0) &&
			    (dest->handle.file = fdopen(newfd, mode)) )
			{
				retval = 1;
			}
			else if( newfd >= 0 )
			{
				close(newfd);
			}
		}
	}
	else
	{
		fprintf(stderr, "Couldn't parse %s\n", spec);
	}
	if( retval )
	{
		/* Copy the URI */
		if( (dest->uri = gc_malloc(strlen(spec) + 1, GC_ALLOC_FIXED)) )
		{
			strcpy(dest->uri, spec);
		}
		else
		{
			closeIOStream(dest);
			retval = 0;
		}
	}
	return( retval );
}

static
int reopenIOStream(ioStream *dest, ioStream *src)
{
	require(dest != 0);
	require(src != 0);
	
	dest->kind = SK_SHARED_FILE;
	dest->handle.file = src->handle.file;
	if( (dest->uri = gc_malloc(strlen(src->uri) + 1, GC_ALLOC_FIXED)) )
	{
		strcpy(dest->uri, src->uri);
	}
	return( 1 );
}

static
void closeIOStream(ioStream *ios)
{
	gc_free(ios->uri);
	switch( ios->kind )
	{
	case SK_SHARED_FILE:
		break;
	case SK_FILE:
		if( ios->handle.file )
			fclose(ios->handle.file);
		break;
	case SK_JSI:
		break;
	}
}

int stdioSubscribe(struct _resource *res,
		   struct _resourceClient *rc,
		   struct _resourceSpec *rs)
{
	stdioClient *sc = (stdioClient *)rc;
	stdioSpec *ss = (stdioSpec *)rs;
	int retval = 0;

	if( !ss )
		ss = &defaultSpec;
	/* Try to open the requested streams */
	if( openIOStream(&sc->streams[STDIO_IN], ss->inURI, 0) &&
	    openIOStream(&sc->streams[STDIO_OUT], ss->outURI, 1) &&
	    /* Reopen if the err stream is the same as the out */
	    (((sc->streams[STDIO_OUT].kind == SK_FILE) &&
	      !strcmp(ss->outURI, ss->errURI)) ?
	     reopenIOStream(&sc->streams[STDIO_ERR], &sc->streams[STDIO_OUT]) :
	     openIOStream(&sc->streams[STDIO_ERR], ss->errURI, 1)) )
	{
		retval = 1;
	}
	else
	{
		closeIOStream(&sc->streams[STDIO_OUT]);
		closeIOStream(&sc->streams[STDIO_ERR]);
	}
	
	return( retval );
}

void stdioUnsubscribe(struct _resource *res,
		      struct _resourceClient *rc)
{
	stdioClient *sc = (stdioClient *)rc;

	closeIOStream(&sc->streams[STDIO_IN]);
	closeIOStream(&sc->streams[STDIO_OUT]);
	closeIOStream(&sc->streams[STDIO_ERR]);
}

void stdioAttachJSIChannel(jsiChannel *jc)
{
	stdioClient *sc = (stdioClient *)
		getTeamInternalResource(0, &metalStdioResource.link);
	ioStream *ios = sc->streams;
	int lpc;

	for( lpc = 0; lpc < STDIO_STREAM_MAX; lpc++ )
	{
		if( ios[lpc].kind == SK_JSI )
			ios[lpc].handle.jsi = jc;
	}
}

void stdioDetachJSIChannel(void)
{
	stdioClient *sc = (stdioClient *)
		getTeamInternalResource(0, &metalStdioResource.link);
	ioStream *ios = sc->streams;
	int lpc;

	for( lpc = 0; lpc < STDIO_STREAM_MAX; lpc++ )
	{
		if( ios[lpc].kind == SK_JSI )
			ios[lpc].handle.jsi = 0;
	}
}

void stdioPrint(int stream, char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	stdioVPrint(stream, fmt, args);
	va_end(args);
}

/* Send a jsiMessage to print out the arguments */
static
void jsiPrint(int stream, ioStream *ios, char *fmt, va_list args)
{
	jsiMessage *jm;

	if( (jm = jsiBuildMessage(JMA_What, JSI_EXECUTE_PROPERTY,
				  JMA_Specifier, JMA_String, "write", "",
				  JMA_Specifier, JMA_String, "out", "",
				  JMA_FormattedString, "data", fmt, &args,
				  JMA_TAG_DONE)) )
	{
		jsiHandlerContext hc = { &jvmHandlerList, ios->handle.jsi };
		
		/* Send the message and wait for the reply */
		if( (jm = jsiDoMessage(&hc, jm)) )
		{
			jsiDeleteMessage(jm);
		}
	}
}

void stdioVPrint(int stream, char *fmt, va_list args)
{
	stdioClient *sc = (stdioClient *)
		getTeamInternalResource(0, &metalStdioResource.link);
	ioStream *ios = &sc->streams[stream];

	switch( ios->kind )
	{
	case SK_SHARED_FILE:
	case SK_FILE:
		vfprintf(ios->handle.file, fmt, args);
		break;
	case SK_JSI:
		if( ios->handle.jsi )
			jsiPrint(stream, ios, fmt, args);
		break;
	default:
		break;
	}
}

static
int jsiRead(ioStream *ios, char *buffer, int size)
{
	jsiHandlerContext hc = { &jvmHandlerList, ios->handle.jsi };
	int retval = -1;
	jsiMessage *jm;

	/* Build and send a jsiMessage to read the given number of bytes */
	if( (jm = jsiDoMessage(
		&hc,
		jsiBuildMessage(JMA_What, JSI_EXECUTE_PROPERTY,
				JMA_Specifier, JMA_String, "read", "",
				JMA_Specifier, JMA_String, "in", "",
				JMA_Integer, "size", size,
				JMA_TAG_DONE))) )
	{
		if( jm->what == JSI_REPLY )
		{
			int actual;
			char *data;

			/* Get the data and actual size */
			if( jsiMessageGet(jm,
					  JMA_Bytes, "data", 0, &data, &size,
					  JMA_Integer, "actual", 0, &actual,
					  JMA_TAG_DONE) )
			{
				if( actual <= size )
				{
					memcpy(buffer, data, actual);
					retval = actual;
				}
			}
		}
		jsiDeleteMessage(jm);
	}
	return( retval );
}

int stdioRead(char *buffer, int size)
{
	stdioClient *sc = (stdioClient *)
		getTeamInternalResource(0, &metalStdioResource.link);
	ioStream *ios = &sc->streams[STDIO_IN];
	int retval = -1;
	
	switch( ios->kind )
	{
	case SK_SHARED_FILE:
	case SK_FILE:
		/*
		 * Use fgets so that it stops when it hits a line feed,
		 * otherwise stdin just gets queued up for a long time.
		 */
		if( fgets(buffer, size, ios->handle.file) )
			retval = strlen(buffer);
		else 
			retval = -1;
		break;
	case SK_JSI:
		if( ios->handle.jsi )
			retval = jsiRead(ios, buffer, size);
		break;
	default:
		break;
	}
	return( retval );
}

static
void jsiWrite(int stream, ioStream *ios, char *buffer, int size)
{
	jsiHandlerContext hc = { &jvmHandlerList, ios->handle.jsi };
	jsiMessage *jm;

	/* Build and send a jsiMessage to write out the buffer */
	if( (jm = jsiDoMessage(
		&hc,
		jsiBuildMessage(JMA_What, JSI_EXECUTE_PROPERTY,
				JMA_Specifier, JMA_String, "write", "",
				JMA_Specifier, JMA_String, "out", "",
				JMA_Bytes, "data", buffer, size,
				JMA_TAG_DONE))) )
	{
		jsiDeleteMessage(jm);
	}
}

void stdioWrite(int stream, char *buffer, int size)
{
	stdioClient *sc = (stdioClient *)
		getTeamInternalResource(0, &metalStdioResource.link);
	ioStream *ios = &sc->streams[stream];

	switch( ios->kind )
	{
	case SK_SHARED_FILE:
	case SK_FILE:
		fwrite(buffer, 1, size, ios->handle.file);
		break;
	case SK_JSI:
		if( ios->handle.jsi )
			jsiWrite(stream, ios, buffer, size);
		break;
	default:
		break;
	}
}

void stdioFlush(int stream)
{
	stdioClient *sc = (stdioClient *)
		getTeamInternalResource(0, &metalStdioResource.link);
	ioStream *ios = &sc->streams[stream];

	switch( ios->kind )
	{
	case SK_SHARED_FILE:
	case SK_FILE:
		fflush(ios->handle.file);
		break;
	default:
		break;
	}
}

void edu_utah_janosvm_resources_StdioSpec_nativeConstructor(
	struct Hedu_utah_janosvm_resources_StdioSpec *jss,
	struct Hjava_lang_String *jInURI,
	struct Hjava_lang_String *jOutURI,
	struct Hjava_lang_String *jErrURI)
{
	stdioSpec *ss = OBJECT2NATIVE(&metalStdioResource.link.specNativeClass,
				      (Hjava_lang_Object *)jss);
	errorInfo einfo;
	int error = 0;

	if( !jInURI || !jOutURI || !jErrURI )
	{
		postException(&einfo, JAVA_LANG(NullPointerException));
		throwError(&einfo);
	}
	jss->found = 0;
	ss->spec.handler = &metalStdioResource.link;
	if( !(ss->inURI = stringJava2C(jInURI)) )
		error = 1;
	if( !(ss->outURI = stringJava2C(jOutURI)) )
		error = 1;
	if( !(ss->errURI = stringJava2C(jErrURI)) )
		error = 1;
	if( error )
	{
		/* The finalizer should free them... right? */
		postOutOfMemory(&einfo);
		throwError(&einfo);
	}
}

void edu_utah_janosvm_resources_StdioSpec_nativeFinalize(
	struct Hedu_utah_janosvm_resources_StdioSpec *jss)
{
	stdioSpec *ss = OBJECT2NATIVE(&metalStdioResource.link.specNativeClass,
				      (Hjava_lang_Object *)jss);

	if( !jss->found )
	{
		gc_free(ss->inURI);
		gc_free(ss->outURI);
		gc_free(ss->errURI);
	}
}

struct Hjava_lang_String *
edu_utah_janosvm_resources_StdioSpec_getInput(
	struct Hedu_utah_janosvm_resources_StdioSpec *jss)
{
	stdioSpec *ss = OBJECT2NATIVE(&metalStdioResource.link.specNativeClass,
				      (Hjava_lang_Object *)jss);
	struct Hjava_lang_String *retval;

	if( !(retval = stringC2Java(ss->inURI)) )
	{
		errorInfo einfo;
		
		postOutOfMemory(&einfo);
		throwError(&einfo);
	}
	return( retval );
}

struct Hjava_lang_String *
edu_utah_janosvm_resources_StdioSpec_getOutput(
	struct Hedu_utah_janosvm_resources_StdioSpec *jss)
{
	stdioSpec *ss = OBJECT2NATIVE(&metalStdioResource.link.specNativeClass,
				      (Hjava_lang_Object *)jss);
	struct Hjava_lang_String *retval;

	if( !(retval = stringC2Java(ss->outURI)) )
	{
		errorInfo einfo;
		
		postOutOfMemory(&einfo);
		throwError(&einfo);
	}
	return( retval );
}

struct Hjava_lang_String *
edu_utah_janosvm_resources_StdioSpec_getError(
	struct Hedu_utah_janosvm_resources_StdioSpec *jss)
{
	stdioSpec *ss = OBJECT2NATIVE(&metalStdioResource.link.specNativeClass,
				      (Hjava_lang_Object *)jss);
	struct Hjava_lang_String *retval;

	if( !(retval = stringC2Java(ss->errURI)) )
	{
		errorInfo einfo;
		
		postOutOfMemory(&einfo);
		throwError(&einfo);
	}
	return( retval );
}

struct Hjava_lang_Object *
edu_utah_janosvm_resources_StdioSpec_clone(
	struct Hedu_utah_janosvm_resources_StdioSpec *jss)
{
	stdioSpec *ss = OBJECT2NATIVE(&metalStdioResource.link.specNativeClass,
				      (Hjava_lang_Object *)jss);
	struct Hedu_utah_janosvm_resources_StdioSpec *retval;
	stdioSpec *retval_ss;

	retval = (struct Hedu_utah_janosvm_resources_StdioSpec *)
		newObject(metalStdioResource.link.specNativeClass.peer);
	ss = OBJECT2NATIVE(&metalStdioResource.link.specNativeClass,
			   (Hjava_lang_Object *)jss);
	retval_ss = OBJECT2NATIVE(&metalStdioResource.link.specNativeClass,
				  (Hjava_lang_Object *)retval);
	retval_ss->spec.handler = &metalStdioResource.link;
	retval_ss->inURI = checkPtr(gc_malloc(strlen(ss->inURI) + 1,
					      GC_ALLOC_FIXED));
	strcpy(retval_ss->inURI, ss->inURI);
	retval_ss->outURI = checkPtr(gc_malloc(strlen(ss->outURI) + 1,
					       GC_ALLOC_FIXED));
	strcpy(retval_ss->outURI, ss->outURI);
	retval_ss->errURI = checkPtr(gc_malloc(strlen(ss->errURI) + 1,
					       GC_ALLOC_FIXED));
	strcpy(retval_ss->errURI, ss->errURI);
	return( (Hjava_lang_Object *)retval );
}

struct Hedu_utah_janosvm_resources_StdioSpec *
edu_utah_janosvm_resources_StdioSpec_current(void)
{
	struct Hedu_utah_janosvm_resources_StdioSpec *retval = 0;
	stdioClient *sc = (stdioClient *)
		getTeamInternalResource(0, &metalStdioResource.link);
	stdioSpec *ss;

	retval = (struct Hedu_utah_janosvm_resources_StdioSpec *)
		newObject(metalStdioResource.link.specNativeClass.peer);
	retval->found = 1;
	ss = OBJECT2NATIVE(&metalStdioResource.link.specNativeClass,
			   (Hjava_lang_Object *)retval);
	ss->spec.handler = &metalStdioResource.link;
	ss->inURI = sc->streams[STDIO_IN].uri;
	ss->outURI = sc->streams[STDIO_OUT].uri;
	ss->errURI = sc->streams[STDIO_ERR].uri;
	return( retval );
}

static
char *copyIOStreamURI(ioStream *ios)
{
	char *retval = 0;
	char buffer[32];

	buffer[0] = '\0';
	switch( ios->kind )
	{
	case SK_NONE:
		strcpy(buffer, "builtin://none");
		break;
	case SK_FILE:
	case SK_SHARED_FILE:
		sprintf(buffer, "fd://%d", fileno(ios->handle.file));
		break;
	case SK_JSI:
		strcpy(buffer, "jsi://");
		break;
	}
	if( (retval = gc_malloc(strlen(buffer) + 1, GC_ALLOC_FIXED)) )
	{
		strcpy(retval, buffer);
	}
	return( retval );
}

struct Hedu_utah_janosvm_resources_StdioSpec *
edu_utah_janosvm_resources_StdioSpec_currentFD()
{
	struct Hedu_utah_janosvm_resources_StdioSpec *retval = 0;
	stdioClient *sc = (stdioClient *)
		getTeamInternalResource(0, &metalStdioResource.link);
	stdioSpec *ss;
	int error = 0;

	retval = (struct Hedu_utah_janosvm_resources_StdioSpec *)
		newObject(metalStdioResource.link.specNativeClass.peer);
	ss = OBJECT2NATIVE(&metalStdioResource.link.specNativeClass,
			   (Hjava_lang_Object *)retval);
	ss->spec.handler = &metalStdioResource.link;
	if( !(ss->inURI = copyIOStreamURI(&sc->streams[STDIO_IN])) )
		error = 1;
	if( !(ss->outURI = copyIOStreamURI(&sc->streams[STDIO_OUT])) )
		error = 1;
	if( !(ss->errURI = copyIOStreamURI(&sc->streams[STDIO_ERR])) )
		error = 1;
	if( error )
	{
		errorInfo einfo;
		
		postOutOfMemory(&einfo);
		throwError(&einfo);
	}
	return( retval );
}
