/*
 * stdioResource.h
 * Resource to managed standard I/O for each team.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _stdio_resource_h
#define _stdio_resource_h

#include <stdio.h>
#include <jsi/jsi.h>

#include "resource.h"

/*
 * An ioStream is used to manage an individual data stream used for standard
 * I/O.
 *
 * `uri' - Description of the type of stream and any additional data.
 *   The URI is of the form `<protocol>://<path>' where protocol is one of:
 *     builtin - The VM's stdio streams.  Possible paths are `null' which
 *       specifies no stream, `stdin' for standard in, `stdout' for standard
 *       out, and `stderr' for standard error.
 *     jsi - A connected jsiChannel, currently, just the one that executed
 *       run of team x.
 *     file - A regular file where path should be the path of the file to
 *       create.
 *     fd - A file descriptor in the VM, the fd given in the path is dup'd
 *       and used to build a FILE object.
 * `kind' - The type of stream.
 * `handle.file' - Pointer to the a regular stdio file handle.
 * `handle.jsi' - Pointer to JSI channel that is connected to something that
 *   can handle jsi messages sent to stdio objects.
 */

enum {
	SK_NONE,
	SK_SHARED_FILE,	/* A shared stdio FILE is used for I/O */
	SK_FILE,	/* An exclusive stdio FILE is used for I/O */
	SK_JSI,		/* A jsiChannel is used for I/O */
	SK_MAX
};

typedef struct ioStream {
	char *uri;
	int kind;
	union {
		FILE *file;
		jsiChannel *jsi;
	} handle;
} ioStream;

/*
 * The root stdio resource.
 */
typedef struct _stdioResource {
	resource link;
} stdioResource;

/* Identifiers for each stdio stream */
enum {
	STDIO_IN,
	STDIO_OUT,
	STDIO_ERR,
	STDIO_STREAM_MAX
};

/*
 * Client for the stdio resource, manages stdin, stdout, and stderr.
 *
 * `client' - Standard resourceClient header.
 * `streams' - Array of ioStreams representing stdin, stdout, and stderr.
 */
typedef struct _stdioClient {
	resourceClient client;
	ioStream streams[STDIO_STREAM_MAX];
} stdioClient;

/*
 * Stdio resource specification.
 *
 * `inURI' - URI describing stdin.
 * `outURI' - URI describing stdout.
 * `errURI' - URI describing stderr.
 */
typedef struct _stdioSpec {
	resourceSpec spec;
	char *inURI;
	char *outURI;
	char *errURI;
} stdioSpec;

/*
 * Attach a jsiChannel to the current team's stdio streams
 */
void stdioAttachJSIChannel(jsiChannel *jc);
/*
 * Detach a jsiChannel from the current team's stdio streams
 */
void stdioDetachJSIChannel(void);

/*
 * Print a formatted string to the given stream.
 */
void stdioPrint(int stream, char *fmt, ...);
/*
 * The va_list form of the above function.
 */
void stdioVPrint(int stream, char *fmt, va_list args);
/*
 * Write a block of memory to the given stream.
 */
void stdioWrite(int stream, char *buffer, int size);
/*
 * Read a block of memory from stdin.
 */
int stdioRead(char *buffer, int size);
/*
 * Flush the buffer of the given stream.
 */
void stdioFlush(int stream);

/*
 * AddModule: stdioResourceModule
 *   ModuleDependency: resourceModule.internalResources
 *   ModuleDependency: memResourceModule.initMetal
 * Transition: initMetal
 *   ModuleDependency: resourceModule.reinstall
 */

#endif
