/*
 * tags.c
 * A simple abstraction for managing key/value pairs.
 *
 * Copyright (c) 2000, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdarg.h>

#include "stringSupport.h"
#include "assert_pp.h"
#include "module.h"
#include "mungWallProxy.h"
#include "tags.h"

#define DEFAULT_LIST_GROWTH 2

/** Module declarations **/
static int initTagsModule(int stage);

module_t tagsModule = {
	"tags",
	0,
	initTagsModule,
	0,
	0
};

/** Section file declarations **/
static int tagsSectionInstall(struct _sectionType *st, int tag, va_list *args);
static int tagsSectionInit(struct _sectionType *st,
			   struct _sectionFileData *sfd);
static void tagsSectionKill(struct _sectionType *st,
			    struct _sectionFileData *sfd);
static int tagsSectionCache(struct _sectionType *st,
			    struct _sectionFileData *sfd,
			    void *ps, char *line, int len);

static sectionTypeFunctions tagsSectionFunctions = {
	tagsSectionInstall,
	0, /* Uninstall */
	tagsSectionInit,
	tagsSectionKill,
	tagsSectionCache,
	0, /* Flush */
	0 /* Walk */
};

tagSectionType tagsFileSectionTemplate;
tagSectionType pairsFileSectionType;

/** Module functions **/
int initTagsModule(int stage)
{
	installSectionType(&tagsFileSectionTemplate.type,
			   STA_Size, sizeof(tagSectionData),
			   STA_Functions, &tagsSectionFunctions,
			   STA_TAG_DONE);
	installSectionType(&pairsFileSectionType.type,
			   STA_Name, "pairs",
			   STA_Size, sizeof(tagSectionData),
			   STA_Functions, &tagsSectionFunctions,
			   STA_TAG_DONE);
	return( 1 );
}

/** Section file functions **/
int tagsSectionInstall(struct _sectionType *st, int tag, va_list *args)
{
	tagSectionType *tfs = (tagSectionType *)st;
	int retval = 1;

	switch( tag )
	{
	case TSA_DefaultArray:
		tfs->defaultArray = va_arg(*args, tagArray *);
		if( tfs->defaultArray )
		{
			/* Add any tagData to the end of our section data */
			st->dataSize +=
				getTagCount(tfs->defaultArray->desc) *
				sizeof(tagData);
		}
		break;
	default:
		break;
	}
	return( retval );
}

int tagsSectionInit(struct _sectionType *st, struct _sectionFileData *sfd)
{
	tagSectionData *tsd = LOCAL_SECTION_DATA(tagSectionData *, st, sfd);
	tagSectionType *tfs = (tagSectionType *)st;
	int retval = 0;
	tagList *tl;

	if( (tl = createTagList()) )
	{
		if( tfs->defaultArray )
		{
			/* Copy the default tags */
			tsd->tags.desc = tfs->defaultArray->desc;
			tsd->tags.data = (tagData *)(tsd + 1);
			copyTagArray(&tsd->tags, tfs->defaultArray);
		}
		else
		{
			tsd->tags.desc = 0;
			tsd->tags.data = 0;
		}
		tsd->taglist = tl;
		retval = 1;
	}
	return( retval );
}

void tagsSectionKill(struct _sectionType *st, struct _sectionFileData *sfd)
{
	tagSectionData *tsd = LOCAL_SECTION_DATA(tagSectionData *, st, sfd);
	
	deleteTagList(tsd->taglist);
}

int tagsSectionCache(struct _sectionType *st, struct _sectionFileData *sfd,
		     void *ps, char *line, int len)
{
	tagSectionData *tsd = LOCAL_SECTION_DATA(tagSectionData *, st, sfd);
	char *tag_name = 0, *tag_data = 0;
	int retval = 1;

	/* Try to parse the line */
	parseSectionLine(ps, &tag_name, &tag_data, 0);
	if( tag_name )
	{
		int index;

		/* Check for a hit in the tagArray first */
		if( (index = getTagIndex(tsd->tags.desc, tag_name)) >= 0 )
		{
			if( !parseTagDataString(&tsd->tags.data[index],
						tsd->tags.desc[index].type,
						tag_data) )
			{
				retval = 0;
			}
		}
		else
		{
			Utf8Const *tag_utf, *data_utf;

			/* Its unexpected add it to the tagList */
			if( (tag_utf = utf8ConstNew(tag_name, -1)) &&
			    (data_utf = utf8ConstNew(tag_data, -1)) )
			{
				if( !addTag(tsd->taglist,
					    tag_utf, TAG_STRING, data_utf) )
				{
					retval = 0;
				}
			}
			else
			{
				utf8ConstRelease(tag_utf);
				retval = 0;
			}
		}
	}
	return( retval );
}

/** Tags functions **/
static
int resizeTagList(tagList *tl, int new_size)
{
	int retval = 1;

	require(tl != 0);
	require(new_size >= 0);
	
	if( new_size > tl->tags.max )
	{
		tagItem *new_items;

		/* Need more space in the array */
		if( (new_items = gc_realloc(tl->tags.data,
					    sizeof(tagItem) *
					    (new_size + DEFAULT_LIST_GROWTH),
					    GC_ALLOC_TAGS)) )
		{
			tl->tags.data = new_items;
			tl->tags.count = new_size;
			tl->tags.max = new_size + DEFAULT_LIST_GROWTH;

			ensure_valid_memory(new_items);
		}
		else
		{
			retval = 0;
		}
	}
	else
	{
		tl->tags.count = new_size;

		ensure(tl->tags.data != 0);
	}
	return( retval );
}

tagList *createTagList(void)
{
	tagList *retval;

	if( (retval = gc_malloc(sizeof(tagList), GC_ALLOC_TAGS)) )
	{
		retval->users = 1;
		/* everything else is just zero ... */
	}
	return( retval );
}

tagList *grabTagList(tagList *tl)
{
	require(tl != 0);
	
	tl->users++;
	return( tl );
}

void deleteTagList(tagList *tl)
{
	if( tl )
	{
		tl->users--;
		/* Only free when there are no more users */
		if( tl->users == 0 )
		{
			int lpc;

			/* Drop any references we might have */
			for( lpc = 0; lpc < tl->tags.count; lpc++ )
			{
				switch( tl->tags.data[lpc].type )
				{
				case TAG_STRING:
					utf8ConstRelease(tl->tags.data[lpc].
							 data.str);
					break;
				case TAG_LIST:
					deleteTagList(tl->tags.data[lpc].
						      data.tl);
					break;
				default:
					break;
				}
			}
			gc_free(tl->tags.data);
			gc_free(tl);
		}
	}
}

tagItem *addTag(tagList *tl, Utf8Const *name, tagType type, ...)
{
	tagItem *retval = 0;
	int slot;
	va_list args;

	require(tl != 0);
	require(name != 0);
	require(type > TAG_NONE);
	require(type < TAG_MAX);
	
	va_start(args, type);
	slot = tl->tags.count;
	if( resizeTagList(tl, tl->tags.count + 1) )
	{
		retval = &tl->tags.data[slot];
		retval->tag = name;
		retval->type = type;
		/* Get the data based on its type */
		switch( type )
		{
		case TAG_BOOLEAN:
			retval->data.b = va_arg(args, int);
			break;
		case TAG_CHARACTER:
			retval->data.c = va_arg(args, int);
			break;
		case TAG_INTEGER:
			retval->data.i = va_arg(args, int);
			break;
		case TAG_FLOAT:
			retval->data.f = (float)va_arg(args, double);
			break;
		case TAG_STRING:
			retval->data.str = va_arg(args, Utf8Const *);
			break;
		case TAG_LIST:
			retval->data.tl = va_arg(args, tagList *);
			break;
		case TAG_ARRAY:
			retval->data.ta = va_arg(args, tagArray *);
			break;
		case TAG_POINTER:
			retval->data.ptr = va_arg(args, void *);
			break;
		default:
			break;
		}
	}
	va_end(args);
	return( retval );
}

tagItem *findTag(tagList *tl, Utf8Const *name, int instance)
{
	tagItem *retval = 0;
	int lpc;

	require(tl != 0);
	require(name != 0);
	require(instance >= 0);
	
	for( lpc = 0; !retval && (lpc < tl->tags.count); lpc++ )
	{
		if( utf8ConstEqual(name, tl->tags.data[lpc].tag) )
		{
			if( !instance )
				retval = &tl->tags.data[lpc];
			else
				instance--;
		}
	}
	return( retval );
}

tagItem *findTagStr(tagList *tl, char *name, int len, int instance)
{
	tagItem *retval = 0;
	int lpc;

	require(tl != 0);
	require(name != 0);
	require((len == -1) || (len >= 0));
	require(instance >= 0);
	
	if( len < 0 )
		len = strlen(name);
	for( lpc = 0; !retval && (lpc < tl->tags.count); lpc++ )
	{
		if( !strncmp(name, tl->tags.data[lpc].tag->data, len) &&
		    (((char *)tl->tags.data[lpc].tag->data)[len] == 0) )
		{
			if( !instance )
				retval = &tl->tags.data[lpc];
			else
				instance--;
		}
	}
	return( retval );
}

int getListTags(tagList *tl, Utf8Const *name, ...)
{
	int instance, retval = 1;
	Utf8Const **utf_ptr;
	tagArray **ta_ptr;
	float *float_ptr;
	tagList **tl_ptr;
	void **void_ptr;
	char *char_ptr;
	int *int_ptr;
	va_list args;
	tagItem *ti;

	require(tl != 0);
	require(name != 0);

	va_start(args, name);
	/* Walk the list of args until theres a null name */
	while( name )
	{
		instance = va_arg(args, int);
		if( (ti = findTag(tl, name, instance)) )
		{
			/* Found the tag set the data */
			switch( ti->type )
			{
			case TAG_BOOLEAN:
				char_ptr = va_arg(args, char *);
				
				*char_ptr = ti->data.b;
				break;
			case TAG_CHARACTER:
				char_ptr = va_arg(args, char *);
				
				*char_ptr = ti->data.c;
				break;
			case TAG_INTEGER:
				int_ptr = va_arg(args, int *);
				
				*int_ptr = ti->data.i;
				break;
			case TAG_FLOAT:
				float_ptr = va_arg(args, float *);
				
				*float_ptr = ti->data.f;
				break;
			case TAG_STRING:
				utf_ptr = va_arg(args, Utf8Const **);
				
				*utf_ptr = ti->data.str;
				break;
			case TAG_LIST:
				tl_ptr = va_arg(args, tagList **);
				
				*tl_ptr = ti->data.tl;
				break;
			case TAG_ARRAY:
				ta_ptr = va_arg(args, tagArray **);
				
				*ta_ptr = ti->data.ta;
				break;
			case TAG_POINTER:
				void_ptr = va_arg(args, void **);
				
				*void_ptr = ti->data.ptr;
				break;
			default:
				break;
			}
		}
		else
		{
			/* Didn't find the tag, skip over the pointer */
			retval = 0;
			switch( ti->type )
			{
			case TAG_BOOLEAN:
				char_ptr = va_arg(args, char *);
				break;
			case TAG_CHARACTER:
				char_ptr = va_arg(args, char *);
				break;
			case TAG_INTEGER:
				int_ptr = va_arg(args, int *);
				break;
			case TAG_FLOAT:
				float_ptr = va_arg(args, float *);
				break;
			case TAG_STRING:
				utf_ptr = va_arg(args, Utf8Const **);
				break;
			case TAG_LIST:
				tl_ptr = va_arg(args, tagList **);
				break;
			case TAG_ARRAY:
				ta_ptr = va_arg(args, tagArray **);
				break;
			case TAG_POINTER:
				void_ptr = va_arg(args, void **);
				break;
			default:
				break;
			}
		}
		name = va_arg(args, Utf8Const *);
	}
	va_end(args);
	return( retval );
}

/* String corresponding to tagType's */
static
char *tag_type_names[] = {
	"none",
	"boolean",
	"character",
	"integer",
	"float",
	"string",
	"taglist",
	"array",
	"pointer"
};

void printTagItem(tagItem *ti)
{
	require(ti != 0);
	
	printf("%s %s = ",
	       tag_type_names[ti->type],
	       ti->tag->data);
	switch( ti->type )
	{
	case TAG_BOOLEAN:
		printf("%s\n", ti->data.b ? "true" : "false");
		break;
	case TAG_CHARACTER:
		printf("`%c'\n", ti->data.c);
		break;
	case TAG_INTEGER:
		printf("%d\n", ti->data.i);
		break;
	case TAG_FLOAT:
		printf("%f\n", ti->data.f);
		break;
	case TAG_STRING:
		printf("%s\n", ti->data.str->data);
		break;
	case TAG_LIST:
		printf("%p\n", ti->data.tl);
		break;
	case TAG_ARRAY:
		printf("%p\n", ti->data.ta);
		break;
	case TAG_POINTER:
		printf("%p\n", ti->data.ptr);
		break;
	default:
		printf("\n");
		break;
	}
}

void initTagArray(tagArray *ta)
{
	require(ta != 0);
	
	/* Nothing to do */
}

void killTagArray(tagArray *ta)
{
	int lpc;

	require(ta != 0);
	
	/* Drop any references */
	for( lpc = 0; ta->desc[lpc].tag; lpc++ )
	{
		switch( ta->desc[lpc].type )
		{
		case TAG_STRING:
			utf8ConstRelease(ta->data[lpc].str);
			break;
		case TAG_LIST:
			deleteTagList(ta->data[lpc].tl);
			break;
		default:
			break;
		}
	}
}

void copyTagArray(tagArray *dest, tagArray *src)
{
	if( dest && src )
	{
		int lpc;

		/* Copy any values and add references when necessary */
		for( lpc = 0; src->desc[lpc].tag; lpc++ )
		{
			dest->data[lpc] = src->data[lpc];
			switch( src->desc[lpc].type )
			{
			case TAG_STRING:
				if( src->data[lpc].str )
					utf8ConstAddRef(src->data[lpc].str);
				break;
			case TAG_LIST:
				if( src->data[lpc].tl )
					src->data[lpc].tl->users++;
				break;
			default:
				break;
			}
		}
	}
}

int getTagCount(tagDesc *td)
{
	int retval = 0;

	if( td )
	{
		for( ; td[retval].tag; retval++ );
	}
	return( retval );
}

int getTagIndex(tagDesc *td, char *name)
{
	int lpc, retval = -1;

	if( td )
	{
		for( lpc = 0; (retval == -1) && td[lpc].tag; lpc++ )
		{
			if( !strcmp(td[lpc].tag, name) )
				retval = lpc;
		}
	}
	return( retval );
}

tagType getTagType(tagArray *ta, char *name)
{
	tagType retval = TAG_NONE;
	int index;

	require(ta != 0);
	require(name != 0);
	
	if( (index = getTagIndex(ta->desc, name)) >= 0 )
	{
		retval = ta->desc[index].type;
	}
	return( retval );
}

void setTagData(tagArray *ta, char *name, ...)
{
	int index;
	va_list args;

	require(ta != 0);
	require(name != 0);
	
	va_start(args, name);
	if( (index = getTagIndex(ta->desc, name)) >= 0 )
	{
		switch( ta->desc[index].type )
		{
		case TAG_BOOLEAN:
			ta->data[index].b = va_arg(args, int);
			break;
		case TAG_CHARACTER:
			ta->data[index].c = va_arg(args, int);
			break;
		case TAG_INTEGER:
			ta->data[index].i = va_arg(args, int);
			break;
		case TAG_FLOAT:
			ta->data[index].f = (float)va_arg(args, double);
			break;
		case TAG_STRING:
			ta->data[index].str = va_arg(args, Utf8Const *);
			break;
		case TAG_LIST:
			ta->data[index].tl = va_arg(args, tagList *);
			break;
		case TAG_ARRAY:
			ta->data[index].ta = va_arg(args, tagArray *);
			break;
		case TAG_POINTER:
			ta->data[index].ptr = va_arg(args, void *);
			break;
		default:
			break;
		}
	}
	va_end(args);
}

tagData getTagData(tagArray *ta, char *name)
{
	tagData retval;
	int index;

	require(ta != 0);
	require(name != 0);

	retval.ptr = 0;
	if( (index = getTagIndex(ta->desc, name)) >= 0 )
	{
		retval = ta->data[index];
	}
	return( retval );
}

int getArrayTags(tagArray *ta, char *name, ...)
{
	int index, retval = 1;
	Utf8Const **utf_ptr;
	tagArray **ta_ptr;
	tagList **tl_ptr;
	float *float_ptr;
	void **void_ptr;
	char *char_ptr;
	int *int_ptr;
	va_list args;

	require(ta != 0);
	require(name != 0);
	
	va_start(args, name);
	/* Walk the arg list until there are no more names to get */
	while( retval && name )
	{
		if( (index = getTagIndex(ta->desc, name)) >= 0 )
		{
			switch( ta->desc[index].type )
			{
			case TAG_BOOLEAN:
				char_ptr = va_arg(args, char *);

				*char_ptr = ta->data[index].b;
				break;
			case TAG_CHARACTER:
				char_ptr = va_arg(args, char *);

				*char_ptr = ta->data[index].c;
				break;
			case TAG_INTEGER:
				int_ptr = va_arg(args, int *);

				*int_ptr = ta->data[index].i;
				break;
			case TAG_FLOAT:
				float_ptr = va_arg(args, float *);

				*float_ptr = ta->data[index].f;
				break;
			case TAG_STRING:
				utf_ptr = va_arg(args, Utf8Const **);

				*utf_ptr = ta->data[index].str;
				break;
			case TAG_LIST:
				tl_ptr = va_arg(args, tagList **);

				*tl_ptr = ta->data[index].tl;
				break;
			case TAG_ARRAY:
				ta_ptr = va_arg(args, tagArray **);

				*ta_ptr = ta->data[index].ta;
				break;
			case TAG_POINTER:
				void_ptr = va_arg(args, void **);

				*void_ptr = ta->data[index].ptr;
				break;
			default:
				break;
			}
			name = va_arg(args, char *);
		}
		else
		{
			retval = 0;
		}
	}
	va_end(args);
	return( retval );
}

int parseTagDataString(tagData *td, int type, char *value)
{
	int retval = 1;

	require(td != 0);
	require(type > TAG_NONE);
	require(type < TAG_LIST);
	require(value != 0);
	
	switch( type )
	{
	case TAG_BOOLEAN:
		{
			int lpc;

			td->b = 0;
			/*
			 * Test against the trueStrings defined by sectionFile
			 */
			for( lpc = 0; trueStrings[lpc]; lpc++ )
			{
				if( !strcmp(value, trueStrings[lpc]) )
				{
					td->b = 1;
					break;
				}
			}
		}
		break;
	case TAG_CHARACTER:
		td->c = value[0];
		break;
	case TAG_INTEGER:
		if( sscanf(value, "%d", &td->i) != 1 )
			retval = 0;
		break;
	case TAG_FLOAT:
		if( sscanf(value, "%f", &td->f) != 1 )
			retval = 0;
		break;
	case TAG_STRING:
		if( !(td->str = utf8ConstNew(value, -1)) )
			retval = 0;
		break;
	case TAG_LIST:
	case TAG_ARRAY:
	case TAG_POINTER:
		break;
	}
	return( retval );
}
