/*
 * tags.h
 * A simple abstraction for managing key/value pairs.
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _tags_h
#define _tags_h

/*
 * The tags module provides a variety of structures and functions for
 * manipulating lists of key/value pairs.  This is mostly useful for managing
 * input from files/users.
 */

#include <misc/sectionFile.h>
#include "gtypes.h"

/* The type of data stored inside of a tag. */
typedef enum {
	TAG_NONE,
	TAG_BOOLEAN,
	TAG_CHARACTER,
	TAG_INTEGER,
	TAG_FLOAT,
	TAG_STRING, /* UTF8 really */
	TAG_LIST,
	TAG_ARRAY,
	TAG_POINTER,
	TAG_MAX
} tagType;

struct _tagList;
struct _tagArray;

/*
 * The tagData union stores the data value for a specific key.
 */
typedef union _tagData {
	char b;
	char c;
	int i;
	float f;
	Utf8Const *str;
	struct _tagList *tl;
	struct _tagArray *ta;
	void *ptr;
} tagData;

/*
 * A tagDesc structure is meant to be statically defined and then used to
 * process expected sets of keys.
 *
 * `tag' - The key value that some value should be attached to.
 * `type' - The expected type of data
 */
typedef struct _tagDesc {
	char *tag;
	tagType type;
} tagDesc;

/*
 * A tagArray structure holds the set of values for tagDesc element.
 *
 * `desc' - NULL terminated array of key descriptions, the value for this key
 *   can then be found at the same index in the `data' array.
 * `data' - The data values for each key.
 */
typedef struct _tagArray {
	tagDesc *desc;
	tagData *data;
} tagArray;

/*
 * A tagItem structure is used to store more dynamic sets of data.
 *
 * `tag' - The key for this item.
 * `type' - The type of data stored in `data'.
 * `data' - The data value attached to this key.
 */
typedef struct _tagItem {
	Utf8Const *tag;
	tagType type;
	tagData data;
} tagItem;

/*
 * A tagList is a vector of tagItems, it can hold a number of tagItems and
 * can be grown/shrunk dynamically.
 *
 * `users' - Reference count for this tagList.
 * `tags.data' - The array of tagItems in this tagList.
 * `tags.count' - The number of items in the `tags.data' array that are in use.
 * `tags.max' - The allocated size of the `tags.data' array.
 */
typedef struct _tagList {
	int users;
	struct {
		tagItem *data;
		int count;
		int max;
	} tags;
} tagList;

/*
 * The tagSectionType is the template sectionType structure that can handle
 * tag lists.
 *
 * `type' - The sectionType header.
 * `defaultArray' - A tagArray which defines any expected keys and their
 *   default values.
 */
typedef struct _tagSectionType {
	sectionType type;
	tagArray *defaultArray;
} tagSectionType;

/*
 * The tagSectionData structure is the local section data that a
 * tagSectionType template will create.
 *
 * `tags' - The tagArray with any expected keys and their values.  This is
 *   filled with the default values before the any values specified in the
 *   file are filled in.
 * `tagList' - Any additional keys/values are stored in this tagList.
 */
typedef struct _tagSectionData {
	tagArray tags;
	tagList *taglist;
} tagSectionData;

/*
 * Function tags for tagSectionType
 */
enum {
	TSA_BASE = STA_USER,
	TSA_DefaultArray,	/* (tagArray *) - The default tagArray */
};

/*
 * Create an empty tagList
 */
tagList *createTagList(void);
/*
 * Delete a tagList and its elements.
 */
void deleteTagList(tagList *tl);
/*
 * Add a reference to the given tagList.
 */
tagList *grabTagList(tagList *tl);
/*
 * Add a tag to the given tag list.
 */
tagItem *addTag(tagList *tl, Utf8Const *name, tagType type, ...);
/*
 * Find a tag with the given name and instance in the tagList
 */
tagItem *findTag(tagList *tl, Utf8Const *name, int instance);
/*
 * Same as above except search based on a regular C string instead of a utf8.
 */
tagItem *findTagStr(tagList *tl, char *name, int len, int instance);
/*
 * Get a number of tags from a tagList, the arguments are the name of the
 * tag, an instance number, and then a pointer to the storage for the data
 * if its found.  If any of the values aren't found then the function returns
 * zero.  However, it will try to get as many of them as possible, so if you
 * initialize the storage with defaults before hand, you needn't worry about
 * failure.
 */
int getListTags(tagList *tl, Utf8Const *name, ...);
/*
 * Print the value of a tag item.
 * (For debugging use only)
 */
void printTagItem(tagItem *ti);

/*
 * Initialize a tagArray
 */
void initTagArray(tagArray *ta);
/*
 * Destruct a tagArray
 */
void killTagArray(tagArray *ta);
/*
 * Copy the values of a tagArray
 */
void copyTagArray(tagArray *dest, tagArray *src);
/*
 * Get the number of tagDesc's in the given NULL terminated array of them.
 */
int getTagCount(tagDesc *td);
/*
 * Get the array index of the given tag, returns -1 if the tag wasn't found.
 */
int getTagIndex(tagDesc *td, char *name);
/*
 * Get the type of a given tag
 */
tagType getTagType(tagArray *ta, char *name);
/*
 * Set the value of a tag
 */
void setTagData(tagArray *ta, char *name, ...);
/*
 * Get the value of a tag
 */
tagData getTagData(tagArray *ta, char *name);
/*
 * Similar to getListTags except on an array and an instance number doesn't
 * have to be specified.
 */
int getArrayTags(tagArray *ta, char *name, ...);
/*
 * Attempt to parse the given string based on the given type.  If the parse
 * is successful the data is stored in the tagData structure and true is
 * returned.
 */
int parseTagDataString(tagData *td, int type, char *value);

/*
 * A sectionType structure that can be used as a template for sectionTypes
 * that just want to work with tagArray's/tagList's.
 */
extern tagSectionType tagsFileSectionTemplate;

extern tagSectionType pairsFileSectionType;

/*
 * AddModule: tagsModule
 *   ModuleDependency: sectionFileModule.templates
 *   Comment: Install the tagsSection template
 */

#endif
