/*
 * jsi.h
 * JanosVM Scripting Interface functions
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jsi_h
#define _jsi_h

/*
 * JanosVM Scripting Interface
 *
 * JSI is a simple messaging interface for controlling a JanosVM instance from
 * another program.  In order to keep the interface flexible and independent of
 * use it only defines a simple message format and the semantics for handling
 * incoming messages.  Therefore, a user is free to control the VM through
 * various scripting languages/shells/programs and we don't have to drag in
 * these large chunks of code.
 */

#include "jsiSystem.h"

struct _jsiChannel;
struct _jsiProvider;
struct _jsiHandlerContext;

/*
 * A jsiConnectionHandler is a function that is called when a connection to
 * an application has been requested.
 *
 * `provider' - The provider that received the connection request.
 * `channel' - A jsiChannel that is connected to the other application.
 */
typedef int (*jsiConnectionFunction)(struct _jsiProvider *provider,
				     struct _jsiChannel *channel);
typedef struct _jsiConnectionHandler {
	jsiConnectionFunction func;
} jsiConnectionHandler;


/*
 * The jsiProviderFunctions structure holds the implementation for a
 * "jsiProvider" which provides a point where outside programs and connect
 * to the application.
 *
 * `names' - The list of names/aliases that this provider implementation
 *   supports.
 * `create' - Create and return a provider object that corresponds to this
 *   implementation.
 * `delete' - Delete a provider object returned by `create'.
 * `attach' - Attach the provider so that it can begin servicing requests
 *   for connections.
 * `detach' - Detach the provider.
 * `handle' - The handler function for this provider, usually run in a
 *   separate thread.
 */
typedef struct _jsiProviderFunctions {
	char **names;
	struct _jsiProvider *(*create)(jsiConnectionHandler *ch);
	void (*delete)(struct _jsiProvider *jp);
	int (*attach)(struct _jsiProvider *jp, char *format);
	void (*detach)(struct _jsiProvider *jp);
	void (*handle)(struct _jsiProvider *jp);
} jsiProviderFunctions;

/*
 * A jsiProvider is used to listen on a port and handle any connections
 * from the outside.  Typically, the response should be to start a thread to
 * handle any incoming messages.
 *
 * `ops' - The implementation functions for this provider.
 * `handleConnection' - The function to call when a connection has been
 *   requested.
 */
typedef struct _jsiProvider {
	jsiProviderFunctions *ops;
	jsiConnectionHandler *handleConnection;
} jsiProvider;

/*
 * The jsiChannelFunctions structure holds the implementation for a
 * "jsiChannel" which provides a conduit for sending and receiving jsi
 * messages.
 *
 * `names' - The list of names/aliases that this provider implementation
 *   supports.
 * `create' - Create and return a channel object that corresponds to this
 *   implementation.
 * `delete' - Delete a provider object returned by `create'.
 * `connect' - Connect this channel to some application.
 * `disconnect' - Disconnect the channel.
 * `sendMessage' - Send a flattened jsiMessage.
 * `doMessage' - Send a message and wait for its reply.
 * `replyMessage' - Send a reply to a message.
 * `recvMessage' - Wait for a message to come in and return its flattened
 *   version.
 */
typedef struct _jsiChannelFunctions {
	char **names;
	struct _jsiChannel *(*create)(void);
	void (*delete)(struct _jsiChannel *jc);
	int (*connect)(struct _jsiChannel *jc, int tag, ...);
	void (*disconnect)(struct _jsiChannel *jc);
	int (*sendMessage)(struct _jsiChannel *jc, char *buffer, int size);
	struct _jsiMessage *(*doMessage)(struct _jsiHandlerContext *hc,
					 struct _jsiMessage *jm);
	int (*replyMessage)(struct _jsiChannel *jc, char *buffer,
			    int bufferSize);
	char *(*recvMessage)(struct _jsiChannel *jrc, int *bufferSize);
	char *(*encode)(struct _jsiChannel *jc);
} jsiChannelFunctions;

/*
 * A jsiChannel is the actual object used to communicate with another
 * application.
 *
 * `ops' - The implementation functions for this channel.
 */
typedef struct _jsiChannel {
	jsiChannelFunctions *ops;
} jsiChannel;

/*
 * A jsiPropertyInfo is a description of a property that exported by some
 * scriptable object.  The structure can then be used by scripting programs
 * to determine what operations are available for that object.
 *
 * `name' - The name of the property.
 * `commands' - A bitmask of the commands that are implemented.  The bits
 *   correspond to the values available in the `what' field of a jsiMessage
 *   (described below).
 * `specifiers' - The specifier associated with the property.
 * `usage' - A description of the usage of this property and commands.
 */
typedef struct _jsiPropertyInfo {
	char *name;
	int commands;
	char *specifiers;
	char *usage;
} jsiPropertyInfo;

/*
 * A jsiStringSequence is a substructure used by jsiMessage to hold a sequence
 * of string values.
 *
 * `data' - An array of pointers to strings.
 * `count' - The number of valid strings in `data'.
 * `max' - The allocated size of the `data' array.
 */
typedef struct _jsiStringSequence {
	char **data;
	int count;
	int max;
} jsiStringSequence;

/*
 * A jsiMessage is the primary structure in JSI, all commands and data are
 * carried in these messages.  Messages are flattened to a host independent
 * byte array, and are instantiated from the same byte array so there are
 * no dependencies on the transport used for the byte array.
 *
 * `what' - Indicates what an object should do in response to the message.
 * `specifiers' - The specifier stack for this message.  A specifier is simply
 *   a name/value pair that is used to direct the message to its destination.
 *   In other words, it "specifies" which object should receive the message.
 * `identifier' - An index of the data items in the message.  Each item
 *   has an identifier string which is used in conjunction with an instance
 *   number to select the item for retrieval.  All data items are encoded as
 *   strings, however, the provided functions will automatically encode/decode,
 *   some data types.
 * `contents.data' - Holds the byte array that encodes the message.  All
 *   data is copied into this buffer when added to the message, and any
 *   specifier/identifier pointers reference positions in this array.
 * `contents.size' - The current used size of the `contents.data' array.
 * `contents.max' - The allocated size of the `contents.data' array.
 */

/* jsiMessage what values */
enum {
	JSI_WHAT_NONE,
	JSI_ERROR,		/* Signal an error */
	JSI_SHUTDOWN,		/* Shutdown the transport */
	JSI_CARRIER,		/* The message only carries data */
	JSI_REPLY,		/* A reply message */
	JSI_GET_SUITES,		/*
				 * Get the jsiPropertyInfo that
				 * describes this property
				 */
	JSI_LIST_PROPERTIES,	/* List sub-properties of this one */
	JSI_GET_PROPERTY,	/* Get the value of the property */
	JSI_SET_PROPERTY,	/* Set the value of the property */
	JSI_CREATE_PROPERTY,	/* Create a new property */
	JSI_DELETE_PROPERTY,	/* Delete a property */
	JSI_EXECUTE_PROPERTY,	/* Execute a property */
	JSI_WHAT_MAX
};

typedef struct _jsiMessage {
	int what;
	jsiStringSequence specifiers;
	jsiStringSequence identifiers;
	struct {
		char *data;
		int size;
		int max;
	} contents;
} jsiMessage;

/* An array of strings corresponding to the names of the `what' values */
extern char *jsiWhatNames[JSI_WHAT_MAX];

struct _jsiHandler;
struct _jsiNamedHandler;

/*
 * The jsiHandlerContext is used to hold any information that a jsi message
 * handler might need while processing the message.
 *
 * `handlers' - The handlers list any messages should be dispatched too.
 * `channel' - The channel the message came in on.
 */
typedef struct _jsiHandlerContext {
	struct _jsiNamedHandler **handlers;
	struct _jsiChannel *channel;
} jsiHandlerContext;

/*
 * The jsiHandlerFunctions structure is used to encode the functions a
 * jsiHandler needs to implement.
 *
 * `resolveSpecifier' - Ask the handler to resolve the next specifier pair
 *   and return the handler that can process them.
 * `handleMessage' - Process the given message, then clear it and build
 *   the reply message in the same jsiMessage object.
 * `getSuites' - Adds any jsiPropertyInfo's describing this handler.
 */
typedef struct _jsiHandlerFunctions {
	struct _jsiHandler *(*resolveSpecifier)(struct _jsiHandler *,
						jsiMessage *,
						char *,
						char *);
	char *(*handleMessage)(struct _jsiHandlerContext *hc,
			       struct _jsiHandler *, jsiMessage *);
	int (*getSuites)(struct _jsiHandler *, jsiMessage *);
} jsiHandlerFunctions;

/*
 * A jsiHandler is used to process and reply to incoming jsiMessages.  Of
 * course, this structure reall only needs to be used at the top level, other
 * handler designs can be implemented at other levels.
 *
 * `next' - Points to the next handler in a list.
 * `funcs' - Refers to the implementation functions for this object.
 */
typedef struct _jsiHandler {
	struct _jsiHandler *next;
	jsiHandlerFunctions *funcs;
} jsiHandler;

/*
 * A jsiNamedHandler is a convenience for building a handler that is specified
 * only by a single name.  It is also used for the set of root handlers in
 * the VM, so the first specifier will always resolve to a jsiNamedHandler.
 *
 * `handler' - The root jsiHandler structure
 * `name' - The name of this handler
 * `children' - Any children handlers that are to be selected by extra
 *   specifiers.
 */
typedef struct _jsiNamedHandler {
	jsiHandler handler;
	char *name;
	struct _jsiNamedHandler *children;
} jsiNamedHandler;

enum {
	FHB_READ,
	FHB_WRITE,
};

enum {
	FHF_READ = (1L << FHB_READ),
	FHF_WRITE = (1L << FHB_WRITE),
};

/*
 * A jsiFileHandler can be used to implement a handler for a standard file.
 *
 * `handler' - Standard jsiHandler for this file object.
 * `fd' - The file descriptor for the opened file.
 * `flags' - Holder for the above FHF_* flags.
 */
typedef struct _jsiFileHandler {
	jsiNamedHandler handler;
	int fd;
	unsigned long flags;
} jsiFileHandler;

/*
 * Based on the given format create a provider of the appropriate type and
 * tell it to attach to the specified location.  For example, a format of
 * `tcp:localhost/8192' will create a TCP provider thats attached to port
 * 8192.
 *
 * `format' - The format of the connection provider to create.
 * `defproto' - The default protocol to use if the format doesn't specify.
 * `ch' - The function to call whenever a new connection is made.
 */
jsiProvider *jsiAttachProviderTo(char *format,
				 char *defproto,
				 jsiConnectionHandler *ch);
/*
 * Destruct a jsiProvider object
 *
 * `jp' - A provider object.
 */
void jsiDeleteProvider(jsiProvider *jp);
/*
 * Detach the provider from the whatever port is was listening on.
 *
 * `jp' - A provider object.
 */
void jsiDetachProvider(jsiProvider *jp);
/*
 * Loop constantly waiting for connections and calling the `handleConnection'
 * function in the jsiProvider object.
 *
 * `jp' - An attached provider object.
 */
void jsiHandleProvider(jsiProvider *jp);

/*
 * Tags for jsiConnectChannel, these are all the ways to connect a channel to
 * a socket.
 */
enum {
	JCA_TAG_DONE,	/* Terminate the list */
	JCA_Formatted,	/*
			 * (char *) - A string of the form
			 * `<transport>:<host>/<port>'.
			 */
	JCA_Encoded,	/*
			 * (char *) - An encoded channel object that is already
			 * connected.
			 */
	JCA_Protocol,	/*
			 * (int) - A JCP_* value representing the transport
			 * protocol.
			 */
	JCA_Host,	/* (char *) - The name of the remote host */
	JCA_Port,	/* (int) - The remote port */
	JCA_SockAddr,	/*
			 * (struct sockaddr_in *) - An initialized sockaddr_in
			 * address to connect to
			 */
	JCA_Socket,	/*
			 * (int) - A socket file descriptor, most likely gotten
			 * from a call to `accept'.
			 */
	JCA_FileName,	/*
			 * (char *) - The base name for a file.
			 */
	JCA_Create,	/*
			 * (int) - Flag indicating that the files should be
			 * created.
			 */
	JCA_TAG_MAX,
};

/*
 * Create and connect a jsiChannel to the given address.  If `defproto' is non
 * NULL then it is used if one isn't specified in the format.
 *
 * `format' - A string used to determine the remote host that the channel
 *  should connect to.  The format of the string is the name of the protocol
 *  followed by a colon and any protocol specific data, or if `defproto' is
 *  non-NULL, then just the protocol specific data.   For example, the format
 *  "tcp:localhost/8192" might be used to specify a TCP connection to the
 *  localhost port number 8192.
 * `defproto' - The default protocol to use in case its not specified in
 *  `format'.
 * returns - A connected jsiChannel or NULL if the connection attempt failed.
 */
jsiChannel *jsiConnectChannelTo(char *format, char *defproto);
/*
 * Reconnect a channel thats been encoded into a string.  This is useful if
 * one program wants to handoff a channel without having to do an explicit
 * disconnect and reconnect.  (Basically, we have to manually do what an exec()
 * would handle automatically).
 *
 * `encodedChannel' - A string as returned by jsiEncodeChannel.
 * returns - A reconnected jsiChannel or NULL if there was a problem.
 */
jsiChannel *jsiReconnectChannel(char *encodedChannel);
/*
 * Disconnect the channel from its peer.
 *
 * `jc' - A jsiChannel to disconnect.
 */
void jsiDisconnectChannel(jsiChannel *jc);
/*
 * Destruct and free the jsiChannel object.  Also disconnects the socket.
 *
 * `jc' - A jsiChannel object.
 */
void jsiDeleteChannel(jsiChannel *jc);
/*
 * Encode a channel into a string so that another program can reconnect to the
 * peer channel.  If the encoding is successful then the given channel is
 * deleted without causing the peer to be disconnected.
 *
 * `jc' - A connected jsiChannel.
 * returns - A jsiMalloc'ed text string that encodes the current state of
 *  the channel.
 */
char *jsiEncodeChannel(jsiChannel *jc);
/*
 * Send a flattened jsiMessage to the peer.
 *
 * `jc' - The channel to send the message over.
 * `buffer' - The flattened message to send.  NOTE: If this data isn't a
 *  flattened message it should atleast have a four byte network order integer
 *  indicating the size of the message since some channel protocols might
 *  need this to find the bounds of the message.
 * `bufferSize' - The size of `buffer'.
 * returns - True if the send worked, false otherwise.
 */
int jsiSendMessage(jsiChannel *jc, char *buffer, int bufferSize);
/*
 * Send the message the given message and wait for its reply, returning the
 * reply on success.  The passed in jsiMessage may be consumed or may be
 * returned as the reply.
 *
 * `hc' - The set of handlers that should handle any interim messages and the
 *  channel that messages will be communicated over.
 * `jm' - The initial message to send over the channel.
 * returns - The reply message or NULL if something went wrong.  NOTE: This
 *  may be equal to `jm'.
 */
jsiMessage *jsiDoMessage(jsiHandlerContext *hc, jsiMessage *jm);
/*
 * Send a reply message back to the requester.
 *
 * `jc' - The channel to send the reply over.
 * `buffer' - The flattened message to send.  NOTE: If this data isn't a
 *  flattened message it should atleast have a four byte network order integer
 *  indicating the size of the message since some channel protocols might
 *  need this to find the bounds of the message.
 * `bufferSize' - The size of `buffer'.
 * returns - True if the send worked, false otherwise.
 */
int jsiReplyMessage(jsiChannel *jc, char *buffer, int bufferSize);
/*
 * Block waiting for a message on the channel.  A jsiMalloc'ed buffer is
 * returned and the size of the buffer is set in `bufferSize'.
 *
 * `jc' - The channel to wait for messages on.
 * `bufferSize' - A pointer to an integer to hold the size of the returned
 *  message.
 * returns - A jsiMalloc'ed buffer containing the next message on the channel
 *  or NULL if something went wrong.
 */
char *jsiRecvMessage(jsiChannel *jc, int *bufferSize);
/*
 * Waits for a reply message, but also dispatches any interim messages that
 * can come in.  For example, when a message is sent the other application
 * might need to send requests back to get additional information to help
 * handle the original request.
 *
 * `hc' - The set of handlers to handle any messages and the channel to send
 *  replies over.
 * returns - A reply or error message, or NULL if something went wrong.
 */
jsiMessage *jsiHandleInterimMessages(jsiHandlerContext *hc);
/*
 * Print out the current state of the jsiChannel object.
 * (For debugging purposes only.)
 */
void jsiPrintChannel(FILE *file, jsiChannel *jc);

/* Tags for jsiCreatePropertyInfo */
enum {
	JPIA_TAG_DONE,		/* Terminate the list */
	JPIA_Internal,		/* (char *) Internally formatted string. */
	JPIA_Name,		/* (char *) The name of the property. */
	JPIA_Commands,		/* (int) The commands the property supports. */
	JPIA_Specifiers,	/* (char *) List of specifiers required. */
	JPIA_Usage,		/* (char *) Description of the property. */
	JPIA_TAG_MAX,
};

/*
 * Dynamically create a jsiPropertyInfo object.  The object is constant, so
 * its value is set at creation only.  Usually, jsiPropertyInfo's are allocated
 * statically, so this is only really used to instantiate them from messages.
 *
 * `tag, ...' - A tag list describing the property.
 */
jsiPropertyInfo *jsiCreatePropertyInfo(int tag, ...);
/*
 * Destruct and free the jsiPropertyInfo.
 */
void jsiDeletePropertyInfo(jsiPropertyInfo *jpi);
/*
 * Find a jsiPropertyInfo in a null terminated array of jsiPropertyInfo's.
 *
 * `jpi' - A NULL terminated array of jsiPropertyInfo's.
 * `name' - The name of the property to search for in `jpi'.
 * returns - The index in `jpi' where the `name' property was found or -1 if
 *  the search failed.
 */
int jsiFindProperty(jsiPropertyInfo *jpi, char *name);
/*
 * Compute the amount of memory required to encode this jsiPropertyInfo as
 * a string.
 *
 * `jpi' - An initialized jsiPropertyInfo object.
 * returns - The amount of memory needed to encode `jpi' as a string.
 */
int jsiComputePropertyInfoSize(jsiPropertyInfo *jpi);
/*
 * Encode the jsiProperyInfo object as a string in the given buffer.
 *
 * `jpi' - An initialized jsiPropertyInfo object.
 * `buffer' - A buffer of atleast jsiComputePropertyInfo(jpi) size.
 */
void jsiEncodePropertyInfo(jsiPropertyInfo *jpi, char *buffer);
/*
 * Print out a jsiPropertyInfo object.
 * (For debugging purposes only)
 */
void jsiPrintProperty(FILE *file, jsiPropertyInfo *pi);

/* Tags for adding/getting values in the jsiMessage */
enum {
	JMA_TAG_DONE,
	JMA_What,
	JMA_Character,
	JMA_Integer,
	JMA_UnsignedInteger,
	JMA_Float,
	JMA_Pointer,
	JMA_String,
	JMA_Bytes,
	JMA_PropertyInfo,
	JMA_PropertyList,
	JMA_FormattedString,
	JMA_Specifier,
	JMA_TAG_MAX,
};

/* XXX Arbitrary message size maximum. */
#define JSI_MAX_MESSAGE_SIZE 32768

/*
 * returns - An empty jsiMessage.
 */
jsiMessage *jsiCreateMessage(void);
/*
 * Create and add values to a message, if all goes well then the constructed
 * message is returned.  For example, the following call:
 *
 * jm = jsiBuildMessage(JMA_What, JSI_EXECUTE_PROPERTY,
 *			JMA_Specifier, JMA_String, "write", "",
 *			JMA_Specifier, JMA_String, "out", "",
 *			JMA_Integer, "size", size,
 *			JMA_TAG_DONE)
 *
 * would construct a message with the specifiers "write" and "out", and the
 * identifier "size".
 *
 * `tag, ...' - A set of tags describing the message to be built.
 * returns - The message with the contents specified in the tag list or NULL
 *  if there wasn't enough memory.
 */
jsiMessage *jsiBuildMessage(int tag, ...);
/*
 * Destruct and free a jsiMessage
 */
void jsiDeleteMessage(jsiMessage *jm);
/*
 * Clear the contents of a message.  Effectively, puts it in the same state
 * as a jsiMessage returned by jsiCreateMessage.
 */
void jsiClearMessage(jsiMessage *jm);
/*
 * Push a specifier onto the message's specifier stack.  The `tag' should be
 * followed by a string indicating the specifiers name, and then by the
 * data value whose type is determined by the value of `tag'.
 */
int jsiPushMessageSpecifier(jsiMessage *jm, int tag, ...);
/*
 * Pop a specifier from the messages stack.
 *
 * `jm' - The message to pop a specifier off of.
 * `out_name' - A pointer to a pointer that will be set with the name of the
 *  specifier at the top of the stack.
 * `out_value' - A pointer to a pointer that will be set with the value of the
 *  specifier at the top of the stack.
 * returns - True if there was a specifier on the stack or false if the stack
 *  was empty.
 */
int jsiPopMessageSpecifier(jsiMessage *jm, char **out_name, char **out_value);
/*
 * Restore a popped specifier back onto the stack.  This is possible since
 * jsiPopMessageSpecifier doesn't damage the popped specifiers.  Note:  This
 * function expects to be called only if there were previously popped
 * specifiers.
 */
void jsiRestoreMessageSpecifier(jsiMessage *jm);
/*
 * Add a list of values to the message.  The tag indicates the type of data
 * to be added, followed by a name string, and finally the data value.  Since,
 * the values eventually end up as strings in the message you don't have to
 * worry about converting data before hand.  Identifier names don't have to
 * be unique since they are further specified with an instance number upon
 * retrieval.  This is handy for sending single dimension arrays.
 *
 * `jm' - The message to add identifers too.
 * `tag, ...' - A tag list describing what to add to `jm'.
 * returns - True if the additions worked, false otherwise.
 */
int jsiMessageAdd(jsiMessage *jm, int tag, ...);
/*
 * Remove a specific instance of an identifier with the given name.
 *
 * `jm' - The message to remove the identifier from.
 * `name' - The name of the identifier to remove.
 * `instance' - The instance of `name' in `jm' to remove.
 * returns - True if the identifier was found and removed, false otherwise.
 */
int jsiMessageRem(jsiMessage *jm, char *name, int instance);
/*
 * Count the number of instances of a specific identifier.
 *
 * `jm' - The message possibly holding the identifier.
 * `name' - The name of the identifier to count.
 * returns - the number instances in the message for the given name.
 */
int jsiMessageIdentifierCount(jsiMessage *jm, char *name);
/*
 * Get the identifier at the given index and its value.
 *
 * `jm' - The message to get the identifier from.
 * `index' - The index of the identifier to retrieve.  Note: This needs to be
 *  a number less than jm->identifiers.count.
 * `out_name' - A pointer to a pointer that will be set with the name of the
 *  identifier.
 * `out_value' -  A pointer to a pointer that will be set with the value of the
 *  identifier.
 */
void jsiMessageGetIdentifier(jsiMessage *jm,
			     int index, char **out_name, char **out_value);
/*
 * Get a list of values from a message.  The tag indicates the type of data
 * you're expecting, the function will then try to parse the internal string
 * to get this data.  Next, is the name of the value and its instance number,
 * followed by a pointer to where the value should be stored.  Strings aren't
 * copied out so they are only valid as long as the message is.  However,
 * any jsiPropertyInfo objects returned are allocated so you'll need to free
 * them after use.
 *
 * returns - True if all the requested identifiers were found, false otherwise.
 */
int jsiMessageGet(jsiMessage *jm, int tag, ...);
/*
 * Flatten a message into a byte array of size `out_bufferSize'.  The buffer
 * returned is still owned by the message, in fact it is just the contents.data
 * array in the message, so it will be free'd when the message is deleted.
 *
 * `jm' - The message to flatten.
 * `out_bufferSize' - The number of bytes in the flattened message.
 * returns - The flattened message.
 */
char *jsiFlattenMessage(jsiMessage *jm, int *out_bufferSize);
/*
 * Restore a flattened message so you can continue to manipulate it.
 *
 * `jm' - A previously flattened jsiMessage.
 */
void jsiRestoreMessage(jsiMessage *jm);
/*
 * Instantiate a message object from a flat, jsiMalloc'ed buffer.  Ownership
 * of the buffer is passed to the message so you're not allowed to modify it.
 * Also, this function will jsiFree the buffer if it fails to instantiate the
 * message.
 */
jsiMessage *jsiInstantiateMessage(char *buffer, int bufferSize);
jsiMessage *jsiInstantiateMessageNoFree(char *buffer, int bufferSize);
/*
 * Print out the state of a jsiMessage.
 * (For debugging purposes only.)
 */
void jsiPrintMessage(FILE *file, jsiMessage *jm);

/*
 * Function tags for jsiInitNamedHandler
 */
enum {
	JNA_TAG_DONE,		/* Terminate the list */
	JNA_Functions,		/*
				 * (jsiHandlerFunctions *) - Structure to store
				 * the handler functions.
				 */
	JNA_Name,		/* (char *) - The name of the handler */
	JNA_ResolveSpecifier,	/* (void *) - The resolve specifier function */
	JNA_HandleMessage,	/* (void *) - The message handler function */
	JNA_GetSuites,		/* (void *) - The get suites function */
	JNA_Child,		/*
				 * (jsiNamedHandler *) - A child handler to be
				 * added to this parent.
				 */
	JNA_TAG_MAX
};

/*
 * Initialize a named handler.
 *
 * `jnh' - The handler to initialize.
 * `tag, ...' - A tag list describing how the handler should be initialized.
 * returns - The `jnh' object, useful when initializing trees of handlers.
 */
jsiNamedHandler *jsiInitNamedHandler(jsiNamedHandler *jnh, int tag, ...);
/*
 * Install a root handler, the top specifier of any message will be matched
 * against this set of root handlers when dispatching messages.
 *
 * `list' - The handler list that `jh' should be put on.
 * `jh' - An initialized handler.
 */
void jsiInstallHandler(jsiNamedHandler **list, jsiNamedHandler *jh);
/*
 * Uninstall a root handler
 *
 * `list' - The handler list to remove `jh' from.
 * `jh' - An installed and initialized handler.
 */
void jsiUninstallHandler(jsiNamedHandler **list, jsiNamedHandler *jh);
/*
 * Dispatch a message, the reply should be encoded directly into the passed
 * in message.  The reply message can then be flattened and sent to the peer.
 *
 * `hc' - The set of handlers that the message should be dispatched to.
 * `jm' - The message to be dispatched.
 */
void jsiDispatchMessage(jsiHandlerContext *hc, jsiMessage *jm);

/* The default implementations of the jsiNamedHandler functions */
jsiHandler *namedResolveSpecifier(jsiHandler *jh, jsiMessage *jm,
				  char *name, char *value);
char *namedHandleMessage(jsiHandlerContext *hc,
			 jsiHandler *jh, jsiMessage *jm);
int namedGetSuites(jsiHandler *jh, jsiMessage *jm);
extern jsiHandlerFunctions namedHandlerFunctions;

/*
 * The registry functions provide a way for programs to get information about
 * each other.  Basically, it just manages a bunch of files which can be used
 * to figure out how to talk to each other, through jsi possibly, or its just
 * a dumping ground for files generated by the programs.
 */

/*
 * The `path' buffer passed to jsiGetRegistryPath should be atleast this size
 * XXX This sucks
 */
#define REGISTRY_PATH_LEN 1024
/*
 * Compute the path for a registered name and sub-name.
 *
 * `path' - An array of atleast REGISTRY_PATH_LEN size that the computed path
 *  will be placed in.
 * `name' - The name of program instance to search for.
 * `export' - The export of the `name' to search for.
 * `create' - A flag indicating whether to create the directory path that is
 *  computed.
 */
void jsiGetRegistryPath(char *path, char *name, char *export, int create);
/*
 * Register an instance name
 *
 * `name' - The name of the program instance to register.
 */
void jsiRegisterName(char *name);
/*
 * Unregister an instance name
 *
 * `name' - The name of the program instance to unregister.
 */
void jsiUnregisterName(char *name);
/*
 * Export some data in the given instance and export name.
 *
 * `name' - The name of the program instance that is exporting something.
 * `export' - The name of the export.
 * `fmt' - A printf like format string which will be the value of the export.
 */
void jsiRegisterExport(char *name, char *export, char *fmt, ...);
/*
 * va_list version of the above function.
 */
void jsiVRegisterExport(char *name, char *export, char *fmt, va_list args);
/*
 * Unregister an export
 *
 * `name' - The name of the program instance that is revoking the export.
 * `export' - The name of the export to revoke.
 */
void jsiUnregisterExport(char *name, char *export);
/*
 * Find an export and get its value.
 *
 * `name' - The name of the program instance that exported something.
 * `export' - The name of the export.
 * returns - The jsiMalloc'ed string containing the value of the export or
 *  NULL if something went wrong.
 */
char *jsiFindExport(char *name, char *export);
/*
 * Test if an export exists.
 *
 * `name' - The name of the program instance that exported something.
 * `export' - The name of the export.
 * returns - The jsiMalloc'ed string containing the value of the export or
 *  NULL if something went wrong.
 */
int jsiExportExists(char *name, char *export);
#define JSI_INTERFACE_EXPORT "jsi.id"
/*
 * Register a JSI interface with the given name, transport, and transport
 * specific address.
 *
 * `name' - The name of the program instance.
 * `transport' - The protocol used to communicate with the program instance.
 * `tid' - The protocol specific data used to describe how to talk to the
 *  program.
 */
void jsiRegisterInterface(char *name, char *transport, char *tid);
/*
 * Unregister a JSI interface instance.
 */
void jsiUnregisterInterface(char *name);
/*
 * Find an interface and return its formatted transport address in a
 * jsiMalloc'ed buffer.
 *
 * `name' - The name of the program instance.
 * returns - A jsiMalloc'ed string that can be passed to jsiConnectChannelTo.
 */
char *jsiFindInterface(char *name);

/* Handler functions that work with regular I/O files */
extern char *jsiHandleFileIOMessage(jsiHandlerContext *hc,
				    jsiHandler *jh, jsiMessage *jm);

extern jsiHandler *jsiResolveFileIOSpecifier(jsiHandler *jh, jsiMessage *jm,
					     char *name, char *data);

extern char *jsiHandleFileMessage(jsiHandlerContext *hc,
				  jsiHandler *jh, jsiMessage *jm);

extern jsiHandler *jsiResolveFileSpecifier(jsiHandler *jh, jsiMessage *jm,
					   char *name, char *data);

int jsiGetFileSuites(jsiHandler *jh, jsiMessage *jm);

#endif /* _jsi_h */
