/*
 * jsiChannel.c
 * Functions specific to jsiChannels
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsi.h"
#include "jsiImpl.h"
#include "jsiStdChannel.h"
#include "jsiPlexChannel.h"

/* The array jsiChannel implementations */
static
jsiChannelFunctions *jsiChannelImplementations[] = {
	&jsiStdChannelImplementation,
	&jsiPlexChannelImplementation,
	0
};

void jsiPrintChannelImplementations(FILE *file)
{
	int lpc;

	require(file != 0);
	
	fprintf(file, "Channels:\n");
	for( lpc = 0; jsiChannelImplementations[lpc]; lpc++ )
	{
		fprintf(file,
			"  %s\n",
			jsiChannelImplementations[lpc]->names[0]);
	}
}

jsiChannelFunctions *jsiFindChannelImplementation(parsedString *protocol)
{
	jsiChannelFunctions *cf, *retval = 0;
	int lpc, lpc2;

	require(protocol != 0);
	require(protocol->data != 0);
	
	/* Find the matching implementation */
	for( lpc = 0; jsiChannelImplementations[lpc] && !retval; lpc++ )
	{
		cf = jsiChannelImplementations[lpc];
		/* Check the implementation aliases */
		for( lpc2 = 0;
		     !retval && cf->names[lpc2];
		     lpc2++ )
		{
			if( !cmpPStrStr(protocol, cf->names[lpc2]) )
			{
				require(cf->create != 0);
				require(cf->delete != 0);
				require(cf->connect != 0);
				
				retval = cf;
			}
		}
	}
	return( retval );
}

jsiChannel *jsiConnectChannelTo(char *format, char *defproto)
{
	parsedString protocol, rest;
	jsiChannel *retval = 0;
	int parsed;

	require(format != 0);

	if( (parsed = parseString(format,
				  SPO_NonEmptyString, &protocol,
				  SPO_Expect, ":",
				  SPO_String, &rest,
				  SPO_End)) ||
	    defproto )
	{
		jsiChannelFunctions *cf = 0;
		jsiChannel *jc;
		
		if( !parsed )
		{
			/* Use the default protocol */
			protocol.data = defproto;
			protocol.len = strlen(defproto);
		}
		if( (cf = jsiFindChannelImplementation(&protocol)) )
		{
			if( (jc = cf->create()) )
			{
				if( cf->connect(jc,
						JCA_Formatted, format,
						JCA_TAG_DONE) )
				{
					retval = jc;
				}
				else
				{
					jsiDeleteChannel(jc);
					jsiLog("Unable to connect to "
					       "destination: %s",
					       format);
				}
			}
		}
		else
		{
			jsiLog("No implementation for protocol `%s'",
			       protocol.data);
		}
	}
	return( retval );
}

jsiChannel *jsiReconnectChannel(char *encodedChannel)
{
	parsedString protocol, rest;
	jsiChannel *retval = 0;

	require(encodedChannel != 0);

	if( parseString(encodedChannel,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_String, &rest,
			SPO_End) )
	{
		jsiChannelFunctions *cf = 0;
		jsiChannel *jc;
		
		if( (cf = jsiFindChannelImplementation(&protocol)) )
		{
			if( (jc = cf->create()) )
			{
				if( cf->connect(jc,
						JCA_Encoded, encodedChannel,
						JCA_TAG_DONE) )
				{
					retval = jc;
				}
				else
				{
					jsiDeleteChannel(jc);
					jsiLog("Unable to reconnect to "
					       "destination: %s",
					       encodedChannel);
				}
			}
		}
		else
		{
			jsiLog("No implementation for protocol `%s'",
			       protocol.data);
		}
	}
	return( retval );
}

void jsiDisconnectChannel(jsiChannel *jc)
{
	require(jc != 0);
	require(jc->ops != 0);
	
	jc->ops->disconnect(jc);
}

void jsiDeleteChannel(jsiChannel *jc)
{
	if( jc )
	{
		require(jc->ops != 0);
		
		jc->ops->delete(jc);
	}
}

char *jsiEncodeChannel(jsiChannel *jc)
{
	char *retval = 0;
	
	require(jc != 0);
	require(jc->ops != 0);

	if( jc->ops->encode )
	{
		retval = jc->ops->encode(jc);
	}
	return( retval );
}

int jsiSendMessage(jsiChannel *jc, char *buffer, int bufferSize)
{
	require(jc != 0);
	require(jc->ops != 0);
	require(jc->ops->sendMessage != 0);
	require(buffer != 0);
	require(bufferSize > 0);
	
	return( jc->ops->sendMessage(jc, buffer, bufferSize) );
}

jsiMessage *jsiDoMessage(jsiHandlerContext *hc, jsiMessage *jm)
{
	require(hc != 0);
	require(hc->channel != 0);
	require(hc->channel->ops != 0);
	require(hc->channel->ops->doMessage != 0);
	require(jm != 0);
	
	return( hc->channel->ops->doMessage(hc, jm) );
}

int jsiReplyMessage(jsiChannel *jc, char *buffer, int bufferSize)
{
	require(jc != 0);
	require(jc->ops != 0);
	require(jc->ops->replyMessage != 0);
	require(buffer != 0);
	require(bufferSize > 0);
	
	return( jc->ops->replyMessage(jc, buffer, bufferSize) );
}

char *jsiRecvMessage(jsiChannel *jc, int *bufferSize)
{
	require(jc != 0);
	require(jc->ops != 0);
	require(jc->ops->recvMessage != 0);
	require(bufferSize != 0);

	return( jc->ops->recvMessage(jc, bufferSize) );
}

jsiMessage *jsiHandleInterimMessages(jsiHandlerContext *hc)
{
	int bufferSize, done = 0;
	jsiMessage *retval = 0;
	char *buffer;

	require(hc != 0);
	require(hc->channel != 0);

	/* Wait for a message as long as there isn't a reply or an error. */
	while( !done && (buffer = jsiRecvMessage(hc->channel, &bufferSize)) )
	{
		if( (retval = jsiInstantiateMessage(buffer, bufferSize)) )
		{
			switch( retval->what )
			{
			case JSI_ERROR:
			case JSI_REPLY:
				/* Got a response to our original message. */
				done = 1;
				break;
			default:
				jsiDispatchMessage(hc, retval);
				if( (buffer = jsiFlattenMessage(retval,
								&bufferSize)) )
				{
					if( jsiReplyMessage(hc->channel,
							    buffer,
							    bufferSize) )
					{
					}
					else
					{
						done = 1;
						jsiLog("Can't send reply to "
						       "interim message");
					}
				}
				else
				{
					done = 1;
					jsiLog("Can't flatten reply to "
					       "interim message");
				}
				jsiDeleteMessage(retval);
				retval = 0;
				break;
			}
		}
		else
		{
			/* jsiInstantiateMessage will free the buffer */
			done = 1;
			jsiLog("Unable to instantiate reply/interim message");
		}
	}
	return( retval );
}
