/*
 * jsiHandler.c
 * Functions for jsiHandlers/jsiNamedHandlers
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsi.h"
#include "jsiImpl.h"

/** BEGIN Default functions for named handlers **/

/* Functions for jsiNamedHandler */
jsiHandlerFunctions namedHandlerFunctions = {
	namedResolveSpecifier,
	namedHandleMessage,
	namedGetSuites
};

jsiHandler *namedResolveSpecifier(jsiHandler *jh, jsiMessage *jm,
				  char *name, char *value)
{
	jsiNamedHandler *jnh = (jsiNamedHandler *)jh, *curr;
	jsiHandler *retval = 0;
	int done = 0;

	require(jh != 0);
	require(jm != 0);
	require(value != 0);

	if( !name )
	{
		/* We're a root handler */
		if( !jsiPopMessageSpecifier(jm, &name, &value) )
		{
			/* No more specifiers, return ourself as the handler */
			retval = jh;
			return( retval );
		}
	}
	/* Try to find a child that matches the given specifier name */
	curr = jnh->children;
	while( curr && !done )
	{
		if( !strcmp(curr->name, name) )
		{
			/* Found a match, pop the next specifier */
			if( jsiPopMessageSpecifier(jm, &name, &value) )
			{
				/*
				 * Ask the child to resolve further specifiers
				 */
				retval = curr->handler.funcs->
					resolveSpecifier(&curr->handler, jm,
							 name, value);
			}
			else
			{
				/*
				 * No more specifiers, the message is directed
				 * at this child.
				 */
				retval = &curr->handler;
			}
			done = 1;
		}
		curr = (jsiNamedHandler *)curr->handler.next;
	}
	if( !retval )
	{
		jsiLog("Couldn't resolve specifier name `%s'", name);
	}
	return( retval );
}

char *namedHandleMessage(jsiHandlerContext *hc, jsiHandler *jh, jsiMessage *jm)
{
	jsiNamedHandler *jnh = (jsiNamedHandler *)jh;
	char *retval = 0;

	require(jh != 0);
	require(jm != 0);
	
	/* Try to provide a reasonable default */
	switch( jm->what )
	{
	case JSI_LIST_PROPERTIES:
		{
			int added = 0, error = 0;
			jsiNamedHandler *curr;

			/* Return a listing of our children */
			curr = jnh->children;
			jsiClearMessage(jm);
			jm->what = JSI_REPLY;
			while( curr && !error )
			{
				error = !jsiMessageAdd(
					jm,
					JMA_String, "result", curr->name,
					JMA_TAG_DONE);
				added = 1;
				curr = (jsiNamedHandler *)curr->handler.next;
			}
			if( !error && !added )
			{
				error = !jsiMessageAdd(
					jm,
					JMA_String, "result", "<empty>",
					JMA_TAG_DONE);
			}
			if( error )
				retval = "No memory";
		}
		break;
	default:
		retval = "Operation not supported";
		break;
	}
	return( retval );
}

int namedGetSuites(jsiHandler *jh, jsiMessage *jm)
{
	return( 1 );
}

/** END Default functions for named handlers **/

/** BEGIN File handler functions **/

jsiPropertyInfo jsiFileRootProperties[] = {
	{ "file", (1L << JSI_CREATE_PROPERTY), "",
	  "name:string - The name of the file to create\n\n"
	  "Create a file\n" },
	{ "file", (1L << JSI_DELETE_PROPERTY), "filename:string",
	  "Delete the given file\n" },
	{ "file", (1L << JSI_LIST_PROPERTIES), "filename:string",
	  "List the properties of the given file.\n" },
	{ "file", (1L << JSI_GET_SUITES), "filename:string",
	  "Get the suites supported by the given file.\n" },
	{ 0 }
};

jsiPropertyInfo jsiFileProperties[] = {
	{ "size", (1L << JSI_GET_PROPERTY), "",
	  "Get the size of the file\n" },
	{ "read", (1L << JSI_EXECUTE_PROPERTY), "",
	  "specifier:string - An array of specifier strings that\n"
	  "correspond to an object on the sending side that can\n"
	  "handle a write method (Default: `out')\n"
	  "specifier-arg:string - The argument for the specifier\n"
	  "at the same index (Default: `')\n"
	  "offset:int - The position to start reading\n"
	  "size:int - If given the number of characters to read out\n"
	  "otherwise it will execute the write method of specifiers\n\n"
	  "Reads and returns the specified number of bytes from the file\n"
	  "or sends the data to the specified object.\n" },
	{ "write", (1L << JSI_EXECUTE_PROPERTY), "",
	  "specifier:string - An array of specifier strings that\n"
	  "correspond to an object on the sending side that can\n"
	  "handle a read method (Default: `in')\n"
	  "specifier-arg:string - The argument for the specifier\n"
	  "at the same index (Default: `')\n"
	  "data:bytes - The data to write to the file\n"
	  "size:int - The number of bytes of `data' to write out\n\n"
	  "Writes the given data to the file or reads the data from\n"
	  "the specified object\n" },
	{ 0 }
};

/* Helper function that pushes specifiers given by attributes */
static
char *pushFileSpecifiers(jsiMessage *dest, jsiMessage *src,
			 char *spec_name, char *spec_data,
			 char *def_name, char *def_data)
{
	char *retval = 0;
	int error = 1;
	
	if( jsiPushMessageSpecifier(dest,
				    JMA_String, spec_name, spec_data,
				    JMA_TAG_DONE) )
	{
		int lpc;

		error = 0;
		for( lpc = 0;
		     !error &&
		     jsiMessageGet(
			src,
			JMA_String, "specifier", lpc, &spec_name,
			JMA_String, "specifier-arg", lpc, &spec_data,
			JMA_TAG_DONE);
		     lpc++ )
		{
			error = !jsiPushMessageSpecifier(
				dest,
				JMA_String, spec_name, spec_data,
				JMA_TAG_DONE);
		}
		if( !error && (lpc == 0) )
		{
			/* Push the default speicifiers on */
			error = !jsiPushMessageSpecifier(
				dest,
				JMA_String, def_name, def_data,
				JMA_TAG_DONE);
		}
	}
	if( error )
	{
		retval = "No memory";
	}
	return( retval );
}

static
char *doReadMsg(jsiHandlerContext *hc, jsiMessage *jm,
		int fd, char *buffer, int size)
{
	int rc = 0, actual;
	char *retval = 0;

	/* Read into the buffer, which is in the jsiMessage. */
	while( !retval &&
	       ((rc = jsiRead(fd, buffer, size, &actual)) == 0) &&
	       (actual > 0) )
	{
		if( jsiMessageAdd(jm,
				  JMA_Integer, "size", actual,
				  JMA_TAG_DONE) )
		{
			char *msg_buffer;
			int msg_size;

			if( (msg_buffer = jsiFlattenMessage(jm, &msg_size)) )
			{
				if( jsiSendMessage(hc->channel,
						   msg_buffer,
						   msg_size) )
				{
					jsiMessage *reply;

					/* Wait for the reply */
					if( (reply =
					     jsiHandleInterimMessages(hc)) &&
					    (reply->what == JSI_REPLY) )
					{
						/*
						 * Restore the message to pre
						 * flattened state.
						 */
						jsiRestoreMessage(jm);
						jsiMessageRem(jm, "size", 0);
						ensure(!jsiMessageGet(jm,
								      JMA_Integer, "size", 0, &actual,
								      JMA_TAG_DONE));
					}
					else
					{
						retval = "Read failed";
					}
					/*
					 * We only care about the `what' part
					 * of the reply.
					 */
					jsiDeleteMessage(reply);
				}
				else
				{
					retval = "Send failed";
				}
			}
			else
			{
				retval = "No memory";
			}
		}
		else
		{
			retval = "No memory";
		}
	}
	if( rc )
		retval = "Read error";
	return( retval );
}

static
char *readData(jsiHandlerContext *hc, jsiMessage *jm, int fd)
{
	int offset = -1, size, actual;
	char *buffer, *retval = 0;
	jsiMessage *io_msg;

	if( jsiMessageGet(jm,
			  JMA_Integer, "size", 0, &size,
			  JMA_TAG_DONE) )
	{
		jsiMessageGet(jm,
			      JMA_Integer, "offset", 0, &offset,
			      JMA_TAG_DONE);
		/* `size' is specified, return the amount they requested. */
		jsiClearMessage(jm);
		/* Add an attribute to the message which will hold the data */
		if( jsiMessageAdd(jm,
				  JMA_What, JSI_REPLY,
				  JMA_Bytes, "data", 0, size,
				  JMA_TAG_DONE) )
		{
			off_t o_actual;
			
			/*
			 * Get the address of the `data' attribute in the
			 * message so we can read the data into it.
			 */
			jsiMessageGet(jm,
				      JMA_Bytes, "data", 0, &buffer, 0,
				      JMA_TAG_DONE);
			if( (offset == -1) ||
			    (jsiSeek(fd, offset, SEEK_SET, &o_actual) == 0) )
			{
				if( jsiRead(fd, buffer, size, &actual) == 0 )
				{
					/* Add the actual amount read */
					if( jsiMessageAdd(
						jm,
						JMA_Integer, "actual", actual,
						JMA_TAG_DONE) )
					{
					}
					else
					{
						retval = "No memory";
					}
				}
				else
				{
					retval = "Read error";
				}
			}
			else
			{
				retval = "Seek error";
			}
		}
		else
		{
			retval = "No memory/read size too large";
		}
	}
	else if( (io_msg = jsiCreateMessage()) )
	{
		/*
		 * No size specified, so we read the file and send to their
		 * specified handler, or the default of out.write().
		 */
		if( !(retval = pushFileSpecifiers(io_msg, jm,
						  "write", "",
						  "out", "")) )
		{
			/*
			 * Some reasonable buffer size to add to the message.
			 * XXX We're gonna be sending messages of this size
			 * even though the read might not fill up the entire
			 * space allocated to the attribute.
			 */
			size = 4096;
			if( jsiMessageAdd(io_msg,
					  JMA_What, JSI_EXECUTE_PROPERTY,
					  JMA_Bytes, "data", 0, size,
					  JMA_TAG_DONE) )
			{
				jsiMessageGet(io_msg,
					      JMA_Bytes, "data", 0, &buffer, 0,
					      JMA_TAG_DONE);
				if( !(retval = doReadMsg(hc, io_msg,
							 fd, buffer, size)) )
				{
					/* Everything went OK */
					jsiClearMessage(jm);
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_String, "result", "ok",
						JMA_TAG_DONE);
				}
			}
			else
			{
				retval = "No memory";
			}
		}
		jsiDeleteMessage(io_msg);
	}
	else
	{
		retval = "No memory";
	}
	return( retval );
}

static
char *doWriteMsg(jsiHandlerContext *hc,
		 char *msg_buffer, int msg_size,
		 int fd)
{
	char *retval = 0;
	int done = 0;

	while( !retval && !done )
	{
		/* Send the read request message */
		if( jsiSendMessage(hc->channel, msg_buffer, msg_size) )
		{
			jsiMessage *reply;

			/* Wait for the reply */
			if( (reply = jsiHandleInterimMessages(hc)) &&
			    (reply->what == JSI_REPLY) )
			{
				int size, actual;
				char *buffer;

				/* Get the read data */
				if( jsiMessageGet(
					reply,
					JMA_Bytes, "data", 0, &buffer, &actual,
					JMA_Integer, "actual", 0, &size,
					JMA_TAG_DONE) )
				{
					/* Be paranoid about the size */
					if( (size < 0) || (size > actual) )
						size = actual;
					/* Check if we're done */
					if( size == 0 )
						done = 1;
					/*
					 * Finally, write the data to the file
					 */
					if( jsiWrite(fd,
						     buffer,
						     size,
						     &actual) == 0 )
					{
					}
					else
					{
						retval = "Write failed";
					}
				}
				else
				{
					retval = "Bad reply";
				}
			}
			else
			{
				retval = "Read failed";
			}
			jsiDeleteMessage(reply);
		}
		else
		{
			retval = "Send failed";
		}
	}
	return( retval );
}

static
char *writeData(jsiHandlerContext *hc, jsiMessage *jm, int fd)
{
	int offset = -1, size = -1, actual;
	char *buffer, *retval = 0;
	jsiMessage *out_msg;

	if( jsiMessageGet(jm,
			  JMA_Bytes, "data", 0, &buffer, &actual,
			  JMA_TAG_DONE) )
	{
		off_t o_actual;
		
		/* The data to write was specified */
		jsiMessageGet(jm,
			      JMA_Integer, "offset", 0, &offset,
			      JMA_Integer, "size", 0, &size,
			      JMA_TAG_DONE);
		if( (size < 0) || (size > actual) )
		{
			size = actual;
		}
		if( (offset == -1) ||
		    (jsiSeek(fd, offset, SEEK_SET, &o_actual) == 0) )
		{
			/* Write it to the file */
			if( jsiWrite(fd, buffer, size, &actual) == 0 )
			{
				require(size == actual); // XXX
				
				jsiClearMessage(jm);
				if( jsiMessageAdd(
					jm,
					JMA_What, JSI_REPLY,
					JMA_Integer, "actual", actual,
					JMA_TAG_DONE) )
				{
				}
				else
				{
					retval = "No memory";
				}
			}
			else
			{
				retval = "Write error";
			}
		}
		else
		{
			retval = "Seek error";
		}
	}
	else if( (out_msg = jsiCreateMessage()) )
	{
		/* We'll need to get the data from their end */
		if( !(retval = pushFileSpecifiers(out_msg, jm,
						  "read", "",
						  "in", "")) )
		{
			char *msg_buffer;
			int msg_size;

			/* Send read requests for `size' bytes */
			size = 4096;
			if( jsiMessageAdd(out_msg,
					  JMA_What, JSI_EXECUTE_PROPERTY,
					  JMA_Integer, "size", size,
					  JMA_TAG_DONE) &&
			    (msg_buffer = jsiFlattenMessage(out_msg,
							    &msg_size)))
			{
				if( !(retval = doWriteMsg(hc,
							  msg_buffer,
							  msg_size,
							  fd)) )
				{
					/* Everything worked */
					jsiClearMessage(jm);
					jsiMessageAdd(
						jm,
						JMA_What, JSI_REPLY,
						JMA_String, "result", "ok",
						JMA_TAG_DONE);
				}
			}
			else
			{
				retval = "No memory";
			}
		}
		jsiDeleteMessage(out_msg);
	}
	else
	{
		retval = "No memory";
	}
	return( retval );
}

char *jsiHandleFileIOMessage(jsiHandlerContext *hc,
			     jsiHandler *jh, jsiMessage *jm)
{
	jsiFileHandler *jfh = (jsiFileHandler *)jh;
	char *spec_name, *spec_value;
	char *retval = 0;

	switch( jm->what )
	{
	case JSI_LIST_PROPERTIES:
		jsiClearMessage(jm);
		if( !jsiMessageAdd(jm,
				   JMA_What, JSI_REPLY,
				   JMA_String, "result", "size",
				   JMA_String, "result", "read",
				   JMA_String, "result", "write",
				   JMA_TAG_DONE) )
		{
			retval = "No memory";
		}
		break;
	case JSI_EXECUTE_PROPERTY:
		if( jsiPopMessageSpecifier(jm, &spec_name, &spec_value) )
		{
			if( !strcmp(spec_name, "read") &&
			    (jfh->flags & FHF_READ) )
			{
				retval = readData(hc, jm, jfh->fd);
			}
			else if( !strcmp(spec_name, "write") &&
				 (jfh->flags & FHF_WRITE) )
			{
				retval = writeData(hc, jm, jfh->fd);
			}
			else
			{
				retval = "Invalid Specifier";
			}
		}
		else
		{
			retval = "Not enough specifiers";	
		}
		break;
	case JSI_GET_SUITES:
		jsiClearMessage(jm);
		if( !jsiMessageAdd(
			jm,
			JMA_What, JSI_REPLY,
			JMA_PropertyList, "result", jsiFileProperties,
			JMA_TAG_DONE) )
		{
			retval = "No memory";
		}
		break;
	default:
		retval = "File doesn't support this operation";
		break;
	}
	return( retval );
}

jsiHandler *jsiResolveFileIOSpecifier(jsiHandler *jh, jsiMessage *jm,
				      char *name, char *data)
{
	jsiHandler *retval = jh;

	return( retval );
}

static jsiHandlerFunctions jsiFileFunctions = {
	jsiResolveFileSpecifier,
	jsiHandleFileMessage,
	jsiGetFileSuites
};

char *jsiHandleFileMessage(jsiHandlerContext *hc,
			   jsiHandler *jh, jsiMessage *jm)
{
	char *spec_name, *spec_data;
	char *name, *retval = 0;
	struct stat st;

	switch( jm->what )
	{
	case JSI_GET_PROPERTY:
		/*
		 * We need to restore the specifier and then pop it again so
		 * we can get the file name.
		 */
		jsiRestoreMessageSpecifier(jm);
		jsiPopMessageSpecifier(jm, &spec_name, &spec_data);
		if( jsiStat(spec_data, &st) != 0 )
		{
			retval = "Bad file name";
		}
		else if( !jsiPopMessageSpecifier(jm, &spec_name, &spec_data) )
		{
			retval = "File attribute not specified for `get'";
		}
		else
		{
			/* Figure out which file attribute they want */
			if( !strcmp(spec_name, "size") )
			{
				jsiClearMessage(jm);
				if( !jsiMessageAdd(
					jm,
					JMA_What, JSI_REPLY,
					JMA_Integer, "result", st.st_size,
					JMA_TAG_DONE) )
				{
					retval = "No memory";
				}
			}
			else
			{
				retval = "Bad file attribute";
			}
		}
		break;
	case JSI_CREATE_PROPERTY:
		if( jsiMessageGet(jm,
				  JMA_String, "name", 0, &name,
				  JMA_TAG_DONE) )
		{
			int fd;
			
			jsiClearMessage(jm);
			/* Create the file */
			if( jsiOpen(name,
				    O_RDWR|O_CREAT|O_TRUNC,
				    0600,
				    &fd) == 0 )
			{
				jsiClose(fd);
				if( !jsiMessageAdd(jm,
						   JMA_What, JSI_REPLY,
						   JMA_String, "result", "ok",
						   JMA_TAG_DONE) )
				{
					retval = "No memory";
				}
			}
			else
			{
				retval = "Unable to create file";
			}
		}
		else
		{
			retval = "No file name given";
		}
		break;
	case JSI_DELETE_PROPERTY:
		/*
		 * We need to restore the specifier and then pop it again so
		 * we can get the file name.
		 */
		jsiRestoreMessageSpecifier(jm);
		jsiPopMessageSpecifier(jm, &spec_name, &spec_data);
		if( jsiRemove(spec_data) == 0 )
		{
			jsiClearMessage(jm);
			if( !jsiMessageAdd(jm,
					   JMA_What, JSI_REPLY,
					   JMA_String, "result", "ok",
					   JMA_TAG_DONE) )
			{
				retval = "No memory";
			}
		}
		else
		{
			retval = "Bad file name";
		}
		break;
	case JSI_EXECUTE_PROPERTY:
		jsiRestoreMessageSpecifier(jm);
		jsiPopMessageSpecifier(jm, 0, &name);
		if( jsiPopMessageSpecifier(jm, &spec_name, &spec_data) )
		{
			int flags = 0, valid = 0;
			jsiFileHandler fh;
			
			if( !strcmp(spec_name, "read") )
			{
				flags = O_RDONLY;
				fh.flags = FHF_READ;
				valid = 1;
			}
			else if( !strcmp(spec_name, "write") )
			{
				flags = O_WRONLY;
				fh.flags = FHF_WRITE;
				valid = 1;
			}
			if( valid &&
			    (jsiOpen(name, flags, 0600, &fh.fd) == 0) )
			{
				jsiInitNamedHandler(
					&fh.handler,
					JNA_Functions, &jsiFileFunctions,
					JNA_TAG_DONE);
				/* Pass this off to the File I/O handler */
				jsiRestoreMessageSpecifier(jm);
				retval = jsiHandleFileIOMessage(hc,
								&fh.handler.
								handler,
								jm);
				jsiClose(fh.fd);
			}
			else if( valid )
			{
				retval = "Couldn't open file";
			}
			else
			{
				retval = "Invalid file I/O specifier";
			}
		}
		else
		{
			retval = "No specifier to execute on file";
		}
		break;
	case JSI_GET_SUITES:
		jsiClearMessage(jm);
		if( !jsiMessageAdd(
			jm,
			JMA_What, JSI_REPLY,
			JMA_PropertyList, "result", jsiFileProperties,
			JMA_TAG_DONE) )
		{
			retval = "No memory";
		}
		break;
	default:
		retval = "File doesn't support this operation";
		break;
	}
	return( retval );
}

jsiHandler *jsiResolveFileSpecifier(jsiHandler *jh, jsiMessage *jm,
				    char *name, char *data)
{
	jsiHandler *retval = 0;

	switch( jm->what )
	{
	case JSI_CREATE_PROPERTY:
	case JSI_DELETE_PROPERTY:
	case JSI_GET_SUITES:
	case JSI_GET_PROPERTY:
	case JSI_EXECUTE_PROPERTY:
		/* We can handle this directly */
		retval = jh;
		break;
	default:
		break;
	}
	return( retval );
}

int jsiGetFileSuites(jsiHandler *jh, jsiMessage *jm)
{
	int retval;

	retval = jsiMessageAdd(
		jm,
		JMA_PropertyList, "result", jsiFileRootProperties,
		JMA_TAG_DONE);
	return( retval );
}

jsiNamedHandler *jsiInitNamedHandler(jsiNamedHandler *jnh, int tag, ...)
{
	va_list args;

	require(jnh != 0);
	
	va_start(args, tag);
	/* Walk the tag list filling in the structure */
	while( tag != JNA_TAG_DONE )
	{
		switch( tag )
		{
		case JNA_Functions:
			jnh->handler.funcs =
				va_arg(args, jsiHandlerFunctions *);
			/* Be nice and set the defaults */
			if( !jnh->handler.funcs->resolveSpecifier )
			{
				jnh->handler.funcs->resolveSpecifier =
					namedResolveSpecifier;
			}
			if( !jnh->handler.funcs->handleMessage )
			{
				jnh->handler.funcs->handleMessage =
					namedHandleMessage;
			}
			if( !jnh->handler.funcs->getSuites )
			{
				jnh->handler.funcs->getSuites =
					namedGetSuites;
			}
			break;
		case JNA_Name:
			jnh->name = va_arg(args, char *);
			break;
		case JNA_Child:
			{
				jsiNamedHandler *handler;

				/*
				 * Just put the child at the front of the list
				 */
				handler = va_arg(args, jsiNamedHandler *);
				handler->handler.next =
					&jnh->children->handler;
				jnh->children = handler;
			}
			break;
		case JNA_ResolveSpecifier:
			jnh->handler.funcs->resolveSpecifier =
				va_arg(args, void *);
			break;
		case JNA_HandleMessage:
			jnh->handler.funcs->handleMessage =
				va_arg(args, void *);
			break;
		case JNA_GetSuites:
			jnh->handler.funcs->getSuites = va_arg(args, void *);
			break;
		default:
			require(tag > JNA_TAG_DONE);
			require(tag < JNA_TAG_MAX);
			break;
		}
		tag = va_arg(args, int);
	}
	va_end(args);
	return( jnh );
}

void jsiInstallHandler(jsiNamedHandler **list, jsiNamedHandler *jnh)
{
	require(list != 0);
	require(jnh != 0);
	
	jnh->handler.next = (jsiHandler *)*list;
	*list = jnh;
}

void jsiUninstallHandler(jsiNamedHandler **list, jsiNamedHandler *jnh)
{
	jsiNamedHandler *curr, **prev;

	require(list != 0);
	require(jnh != 0);
	
	prev = list;
	curr = *list;
	while( curr && (curr != jnh) )
	{
		prev = (jsiNamedHandler **)&curr->handler.next;
		curr = (jsiNamedHandler *)curr->handler.next;
	}
	if( curr )
		*prev = (jsiNamedHandler *)curr->handler.next;
}

jsiNamedHandler *jsiFindHandler(jsiNamedHandler **list, char *name)
{
	jsiNamedHandler *curr, *retval = 0;

	require(list != 0);
	require(name != 0);
	
	curr = *list;
	while( curr && !retval )
	{
		if( !strcmp(curr->name, name) )
			retval = curr;
		curr = (jsiNamedHandler *)curr->handler.next;
	}
	return( retval );
}

void jsiDispatchMessage(jsiHandlerContext *hc, jsiMessage *jm)
{
	char *spec_name, *spec_value, *error_msg = 0;

	require(jm != 0);

	if( jsiPopMessageSpecifier(jm, &spec_name, &spec_value) )
	{
		jsiHandler *jh;

		/* Find a matching root handler for the specifier */
		if( (jh = (jsiHandler *)jsiFindHandler(hc->handlers,
						       spec_name)) )
		{
			/* Try to resolve to a more specific handler */
			jh = jh->funcs->resolveSpecifier(jh, jm,
							 0,
							 spec_value);
			if( jh )
			{
				/* Ask the handler to process the msg */
				error_msg = jh->funcs->handleMessage(hc,
								     jh,
								     jm);
				/* jh is unusable now */

				ensure(error_msg ||
				       (jm->what == JSI_REPLY) ||
				       (jm->what == JSI_ERROR));
			}
			else if( jm->what != JSI_ERROR )
			{
				error_msg = "Invalid specifier";
			}
		}
		else
		{
			error_msg = "Unknown root handler";
		}
	}
	else
	{
		/*
		 * There are no specifiers, so either they want a list of the
		 * roots, or its a bad message
		 */
		switch( jm->what )
		{
		case JSI_LIST_PROPERTIES:
			{
				jsiNamedHandler *nh;
				int error = 0;

				jsiClearMessage(jm);
				jm->what = JSI_REPLY;
				nh = *hc->handlers;
				while( nh && !error )
				{
					error = !jsiMessageAdd(
						jm,
						JMA_String, "result", nh->name,
						JMA_TAG_DONE);
					nh = (jsiNamedHandler *)
						nh->handler.next;
				}
				if( error )
					error_msg = "Out of memory";
			}
			break;
		case JSI_GET_SUITES:
			{
				jsiNamedHandler *nh;
				int error = 0;
				
				jsiClearMessage(jm);
				jm->what = JSI_REPLY;
				nh = *hc->handlers;
				while( nh && !error )
				{
					if( nh->handler.funcs->getSuites )
					{
						error = !nh->handler.funcs->
							getSuites(&nh->handler,
								  jm);
					}
					nh = (jsiNamedHandler *)
						nh->handler.next;
				}
				if( error )
					error_msg = "Out of memory";
			}
			break;
		default:
			error_msg = "Operation not supported on the root "
			    "property.";
			break;
		}
	}
	if( error_msg )
	{
		/* Make an error reply message */
		jsiClearMessage(jm);
		jsiMessageAdd(jm,
			      JMA_What, JSI_ERROR,
			      JMA_String, "why", error_msg,
			      JMA_TAG_DONE);
	}
}
