/*
 * jsiMessage.c
 * Functions for manipulating jsiMessages
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsi.h"
#include "jsiImpl.h"

/* Amount to use when allocating extra space in arrays */
#define JSI_ARRAY_GROWTH 3
/* The initial size of a jsiMessage buffer */
#define JSI_BUFFER_INIT_SIZE 1024
/* Amount to grow the buffer when allocating extra space */
#define JSI_BUFFER_GROWTH 512

/* The size of an integer when encoded in the message buffer */
#define SIZEOF_INT 4
/* Handy macro for aligning the size to an integer */
#define ALIGN_INT(size) (((size) + (SIZEOF_INT - 1)) & ~(SIZEOF_INT - 1))

/*
 * A jsiMessageHeader is placed at the head of every message so that we can
 * figure out what its contents are.
 *
 * `messageSize' - The size of the message, including the size of the header.
 * `what' - Same as jsiMessage.what.
 * `specifierCount' - The number of specifiers in the messsage.
 * `identifierCount' - The number of identifiers in the message.
 */
typedef struct _jsiMessageHeader {
	int messageSize;
	int what;
	int specifierCount;
	int identifierCount;
} jsiMessageHeader;

static int jsiAddTag(jsiMessage *jm, jsiStringSequence *ss,
		     int tag, va_list *args_ptr);

#if !defined(NO_PRECONDITIONS) || !defined(NO_POSTCONDITIONS)
static
int jsiMessageInvariant(jsiMessage *jm)
{
	require(jm != 0);
	require(jm->what > JSI_WHAT_NONE);
	require(jm->what < JSI_WHAT_MAX);
	if( jm->specifiers.data )
	{
		require(jm->specifiers.count >= 0);
		require(jm->specifiers.max >= 0);
		require(jm->specifiers.count <= jm->specifiers.max);
	}
	else
	{
		require(jm->specifiers.count == 0);
		require(jm->specifiers.max == 0);
	}
	if( jm->identifiers.data )
	{
		require(jm->identifiers.count >= 0);
		require(jm->identifiers.max >= 0);
		require(jm->identifiers.count <= jm->identifiers.max);
	}
	else
	{
		require(jm->identifiers.count == 0);
		require(jm->identifiers.max == 0);
	}
	require(jm->contents.size >= sizeof(jsiMessageHeader));
	require(jm->contents.size <= jm->contents.max);
	
	return( 1 );
}
#endif

/* Resize the message buffer to the given size */
static
int jsiResizeMessage(jsiMessage *jm, int new_size)
{
	int retval = 1;

	require(jm != 0);
	require(new_size >= 0);
	
	if( new_size > jm->contents.max )
	{
		char *new_data;

		/* We need to reallocate a larger array */
		if( (new_data = jsiRealloc(jm->contents.data,
					   sizeof(char) *
					   (new_size + JSI_BUFFER_GROWTH))) )
		{
			int ptr_diff, lpc;

			/* Adjust any internal pointers */
			ptr_diff = new_data - jm->contents.data;
			for( lpc = 0; lpc < jm->specifiers.count; lpc++ )
			{
				jm->specifiers.data[lpc] += ptr_diff;
			}
			for( lpc = 0; lpc < jm->identifiers.count; lpc++ )
			{
				jm->identifiers.data[lpc] += ptr_diff;
			}
			/* Update the buffer values */
			jm->contents.data = new_data;
			jm->contents.size = new_size;
			jm->contents.max = new_size + JSI_BUFFER_GROWTH;
		}
		else
		{
			retval = 0;
		}
	}
	else
	{
		jm->contents.size = new_size;
	}

	ensure(!retval || (jm->contents.size == new_size));
	ensure(jm->contents.size <= jm->contents.max);
	
	return( retval );
}

/*
 * Append a byte array to the message.  If `buffer' is NULL then we just
 * increase the size of the message, with proper alignment, and don't copy
 * anything.
 */
static
int jsiAppendMessage(jsiMessage *jm, char *buffer, int bufferSize)
{
	int alignSize, start, retval = 0;

	require(jm != 0);
	require(bufferSize > 0);

	/* Make sure the bufferSize is aligned properly */
	alignSize = ALIGN_INT(bufferSize);
	/*
	 * Note that we record the offset since we might realloc and a pointer
	 * saved here would be bogus.
	 */
	start = jm->contents.size;
	if( jsiResizeMessage(jm, jm->contents.size + alignSize) )
	{
		if( buffer )
			memcpy(&jm->contents.data[start], buffer, bufferSize);
		retval = 1;
	}
	return( retval );
}

/*
 * Append a byte array to the message, but we also encode the size of the
 * array before the data
 */
static
int jsiAddData(jsiMessage *jm, char *buffer, int bufferSize)
{
	int size, retval = 1;

	require(jm != 0);
	require(bufferSize > 0);

	size = htonl(bufferSize);
	retval = jsiAppendMessage(jm, (char *)&size, SIZEOF_INT);
	retval = retval && jsiAppendMessage(jm, buffer, bufferSize);
	return( retval );
}

/*
 * Append a string to the message and return the address of the string in the
 * message buffer
 */
static
char *jsiAddString(jsiMessage *jm, char *name)
{
	int offset, hlen, nlen, good = 1;
	char *retval = 0;

	require(jm != 0);
	require(name != 0);
	
	hlen = strlen(name) + 1;
	nlen = htonl(hlen);
	/* Add the size of the string first */
	good = jsiAppendMessage(jm, (char *)&nlen, SIZEOF_INT);
	/* Now get the offset of the string in the buffer */
	offset = jm->contents.size;
	good = good && jsiAppendMessage(jm, name, hlen);
	/* We get the address here since the buffer might've been realloc'ed */
	if( good )
		retval = &jm->contents.data[offset];
	else
		retval = 0;
	return( retval );
}

/* Add a name to the message and its stringSequence owner */
static
int jsiAddName(jsiMessage *jm, jsiStringSequence *ss, char *name)
{
	int retval = 1;
	char *name_cp;

	require(jm != 0);
	require(ss != 0);
	require(name != 0);

	/* Add the string to the message */
	if( !(name_cp = jsiAddString(jm, name)) )
		retval = 0;
	/* Make sure theres space in the stringSequence */
	if( retval && ((ss->count + 1) > ss->max) )
	{
		char **new_names;

		if( (new_names = jsiRealloc(ss->data,
					    sizeof(char *) *
					    (ss->max + JSI_ARRAY_GROWTH))) )
		{
			ss->data = new_names;
			ss->max += JSI_ARRAY_GROWTH;
		}
		else
		{
			retval = 0;
		}
	}
	/* Finally, add the internal string to the sequence */
	if( retval )
	{
		int slot;
		
		slot = ss->count;
		ss->data[slot] = name_cp;
		ss->count++;
	}
	return( retval );
}

jsiMessage *jsiCreateMessage()
{
	jsiMessage *retval;

	/* Allocate the jsiMessage object and an initial buffer */
	if( (retval = jsiMalloc(sizeof(jsiMessage))) &&
	    (retval->contents.data = jsiMalloc(JSI_BUFFER_INIT_SIZE)) )
	{
		retval->contents.max = JSI_BUFFER_INIT_SIZE;
		jsiClearMessage(retval);
	}
	else
	{
		jsiFree(retval);
		retval = 0;
	}
	return( retval );
}

jsiMessage *jsiBuildMessage(int tag, ...)
{
	jsiMessage *retval;
	va_list args;

	va_start(args, tag);
	if( (retval = jsiCreateMessage()) )
	{
		int good = 1;
		
		while( good && (tag != JMA_TAG_DONE) )
		{
			good = jsiAddTag(retval,
					 &retval->identifiers,
					 tag,
					 &args);
			tag = va_arg(args, int);
		}
		if( !good )
		{
			jsiDeleteMessage(retval);
			retval = 0;
		}
	}
	va_end(args);

	ensure(!retval || jsiMessageInvariant(retval));
	return( retval );
}

void jsiDeleteMessage(jsiMessage *jm)
{
	if( jm )
	{
		require(jsiMessageInvariant(jm));
		
		jsiFree(jm->specifiers.data);
		jsiFree(jm->identifiers.data);
		jsiFree(jm->contents.data);
		jsiFree(jm);
	}
}

void jsiClearMessage(jsiMessage *jm)
{
	require(jm != 0);

	/* Reset any internal values */
	jm->what = JSI_ERROR;
	jm->specifiers.count = 0;
	jm->identifiers.count = 0;
	/* Always allocate space for the header */
	jm->contents.size = sizeof(jsiMessageHeader);

	ensure(jsiMessageInvariant(jm));
}

/* Add a name/value pair to the message */
static
int jsiAddTag(jsiMessage *jm, jsiStringSequence *ss,
	      int tag, va_list *args_ptr)
{
	va_list args = *args_ptr;
	int retval = 1;
	char *name;

	require(jm != 0);
	require(ss != 0);

	/* Figure out what type to add */
	switch( tag )
	{
	case JMA_What:
		jm->what = va_arg(args, int);
		break;
	case JMA_Character:
		{
			char value[2];
			
			name = va_arg(args, char *);
			retval = jsiAddName(jm, ss, name);
			value[0] = va_arg(args, int);
			value[1] = 0;
			retval = retval && jsiAddData(jm, value, 2);
		}
		break;
	case JMA_Integer:
		{
			char value[24];
			int len;
			
			name = va_arg(args, char *);
			retval = jsiAddName(jm, ss, name);
			len = sprintf(value, "%d", va_arg(args, int));
			retval = retval && jsiAddData(jm, value, len + 1);
		}
		break;
	case JMA_UnsignedInteger:
		{
			char value[24];
			int len;
			
			name = va_arg(args, char *);
			retval = jsiAddName(jm, ss, name);
			len = sprintf(value, "%u", va_arg(args, int));
			retval = retval && jsiAddData(jm, value, len + 1);
		}
		break;
	case JMA_Float:
		{
			char value[24];
			int len;
			
			name = va_arg(args, char *);
			retval = jsiAddName(jm, ss, name);
			len = sprintf(value, "%f", va_arg(args, double));
			retval = retval && jsiAddData(jm, value, len + 1);
		}
		break;
	case JMA_Pointer:
		{
			char value[24];
			int len;
			
			name = va_arg(args, char *);
			retval = jsiAddName(jm, ss, name);
			len = sprintf(value, "%p", va_arg(args, void *));
			retval = retval && jsiAddData(jm, value, len + 1);
		}
		break;
	case JMA_String:
		{
			char *value;
			int len;
			
			name = va_arg(args, char *);
			retval = jsiAddName(jm, ss, name);
			value = va_arg(args, char *);
			if( value == 0 )
				value = "";
			len = strlen(value);
			retval = retval && jsiAddData(jm, value, len + 1);
		}
		break;
	case JMA_Bytes:
		{
			int len, start;
			char *value;
			
			name = va_arg(args, char *);
			retval = jsiAddName(jm, ss, name);
			value = va_arg(args, char *);
			len = va_arg(args, int);
			start = jm->contents.size + SIZEOF_INT;
			retval = retval && jsiAddData(jm, 0, len + 1);
			if( retval )
			{
				if( value )
				{
					memcpy(&jm->contents.data[start],
					       value,
					       len);
				}
				jm->contents.data[start + len] = '\0';
			}
		}
		break;
	case JMA_PropertyInfo:
		{
			jsiPropertyInfo *pi;
			int len, start;
			char *str;
			
			name = va_arg(args, char *);
			retval = jsiAddName(jm, ss, name);
			pi = va_arg(args, jsiPropertyInfo *);
			len = jsiComputePropertyInfoSize(pi);
			start = jm->contents.size + SIZEOF_INT;
			retval = retval && jsiAddData(jm, 0, len);
			if( retval )
			{
				str = &jm->contents.data[start];
				jsiEncodePropertyInfo(pi, str);
			}
		}
		break;
	case JMA_PropertyList:
		{
			jsiPropertyInfo *pi;
			int lpc, len, start;
			char *str;
			
			name = va_arg(args, char *);
			pi = va_arg(args, jsiPropertyInfo *);
			for( lpc = 0; retval && pi[lpc].name; lpc++ )
			{
				retval = jsiAddName(jm, ss, name);
				len = jsiComputePropertyInfoSize(&pi[lpc]);
				start = jm->contents.size + SIZEOF_INT;
				retval = retval && jsiAddData(jm, 0, len);
				if( retval )
				{
					str = &jm->contents.data[start];
					jsiEncodePropertyInfo(&pi[lpc], str);
				}
			}
		}
		break;
	case JMA_FormattedString:
		{
			int start, bufSize = 1024;
			va_list *fmtArgs;
			char *dest, *format;

			name = va_arg(args, char *);
			retval = jsiAddName(jm, ss, name);
			format = va_arg(args, char *);
			fmtArgs = va_arg(args, va_list *);
			start = jm->contents.size + SIZEOF_INT;
			retval = retval && jsiAddData(jm, 0, bufSize);
			if( retval )
			{
				dest = &jm->contents.data[start];
				/* XXX */
#ifdef HAVE_VSNPRINTF
				vsnprintf(dest, bufSize, format, *fmtArgs);
#else
				vsprintf(dest, format, *fmtArgs);
#endif
			}
		}
		break;
	case JMA_Specifier:
		{
			tag = va_arg(args, int);
			/* er... Is this a good idea? */
			retval = jsiAddTag(jm,
					   &jm->specifiers,
					   tag,
					   &args);
		}
		break;
	default:
		break;
	}
	*args_ptr = args;
	return( retval );
}

void jsiMessageGetIdentifier(jsiMessage *jm,
			     int index, char **out_name, char **out_value)
{
	/* Should the index checks be all the time? */
	require(index >= 0);
	require(index < jm->identifiers.count);
	require(out_name);
	require(out_value);

	*out_name = jm->identifiers.data[index];
	*out_value = jm->identifiers.data[index] +
		ALIGN_INT(strlen(*out_name) + 1) +
		SIZEOF_INT;
}

/* Get the string value for a given name and instance */
static
int jsiIdentifierSlot(jsiMessage *jm, char *name, int instance,
		      char **out_data)
{
	int lpc, retval = -1;

	require(jm != 0);
	require(name != 0);
	require(out_data != 0);

	if( instance >= 0 )
	{
		/* Count forwards to find the right instance */
		for( lpc = 0;
		     (retval < 0) && (lpc < jm->identifiers.count);
		     lpc++ )
		{
			if( !strcmp(jm->identifiers.data[lpc], name) )
			{
				if( !instance )
				{
					int name_len;
					
					/*
					 * The data value should be after this
					 * name and the integer specifying
					 * the size of the value.
					 */
					name_len = ALIGN_INT(ntohl(*(int *)(
						jm->identifiers.data[lpc] -
						SIZEOF_INT)));
					*out_data = jm->identifiers.data[lpc] +
						name_len +
						SIZEOF_INT;
					retval = lpc;
				}
				else
				{
					instance--;
				}
			}
		}
	}
	else
	{
		/* Count backwards to find the right instance */
		for( lpc = jm->identifiers.count - 1;
		     (retval < 0) && (lpc >= 0);
		     lpc-- )
		{
			if( !strcmp(jm->identifiers.data[lpc], name) )
			{
				if( instance == -1 )
				{
					int name_len;
					
					name_len = ALIGN_INT(ntohl(*(int *)(
						jm->identifiers.data[lpc] -
						SIZEOF_INT)));
					*out_data = jm->identifiers.data[lpc] +
						name_len +
						SIZEOF_INT;
					retval = lpc;
				}
				else
				{
					instance++;
				}
			}
		}
	}
	return( retval );
}

int jsiPushMessageSpecifier(jsiMessage *jm, int tag, ...)
{
	va_list args;
	int retval;

	require(jm != 0);
	
	va_start(args, tag);
	retval = jsiAddTag(jm, &jm->specifiers, tag, &args);
	va_end(args);
	return( retval );
}

int jsiPopMessageSpecifier(jsiMessage *jm, char **out_name, char **out_value)
{
	int retval = 0;

	require(jm != 0);
	
	if( jm->specifiers.count )
	{
		char *name;
		
		/*
		 * Just decrease the specifier count and return the pointers
		 * into our buffer.
		 */
		name = jm->specifiers.data[jm->specifiers.count - 1];
		if( out_name )
			*out_name = name;
		if( out_value )
		{
			*out_value = name +
				ALIGN_INT(strlen(name) + 1) +
				SIZEOF_INT;
		}
		jm->specifiers.count--;
		retval = 1;
	}
	return( retval );
}

void jsiRestoreMessageSpecifier(jsiMessage *jm)
{
	require(jm != 0);
	require(jm->specifiers.count < jm->specifiers.max);
	
	jm->specifiers.count++;
}

int jsiMessageAdd(jsiMessage *jm, int tag, ...)
{
	int retval = 1;
	va_list args;

	require(jm != 0);
	
	va_start(args, tag);
	while( retval && (tag != JMA_TAG_DONE) )
	{
		retval = jsiAddTag(jm, &jm->identifiers, tag, &args);
		tag = va_arg(args, int);
	}
	va_end(args);
	return( retval );
}

int jsiMessageRem(jsiMessage *jm, char *name, int instance)
{
	int slot, retval = 0;
	char *data;

	if( (slot = jsiIdentifierSlot(jm, name, instance, &data)) >= 0 )
	{
		int lpc, name_len, data_len, total_len;
		char *start;

		start = jm->identifiers.data[slot] - SIZEOF_INT;
		name_len = ALIGN_INT(ntohl(*(int *)start));
		data_len = ALIGN_INT(ntohl(*(int *)(start +
						    SIZEOF_INT +
						    name_len)));
		total_len = SIZEOF_INT + name_len + SIZEOF_INT + data_len;
		memcpy(start,
		       start + total_len,
		       jm->contents.size - ((start - jm->contents.data) +
					    total_len));
		jm->identifiers.count--;
		jm->contents.size -= total_len;
		for( lpc = slot; lpc < jm->identifiers.count; lpc++ )
		{
			jm->identifiers.data[lpc] =
				jm->identifiers.data[lpc + 1] -
				total_len;
		}
		retval = 1;
	}
	return( retval );
}

int jsiMessageGet(jsiMessage *jm, int tag, ...)
{
	char *data, *name, *char_ptr, **str_ptr;
	int slot, instance, retval = 1;
	jsiPropertyInfo **pi_ptr;
	double *double_ptr;
	void **void_ptr;
	int *int_ptr;
	va_list args;

	require(jm != 0);
	
	va_start(args, tag);
	/* Walk over the tag list and try to find all the requested values */
	while( tag != JMA_TAG_DONE )
	{
		switch( tag )
		{
		case JMA_What:
			int_ptr = va_arg(args, int *);
			*int_ptr = jm->what;
			break;
		case JMA_Character:
			name = va_arg(args, char *);
			instance = va_arg(args, int);
			char_ptr = va_arg(args, char *);
			if( (slot = jsiIdentifierSlot(jm, name,
						      instance, &data)) >= 0 )
			{
				*char_ptr = data[0];
			}
			else
			{
				retval = 0;
			}
			break;
		case JMA_Integer:
			name = va_arg(args, char *);
			instance = va_arg(args, int);
			int_ptr = va_arg(args, int *);
			if( (slot = jsiIdentifierSlot(jm, name,
						      instance, &data)) >= 0 )
			{
				if( sscanf(data, "%d", int_ptr) == 0 )
					retval = 0;
			}
			else
			{
				retval = 0;
			}
			break;
		case JMA_UnsignedInteger:
			name = va_arg(args, char *);
			instance = va_arg(args, int);
			int_ptr = va_arg(args, int *);
			if( (slot = jsiIdentifierSlot(jm, name,
						      instance, &data)) >= 0 )
			{
				if( sscanf(data, "%u", int_ptr) == 0 )
					retval = 0;
			}
			else
			{
				retval = 0;
			}
			break;
		case JMA_Float:
			name = va_arg(args, char *);
			instance = va_arg(args, int);
			double_ptr = va_arg(args, double *);
			if( (slot = jsiIdentifierSlot(jm, name,
						      instance, &data)) >= 0 )
			{
				if( sscanf(data, "%lf", double_ptr) == 0 )
					retval = 0;
			}
			else
			{
				retval = 0;
			}
			break;
		case JMA_Pointer:
			name = va_arg(args, char *);
			instance = va_arg(args, int);
			void_ptr = va_arg(args, void **);
			if( (slot = jsiIdentifierSlot(jm, name,
						      instance, &data)) >= 0 )
			{
				if( sscanf(data, "%p", void_ptr) == 0 )
					retval = 0;
			}
			else
			{
				retval = 0;
			}
			break;
		case JMA_String:
			name = va_arg(args, char *);
			instance = va_arg(args, int);
			str_ptr = va_arg(args, char **);
			if( (slot = jsiIdentifierSlot(jm, name,
						      instance, &data)) >= 0 )
			{
				*str_ptr = data;
			}
			else
			{
				retval = 0;
			}
			break;
		case JMA_Bytes:
			name = va_arg(args, char *);
			instance = va_arg(args, int);
			str_ptr = va_arg(args, char **);
			int_ptr = va_arg(args, int *);
			if( (slot = jsiIdentifierSlot(jm, name,
						      instance, &data)) >= 0 )
			{
				*str_ptr = data;
				if( int_ptr )
				{
					int *size_ptr;

					size_ptr = (int *)(data - SIZEOF_INT);
					*int_ptr = ntohl(*size_ptr) - 1;
				}
				else
				{
					/* They don't care */
				}
			}
			else
			{
				retval = 0;
			}
			break;
		case JMA_PropertyInfo:
			name = va_arg(args, char *);
			instance = va_arg(args, int);
			pi_ptr = va_arg(args, jsiPropertyInfo **);
			if( (slot = jsiIdentifierSlot(jm, name,
						      instance, &data)) >= 0 )
			{
				if( !(*pi_ptr = jsiCreatePropertyInfo(
					JPIA_Internal, data,
					JPIA_TAG_DONE)) )
				{
					retval = 0;
				}
			}
			else
			{
				retval = 0;
			}
			break;
		default:
			require(tag > JMA_TAG_DONE);
			require(tag < JMA_TAG_MAX);
			break;
		}
		tag = va_arg(args, int);
	}
	va_end(args);
	return( retval );
}

int jsiMessageIdentifierCount(jsiMessage *jm, char *name)
{
	int lpc, retval = 0;

	for( lpc = 0; lpc < jm->identifiers.count; lpc++ )
	{
		if( !strcmp(jm->identifiers.data[lpc], name) )
			retval++;
	}
	return( retval );
}

char *jsiFlattenMessage(jsiMessage *jm, int *out_bufferSize)
{
	int good = 1, lpc, offset;
	jsiMessageHeader *jmh;
	char *retval = 0;

	require(jm != 0);
	require(out_bufferSize != 0);

	/* Get a pointer to the message header at the base of the buffer */
	jmh = (jsiMessageHeader *)jm->contents.data;
	/* Write the network encoded what integer */
	jmh->what = htonl(jm->what);
	/*
	 * The contents of the message is just a bunch of strings so we
	 * encode the offsets of all the specifiers at the end of the message
	 * so we can tell them apart from the regular values.
	 */
	for( lpc = 0; good && (lpc < jm->specifiers.count); lpc++ )
	{
		offset = htonl(jm->specifiers.data[lpc] - jm->contents.data);
		good = jsiAppendMessage(jm, (char *)&offset, sizeof(offset));
	}
	/* Check if we're good to go and fill in the rest of the header */
	if( good )
	{
		jmh->messageSize = htonl(jm->contents.size);
		jmh->specifierCount = htonl(jm->specifiers.count);
		jmh->identifierCount = htonl(jm->identifiers.count);
		(*out_bufferSize) = jm->contents.size;
		retval = jm->contents.data;
	}
	return( retval );
}

void jsiRestoreMessage(jsiMessage *jm)
{
	jm->contents.size -= (sizeof(int) * jm->specifiers.count);
}

/*
 * Decode a flattened message buffer and index all of the
 * identifiers/specifiers.
 */
static
int jsiUnflattenMessage(jsiMessage *jm)
{
	int *specOffsets, currSpec = 0;
	int len, lpc, retval = 1;
	char *pos, *nextSpec;

	require(jm != 0);

	/*
	 * Get a pointer to the list of specifier offsets at the end of
	 * the message
	 */
	specOffsets = (int *)&jm->contents.data[jm->contents.size];
	/* Get the next specifier offset we're expecting */
	if( jm->specifiers.max > 0 )
		nextSpec = jm->contents.data + ntohl(specOffsets[0]);
	else
		nextSpec = 0;
	pos = jm->contents.data + sizeof(jsiMessageHeader);
	/*
	 * We're expecting a known number of identifiers and specifiers in the
	 * contents of the message
	 */
	for( lpc = 0;
	     retval && (lpc < (jm->identifiers.max + jm->specifiers.max));
	     lpc++ )
	{
		/* Get the size of the data item in host format */
		len = ntohl(*((int *)pos));
		/* Skip to the start of the string */
		pos += SIZEOF_INT;
		if( (len <= 0) ||
		    ((pos + len) > (jm->contents.data + jm->contents.size)) )
		{
			/* The length is bogus */
			retval = 0;
			jsiLog("Length of tag name at offset %d in message %p "
			       "is %d which is invalid.",
			       pos - jm->contents.data,
			       jm,
			       len);
		}
		else if( pos == nextSpec )
		{
			/* Its a specifier */
			currSpec++;
			/* Check for any additional specifiers */
			if( currSpec < jm->specifiers.max )
			{
				nextSpec = jm->contents.data +
					ntohl(specOffsets[currSpec]);
			}
			else
			{
				nextSpec = 0;
			}
			jm->specifiers.data[jm->specifiers.count] = pos;
			/* Make sure its terminated */
			jm->specifiers.data[jm->specifiers.count][len - 1] =
				'\0';
			jm->specifiers.count++;
		}
		else if( jm->identifiers.count < jm->identifiers.max )
		{
			/* A regular value */
			jm->identifiers.data[jm->identifiers.count] = pos;
			/* Make sure its terminated */
			jm->identifiers.data[jm->identifiers.count][len - 1] =
				'\0';
			jm->identifiers.count++;
		}
		else
		{
			/* Too many identifiers */
			retval = 0;
			jsiLog("Extra identifiers in message %p", jm);
		}
		if( retval )
		{
			/* Skip the aligned size of the name */
			pos += ALIGN_INT(len);
			/* Get the length of the value */
			len = ntohl(*((int *)pos));
			if( (len <= 0) &&
			    ((pos + len) >
			     (jm->contents.data + jm->contents.size)) )
			{
				retval = 0;
				jsiLog("Length of tag value at offset %d in "
				       "message %p is %d which is invalid.",
				       pos - jm->contents.data,
				       jm,
				       len);
			}
			else
			{
				char *val;
				
				/*
				 * Skip the aligned size of the value and its
				 * length integer
				 */
				val = pos + SIZEOF_INT;
				/* Make sure its terminated */
				val[len - 1] = '\0';
				pos = val + ALIGN_INT(len);
			}
		}
	}
	if( (jm->specifiers.count != jm->specifiers.max) ||
	    (jm->identifiers.count != jm->identifiers.max) )
	{
		retval = 0;
		jsiLog("Invalid specifier/identifier count in message %p",
		       jm);
	}
	return( retval );
}

jsiMessage *jsiInstantiateMessageNoFree(char *buffer, int bufferSize)
{
	jsiMessage *retval = 0;
	jsiMessageHeader *jmh;

	require(buffer != 0);

	/* Get a pointer to the header and decode it */
	jmh = (jsiMessageHeader *)buffer;
	jmh->messageSize = ntohl(jmh->messageSize);
	jmh->what = ntohl(jmh->what);
	jmh->specifierCount = ntohl(jmh->specifierCount);
	jmh->identifierCount = ntohl(jmh->identifierCount);
	if( (jmh->messageSize == bufferSize) &&
	    (jmh->specifierCount >= 0) &&
	    (jmh->identifierCount >= 0) &&
	    (jmh->what > JSI_WHAT_NONE) &&
	    (jmh->what < JSI_WHAT_MAX) )
	{
		/* Allocate the jsiMessage object and its string sequences */
		if( (retval = jsiMalloc(sizeof(jsiMessage))) &&
		    (!jmh->specifierCount ||
		     (retval->specifiers.data =
		      jsiMalloc(sizeof(char *) * jmh->specifierCount))) &&
		    (!jmh->identifierCount ||
		     (retval->identifiers.data =
		      jsiMalloc(sizeof(char *) * jmh->identifierCount))) )
		{
			/* Copy important data from the header */
			retval->what = jmh->what;
			retval->specifiers.max = jmh->specifierCount;
			retval->identifiers.max = jmh->identifierCount;
			retval->contents.data = buffer;
			/*
			 * Note:  If specifierCount is bogus than its possible
			 * to get a negative size.  However, the unflatten
			 * will fail properly.
			 */
			retval->contents.size = bufferSize -
				(sizeof(int) * jmh->specifierCount);
			retval->contents.max = bufferSize;
			/* Try to decode the message */
			if( !jsiUnflattenMessage(retval) )
			{
				/*
				 * Need to clear the buffer we aren't supposed
				 * to free it.
				 */
				retval->contents.data = 0;
				jsiDeleteMessage(retval);
				retval = 0;
			}
		}
		else
		{
			jsiFree(retval->specifiers.data);
			jsiFree(retval);
			retval = 0;
			jsiLog("Unable to allocate memory for message");
		}
	}
	else
	{
		jsiLog("Invalid message header");
	}
	return( retval );
}

jsiMessage *jsiInstantiateMessage(char *buffer, int bufferSize)
{
	jsiMessage *retval;

	if( !(retval = jsiInstantiateMessageNoFree(buffer, bufferSize)) )
	{
		jsiFree(buffer);
	}
	return( retval );
}

char *jsiWhatNames[JSI_WHAT_MAX] = {
	"",
	"error",
	"shutdown",
	"carrier",
	"reply",
	"get-suites",
	"list-properties",
	"get-property",
	"set-property",
	"create-property",
	"delete-property",
	"execute-property",
};

void jsiPrintMessage(FILE *file, jsiMessage *jm)
{
	int lpc;

	require(file != 0);
	require(jm != 0);
	
	fprintf(file,
		"jsiMessage[what=%s; specifiers={",
		jsiWhatNames[jm->what]);
	for( lpc = 0; lpc < jm->specifiers.count; lpc++ )
	{
		fprintf(file, "%s%s",
			(lpc > 0) ? ", " : "",
			jm->specifiers.data[lpc]);
	}
	fprintf(file, "}; values={");
	for( lpc = 0; lpc < jm->identifiers.count; lpc++ )
	{
		fprintf(file, "%s%s=%s",
			(lpc > 0) ? ", " : "",
			jm->identifiers.data[lpc],
			jm->identifiers.data[lpc] +
			ALIGN_INT(strlen(jm->identifiers.data[lpc]) + 1) +
			SIZEOF_INT);
	}
	fprintf(file, "}]\n");
}
