/*
 * jsiPlexChannel.c
 * Channel implementation that runs over unreliable links.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsi.h"
#include "jsiImpl.h"
#include "jsiPlexChannel.h"

#ifndef MIN
#define MIN(x, y) ((x) < (y) ? (x) : (y))
#endif

#ifndef MAX
#define MAX(x, y) ((x) > (y) ? (x) : (y))
#endif

#ifndef ABS
#define ABS(x) ((x) < 0 ? -x : x)
#endif

void jsiPrintPlexChannel(FILE *file, jsiPlexChannel *pc)
{
	require(file != 0);
	require(pc != 0);
	
	fprintf(file,
		"jsiPlexChannel(%p)[next=%p; window={",
		pc,
		pc->next);
	jsiPrintPlexWindow(file, &pc->window);
	fprintf(file,
		"}; channelID=%d; nextMessageID=%d; recvBuffer={"
		"messageID=%d; start=%p; size=%d; remaining=%d}]",
		pc->channelID,
		pc->nextMessageID,
		pc->recvBuffer.messageID,
		pc->recvBuffer.start,
		pc->recvBuffer.size,
		pc->recvBuffer.remaining);
}

struct {
#define JSI_PLEX_MAX_CHANNEL_IDS 256
#define JSI_PLEX_FREE_CHANNEL_ID 0
#define JSI_PLEX_CHANNEL_ID_INDEX_MASK 0x00ff
#define JSI_PLEX_CHANNEL_ID_DELAY_OFFSET 8
#define JSI_PLEX_CHANNEL_ID_DELAY_MASK 0xff00
	unsigned short ids[JSI_PLEX_MAX_CHANNEL_IDS];
#define JSI_PLEX_CHANNEL_TABLE_SIZE 31
	jsiPlexChannel *table[JSI_PLEX_CHANNEL_TABLE_SIZE];
	JSI_LOCK_DECL
} plexChannels = {
	{ 1 /* Allocate the zero'th ID */ }
};

/*
 * returns - A new channel identifier for use with outgoing connections.
 */
static
unsigned short allocChannelID(void)
{
	unsigned int start, lpc, slot, end, retval = 0;
	int iLockRoot;
	
	jsiLock(&plexChannels);
#ifdef JSI_RANDOM_CHANNEL_IDS
	srandom(time(0));
	start = random() % JSI_PLEX_MAX_CHANNEL_IDS;
	if( start == 0 )
		start = 1;
	end = start + JSI_PLEX_MAX_CHANNEL_IDS;
#else
	start = 1;
	end = JSI_PLEX_MAX_CHANNEL_IDS;
#endif
	for( lpc = start; !retval && (lpc < end); lpc++ )
	{
		slot = lpc % JSI_PLEX_MAX_CHANNEL_IDS;
		if( (plexChannels.ids[slot] & JSI_PLEX_CHANNEL_ID_INDEX_MASK)
		    == 0 )
		{
			/* It's free */
			plexChannels.ids[slot] +=
				(1L << JSI_PLEX_CHANNEL_ID_DELAY_OFFSET);
			plexChannels.ids[slot] |= slot;
			retval = plexChannels.ids[slot];
		}
	}
	if( retval )
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
		fprintf(stderr,
			"Allocated channel id: %d (delay=%d; slot=%d)\n",
			retval,
			(retval & JSI_PLEX_CHANNEL_ID_DELAY_MASK) >>
			JSI_PLEX_CHANNEL_ID_DELAY_OFFSET,
			retval & JSI_PLEX_CHANNEL_ID_INDEX_MASK);
#endif
	}
	else
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr, "Failed to allocated channel ID\n");
#endif
	}
	jsiUnlock(&plexChannels);
	return( retval );
}

/*
 * Free a previously allocated channel identifier.
 *
 * `id' - An channel identifier as returned by allocChannelID.
 */
static
void freeChannelID(unsigned short id)
{
	int iLockRoot;

	require(id != 0);
	require(plexChannels.ids[id & JSI_PLEX_CHANNEL_ID_INDEX_MASK] != 0);
	
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
	fprintf(stderr,
		"Freeing channel id: %d (delay=%d; slot=%d)\n",
		id,
		(id & JSI_PLEX_CHANNEL_ID_DELAY_MASK) >>
		JSI_PLEX_CHANNEL_ID_DELAY_OFFSET,
		id & JSI_PLEX_CHANNEL_ID_INDEX_MASK);
#endif
	jsiLock(&plexChannels);
	plexChannels.ids[id & JSI_PLEX_CHANNEL_ID_INDEX_MASK] &=
		~JSI_PLEX_CHANNEL_ID_INDEX_MASK;
	jsiUnlock(&plexChannels);
}

/*
 * Compute the hash for a given channel identifier and link hash.
 *
 * `channelID' - The channel identifier to hash.
 * `linkHash' - The jsiPlexLink hash value for the peer.
 * returns - An index into plexChannels.table.
 */
static
int hashPlexChannel(unsigned short channelID, int linkHash)
{
	int retval;

	retval = ((channelID + linkHash) & 0x7fffffff) %
		JSI_PLEX_CHANNEL_TABLE_SIZE;

	ensure(retval >= 0);
	ensure(retval < JSI_PLEX_CHANNEL_TABLE_SIZE);
	
	return( retval );
}

void jsiAddPlexChannel(jsiPlexChannel *pc, jsiPlexOutLink *pol)
{
	int hash, iLockRoot;

	require(pc != 0);
	require(!(pc->flags & PCF_PUBLISHED));
	require(pol != 0);

#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
	fprintf(stderr, "Adding plex channel ");
	jsiPrintPlexChannel(stderr, pc);
	fprintf(stderr, "\n");
#endif
	pc->flags |= PCF_PUBLISHED;
	jsiLock(&plexChannels);
	hash = hashPlexChannel(pc->channelID, pol->hashValue);
	pc->next = plexChannels.table[hash];
	plexChannels.table[hash] = pc;
	jsiUnlock(&plexChannels);
}

void jsiRemPlexChannel(jsiPlexChannel *pc, jsiPlexOutLink *pol)
{
	jsiPlexChannel *curr, **prev;
	int hash, iLockRoot;

	require(pc != 0);
	require(pc->flags & PCF_PUBLISHED);
	require(pol != 0);
	
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
	fprintf(stderr, "Removing plex channel ");
	jsiPrintPlexChannel(stderr, pc);
	fprintf(stderr, "\n");
#endif
	hash = hashPlexChannel(pc->channelID, pol->hashValue);
	jsiLock(&plexChannels);
	curr = plexChannels.table[hash];
	prev = &plexChannels.table[hash];
	while( curr != pc )
	{
		prev = &curr->next;
		curr = curr->next;
	}
	if( curr == pc )
	{
		*prev = curr->next;
	}
	else
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr,
			"Trying to remove channel that wasn't added %p\n",
			pc);
#endif
	}
	jsiUnlock(&plexChannels);
	pc->flags &= ~PCF_PUBLISHED;
}

jsiPlexChannel *jsiFindPlexChannel(unsigned short channelID,
				   jsiPlexLinkContext *plc)
{
	jsiPlexChannel *pc, *retval = 0;
	int hash, iLockRoot;

	require(channelID != 0);
	require(plc != 0);

	hash = hashPlexChannel(channelID, plc->hashValue);
	jsiLock(&plexChannels);
	pc = plexChannels.table[hash];
	while( pc && !retval )
	{
		if( (pc->channelID == channelID) &&
		    jsiComparePlexOutLink(pc->window.link, plc) )
		{
			retval = pc;

			ensure(retval->flags & PCF_PUBLISHED);
		}
		pc = pc->next;
	}
	jsiUnlock(&plexChannels);
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_WARNINGS
	if( retval == 0 )
	{
		fprintf(stderr,
			"Unable to find channel for %d/(%p %x)\n",
			channelID,
			plc,
			plc->hashValue);
	}
#endif
	return( retval );
}

jsiChannel *jsiCreatePlexChannel(void)
{
	jsiChannel *retval = 0;
	jsiPlexChannel *pc;

	if( (pc = (jsiPlexChannel *)jsiMalloc(sizeof(jsiPlexChannel))) )
	{
		pc->channel.ops = &jsiPlexChannelImplementation;
		pc->next = 0;
		pc->flags = 0;
		jsiPlexWindowInit(&pc->window);
		/* Only allocate an ID if this side initiates the connection */
		pc->channelID = 0;
		pc->nextMessageID = 0;
		pc->recvBuffer.messageID = 0;
		pc->recvBuffer.start = 0;
		pc->recvBuffer.size = 0;
		pc->recvBuffer.remaining = 0;
		retval = &pc->channel;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
		fprintf(stderr, "Created channel ");
		jsiPrintPlexChannel(stderr, pc);
		fprintf(stderr, "\n");
#endif
	}
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_WARNINGS
	if( retval == 0 )
		fprintf(stderr, "Unable to allocate channel\n");
#endif
	return( retval );
}

static
void jsiDeletePlexChannel(jsiChannel *jc)
{
	jsiPlexChannel *pc = (jsiPlexChannel *)jc;
	
	require(jc != 0);

#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
	fprintf(stderr, "Deleting channel ");
	jsiPrintPlexChannel(stderr, pc);
	fprintf(stderr, "\n");
#endif
	jsiDisconnectChannel(jc);
	jsiFree(pc->recvBuffer.start);
	jsiFree(jc);
}

static
jsiPlexOutLink *jsiConnectPlexFormatted(jsiPlexChannel *pc, char *fmt)
{
	parsedString protocol, link, rest;
	jsiPlexOutLink *retval = 0;
	int iLockRoot;

	require(pc != 0);
	require(fmt != 0);

	if( parseString(fmt,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_NonEmptyString, &link,
			SPO_Expect, ":",
			SPO_String, &rest,
			SPO_End) )
	{
		jsiPlexLink *pl;

		pl = jsiFindPlexLink(link.data, link.len);
		if( pl )
		{
			jsiObtainPlexLink(pl);
			retval = jsiGetPlexOutLinkFromFormat(pl, link.data);
			jsiReleasePlexLink(pl);
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
			fprintf(stderr, "Connecting channel ");
			jsiPrintPlexChannel(stderr, pc);
			fprintf(stderr, " to %s thru ", fmt);
			if( retval )
				jsiPrintPlexOutLink(stderr, retval);
			else
				fprintf(stderr, "nothing");
			fprintf(stderr, "\n");
#endif
		}
	}
	return( retval );
}

static
int jsiConnectPlexChannel(jsiChannel *jc, int tag, ...)
{
	jsiPlexChannel *pc = (jsiPlexChannel *)jc;
	int iLockRoot, retval = 0;
	jsiPlexOutLink *pol = 0;
	char *enc = 0, *fmt = 0;
	va_list args;

	require(jc != 0);
	
	va_start(args, tag);
	jsiDisconnectChannel(jc);
	while( tag != JCA_TAG_DONE )
	{
		switch( tag )
		{
		case JCA_Formatted:
			fmt = va_arg(args, char *);
			pol = jsiConnectPlexFormatted(pc, fmt);
			break;
		case JCA_Encoded:
			enc = va_arg(args, char *);
			break;
		default:
			break;
		}
		tag = va_arg(args, int);
	}
	if( enc )
	{
		parsedString protocol, link, linkData;
		jsiPlexWindow *pw = &pc->window;

		if( parseString(enc,
				SPO_NonEmptyString, &protocol,
				SPO_Expect, ":",
				SPO_Short, &pc->channelID,
				SPO_Expect, "/",
				SPO_Short, &pc->nextMessageID,
				SPO_Expect, "/",
				SPO_Integer, &pw->state,
				SPO_Expect, "/",
				SPO_Byte, &pw->lastAckReceived,
				SPO_Expect, "/",
				SPO_Byte, &pw->lastFrameSent,
				SPO_Expect, "/",
				SPO_Byte, &pw->nextFrameExpected,
				SPO_Expect, "/",
				SPO_NonEmptyString, &link,
				SPO_Expect, ":",
				SPO_NonEmptyString, &linkData,
				SPO_End) )
		{
			jsiPlexLink *pl;

			pl = jsiFindPlexLink(link.data, link.len);
			if( pl )
			{
				jsiObtainPlexLink(pl);
				pol = jsiGetPlexOutLinkFromEncoding(pl,
								    link.data);
				jsiReleasePlexLink(pl);
				if( pol )
				{
					jsiAddPlexChannel(pc, pol);
					pw->link = pol;
					retval = 1;
				}
			}
		}
		else
		{
			fprintf(stderr, "Couldn't parse encoding\n");
		}
	}
	else if( pol && (pc->channelID = allocChannelID()) )
	{
		pc->flags |= PCF_LOCAL;
		jsiAddPlexChannel(pc, pol);
		if( !(retval = jsiPlexWindowConnect(&pc->window,
						    pol,
						    pc->channelID)) )
		{
			jsiPlexLink *pl = pol->link;

			jsiRemPlexChannel(pc, pol);
			pc->flags &= ~PCF_LOCAL;
			freeChannelID(pc->channelID);
			pc->channelID = 0;
			jsiObtainPlexLink(pl);
			jsiDropPlexOutLink(pol);
			jsiReleasePlexLink(pl);
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
			fprintf(stderr, "Connection failed\n");
#endif
		}
	}
	else
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr,
			"Unable to create jsiPlexOutLink for connection\n");
#endif
	}
	va_end(args);
	return( retval );
}

static
void jsiDisconnectPlexChannel(jsiChannel *jc)
{
	jsiPlexChannel *pc = (jsiPlexChannel *)jc;
	
	require(pc != 0);
	
	if( pc->window.link )
	{
		jsiPlexOutLink *pol = pc->window.link;
		jsiPlexLink *pl = pol->link;
		int iLockRoot;
		
		jsiPlexWindowDisconnect(&pc->window, pc->channelID);
		jsiRemPlexChannel(pc, pol);
		if( pc->flags & PCF_LOCAL )
		{
			freeChannelID(pc->channelID);
			pc->flags &= ~PCF_LOCAL;
		}
		pc->channelID = 0;
		jsiObtainPlexLink(pl);
		jsiDropPlexOutLink(pol);
		jsiReleasePlexLink(pl);
	}
	else
	{
		/* Already disconnected */
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
		fprintf(stderr, "Already disconnected\n");
#endif
	}
	
	ensure(pc->window.state == PWS_CLOSED);
	ensure(!(pc->flags & PCF_PUBLISHED));
}

int jsiPlexSendMessage(jsiChannel *jc, char *buffer, int size)
{
	int mid, offset = 0, packetSize, retval = 1, iLockRoot;
	jsiPlexChannel *pc = (jsiPlexChannel *)jc;
	jsiPlexOutLink *pol = pc->window.link;
	jsiPlexLinkPacket *plp;
	
	require(jc != 0);
	require(buffer != 0);
	require(size >= 0);

	jsiObtainPlexLink(pol->link);
	mid = pc->nextMessageID++;
	/* Packetize the message */
	while( size && retval )
	{
		require(size > 0);
		require(offset >= 0);

		packetSize = MIN(size, pol->link->maxPacketSize);
		if( (plp = jsiAllocPlexLinkPacket(pol, packetSize)) )
		{
			plp->header.fragOffset = offset;
			plp->header.totalLength = offset + size;
			plp->header.messageID = mid;
			plp->header.flags = 0;
			plp->header.channelID = pc->channelID;
			memcpy(plp->body, &buffer[offset], packetSize);
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
			fprintf(stderr, "Sending packet: ");
			jsiPrintPlexLinkPacket(stderr, plp);
			fprintf(stderr, " from ");
			jsiPrintPlexChannel(stderr, pc);
			fprintf(stderr, "\n");
#endif
			retval = jsiPlexWindowSendPacket(&pc->window, plp);
			size -= packetSize;
			offset += packetSize;
		}
		else
		{
			retval = 0;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
			fprintf(stderr,
				"Failed to allocate send packet for channel ");
			jsiPrintPlexChannel(stderr, pc);
			fprintf(stderr, "\n");
#endif
		}
	}
	jsiReleasePlexLink(pol->link);
	if( !retval )
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr, "Send message failed\n");
#endif
		jsiDisconnectChannel(jc);
	}
	return( retval );
}

static
jsiMessage *jsiPlexDoMessage(jsiHandlerContext *hc, jsiMessage *jm)
{
	jsiMessage *retval = 0;

	require(hc != 0);
	
	if( jm )
	{
		int bufferSize;
		char *buffer;

		if( (buffer = jsiFlattenMessage(jm, &bufferSize)) )
		{
			if( jsiSendMessage(hc->channel, buffer, bufferSize) )
			{
				retval = jsiHandleInterimMessages(hc);
			}
		}
		jsiDeleteMessage(jm);
	}
	return( retval );
}

static
int jsiPlexReplyMessage(jsiChannel *jc, char *buffer, int size)
{
	return( jsiPlexSendMessage(jc, buffer, size) );
}

static
char *jsiPlexRecvMessage(jsiChannel *jc, int *bufferSize)
{
	jsiPlexChannel *pc = (jsiPlexChannel *)jc;
	char *retval = 0;

	require(jc != 0);
	require(bufferSize != 0);

	retval = jsiPlexWindowWaitForMessage(pc, bufferSize);
	return( retval );
}

static
char *jsiPlexEncodeChannel(jsiChannel *jc)
{
	jsiPlexChannel *pc = (jsiPlexChannel *)jc;
	jsiPlexOutLink *pol;
	jsiPlexWindow *pw;
	char *retval = 0;
	int iLockRoot;
	
	require(jc != 0);

	/* XXX This is still very experimental */
	pw = &pc->window;
	pol = pw->link;
	jsiObtainPlexLink(pol->link);
	if( pc->window.state == PWS_CONNECTED )
	{
		char *encodedLink;

		/*
		 * Make sure we're not going to lose track of any packets
		 * that have been sent recently.
		 */
		while( pw->lastAckReceived != pw->lastFrameSent )
		{
			jsiWaitForPlexLink(pol->link, 100);
		}
		
		if( (encodedLink = jsiEncodePlexOutLink(pol)) )
		{
			if( (retval = jsiMalloc(4 + 1 +	/* plex */
						5 + 1 +	/* channelID */
						5 + 1 +	/* nextMessageID */
						5 + 1 +	/* state */
						3 + 1 +	/* lastAckReceived */
						3 + 1 +	/* lastFrameSent */
						3 + 1 +	/* nextFrameExpected */
						strlen(encodedLink) +
						1)) )
			{
				sprintf(retval,
					"plex:%d/%d/%d/%d/%d/%d/%s",
					pc->channelID,
					pc->nextMessageID,
					pw->state,
					pw->lastAckReceived,
					pw->lastFrameSent,
					pw->nextFrameExpected,
					encodedLink);
				pw->state = PWS_CLOSED;
			}
			jsiFree(encodedLink);
		}
	}
	jsiReleasePlexLink(pol->link);
	if( retval )
	{
		jsiDeleteChannel(jc);
	}
	return( retval );
}

/* Mapping of protocol identifiers to strings */
static char *jsiPlexChannelNames[] = {
	"plex",
	0
};

jsiChannelFunctions jsiPlexChannelImplementation = {
	jsiPlexChannelNames,
	jsiCreatePlexChannel,
	jsiDeletePlexChannel,
	jsiConnectPlexChannel,
	jsiDisconnectPlexChannel,
	jsiPlexSendMessage,
	jsiPlexDoMessage,
	jsiPlexReplyMessage,
	jsiPlexRecvMessage,
	jsiPlexEncodeChannel
};
