/*
 * jsiPlexChannel.h
 * Channel implementation that runs over unreliable links.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jsi_plex_channel_h
#define _jsi_plex_channel_h

#include "jsiPlexLink.h"
#include "jsiPlexWindow.h"

enum {
	PCB_LOCAL,	/*
			 * Indicates that the channel ID was allocated from
			 * the local pool and not from the remote host.
			 */
	PCB_PUBLISHED,	/*
			 * Indicates the channel has been published and can
			 * be found via jsiFindPlexChannel.
			 */
};

enum {
	PCF_LOCAL = (1L << PCB_LOCAL),
	PCF_PUBLISHED = (1L << PCB_PUBLISHED),
};

/*
 * The jsiPlexChannel is a jsiChannel implementation that runs over unreliable
 * link protocols.
 *
 * `channel' - The jsiChannel header.
 * `next' - Link used when the channel is published in the global table.
 * `flags' - Storage for the above flags.
 * `window' - The sliding window control structure.
 * `channelID' - The local link unique identifier for this channel.
 * `nextMessageID' - The next identifier for a message sent over the channel.
 *   This is just a monotonically increasing value used to group together
 *   packet fragments and to ensure proper message ordering.
 * `recvBuffer.messageID' - The identifier for the message being received.
 * `recvBuffer.start' - The start of memory allocated for the message.
 * `recvBuffer.size' - The total size of the message.
 * `recvBuffer.remaining' - The number of bytes left to be received to complete
 *   the message.
 */
typedef struct _jsiPlexChannel {
	jsiChannel channel;

	struct _jsiPlexChannel *next;
	unsigned long flags;
	jsiPlexWindow window;

	unsigned short channelID;
	
	unsigned int nextMessageID;
	
	struct {
		int messageID;
		char *start;
		int size;
		int remaining;
	} recvBuffer;
} jsiPlexChannel;

/*
 * Print the contents of the given jsiPlexChannel.
 *
 * `file' - The file to print out too.
 * `pc' - The jsiPlexChannel to print out.
 */
void jsiPrintPlexChannel(FILE *file, jsiPlexChannel *pc);
/*
 * Create a new jsiPlexChannel object.
 *
 * returns - A empty, disconnected jsiPlexChannel.
 */
jsiChannel *jsiCreatePlexChannel(void);
/*
 * Publish the given channel as bound to a specific destination.  This is
 * needed so that the jsiPlexWindow code can locate the destination channel
 * of any packets it processes.
 *
 * `pc' - An unpublished channel.
 * `pol' - A jsiPlexOutLink object which refers to the remote host.
 */
void jsiAddPlexChannel(jsiPlexChannel *pc, jsiPlexOutLink *pol);
/*
 * Remove the given channel from the global table.
 *
 * `pc' - A published channel.
 * `pol' - A jsiPlexOutLink object which refers to the remote host.
 */
void jsiRemPlexChannel(jsiPlexChannel *pc, jsiPlexOutLink *pol);
/*
 * Find a channel with the given channel identifier and connected to the given
 * remote host.
 *
 * `channelID' - The identifier for the jsiPlexChannel.
 * `context' - A description of the remote host that will be compared against
 *   the jsiPlexOutLink objects of the channels in the table.
 */
jsiPlexChannel *jsiFindPlexChannel(unsigned short channelID,
				   jsiPlexLinkContext *context);

extern jsiChannelFunctions jsiPlexChannelImplementation;

#define JSI_PLEX_DEBUG_NONE	0
#define JSI_PLEX_DEBUG_ERRORS	1
#define JSI_PLEX_DEBUG_WARNINGS	2
#define JSI_PLEX_DEBUG_SYSTEM	3
#define JSI_PLEX_DEBUG_PACKET	4
#ifndef JSI_PLEX_DEBUG
#define JSI_PLEX_DEBUG		JSI_PLEX_DEBUG_NONE
#endif


#endif
