/*
 * jsiPlexLink.h
 * Link layer for plex channels
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsi.h"
#include "jsiImpl.h"
#include "jsiPlexLink.h"
#include "jsiPlexChannel.h"

/* List of link implementations */
static
jsiPlexLink *plexLinks;

void jsiPrintPlexLinkHeader(FILE *file, jsiPlexLinkHeader *plh)
{
	require(file != 0);
	require(plh != 0);
	
	fprintf(file,
		"jsiPlexLinkHeader(%p)[fragOffset=%d; totalLength=%d; "
		"messageID=%d; flags=0x%x; seqNum=%d; ackNum=%d; "
		"channelID=%d]",
		plh,
		plh->fragOffset,
		plh->totalLength,
		plh->messageID,
		plh->flags,
		plh->seqNum,
		plh->ackNum,
		plh->channelID);
}

void jsiPrintPlexLinkPacket(FILE *file, jsiPlexLinkPacket *plp)
{
	require(file != 0);
	require(plp != 0);

	fprintf(file,
		"jsiPlexLinkPacket(%p)[packetSize=%d; header=",
		plp,
		plp->packetSize);
	jsiPrintPlexLinkHeader(file, &plp->header);
	if( plp->packetSize > sizeof(jsiPlexLinkHeader) )
	{
		fprintf(file,
			"; body={ %02x %02x %02x %02x %02x %02x %02x %02x }",
			plp->body[0],
			plp->body[1],
			plp->body[2],
			plp->body[3],
			plp->body[4],
			plp->body[5],
			plp->body[6],
			plp->body[7]);
	}
	fprintf(file, "]");
}

void jsiAddPlexLink(jsiPlexLink *pl)
{
	require(pl != 0);

#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
	fprintf(stderr, "Adding plex link %p:%s\n", pl, pl->names[0]);
#endif
	pl->next = plexLinks;
	plexLinks = pl;
}

jsiPlexLink *jsiFindPlexLink(char *name, int len)
{
	jsiPlexLink *pl, *retval = 0;
	int lpc;

	require(name != 0);
	require((len == -1) || (len >= 0));

	if( len == -1 )
		len = strlen(name);
	pl = plexLinks;
	while( pl && !retval )
	{
		for( lpc = 0; pl->names[lpc] && !retval; lpc++ )
		{
			if( !strncmp(pl->names[lpc], name, len) &&
			    (pl->names[lpc][len] == '\0') )
			{
				retval = pl;
			}
		}
		pl = pl->next;
	}
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
	if( !retval )
	{
		fprintf(stderr, "Couldn't find link %s\n", name);
	}
#endif
	return( retval );
}

void jsiPrintPlexLink(FILE *file, jsiPlexLink *pl)
{
	int lpc;

	require(pl != 0);
	require(file != 0);
	
	fprintf(file, "jsiPlexLink(%p)[names={", pl);
	for( lpc = 0; pl->names[lpc]; lpc++ )
	{
		fprintf(file, " `%s' ", pl->names[lpc]);
	}
	fprintf(file,
		"}; maxPacketSize=%d",
		pl->maxPacketSize);
	if( pl->ops->print )
		pl->ops->print(pl, file);
	fprintf(file, "]");
}

void jsiPrintPlexLinks(FILE *file)
{
	jsiPlexLink *pl;

	pl = plexLinks;
	while( pl )
	{
		jsiPrintPlexLink(file, pl);
		fprintf(file, "\n");
		pl = pl->next;
	}
}

void jsiPrintPlexInLink(FILE *file, jsiPlexInLink *pil)
{
	require(pil != 0);
	require(file != 0);
	
	fprintf(file,
		"jsiPlexInLink(%p)[link=%p; provider=%p",
		pil,
		pil->link,
		pil->provider);
	if( pil->link->ops->printInLink )
		pil->link->ops->printInLink(pil, file);
	fprintf(file, "]");
}

void jsiPrintPlexOutLink(FILE *file, jsiPlexOutLink *pol)
{
	require(pol != 0);
	require(file != 0);
	
	fprintf(file,
		"jsiPlexOutLink(%p)[link=%p; flags=%lx; "
		"hashValue=%d; users=%d",
		pol,
		pol->link,
		pol->flags,
		pol->hashValue,
		pol->users);
	if( pol->link->ops->printOutLink )
		pol->link->ops->printOutLink(pol, file);
	fprintf(file, "]");
}

jsiPlexInLink *jsiGetPlexInLink(jsiPlexLink *pl, char *format)
{
	jsiPlexInLink *retval = 0;

	require(pl != 0);
	require(pl->ops != 0);
	require(pl->ops->getInLink != 0);
	require(format != 0);

	retval = pl->ops->getInLink(pl, format);

	ensure(!retval || (retval->link != 0));
	
	return( retval );
}

void jsiDropPlexInLink(jsiPlexInLink *pil)
{
	require(pil != 0);
	require(pil->link != 0);

	pil->link->ops->dropInLink(pil);
}

static
void fillPacketPool(jsiPlexOutLink *pol)
{
	int lpc;

	for( lpc = 0; lpc < JSI_PLEX_DEFAULT_PACKET_POOL_SIZE; lpc++ )
	{
		jsiPlexLinkPacket *plp;
		
		if( (plp = (jsiPlexLinkPacket *)
		     jsiMalloc(sizeof(jsiPlexLinkPacket) +
			       pol->link->maxPacketSize)) )
		{
			plp->next = pol->packetPool;
			pol->packetPool = plp;
		}
	}
}

jsiPlexOutLink *jsiGetPlexOutLinkFromFormat(jsiPlexLink *pl, char *format)
{
	jsiPlexOutLink *retval = 0;

	require(pl != 0);
	require(format != 0);
	
	if( (retval = pl->ops->getOutLinkFromFormat(pl, format)) )
	{
		if( retval->packetPool == 0 )
		{
			fillPacketPool(retval);
		}
	}

	ensure(!retval || (retval->link != 0));
	
	return( retval );
}

jsiPlexOutLink *jsiGetPlexOutLinkFromEncoding(jsiPlexLink *pl, char *encoding)
{
	jsiPlexOutLink *retval = 0;

	require(pl != 0);
	require(encoding != 0);
	
	if( (retval = pl->ops->getOutLinkFromEncoding(pl, encoding)) )
	{
		if( retval->packetPool == 0 )
		{
			fillPacketPool(retval);
		}
	}

	ensure(!retval || (retval->link != 0));
	
	return( retval );
}

jsiPlexOutLink *jsiGetPlexOutLinkFromContext(jsiPlexLink *pl,
					     jsiPlexLinkContext *context)
{
	jsiPlexOutLink *retval = 0;

	require(pl != 0);
	require(context != 0);
	
	if( (retval = pl->ops->getOutLinkFromContext(pl, context)) )
	{
		if( retval->packetPool == 0 )
		{
			fillPacketPool(retval);
		}
	}

	ensure(!retval || (retval->link != 0));
	
	return( retval );
}

void jsiDropPlexOutLink(jsiPlexOutLink *pol)
{
	require(pol != 0);
	require(pol->link != 0);

	pol->link->ops->dropOutLink(pol);
}

int jsiComparePlexOutLink(jsiPlexOutLink *pol, jsiPlexLinkContext *plc)
{
	require(pol != 0);
	require(plc != 0);
	
	return( pol->link->ops->compareOutLink(pol, plc) );
}

jsiChannel *jsiHandlePlexLink(jsiPlexInLink *pil)
{
	jsiChannel *retval;
	
	require(pil != 0);
	require(pil->link != 0);
	
	retval = pil->link->ops->handle(pil);

	return( retval );
}

jsiPlexLinkPacket *jsiAllocPlexLinkPacket(jsiPlexOutLink *pol, int size)
{
	jsiPlexLinkPacket *retval = 0;

	require(pol != 0);
	require(size >= 0);
	require(size <= pol->link->maxPacketSize);

	if( pol->packetPool == 0 )
	{
		fillPacketPool(pol);
	}
	if( (retval = pol->packetPool) )
	{
		pol->packetPool = retval->next;
		retval->next = 0;
		retval->packetSize = sizeof(jsiPlexLinkHeader) + size;
	}
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
	if( retval == 0 )
		fprintf(stderr, "Unable to allocate packet from %p\n", pol);
#endif
	return( retval );
}

void jsiFreePlexLinkPacket(jsiPlexOutLink *pol, jsiPlexLinkPacket *plp)
{
	require(pol != 0);

	if( plp )
	{
		plp->next = pol->packetPool;
		pol->packetPool = plp;
	}
}

int jsiWritePlexLink(jsiPlexOutLink *pol, jsiPlexLinkPacket *plp)
{
	int retval = 0;

	require(pol != 0);
	require(plp != 0);

	if( pol->flags & POLF_UP )
	{
		/* Be nice, do the byte swapping here. */
		plp->header.fragOffset = htons(plp->header.fragOffset);
		plp->header.totalLength = htons(plp->header.totalLength);
		plp->header.messageID = htons(plp->header.messageID);
		plp->header.channelID = htons(plp->header.channelID);
		retval = pol->link->ops->write(pol,
					       &plp->header,
					       plp->packetSize);
	}
	else
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr, "Link down, unable to write packet ");
		jsiPrintPlexLinkPacket(stderr, plp);
		fprintf(stderr, " to ");
		jsiPrintPlexOutLink(stderr, pol);
		fprintf(stderr, "\n");
#endif
	}
	
	ensure((retval == 0) || (retval == 1));
	
	return( retval );
}

int jsiRewritePlexLink(jsiPlexOutLink *pol, jsiPlexLinkPacket *plp)
{
	int retval = 0;
	
	require(pol != 0);
	require(plp != 0);

	if( pol->flags & POLF_UP )
	{
		retval = pol->link->ops->write(pol,
					       &plp->header,
					       plp->packetSize);
	}
	else
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr, "Link down, unable to rewrite packet ");
		jsiPrintPlexLinkPacket(stderr, plp);
		fprintf(stderr, " to ");
		jsiPrintPlexOutLink(stderr, pol);
		fprintf(stderr, "\n");
#endif
	}
	
	ensure((retval == 0) || (retval == 1));
	
	return( retval );
}
