/*
 * jsiPlexLink.h
 * Link layer abstraction for plex channels
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jsi_plex_link_h
#define _jsi_plex_link_h

/*
 * The jsiPlexLink module provides an abstraction layer and support functions
 * for the actual unreliable link that jsiMessages will be multiplexed over.
 */

#include "jsiPlexProvider.h"

/*
 * Flags for packets sent over a plex link...  Hmm, they're remarkably similar
 * to TCP flags...
 */
enum {
	PHB_SYN,
	PHB_ACK,
	PHB_FIN,
	PHB_POKE,	/* Used to test the liveness of the connection. */
};

enum {
	PHF_SYN = (1L << PHB_SYN),
	PHF_ACK = (1L << PHB_ACK),
	PHF_FIN = (1L << PHB_FIN),
	PHF_POKE = (1L << PHB_POKE),
};

/*
 * Header structure for packets sent over a plex link.
 *
 * `fragOffset' - The offset in the message that this packet's payload
 *   contains.  Sent as network order on the link.
 * `totalLength' - The total length of the message.  Sent as network order on
 *   the link.
 * `messageID' - The identifier for this message.  Sent as network order on
 *   the link.
 * `flags' - Storage for the above flags.
 * `seqNum' - The sequence number for this packet, its counted as packets, not
 *   bytes like in TCP.  This isn't set for an ACK.
 * `ackNum' - If PHF_ACK is set then this is the acknowledgement for the given
 *   packet and all previous.
 * `channelID' - The local link unique channel identifier.  Sent as network
 *   order on the link.
 */
typedef struct _jsiPlexLinkHeader {
	unsigned short fragOffset;
	unsigned short totalLength;
	unsigned short messageID;
	unsigned char flags;
	unsigned char seqNum;
	unsigned char ackNum;
	unsigned char _pad; // XXX Window size?
	unsigned short channelID;
} jsiPlexLinkHeader;

/*
 * A packet structure used when interacting with the link.
 *
 * `next' - Link when the object is in a list.
 * `packetSize' - The size of the packet, including the header.  NOTE: This is
 *  not part of the transmission.
 * `header' - The packet header.
 * `body' - The packet body.
 */
typedef struct _jsiPlexLinkPacket {
	struct _jsiPlexLinkPacket *next;
	unsigned short packetSize;
	
	jsiPlexLinkHeader header;
	char body[0];
} jsiPlexLinkPacket;

struct _jsiPlexLink;
struct _jsiPlexInLink;
struct _jsiPlexOutLink;
struct _jsiPlexLinkContext;

/*
 * The set of functions that need to be implemented by a plex link.  These
 * are all wrapped by functions specified below.
 */
typedef struct _jsiPlexLinkFunctions {
	void (*print)(struct _jsiPlexLink *pl, FILE *file);
	void (*printInLink)(struct _jsiPlexInLink *pil, FILE *file);
	void (*printOutLink)(struct _jsiPlexOutLink *pol, FILE *file);

	void (*obtainLink)(struct _jsiPlexLink *pl, int *iLockRoot);
	void (*releaseLink)(struct _jsiPlexLink *pl, int *iLockRoot);

#define JSI_PLEX_TIME_STEP (1000)
	int (*waitForLink)(struct _jsiPlexLink *pl, int ms_timeout);
	void (*signalLink)(struct _jsiPlexLink *pl);

	jsiChannel *(*handle)(struct _jsiPlexInLink *pl);
	int (*write)(struct _jsiPlexOutLink *pl, void *buffer, int len);

	struct _jsiPlexInLink *(*getInLink)(struct _jsiPlexLink *, char *);
	void (*dropInLink)(struct _jsiPlexInLink *);
	struct _jsiPlexOutLink *(*getOutLinkFromFormat)(struct _jsiPlexLink *,
							char *);
	struct _jsiPlexOutLink *(*getOutLinkFromContext)(
		struct _jsiPlexLink *, struct _jsiPlexLinkContext *);
	void (*dropOutLink)(struct _jsiPlexOutLink *);
	int (*compareOutLink)(struct _jsiPlexOutLink *,
			      struct _jsiPlexLinkContext *);
	char *(*encodeOutLink)(struct _jsiPlexOutLink *);
	struct _jsiPlexOutLink *(*getOutLinkFromEncoding)(
		struct _jsiPlexLink *, char *enc);
	int (*processPackets)(struct _jsiPlexOutLink *, int);
} jsiPlexLinkFunctions;

/*
 * Root structure describing a plex link implementation.
 *
 * `next' - Link for plex link in the list.
 * `names' - NULL terminated array of names for this plex link.
 * `ops' - The function that implement this plex link.
 * `maxPacketSize' - The maximum packet size that can be sent via this plex
 *   link.
 */
typedef struct _jsiPlexLink {
	struct _jsiPlexLink *next;
	char **names;
	jsiPlexLinkFunctions *ops;
	unsigned short maxPacketSize;
} jsiPlexLink;

/*
 * Base structure for describing a host that is reachable via the plex link.
 * For example, the context for UDP would be the destination IP address and
 * port.
 *
 * `hashValue' - An integer value that can be used for doing a hash lookup on
 *   this context.  This should match the hashValue stored in the
 *   jsiPlexOutLink that this context matches.
 */
typedef struct _jsiPlexLinkContext {
	int hashValue;
} jsiPlexLinkContext;

enum {
	PILB_UP,
};

enum {
	PILF_UP = (1L << PILB_UP),
};

/*
 * Base structure for an object that accepts connection request on the plex
 * link.
 *
 * `link' - The plex link that implements this input link.
 * `flags' - Flags for the input side of the link.
 * `provider' - The jsiProvider that should will handle new connections.
 */
typedef struct _jsiPlexInLink {
	jsiPlexLink *link;
	unsigned long flags;
	jsiProvider *provider;
} jsiPlexInLink;

enum {
	/*
	 * The link is up and functioning correctly.
	 */
	POLB_UP,
	/*
	 * Indicates that there isn't a separate handler thread for packets
	 * coming in on this link, so they must be handled by the same thread
	 * that sends requests.
	 */
	POLB_BIDIRECTIONAL,
	/*
	 * XXX A hack to allow the normal handler thread to send any disconnect
	 * packets and process the replies.
	 */
	POLB_TEMPORARY_BIDIRECTIONAL,
	/*
	 * Used with BIDIRECTIONAL to indicate when the upper layers are
	 * expecting a packet to come in.
	 */
	POLB_EXPECTING_PACKET,
};

enum {
	POLF_UP = (1L << POLB_UP),
	POLF_BIDIRECTIONAL = (1L << POLB_BIDIRECTIONAL),
	POLF_TEMPORARY_BIDIRECTIONAL = (1L << POLB_TEMPORARY_BIDIRECTIONAL),
	POLF_EXPECTING_PACKET = (1L << POLB_EXPECTING_PACKET),
};

/*
 * Base structure for an object that is used to send bytes to a host on the
 * plex link.  For example, the output link for UDP would hold the
 * destination IP address and port and communicate via sendto.
 *
 * `link' - The plex link that implements this input link.
 * `flags' - Flags for the output side of the link.
 * `hashValue' - An integer value that can be used for doing a hash lookup on
 *   this output link.  This should match the hashValue stored in the
 *   jsiPlexLinkContext that this output link matches.
 * `users' - Reference count for this output link.
 * `sendQueue' - Set of packets that is maintained by this wrapper code, so
 *   packets are preallocated and resending is handled automatically.
 * `sendQueue.usedBuffers' - Array of pointers to packets to be sent over this
 *   link.  These pointers just reference their freeBuffers counterparts.
 * `sendQueue.freeBuffers' - Array of pointers to preallocated packets that
 *   are available to hold data to be sent over the link.
 * `sendQueue.inUse' - The number of buffers that are in use at the moment.
 */
typedef struct _jsiPlexOutLink {
	jsiPlexLink *link;
	unsigned long flags;
	int hashValue;
	int users;
#define	JSI_PLEX_DEFAULT_PACKET_POOL_SIZE 8
	jsiPlexLinkPacket *packetPool;
} jsiPlexOutLink;

/*
 * Print the contents of the given packet header.
 *
 * `file' - The file to output the contents to.
 * `plh' - An initialized plex link packet header.
 */
void jsiPrintPlexLinkHeader(FILE *file, jsiPlexLinkHeader *plh);
/*
 * Print the contents of the given plex link packet.
 *
 * `file' - The file to output the contents to.
 * `plp' - An initialized plex link packet.
 */
void jsiPrintPlexLinkPacket(FILE *file, jsiPlexLinkPacket *plp);

/*
 * Make a plex link implementation available for use.
 *
 * `pl' - An initialized jsiPlexLink object.
 */
void jsiAddPlexLink(jsiPlexLink *pl);
/*
 * Lookup a plex link implementation.
 *
 * `name' - The name of the plex link implementation.
 * returns - The jsiPlexLink object that implements the given name or NULL if
 *  none was found.
 */
jsiPlexLink *jsiFindPlexLink(char *name, int len);

/*
 * Functions for printing out various data structures.
 */
void jsiPrintPlexLink(FILE *file, jsiPlexLink *pl);
void jsiPrintPlexLinks(FILE *file);
void jsiPrintPlexInLink(FILE *file, jsiPlexInLink *pil);
void jsiPrintPlexOutLink(FILE *file, jsiPlexOutLink *pol);

/*
 * Lock the plex link object.
 *
 * `x' - The jsiPlexLink to lock.
 * `iLockRoot' - A pointer to a stack variable, its a kaffe thing.
 */
#define jsiObtainPlexLink(x) _jsiObtainPlexLink(x, &iLockRoot)
/*
 * Unlock the plex link object.
 *
 * `x' - The jsiPlexLink to unlock.
 */
#define jsiReleasePlexLink(x) _jsiReleasePlexLink(x, &iLockRoot)
/*
 * Wait for a plex link object to be signalled or for a timeout.
 *
 * `x' - The jsiPlexLink to wait for.
 * `y' - The number of milliseconds to wait for a signal.
 * returns - True if the timeout fired, false if the object was signalled.
 */
#define jsiWaitForPlexLink(x, y) _jsiWaitForPlexLink(x, y)
/*
 * Signal anyone waiting on the given plex link object.
 *
 * `x' - The jsiPlexLink to signal.
 */
#define jsiSignalPlexLink(x) _jsiSignalPlexLink(x)

#ifndef JSI_INLINE

void _jsiObtainPlexLink(jsiPlexLink *pl, int *iLockRoot);
void _jsiReleasePlexLink(jsiPlexLink *pl, int *iLockRoot);
void _jsiWaitForPlexLink(jsiPlexLink *pl, int ms_timeout);
void _jsiSignalPlexLink(jsiPlexLink *pl);

/*
 * Compare the given output link and context to determine if they point at the
 * same destination.
 *
 * `pol' - The previously constructed output link.
 * `plc' - The link context as given to jsiPlexWindowRecvPacket.
 * returns - True if they point to the same destination, false otherwise.
 */
int jsiComparePlexOutLink(jsiPlexOutLink *pol, jsiPlexLinkContext *plc);
/*
 * Encode the output link's destination into a string.  The encoding
 * should start with the plex link's name followed by a colon and any link
 * data needed to reconstruct the link later on.
 *
 * `pol' - The output link to encode.
 * returns - A string that can be passed to jsiGetPlexOutLinkFromEncoding to
 *  construct a link equivalent to `pol'.
 */
char *jsiEncodePlexOutLink(jsiPlexOutLink *pol);
int jsiProcessPlexOutLinkPackets(jsiPlexOutLink *pol, int ms_timeout);

#else

JSI_INLINE
void _jsiObtainPlexLink(jsiPlexLink *pl, int *iLockRoot)
{
	pl->ops->obtainLink(pl, iLockRoot);
}

JSI_INLINE
void _jsiReleasePlexLink(jsiPlexLink *pl, int *iLockRoot)
{
	pl->ops->releaseLink(pl, iLockRoot);
}

JSI_INLINE
void _jsiWaitForPlexLink(jsiPlexLink *pl, int ms_timeout)
{
	pl->ops->waitForLink(pl, ms_timeout);
}

JSI_INLINE
void _jsiSignalPlexLink(jsiPlexLink *pl)
{
	pl->ops->signalLink(pl);
}

JSI_INLINE
int jsiComparePlexOutLink(jsiPlexOutLink *pol, jsiPlexLinkContext *plc)
{
	return( pol->link->ops->compareOutLink(pol, plc) );
}

JSI_INLINE
char *jsiEncodePlexOutLink(jsiPlexOutLink *pol)
{
	return( pol->link->ops->encodeOutLink(pol) );
}

JSI_INLINE
int jsiProcessPlexOutLinkPackets(jsiPlexOutLink *pol, int ms_timeout)
{
	return( pol->link->ops->processPackets(pol, ms_timeout) );
}

#endif

/*
 * Get an input link that corresponds to the given format.
 *
 * require(pl is locked)
 * `pl' - The plex link implemenation to get the input link from.
 * `format' - A string of the format "<link-name>:<link-specific-data>" to
 *  use when creating the input link.  For example, "udp:localhost/8192" could
 *  specify a UDP connection listening on port 8192 on the current machine.
 *  NOTE: The jsiPlexChannel code strips off the preceding "plex:" so the
 *  format on the command line is actually "plex:<link-name>:<link-data>".
 * returns - A jsiPlexInLink object that is listening on the specified link or
 *  NULL if the link couldn't be set up.
 */
jsiPlexInLink *jsiGetPlexInLink(jsiPlexLink *pl, char *format);
/*
 * Drop a reference to the given input link.
 *
 * `pil' - The input link to release.
 */
void jsiDropPlexInLink(jsiPlexInLink *pil);
/*
 * Get an output link that corresponds to the given format.
 *
 * require(pl is locked)
 * `pl' - The plex link implementation to get the output link from.
 * `format' - A string of the format "<link-name>:<link-specific-data>" to
 *  use when creating the output link.
 * returns - A jsiPlexOutLink object that is pointed at the specified
 *  destination or NULL if the link couldn't be set up.
 */
jsiPlexOutLink *jsiGetPlexOutLinkFromFormat(jsiPlexLink *pl, char *format);
/*
 * Get an output link as described by the given encoding.
 *
 * require(pl is locked)
 * `pl' - The plex link implementation to get the output link from.
 * `enc' - A string encoding as returned by encodeOutLink.
 * returns - A jsiPlexOutLink object that is pointed at the specified
 *  destination or NULL if the link couldn't be set up.
 */
jsiPlexOutLink *jsiGetPlexOutLinkFromEncoding(jsiPlexLink *pl, char *enc);
/*
 * Get an output link from the given link context.  This is typically used by
 * the packet receiver code to create an output path for any reply packets.
 *
 * require(pl is locked)
 * `pl' - The plex link implementation to get the output link from.
 * `context' - A context generated by the plex link implemenation that
 *  specifies another host on the link.
 * returns - A jsiPlexOutLink object that is pointed at the specified
 *  destination or NULL if the link couldn't be set up.
 */
jsiPlexOutLink *jsiGetPlexOutLinkFromContext(jsiPlexLink *pl,
					     jsiPlexLinkContext *context);
/*
 * Drop a reference to the given output link.
 *
 * `pol' - The link to drop the reference count on.
 */
void jsiDropPlexOutLink(jsiPlexOutLink *pol);

/*
 * Execute the main handler loop for the input link.  A typical implementation
 * would wait for packets on the link and call jsiPlexWindowRecvpacket to
 * process any received packets.
 *
 * `pil' - The jsiPlexInLink to process packets from.
 * returns - A newly accepted connection.
 */
jsiChannel *jsiHandlePlexLink(jsiPlexInLink *pil);
/*
 * Allocate a packet from the output's sendQueue, if all of the packets are
 * currently in use then the link is flushed until there is one free or the
 * flush fails and the link is considered down.
 *
 * require(pl is locked)
 * `pol' - The jsiPlexOutLink to allocate the packet from.
 * `size' - The size of the body of the packet.
 * returns - A jsiPlexLinkPacket that is 
 */
jsiPlexLinkPacket *jsiAllocPlexLinkPacket(jsiPlexOutLink *pol, int size);
/*
 * Release a packet back onto the output's sendQueue so that it can be reused.
 *
 * require(pl is locked)
 * `pol' - The jsiPlexOutLink to release the packet to.
 * `plp' - The jsiPlexLinkPacket that was allocated from the given link.
 */
void jsiFreePlexLinkPacket(jsiPlexOutLink *pol, jsiPlexLinkPacket *plp);
/*
 * Write a packet out on the given output link.
 *
 * require(pl is locked)
 * `pol' - The jsiPlexOutLink to send the packet through.
 * `packet' - The packet to send over the link.  NOTE: If this packet isn't
 *  allocated with jsiAllocPlexLinkPacket than it will not be automatically
 *  resent on a flush.  Also, this puts any data in the header in network
 *  order.
 */
int jsiWritePlexLink(jsiPlexOutLink *pol, jsiPlexLinkPacket *packet);
/*
 * Rewrite a packet out on the given output link.
 *
 * require(pl is locked)
 * `pol' - The jsiPlexOutLink to send the packet through.
 * `packet' - The packet to send over the link.  This packet should already
 *  have gone through jsiWritePlexLink.
 */
int jsiRewritePlexLink(jsiPlexOutLink *pol, jsiPlexLinkPacket *packet);

#endif
