/*
 * jsiPlexUDP.h
 * jsiPlexLink for UDP over sockets
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsi.h"
#include "jsiImpl.h"
#include "jsiPlexLink.h"
#include "jsiPlexChannel.h"
#include "jsiPlexWindow.h"
#include "jsiPlexUDP.h"

#include "kaffevm/locks.h"

static struct {
	struct _iLock *lock;
	jsiPlexUDPIn *links;
} inLinks;

static
void obtainUDPLink(jsiPlexLink *pl, int *iLockRoot)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;

	_lockMutex(&pu->lock, iLockRoot);
}

static
void releaseUDPLink(jsiPlexLink *pl, int *iLockRoot)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;

	_unlockMutex(&pu->lock, iLockRoot);
}

static
int waitForUDPLink(jsiPlexLink *pl, int ms_timeout)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;
	int retval;

	retval = _waitCond(&pu->lock, ms_timeout);
	return( retval );
}

static
void signalUDPLink(jsiPlexLink *pl)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;

	_signalCond(&pu->lock);
}

struct recvContext {
	jsiPlexLinkContext plc;
	jsiPlexUDPIn *pui;
	jsiPlexUDPOut *puo;
	int fromlen;
	struct sockaddr_in saddr;
};

static
jsiChannel *handleUDPLink(jsiPlexInLink *pil)
{
	jsiPlexUDPIn *pui = (jsiPlexUDPIn *)pil;
	int iLockRoot, done = 0;
	jsiChannel *retval = 0;
	struct recvContext rc;
	int packetSize;

	rc.pui = pui;
	rc.puo = 0;
	while( !done && !retval )
	{
		rc.fromlen = sizeof(rc.saddr);
		if( !jsiRecvFrom(pui->fd,
				 &pui->recvBuffer.data.header,
				 sizeof(jsiPlexLinkHeader) +
				 JSI_PLEX_UDP_MAX_BODY_SIZE,
				 0,
				 (struct sockaddr *)&rc.saddr,
				 &rc.fromlen,
				 &packetSize) )
		{
			jsiPlexLinkPacket *plp = &pui->recvBuffer.data;
			
			plp->packetSize = packetSize;
			plp->header.fragOffset = ntohs(plp->header.fragOffset);
			plp->header.totalLength =
				ntohs(plp->header.totalLength);
			plp->header.messageID = ntohs(plp->header.messageID);
			plp->header.channelID = ntohs(plp->header.channelID);
			rc.plc.hashValue = rc.saddr.sin_addr.s_addr;
		        jsiObtainPlexLink(pil->link);
			retval = jsiPlexWindowRecvPacket(pil, 0, &rc.plc, plp);
			jsiReleasePlexLink(pil->link);
		}
		else
		{
			/* Error receiving, link is down? */
			pil->flags &= ~PILF_UP;
			done = 1;
		}
	}
	return( retval );
}

static
int processUDPPackets(jsiPlexOutLink *pol, int ms_timeout)
{
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;
	int packetSize, done = 0, retval = 1;
	struct recvContext rc;
	int iLockRoot;

	memset(&rc, 0, sizeof(rc));
	rc.puo = puo;
	while( !done )
	{
		rc.fromlen = sizeof(rc.saddr);
		if( !jsiRecvFromTimeout(puo->fd,
					&puo->recvBuffer.data.header,
					sizeof(jsiPlexLinkHeader) +
					JSI_PLEX_UDP_MAX_BODY_SIZE,
					0,
					(struct sockaddr *)&rc.saddr,
					&rc.fromlen,
					ms_timeout,
					&packetSize) )
		{
			jsiPlexLinkPacket *plp = &puo->recvBuffer.data;
			jsiChannel *jc;
			
			plp->packetSize = packetSize;
			plp->header.fragOffset = ntohs(plp->header.fragOffset);
			plp->header.totalLength =
				ntohs(plp->header.totalLength);
			plp->header.messageID = ntohs(plp->header.messageID);
			plp->header.channelID = ntohs(plp->header.channelID);
			rc.plc.hashValue = rc.saddr.sin_addr.s_addr;
			jsiObtainPlexLink(pol->link);
			jc = jsiPlexWindowRecvPacket(0, pol, &rc.plc, plp);
			jsiReleasePlexLink(pol->link);
			if( jc )
				jsiDeleteChannel(jc);
			if( !(pol->flags & POLF_EXPECTING_PACKET) )
				done = 1;
		}
		else
		{
			done = 1;
			retval = 0;
		}
	}
	return( retval );
}

static
int writeUDPLink(jsiPlexOutLink *pol, void *buffer, int len)
{
	int tolen = sizeof(struct sockaddr_in), retval = 0;
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
	int rc;
#endif
	
	if( (
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
	     rc =
#endif
	     jsiSendTo(puo->fd, buffer, len, 0,
		       (struct sockaddr *)&puo->dest, tolen )) == 0 )
	{
		retval = 1;
	}
	else
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr,
			"Got error %d on fd %d, taking down link.\n",
			rc,
			puo->fd);
#endif
		pol->flags &= ~POLF_UP;
	}
	return( retval );
}

static
jsiPlexInLink *getUDPIn(jsiPlexLink *pl, char *format)
{
	parsedString protocol, host;
	jsiPlexInLink *retval = 0;
	int port;

	if( parseString(format,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_NonEmptyString, &host,
			SPO_Expect, "/",
			SPO_Integer, &port,
			SPO_End) )
	{
		jsiPlexUDPIn *pui;
		int sock = -1;
		
		if( jsiSocket(AF_INET, SOCK_DGRAM, 0, &sock) == 0 )
		{
			struct sockaddr_in saddr;
			int on_off = 1;

			memset(&saddr, 0, sizeof(saddr));
			saddr.sin_family = AF_INET;
			saddr.sin_addr.s_addr = htonl(INADDR_ANY);
			saddr.sin_port = htons(port);
			jsiSetSockOpt(sock, SOL_SOCKET, SO_REUSEADDR,
				      &on_off, sizeof( int ));
			if( !jsiBind(sock,
				     (struct sockaddr *)&saddr,
				     sizeof(saddr)) )
			{
			}
			else
			{
				jsiClose(sock);
				sock = -1;
			}
		}
		if( (sock != -1) &&
		    (pui = (jsiPlexUDPIn *)jsiMalloc(sizeof(jsiPlexUDPIn))) )
		{
			int iLockRoot;
			
			pui->link.link = pl;
			pui->link.flags = PILF_UP;
			pui->link.provider = 0;
			pui->fd = sock;
			pui->srcPort = port;
			lockMutex(&inLinks);
			pui->next = inLinks.links;
			inLinks.links = pui;
			unlockMutex(&inLinks);
			retval = &pui->link;
		}
	}
	return( retval );
}

static
void dropUDPIn(jsiPlexInLink *pil)
{
	jsiPlexUDPIn *pui = (jsiPlexUDPIn *)pil;

	/* XXX Ugh, destruction of the input link needs to be handled better */
	jsiClose(pui->fd);
	jsiFree(pil);
}

static
jsiPlexOutLink *getUDPOut(jsiPlexLink *pl, struct sockaddr_in *saddr)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;
	jsiPlexUDPOut *puo, *match = 0;
	jsiPlexOutLink *retval = 0;

	puo = pu->outLinks;
	while( puo && !match )
	{
		if( !memcmp(&puo->dest, saddr, sizeof(struct sockaddr_in)) )
			match = puo;
		puo = puo->next;
	}
	if( match )
	{
		match->link.users++;
		retval = &match->link;
	}
	else if( (puo = (jsiPlexUDPOut *)jsiMalloc(sizeof(jsiPlexUDPOut))) )
	{
		puo->link.link = pl;
		puo->link.hashValue = saddr->sin_addr.s_addr;
		puo->link.users = 1;
		puo->link.flags = POLF_UP;
		puo->next = pu->outLinks;
		pu->outLinks = puo;
		puo->dest = *saddr;
		retval = &puo->link;
	}
	return( retval );
}

static
jsiPlexOutLink *getUDPOutFromFormat(jsiPlexLink *pl, char *format)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;
	parsedString protocol, host;
	jsiPlexOutLink *retval = 0;
	int port;

	if( parseString(format,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_NonEmptyString, &host,
			SPO_Expect, "/",
			SPO_Integer, &port,
			SPO_End) )
	{
		struct sockaddr_in host_addr;
		struct hostent *host_ent;
		int valid = 1;
		char save;

		save = host.data[host.len];
		host.data[host.len] = '\0';
		memset(&host_addr, 0, sizeof(host_addr));
#if defined(BSD44)
		host_addr.sin_len = sizeof(host_addr);
#endif
		host_addr.sin_family = AF_INET;
		if( !jsiGetHostByName(host.data, &host_ent) )
		{
			memcpy((char *)&host_addr.sin_addr.s_addr,
			       host_ent->h_addr,
			       host_ent->h_length);
		}
		else if( (host_addr.sin_addr.s_addr =
			  inet_addr(host.data)) == -1 )
		{
			/* Its not a number address either */
			valid = 0;
		}
		host_addr.sin_port = htons(port);
		host.data[host.len] = save;
		if( valid )
		{
			if( (retval = getUDPOut(pl, &host_addr)) )
			{
				require(pu->fd >= 0);
				require(pu->fd < FD_SETSIZE);
				
				((jsiPlexUDPOut *)retval)->fd = pu->fd;
				((jsiPlexUDPOut *)retval)->srcPort =
					pu->srcPort;
			}
		}
	}
	return( retval );
}

static
jsiPlexOutLink *getUDPOutFromContext(jsiPlexLink *pl, jsiPlexLinkContext *plc)
{
	struct recvContext *rc = (struct recvContext *)plc;
	jsiPlexOutLink *retval;

	if( (retval = getUDPOut(pl, &rc->saddr)) )
	{
		require(rc->pui || rc->puo);
		
		if( rc->pui )
		{
			require(rc->pui->fd >= 0);
			require(rc->pui->fd < FD_SETSIZE);
			
			((jsiPlexUDPOut *)retval)->fd = rc->pui->fd;
			((jsiPlexUDPOut *)retval)->srcPort = rc->pui->srcPort;
		}
		else
		{
			require(rc->puo->fd >= 0);
			require(rc->puo->fd < FD_SETSIZE);
			
			((jsiPlexUDPOut *)retval)->fd = rc->puo->fd;
			((jsiPlexUDPOut *)retval)->srcPort = rc->puo->srcPort;
		}
	}
	return( retval );
}

static
void dropUDPOut(jsiPlexOutLink *pol)
{
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;

	require(puo != 0);
	require(puo->link.users > 0);

	puo->link.users--;
	if( puo->link.users == 0 )
	{
		jsiPlexLinkPacket *plp, *plp_next;
		jsiPlexUDPOut *curr, **prev;
		
		curr = ((jsiPlexUDP *)pol->link)->outLinks;
		prev = &((jsiPlexUDP *)pol->link)->outLinks;
		while( curr != puo )
		{
			prev = &curr->next;
			curr = curr->next;
		}
		*prev = curr->next;
		plp = pol->packetPool;
		while( plp )
		{
			plp_next = plp->next;
			jsiFree(plp);
			plp = plp_next;
		}
		jsiFree(puo);
	}
}

int compareUDPOut(jsiPlexOutLink *pol, jsiPlexLinkContext *plc)
{
	struct recvContext *rc = (struct recvContext *)plc;
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;
	int retval = 0;

	if( (puo->dest.sin_addr.s_addr == rc->saddr.sin_addr.s_addr) &&
	    (puo->dest.sin_port == rc->saddr.sin_port) )
	{
		retval = 1;
	}
	return( retval );
}

char *encodeUDPOut(jsiPlexOutLink *pol)
{
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;
	char *retval = 0;

	if( (retval = jsiMalloc(3 + 1 +
				15 + 1 +
				5 +
				1)) )
	{
		sprintf(retval,
			"udp:%d.%d.%d.%d/%d/%d",
			((char *)&puo->dest.sin_addr.s_addr)[0] & 0xff,
			((char *)&puo->dest.sin_addr.s_addr)[1] & 0xff,
			((char *)&puo->dest.sin_addr.s_addr)[2] & 0xff,
			((char *)&puo->dest.sin_addr.s_addr)[3] & 0xff,
			ntohs(puo->dest.sin_port),
			puo->srcPort);
	}
	return( retval );
}

static
jsiPlexOutLink *getUDPOutFromEncoding(jsiPlexLink *pl, char *enc)
{
	parsedString protocol, host;
	jsiPlexOutLink *retval = 0;
	int srcPort,port;

	if( parseString(enc,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_NonEmptyString, &host,
			SPO_Expect, "/",
			SPO_Integer, &port,
			SPO_Expect, "/",
			SPO_Integer, &srcPort,
			SPO_End) )
	{
		struct sockaddr_in host_addr;
		char save;
		
		save = host.data[host.len];
		host.data[host.len] = '\0';
		memset(&host_addr, 0, sizeof(host_addr));
#if defined(BSD44)
		host_addr.sin_len = sizeof(host_addr);
#endif
		host_addr.sin_family = AF_INET;
		if( (host_addr.sin_addr.s_addr = inet_addr(host.data)) != -1 )
		{
			host_addr.sin_port = htons(port);
			if( (retval = getUDPOut(pl, &host_addr)) )
			{
				jsiPlexUDPIn *pui;
				int iLockRoot;

				lockMutex(&inLinks);
				pui = inLinks.links;
				while( pui && (pui->srcPort != srcPort) )
				{
					pui = pui->next;
				}
				unlockMutex(&inLinks);
				if( pui )
				{
					((jsiPlexUDPOut *)retval)->fd =
						pui->fd;
				}
				else
				{
					dropUDPOut(retval);
					retval = 0;
				}
			}
		}
		host.data[host.len] = save;
	}
	return( retval );
}

jsiPlexLinkFunctions jsiPlexUDPFunctions = {
	0,
	0,
	0,
	obtainUDPLink,
	releaseUDPLink,
	waitForUDPLink,
	signalUDPLink,
	handleUDPLink,
	writeUDPLink,
	getUDPIn,
	dropUDPIn,
	getUDPOutFromFormat,
	getUDPOutFromContext,
	dropUDPOut,
	compareUDPOut,
	encodeUDPOut,
	getUDPOutFromEncoding,
	processUDPPackets,
};

static
char *plexUDPNames[] = {
	"udp",
	0
};

jsiPlexLink *jsiCreatePlexUDP()
{
	jsiPlexLink *retval = 0;
	int sock;

	if( jsiSocket(AF_INET, SOCK_DGRAM, 0, &sock) == 0 )
	{
		jsiPlexUDP *pu;

		if( (pu = (jsiPlexUDP *)jsiMalloc(sizeof(jsiPlexUDP))) )
		{
			pu->link.next = 0;
			pu->link.names = plexUDPNames;
			pu->link.ops = &jsiPlexUDPFunctions;
			pu->link.maxPacketSize = JSI_PLEX_UDP_MAX_BODY_SIZE;
			pu->lock = 0;
			pu->fd = sock;
			pu->outLinks = 0;
			retval = &pu->link;
		}
		else
		{
			jsiClose(sock);
		}
	}
	return( retval );
}
