/*
 * jsiPlexWindow.c
 * Sliding window implementation for plex channels
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsi.h"
#include "jsiImpl.h"
#include "jsiPlexChannel.h"
#include "jsiPlexWindow.h"

#define JSI_PLEX_INITIAL_RESEND_TIMEOUT (1 * 1000)
#define JSI_PLEX_SYN_TIMEOUT (75 * 1000)
#define JSI_PLEX_FIN_TIMEOUT (75 * 1000)
#define JSI_PLEX_RECV_TIMEOUT (10 * 1000)
#define JSI_PLEX_POKE_TIMEOUT (60 * 1000)
//#define JSI_TIME_WAIT_TIMEOUT 2 * 1000
#define JSI_TIME_WAIT_TIMEOUT 100

/* String representations of the window states */
static
char *stateNames[] = {
	"",
	"closed",
	"syn_sent",
	"syn_recv",
	"connected",
	"fin_wait_1",
	"fin_wait_2",
	"closing",
	"time_wait",
	"close_wait",
	"last_ack",
};

void jsiPrintPlexWindow(FILE *file, jsiPlexWindow *pw)
{
	jsiPlexLinkPacket *plp;
	int lpc;

	require(file != 0);
	require(pw != 0);
	
	fprintf(file,
		"jsiPlexWindow(%p)[state=%s; link=%p; sendWindow={",
		pw,
		stateNames[pw->state],
		pw->link);
	plp = pw->sendWindow;
	while( plp )
	{
		fprintf(file,
			" %p(%d)={",
			plp,
			plp->packetSize);
		jsiPrintPlexLinkPacket(file, plp);
		fprintf(file, "} ");
		plp = plp->next;
	}
	fprintf(file, "}; recvWindow={");
	for( lpc = 0; lpc < JSI_PLEX_WINDOW_SIZE; lpc++ )
	{
		fprintf(file,
			" %d ",
			pw->recvWindow[lpc]);
	}
	fprintf(file,
		"}; lastAckReceived=%d; lastFrameSent=%d; "
		"nextFrameExpected=%d]",
		pw->lastAckReceived,
		pw->lastFrameSent,
		pw->nextFrameExpected);
}

/*
 * Conjure up a packet sequence number.
 *
 * returns - A new sequence number.
 */
static
unsigned char conjureSeqNum(void)
{
	unsigned char retval;

#ifdef JSI_RANDOM_SEQ_NUMBERS
	retval = random();
#else
	retval = 20;
#endif
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
	fprintf(stderr, "Conjured sequence number: %d\n", retval);
#endif
	return( retval );
}

/*
 * Test a sequence number to see if its in the sliding window.
 *
 * `seqNum' - The sequence number to test.
 * `seqMin' - The minimum sequence number in the window.
 * `seqMax' - The maximum sequence number in the window.  NOTE: It does not
 *  have to be true that seqMin < seqMax since the function will handle the
 *  wrap around properly.
 * returns - True if seqNum is inside the window, false otherwise.
 */
static
int inPlexWindow(unsigned int seqNum, unsigned int seqMin, unsigned int seqMax)
{
	unsigned int pos, maxpos;
	int retval = 0;

	pos = seqNum - seqMin;
	maxpos = seqMax - seqMin + 1;
	if( pos < maxpos )
		retval = 1;
	return( retval );
}

void jsiPlexWindowInit(jsiPlexWindow *pw)
{
	int seqNum, lpc;
	
	require(pw != 0);

	pw->state = PWS_CLOSED;
	pw->link = 0;
	pw->sendWindow = 0;
	pw->sendWindowTail = &pw->sendWindow;
	pw->resendTimeout = JSI_PLEX_INITIAL_RESEND_TIMEOUT;
	for( lpc = 0; lpc < JSI_PLEX_WINDOW_SIZE; lpc++ )
	{
		pw->recvWindow[lpc] = 0;
	}
	seqNum = conjureSeqNum();
	/* Start off in a stable state. */
	pw->lastAckReceived = seqNum;
	pw->lastFrameSent = seqNum;
	/*
	 * This value is just ignored until we're connected to the peer window.
	 */
	pw->nextFrameExpected = 0;
}

void jsiPlexWindowKill(jsiPlexWindow *pw)
{
	require(pw != 0);
}

/*
 * Drop the connection to the peer.
 *
 * require(pw->link->link is locked)
 * `pw' - An initialized jsiPlexWindow.
 * `forceful' - Forcefully drop the connection, don't negotiate with the other
 *  side, just set the window to the closed state and remove any queued
 *  packets.
 */
static
void dropLink(jsiPlexWindow *pw, int forceful)
{
	jsiPlexLinkPacket *plp, *plp_next;

	require(pw != 0);
	
	if( forceful )
		pw->state = PWS_CLOSED;
	plp = pw->sendWindow;
	while( plp )
	{
		require(forceful);
		
		plp_next = plp->next;
		jsiFreePlexLinkPacket(pw->link, plp);
		plp = plp_next;
	}
	pw->sendWindow = 0;
	pw->sendWindowTail = &pw->sendWindow;
	pw->link = 0; /* Its freed by jsiDisconnectPlexChannel */
	
	ensure(pw->state == PWS_CLOSED);
}

/*
 * Resend the any packets still in the send window.
 */
static
int resendPlexWindow(jsiPlexWindow *pw)
{
	jsiPlexLinkPacket *plp;
	jsiPlexOutLink *pol;
	int retval = 1;
	int iLockRoot;
	
	require(pw != 0);

	pol = pw->link;
	/* Resend the packets */
	jsiObtainPlexLink(pol->link);
	plp = pw->sendWindow;
	while( retval && plp )
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
		fprintf(stderr, "Resending packet ");
		jsiPrintPlexLinkPacket(stderr, plp);
		fprintf(stderr, " thru ");
		jsiPrintPlexWindow(stderr, pw);
		fprintf(stderr, "\n");
#endif
		retval = jsiRewritePlexLink(pol, plp);
		plp = plp->next;
	}
	jsiReleasePlexLink(pol->link);

	return( retval );
}

/*
 * Free any packets in the send window that have been ACK'ed.
 */
static
void flushPlexWindow(jsiPlexWindow *pw, unsigned char ack)
{
	jsiPlexLinkPacket *curr, *next;
	
	require(inPlexWindow(ack,
			     pw->lastAckReceived + 1,
			     pw->lastFrameSent));

	curr = pw->sendWindow;
	while( curr && (curr->header.seqNum != ((ack + 1) & 0xff)) )
	{
		next = curr->next;
		jsiFreePlexLinkPacket(pw->link, curr);
		curr = next;
	}
	if( curr )
	{
		pw->sendWindow = curr;
	}
	else
	{
		pw->sendWindow = 0;
		pw->sendWindowTail = &pw->sendWindow;
	}
}

/*
 * Send a poke message to the other side to test the connection liveness.
 */
static
void pokePlexWindow(jsiPlexWindow *pw, unsigned short channelID, int ack)
{
	jsiPlexLinkPacket pokePacket;
	jsiPlexOutLink *pol;
	
	require(pw != 0);
	require(channelID != 0);

	pol = pw->link;
	pokePacket.packetSize = sizeof(jsiPlexLinkHeader);
	pokePacket.header.fragOffset = 0;
	pokePacket.header.totalLength = 0;
	pokePacket.header.messageID = 0;
	pokePacket.header.flags = PHF_POKE;
	if( ack )
		pokePacket.header.flags |= PHF_ACK;
	pokePacket.header.seqNum = 0;
	pokePacket.header.ackNum = 0;
	pokePacket.header.channelID = channelID;
	jsiWritePlexLink(pol, &pokePacket);
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
	fprintf(stderr, "Sending poke on channel %d.\n", channelID);
#endif
}

int jsiPlexWindowWaitFor(jsiPlexWindow *pw,
			 plex_window_state_t expected_state,
			 unsigned long fail_state,
			 int timeout)
{
	int iLockRoot, retval = 0;

	require(pw != 0);
	require(expected_state > PWS_MIN);
	require(expected_state < PWS_MAX);
	require(timeout > 0);
	
	if( pw->link->flags & POLF_BIDIRECTIONAL )
		pw->link->flags |= POLF_EXPECTING_PACKET;
	jsiLock(pw);
	while( (timeout > 0) && !retval )
	{
		if( pw->state == expected_state )
		{
			/* Got the proper reply */
			retval = 1;
		}
		else if( (1L << pw->state) & fail_state )
		{
			/* Failure, sadness */
			timeout = 0;
		}
		else if( pw->link->flags & POLF_EXPECTING_PACKET )
		{
			jsiUnlock(pw);
			if( !jsiProcessPlexOutLinkPackets(pw->link,
							  pw->resendTimeout) )
			{
				timeout -= pw->resendTimeout;
			}
			jsiLock(pw);
		}
		else
		{
			if( !jsiWait(pw, pw->resendTimeout) )
				timeout -= pw->resendTimeout;
		}
		if( (timeout > 0) && !retval )
		{
			int link_up;
			
			jsiUnlock(pw);
			link_up = resendPlexWindow(pw);
			jsiLock(pw);
			if( !link_up )
			{
				/* Link is dead */
				timeout = 0;
			}
		}
	}
	pw->link->flags &= ~POLF_EXPECTING_PACKET;
	jsiUnlock(pw);
	
	ensure((timeout <= 0) ||
	       (pw->state == expected_state) ||
	       ((1L << pw->state) & fail_state));
	
	return( retval );
}

char *jsiPlexWindowWaitForMessage(jsiPlexChannel *pc, int *bufferSize)
{
	int pokeTimeout = JSI_PLEX_POKE_TIMEOUT;
	int timeout = JSI_PLEX_RECV_TIMEOUT;
	jsiPlexOutLink *pol;
	jsiPlexWindow *pw;
	char *retval = 0;
	int iLockRoot;

	require(pc != 0);
	require(bufferSize != 0);

	pw = &pc->window;
	pol = pw->link;
	jsiLock(pw);
	if( pw->link->flags & POLF_TEMPORARY_BIDIRECTIONAL )
	{
		/*
		 * Something is waiting for a message so the channel has made
		 * it out of the provider thread.
		 */
		pw->link->flags &= ~(POLF_BIDIRECTIONAL |
				     POLF_TEMPORARY_BIDIRECTIONAL);
	}
	if( pw->link->flags & POLF_BIDIRECTIONAL )
		pw->link->flags |= POLF_EXPECTING_PACKET;
	while( (timeout > 0) &&
	       (pw->state == PWS_CONNECTED) &&
	       (!pc->recvBuffer.start || pc->recvBuffer.remaining) )
	{
		if( pw->link->flags & POLF_EXPECTING_PACKET )
		{
			jsiUnlock(pw);
			if( jsiProcessPlexOutLinkPackets(pw->link,
							 pw->resendTimeout) )
			{
			}
			else if( pw->sendWindow != 0 )
			{
				/* Still waiting for ACK's... */
				timeout -= pw->resendTimeout;
			}
			else
			{
				/*
				 * Nothing to send but we should eventually
				 * test to make sure the channel is still
				 * alive.
				 */
				pokeTimeout -= pw->resendTimeout;
			}
			jsiLock(pw);
			pw->link->flags |= POLF_EXPECTING_PACKET;
		}
		else
		{
			if( !jsiWait(pw, pw->resendTimeout) )
			{
				if( pw->sendWindow != 0 )
					timeout -= pw->resendTimeout;
				else
					pokeTimeout -= pw->resendTimeout;
			}
		}
		if( (timeout > 0) && (pw->state == PWS_CONNECTED) &&
		    pw->sendWindow &&
		    (!pc->recvBuffer.start || pc->recvBuffer.remaining) )
		{
			int link_up;
			
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
			fprintf(stderr,
				"WaitForMessage timeout %d on ",
				timeout);
			jsiPrintPlexChannel(stderr, pc);
			fprintf(stderr, "\n");
#endif
			/* Timeout */
			jsiUnlock(pw);
			link_up = resendPlexWindow(pw);
			jsiLock(pw);
			if( !link_up )
			{
				/* Link is dead */
				timeout = 0;
			}
		}
		else if( (pokeTimeout <= 0) && (pw->state == PWS_CONNECTED) )
		{
			int retries = 3;

			/* Retry sending the poke a few times. */
			while( retries && !pw->poked )
			{
				jsiUnlock(pw);
				jsiObtainPlexLink(pw->link->link);
				pokePlexWindow(pw, pc->channelID, 0);
				jsiReleasePlexLink(pw->link->link);
				if( pw->link->flags & POLF_EXPECTING_PACKET )
				{
					jsiProcessPlexOutLinkPackets(
						pw->link,
						JSI_PLEX_POKE_TIMEOUT);
					pw->link->flags |=
						POLF_EXPECTING_PACKET;
				}
				jsiLock(pw);
				if( !(pw->link->flags &
				      POLF_EXPECTING_PACKET) )
				{
					jsiWait(pw, JSI_PLEX_POKE_TIMEOUT);
				}
				retries--;
			}
			if( !pw->poked )
			{
				/* No reply, assume the channel is dead. */
				timeout = 0;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
				fprintf(stderr, "Poke timed out for channel ");
				jsiPrintPlexChannel(stderr, pc);
				fprintf(stderr, "\n");
#endif
			}
			else
			{
				/*
				 * Got a reply, the response is just taking
				 * a long time.
				 */
				pw->poked = 0;
				pokeTimeout = JSI_PLEX_POKE_TIMEOUT;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
				fprintf(stderr, "Connection still alive ");
				jsiPrintPlexChannel(stderr, pc);
				fprintf(stderr, "\n");
#endif
			}
		}
	}
	if( pc->recvBuffer.start && (pc->recvBuffer.remaining == 0) )
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
		fprintf(stderr,
			"Received message (data=%p; size=%d) for ",
			pc->recvBuffer.start,
			pc->recvBuffer.size);
		jsiPrintPlexChannel(stderr, pc);
		fprintf(stderr, "\n");
#endif
		retval = pc->recvBuffer.start;
		*bufferSize = pc->recvBuffer.size;
		pc->recvBuffer.start = 0;
		pc->recvBuffer.size = 0;
	}
	else if( timeout <= 0 )
	{
		/* There was a timeout error, assume the channel is dead */
		*bufferSize = 0;
		pw->state = PWS_CLOSED;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr, "WaitForMessage closed on ");
		jsiPrintPlexChannel(stderr, pc);
		fprintf(stderr, "\n");
#endif
	}
	jsiUnlock(pw);
	return( retval );
}

int jsiPlexWindowConnect(jsiPlexWindow *pw, jsiPlexOutLink *pol, int channel)
{
	int retval = 0, iLockRoot;
	jsiPlexLinkPacket *plp;

	require(pw != 0);
	require(pw->state == PWS_CLOSED);
	require(pol != 0);
	
	jsiObtainPlexLink(pol->link);
	if( (plp = jsiAllocPlexLinkPacket(pol, 0)) )
	{
		plp->header.fragOffset = 0;
		plp->header.totalLength = 0;
		plp->header.messageID = 0;
		plp->header.flags = PHF_SYN;
		plp->header.channelID = channel;
		pw->link = pol;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
		fprintf(stderr, "Sending SYN ");
		jsiPrintPlexLinkPacket(stderr, plp);
		fprintf(stderr, " from ");
		if( pw )
			jsiPrintPlexWindow(stderr, pw);
		else
			fprintf(stderr, "(unknown)");
		fprintf(stderr, "\n");
#endif
		/*
		 * Send off the SYN and wait for the receiver code to finish
		 * the connection setup.
		 */
		if( jsiPlexWindowSendPacket(pw, plp) )
		{
			pw->state = PWS_SYN_SENT;
			pol->flags |= POLF_BIDIRECTIONAL;
			jsiReleasePlexLink(pol->link);
			if( jsiPlexWindowWaitFor(pw,
						 PWS_CONNECTED,
						 (1L << PWS_CLOSED) |
						 (1L << PWS_CLOSE_WAIT),
						 JSI_PLEX_SYN_TIMEOUT) )
			{
				retval = 1;
			}
			jsiObtainPlexLink(pol->link);
		}
		if( !retval )
		{
			/* Somethings wrong, forcefully drop the connection */
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
			fprintf(stderr, "Error connecting window ");
			jsiPrintPlexWindow(stderr, pw);
			fprintf(stderr, "\n");
#endif
			switch( pw->state )
			{
			case PWS_CLOSE_WAIT:
				jsiReleasePlexLink(pol->link);
				jsiPlexWindowDisconnect(pw, channel);
				jsiObtainPlexLink(pol->link);
				break;
			default:
				break;
			}
			dropLink(pw, 1);
		}
	}
	jsiReleasePlexLink(pol->link);
	return( retval );
}

void jsiPlexWindowDisconnect(jsiPlexWindow *pw, int channel)
{
	jsiPlexOutLink *pol = pw->link;
	jsiPlexLinkPacket *plp;
	int iLockRoot;

	require(pw != 0);

#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
	fprintf(stderr, "Disconnecting window ");
	jsiPrintPlexWindow(stderr, pw);
	fprintf(stderr, "\n");
#endif
	switch( pw->state )
	{
	case PWS_CLOSED:
		/* Already disconnected, but make sure */
		jsiObtainPlexLink(pol->link);
		dropLink(pw, 1);
		jsiReleasePlexLink(pol->link);
		break;
	case PWS_SYN_RECV:
	case PWS_CONNECTED:
	case PWS_CLOSE_WAIT:
		jsiObtainPlexLink(pol->link);
		if( (plp = jsiAllocPlexLinkPacket(pw->link, 0)) )
		{
			plp->packetSize = sizeof(jsiPlexLinkHeader);
			plp->header.fragOffset = 0;
			plp->header.totalLength = 0;
			plp->header.messageID = 0;
			plp->header.flags = PHF_FIN;
			plp->header.channelID = channel;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
			fprintf(stderr, "Sending FIN for window ");
			jsiPrintPlexWindow(stderr, pw);
			fprintf(stderr, "\n");
#endif
			/*
			 * Send the FIN and wait for the receiver code to
			 * handle the disconnect.
			 */
			if( jsiPlexWindowSendPacket(pw, plp) )
			{
				require((pw->state == PWS_CLOSE_WAIT) ||
					(pw->state == PWS_SYN_RECV) ||
					(pw->state == PWS_CONNECTED));
				
				if( pw->state == PWS_CLOSE_WAIT )
					pw->state = PWS_LAST_ACK;
				else
					pw->state = PWS_FIN_WAIT_1;
				jsiReleasePlexLink(pol->link);
				jsiPlexWindowWaitFor(pw,
						     PWS_TIME_WAIT,
						     (1L << PWS_CLOSED),
						     JSI_PLEX_FIN_TIMEOUT);
				jsiLock(pw);
				if( pw->state == PWS_TIME_WAIT )
				{
					/*
					 * Do the TIME_WAIT, since we might
					 * not be an OS we don't want to hang
					 * around for an ultra long time like
					 * TCP does, so we cheat and get out
					 * real fast.
					 */
					jsiWait(pw, JSI_TIME_WAIT_TIMEOUT);
					pw->state = PWS_CLOSED;
				}
				else
				{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
					fprintf(stderr,
						"Passed the time_wait\n");
#endif
				}
				jsiUnlock(pw);
				jsiObtainPlexLink(pol->link);
			}
		}
		/* Drop it like a bad habit */
		dropLink(pw, 1);
		jsiReleasePlexLink(pol->link);
		
		ensure(pw->state == PWS_CLOSED);
		break;
	default:
		// require(0);
		break;
	}
}

static
void primeAckPacket(jsiPlexWindow *pw, jsiPlexLinkPacket *plp)
{
	int slot;
	
	/* Find any new packets to acknowledge */
	slot = pw->nextFrameExpected % JSI_PLEX_WINDOW_SIZE;
	while( pw->recvWindow[slot] > 0 )
	{
		pw->recvWindow[slot] = 0;
		pw->nextFrameExpected++;
		slot = pw->nextFrameExpected % JSI_PLEX_WINDOW_SIZE;
	}
	
	/* Append any ACK information to the packet */
	plp->header.flags |= PHF_ACK;
	plp->header.ackNum = pw->nextFrameExpected - 1;
}

int jsiPlexWindowSendPacket(jsiPlexWindow *pw, jsiPlexLinkPacket *plp)
{
	int retval = 0;

	plp->header.seqNum = ++pw->lastFrameSent;
	if( !(plp->header.flags & PHF_SYN) )
	{
		primeAckPacket(pw, plp);
	}
	/* Record the send in the window so we can free the packet later on. */
	*pw->sendWindowTail = plp;
	pw->sendWindowTail = &plp->next;
	plp->next = 0;

	/* Finally, write out the packet */
	retval = jsiWritePlexLink(pw->link, plp);
	return( retval );
}

int jsiPlexWindowSendAckPacket(jsiPlexWindow *pw, jsiPlexLinkPacket *plp)
{
	jsiPlexLinkPacket ackPacket;
	jsiPlexOutLink *pol;
	int retval = 1;

	require(pw != 0);
	require(plp != 0);

	pol = pw->link;
	ackPacket.packetSize = sizeof(jsiPlexLinkHeader);
	ackPacket.header.fragOffset = 0;
	ackPacket.header.totalLength = 0;
	ackPacket.header.messageID = 0;
	ackPacket.header.flags = plp->header.flags;
	if( plp->header.flags & PHF_SYN )
		ackPacket.header.seqNum = ++pw->lastFrameSent;
	else
		ackPacket.header.seqNum = 0;
	primeAckPacket(pw, &ackPacket);
	ackPacket.header.channelID = plp->header.channelID;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
	fprintf(stderr, "Sending ACK ");
	jsiPrintPlexLinkPacket(stderr, &ackPacket);
	fprintf(stderr, " from ");
	jsiPrintPlexWindow(stderr, pw);
	fprintf(stderr, "\n");
#endif
	/*
	 * We're sending the ACK manually so the plex link won't help with
	 * resending.
	 */
	if( !(retval = jsiWritePlexLink(pol, &ackPacket)) )
	{
		int tries = 5;
		
		while( !jsiRewritePlexLink(pol, &ackPacket) && tries )
		{
			tries--;
		}
		if( tries == 0 )
		{
			retval = 0;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
			fprintf(stderr, "ACK send for ");
			jsiPrintPlexLinkPacket(stderr, &ackPacket);
			fprintf(stderr, " failed\n");
#endif
		}
	}
	return( retval );
}

void plexSetup(jsiPlexLink *pl,
	       jsiPlexChannel *pc,
	       jsiPlexLinkContext *plc,
	       jsiPlexLinkPacket *plp)
{
	jsiPlexOutLink *pol;
	
	if( plp->header.flags & PHF_ACK )
	{
		if( pc )
		{
			jsiPlexWindow *pw = &pc->window;
			
			/*
			 * We requested the connection, or its a duplicate
			 * SYN|ACK on the server.
			 */
			if( pw->state == PWS_SYN_SENT )
			{
				int iLockRoot;
				
				flushPlexWindow(pw, plp->header.ackNum);
				pw->lastAckReceived = plp->header.ackNum;
				pw->nextFrameExpected = plp->header.seqNum + 1;
				plp->header.flags = 0;
				if( jsiPlexWindowSendAckPacket(pw, plp) )
				{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
					fprintf(stderr, "Channel ");
					jsiPrintPlexChannel(stderr, pc);
					fprintf(stderr, " is now connected\n");
#endif
					pw->state = PWS_CONNECTED;
				}
				else
				{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
					fprintf(stderr,
						"Error sending ACK for ");
					jsiPrintPlexChannel(stderr, pc);
					fprintf(stderr, " when synching\n");
#endif
					pw->state = PWS_CLOSED;
					pw->link = 0;
				}
				jsiLock(pw);
				pw->link->flags &= ~POLF_EXPECTING_PACKET;
				jsiSignal(pw);
				jsiUnlock(pw);
			}
			else
			{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
				fprintf(stderr,
					"Channel %d already connected\n",
					pc->channelID);
#endif
				plp->header.flags = 0;
				jsiPlexWindowSendAckPacket(pw, plp);
			}
		}
		else
		{
			/* error */
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
			fprintf(stderr, "Channel closed before syn+ack?\n");
#endif
		}
	}
	else if( pc )
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr,
			"Resending sync ACK for channel %d\n",
			plp->header.channelID);
#endif
		pc->window.lastFrameSent--;
		jsiPlexWindowSendAckPacket(&pc->window, plp);
	}
	else if( (pol = jsiGetPlexOutLinkFromContext(pl, plc)) )
	{
		jsiPlexChannel *pc;
		
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
		fprintf(stderr,	"Sending sync ACK\n");
#endif
		if( (pc = (jsiPlexChannel *)jsiCreatePlexChannel()) )
		{
			jsiPlexWindow *pw = &pc->window;

			pw->state = PWS_SYN_RECV;
			pw->link = pol;
			pc->channelID = plp->header.channelID;
			pw->nextFrameExpected = plp->header.seqNum + 1;
			jsiAddPlexChannel(pc, pol);
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
			fprintf(stderr,	"Created peer ");
			jsiPrintPlexChannel(stderr, pc);
			fprintf(stderr, "\n");
#endif
			jsiPlexWindowSendAckPacket(pw, plp);
		}
		else
		{
			jsiDropPlexOutLink(pol);
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
			fprintf(stderr, "Can't allocate out channel\n");
#endif
		}
	}
	else
	{
		/* error */
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr, "Can't allocate out link\n");
#endif
	}
}

static
jsiChannel *handleAckPacket(jsiPlexInLink *pil,
			    jsiPlexLink *pl,
			    jsiPlexChannel *pc,
			    jsiPlexLinkPacket *plp)
{
	jsiChannel *retval = 0;
	jsiPlexWindow *pw;

	require(pc != 0);
	require(plp != 0);

	pw = &pc->window;
	/* Check if its an ACK we care about */
	if( !inPlexWindow(plp->header.ackNum,
			  pw->lastAckReceived + 1,
			  pw->lastFrameSent) )
	{
		return( retval );
	}
	if( pw->state == PWS_SYN_RECV )
	{
		pw->state = PWS_CONNECTED;
		pc->window.link->flags |=
			POLF_BIDIRECTIONAL |
			POLF_TEMPORARY_BIDIRECTIONAL;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
		fprintf(stderr, "Server channel connected ");
		jsiPrintPlexChannel(stderr, pc);
		fprintf(stderr, "\n");
#endif
		retval = (jsiChannel *)pc;
	}
	else
	{
		flushPlexWindow(pw, plp->header.ackNum);
		pw->lastAckReceived = plp->header.ackNum;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
		fprintf(stderr, "Received ACK ");
		jsiPrintPlexLinkPacket(stderr, plp);
		fprintf(stderr, " for ");
		jsiPrintPlexChannel(stderr, pc);
		fprintf(stderr, "\n");
#endif
		jsiSignalPlexLink(pl);
		switch( pw->state )
		{
		case PWS_CLOSING:
		case PWS_LAST_ACK:
		case PWS_FIN_WAIT_1:
			if( (pw->lastAckReceived == pw->lastFrameSent) ||
			    (pw->lastAckReceived == (pw->lastFrameSent - 1)) )
			{
				int iLockRoot;
				
				switch( pw->state )
				{
				case PWS_CLOSING:
					pw->state = PWS_TIME_WAIT;
					break;
				case PWS_LAST_ACK:
					pw->state = PWS_CLOSED;
					break;
				case PWS_FIN_WAIT_1:
					pw->state = PWS_FIN_WAIT_2;
					break;
				default:
					assert(0);
					break;
				}
				jsiLock(pw);
				pw->link->flags &= ~POLF_EXPECTING_PACKET;
				jsiSignal(pw);
				jsiUnlock(pw);
			}
			break;
		default:
			break;
		}
	}
	return( retval );
}

/*
 * require(pol->link is locked)
 */
static
int handleDataPacket(jsiPlexChannel *pc, jsiPlexLinkPacket *plp)
{
	int seqNum, iLockRoot, retval = 1;
	jsiPlexWindow *pw = &pc->window;

	require(pc != 0);
	require(plp != 0);

	seqNum = plp->header.seqNum;
	jsiLock(pw);
	/* Check if we need to allocate a buffer for the message */
	if( (pc->recvBuffer.start == 0) && plp->header.totalLength )
	{
		int msgSize;

		msgSize = plp->header.totalLength;
		if( (pc->recvBuffer.start = jsiMalloc(msgSize)) )
		{
			pc->recvBuffer.messageID = plp->header.messageID;
			pc->recvBuffer.size = msgSize;
			pc->recvBuffer.remaining = msgSize;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
			fprintf(stderr,
				"Allocated receive buffer %p:%d for "
				"message %d\n",
				pc->recvBuffer.start,
				pc->recvBuffer.size,
				pc->recvBuffer.messageID);
#endif
		}
		else if( pw->recvWindow[seqNum % JSI_PLEX_WINDOW_SIZE] > -5 )
		{
			/*
			 * Bah, no memory, delay sending the ACK in hopes we
			 * can get some more in the future.
			 */
			retval = 0;
			pw->recvWindow[seqNum % JSI_PLEX_WINDOW_SIZE]--;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
			fprintf(stderr, "No memory, delaying ACK\n");
#endif
		}
		else
		{
			/* Its not happening, blow the channel away. */
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
			fprintf(stderr, "No memory, killing channel\n");
#endif
			pw->state = PWS_CLOSED; // XXX
			jsiSignal(pw);
		}
	}
	if( pc->recvBuffer.start )
	{
		if( pc->recvBuffer.messageID == plp->header.messageID )
		{
			int bodySize = plp->packetSize -
				sizeof(jsiPlexLinkHeader);

#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
			fprintf(stderr,
				"Received segment (%d:%d/%d) for message %d\n",
				plp->header.fragOffset,
				bodySize,
				plp->header.totalLength,
				pc->recvBuffer.messageID);
#endif
			/* Store the packet segment */
			memcpy(&pc->recvBuffer.start[plp->header.fragOffset],
			       plp->body,
			       bodySize);
			pc->recvBuffer.remaining -= bodySize;
			if( pc->recvBuffer.remaining == 0 )
			{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_SYSTEM
				fprintf(stderr,
					"Message %d completed, informing "
					"application.\n",
					pc->recvBuffer.messageID);
#endif
				pw->link->flags &= ~POLF_EXPECTING_PACKET;
				jsiSignal(pw);
			}
		}
		else
		{
			/*
			 * Must be part of a new message, delay the ACK for a
			 * bit, so the handler can eat the message.
			 */
			retval = 0;
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
			fprintf(stderr,
				"No buffer space for new message %d\n",
				plp->header.messageID);
#endif
		}
	}
	else
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr,
			"No buffer space allocated for %d?\n",
			plp->header.messageID);
#endif
	}
	jsiUnlock(pw);
	return( retval );
}

void plexTeardown(jsiPlexChannel *pc,
		  jsiPlexLinkPacket *plp)
{
	unsigned char oldAckNum = plp->header.ackNum;
	jsiPlexWindow *pw = &pc->window;
	int iLockRoot;
	
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
	fprintf(stderr, "Received fin ");
	jsiPrintPlexLinkPacket(stderr, plp);
	fprintf(stderr, " for ");
	jsiPrintPlexWindow(stderr, pw);
	fprintf(stderr, "\n");
#endif
	switch( pw->state )
	{
	case PWS_CONNECTED:
	case PWS_FIN_WAIT_1:
	case PWS_FIN_WAIT_2:
		plp->header.flags = 0;
		if( jsiPlexWindowSendAckPacket(pw, plp) )
		{
			switch( pw->state )
			{
			case PWS_CONNECTED:
				pw->state = PWS_CLOSE_WAIT;
				break;
			case PWS_FIN_WAIT_1:
				if( oldAckNum == pw->lastFrameSent )
				{
					pw->state = PWS_TIME_WAIT;
				}
				else
				{
					pw->state = PWS_CLOSING;
				}
				break;
			case PWS_FIN_WAIT_2:
				pw->state = PWS_TIME_WAIT;
				break;
			default:
				require(0);
				break;
			}
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
			fprintf(stderr, "Signalling app of ");
			jsiPrintPlexWindow(stderr, pw);
			fprintf(stderr, "\n");
#endif
			/* Signal and wait for app to close */
			jsiLock(pw);
			pw->link->flags &= ~POLF_EXPECTING_PACKET;
			jsiSignal(pw);
			jsiUnlock(pw);
		}
		else
		{
			/* XXX */
			fprintf(stderr, "Unable to send ACK for FIN\n");
		}
		break;
	case PWS_CLOSE_WAIT:
	case PWS_CLOSING:
	case PWS_TIME_WAIT:
	case PWS_LAST_ACK:
		fprintf(stderr, "Resending ACK for FIN\n");
		plp->header.flags = 0;
		jsiPlexWindowSendAckPacket(pw, plp);
		break;
	default:
		break;
	}
}

jsiChannel *jsiPlexWindowRecvPacket(jsiPlexInLink *pil,
				    jsiPlexOutLink *pol,
				    jsiPlexLinkContext *plc,
				    jsiPlexLinkPacket *plp)
{
	jsiPlexLink *pl = pil ? pil->link : pol->link;
	int seqNum = plp->header.seqNum;
	jsiChannel *retval = 0;
	jsiPlexChannel *pc;
	jsiPlexWindow *pw;
	
	pc = jsiFindPlexChannel(plp->header.channelID, plc);
	pw = &pc->window;
	if( plp->header.flags & PHF_SYN )
	{
		plexSetup(pl, pc, plc, plp);
		return( retval );
	}
	if( !pc )
	{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
		fprintf(stderr,
			"No destination channel for %d\n",
			plp->header.channelID);
#endif
		return( retval );
	}
	if( plp->header.flags & PHF_POKE )
	{
		int iLockRoot;
		
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
		fprintf(stderr, "Received poke for");
		jsiPrintPlexChannel(stderr, pc);
		fprintf(stderr, "\n");
#endif
		if( plp->header.flags & PHF_ACK )
		{
			/* Got the reply to our poke, signal the waiter. */
			jsiLock(&pc->window);
			pc->window.poked = 1;
			jsiSignal(&pc->window);
			jsiUnlock(&pc->window);
		}
		else
		{
			/* The other side is testing connection liveness. */
			pokePlexWindow(&pc->window, pc->channelID, 1);
		}
		return( retval );
	}
	if( plp->header.flags & PHF_FIN )
	{
		/* XXX Use receive window to record the fin */
		if( pw->nextFrameExpected == seqNum )
		{
			pw->recvWindow[seqNum % JSI_PLEX_WINDOW_SIZE] = 1;
			plexTeardown(pc, plp);
		}
	}
	if( plp->header.totalLength != 0 )
	{
		if( inPlexWindow(plp->header.seqNum,
				 pw->nextFrameExpected,
				 pw->nextFrameExpected +
				 JSI_PLEX_WINDOW_SIZE - 1) )
		{
			int sendAck = 1;
			
			/* Its a data packet we care about */
			if( pw->recvWindow[seqNum % JSI_PLEX_WINDOW_SIZE] > 0 )
			{
				/*
				 * Got the packet already, but ACK might've
				 * been lost, resend it.
				 */
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
				fprintf(stderr, "Ignoring duplicate data packet: ");
				jsiPrintPlexLinkPacket(stderr, plp);
				fprintf(stderr, " destined for ");
				jsiPrintPlexChannel(stderr, pc);
				fprintf(stderr, "\n");
#endif
			}
			else
			{
				sendAck = handleDataPacket(pc, plp);
			}
			if( sendAck )
			{
				pw->recvWindow[seqNum % JSI_PLEX_WINDOW_SIZE] = 1;
				if( seqNum == pw->nextFrameExpected )
				{
					jsiPlexWindowSendAckPacket(pw,
								   plp);
				}
				else
				{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
					fprintf(stderr,
						"Delaying ACK for %d ",
						plp->header.seqNum);
					jsiPrintPlexChannel(stderr, pc);
					fprintf(stderr, "\n");
#endif
				}
			}
			else
			{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_PACKET
				fprintf(stderr,
					"Delaying ACK for %d\n",
					plp->header.seqNum);
#endif
			}
		}
		else if( inPlexWindow(plp->header.seqNum,
				      pw->nextFrameExpected -
				      JSI_PLEX_WINDOW_SIZE,
				      pw->nextFrameExpected) )
		{
			jsiPlexWindowSendAckPacket(pw, plp);
		}
		else
		{
#if JSI_PLEX_DEBUG >= JSI_PLEX_DEBUG_ERRORS
			fprintf(stderr,	"Dropping packet: ");
			jsiPrintPlexLinkPacket(stderr, plp);
			fprintf(stderr, " destined for ");
			jsiPrintPlexChannel(stderr, pc);
			fprintf(stderr, "\n");
#endif
		}
	}
	if( plp->header.flags & PHF_ACK )
	{
		retval = handleAckPacket(pil, pl, pc, plp);
	}
	return( retval );
}
