/*
 * jsiPlexWindow.h
 * Sliding window implementation for plex channels
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _jsi_plex_window_h
#define _jsi_plex_window_h

/*
 * This module does the actual work of multiplexing connections reliable
 * connections over a given link.  The implementation is basically a derivative
 * of TCP except its packet based instead of a stream and much simpler.
 */

#include "jsiPlexLink.h"

struct _jsiPlexChannel;

/* Window states, see TCP state diagram */
typedef enum {
	PWS_MIN,
	PWS_CLOSED,
	PWS_SYN_SENT,
	PWS_SYN_RECV,
	PWS_CONNECTED,
	PWS_FIN_WAIT_1,
	PWS_FIN_WAIT_2,
	PWS_CLOSING,
	PWS_TIME_WAIT,
	PWS_CLOSE_WAIT,
	PWS_LAST_ACK,
	PWS_MAX,
} plex_window_state_t;

/*
 * The sliding window control structure for plex channels.
 *
 * `state' - One of the above state values.
 * `link' - The output link that is pointed at the remote host.
 * `sendWindow' - Array of pointers to packets that have been sent through the
 *   link, these packets are then removed and returned to the output link
 *   when they have been acknowledged by the other side.
 * `recvWindow' - Array of integers that indicate whether the packet has been
 *   received or if the value is less than zero, how many times we've delayed
 *   the ACK because of lack of memory to store the packet.
 * `lastAckReceived' - The last sequence number that was acknowledged.
 * `lastFrameSent' - The sequence number of the last frame sent.
 * `nextFrameExpected' - The next sequence number expected from the other side.
 * `poked' - Tracks receipt of poke replies.
 */
typedef struct _jsiPlexWindow {

	JSI_LOCK_DECL

	plex_window_state_t state;
	jsiPlexOutLink *link;

#define JSI_PLEX_WINDOW_SIZE 8
	jsiPlexLinkPacket *sendWindow;
	jsiPlexLinkPacket **sendWindowTail;
	unsigned int resendTimeout;

	int recvWindow[JSI_PLEX_WINDOW_SIZE];
	int ackTimeOut;

	unsigned char lastAckReceived;
	unsigned char lastFrameSent;

	unsigned char nextFrameExpected;

	unsigned char poked;
} jsiPlexWindow;

/*
 * Print the contents of the given jsiPlexWindow.
 *
 * `file' - The file to write the contents too.
 * `pw' - An initialized window to print out.
 */
void jsiPrintPlexWindow(FILE *file, jsiPlexWindow *pw);
/*
 * Initialize a jsiPlexWindow object.
 *
 * `pw' - An uninitialized jsiPlexWindow object.
 */
void jsiPlexWindowInit(jsiPlexWindow *pw);
/*
 * Finalize a jsiPlexWindow object.
 *
 * `pw' - An initialized jsiPlexWindow object.
 */
void jsiPlexWindowKill(jsiPlexWindow *pw);
/*
 * Attempt to connect the channel to a peer on the remote host.
 *
 * `pw' - The window that should be connected to a peer.
 * `pol' - An output link that points to the remote host.
 * `channel' - The unique channel identifier.
 * returns - True if the connection succeeded or false if it failed.
 */
int jsiPlexWindowConnect(jsiPlexWindow *pw, jsiPlexOutLink *pol, int channel);
/*
 * Disconnect the channel from its remote peer.
 *
 * `pw' - A connected jsiPlexWindow object.
 * `channel' - The channel identifier that was used to connect the window
 */
void jsiPlexWindowDisconnect(jsiPlexWindow *pw, int channel);
/*
 * Send a packet to the peer on the remote host and track it in the sendWindow
 * array until it has been acknowledged.
 *
 * `pw' - A connected jsiPlexWindow that should be responsible for the packet.
 * `plp' - A packet object as returned from jsiAllocPlexLinkPacket that
 *   has been initialized, except for seqNum and ackNum.
 * returns - True if the write succeeded, false otherwise.
 */
int jsiPlexWindowSendPacket(jsiPlexWindow *pw, jsiPlexLinkPacket *plp);
/*
 * Send an ACK packet to the peer.
 *
 * `pw' - A connected jsiPlexWindow object.
 * `plp' - The received packet that is being acknowledged.
 * return - True if the write succeeded, false otherwise.
 */
int jsiPlexWindowSendAckPacket(jsiPlexWindow *pw,
			       jsiPlexLinkPacket *plp);
/*
 * The callback for packets received by the plex link handler.
 *
 * `pil' - The input link that received the packet.  Needed for access to its
 *  jsiProvider and jsiPlexLink objects.
 * `plc' - A plex link context which describes the source of the packet.
 * `plp' - The received packet.
 * returns - A newly connected channel.
 */
jsiChannel *jsiPlexWindowRecvPacket(jsiPlexInLink *pil,
				    jsiPlexOutLink *pol,
				    jsiPlexLinkContext *plc,
				    jsiPlexLinkPacket *plp);
/*
 * Wait for a message to be received on the given channel.
 *
 * `pc' - The channel to expect a message on.
 * `bufferSize' - A pointer to an integer that will be set to the size of the
 *  returned message buffer.
 * returns - A buffer containing the received message or NULL if there channel
 *  became disconnected.
 */
char *jsiPlexWindowWaitForMessage(struct _jsiPlexChannel *pc, int *bufferSize);

#endif
