/*
 * jsiPropertyInfo.c
 * Functions for manipulating jsiPropertyInfo's
 *
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsi.h"
#include "jsiImpl.h"

jsiPropertyInfo *jsiCreatePropertyInfo(int tag, ...)
{
	parsedString name, specifiers, usage;
	jsiPropertyInfo *retval = 0;
	int error = 0, commands;
	char *internal = 0;
	va_list args;

	va_start(args, tag);
	/* Walk the tag list and figure out how much space we'll need */
	while( tag != JPIA_TAG_DONE )
	{
		switch( tag )
		{
		case JPIA_Internal:
			internal = va_arg(args, char *);
			break;
		case JPIA_Name:
			name.data = va_arg(args, char *);
			name.len = strlen(name.data);
			break;
		case JPIA_Commands:
			commands = va_arg(args, int);
			break;
		case JPIA_Specifiers:
			specifiers.data = va_arg(args, char *);
			specifiers.len = strlen(specifiers.data);
			break;
		case JPIA_Usage:
			usage.data = va_arg(args, char *);
			usage.len = strlen(usage.data);
			break;
		}
		tag = va_arg(args, int);
	}
	/* Parse the internal format */
	if( internal )
	{
		error = !parseString(internal,
				     SPO_Expect, "<:",
				     SPO_String, &name,
				     SPO_Expect, ":><:",
				     SPO_HexInteger, &commands,
				     SPO_Expect, ":><:",
				     SPO_String, &specifiers,
				     SPO_Expect, ":><:",
				     SPO_String, &usage,
				     SPO_Expect, ":>",
				     SPO_End);
	}
	/*
	 * If we're good to go, then allocate the jsiPropertyInfo object and
	 * its values in one shot.
	 */
	if( !error &&
	    (retval = jsiMalloc(sizeof(jsiPropertyInfo) +
				name.len + 1 +
				specifiers.len + 1 +
				usage.len + 1)) )
	{
		/* Setup the object */
		retval->name = (char *)(retval + 1);
		strncpy(retval->name, name.data, name.len);
		retval->name[name.len] = 0;
		retval->commands = commands;
		retval->specifiers = retval->name + name.len + 1;
		strncpy(retval->specifiers, specifiers.data, specifiers.len);
		retval->specifiers[specifiers.len] = 0;
		retval->usage = retval->specifiers + specifiers.len + 1;
		strncpy(retval->usage, usage.data, usage.len);
		retval->usage[usage.len] = 0;
	}
	va_end(args);
	return( retval );
}

void jsiDeletePropertyInfo(jsiPropertyInfo *jpi)
{
	jsiFree(jpi);
}

int jsiFindProperty(jsiPropertyInfo *jpi, char *name)
{
	int lpc, retval = -1;

	require(jpi != 0);
	require(name != 0);

	/*
	 * Walk over the array until we find it or reach a null or empty string
	 */
	for( lpc = 0;
	     (retval < 0) && jpi[lpc].name && jpi[lpc].name[0];
	     lpc++ )
	{
		if( !strcmp(jpi[lpc].name, name) )
			retval = lpc;
	}
	return( retval );
}

int jsiComputePropertyInfoSize(jsiPropertyInfo *pi)
{
	int retval;

	require(pi != 0);
	
	retval =
		2 + strlen(pi->name) +			/* <name> */
		4 + 2 + 8 +				/* :0x<commands> */
		4 + strlen(pi->specifiers) +		/* :<specifiers> */
		4 + strlen(pi->usage) +			/* :<usage> */
		2 + 1;					/* \0' */
	return( retval );
}

void jsiEncodePropertyInfo(jsiPropertyInfo *pi, char *buffer)
{
	require(pi != 0);
	require(buffer != 0);
	
	sprintf(buffer,
		"<:%s:><:0x%08x:><:%s:><:%s:>",
		pi->name,
		pi->commands,
		pi->specifiers,
		pi->usage);

	ensure((strlen(buffer) + 1) == jsiComputePropertyInfoSize(pi));
}

void jsiPrintProperty(FILE *file, jsiPropertyInfo *pi)
{
	int lpc;

	require(file != 0);
	require(pi != 0);
	
	fprintf(file, "jsiPropertyInfo[name=`%s'; commands={", pi->name);
	for( lpc = JSI_WHAT_NONE + 1; lpc < JSI_WHAT_MAX; lpc++ )
	{
		if( pi->commands & (1L << lpc) )
		{
			fprintf(file,
				"%s%s",
				(lpc > 0) ? "|" : "",
				jsiWhatNames[lpc]);
		}
	}
	fprintf(file, "}; usage=`%s']", pi->usage);
}
