/*
 * jsiRegistry.h
 * Functions for managing a simple registry
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsi.h"
#include "jsiImpl.h"

/* Default directories to store registry information */
#define DEFAULT_SYSTEM_ROOT "/etc"
#define HOME_DIRECTORY_ROOT ".janosvm"
#define HOME_DIRECTORY_REGISTRY "/registry"

#ifndef DEFFILEMODE
#define DEFFILEMODE     (S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)
#endif

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

static char *userHome = 0;

void jsiSetRegistryHome(char *home)
{
	userHome = home;
}

/* Copy the path to `name' into `path'. */
void jsiGetRegistryPath(char *path, char *name, char *sub, int create)
{
	char *homeDir;

	require(path != 0);

	/* We allow sub == 0 so its easy to create the directories */

	path[0] = '\0';
	if( name )
	{
		if( userHome )
		{
			/* User specified home, make the instance directory. */
			sprintf(path, "%s/%s", userHome, name);
			if( create )
				jsiMkDir(path, DEFFILEMODE|S_IEXEC);
			/* Look for the file inside that. */
			if( sub )
			{
				strcat(path, "/");
				strcat(path, sub);
			}
		}
		else if( (homeDir = getenv("HOME")) )
		{
			/*
			 * Look in the users `$HOME/.janosvm/registry/'
			 * directory
			 */
			sprintf(path, "%s/" HOME_DIRECTORY_ROOT, homeDir);
			if( create )
				jsiMkDir(path, DEFFILEMODE|S_IEXEC);
			strcat(path, HOME_DIRECTORY_REGISTRY);
			if( create )
				jsiMkDir(path, DEFFILEMODE|S_IEXEC);
			strcat(path, "/");
			strcat(path, name);
			if( create )
				jsiMkDir(path, DEFFILEMODE|S_IEXEC);
			if( sub )
			{
				strcat(path, "/");
				strcat(path, sub);
			}
		}
		else
		{
			/* We don't create dirs for the system registry... */
			sprintf(path,
				DEFAULT_SYSTEM_ROOT
				HOME_DIRECTORY_REGISTRY
				"/%s",
				name);
			if( sub )
			{
				strcat(path, "/");
				strcat(path, sub);
			}
			
		}
	}
	if( sub && (path[0] == '\0') )
	{
		/* No hit, try the current directory */
		strcpy(path, sub);
	}
}

void jsiRegisterName(char *name)
{
	char path[MAXPATHLEN];

	/* Use jsiGetRegistryPath to create the directories for us. */
	jsiGetRegistryPath(path, name, 0, 1);
}

void jsiUnregisterName(char *name)
{
	/*
	 * ... Should we delete the dir?  It might still have stuff the user
	 * wants to look at.
	 */
}

void jsiVRegisterExport(char *name, char *export, char *fmt, va_list args)
{
	char path[MAXPATHLEN];
	FILE *file;
	
	require(export != 0);
	
	jsiGetRegistryPath(path, name, export, 1);
	if( (file = fopen(path, "w")) )
	{
		if( fmt )
		{
			vfprintf(file, fmt, args);
			fclose(file);
		}
	}
}

void jsiRegisterExport(char *name, char *export, char *fmt, ...)
{
	va_list args;
	
	require(export != 0);

	va_start(args, fmt);
	jsiVRegisterExport(name, export, fmt, args);
	va_end(args);
}

void jsiUnregisterExport(char *name, char *export)
{
	char path[MAXPATHLEN];
	
	require(export != 0);

	jsiGetRegistryPath(path, name, JSI_INTERFACE_EXPORT, 0);
	jsiRemove(path);
}

void jsiRegisterInterface(char *name, char *transport, char *tid)
{
	require(transport != 0);
	require(tid != 0);

	/* XXX This is compatible with jsiChannel's formatting */
	jsiRegisterExport(name, JSI_INTERFACE_EXPORT, "%s:%s", transport, tid);
}

void jsiUnregisterInterface(char *name)
{
	jsiUnregisterExport(name, JSI_INTERFACE_EXPORT);
}

char *jsiFindExport(char *name, char *export)
{
	char path[MAXPATHLEN];
	char *retval = 0;
	FILE *file;

	require(export != 0);
	
	jsiGetRegistryPath(path, name, export, 0);
	if( (file = fopen(path, "r")) )
	{
		int size;

		/* Get the size of the file and read it in. */
		fseek(file, 0, SEEK_END);
		size = ftell(file);
		if( (retval = jsiMalloc(size + 1)) )
		{
			fseek(file, 0, SEEK_SET);
			fread(retval,
			      sizeof(char),
			      size,
			      file);
			retval[size] = 0;
		}
		fclose(file);
		/* Do some simple checking on the string we got */
		if( retval )
		{
			if( strlen(retval) != size )
			{
				jsiFree(retval);
				retval = 0;
			}
		}
	}
	return( retval );
}

int jsiExportExists(char *name, char *export)
{
	char path[MAXPATHLEN];
	int retval = 0;
	FILE *file;

	require(export != 0);
	
	jsiGetRegistryPath(path, name, export, 0);
	if( (file = fopen(path, "r")) )
	{
		retval = 1;
		fclose(file);
	}
	return( retval );
}

char *jsiFindInterface(char *name)
{
	char *retval;
	
	retval = jsiFindExport(name, JSI_INTERFACE_EXPORT);
	return( retval );
}
