/*
 * jsiStdChannel.c
 * Functions specific to jsiStdChannels
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsi.h"
#include "jsiImpl.h"
#include "jsiStdChannel.h"

/* Mapping of protocol identifiers to strings */
static char *jsiStdChannelNames[] = {
	"",
	"self",
	"udp",
	"tcp",
	"file",
	0
};

/*
 * XXX Copied from jsiMessage since we need to unmarshall the first int to
 * get the size of the message.
 */
#define SIZEOF_INT 4

jsiChannel *jsiCreateStdChannel(void)
{
	jsiChannel *retval = 0;
	jsiStdChannel *jsc;

	if( (jsc = (jsiStdChannel *)jsiMalloc(sizeof(jsiStdChannel))) )
	{
		/* Default to tcp */
		jsc->channel.ops = &jsiStdChannelImplementation;
		jsc->protocol = JCP_TCP;
		jsc->inputFD = -1;
		jsc->outputFD = -1;
		retval = &jsc->channel;
	}
	return( retval );
}

static
void jsiDeleteStdChannel(jsiChannel *jc)
{
	jsiStdChannel *jsc = (jsiStdChannel *)jc;
	
	if( jc )
	{
		/* Make sure we're disconnected */
		if( jsc->inputFD >= 0 )
			jsiDisconnectChannel(jc);
		jsiFree(jsc);
	}
}

/* Setup a listener port for UDP */
static
int jsiBindPort(jsiStdChannel *jc, int port)
{
	int sock, retval = 0;

	require(jc != 0);
	require(jc->protocol == JCP_UDP);
	
	if( !jsiSocket(AF_INET, SOCK_DGRAM, 0, &sock) )
	{
		struct sockaddr_in saddr;
		int on_off = 1;

		memset(&saddr, 0, sizeof(saddr));
		saddr.sin_family = AF_INET;
		saddr.sin_addr.s_addr = htonl(INADDR_ANY);
		saddr.sin_port = htons(port);
		jsiSetSockOpt(sock, SOL_SOCKET, SO_REUSEADDR,
			      &on_off, sizeof( int ));
		if( !jsiBind(sock, (struct sockaddr *)&saddr, sizeof(saddr)) )
		{
			jc->inputFD = sock;
			jc->outputFD = sock;
			retval = 1;
		}
		else
		{
			jsiClose(sock);
		}
	}
	return( retval );
}

/* Connect using a sockaddr_in value */
static
int jsiConnectSockAddr(jsiStdChannel *jc, struct sockaddr_in *host_addr)
{
	int type = SOCK_STREAM, sock, retval = 0;

	require(jc != 0);
	require(host_addr != 0);
	
	switch( jc->protocol )
	{
	case JCP_UDP:
		type = SOCK_DGRAM;
		break;
	case JCP_TCP:
		type = SOCK_STREAM;
		break;
	}
	if( !jsiSocket(AF_INET, type, 0, &sock) )
	{
		if( !jsiConnect(sock,
				(struct sockaddr *)host_addr,
				sizeof(*host_addr)) )
		{
			jc->inputFD = sock;
			jc->outputFD = sock;
			retval = 1;
		}
		else
		{
			jsiClose(sock);
		}
	}
	return( retval );
}

/* Connect using a host name and port */
static
int jsiConnectHostPort(jsiStdChannel *jc, char *hostname, int hostPort)
{
	struct sockaddr_in host_addr;
	struct hostent *host_ent;
	int retval = 1;

	require(jc != 0);
	require(hostname != 0);
	require(hostPort >= 0);
	
	memset(&host_addr, 0, sizeof(host_addr));
	host_addr.sin_family = AF_INET;
	/* Check if hostname is a valid DNS name */
	if( !jsiGetHostByName(hostname, &host_ent) )
	{
		memcpy((char *)&host_addr.sin_addr.s_addr,
		       host_ent->h_addr,
		       host_ent->h_length);
	}
	else if( (host_addr.sin_addr.s_addr = inet_addr(hostname)) == -1 )
	{
		/* Its not a number address either */
		retval = 0;
	}
	if( retval )
	{
		host_addr.sin_port = htons(hostPort);
		retval = jsiConnectSockAddr(jc, &host_addr);
	}
	return( retval );
}

static
int jsiConnectFile(jsiStdChannel *jc, char *basename, int do_create)
{
	static char *input_name_template = "%s.to";
	static char *output_name_template = "%s.from";
	int len, retval = 0;
	char *filename;

	require(jc != 0);
	
	len = strlen(basename) + strlen(output_name_template) + 1;
	if( (filename = jsiMalloc(len)) )
	{
		int flags;
		
		retval = 1;
		sprintf(filename, input_name_template, basename);
		if( do_create )
			flags = O_RDONLY|O_CREAT;
		else
			flags = O_WRONLY|O_APPEND;
		if( jsiOpen(filename, flags, 0600, &jc->inputFD) != 0 )
			retval = 0;
		sprintf(filename, output_name_template, basename);
		if( do_create )
			flags = O_WRONLY|O_APPEND|O_CREAT;
		else
			flags = O_RDONLY;
		if( jsiOpen(filename, flags, 0600, &jc->outputFD) != 0 )
			retval = 0;
		if( retval )
		{
			off_t dummy;
			
			jsiSeek(jc->inputFD, 0, SEEK_END, &dummy);
			jsiSeek(jc->outputFD, 0, SEEK_END, &dummy);
			if( !do_create )
			{
				int tmp;

				tmp = jc->inputFD;
				jc->inputFD = jc->outputFD;
				jc->outputFD = tmp;
			}
			retval = 1;
		}
		else
		{
			if( jc->inputFD != -1 )
				jsiClose(jc->inputFD);
			if( jc->outputFD != -1 )
				jsiClose(jc->outputFD);
		}
	}
	return( retval );
}

/* Connect using a formatted address */
static
int jsiConnectFormatted(jsiStdChannel *jc, char *fmt, int do_create)
{
	parsedString protocol, path;
	int port = -1, retval = 0;

	require(jc != 0);
	require(fmt != 0);

	if( parseString(fmt,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_String, &path,
			SPO_End) )
	{
		int lpc, found = 0;
		
		/*
		 * Check to see if the protocol name matches something we
		 * support.
		 */
		for( lpc = 0; (lpc < JCP_MAX) && !found; lpc++ )
		{
			if( !cmpPStrStr(&protocol, jsiStdChannelNames[lpc]) )
			{
				jc->protocol = lpc;
				found = 1;
			}
		}
		if( found )
		{
			switch( jc->protocol )
			{
			case JCP_SELF:
				retval = 1;
				break;
			case JCP_UDP:
			case JCP_TCP:
				if( parseString(fmt,
						SPO_NonEmptyString, &protocol,
						SPO_Expect, ":",
						SPO_NonEmptyString, &path,
						SPO_Expect, "/",
						  SPO_Integer, &port,
						  SPO_End,
						SPO_End) )
				{
					path.data[path.len] = 0;
					retval = jsiConnectHostPort(jc,
								    path.data,
								    port);
					path.data[path.len] = '/';
				}
				else
				{
					jsiLog("Bad %s format string: `%s'",
					       jsiStdChannelNames[jc->
								 protocol],
					       fmt);
				}
				break;
			case JCP_FILE:
				if( parseString(fmt,
						SPO_NonEmptyString, &protocol,
						SPO_Expect, ":",
						SPO_NonEmptyString, &path,
						SPO_End) )
				{
					retval = jsiConnectFile(jc,
								path.data,
								do_create);
				}
				else
				{
					jsiLog("No file name specified in "
					       "`file:' format.");
				}
				break;
			default:
				assert(0);
				break;
			}
		}
	}
	return( retval );
}

int jsiConnectStdChannel(jsiChannel *jc, int tag, ...)
{
	jsiStdChannel *jsc = (jsiStdChannel *)jc;
	int do_create = 0, port = -1, retval = 0;
	char *host = 0, *fmt = 0, *filename = 0;
	struct sockaddr_in *host_addr = 0;
	va_list args;

	require(jc != 0);

	va_start(args, tag);
	/* Disconnect from any previous connections */
	if( jsc->inputFD >= 0 )
		jsiDisconnectChannel(&jsc->channel);
	/* Walk the tag list */
	while( tag != JCA_TAG_DONE )
	{
		switch( tag )
		{
		case JCA_Formatted:
			fmt = va_arg(args, char *);
			break;
		case JCA_Protocol:
			jsc->protocol = va_arg(args, int);
			
			ensure(jsc->protocol > JCP_NONE);
			ensure(jsc->protocol < JCP_MAX);
			break;
		case JCA_Host:
			host = va_arg(args, char *);
			break;
		case JCA_Port:
			port = va_arg(args, int);
			break;
		case JCA_SockAddr:
			host_addr = va_arg(args, struct sockaddr_in *);
			break;
		case JCA_Socket:
			jsc->inputFD = va_arg(args, int);
			jsc->outputFD = jsc->inputFD;
			/* No other processing needed */
			break;
		case JCA_FileName:
			filename = va_arg(args, char *);
			break;
		case JCA_Create:
			do_create = va_arg(args, int);
			break;
		default:
			require(tag > JCA_TAG_DONE);
			require(tag < JCA_TAG_MAX);
			break;
		}
		tag = va_arg(args, int);
	}
	/* Figure out which way to connect to the peer */
	if( fmt )
		retval = jsiConnectFormatted(jsc, fmt, do_create);
	else if( host )
		retval = jsiConnectHostPort(jsc, host, port);
	else if( host_addr )
		retval = jsiConnectSockAddr(jsc, host_addr);
	else if( port >= 0 )
		retval = jsiBindPort(jsc, port);
	else if( jsc->inputFD >= 0 )
		retval = 1;
	else if( filename )
		retval = jsiConnectFile(jsc, filename, do_create);
	else if( jsc->protocol == JCP_SELF )
		retval = 1;
	va_end(args);
	return(retval);
}

static
void jsiDisconnectStdChannel(jsiChannel *jc)
{
	jsiStdChannel *jsc = (jsiStdChannel *)jc;
	
	require(jc != 0);
	
	if( jsc->inputFD >= 0 )
	{
		if( jsc->protocol == JCP_TCP )
			jsiShutdown(jsc->inputFD, 2);
		jsiClose(jsc->inputFD);
		if( jsc->inputFD != jsc->outputFD )
			jsiClose(jsc->outputFD);
		/* jc->protocol not reset */
		jsc->inputFD = -1;
		jsc->outputFD = -1;
	}
}

/* Send a buffer over UDP to the given destination */
static
int jsiSendUDP(jsiStdChannel *jc, char *buffer, int bufferSize)
{
	int retval = 0;

	require(jc != 0);
	require(jc->outputFD != -1);
	require(jc->protocol == JCP_UDP);
	require(buffer != 0);
	
	retval = !jsiSend(jc->outputFD, buffer, bufferSize, 0);
	return( retval );
}

/* Send a buffer over a stream */
static
int jsiSendStream(jsiStdChannel *jc, char *buffer, int bufferSize)
{
	int bytes_out, bytes_left = bufferSize, retval = 0;

	require(jc != 0);
	require(jc->outputFD != -1);
	require((jc->protocol == JCP_TCP) || (jc->protocol == JCP_FILE));
	require(buffer != 0);

	/* Send the whole buffer out */
	while( bytes_left &&
	       !jsiWrite(jc->outputFD,
			 &buffer[bufferSize - bytes_left],
			 bytes_left, &bytes_out) )
	{
		bytes_left -= bytes_out;
	}
	if( bytes_left == 0 )
		retval = 1;
	return( retval );
}

static
int jsiStdSendMessage(jsiChannel *jc, char *buffer, int bufferSize)
{
	jsiStdChannel *jsc = (jsiStdChannel *)jc;
	int retval = 0;

	require(jc != 0);
	require(jsc->outputFD != -1);
	require(buffer != 0);

	switch( jsc->protocol )
	{
	case JCP_UDP:
		retval = jsiSendUDP(jsc, buffer, bufferSize);
		break;
	case JCP_TCP:
	case JCP_FILE:
		retval = jsiSendStream(jsc, buffer, bufferSize);
		break;
	}
	return( retval );
}

static
jsiMessage *jsiDoIPMessage(jsiHandlerContext *hc, jsiMessage *jm)
{
	jsiMessage *retval = 0;

	if( jm )
	{
		int bufferSize;
		char *buffer;
		
		if( (buffer = jsiFlattenMessage(jm, &bufferSize)) )
		{
			if( jsiSendMessage(hc->channel, buffer, bufferSize) )
			{
				retval = jsiHandleInterimMessages(hc);
			}
		}
		jsiDeleteMessage(jm);
	}
	return( retval );
}

static
jsiMessage *jsiStdDoMessage(jsiHandlerContext *hc, jsiMessage *jm)
{
	jsiStdChannel *jsc = (jsiStdChannel *)hc->channel;
	jsiMessage *retval = 0;

	if( jm )
	{
		switch( jsc->protocol )
		{
		case JCP_SELF:
			jsiDispatchMessage(hc, jm);
			retval = jm;
			break;
		case JCP_UDP:
		case JCP_TCP:
		case JCP_FILE:
			retval = jsiDoIPMessage(hc, jm);
			break;
		}
	}
	return( retval );
}

static
int jsiStdReplyMessage(jsiChannel *jc, char *buffer, int bufferSize)
{
	jsiStdChannel *jsc = (jsiStdChannel *)jc;
	int retval = 0;

	require(jc != 0);
	require(buffer != 0);

	switch(jsc->protocol)
	{
	case JCP_UDP:
		retval = jsiSendUDP(jsc, buffer, bufferSize);
		break;
	case JCP_TCP:
	case JCP_FILE:
		retval = jsiSendStream(jsc, buffer, bufferSize);
		break;
	}
	return( retval );
}

/* Receive a message over UDP */
static
char *jsiRecvUDP(jsiStdChannel *jc, int *bufferSize)
{
	char *retval;

	require(jc != 0);
	require(jc->inputFD != -1);
	require(bufferSize != 0);

	/* Allocate the maximum allowable message size */
	if( (retval = jsiMalloc(JSI_MAX_MESSAGE_SIZE)) )
	{
		/* Blast the packet into the buffer */
		if( jsiRecvFrom(jc->inputFD,
				retval,
				JSI_MAX_MESSAGE_SIZE,
				0,
				0,
				0,
				bufferSize) )
		{
			jsiFree(retval);
			retval = 0;
		}
	}
	return( retval );
}

/* Receive a message over a stream */
static
char *jsiRecvStream(jsiStdChannel *jc, int *bufferSize)
{
	int messageSize, hostSize, bytes_left, bytes_in;
	char *retval = 0;

	require(jc != 0);
	require(jc->inputFD != -1);
	require((jc->protocol == JCP_TCP) || (jc->protocol == JCP_FILE));
	require(bufferSize != 0);

	/*
	 * Read the first integer in the message, this should be the size of
	 * the entire message.
	 */
	bytes_left = SIZEOF_INT;
	while( bytes_left &&
	       !((jc->protocol == JCP_TCP) ?
		 jsiRead(jc->inputFD,
			 &((char *)&messageSize)[SIZEOF_INT - bytes_left],
			 bytes_left,
			 &bytes_in) :
		 jsiReadAndWait(jc->inputFD,
				&((char *)&messageSize)[SIZEOF_INT -
						       bytes_left],
				bytes_left,
				50, /* Sample time */
				&bytes_in)) &&
	       bytes_in )
	{
		bytes_left -= bytes_in;
	}
	hostSize = ntohl(messageSize);
	/* Allocate a buffer to hold the message */
	if( (bytes_left == 0) && (retval = jsiMalloc(hostSize)) )
	{
		int iter = 0;
		
		/*
		 * Read in the rest of the message, the size includes the
		 * first integer, so we have to subtract its size.
		 */
		*bufferSize = hostSize;
		((int *)retval)[0] = messageSize;
		bytes_left = hostSize - SIZEOF_INT;
		while( bytes_left &&
		       !((jc->protocol == JCP_TCP) ?
			 jsiRead(jc->inputFD,
				 &retval[*bufferSize - bytes_left],
				 bytes_left,
				 &bytes_in) :
			 jsiReadAndWait(jc->inputFD,
					&retval[*bufferSize - bytes_left],
					bytes_left,
					50, /* Sample time */
					&bytes_in)) &&
		       bytes_in )
		{
			bytes_left -= bytes_in;
			iter ++;
		}
		if( bytes_left != 0 )
		{
			jsiFree(retval);
			retval = 0;
		}
	}
	return( retval );
}

static
char *jsiStdRecvMessage(jsiChannel *jc, int *bufferSize)
{
	jsiStdChannel *jsc = (jsiStdChannel *)jc;
	char *retval = 0;

	require(jsc->inputFD != -1);
	require(bufferSize != 0);
	
	switch( jsc->protocol )
	{
	case JCP_UDP:
		retval = jsiRecvUDP(jsc, bufferSize);
		break;
	case JCP_TCP:
	case JCP_FILE:
		retval = jsiRecvStream(jsc, bufferSize);
		break;
	}
	return( retval );
}

void jsiPrintChannel(FILE *file, jsiChannel *jc)
{
	jsiStdChannel *jsc = (jsiStdChannel *)jc;
	
	fprintf(file,
		"jsiChannel[protocol=%s; inputFD=%d; outputFD=%d]",
		jsiStdChannelNames[jsc->protocol],
		jsc->inputFD,
		jsc->outputFD);
}

jsiChannelFunctions jsiStdChannelImplementation = {
	jsiStdChannelNames,
	jsiCreateStdChannel,
	jsiDeleteStdChannel,
	jsiConnectStdChannel,
	jsiDisconnectStdChannel,
	jsiStdSendMessage,
	jsiStdDoMessage,
	jsiStdReplyMessage,
	jsiStdRecvMessage
};
