/*
 * jsiStdProvider.c
 * Provider for jsiStdChannels
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "jsi.h"
#include "jsiImpl.h"
#include "jsiStdChannel.h"
#include "jsiStdProvider.h"

static
jsiProvider *jsiCreateStdProvider(jsiConnectionHandler *ch)
{
	jsiProvider *retval = 0;
	jsiStdProvider *sp;

	if( (sp = (jsiStdProvider *)jsiMalloc(sizeof(jsiStdProvider))) )
	{
		sp->provider.ops = &jsiStdProviderImplementation;
		sp->provider.handleConnection = ch;
		sp->socket = -1;
		retval = &sp->provider;
	}
	return( retval );
}

static
void jsiDeleteStdProvider(jsiProvider *jp)
{
	if( jp )
	{
		jsiDetachProvider(jp);
		jsiFree(jp);
	}
}

static
int jsiAttachStdProvider(jsiProvider *jp, char *format)
{
	jsiStdProvider *jsp = (jsiStdProvider *)jp;
	parsedString protocol, host;
	int port, sock, retval = 0;

	require(jp != 0);
	require(format != 0);

	if( parseString(format,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_NonEmptyString, &host,
			SPO_Expect, "/",
			SPO_Integer, &port,
			SPO_End) )
	{
		/* Setup a tcp socket to listen on. */
		if( !jsiSocket(AF_INET, SOCK_STREAM, 0, &sock) )
		{
			struct sockaddr_in saddr;
			int on_off = 1;
			
			memset(&saddr, 0, sizeof(saddr));
			saddr.sin_family = AF_INET;
			saddr.sin_addr.s_addr = htonl(INADDR_ANY);
			saddr.sin_port = htons(port);
			/*
			 * Its annoying to have the bind fail if we're doing a
			 * lot of restarts, so set reuseaddr to true.
			 */
			jsiSetSockOpt(sock, SOL_SOCKET, SO_REUSEADDR,
				      &on_off, sizeof( int ));
			if( !jsiBind(sock,
				     (struct sockaddr *)&saddr,
				     sizeof(saddr)) &&
			    !jsiListen(sock, 5) )
			{
				jsp->socket = sock;
				retval = 1;
			}
			else
			{
				jsiClose(sock);
			}
		}
	}
	return( retval );
}

static
void jsiDetachStdProvider(jsiProvider *jp)
{
	jsiStdProvider *jsp = (jsiStdProvider *)jp;
	
	require(jp != 0);
	
	if( jsp->socket >= 0 )
	{
		jsiShutdown(jsp->socket, 2);
		jsiClose(jsp->socket);
		jsp->socket = -1;
	}
}

static
void jsiHandleStdProvider(jsiProvider *jp)
{
	jsiStdProvider *jsp = (jsiStdProvider *)jp;
	
	require(jp != 0);
	require(jp->handleConnection != 0);
	
	/* Make sure we're connected... */
	if( jsp->socket >= 0 )
	{
		int peer_len, peer_sock, done = 0;
		struct sockaddr_in peer_addr;

		/*
		 * Just keep trying to accept connections until we get an
		 * error.  Shutting down this loop externally would be nice
		 * to have though.
		 */
		while( !done )
		{
			peer_len = sizeof(peer_addr);
			if( !jsiAccept(jsp->socket,
				       (struct sockaddr *)&peer_addr,
				       &peer_len,
				       &peer_sock) )
			{
				jsiChannel *jc;

				if( (jc = jsiCreateStdChannel()) )
				{
					jsiConnectStdChannel(
						jc,
						JCA_Socket, peer_sock,
						JCA_TAG_DONE);
					/*
					 * Ask the handler to do something
					 * with this new connection.
					 */
					done = jp->handleConnection->func(jp,
									  jc);
				}
				else
				{
					jsiClose(peer_sock);
					jsiLog("Can't create channel for new"
					       " connection");
				}
			}
			else
			{
				done = 1;
				jsiLog("Accept failed, exiting provider loop");
			}
		}
	}
}

static
char *jsiStdProviderNames[] = {
	"tcp",
	0
};

jsiProviderFunctions jsiStdProviderImplementation = {
	jsiStdProviderNames,
	jsiCreateStdProvider,
	jsiDeleteStdProvider,
	jsiAttachStdProvider,
	jsiDetachStdProvider,
	jsiHandleStdProvider
};
