/*
 * heyActions.c
 * Actions that jsihey can perform on a reply message
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "config.h"

#include "assert_pp.h"

#include "heyActions.h"

static int actionDefault(heyMsgAction *hma);
static int actionPrint(heyMsgAction *hma);
static int actionForInstance(heyMsgAction *hma);
static int actionPrintSuites(heyMsgAction *hma);

/* Mappings for all actions to their integer values */
stringIntMap actionMappings[] = {
	{ "forinstance", ACTION_FOR_INSTANCE },
	{ "print", ACTION_PRINT },
	{ "nothing", ACTION_NOTHING },
	{ "ignore", ACTION_NONE },
	{ "", ACTION_PRINT_SUITES },
	{ 0, ACTION_ERROR }
};

int doAction(heyMsgAction *hma)
{
	int retval = 0;

	require(hma != 0);
	
	switch( hma->action )
	{
	case ACTION_NONE:
		retval = 1;
		break;
	case ACTION_NOTHING:
		retval = 1;
		break;
	case ACTION_DEFAULT:
		retval = actionDefault(hma);
		break;
	case ACTION_FOR_INSTANCE:
		retval = actionForInstance(hma);
		break;
	case ACTION_PRINT:
		retval = actionPrint(hma);
		break;
	case ACTION_PRINT_SUITES:
		retval = actionPrintSuites(hma);
		break;
	}
	return( retval );
}

/* The default action, just print out the reply in some format. */
static
int actionDefault(heyMsgAction *hma)
{
	int lpc, retval = 0;
	char *name, *value;
	jsiMessage *jm;

	require(hma != 0);
	require(hma->msg != 0);

	jm = hma->msg;
	switch( jm->what )
	{
	case JSI_ERROR:
		printf("Error:\n");
		break;
	case JSI_REPLY:
		printf("Normal reply:\n");
		retval = 1;
		break;
	}
	for( lpc = 0; lpc < jm->identifiers.count; lpc++ )
	{
		jsiMessageGetIdentifier(jm, lpc, &name, &value);
		printf("  %s = `%s'\n",
		       name,
		       value);
	}
	return( retval );
}

/* Our own printf so we can handle our own escapes */
static
void localPrintf(char *str)
{
	require(str != 0);
	
	while( *str )
	{
		switch( *str )
		{
		case '\\':
			switch( *(str + 1) )
			{
			case 0:
				return;
			case 'n':
				putchar('\n');
				break;
			case 't':
				putchar('\t');
				break;
			default:
				putchar(*(str + 1));
				break;
			}
			str++;
			break;
		default:
			putchar(*str);
			break;
		}
		str++;
	}
}

/*
 * Print a values of a message identifier through a range using the given
 * string to separate each instance.
 */
static
int printMessageValues(jsiMessage *jm, char *name, int kind, char *format,
		       int index_start, char *sep, int index_end)
{
	int lpc, lpc2, len, retval = 1;
	localVariable *lv;
	union {
		char c;
		int i;
		double f;
		char *s;
	} value;

	require(jm != 0);
	require(name != 0);
	require(format != 0);
	
	if( index_start > index_end )
	{
		index_start = -jsiMessageIdentifierCount(jm, name);
	}
	/* Walk through each instance in range */
	for( lpc = index_start;
	     (lpc <= index_end) && retval;
	     lpc++ )
	{
		if( ((kind == JMA_Bytes) &&
		     (retval = jsiMessageGet(jm,
					     kind, name, lpc, &value, &len,
					     JMA_TAG_DONE))) ||
		    ((kind != JMA_Bytes) &&
		     (retval = jsiMessageGet(jm,
					     kind, name, lpc, &value,
					     JMA_TAG_DONE))) )
		{
			if( (lpc > index_start) && sep )
				localPrintf(sep);
			/*
			 * This is kinda silly, we want to use the format given
			 * on the command line for printf, so we need to put
			 * a `%' at the start so it will accept it.  However,
			 * this could also be the end of `name' so we need
			 * to null out the character below to make it a
			 * regular string again...
			 */
			format[0] = '%';
			switch( kind )
			{
			case JMA_Bytes:
				for( lpc2 = 0; lpc2 < len; lpc2++ )
				{
					printf("%02x", value.s[lpc2] & 0xff);
				}
				break;
			case JMA_Character:
				printf(format, value.c);
				break;
			case JMA_Integer:
				printf(format, value.i);
				break;
			case JMA_Float:
				printf(format, value.f);
				break;
			case JMA_String:
				/* XXX This isn't very elegant... */
				len = strlen(value.s);
				for( lpc2 = 0; lpc2 < len; lpc2++ )
				{
					char ch;
					
					ch = value.s[lpc2] & 0xff;
					if( isprint(ch) )
						putchar(ch);
					else
						printf("\\x%02x", ch);
				}
				break;
			default:
				break;
			}
			format[0] = 0;
		}
		else if( (lv = findVariable(name)) )
		{
			switch( kind )
			{
			case JMA_Integer:
				printf("%d", lv->value);
				/* It got set to zero by the jsiMessageGet. */
				retval = 1;
				break;
			default:
				printError("Invalid format for local "
					   "variable.");
				retval = 0;
				break;
			}
		}
		else
		{
			printError("Couldn't find message identifier `%s' "
				   "instance %d", name, lpc);
		}
	}
	return( retval );
}

/*
 * Parse the message identifier format in the print string.  See the usage
 * message for a description of the format.
 */
static
int parseMessageVariable(jsiMessage *jm, char **fmt_ptr)
{
	char *name_start, *name_end, *pformat_start, *pformat_end;
	int kind, index_start = 0, index_end = 0, retval = 0;
	char *sep_start = 0, *sep_end = 0;
	char *fmt = *fmt_ptr;
	char type, name_save;
	localVariable *lv;

	require(jm != 0);
	require(fmt_ptr != 0);
	
	/* parsing `%(<name>:' || `%(<name>[' */
	fmt++;
	switch( *fmt )
	{
	case '(':
		break;
	case '%':
		putchar('%');
		*fmt_ptr = fmt;
		return( 1 );
	default:
		printError("Illegal character following %%: %c", *fmt);
		return( 0 );
	}
	fmt++;
	name_start = fmt;
	for( name_end = name_start;
	     *name_end && (*name_end != ':') && (*name_end != '[');
	     name_end++ );
	name_save = *name_end;
	switch( *name_end )
	{
	case ':':
		*name_end = 0;
		pformat_start = name_end;
		break;
	case '[':
		{
			char *range_start, *range_end;

			/* parsing `<index>]' || `<index>.<sep>.<index>]:' */
			*name_end = 0;
			range_start = name_end + 1;
			for( range_end = range_start;
			     *range_end && (*range_end != ']') &&
				     (*range_end != '.');
			     range_end++ );
			switch( *range_end )
			{
			case ']':
				*range_end = 0;
				/* Check for integer index or variable */
				if( sscanf(range_start,
					   "%d",
					   &index_start) == 1 )
				{
				}
				else if( (lv = findVariable(range_start)) )
				{
					index_start = lv->value;
				}
				else
				{
					printError("Unknown index value: %s",
						   range_start);
					return( 0 );
				}
				*range_end = ']';
				index_end = index_start;
				break;
			case '.':
				*range_end = 0;
				/*
				 * Its a range, get the start/end and separator
				 */
				sep_start = range_end + 1;
				if( (range_start == range_end) ||
				    (sscanf(range_start,
					    "%d",
					    &index_start) == 1) )
				{
					/*
					 * Integer or nothing, which implicitly
					 * means `0'
					 */
				}
				else if( (lv = findVariable(range_start)) )
				{
					index_start = lv->value;
				}
				else
				{
					printError("Unknown range value: %s",
						   range_start);
					return( 0 );
				}
				*range_end = '.';
				for( sep_end = sep_start;
				     *sep_end && (*sep_end != '.');
				     sep_end++ );
				if( *sep_end != '.' )
				{
					printError("No terminating '.' in "
						   "print range.");
					return( 0 );
				}
				*sep_end = 0;
				range_start = sep_end + 1;
				for( range_end = range_start;
				     *range_end && (*range_end != ']');
				     range_end++ );
				if( *range_end != ']' )
				{
					printError("No terminating ']' in "
						   "print range.");
					return( 0 );
				}
				*range_end = 0;
				if( range_start == range_end )
				{
					/* No index implies the last one */
					index_end =
						jsiMessageIdentifierCount(
							jm, name_start) - 1;
				}
				else if( sscanf(range_start,
						"%d",
						&index_end) == 1 )
				{
				}
				else if( (lv = findVariable(range_start)) )
				{
					index_end = lv->value;
				}
				else
				{
					printError("Unknown range value: %s",
						   range_start);
					return( 0 );
				}
				*range_end = ']';
				break;
			default:
				printError("No terminating ']' in "
					   "print range.");
				return( 0 );
			}
			if( *(range_end + 1) != ':' )
			{
				printError("No print format given.");
				return( 0 );
			}
			pformat_start = range_end + 1;
		}
		break;
	default:
		printError("No print format given.");
		return( 0 );
	}
	/*
	 * parsing `<pformat>)'  (Note: we don't allow `%'s in the format
	 * since it will be passed directly to printf, and they could screw
	 * us up.)
	 */
	for( pformat_end = pformat_start + 1;
	     *pformat_end && (*pformat_end != ')') && (*pformat_end != '%');
	     pformat_end++ );
	if( *pformat_end != ')' )
	{
		printError("No print format given.");
		return( 0 );
	}

	/*
	 * Successful parse, now figure out what type of data to get from the
	 * message.  We just look at the last character of the format to
	 * see if its a valid.
	 */
	*name_end = 0;
	*pformat_end = 0;
	type = *(pformat_end - 1);
	switch( type )
	{
	case 'b':
		kind = JMA_Bytes;
		break;
	case 'c':
		kind = JMA_Character;
		break;
	case 'd':
	case 'x':
		kind = JMA_Integer;
		break;
	case 'f':
		kind = JMA_Float;
		break;
	case 's':
		kind = JMA_String;
		break;
	default:
		printError("Unknown print format: %c", type);
		return( 0 );
	}
	/* Print out the identifier range */
	retval = printMessageValues(jm, name_start, kind,
				    pformat_start, index_start,
				    sep_start, index_end);
	/* Restore the format string so we still can process it again later */
	*name_end = name_save;
	*pformat_start = ':';
	*pformat_end = ')';
	if( sep_end )
		*sep_end = '.';
	/* Move the format pointer forward */
	*fmt_ptr = pformat_end;
	return( retval );
}

/*
 * The `print' action handler.  We provide a printf like formatter so users
 * can specify a custom print out if the default one is insufficient.
 */
static
int actionPrint(heyMsgAction *hma)
{
	char **actionValues = hma->actionValues;
	jsiMessage *jm = hma->msg;
	int lpc, retval = 1;

	require(hma != 0);
	
	/* Treat any additional arguments as strings to be printed */
	for( lpc = 0; lpc < hma->actionValueCount; lpc++ )
	{
		char *fmt, *pos;

		fmt = actionValues[lpc];
		pos = fmt;
		while( *pos && retval )
		{
			switch( *pos )
			{
			case '%':
				/* A msg-id format request */
				if( !(retval =
				      parseMessageVariable(jm, &pos)) )
				{
#if defined(HAVE_FPURGE)
					fpurge(stdout);
#endif
					printError("Error in print for "
						   "message:");
					actionDefault(hma);
				}
				break;
			case '\\':
				/* an escape */
				switch( *(pos + 1) )
				{
				case 0:
					return( 0 );
				case 'n':
					putchar('\n');
					break;
				case 't':
					putchar('\t');
					break;
				default:
					putchar(*(pos + 1));
					break;
				}
				pos++;
				break;
			default:
				putchar(*pos);
				break;
			}
			pos++;
		}
	}
	return( retval );
}

/*
 * The `forinstance' action handler.  This is meant to be used in conjunction
 * with `print' so that you can have a more flexible printing than just the
 * separator string in an instance range.
 */
static
int actionForInstance(heyMsgAction *hma)
{
	int retval = 0;

	require(hma != 0);
	
	if( hma->actionValueCount > 3 )
	{
		heyMsgAction subHma;
		localVariable lv;
		char *name;
		int count;

		/* parsing `<lv.name> in <name>' */
		lv.name = hma->actionValues[0];
		lv.value = 0;
		if( strcmp(hma->actionValues[1], "in") )
		{
#if defined(HAVE_FPURGE)
			fpurge(stdout);
#endif
			printError("Missing `in' operator in forinstance.");
			actionDefault(hma);
			return( 0 );
		}
		name = hma->actionValues[2];
		/* Count the identifier instances */
		count = jsiMessageIdentifierCount(hma->msg, name);
		if( (subHma.action = mapStringToInt(actionMappings,
						    hma->actionValues[3])) )
		{
			/* Create a new heyMsgAction for the `print' */
			subHma.msg = hma->msg;
			subHma.actionValues = hma->actionValues + 4;
			subHma.actionValueCount = hma->actionValueCount - 4;
			/* Add our iterator variable */
			addVariable(&lv);
			retval = 1;
			/* Do the loop and call our sub action */
			for( lv.value = 0;
			     (lv.value < count) && retval;
			     lv.value++ )
			{
				retval = doAction(&subHma);
			}
			remVariable(&lv);
		}
	}
	else
	{
#if defined(HAVE_FPURGE)
		fpurge(stdout);
#endif
		printError("Not enough forinstance arguments.");
		printError("Usage: forinstance <local name> in <msg id>.");
		actionDefault(hma);
	}
	return( retval );
}

extern stringIntMap verbMappings[];

static
void printSupportedMessages(int commands)
{
	int lpc, p_count = 0;

	for( lpc = 0; lpc < (sizeof(int) * 8); lpc++ )
	{
		if( commands & (1L << lpc) )
		{
			printf("%s%s",
			       (p_count > 0) ? ", " : "  ",
			       mapIntToString(verbMappings, lpc));
			p_count++;
		}
	}
	printf("\n");
}

static
void printPropertyUsage(char *usage)
{
	char *pos = usage;
	
	printf("Usage (arguments, if any, followed by a description):\n  ");
	while( *pos )
	{
		switch( *pos )
		{
		case '\n':
			putc('\n', stdout);
			putc(' ', stdout);
			putc(' ', stdout);
			break;
		case '\t':
			putc(' ', stdout);
			putc(' ', stdout);
			break;
		default:
			putc(*pos, stdout);
			break;
		}
		pos++;
	}
	printf("\n");
}

static
int actionPrintSuites(heyMsgAction *hma)
{
	jsiMessage *jm = hma->msg;
	int retval = 1;

	if( jm->what == JSI_ERROR )
	{
		retval = actionDefault(hma);
	}
	else
	{
		char *id_name, *id_value;
		jsiPropertyInfo *pi;
		int lpc;
		
		for( lpc = 0; lpc < jm->identifiers.count; lpc++ )
		{
			jsiMessageGetIdentifier(jm, lpc, &id_name, &id_value);
			if( (pi = jsiCreatePropertyInfo(JPIA_Internal,
							id_value,
							JPIA_TAG_DONE)) )
			{
				printf("Property:  %s\n"
				       "  Supported verbs:",
				       pi->name);
				printSupportedMessages(pi->commands);
				if( pi->specifiers[0] )
				{
					printf("  Specifiers:  %s\n",
					       pi->specifiers);
				}
				if( pi->usage[0] )
					printPropertyUsage(pi->usage);
			}
		}
	}
	return( retval );
}
