/* -*- C -*- */

/*
 * jsiImpl-inline.h
 * Low level functions required by the JanosVM Scripting Interface layer
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _unix_jsi_impl_h
#define _unix_jsi_impl_h

/* Inline jsiImpl functions that are implemented with plain posix calls */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/stat.h>

JSI_INLINE
void *jsiMalloc(int size)
{
	void *retval;

	if( (retval = malloc(size)) )
		memset(retval, 0, size);
	return( retval );
}

JSI_INLINE
void *jsiRealloc(void *old, int size)
{
	return( realloc(old, size) );
}

JSI_INLINE
void jsiFree(void *mem)
{
	free(mem);
}

JSI_INLINE
void _jsiLock(JSI_LOCK_PARAM lock, int *iLockRoot)
{
}

JSI_INLINE
void _jsiUnlock(JSI_LOCK_PARAM lock, int *iLockRoot)
{
}

int _jsiWait(JSI_LOCK_PARAM lock, int ms_timeout);

void _jsiSignal(JSI_LOCK_PARAM lock);

JSI_INLINE
int jsiOpen(char *filename, int flags, int mode, int *out_fd)
{
	int retval = 0;

	if( (*out_fd = open(filename, flags, mode)) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiFStat(int fd, struct stat *st)
{
	return( fstat(fd, st) );
}

JSI_INLINE
int jsiStat(char *filename, struct stat *st)
{
	return( stat(filename, st) );
}

JSI_INLINE
int jsiRemove(char *filename)
{
	return( remove(filename) );
}

JSI_INLINE
int jsiMkDir(char *dirname, int mode)
{
	return( mkdir(dirname, mode) );
}

JSI_INLINE
int jsiRmDir(char *dirname)
{
	return( rmdir(dirname) );
}

JSI_INLINE
int jsiSocket(int domain, int type, int protocol, int *out_fd)
{
	int retval = 0;

	if( (*out_fd = socket(domain, type, protocol)) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiConnect(int sockfd, struct sockaddr *serv, int addrlen)
{
	int retval = 0;

	if( connect(sockfd, serv, addrlen) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiBind(int sockfd, struct sockaddr *my, int addrlen)
{
	int retval = 0;

	if( bind(sockfd, my, addrlen) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiListen(int sockfd, int backlog)
{
	int retval = 0;

	if( listen(sockfd, backlog) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiAccept(int sockfd, struct sockaddr *addr, int *addrlen, int *out_fd)
{
	int retval = 0;

	if( (*out_fd = accept(sockfd, addr, addrlen)) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiRecvFrom(int fd, void *buffer, int len, unsigned int flags,
		struct sockaddr *from, int *fromlen, int *out_read)
{
	int retval = 0;

	if( (*out_read = recvfrom(fd, buffer, len, flags, from, fromlen)) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiSend(int fd, void *buffer, int len, unsigned int flags)
{
	int retval = 0;

	if( send(fd, buffer, len, flags) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiSendTo(int fd, void *buffer, int len, unsigned int flags,
	      struct sockaddr *to, int tolen)
{
	int retval = 0;

	if( sendto(fd, buffer, len, flags, to, tolen) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiSeek(int fd, off_t offset, int whence, off_t *out_offset)
{
	int retval = 0;

	if( (*out_offset = lseek(fd, offset, whence)) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiRead(int fd, void *buffer, size_t bufferSize, size_t *out_read)
{
	int retval = 0;

	if( (*out_read = read(fd, buffer, bufferSize)) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiReadAndWait(int fd, void *buffer, size_t bufferSize, int sample_time,
		   size_t *out_read)
{
	int rc, done = 0, retval = 0;
	fd_set read_fds;

	while( !done )
	{
		rc = read(fd, buffer, bufferSize);
		switch( rc )
		{
		case -1:
			done = 1;
			retval = errno;
			break;
		case 0:
			FD_ZERO(&read_fds);
			FD_SET(fd, &read_fds);
			select(fd + 1, &read_fds, 0, 0, 0);
			break;
		default:
			*out_read = rc;
			done = 1;
			break;
		}
	}
	return( retval );
}

JSI_INLINE
int jsiWrite(int fd, void *buffer, size_t bufferSize, size_t *out_written)
{
	int retval = 0;

	if( (*out_written = write(fd, buffer, bufferSize)) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiShutdown(int fd, int how)
{
	int retval = 0;

	if( shutdown(fd, how) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiClose(int fd)
{
	int retval = 0;

	if( close(fd) == -1 )
		retval = errno;
	return( retval );
}

JSI_INLINE
int jsiGetHostByName(char *name, struct hostent **he)
{
	int retval = 0;

	if( (*he = gethostbyname(name)) == 0 )
		retval = h_errno;
	return( retval );
}

JSI_INLINE
int jsiGetHostByAddr(char *addr, int len, int type, struct hostent **he)
{
	int retval = 0;

	if( (*he = gethostbyaddr(addr, len, type)) )
		retval = h_errno;
	return( retval );
}

JSI_INLINE
int jsiSetSockOpt(int sock, int level, int optname, void *optval, int optlen)
{
	int retval = 0;

	if( setsockopt(sock, level, optname, optval, optlen) == -1 )
		retval = errno;
	return( retval );
}

#endif
