/*
 * jsi.h
 * JanosVM Scripting Interface functions
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>

#include "assert_pp.h"
#include "jsiImpl.h"

#include "jsi.h"
#include "jsiPlexLink.h"
#include "jsiPlexChannel.h"
#include "jsiPlexWindow.h"
#include "jsiPlexUDP.h"

#include <misc/stringParsing.h>

static
jsiPlexUDP *plexUDP;

static
int jsiSignalled = 0;

void _jsiLock(JSI_LOCK_PARAM lock, int *iLockRoot)
{
}

void _jsiUnlock(JSI_LOCK_PARAM lock, int *iLockRoot)
{
}

void _jsiSignal(JSI_LOCK_PARAM lock)
{
	jsiSignalled = 1;
}

static
void obtainUDPLink(jsiPlexLink *pl, int *iLockRoot)
{
}

static
void releaseUDPLink(jsiPlexLink *pl, int *iLockRoot)
{
}

struct recvContext {
	jsiPlexLinkContext plc;
	int fromlen;
	struct sockaddr_in saddr;
};

static
int waitForUDPLink(jsiPlexLink *pl, int ms_timeout)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;
	struct timeval *tv = 0, tv_data;
	fd_set read_fds;
	int retval = 1;
	
	jsiSignalled = 0;
	while( !jsiSignalled && (retval != 0) )
	{
		if( ms_timeout )
		{
			tv = &tv_data;
			tv->tv_sec = ms_timeout / 1000;
			tv->tv_usec = (ms_timeout % 1000) * 1000;
		}
		FD_ZERO(&read_fds);
		FD_SET(pu->fd, &read_fds);
		retval = select(pu->fd + 1, &read_fds, 0, 0, tv);
		if( retval > 0 )
		{
			struct recvContext rc;
			jsiPlexUDPIn pui;
			int packetSize;

			memset(&rc, 0, sizeof(rc));
			pui.link.link = pl;
			rc.fromlen = sizeof(rc.saddr);
			if( !jsiRecvFrom(pu->fd,
					 &pui.recvBuffer.data.header,
					 sizeof(jsiPlexLinkHeader) +
					 JSI_PLEX_UDP_MAX_BODY_SIZE,
					 0,
					 (struct sockaddr *)&rc.saddr,
					 &rc.fromlen,
					 &packetSize) )
			{
				jsiPlexLinkPacket *plp = &pui.recvBuffer.data;
				
				plp->packetSize = packetSize;
				plp->header.fragOffset =
					ntohs(plp->header.fragOffset);
				plp->header.totalLength =
					ntohs(plp->header.totalLength);
				plp->header.messageID =
					ntohs(plp->header.messageID);
				plp->header.channelID =
					ntohs(plp->header.channelID);
				rc.plc.hashValue = rc.saddr.sin_addr.s_addr;
				jsiPlexWindowRecvPacket(&pui.link,
							0,
							&rc.plc,
							plp);
			}
			else
			{
				jsiSignalled = 1;
			}
		}
		else if( retval < 0 )
		{
			fprintf(stderr,
				"Error in select: %s\n",
				strerror(errno));
		}
	}
	jsiSignalled = 0;
	return( retval );
}

static
int processUDPPackets(jsiPlexOutLink *pol, int ms_timeout)
{
	return( waitForUDPLink(pol->link, ms_timeout) );
}

int _jsiWait(JSI_LOCK_PARAM lock, int ms_timeout)
{
	return( waitForUDPLink(&plexUDP->link, ms_timeout) );
}

static
void signalUDPLink(jsiPlexLink *pl)
{
	jsiSignalled = 1;
}

static
jsiChannel *handleUDPLink(jsiPlexInLink *pil)
{
	return( 0 );
}

static
int writeUDPLink(jsiPlexOutLink *pol, void *buffer, int len)
{
	int tolen = sizeof(struct sockaddr_in), retval = 0;
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;

	if( jsiSendTo(((jsiPlexUDP *)puo->link.link)->fd, buffer, len, 0,
		      (struct sockaddr *)&puo->dest, tolen ) == 0 )
	{
		retval = 1;
	}
	else
	{
		pol->flags &= ~POLF_UP;
	}
	return( retval );
}

static
jsiPlexInLink *getUDPIn(jsiPlexLink *pl, char *format)
{
	parsedString protocol, host;
	jsiPlexInLink *retval = 0;
	int port;

	if( parseString(format,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_NonEmptyString, &host,
			SPO_Expect, "/",
			SPO_Integer, &port,
			SPO_End) )
	{
		jsiPlexUDPIn *pui;
		int sock = -1;
		
		if( jsiSocket(AF_INET, SOCK_DGRAM, 0, &sock) == 0 )
		{
			struct sockaddr_in saddr;
			int on_off = 1;

			memset(&saddr, 0, sizeof(saddr));
			saddr.sin_family = AF_INET;
			saddr.sin_addr.s_addr = htonl(INADDR_ANY);
			saddr.sin_port = htons(port);
			jsiSetSockOpt(sock, SOL_SOCKET, SO_REUSEADDR,
				      &on_off, sizeof( int ));
			if( !jsiBind(sock,
				     (struct sockaddr *)&saddr,
				     sizeof(saddr)) )
			{
			}
			else
			{
				jsiClose(sock);
				sock = -1;
			}
		}
		if( (sock != -1) &&
		    (pui = (jsiPlexUDPIn *)jsiMalloc(sizeof(jsiPlexUDPIn))) )
		{
			pui->link.link = pl;
			pui->link.provider = 0;
			pui->fd = sock;
			retval = &pui->link;
		}
	}
	return( retval );
}

static
void dropUDPIn(jsiPlexInLink *pil)
{
	jsiPlexUDPIn *pui = (jsiPlexUDPIn *)pil;

	jsiClose(pui->fd);
	jsiFree(pil);
}

static
jsiPlexOutLink *getUDPOut(jsiPlexLink *pl, struct sockaddr_in *saddr)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;
	jsiPlexUDPOut *puo, *match = 0;
	jsiPlexOutLink *retval = 0;

	puo = pu->outLinks;
	while( puo && !match )
	{
		if( !memcmp(&puo->dest, saddr, sizeof(struct sockaddr_in)) )
			match = puo;
		puo = puo->next;
	}
	if( match )
	{
		match->link.users++;
		retval = &match->link;
	}
	else if( (puo = (jsiPlexUDPOut *)jsiMalloc(sizeof(jsiPlexUDPOut))) )
	{
		puo->link.link = pl;
		puo->link.flags = POLF_UP;
		puo->link.hashValue = saddr->sin_addr.s_addr;
		puo->link.users = 1;
		puo->link.packetPool = 0;
		puo->next = pu->outLinks;
		pu->outLinks = puo;
		puo->dest = *saddr;
		retval = &puo->link;
	}
	return( retval );
}

static
jsiPlexOutLink *getUDPOutFromFormat(jsiPlexLink *pl, char *format)
{
	parsedString protocol, host;
	jsiPlexOutLink *retval = 0;
	int port;

	if( parseString(format,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_NonEmptyString, &host,
			SPO_Expect, "/",
			SPO_Integer, &port,
			SPO_End) )
	{
		struct sockaddr_in host_addr;
		struct hostent *host_ent;
		int valid = 1;
		char save;

		save = host.data[host.len];
		host.data[host.len] = '\0';
		memset(&host_addr, 0, sizeof(host_addr));
#if !defined(__linux__)
		host_addr.sin_len = sizeof(host_addr);
#endif
		host_addr.sin_family = AF_INET;
		if( !jsiGetHostByName(host.data, &host_ent) )
		{
			memcpy((char *)&host_addr.sin_addr.s_addr,
			       host_ent->h_addr,
			       host_ent->h_length);
		}
		else if( (host_addr.sin_addr.s_addr =
			  inet_addr(host.data)) == -1 )
		{
			/* Its not a number address either */
			valid = 0;
		}
		host_addr.sin_port = htons(port);
		host.data[host.len] = save;
		if( valid )
			retval = getUDPOut(pl, &host_addr);
	}
	return( retval );
}

static
jsiPlexOutLink *getUDPOutFromContext(jsiPlexLink *pl, jsiPlexLinkContext *plc)
{
	struct recvContext *rc = (struct recvContext *)plc;
	jsiPlexOutLink *retval;

	retval = getUDPOut(pl, &rc->saddr);
	return( retval );
}

static
void dropUDPOut(jsiPlexOutLink *pol)
{
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;

	require(puo != 0);
	require(puo->link.users > 0);

	puo->link.users--;
	if( puo->link.users == 0 )
	{
		jsiPlexLinkPacket *plp, *plp_next;
		jsiPlexUDPOut *curr, **prev;

		curr = ((jsiPlexUDP *)pol->link)->outLinks;
		prev = &((jsiPlexUDP *)pol->link)->outLinks;
		while( curr != puo )
		{
			prev = &curr->next;
			curr = curr->next;
		}
		*prev = curr->next;
		plp = pol->packetPool;
		while( plp )
		{
			plp_next = plp->next;
			jsiFree(plp);
			plp = plp_next;
		}
		jsiFree(puo);
	}
}

int compareUDPOut(jsiPlexOutLink *pol, jsiPlexLinkContext *plc)
{
	struct recvContext *rc = (struct recvContext *)plc;
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;
	int retval = 0;

	if( (puo->dest.sin_addr.s_addr == rc->saddr.sin_addr.s_addr) &&
	    (puo->dest.sin_port == rc->saddr.sin_port) )
		retval = 1;
	return( retval );
}

jsiPlexLinkFunctions jsiPlexUDPFunctions = {
	0,
	0,
	0,
	obtainUDPLink,
	releaseUDPLink,
	waitForUDPLink,
	signalUDPLink,
	handleUDPLink,
	writeUDPLink,
	getUDPIn,
	dropUDPIn,
	getUDPOutFromFormat,
	getUDPOutFromContext,
	dropUDPOut,
	compareUDPOut,
	0,
	0,
	processUDPPackets
};

static
char *plexUDPNames[] = {
	"udp",
	0
};

jsiPlexLink *jsiCreatePlexUDP()
{
	jsiPlexLink *retval = 0;
	int sock;

	if( jsiSocket(AF_INET, SOCK_DGRAM, 0, &sock) == 0 )
	{
		jsiPlexUDP *pu;

		if( (pu = (jsiPlexUDP *)jsiMalloc(sizeof(jsiPlexUDP))) )
		{
			pu->link.next = 0;
			pu->link.names = plexUDPNames;
			pu->link.ops = &jsiPlexUDPFunctions;
			pu->link.maxPacketSize = JSI_PLEX_UDP_MAX_BODY_SIZE;
			pu->fd = sock;
			pu->outLinks = 0;
			plexUDP = pu; // XXX
			retval = &pu->link;
		}
	}
	return( retval );
}
