/*
 * main.c
 * The main code for the jsihey command line scripting tool
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/*
 * The jsihey program is a command line interface to the scripting protocol
 * used by the JVM.  Its a very basic program that is only intended as an
 * example usage of the JSI code and to provide a lowest common denominator
 * when scripting is required.  A real scripting language with a JSI extension
 * should be preferred over this feeble program.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "assert_pp.h"

#include "jsi.h"
#include "jsiImpl.h"

#include "jsiPlexUDP.h"

#include "heyTools.h"
#include "heyActions.h"
#include "heyHandlers.h"

#define RETCODE_OK 0
#define RETCODE_MSG_ERROR 10
#define RETCODE_COMM_ERROR 20
#define RETCODE_USAGE_ERROR 30
#define RETCODE_INTERNAL 100

int retcode = RETCODE_USAGE_ERROR;

/* The name used to execute us */
char *progName = "jsihey";

/* A BNF like usage message for the user */
char *usageFormat =
"usage - \n\n"
"\t%s: destination msg\n"
"\t;\n"
"\n"
"\tmsg: `getsuites' gs-specifier-list\n"
"\t   | `list' opt-specifier-list opt-action\n"
"\t   | `get' specifier-list opt-action\n"
"\t   | `set' specifier-list `to' value\n"
"\t   | `create' specifier-list args opt-action\n"
"\t   | `delete' specifier-list\n"
"\t   | `execute' specifier-list args opt-action\n"
"\t   | `shutdown' or `exit' or `quit'\n"
"\t   ;\n"
"\n"
"\tdestination: registered-name\n"
"\t           | transport ':' host '/' port\n"
"\t           ;\n"
"\n"
"\tgs-specifier-list:\n"
"\t                 | `of' specifier-list\n"
"\t                 ;\n"
"\n"
"\topt-specifier-list:\n"
"\t                  | specifier-list\n"
"\t                  ;\n"
"\n"
"\tspecifier-list: specifier\n"
"\t              | specifier-list `of' specifier\n"
"\t              ;\n"
"\n"
"\tspecifier: name\n"
"\t         | name value\n"
"\t         ;\n"
"\n"
"\targs:\n"
"\t    | `with' args-list\n"
"\t    ;\n"
"\n"
"\targs-list: name value\n"
"\t         | name `=' value\n"
"\t         | args-list name value\n"
"\t         | args-list name `=' value\n"
"\t         ;\n"
"\n"
"\topt-action:\n"
"\t          | `then' action\n"
"\t          ;\n"
"\n"
"\taction: `forinstance' iterator `in' msg-id action\n"
"\t      | `print' `\"'print-format`\"'\n"
"\t      | `ignore'\n"
"\t      | `nothing'\n"
"\t      ;\n"
"\n"
"\tprint-format:\n"
"\t            | print-format plain-text\n"
"\t            | print-format `%%(' msg-id range `:' printf-format `)'\n"
"\t            ;\n"
"\n"
"\trange:\n"
"\t     | `[' index `]'\n"
"\t     | `[' start-index `.' separator-string `.' end-index `]'\n"
"\t     ;\n";

/* Print the usage message */
static
void printUsage(void)
{
	fprintf(stderr, usageFormat, progName);
}

/*
 * Figure out the destination of the message.  The destination string can
 * either be a registry value or a formatted address.
 */
static
jsiChannel *getMessageDestination(char *dest)
{
	char *fileStr = 0;
	jsiChannel *retval = 0;

	require(dest != 0);
	
	/* Look for it in the registry */
	if( (fileStr = jsiFindInterface(dest)) )
	{
		dest = fileStr;
	}
	if( dest )
	{
		if( !(retval = jsiConnectChannelTo(dest, "tcp")) )
		{
			printError("Couldn't connect to destination `%s'",
				   dest);
		}
		/* If fileStr != 0, free the string we got from the registry */
		jsiFree(fileStr);
	}
	return( retval );
}

/* A simple set of states for parsing the command line */
typedef enum {
	EXPECT_NONE,		/* expecting nothing */
	EXPECT_SPECIFIER_NAME,	/* expecting the specifier name */
	EXPECT_SPECIFIER_VALUE,	/* expecting specifier value or `of' */
	EXPECT_OF,		/* expecting `of' */
	EXPECT_PREPOSITION,	/* expecting a preposition */
	EXPECT_SET_DATA,	/* expecting the value for a `set' message */
	EXPECT_ARG_NAME,	/* expecting the name of an argument */
	EXPECT_ARG_VALUE,	/* expecting the value of an argument */
	EXPECT_THEN_ACTION,	/* expecting action for `then' */
	EXPECT_MAX
} expect_state_t;

/* Handy strings to tell the user what we were expecting */
static
char *expectStrings[] = {
	0,
	"specifier name",
	"specifier value",
	"of",
	"preposition",
	"value to set in property",
	"argument name",
	"argument value",
};

/* Print out a parsing error message */
static
void parseError(expect_state_t expect, char *msg, ...)
{
	va_list args;

	require(expect >= 0);
	require(expect < EXPECT_MAX);
	
	va_start(args, msg);
	if( msg[0] != 0 )
	{
		fprintf(stderr, "Parse error: ");
		vfprintf(stderr, msg, args);
		fprintf(stderr, ".\n");
	}
	if( expectStrings[expect] )
		fprintf(stderr, "Expecting %s\n", expectStrings[expect]);
	va_end(args);
	printUsage();
}

/* Mappings for all of the `what' values we care about */
stringIntMap verbMappings[] = {
	{ "shutdown", JSI_SHUTDOWN },
	{ "quit", JSI_SHUTDOWN },
	{ "exit", JSI_SHUTDOWN },
	{ "getsuites", JSI_GET_SUITES },
	{ "list", JSI_LIST_PROPERTIES },
	{ "get", JSI_GET_PROPERTY },
	{ "set", JSI_SET_PROPERTY },
	{ "create", JSI_CREATE_PROPERTY },
	{ "delete", JSI_DELETE_PROPERTY },
	{ "execute", JSI_EXECUTE_PROPERTY },
	{ 0, JSI_ERROR }
};

/*
 * Check for a preposition (of/with/to) and set the state accordingly.  It
 * also checks for `then' and ends processing (yes I know its not a
 * preposition).
 */
static
int handlePreposition(jsiMessage *jm, expect_state_t *expect,
		      char *spec_name, char *spec_value)
{
	expect_state_t start_expect = *expect;
	char *preposition = "";
	int retval = 1;

	require(jm != 0);
	require(expect != 0);
	require(*expect >= 0);
	require(*expect < EXPECT_MAX);
	require(spec_name != 0);

	switch( *expect )
	{
	case EXPECT_SPECIFIER_NAME:
		/* We just got a specifier value so we might get a prep here */
		preposition = spec_name;
		break;
	case EXPECT_SPECIFIER_VALUE:
		/*
		 * If there specifier name is enough they can use a prep here
		 */
		preposition = spec_value;
		break;
	case EXPECT_PREPOSITION:
		preposition = spec_name;
		break;
	default:
		assert(0);
		break;
	}
	/* Check for a prep. */
	if( !strcasecmp(preposition, "of") )
	{
		spec_value = "";
		*expect = EXPECT_SPECIFIER_NAME;
	}
	else if( !strcasecmp(preposition, "with") )
	{
		spec_value = "";
		/* Make sure we're doing an `execute' or 'create' */
		if( (jm->what == JSI_EXECUTE_PROPERTY) ||
		    (jm->what == JSI_CREATE_PROPERTY) )
		{
			*expect = EXPECT_ARG_NAME;
		}
		else
		{
			parseError(EXPECT_NONE,
				   "`%s' doesn't take arguments",
				   mapIntToString(verbMappings, jm->what));
			retval = 0;
		}
	}
	else if( !strcasecmp(preposition, "to") )
	{
		spec_value = "";
		/* Make sure we're doing a `set' */
		if( jm->what == JSI_SET_PROPERTY )
		{
			*expect = EXPECT_SET_DATA;
		}
		else
		{
			parseError(*expect,
				   "`to' is only used with `set'");
			retval = 0;
		}
	}
	else if( !strcasecmp(preposition, "then") )
	{
		spec_value = "";
		/* Make sure there are specifiers for all but a `list' */
		if( (jm->what != JSI_LIST_PROPERTIES) &&
		    (*expect != EXPECT_SPECIFIER_VALUE) &&
		    !jm->specifiers.count )
		{
			parseError(*expect,
				   "Action used without a specifier");
			retval = 0;
		}
		else
		{
			*expect = EXPECT_THEN_ACTION;
		}
	}
	else /* No preposition */
	{
		switch( *expect )
		{
		case EXPECT_SPECIFIER_NAME:
			*expect = EXPECT_SPECIFIER_VALUE;
			break;
		case EXPECT_SPECIFIER_VALUE:
			*expect = EXPECT_PREPOSITION;
			break;
		case EXPECT_PREPOSITION:
			*expect = EXPECT_SPECIFIER_VALUE;
			break;
		default:
			assert(0);
			break;
		}
	}
	if( retval )
	{
		switch( start_expect )
		{
		case EXPECT_SPECIFIER_NAME:
			if( (*expect != EXPECT_SPECIFIER_VALUE) &&
			    (*expect != EXPECT_THEN_ACTION) &&
			    (*expect != EXPECT_ARG_NAME) &&
			    (*expect != EXPECT_SET_DATA) )
			{
				/* We didn't get a specifier name, tell them */
				parseError(start_expect, "");
				retval = 0;
			}
			break;
		case EXPECT_SPECIFIER_VALUE:
			/*
			 * We started out expecting a specifier value, so we
			 * must've gotten it or there was a prep, push the
			 * specifier onto the message.
			 */
			if( spec_name[0] )
			{
				retval = jsiPushMessageSpecifier(
					jm, JMA_String, spec_name, spec_value);
			}
			break;
		case EXPECT_PREPOSITION:
			if( (*expect != EXPECT_SPECIFIER_NAME) &&
			    (*expect != EXPECT_SPECIFIER_VALUE) &&
			    (*expect != EXPECT_THEN_ACTION) &&
			    (*expect != EXPECT_ARG_NAME) &&
			    (*expect != EXPECT_SET_DATA) )
			{
				/* We didn't get a specifier name, tell them */
				parseError(start_expect, "");
				retval = 0;
			}
			break;
		default:
			assert(0);
			break;
		}
	}
	return( retval );
}

/*
 * Process a possible argument name.  This can either be a `then' which means
 * they want to perform an action, or there could be an equals sign buried
 * in the name so we'll need to split the name into its real name and value
 * and add it to them message.
 */
static
int splitArgName(jsiMessage *jm, char *arg_name, expect_state_t *expect)
{
	int retval = 1;
	char *eq;

	require(jm != 0);
	require(arg_name != 0);
	require(expect != 0);
	require(*expect >= 0);
	require(*expect < EXPECT_MAX);
	
	if( !strcasecmp(arg_name, "then") )
	{
		/* Switch to processing actions */
		*expect = EXPECT_THEN_ACTION;
	}
	else if( (eq = strchr(arg_name, '=')) )
	{
		/*
		 * The argument is of the form `name=value' or `name= value'.
		 */
		*eq = 0;
		if( *(eq + 1) )
		{
			/* The value is part of this argument */
			retval = jsiMessageAdd(jm,
					       JMA_String, arg_name, eq + 1,
					       JMA_TAG_DONE);
			*expect = EXPECT_ARG_NAME;
		}
		else
		{
			/*
			 * There's a space between the equals and the value,
			 * so we just null out the equal and wait for the
			 * argument value.
			 */
			*expect = EXPECT_ARG_VALUE;
		}
	}
	else
	{
		/* Just a plain name */
		*expect = EXPECT_ARG_VALUE;
	}
	return( retval );
}

/* Construct a jsiMessage from the command line arguments */
static
int buildMessage(heyMsgAction *hma, int argc, char *argv[])
{
	int error = 0, what, retval = 0;
	jsiMessage *jm = 0;

	require(hma != 0);

	if( argc < 1 )
	{
		printError("No verb specified");
	}
	else if( !(what = mapStringToInt(verbMappings, argv[0])) )
	{
		/* Be nice and tell them what the valid verbs are */
		printError("`%s' isn't a valid verb, use one of the "
			   "following:\n"
			   "\tgetsuites - Get suites a property supports\n"
			   "\tlist - List the children of a property\n"
			   "\tget - Get the value of a property\n"
			   "\tset - Set the value of a property\n"
			   "\tcreate - Create a property\n"
			   "\tdelete - Delete a property\n"
			   "\texecute - Execute a property\n"
			   "\tshutdown - Shutdown the destination\n",
			   argv[0]);
	}
	else if( (jm = jsiCreateMessage()) )
	{
		expect_state_t expect = EXPECT_SPECIFIER_NAME;
		char *arg_name = 0, *arg_value;
		int lpc;

		hma->msg = jm;
		switch( what )
		{
		case JSI_SHUTDOWN:
			hma->action = ACTION_NONE;
			break;
		case JSI_GET_SUITES:
			hma->action = ACTION_PRINT_SUITES;
			arg_name = "";
			expect = EXPECT_OF;
			break;
		default:
			hma->action = ACTION_DEFAULT;
			break;
		}
		jm->what = what;
		/* Parse the message arguments */
		for( lpc = 1;
		     (lpc < argc) && (expect != EXPECT_NONE) && !error;
		     lpc++ )
		{
			switch( expect )
			{
			case EXPECT_SPECIFIER_NAME:
				/* Expecting a specifier name in this arg... */
				arg_name = argv[lpc];
				error = !handlePreposition(jm, &expect,
							   arg_name, 0);
				break;
			case EXPECT_SPECIFIER_VALUE:
				/* Expecting a specifier value in this arg.. */
				arg_value = argv[lpc];
				error = !handlePreposition(jm, &expect,
							   arg_name,
							   arg_value);
				break;
			case EXPECT_OF:
				if( !strcmp(argv[lpc], "of") )
				{
					expect = EXPECT_SPECIFIER_NAME;
				}
				else
				{
					parseError(EXPECT_OF, "");
					error = 1;
				}
				break;
			case EXPECT_PREPOSITION:
				arg_name = argv[lpc];
				error = !handlePreposition(jm, &expect,
							   arg_name, 0);
				break;
			case EXPECT_SET_DATA:
				/* Expecting the value of the `set' */
				error = !jsiMessageAdd(
					jm,
					JMA_String, "data", argv[lpc],
					JMA_TAG_DONE);
				expect = EXPECT_NONE;
				break;
			case EXPECT_ARG_NAME:
				/* Expecting the name of an argument */
				arg_name = argv[lpc];
				if( !splitArgName(jm, arg_name, &expect) )
					error = 1;
				break;
			case EXPECT_ARG_VALUE:
				/* Expecting the value of the argument */
				if( !strcmp(argv[lpc], "=") )
				{
					/* Just ignore an equals sign */
				}
				else
				{
					arg_value = argv[lpc];
					error = !jsiMessageAdd(
						jm,
						JMA_String,
						arg_name, arg_value,
						JMA_TAG_DONE);
					expect = EXPECT_ARG_NAME;
				}
				break;
			case EXPECT_THEN_ACTION:
				if( (hma->action =
				     mapStringToInt(actionMappings,
						    argv[lpc])) )
				{
					/* Check for anymore args */
					if( (lpc + 1) < argc )
						hma->actionValues =
							argv + lpc + 1;
					else
						hma->actionValues = 0;
					hma->actionValueCount = argc -
						(lpc + 1);
					expect = EXPECT_NONE;
				}
				else
				{
					printError("`%s' is not a valid "
						   "action", argv[lpc]);
					error = 1;
				}
				break;
			default:
				retcode = RETCODE_INTERNAL;
				printError("Internal error, sorry...");
				error = 1;
				break;
			}
		}
		if( !error )
		{
			/*
			 * Any verbs that only need a specifier might need
			 * this extra push since there's no other args to
			 * process.
			 */
			if( arg_name &&
			    (arg_name[0] != '\0') &&
			    (expect == EXPECT_SPECIFIER_VALUE) )
			{
				error = !jsiPushMessageSpecifier(
					jm,
					JMA_String, arg_name, "");
			}
			/* Do some extra checks */
			switch( what )
			{
			case JSI_SHUTDOWN:
			case JSI_GET_SUITES:
			case JSI_LIST_PROPERTIES:
			case JSI_GET_PROPERTY:
				if( (expect == EXPECT_SPECIFIER_NAME) ||
				    (expect == EXPECT_SPECIFIER_VALUE) ||
				    (expect == EXPECT_PREPOSITION) ||
				    (expect == EXPECT_NONE) ||
				    (expect == EXPECT_OF) )
				{
				}
				else
				{
					/* Too much */
					parseError(expect,
						   "Too many arguments for %s",
						   mapIntToString(verbMappings,
								  what));
					error = 1;
				}
				break;
			case JSI_SET_PROPERTY:
				if( (expect == EXPECT_NONE) && (lpc < argc) )
				{
					/* Too many */
					parseError(expect,
						   "Too many arguments for a"
						   " set");
					error = 1;
				}
				else if( expect != EXPECT_NONE )
				{
					/* Not enough */
					parseError(expect,
						   "No value to set");
					error = 1;
				}
				break;
			case JSI_EXECUTE_PROPERTY:
			case JSI_CREATE_PROPERTY:
				if( (expect != EXPECT_SPECIFIER_NAME) &&
				    (expect != EXPECT_SPECIFIER_VALUE) &&
				    (expect != EXPECT_PREPOSITION) &&
				    (expect != EXPECT_ARG_NAME) &&
				    (expect != EXPECT_NONE) )
				{
					/* Missing something */
					parseError(expect,
						   "Missing argument value");
					error = 1;
				}
				break;
			}
		}
		if( error )
		{
			/* Cleanup */
			jsiDeleteMessage(jm);
			retval = 0;
		}
		else
		{
			retval = 1;
		}
	}
	else
	{
		printError("No memory");
	}
	return( retval );
}

/* Send the message and return the channel to receive the reply on. */
static
jsiChannel *sendHeyMessage(heyMsgAction *hma, int argc, char **argv)
{
	jsiChannel *retval = 0;

	require(hma != 0);

	if( buildMessage(hma, argc - 2, &argv[2]) )
	{
		int msgSize;
		char *buf;

		if( !strcasecmp(argv[1], "selftest") )
		{
			/* For testing purposes only */
			doAction(hma);
		}
		else if( (buf = jsiFlattenMessage(hma->msg, &msgSize)) )
		{
			if( (retval = getMessageDestination(argv[1])) )
			{
				if( !jsiSendMessage(retval, buf, msgSize) )
				{
					jsiDeleteChannel(retval);
					retval = 0;
					retcode = RETCODE_COMM_ERROR;
					printError("Couldn't send "
						   "the message");
				}
			}
		}
		else
		{
			printError("No memory to flatten the message");
		}
		jsiDeleteMessage(hma->msg);
	}
	return( retval );
}

/* Receive and print the reply message */
static
void printHeyReply(heyMsgAction *hma, jsiChannel *recv)
{
	jsiHandlerContext hc = { &heyHandlerList, recv };
	int replySize, done = 0;
	char *replyBuf = 0;

	require(hma != 0);
	require(recv != 0);

	retcode = RETCODE_COMM_ERROR;
	while( !done && (replyBuf = jsiRecvMessage(recv, &replySize)) )
	{
		jsiMessage *rmsg;
		
		if( (rmsg = jsiInstantiateMessage(replyBuf, replySize)) )
		{
			switch( rmsg->what )
			{
			case JSI_ERROR:
			case JSI_REPLY:
				/* Act on the reply message */
				hma->msg = rmsg;
				if( doAction(hma) )
				{
					if( rmsg->what == JSI_ERROR )
						retcode = RETCODE_MSG_ERROR;
					else
						retcode = RETCODE_OK;
				}
				else
				{
					retcode = RETCODE_USAGE_ERROR;
				}
				done = 1;
				break;
			default:
				/* Its a regular message, try to handle it. */
				jsiDispatchMessage(&hc, rmsg);
				if( (replyBuf =
				     jsiFlattenMessage(rmsg, &replySize)) )
				{
					jsiReplyMessage(recv,
							replyBuf,
							replySize);
				}
				break;
			}
			jsiDeleteMessage(rmsg);
		}
		else
		{
			printError("Couldn't instantiate message");
			done = 1;
		}
	}
	if( !done && (replyBuf == 0) )
	{
		printError("Couldn't receive reply message");
	}
}

int main(int argc, char *argv[])
{
	jsiPlexLink *pl;

	pl = jsiCreatePlexUDP();
	jsiAddPlexLink(pl);
	progName = argv[0];
	installHeyHandlers();
	if( argc >= 3 )
	{
		heyMsgAction hma;
		jsiChannel *dest;
		
		if( (dest = sendHeyMessage(&hma, argc, argv)) &&
		    (hma.action != ACTION_NONE) )
		{
			printHeyReply(&hma, dest);
		}
		else if( hma.action == ACTION_NONE )
		{
			retcode = RETCODE_OK;
		}
		jsiDeleteChannel(dest);
	}
	else
	{
		printUsage();
	}
	return( retcode );
}

void pp_failed(char *file, unsigned int line, 
	       char* func, char *type, char *expr)
{
#if 0
	/* A more assert-like failure message */
     	fprintf(stderr, "%s:%u:%s(): %s failed: %s\n",
	        file, line, func, type, expr);
#else
	/* A more verbose message */
	fprintf(stderr, "%s %s failed in\n"
		"  file: %s at line: %d in function: %s\n",
		type, expr, file, line, func);
#endif /* 1 */
	abort();
}
