/*
 * Copyright (c) 1998, 1999, 2000, 2001
 *     The University of Utah. All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * Contributed by the Flux Research Group at the University of Utah.
 * Authors: Godmar Back, Patrick Tullmann, Edouard G. Parmelan
 */
/*
 * debug.c
 *
 * A dynamic debugging framework for Kaffe.  Through the magic
 * of hideous macros, we can control the debugging output of
 * Kaffe in a couple of ways.  First, it can be completely
 * disabled in which case all of the code "just goes away",
 * or, if its enabled, the area of code to debug is dynamically
 * chosen at run time.
 */

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "jtypes.h"
#include "gtypes.h"
#include "gc.h"
#include "debug.h"
#include "module.h"
#include "assert_pp.h"
#include <ctype.h>
#include <stdarg.h>

/* Default debugging mask to use (if debug is enabled) */
#define DEFAULT_DEBUG_MASK	DBG_NONE

/*
 * XXX all this "debugging printf" stuff is being used outside of
 * the debugging support.  Move it elsewhere.
 */
static int debugSysInit(int ignored);

static char *debugBuffer = NULL;
static volatile int bufferBegin = 0;
static int bufferSz = 1024;
static int bufferOutput = 0;

#if defined(NDEBUG) || !defined(DEBUG)
/* --- Debugging is NOT enabled --- */

/* Don't waste space with the debugging functions */

void dbgSetMask(debugmask_t m) { }
int dbgSetMaskStr(char *s) { return 0; }
void dPrintMemory(void *mem, int size) { }

#else /* Actually define the functions */
/* --- Debugging is enabled --- */

static int debugSysInit(int ignored);
debugmask_t kaffevmDebugMask = DEFAULT_DEBUG_MASK;

void
dbgSetMask(debugmask_t mask)
{
	kaffevmDebugMask = mask;
}

/*
 * Associate strings with an option or set of
 * options.
 */
static struct debug_opts
{
	char *name;
	debugmask_t mask;
	char *desc;
} debug_opts[] =
{
	/* XXX Clean these names up, re-order them and make them consistent. */
#define D(name, str) { #name, DBG_ ## name, str }
	D(NONE, "Nothing"),
	D(SLOWLOCKS,  "Heavy lock/unlock operations"),
	D(VMCONDS,  "Show condition variable operations called by VM"),
	D(NEWINSTR, "Show softnew_ instructions (NEW, NEWARRAY, ANEWARRAY)"),
	D(VMTHREAD, "Show some java.lang.Thread operations called by VM"),
	D(JTHREAD,  "Show jthread operations--jthreads only"),
	D(JTHREADDETAIL,  "Show jthread operations (more detail)"),
	D(JTHREADNOPREEMPT,  "Disable preemption in--jthreads only"),
	D(EXCEPTION, "Debug exceptions, don't catch traps"),
	D(STACKTRACE, "Debug stack trace inspection"),
	D(INIT,     "Show initialization steps."),
	D(INITCLASSPATH,     "Show classpath initialization."),
	D(BREAKONEXIT, "Cause an exception before exiting. (Useful under GDB)"),
	D(GCPRIM,   "Debug gc_primitive_*"),
	D(GCALLOC,   "Debug gc_heap_alloc*"),
	D(GCFREE,   "Debug gc_heap_free*"),
	D(GCSYSALLOC,   "Show allocations of system memory"),
	D(GCWALK,   "Show gc walking"),
	D(GCPRECISE, "Debug precise collection of the heap."),
	D(GCSTAT,   "count instances of class and array types"),
	D(GCDIAG,   "Perform diagnostic checks on heap" ),
	D(GCHEAPS,  "Debug multi-heap collection" ),
	D(GCREFS,   "Show when global refs are born and die"),
	D(GCFIN,    "trace finalizable object lifecycle"),
	{ "GCMEM", DBG_GCPRIM|DBG_GCALLOC|DBG_GCFREE|DBG_GCSYSALLOC|DBG_GCSTAT,
			"All allocation and free operations in gc-mem" },
	D(SLACKANAL,   "Print internal fragmentation statistics."),

	D(ASYNCSTDIO, "Make stdio fds asynchronous despite debugging."),
	D(CATCHOUTOFMEM, "Catch recursive out of memory exceptions."),
	D(JARFILES, "Debug reading JAR files in jar.c."),
	D(CODEATTR, "Show code attributes during class file parsing."),
	D(INT_INSTR, "Show instructions. (interpreter)"),
	D(INT_NATIVE, "Show call to native methods. (interpreter)"),
	D(INT_RETURN, "Show return from function. (interpreter)"),
	D(INT_VMCALL, "Show call to virtualMachine. (interpreter)"),
	D(INT_CHECKS, "Show various checks. (interpreter)"),
	D(ELOOKUP, "Debug exception lookup."),
	D(FLOOKUP, "Debug field lookup."),
	D(MLOOKUP, "Debug method lookup."),
	D(CLASSLOOKUP, "Debug class lookup."),
	D(JIT, 	"Debug JIT compiler--show emitted instructions."),
	D(MOREJIT, 	"Debug JIT compiler--show callinfo."),
	D(NOGC,	"Turn garbage collection off."),
	D(STATICINIT,	"Announce when static initializers are called."),
	D(CLASSFILE,	"Show when methods and fields are added."),
	D(RESERROR,	"Show error in class file resolution."),
	D(VMCLASSLOADER,"Announce when VM calls class loaders.."),
	D(LOCKCONTENTION,"Show when a lock is contended."),
	D(CODEANALYSE, "Debug code analysis/verification."),
	D(CLASSGC, 	"Debug class garbage collection."),
	D(NEWOBJECT, 	"Show when new objects are allocated."),
	D(FINALIZE, 	"Debug finalization."),
	D(NATIVELIB, 	"Debug native library operations."),
	D(NATIVENET,	"Show network accesses from the native code."),
	D(SHOWEXCEPTIONS, "Show execptions that are generated"),
	D(INVOKE,	"Show exceptions in Method.invoke, and at top level"),
	D(MUNGWALL,	"Attempt to detect buffer underflow/overflow"),
	
	D(AWT_MEM,	"AWT subsystem."),
	D(AWT_CLR,	"AWT subsystem."),
	D(AWT_EVT,	"AWT subsystem."),
	D(AWT_IMG,	"AWT subsystem."),
	D(AWT_WND,	"AWT subsystem."),
	D(AWT_GRA,	"AWT subsystem."),
	D(AWT_FNT,	"AWT subsystem."),
	D(AWT,		"Debug all of AWT."),

	D(GCJ,		"Debug GCJ support."),
	D(GCJMORE,	"Debug GCJ support (additional msg)."),

	D(REGFORCE,	"Debug forced registers framework."),

	/* you can define combinations too */
	{ "lookup", DBG_MLOOKUP|DBG_ELOOKUP|DBG_FLOOKUP,
			"Various lookup operations" },

	{ "gcj", DBG_GCJ|DBG_GCJMORE,
			"Debug GCJ support in detailed" },
	{ "thread", DBG_JTHREAD|DBG_SLOWLOCKS|DBG_VMCONDS,
			"Thread operations and locking operations" },

	{ "intrp", DBG_INT_NATIVE|DBG_INT_RETURN|DBG_INT_VMCALL,
			"Calls of interpreter (without instructions)" },
	{ "deadlock", DBG_LOCKCONTENTION, "Debug deadlocks" },
	{ "classgc", DBG_CLASSGC|DBG_FINALIZE,
			"Debug finalization (show finalization)" },
	{ "intrpA",
		DBG_INT_CHECKS|DBG_INT_INSTR|DBG_INT_NATIVE|
		DBG_INT_RETURN|DBG_INT_VMCALL,
			"Complete interpreter trace" },

	/* special options */
	{ "buffer", 0, "Log output to an internal buffer instead of stderr" },
	{ "dump", 0, "Print the buffer log when the program exits" },
	/* These must be last, as they are markers for some GDB macros. */
	D(ALL, "Everything under the sun...."),
	D(ANY, "Ibid.")
};

static void debugToBuffer(int size);
void printDebugBuffer(void);

#define NELEMS(a)	(sizeof (a) / sizeof(a[0]))

#if !defined(NO_POSTCONDITIONS) || !defined(NO_PRECONDITIONS)
/* 
 * The routine which prints failure messages.  
 *
 * Macros that uses this are defined in include/assert_pp.h
 */
void 
pp_failed(char *file, unsigned int line, 
	  char* func, char *type, char *expr)
{
#if 1
	/* A more assert-like failure message */
     	kaffe_dprintf("%s:%u:%s(): %s failed: %s\n",
		      file, line, func, type, expr);
#else
	/* A more verbose message */
	kaffe_dprintf("%s %s failed in\n"
		      "  file: %s at line: %d in function: %s\n",
		      type, expr, file, line, func);
#endif /* 1 */
	abort();
}
#endif /* !defined(NO_POSTCONDITIONS) || !defined(NO_PRECONDITIONS) */

void dPrintMemory(void *mem, int size)
{
	char *ptr = (char *)mem;
	int lpc, pos, temp_pos;
	char the_char;
	
	require( mem != 0 );
	require( size >= 0 );
	
	kaffe_dprintf("Printing memory at %p, size %d:\n", mem, size );
	for( pos = 0; pos < size; )
	{
		kaffe_dprintf("%p:", &ptr[pos]);
		for( lpc = 0; lpc < 16; lpc++, pos++ )
		{
			if( (lpc % 4) == 0 )
				kaffe_dprintf(" ");
			if( pos < size )
				kaffe_dprintf("%02x", ptr[pos] & 0xff);
			else
				kaffe_dprintf("  ");
		}
		kaffe_dprintf(" ");
		for( lpc = 0, temp_pos = (pos - 16);
		     lpc < 16;
		     lpc++, temp_pos++ )
		{
			if( (temp_pos < size) && isprint(ptr[temp_pos]) )
				the_char = ptr[temp_pos];
			else
				the_char = '.';
			kaffe_dprintf("%c", the_char);
		}
		kaffe_dprintf("\n");
	}
}

int
dbgSetMaskStr(char *mask_str)
{
	int i;
	char *separators = "|,";
	char *opt;

	opt = strtok(mask_str, separators);

	if (!opt) {
		kaffevmDebugMask = DEFAULT_DEBUG_MASK;
		return 1;
	}

	/* Need to init buffer for kaffe_dprintf... */
	debugSysInit(0);

	/* Special target 'list' lists all the defined options */
	if (!strcasecmp(opt, "list")) {
		kaffe_dprintf("Available debug opts: \n");
		kaffe_dprintf("  %-15s\t%16s  %s\n", "Option", "Mask", "Description");
		for (i = 0; i < NELEMS(debug_opts); i++)
			if (debug_opts[i].mask>>32)
			{
				kaffe_dprintf("  %-15s\t%8X%08X  %s\n", 
					      debug_opts[i].name,
					      (int)(debug_opts[i].mask>>32), 
					      (int)(debug_opts[i].mask),
					      debug_opts[i].desc);
			}
			else
			{
				kaffe_dprintf("  %-15s\t        %8X  %s\n", 
					      debug_opts[i].name,
					      (int)(debug_opts[i].mask), 
					      debug_opts[i].desc);
			}
		return 0;
	}


	while (opt) {
		if (!strcasecmp(opt, "buffer"))
			debugToBuffer(64 * 1024);
		else if (!strcasecmp(opt, "dump"))
			atexit(printDebugBuffer);
		else
		{
			int set;

			switch(opt[0]) {
			case '-':
				opt++;
				set = 0;
				break;

			case '+':
				opt++;
				/* FALLTHRU */
			default:
				set = 1;
			}

			for (i = 0; i < NELEMS(debug_opts); i++)
				if (!strcasecmp(opt, debug_opts[i].name))
				{
					if (set)
						kaffevmDebugMask |= debug_opts[i].mask;
					else
						kaffevmDebugMask &= ~debug_opts[i].mask;
					break;
				}
		
			/* Allow bit shifts. */
			if (i == (sizeof debug_opts)/(sizeof(debug_opts[0]))) {
				char *endp;
				int bit;

				bit = strtoul(opt, &endp, 0);
				/* Be polite. */
				if (*endp != '\0') {
					dprintf("Unknown flag (%s) passed to -vmdebug\n",
						opt);
					return 0;
				}
				if (set)
					kaffevmDebugMask |= DBG_BIT(bit);
				else
					kaffevmDebugMask &= ~DBG_BIT(bit);
			}
		}

		/* Get next opt */
		opt = strtok(NULL, separators);
	}

	if (kaffevmDebugMask & DBG_JIT) {
#if defined(TRANSLATOR) && defined(DEBUG)
		extern int jit_debug;
		jit_debug = 1;
#else
		dprintf(
			"You cannot debug the JIT in interpreter mode \n");
#endif
	}
	return 1;
}

/*
 * create a buffer in which debugging output is written
 */
static void
debugToBuffer(int size)
{
	bufferSz = size;
	debugBuffer = malloc(size);
	bufferOutput = 1;
	assert(debugBuffer != NULL);
}
#endif

/*
 * The following function is invoked at exit time.
 * It deliberately caused a trap, to give control to gdb.
 */
static void
debugExitHook(void)
{
	/*
	 * this is a hook for catching exits from GDB.
	 * make this dependent on the selection of this option
	 */
	DBG(BREAKONEXIT, DBGGDBBREAK())
}

module_t debugModule = {
	"debug",
	0,
	debugSysInit,
	0,
	0
};

/*
 * initialize debugging system
 *
 * Note that this may be invoked before the module init
 * system gets around to it.  (E.g., if debugging the
 * initialization sequence.)
 */
static int
debugSysInit(int ignored)
{
	int retval = 0;
	
#if defined(TRANSLATOR) && defined(DEBUG)
	{
		extern int jit_debug;
		if (getenv("JIT_DEBUG"))
			jit_debug = 1;
	}
#endif
	
	/*
	 * If buffer has already been allocated, then this
	 * function has already been run, just return success.
	 */
	if (debugBuffer != NULL)
		return 1;


	debugBuffer = malloc(bufferSz);

	if (debugBuffer != NULL);
	{
		atexit(debugExitHook);
		retval = 1;
	}

	return retval;
}

/*
 * When debugging, printf should use fprintf() to avoid
 * threading/blocking problems.
 */
void
printDebugBuffer(void)
{
	int i = 0;
	int end = bufferBegin;

	assert(debugBuffer);

	i = bufferBegin;
	assert(i != 0);

	while(i < bufferSz)
		putc(debugBuffer[i++], stdout);
	i = 0;
	while(i < end)
		putc(debugBuffer[i++], stdout);
}

int
kaffe_dprintf(const char *fmt, ...)
{
	int lastBegin;
	int n;
	int max;
	va_list args;

	assert(debugBuffer);

	va_start(args, fmt);

	lastBegin = bufferBegin;
#ifdef HAVE_VSNPRINTF
	max = bufferSz - bufferBegin - 1;
	n = vsnprintf(debugBuffer + bufferBegin, max, fmt, args);

	/* The return value is bytes *needed* not bytes *used* */
	if (n > max)
		n = max;
#else
	n = vsprintf(debugBuffer + bufferBegin, fmt, args);
#endif
	bufferBegin += n;
	assert(bufferBegin < bufferSz);/* XXX */

	if (bufferOutput) {
		if (bufferBegin >= (bufferSz - 60))
			bufferBegin = 0;
	} else if( lastBegin == 0 ) {
		/* Keep trying to write.  Should we pause(), or
		 * sigsuspend(), or do something based on config
		 * defintions?
		 */
		max = 0;
		while (max < bufferBegin) {
			int w =  write(2,
				       debugBuffer + max,
				       bufferBegin - max);
			if (w >= 0)
				/* ignore errors */
				max += w;
		}
		bufferBegin = 0;
	}
	va_end(args);

	return n;
}
