/*
 * external.c
 * Handle method calls to other languages.
 *
 * Copyright (c) 1996, 1997, 1998, 1999
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "assert_pp.h"
#include "debug.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"
#include "gtypes.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "classMethod.h"
#include "slots.h"
#include "external.h"
#include "errors.h"
#include "exception.h"
#include "slib.h"
#include "system.h"
#include "support.h"
#include "md.h"
#include "system.h"
#include "jthread.h"
#include "jsignal.h"
#include "stats.h"
#define LT_NON_POSIX_NAMESPACE
#include "ltdl.h"
#include "module.h"
#include <janosvm/nameSpace.h>
#include <janosvm/jteam.h>
#include <janosvm/stdioResource.h>
#include <misc/feedback.h>

/** Module declarations **/
static int initExternalModule(int stage);

module_t externalModule = {
	"external",
	0,
	initExternalModule,
	0,
	0
};

#ifndef STUB_PREFIX
#define STUB_PREFIX ""
#define STUB_PREFIX_LEN 0
#endif

#ifndef STUB_POSTFIX
#define STUB_POSTFIX ""
#endif

#ifndef LIBRARYHANDLE
#define LIBRARYHANDLE lt_dlhandle
#endif

#ifndef LIBRARYINIT
static inline lt_ptr_t kdlmalloc(size_t len) {
	void *ptr = gc_malloc(len, GC_ALLOC_EXTERNAL);
	addToCounter(&ltmem, "vmmem-libltdl", 1, GCSIZEOF(ptr));
	return (ptr);
}
static inline void kdlfree(lt_ptr_t ptr) { 
	addToCounter(&ltmem, "vmmem-libltdl", 1, -((jlong)GCSIZEOF(ptr)));
	gc_free(ptr); 
}
#define LIBRARYINIT() ((lt_dlmalloc=kdlmalloc),(lt_dlfree=kdlfree),lt_dlinit())
#endif

#ifndef LIBRARYPREFIX
#define LIBRARYPREFIX "lib"
#endif

#ifndef LIBRARYSUFFIX
#define LIBRARYSUFFIX ""
#endif

#ifndef LIBRARYLOAD
#define LIBRARYLOAD(desc,filename)	((desc)=lt_dlopenext((filename)))
#endif

#ifndef LIBRARYUNLOAD
#define LIBRARYUNLOAD(desc)		(lt_dlclose(desc))
#endif

#ifndef LIBRARYERROR
#define LIBRARYERROR() lt_dlerror()
#endif

static struct _libHandle {
	LIBRARYHANDLE	desc;
	char*		name;
	int		ref;
} libHandle[MAXLIBS];

#ifndef LIBRARYFUNCTION
static inline lt_ptr_t findLibraryFunction(const char *name) {
  int i = 0;
  lt_ptr_t ptr = 0;

  while (!ptr && libHandle[i].ref && i < MAXLIBS) {
    ptr = lt_dlsym(libHandle[i].desc, name);
    ++i;
  }

  return ptr;
}

#define LIBRARYFUNCTION(ptr,name) ((ptr)=findLibraryFunction(name))
#endif

static char *libraryPath;
static int maxLibraryPathLen;

extern jint Kaffe_JNI_native(Method*);

/*
 * Error stub function.  Point unresolved link errors here to avoid
 * problems.
 */
static
void*
error_stub(void)
{
	return (0);
}

static
int initExternalModule(int stage)
{
	int len = 0, retval = 0;
	char* lpath;

	/* Get a library path from somewhere */
	lpath = (char*)Kaffe_JavaVMArgs[0].libraryhome;
	if (lpath == 0) {
		lpath = getenv(LIBRARYPATH);
	}

	if (lpath != 0) {
		len = strlen(lpath);
	}

	/* Build a library path from the given library path. */
	if( (libraryPath = gc_malloc(len + 1, GC_ALLOC_EXTERNAL)) )
	{
		int foundNative = 0, foundKaffeVM = 0, foundKaffeMisc = 0;
		
		maxLibraryPathLen = len;
		addToCounter(&ltmem,
			     "vmmem-libltdl", 1, GCSIZEOF(libraryPath));
		if (lpath != 0) {
			strcat(libraryPath, lpath);
		}

		LIBRARYINIT();

		/* Load the standard native library */
		if( findLibrary(NATIVELIBRARY) >= 0 )
			foundNative = 1;
		/*
		 * JanosVM: libkaffevm contains some native symbols needed by
		 * java code so we just load ourselves to get the symbols into
		 * ltdl.  This is pretty silly.
		 */
		if( findLibrary(VMLIBRARY) >= 0 )
			foundKaffeVM = 1;
		if( findLibrary(MISCLIBRARY) >= 0 )
			foundKaffeMisc = 1;
		
		if( foundNative && foundKaffeVM )
		{
			/* Found our libraries, we're good to go */
			retval = 1;
		}
		else
		{
			if( !foundNative )
			{
				fprintf(stderr,
					"Failed to locate %s library\n",
					NATIVELIBRARY);
			}
			if( !foundKaffeVM )
			{
				fprintf(stderr,
					"Failed to locate %s library\n",
					VMLIBRARY);
			}
			if( !foundKaffeMisc )
			{
				fprintf(stderr,
					"Failed to locate %s library\n",
					MISCLIBRARY);
			}
			fprintf(stderr,	"Library path is %s\n", libraryPath);
			fprintf(stderr, "Aborting.\n");
			fflush(stderr);
		}
	}
	else
	{
		fprintf(stderr,	"Not enough memory\n");
		fflush(stderr);
	}
	return( retval );
}

int
loadNativeLibrary(char* lib, char *errbuf, size_t errsiz)
{
	int retval = -1;

	require(lib != 0);

	/* The kernel manages libraries */
	if( switchToTeam(kernelTeam, 0) )
	{
		retval = loadNativeLibrary2(lib, 1, errbuf, errsiz);
		returnFromTeam();
	}
	else if( errbuf )
	{
		strncpy(errbuf, "Can't visit kernel\n", errsiz);
		errbuf[errsiz - 1] = '\0';
	}

	ensure((retval == -1) || (retval >= 0));
	
	return( retval );
}

/* Get the base name of the library */
static
char *getLibraryBaseName(char *path)
{
	char *pos, *retval = 0;
	int sep_len;

	require(path != 0);
	
	/* Strip off the path to the library */
	sep_len = strlen(file_separator);
	for( pos = path + strlen(path);
	     (pos >= path) && (strncmp(pos, file_separator, sep_len) != 0);
	     pos-- );
	/* Strip off the library prefix */
	if( pos >= path )
		retval = pos + 1 + strlen(LIBRARYPREFIX);

	ensure((retval == 0) || strstr(path, retval));
	
	return( retval );
}

/*
 * Link in a native library. If successful, returns an index >= 0 that
 * can be passed to unloadNativeLibrary(). Otherwise, returns -1 and
 * fills errbuf (if not NULL) with the error message. Assumes synchronization.
 */
int
loadNativeLibrary2(char* path, int default_refs, char *errbuf, size_t errsiz)
{
	struct _libHandle *lib;
	int index;

	require(path != 0);
	require(default_refs >= 0);

	/* Find a library handle.  If we find the library has already
	 * been loaded, don't bother to get it again, just increase the
	 * reference count.
	 */
	for (index = 0; index < MAXLIBS; index++) {
		lib = &libHandle[index];
		if (lib->desc == 0) {
			goto open;
		}
		if (strcmp(lib->name, path) == 0) {
			lib->ref++;
DBG(NATIVELIB,
			dprintf("Native lib %s\n"
			    "\tLOAD desc=%p index=%d ++ref=%d\n",
			    lib->name, lib->desc, index, lib->ref);
    )
			return index;
		}
	}
	if (errbuf != 0) {
		strncpy(errbuf, "Too many open libraries", errsiz);
		errbuf[errsiz - 1] = '\0';
	}
	return -1;

	/* Open the library */
	open:

#if 0
	/* If this file doesn't exist, ignore it */
	if (access(path, R_OK) != 0) {
		if (errbuf != 0) {
			strncpy(errbuf, SYS_ERROR(errno), errsiz);
			errbuf[errsiz - 1] = '\0';
		}
		return -1;
	}
#endif
/* if we tested for existence here, libltdl wouldn't be able to look
   for system-dependent library names */

DBG(NATIVELIB,
    	dprintf("Try to load %s...\n", path);
    )

	blockAsyncSignals();
	LIBRARYLOAD(lib->desc, path);
	unblockAsyncSignals();

	if (lib->desc == 0) {
		const char *err = LIBRARYERROR();

		if (err == 0) {
			err = "Unknown error";
		}
		if (errbuf != 0) {
			strncpy(errbuf, err, errsiz);
			errbuf[errsiz - 1] = '\0';
		}
		return -1;
	}

	lib->ref = default_refs;
	lib->name = gc_malloc(strlen(path) + 1, GC_ALLOC_EXTERNAL);
	addToCounter(&ltmem, "vmmem-libltdl", 1, GCSIZEOF(lib->name));
	strcpy(lib->name, path);

DBG(NATIVELIB,
	dprintf("Native lib %s\n"
	    "\tLOAD desc=%p index=%d ++ref=%d\n",
	    lib->name, lib->desc, index, lib->ref);
    )
#if defined(KAFFE_FEEDBACK)
	feedbackLibrary(path, true);
#endif

	/* Execute any module initializers */
	{
		char *primName;

		/*
		 * The moduleInitList variables will clash in a statically
		 * linked environment so we must use a unique name for each.
		 * So the convention is use the base name of the library to
		 * prefix the variables.
		 */
		if( (primName = getLibraryBaseName(path)) )
		{
			char miName[MAXLIBPATH]; // XXX
			moduleInitializer_t *mi;
			
			sprintf(miName, "%sModuleInitList", primName);
			if( (mi = loadNativeLibrarySym(miName)) )
			{
				if( !initModules(mi) )
				{
					fprintf(stderr,
						"Library initializer(s) "
						"failed for %s.\n",
						path);
					unloadNativeLibrary(index);
					return -1;
				}
			}
		}
	}

	return index;
}

/*
 * Unlink a native library. Assumes synchronization.
 * Note that libnative is always at index zero and should
 * never be unloaded. So index should never equal zero here.
 */
void
unloadNativeLibrary(int index)
{
	struct _libHandle *lib;

	require(index >= 0);
	require(index < MAXLIBS);
	
	/* XXX We should manage libs on a per team basis */
	if( !switchToTeam(kernelTeam, 0) )
	{
		fprintf(stderr, "Unable to visit kernel to unload library!");
		return;
	}
	{
		lib = &libHandle[index];
		
		DBG(NATIVELIB,
		    dprintf("Native lib %s\n"
			    "\tUNLOAD desc=%p index=%d --ref=%d\n",
			    lib->name, lib->desc, index, lib->ref - 1);
		    )

		ensure(lib->desc != 0);
		ensure(lib->ref > 0);
	}
	lib->ref--;
	if (lib->ref == 0) {
		blockAsyncSignals();
		LIBRARYUNLOAD(lib->desc);
		unblockAsyncSignals();
		gc_free(lib->name);
		lib->desc = 0;
	}
	returnFromTeam();
}

/*
 * Get pointer to symbol from symbol name.
 */
void*
loadNativeLibrarySym(char* name)
{
	void* func;

	blockAsyncSignals();
	LIBRARYFUNCTION(func, name);
	unblockAsyncSignals();
	
	return (func);
}

bool
native(Method* m, errorInfo *einfo)
{
	char stub[MAXSTUBLEN];
	const char* s;
	jteam *owner;
	int i;
	void* func;

	owner = gc_get_resource(m)->client.owner;
	if( kernelTeam && (owner != kernelTeam) )
	{
		stdioPrint(STDIO_ERR,
			   "Trying to bind user class to native method\n");
		postExceptionMessage(
			einfo,
			JAVA_LANG(UnsatisfiedLinkError),
			"Failed to locate native function:\t%s.%s%s",
			getClassView(m->class)->name->data,
			m->name->data,
			METHOD_SIGD(m));
		return( false );
	}
	
	/* Construct the stub name */
	strcpy(stub, STUB_PREFIX);
	s = getClassView(m->class)->name->data;
	for (i = STUB_PREFIX_LEN; *s != 0; s++, i++) {
		if (*s == '/') {
			stub[i] = '_';
		}
		else {
			stub[i] = *s;
		}
	}
	stub[i] = '_';
	stub[i+1] = 0;
	strcat(stub, m->name->data);
	strcat(stub, STUB_POSTFIX);

DBG(NATIVELIB,	
	dprintf("Method = %s.%s%s\n", getClassView(m->class)->name->data, 
		m->name->data, METHOD_SIGD(m));
	dprintf("Native stub = '%s'\n", stub);
    )

	/* Find the native method */
	func = loadNativeLibrarySym(stub);
	if (func != 0) {
		/* Fill it in */
		extern void Kaffe_KNI_wrapper(Method* xmeth, void* func);
		Kaffe_KNI_wrapper(m, func);
		return (true);
	}

	/* Try to locate the nature function using the JNI interface */
        if (Kaffe_JNI_native(m)) {
                return (true);
        }

DBG(NATIVELIB,
	dprintf("Failed to locate native function:\n\t%s.%s%s\n",
		getClassView(m->class)->name->data,
		m->name->data, METHOD_SIGD(m));
    )
	SET_METHOD_NATIVECODE(m, (void*)error_stub);

	postExceptionMessage(einfo, JAVA_LANG(UnsatisfiedLinkError),
		"Failed to locate native function:\t%s.%s%s",
		getClassView(m->class)->name->data,
			     m->name->data, METHOD_SIGD(m));
	return (false);
}

/*
 * Return the library path.
 */
char*
getLibraryPath(void)
{
	return (libraryPath);
}

int
findLibrary(char *baseName)
{
	int retval = -1;
	char *libPath;

	/* Allocate some scratch to build path names */
	if( (libPath = (char *)gc_malloc(maxLibraryPathLen +
					 1 +
					 strlen(LIBRARYPREFIX) +
					 strlen(baseName) +
					 strlen(LIBRARYSUFFIX) +
					 1,
					 GC_ALLOC_EXTERNAL)) )
	{
		char *pathStart, *pathEnd;
		int libUsed, sepLen;

		sepLen = strlen(path_separator);
		pathStart = libraryPath;
		/*
		 * Walk over the library paths and try to load the requested
		 * one from them
		 */
		while( (retval < 0) && pathStart )
		{
			libUsed = 0;
			/* Copy the path into libPath */
			if( (pathEnd = strstr(pathStart, path_separator)) )
			{
				libUsed += pathEnd - pathStart;
				strncpy(libPath, pathStart, libUsed);
				pathStart = pathEnd + sepLen;
			}
			else
			{
				strcpy(libPath, pathStart);
				libUsed += strlen(pathStart);
				pathStart = 0;
			}
			/* Append the request lib name */
			sprintf(&libPath[libUsed],
				"/%s%s%s",
				LIBRARYPREFIX,
				baseName,
				LIBRARYSUFFIX);
			/* Try to look it up */
			retval = loadNativeLibrary(libPath, NULL, 0);
		}
		gc_free(libPath);
	}
	return( retval );
}
