/*
 * gc.h
 * The garbage collector.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#ifndef __gc_h
#define __gc_h

#include "gtypes.h"
#include "hashtab.h"

/* 
 * By default, use write barriers; can be overridden by passed defines
 */
#ifndef GC_WRITE_BARRIERS
#define	GC_WRITE_BARRIERS	1
#endif

/* 
 * We do not support incremental collection as this time.
 */
#undef	GC_INCREMENTAL

typedef struct _Collector Collector;
typedef void (*walk_func_t)(struct _Collector*, void*, uint32);
typedef void (*final_func_t)(struct _Collector*, void*);
typedef void (*destroy_func_t)(struct _Collector*, void*);

#define	GC_OBJECT_NORMAL	((final_func_t)0)
#define	GC_OBJECT_FIXED		((final_func_t)1)

/* Use the incremental garbage collector */
#ifndef KAFFEH
#include "mem/gc-incremental.h"
#endif

/*
 * Garbage collector interface.
 */
#define	GC_ALLOC_JAVASTRING	0
#define	GC_ALLOC_NOWALK		1
#define	GC_ALLOC_NORMALOBJECT	2
#define	GC_ALLOC_PRIMARRAY	3
#define	GC_ALLOC_REFARRAY	4
#define	GC_ALLOC_CLASSOBJECT	5
#define	GC_ALLOC_FINALIZEOBJECT	6
#define	GC_ALLOC_BYTECODE	7
#define	GC_ALLOC_EXCEPTIONTABLE	8
#define	GC_ALLOC_JITCODE	9
#define	GC_ALLOC_STATICDATA	10
#define	GC_ALLOC_CONSTANT	11
#define	GC_ALLOC_FIXED		12
#define	GC_ALLOC_DISPATCHTABLE	13
#define	GC_ALLOC_METHOD		14
#define	GC_ALLOC_FIELD		15
#define	GC_ALLOC_UTF8CONST	16
#define	GC_ALLOC_INTERFACE	17
#define	GC_ALLOC_LOCK		18
#define	GC_ALLOC_THREADCTX	19
#define	GC_ALLOC_REF		20
#define	GC_ALLOC_JITTEMP	21
#define	GC_ALLOC_JAVALOADER	22
#define GC_ALLOC_CLASSPATH	23
#define GC_ALLOC_CODE_ANALYSE	24
#define GC_ALLOC_EXCEPTION_DATA	25
#define GC_ALLOC_FILE_TYPE	26
#define GC_ALLOC_INFLATE	27
#define GC_ALLOC_JAR		28
#define GC_ALLOC_TAGS		29
#define GC_ALLOC_METHOD_CACHE	30
#define GC_ALLOC_STRING		31
#define GC_ALLOC_SUPPORT	32
#define GC_ALLOC_UTF_TEMP	33
#define GC_ALLOC_STACK_TRACE	34
#define GC_ALLOC_EXTERNAL	35
#define GC_ALLOC_NAME_SPACE	36
#define GC_ALLOC_RESOURCE	37
#define GC_ALLOC_SOFT		38
#define GC_ALLOC_CLASS_GROUP	39
#define GC_ALLOC_IMPL_TABLE	40
#define GC_ALLOC_SIGNATURE	41
#define GC_ALLOC_JTEAM		42
#define GC_ALLOC_GC_MAP		43
#define GC_ALLOC_EXCEPTION_TEMP	44
#define	GC_ALLOC_SOURCE_FILE	45
#define	GC_ALLOC_IF2ITABLE	46
#define	GC_ALLOC_ITABLE2DTABLE	47
#define	GC_ALLOC_TEMP_STRING	48
#define	GC_ALLOC_CODE		49
#define	GC_ALLOC_THREADSTACK	50
#define	GC_ALLOC_JSI		51
#define GC_ALLOC_TRAMPOLINE	52
#define	GC_ALLOC_MAX_INDEX	53

/*
 * Define a COM-like GC interface.
 */
struct GarbageCollectorInterface_Ops;

struct _Collector {
	struct GarbageCollectorInterface_Ops *ops;
};

struct GarbageCollectorInterface_Ops {

	void*   reserved1;
	void*   reserved2;
	void*   reserved3;
	
	/* Global reference (root) management. */
	bool	(*addRef)(Collector *, const void *);
	bool    (*rmRef)(Collector *, const void *);
	void	(*walkRefs)(Collector *);
	
	void*	(*malloc)(Collector *, size_t size, int type);
	void*	(*realloc)(Collector *, void* addr, size_t size, int type);
	void	(*free)(Collector *, void* addr);

	void	(*invoke)(Collector *, int mustgc);
	void	(*invokeFinalizer)(Collector *);
	void	(*init)(Collector *);
	int	(*enable)(Collector *);

	void	(*markAddress)(Collector *, const void* addr,
			       const void *from);
	void	(*markObject)(Collector *, const void* obj,
			      const void *from);
	void	(*finalizeObject)(Collector *, void *obj);
	uint32	(*getObjectSize)(Collector *, const void* obj);
	const char* (*getObjectDescription)(Collector *, const void* obj);
	int	(*getObjectIndex)(Collector *, const void* obj);
	void*	(*getObjectBase)(Collector *, const void* obj);

	void	(*walkMemory)(Collector *, void *addr);
	/* for_addr is a gcable object, addr may not be */
	void	(*walkConservative)(Collector *, 
				    const void* addr, uint32 length,
				    const void *for_addr);

	void	(*registerFixedTypeByIndex)(Collector *, 
			int index, const char *description);

	void 	(*registerGcTypeByIndex)(Collector *, 
			int index,
			walk_func_t walk, final_func_t final, 
			destroy_func_t destroy, const char
					 *description);
	struct Hjava_lang_Throwable *(*throwOOM)(Collector *);
	hashtab_t (*getStringTable)(Collector *);
	void	(*setStringTable)(Collector *, hashtab_t);
    	struct Hjava_lang_Object *(*copyObject)(Collector *,
						struct Hjava_lang_Object *obj);
	void	(*printInfo)(Collector *);
};

Collector* createGC(void (*_walkRootSet)(Collector*));

/*
 * Convenience macros
 */
#define GC_add_ref(G, mem)			\
    ((G)->ops->addRef)((Collector*)(G), (mem))
#define GC_rm_ref(G, mem)			\
    ((G)->ops->rmRef)((Collector*)(G), (mem))
#define GC_walk_refs(G)				\
    ((G)->ops->walkRefs)((Collector*)(G))

#define GC_malloc(G, size, type)	\
    ((G)->ops->malloc)((Collector*)(G), (size), (type))
#define GC_realloc(G, addr, size, type)	\
    ((G)->ops->realloc)((Collector*)(G), (addr), (size), (type))
#define GC_free(G, addr)		\
    ((G)->ops->free)((Collector*)(G), (addr))
#define GC_invoke(G, mustgc)		\
    ((G)->ops->invoke)((Collector*)(G), (mustgc))
#define GC_invokeFinalizer(G)		\
    ((G)->ops->invokeFinalizer)((Collector*)(G))
#define GC_init(G)		\
    ((G)->ops->init)((Collector*)(G))
#define GC_enable(G)		\
    ((G)->ops->enable)((Collector*)(G))
#define GC_throwOOM(G)		\
    ((G)->ops->throwOOM)((Collector*)(G))
#define GC_finalizeObject(G, addr) \
    ((G)->ops->finalizeObject)((Collector *)(G), (addr))
#define GC_markAddress(G, addr, from)		\
    ((G)->ops->markAddress)((Collector*)(G), (addr), (from))
static inline void GC_markObject(void *g, const void *addr, const void *from)
{
	if (addr)
		((Collector*) g)->ops->markObject((Collector*) g, addr, from);
}
#define GC_getObjectSize(G, obj)	\
    ((G)->ops->getObjectSize)((Collector*)(G), (obj))
#define GC_getObjectDescription(G, obj)	\
    ((G)->ops->getObjectDescription)((Collector*)(G), (obj))
#define GC_getObjectIndex(G, obj)	\
    ((G)->ops->getObjectIndex)((Collector*)(G), (obj))
#define GC_getObjectBase(G, obj)	\
    ((G)->ops->getObjectBase)((Collector*)(G), (obj))
#define GC_walkMemory(G, addr)	\
    ((G)->ops->walkMemory)((Collector*)(G), (addr))
#define GC_walkConservative(G, addr, len, for_addr)		\
    ((G)->ops->walkConservative)((Collector*)(G), (addr), (len), (for_addr))
#define GC_registerFixedTypeByIndex(G, idx, desc)	\
    ((G)->ops->registerFixedTypeByIndex)((Collector*)(G),   \
				(idx), (desc))
#define GC_registerGcTypeByIndex(G, idx, walk, final, destroy, desc)	\
    ((G)->ops->registerGcTypeByIndex)((Collector*)(G), 	     \
				(idx), (walk), (final), (destroy), (desc))
#define GC_getStringTable(G) 	\
    ((G)->ops->getStringTable)((Collector*)(G))
#define GC_setStringTable(G, H) 	\
    ((G)->ops->setStringTable)((Collector*)(G), (H))
#define GC_copyObject(G, O) 	\
    ((G)->ops->copyObject)((Collector*)(G), (O))
#define GC_printInfo(G)	\
    ((G)->ops->printInfo)((Collector*)(G))

/*
 * Compatibility macros to access GC functions
 */
extern Collector* main_collector;

#ifdef KAFFEH
#define currentCollector() main_collector
#define teamCollector(team) main_collector
#else
#include <janosvm/memResource.h>
#endif

#define	gc_malloc(A,B)	    GC_malloc((Collector *)currentCollector(),A,B)
#define	gc_calloc(A,B,C)    GC_malloc((Collector *)currentCollector(),(A)*(B),C)
#define gc_realloc(A,B,C)   GC_realloc((Collector *)currentCollector(),A,B,C)
#define	gc_free(A)	    GC_free((Collector *)currentCollector(),(A))
#define	gc_free_in_heap(A)  GC_free(gc_get_heap(A),(A))
#define explicit_malloc(A)  GC_malloc((Collector *)currentCollector(),A,GC_ALLOC_FIXED)
#define gc_copy_object(A)   GC_copyObject((Collector *)currentCollector(),A)

#define	invokeGC()	    GC_invoke((Collector *)currentCollector(),1)
#define	adviseGC()	    GC_invoke((Collector *)currentCollector(),0)
#define	invokeFinalizer()   GC_invokeFinalizer((Collector *)currentCollector())
#define gc_add_ref(mem)	    GC_add_ref((Collector *)currentCollector(),mem)
#define gc_rm_ref(mem)	    GC_rm_ref((Collector *)currentCollector(),mem)
#define gc_throwOOM()	    GC_throwOOM((Collector *) currentCollector())

extern char* describeObject(const void* mem);
extern Collector * gc_get_heap(void *a);
struct _memResourceClient;
extern struct _memResourceClient *gc_get_resource(void *a);
extern Collector *newCollector(struct _memResourceClient *mrc,
			       char *policyClass);
struct _jsiMessage;
extern int gc_report(Collector *gcif, struct _jsiMessage *jm);

/*
 * AddModule: gcModule
 *   ModuleDependency: machineDependentModule memResourceModule.initMetal
 *   Comment: Setup the main collector
 * Transition: md
 * Transition: heap
 * Transition: team
 *   ModuleDependency: threadModule jteamModule
 *   Comment: Enable the collector if we're running in metal mode.
 */

#endif
