/*
 * gcProxy.c
 * A little bit of infrastructure to help with nterposing on the gc.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdarg.h>

#include "assert_pp.h"

#include "gcProxy.h"

static
void copyCollectorOps(struct GarbageCollectorInterface_Ops *dest,
		      struct GarbageCollectorInterface_Ops *src)
{
	require(dest != 0);
	require(src != 0);
	
	dest->addRef = src->addRef;
	dest->rmRef = src->rmRef;
	dest->walkRefs = src->walkRefs;
	dest->malloc = src->malloc;
	dest->realloc = src->realloc;
	dest->free = src->free;
	dest->invoke = src->invoke;
	dest->invokeFinalizer = src->invokeFinalizer;
	dest->init = src->init;
	dest->enable = src->enable;
	dest->markAddress = src->markAddress;
	dest->markObject = src->markObject;
	dest->finalizeObject = src->finalizeObject;
	dest->getObjectSize = src->getObjectSize;
	dest->getObjectDescription = src->getObjectDescription;
	dest->getObjectIndex = src->getObjectIndex;
	dest->getObjectBase = src->getObjectBase;
	dest->walkMemory = src->walkMemory;
	dest->walkConservative = src->walkConservative;
	dest->registerFixedTypeByIndex = src->registerFixedTypeByIndex;
	dest->registerGcTypeByIndex = src->registerGcTypeByIndex;
	dest->throwOOM = src->throwOOM;
	dest->getStringTable = src->getStringTable;
	dest->setStringTable = src->setStringTable;
	dest->copyObject = src->copyObject;
	dest->printInfo = src->printInfo;

	/* Make sure we've filled out the entire interface array */
#if !defined(NO_POSTCONDITIONS)
	{
		void **ptr = (void **)dest;
		int lpc, size = sizeof(struct GarbageCollectorInterface_Ops) /
			sizeof(void *);

		/*
		 * Start at index three since we happen to know that the
		 * first three slots are reserved.
		 */
		for( lpc = 3; lpc < size; lpc++ )
		{
			ensure(ptr[lpc] != 0);
		}
	}
#endif
}

gcProxy *wrapCollector(Collector *gc, size_t size, int tag, ...)
{
	gcProxy *retval;

	require(gc != 0);

	if( (retval = gc_malloc(size, GC_ALLOC_FIXED)) )
	{
		va_list args;

		va_start(args, tag);
		wrapCollector_mem_v(gc, retval, tag, args);
		va_end(args);
	}
	return( retval );
}

gcProxy *wrapCollector_mem(Collector *gc, void *mem, int tag, ...)
{
	gcProxy *retval = mem;
	va_list args;

	require(gc != 0);
	require(mem != 0);
	
	va_start(args, tag);
	wrapCollector_mem_v(gc, mem, tag, args);
	va_end(args);
	return( retval );
}

gcProxy *wrapCollector_mem_v(Collector *gc, void *mem, int tag, va_list args)
{
	gcProxy *retval = mem;

	require(gc != 0);
	require(mem != 0);
	
	copyCollectorOps(&retval->ops, gc->ops);
	retval->originalOps = gc->ops;
	gc->ops = &retval->ops;
	while( tag != GPA_TAG_DONE )
	{
		switch( tag )
		{
		case GPA_AddRef:
			retval->ops.addRef = va_arg(args, void *);
			break;
		case GPA_RmRef:
			retval->ops.rmRef = va_arg(args, void *);
			break;
		case GPA_WalkRefs:
			retval->ops.walkRefs = va_arg(args, void *);
			break;
		case GPA_Malloc:
			retval->ops.malloc = va_arg(args, void *);
			break;
		case GPA_Realloc:
			retval->ops.realloc = va_arg(args, void *);
			break;
		case GPA_Free:
			retval->ops.free = va_arg(args, void *);
			break;
		case GPA_Invoke:
			retval->ops.invoke = va_arg(args, void *);
			break;
		case GPA_InvokeFinalizer:
			retval->ops.invokeFinalizer = va_arg(args, void *);
			break;
		case GPA_Init:
			retval->ops.init = va_arg(args, void *);
			break;
		case GPA_Enable:
			retval->ops.enable = va_arg(args, void *);
			break;
		case GPA_MarkAddress:
			retval->ops.markAddress = va_arg(args, void *);
			break;
		case GPA_MarkObject:
			retval->ops.markObject = va_arg(args, void *);
			break;
		case GPA_FinalizeObject:
			retval->ops.finalizeObject = va_arg(args, void *);
			break;
		case GPA_GetObjectSize:
			retval->ops.getObjectSize = va_arg(args, void *);
			break;
		case GPA_GetObjectDescription:
			retval->ops.getObjectDescription =
				va_arg(args, void *);
			break;
		case GPA_GetObjectIndex:
			retval->ops.getObjectIndex = va_arg(args, void *);
			break;
		case GPA_GetObjectBase:
			retval->ops.getObjectBase = va_arg(args, void *);
			break;
		case GPA_WalkMemory:
			retval->ops.walkMemory = va_arg(args, void *);
			break;
		case GPA_WalkConservative:
			retval->ops.walkConservative = va_arg(args, void *);
			break;
		case GPA_ThrowOOM:
			retval->ops.throwOOM = va_arg(args, void *);
			break;
		case GPA_GetStringTable:
			retval->ops.getStringTable = va_arg(args, void *);
			break;
		case GPA_SetStringTable:
			retval->ops.setStringTable = va_arg(args, void *);
			break;
		case GPA_PrintInfo:
			retval->ops.printInfo = va_arg(args, void *);
			break;
		}
		tag = va_arg(args, int);
	}
	return( retval );
}

void unwrapCollector(Collector *gc)
{
	if( gc )
	{
		gcProxy *gp;
		
		gp = (gcProxy *)gc->ops;
		gc->ops = gp->originalOps;
		gc_free(gp);
	}
}
