/*
 * gcProxy.h
 * A little bit of infrastructure to help with interposing on the gc.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _gc_proxy_h
#define _gc_proxy_h

#include "gc.h"

/*
 * The gcProxy object wraps a garbage collector interface so that you can
 * interpose on its operations.  For example, you could keep stats on certain
 * functions or perform manipulations on the memory thats being used.
 *
 * `ops' - The garbage collector operations, they can point to the original
 *   ones or customs.
 * `originalOps' - Points to the original operations structure the gc had
 *   before it was swapped with the pointer to this gcProxy object.
 */
typedef struct _gcProxy {
	struct GarbageCollectorInterface_Ops ops;
	struct GarbageCollectorInterface_Ops *originalOps;
} gcProxy;

/* Tags for the set of functions that can be overridden */
enum {
	GPA_TAG_DONE,
	GPA_AddRef,
	GPA_RmRef,
	GPA_WalkRefs,
	GPA_Malloc,
	GPA_Realloc,
	GPA_Free,
	GPA_Invoke,
	GPA_InvokeFinalizer,
	GPA_Init,
	GPA_Enable,
	GPA_MarkAddress,
	GPA_MarkObject,
	GPA_FinalizeObject,
	GPA_GetObjectSize,
	GPA_GetObjectDescription,
	GPA_GetObjectIndex,
	GPA_GetObjectBase,
	GPA_WalkMemory,
	GPA_WalkConservative,
	GPA_ThrowOOM,
	GPA_GetStringTable,
	GPA_SetStringTable,
	GPA_PrintInfo,
};

/*
 * Wrap a Collector object with one thats of the given size.  Currently this
 * involves swapping the GarbageCollectorInterface_Ops pointer in the Collector
 * object with one that points to a gcProxy object.
 *
 * `gc' - The garbage collector interface to wrap.  You can currently only
 *   wrap collectors once so this must be a bare one.
 * `size' - The size of gcProxy object to allocate, if you have a custom
 *   structure than pass its size, otherwise just do sizeof(gcProxy).
 * `tag' - One of the above tags for initializing the function pointers.
 * returns - The allocated/initialized gcProxy object or NULL if something
 *  failed.
 */
gcProxy *wrapCollector(Collector *gc, size_t size, int tag, ...);
/*
 * Similar to the above function except you can pass the memory that will be
 * used for the gcProxy object.
 */
gcProxy *wrapCollector_mem(Collector *gc, void *mem, int tag, ...);
/*
 * The va_list version of the above function.
 */
gcProxy *wrapCollector_mem_v(Collector *gc, void *mem, int tag, va_list args);
/*
 * Free the memory allocated by the wrapCollector function, don't call this
 * if you used of the wrapCollector_mem functions.
 */
void unwrapCollector(Collector *gc);

#endif
