/*
 * jar.c
 * Handle JAR input files.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "config.h"
#include "assert_pp.h"
#include "debug.h"
#include "config-std.h"
#include "config-io.h"
#include "config-mem.h"
#include "gtypes.h"
#include "jsyscall.h"
#include "inflate.h"
#include "jar.h"
#include "gc.h"
#include "stats.h"
#include "files.h"
#include "mungWallProxy.h"

/* Undefine this to make jar files mutable during the vm lifetime */
// #define STATIC_JAR_FILES

/* Kaffeh stupidity */
#if defined(KAFFEH)
#define pushTeamReservation(team) (jteamReservation *)1
#define popTeamReservation(team)
#define switchToTeam(team,cat) 1
#define returnFromTeam()
#define teamCollector(team) main_collector
typedef struct _jteam jteam;
typedef struct _jteamReservation jteamReservation;
#else
#include <janosvm/memResource.h>
#include <janosvm/logResource.h>
#include <janosvm/jteam.h>
#include <janosvm/jteamReservation.h>
#endif

#if defined(KAFFEH)
#undef  initStaticLock
#define initStaticLock(x)
#undef  staticLockIsInitialized
#define staticLockIsInitialized(x)	1
#undef  lockStaticMutex
#undef  unlockStaticMutex
#define unlockStaticMutex(x)
#define lockStaticMutex(x)
#undef  lockMutex
#undef  unlockMutex
#define lockMutex(x)
#define unlockMutex(x)
#endif

/*
 * The jarCache keeps a list of all the jarFiles cached in the system.
 * However, since some JAR files are opened and closed frequently we don't
 * actively flush unused files from the system unless there are more than
 * JAR_FILE_CACHE_MAX jarFiles in the system.  This means we'll keep some
 * jarFiles in memory longer than we should but it helps to avoid the
 * constant opening/closing some java code does.
 */
static struct _jarCache {
#if !defined(KAFFEH)
	iLock *lock;
#endif
	jarFile *files;
#define JAR_FILE_CACHE_MAX 12
	unsigned int count;
} jarCache;

/*
 * Hash a file name, the hash value is stored in what `hash' points to.
 */
static void hashName(const char *name, unsigned int *hash)
{
	require(name != 0);
	require(hash != 0);
	
	for( (*hash) = 0; *name; name++ )
		(*hash) = (31 * (*hash)) + (*name);
}

/*
 * Find a cached jarFile object.  If the file is found, it is returned and its
 * user count is incremented.
 */
static jarFile *findJarFile(char *name)
{
	jarFile *curr, **prev, *retval = 0;
#if !defined(KAFFEH)
	int iLockRoot;
#endif

	require(name != 0);
	
	lockStaticMutex(&jarCache.lock);
	curr = jarCache.files;
	prev = &jarCache.files;
	while( curr && !retval )
	{
		require_valid_memory(curr);
		require(curr->fileName != 0);
		require(curr->users >= 0);
		
		if( !strcmp(curr->fileName, name) )
		{
			/* unlink it... */
			*prev = curr->next;
			/* and move it to the front */
			curr->next = jarCache.files;
			jarCache.files = curr;
			/* Return this node and increment the user count */
			retval = curr;
			retval->users++;

			ensure(retval->users >= 1);
		}
		prev = &curr->next;
		curr = curr->next;
	}
	unlockStaticMutex(&jarCache.lock);
	return( retval );
}

/*
 * Free the contents of the entry table.
 */
static void collectEntryTable(jarFile *jf)
{
	require(jf != 0);
	require_valid_memory(jf);
	require(jf->users == 0);
	
	if( jf->table )
	{
		addToCounter(&jarmem, "vmmem-jar files",
			     1, -(jlong)GCSIZEOF(jf->table));
		gc_free(jf->table);
		jf->table = 0;
	}
}

/*
 * Free a jarFile structure and its child objects.
 */
static void collectJarFile(jarFile *jf)
{
	require(jf != 0);
	require_valid_memory(jf);
	require(jf->users == 0);
	require(!(jf->flags & JFF_CACHED));
	
	collectEntryTable(jf);
	/* Make sure we free everything */
	if( jf->fd != -1 )
	{
		/* Its not mmap'ed so check for a snapshot */
		if( jf->flags & JFF_SNAPSHOT )
		{
			/* Delete the temporary file and the alloc'ed name */
			KREMOVE(jf->fileName);
			gc_free(jf->fileName);
		}
		/* Close the file */
		KCLOSE(jf->fd);
		jf->fd = -1;
	}
#ifdef HAVE_MMAP
	if( jf->data != MAP_FAILED )
	{
		pp_decl(int rc = ) munmap(jf->data, jf->size);

		ensure(rc == 0);
	}
#endif
	addToCounter(&jarmem, "vmmem-jar files", 1, -(jlong)GCSIZEOF(jf));
	gc_free(jf);
}

/*
 * Cache a jarFile object, if the passed in object matches one in the cache
 * it will be thrown away, and the cached one is returned, with its user
 * count incremented.
 */
static jarFile *cacheJarFile(jarFile *jf)
{
	jarFile *curr, **prev, **lru = 0, *dead_jar = 0, *retval = jf;
	int already_cached = 0;
#if !defined(KAFFEH)
	int iLockRoot;
#endif

	require(jf != 0);
	require_valid_memory(jf);
	require(!(jf->flags & JFF_CACHED));
	
	lockStaticMutex(&jarCache.lock);
	/*
	 * Walk the cache, if the file we're trying to cache already matches
	 * one in the cache then we can just use it.  Otherwise, we need to
	 * make room in the cache and continue.
	 */
	curr = jarCache.files;
	prev = &jarCache.files;
	while( curr && !already_cached )
	{
		require_valid_memory(curr);
		require(curr->fileName != 0);
		require(curr->users >= 0);
		
		/* Look for a matching JAR file */
		if( !strcmp(curr->fileName, jf->fileName) )
		{
			/* Names match, check the dates */
			if( curr->lastModified == jf->lastModified )
			{
				/*
				 * They're the same, unlink it from the cache
				 * so we can put it at the front later on.
				 */
				*prev = curr->next;
				retval = curr;
				retval->users++;
			}
			else
			{
				/*
				 * The modified time is different, purge our
				 * cached version and proclaim this one the
				 * canonical version.
				 *
				 * XXX Hmm...  Do we care how the two dates
				 * relate?  Since this is the one now being
				 * loaded from disk it "must" be the one
				 * the user wants.  Bleh, too bad the
				 * semantics don't seem to have been defined.
				 */
				*prev = curr->next;
				curr->flags &= ~JFF_CACHED;
				dead_jar = curr;
			}
			/*
			 * `jf' is redundant so the number of cached files
			 * isn't going to change.
			 */
			already_cached = 1;

			ensure(retval->users >= 1);
		}
		else if( curr->users == 0 )
		{
			/*
			 * Record the least recently used file in case we need
			 * to eject someone.
			 */
			lru = prev;
		}
		prev = &curr->next;
		curr = curr->next;
	}
	if( !already_cached )
	{
		/*
		 * Cache the file if theres still room rather than ejecting
		 * the lru or if theres no room and no lru.
		 */
		if( (jarCache.count < JAR_FILE_CACHE_MAX) || !lru )
		{
			/* Adding a new cache node */
			jarCache.count++;
		}
		else
		{
			/*
			 * Theres an unused jarFile, unlink the least
			 * recently used one.
			 */
			dead_jar = *lru;
			*lru = dead_jar->next;
			dead_jar->flags &= ~JFF_CACHED;
		}
	}
	/* Put the file at the start of the cache */
	retval->next = jarCache.files;
	jarCache.files = retval;
	retval->flags |= JFF_CACHED;
	unlockStaticMutex(&jarCache.lock);
	/*
	 * Free the redundant/excess files outside of the lock.
	 *
	 * NOTE: The dead_jar test must come first since the file could
	 * already be cached and get superceded by a newer version of the
	 * file.
	 */
	if( dead_jar )
		collectJarFile(dead_jar);
	else if( already_cached )
		collectJarFile(jf);

	ensure(retval != 0);
	
	return( retval );
}

/*
 * Remove a JAR file from the cache.
 */
static void removeJarFile(jarFile *jf)
{
	jarFile *curr, **prev;
#if !defined(KAFFEH)
	int iLockRoot;
#endif

	require(jf != 0);
	require_valid_memory(jf);

	/* Make sure its actually in the cache. */
	if( jf->flags & JFF_CACHED )
	{
		lockStaticMutex(&jarCache.lock);
		{
			curr = jarCache.files;
			prev = &jarCache.files;
			/* Find `jf' on the list and... */
			while( curr != jf )
			{
				require(curr != 0);
				require_valid_memory(curr);
				
				prev = &curr->next;
				curr = curr->next;
			}
			/* unlink it */
			*prev = curr->next;
			jf->next = 0;
			jf->flags &= ~JFF_CACHED;
			jarCache.count--;
			
			ensure(jarCache.count >= 0);
		}
		unlockStaticMutex(&jarCache.lock);
	}
}

void flushJarCache(void)
{
	jarFile **prev, *curr, *next;
#if !defined(KAFFEH)
	int iLockRoot;
#endif

	lockStaticMutex(&jarCache.lock);
	curr = jarCache.files;
	prev = &jarCache.files;
	while( curr )
	{
		next = curr->next;
		if( curr->users == 0 )
		{
			*prev = next;
			curr->flags &= ~JFF_CACHED;
			collectJarFile(curr);
		}
		else
		{
			prev = &curr->next;
		}
		curr = next;
	}
	unlockStaticMutex(&jarCache.lock);
}

/*
 * Convenient read function that operates on regular or mmap'ed files.
 * This also takes an `instantiation' function which is used to convert
 * any data into the proper byte order and alignment.
 */
static inline int jarRead(jarFile *jf, char *buf, off_t len,
			  int (*ins_func)(uint8 *dest, uint8 *src))
{
	int retval = -1;

	require(jf != 0);
	require(buf != 0);
	
#ifdef HAVE_MMAP
	if( jf->data != MAP_FAILED )
	{
		/* The requested length should be bound by the actual */
		if( (jf->offset + len) > jf->size )
			len = jf->size - jf->offset;
		if( len <= 0 )
		{
			retval = 0;
		}
		else
		{
			if( ins_func )
			{
				/*
				 * We optimize this to use the instantiation
				 * function to do the copying instead of
				 * doing a memcpy and then moving stuff
				 * around.
				 */
				jf->offset += ins_func(buf,
						       jf->data +
						       jf->offset);
			}
			else
			{
				/* Just copy the bits */
				memcpy(buf, jf->data + jf->offset, len);
				jf->offset += len;
			}
			retval = len;
		}
	}
	else
#endif
	{
		ssize_t bytes_left, bytes_read;
		int rc = 0;

		bytes_left = len;
		/* XXX is this loop necessary? */
		while( bytes_left &&
		       (rc = KREAD(jf->fd,
				   &buf[len - bytes_left],
				   bytes_left,
				   &bytes_read)) &&
		       bytes_read )
		{
			bytes_left -= bytes_read;
		}
		if( rc )
		{
			jf->error = SYS_ERROR(rc);
		}
		else
		{
			/* Instantiate the memory */
			if( ins_func )
				ins_func(buf, buf);
			retval = len;
		}
	}
	return( retval );
}

/*
 * Instantiate a structure that was encoded in a flat file.  We use a define
 * to make sure the instantiation function is inlined.
 */
#ifdef HAVE_MMAP
#define jarInstantiate(jf, buf, ins_func) \
{ \
	if( jf->data != MAP_FAILED ) \
	{ \
		jf->offset += ins_func(buf, jf->data + jf->offset); \
	} \
	else \
	{ \
		ins_func(buf, buf); \
	} \
}
#else
#define jarInstantiate(jf, buf, ins_func) \
{ \
	ins_func(buf, buf); \
}
#endif

/*
 * Convenient seek function that operates on regular or mmap'ed files.
 */
static inline off_t jarSeek(jarFile *jf, off_t offset, int whence)
{
	off_t retval = (off_t)-1;

	require(jf != 0);
	
#ifdef HAVE_MMAP
	if( jf->data != (char*)-1 )
	{
		off_t pos;
		
		switch( whence )
		{
		case SEEK_CUR:
			pos = jf->offset + offset;
			break;
		case SEEK_SET:
			pos = offset;
			break;
		case SEEK_END:
			pos = jf->size + offset;
			break;
		default:
			ABORT();
			break;
		}
		if( (pos >= 0) && (pos < jf->size) )
		{
			jf->offset = pos;
			retval = pos;
		}
	}
	else
#endif
	{
		off_t off;
		int rc;
		
		rc = KLSEEK(jf->fd, offset, whence, &off);
		if( rc )
			jf->error = SYS_ERROR(rc);
		else
			retval = off;
	}
	return( retval );
}

/* Macro used below to create a 32 bit value from a uint8 buffer */
#define copy32le(dest, buf, index) \
do { \
	register uint32 tmp; \
	\
	tmp = (((buf)[(index) + 3] << 24) | \
	       ((buf)[(index) + 2] << 16) | \
	       ((buf)[(index) + 1] << 8) | \
	       ((buf)[(index) + 0])); \
	dest = tmp; \
} while(0);

/* Macro used below to create a 16 bit value from a uint8 buffer */
#define copy16le(dest, buf, index) \
do { \
	register uint16 tmp; \
	\
	tmp = (((buf)[(index) + 1] << 8) | \
	       ((buf)[(index) + 0])); \
	dest = tmp; \
} while(0);

/*
 * Unfortunately, the data members of the zip headers aren't aligned properly
 * so we can't just blast the data into a structure.  So we have to use these
 * instantiation functions to copy the data and do any byte swapping.
 *
 * Note 1: These functions work from high to low memory so that we're
 * basically just shuffling the bits around and not using extra storage.
 *
 * Note 2: We're hoping that the compiler is smart enough to drop the unused
 * data members when inlining...
 */
static inline int
instantiateCentralDir(uint8 *dest, uint8 *buf)
{
	jarCentralDirectoryRecord *cdr = (jarCentralDirectoryRecord *)dest;

	require(dest != 0);
	require(buf != 0);
	
	copy32le(cdr->relativeLocalHeaderOffset, buf, 42);
	copy32le(cdr->externalFileAttribute, buf, 38); /* Not used */
	copy16le(cdr->internalFileAttribute, buf, 36); /* Not used */
	copy16le(cdr->diskNumberStart, buf, 34); /* Not used */
	copy16le(cdr->fileCommentLength, buf, 32);
	copy16le(cdr->extraFieldLength, buf, 30);
	copy16le(cdr->fileNameLength, buf, 28);
	copy32le(cdr->uncompressedSize, buf, 24);
	copy32le(cdr->compressedSize, buf, 20);
	copy32le(cdr->crc, buf, 16); /* Not used */
	copy16le(cdr->lastModifiedDate, buf, 14);
	copy16le(cdr->lastModifiedTime, buf, 12);
	copy32le(cdr->compressionMethod, buf, 10);
	copy16le(cdr->flags, buf, 8); /* Not used */
	copy16le(cdr->versionExtract, buf, 6); /* Not used */
	copy16le(cdr->versionMade, buf, 4); /* Not Used */
	return( FILE_SIZEOF_CENTRALDIR );
}

static inline int
instantiateLocalHeader(uint8 *dest, uint8 *buf)
{
	jarLocalHeader *lh = (jarLocalHeader *)dest;

	require(dest != 0);
	require(buf != 0);
	
	copy16le(lh->extraFieldLength, buf, 28);
	copy16le(lh->fileNameLength, buf, 26);
	copy32le(lh->uncompressedSize, buf, 22); /* Not used */
	copy32le(lh->compressedSize, buf, 18); /* Not used */
	copy32le(lh->crc, buf, 14); /* Not used */
	copy16le(lh->lastModifiedDate, buf, 12); /* Not used */
	copy16le(lh->lastModifiedTime, buf, 10); /* Not used */
	copy16le(lh->compressionMethod, buf, 8); /* Not used */
	copy16le(lh->flags, buf, 6); /* Not used */
	copy16le(lh->versionExtract, buf, 4); /* Not used */
	return( FILE_SIZEOF_LOCALHEADER );
}

static inline int
instantiateCentralDirEnd(uint8 *dest, uint8 *buf)
{
	jarCentralDirectoryEnd *cde = (jarCentralDirectoryEnd *)dest;

	require(dest != 0);
	require(buf != 0);
	
	copy16le(cde->commentLength, buf, 18); /* Not used */
	copy32le(cde->offsetOfDirectory, buf, 16);
	copy32le(cde->sizeOfDirectory, buf, 12); /* Not used */
	copy16le(cde->nrOfEntriesInDirectory, buf, 10);
	copy16le(cde->nrOfEntriesInThisDirectory, buf, 8); /* Not used */
	copy16le(cde->numberOfDiskWithDirectory, buf, 6); /* Not used */
	copy16le(cde->numberOfDisk, buf, 4); /* Not used */
	return( FILE_SIZEOF_CENTRALEND );
}

static inline int instantiateSignature(uint8 *dest, uint8 *buf)
{
	uint32 *sig = (uint32 *)dest;

	require(dest != 0);
	require(buf != 0);
	
	copy32le(sig[0], buf, 0);
	/*
	 * We're instantiating this structure in multiple stages so we don't
	 * move the file pointer.
	 */
	return( 0 );
}

/*
 * Read a JAR header from the file and check it's signature.
 */
static int readJarHeader(jarFile *jf, uint32 sig, void *buf, int len)
{
	int retval = 0;

	require(jf != 0);
	require((sig == CENTRAL_HEADER_SIGNATURE) ||
		(sig == LOCAL_HEADER_SIGNATURE) ||
		(sig == CENTRAL_END_SIGNATURE));
	require(buf != 0);
	require(len >= 0);
	
	if( jarRead(jf, buf, len, instantiateSignature) == len )
	{
		/* Check the signature */
		if( sig == ((uint32 *)buf)[0] )
		{
			retval = 1;
		}
		else
		{
			jf->error = "Bad signature";
#if !defined(KAFFEH)
			teamLogEntry("jar",
				     LEA_String, "Error", "Bad signature", -1,
				     LEA_Integer, "expecting", sig,
				     LEA_Integer, "found", ((uint32 *)buf)[0],
				     LEA_String, "filename", jf->fileName, -1,
				     LEA_TAG_DONE);
#endif
		}
	}
	return( retval );
}

/*
 * Add a jarEntry to the given jarFile.
 */
static void addJarEntry(jarFile *jf, jarEntry *je)
{
	unsigned int hash;

	require(jf != 0);
	require_valid_memory(jf);
	require(jf->table != 0);
	require(je != 0);
	require(je->fileName != 0);
	
	hashName(je->fileName, &hash);
	hash = hash % jf->tableSize;
	je->next = jf->table[hash];
	jf->table[hash] = je;
}

/*
 * Create a jarEntry structure for the current central directory record in
 * the file.
 */
static int initJarEntry(jarFile *jf, jarEntry *je, char **name_strings)
{
	jarCentralDirectoryRecord cdr;
	int retval = 0;

	require(jf != 0);
	require(je != 0);
	require(name_strings != 0);
	require(*name_strings != 0);

	/* Read the header */
	if( readJarHeader(jf, CENTRAL_HEADER_SIGNATURE, &cdr,
			  FILE_SIZEOF_CENTRALDIR) )
	{
		int read_size;

		/* Instantiate the header */
		jarInstantiate(jf, (uint8 *)&cdr, instantiateCentralDir);
		je->next = 0;
		/* Allocate space for our name */
		(*name_strings) -= cdr.fileNameLength + 1;
		je->fileName = *name_strings;
		/* Copy whatever we care about from the file record */
		je->dosTime = (cdr.lastModifiedDate << 16) |
			cdr.lastModifiedTime;
		je->localHeaderOffset = cdr.relativeLocalHeaderOffset;
		je->uncompressedSize = cdr.uncompressedSize;
		je->compressedSize = cdr.compressedSize;
		je->compressionMethod = cdr.compressionMethod;
		/* Read the file name */
		if( (read_size = jarRead(jf,
					 je->fileName,
					 cdr.fileNameLength,
					 0)) >= 0 )
		{
			/* Make sure its terminated */
			je->fileName[cdr.fileNameLength] = 0;
			/*
			 * Skip the extra field and comment, we should
			 * now be positioned at the next directory
			 * record, if there is one.
			 */
			if( jarSeek(jf,
				    cdr.extraFieldLength +
				    cdr.fileCommentLength,
				    SEEK_CUR) > 0 )
			{
				retval = 1;

				ensure(strlen(je->fileName) ==
				       cdr.fileNameLength);
			}
		}

		ensure_valid_memory(jf->table);
	}
	else
	{
		jf->error = "Bad central record signature";
	}
	return( retval );
}

/*
 * Find the central directory end, find out the number of central directory
 * entries and seek to the start of them.
 */
static int getCentralDirCount(jarFile *jf, unsigned int *out_dir_size)
{
	int pos, retval = -1;

	require(jf != 0);
	require(out_dir_size != 0);
	
	/* The central directory end is at the end of the file */
	if( (pos = jarSeek(jf, -FILE_SIZEOF_CENTRALEND, SEEK_END)) > 0 )
	{
		jarCentralDirectoryEnd cde;

		if( readJarHeader(jf, CENTRAL_END_SIGNATURE,
				  &cde, FILE_SIZEOF_CENTRALEND) )
		{
			jarInstantiate(jf, (uint8 *)&cde,
				       instantiateCentralDirEnd);
			if( cde.nrOfEntriesInDirectory >=
			    (cde.sizeOfDirectory / FILE_SIZEOF_CENTRALDIR) )
			{
				jf->error = "Entry count doesn't match directory size";
#if !defined(KAFFEH)
				teamLogEntry("jar",
					     LEA_String, "Error", "Entry count doesn't match directory size", -1,
					     LEA_Integer, "size", cde.sizeOfDirectory,
					     LEA_Integer, "count", cde.nrOfEntriesInDirectory,
					     LEA_String, "filename", jf->fileName, -1,
					     LEA_TAG_DONE);
#endif
			}
			else if( cde.sizeOfDirectory > pos )
			{
				jf->error = "Impossibly large directory size";
#if !defined(KAFFEH)
				teamLogEntry("jar",
					     LEA_String, "Error", "Impossibly large directory size", -1,
					     LEA_Integer, "size", cde.sizeOfDirectory,
					     LEA_Integer, "central-end-position", pos,
					     LEA_String, "filename", jf->fileName, -1,
					     LEA_TAG_DONE);
#endif
			}
			else if( jarSeek(jf,
					 cde.offsetOfDirectory,
					 SEEK_SET) >= 0 )
			{
				*out_dir_size = cde.sizeOfDirectory;
				retval = cde.nrOfEntriesInDirectory;
			}
		}
		else
		{
			jf->error = "Failed to find end of JAR record";
		}
	}
	return( retval );
}

/*
 * Read the central directory records from the file
 */
static int readJarEntries(jarFile *jf)
{
	unsigned int dir_size;
	int retval = 0;

	require(jf != 0);
	
	if( (jf->count = getCentralDirCount(jf, &dir_size)) >= 0 )
	{
		unsigned int table_size;

		/*
		 * Compute a sensible size for the hash table base on the
		 * number of entries in the jar.
		 */
		jf->tableSize = (jf->count + 3) / 4;
		/*
		 * We want a single block of memory for all jarEntry's and
		 * their names.
		 */
		/* Hash table buckets */
		table_size = sizeof(jarEntry *) * jf->tableSize;
		/* Add the total size of the central directory records */
		table_size += dir_size;
		/* ... however, we don't read the whole central dir struct */
		table_size -= jf->count * FILE_SIZEOF_CENTRALDIR;
		/* ... we use jarEntry's */
		table_size += jf->count * sizeof(jarEntry);
		/* ... with NULL terminated name strings. */
		table_size += jf->count * 1;
		if( (jf->table = gc_malloc(table_size, GC_ALLOC_JAR)) )
		{
			char *name_strings;
			jarEntry *je;
			int lpc;

			/* jarEntry's are right after the table */
			je = (jarEntry *)(jf->table + jf->tableSize);
			/* names are at the end of the memory block */
			name_strings = ((char *)jf->table) + table_size;
			retval = 1;
			/*
			 * Read in the central directory records and add them
			 * to the hash table.
			 */
			for( lpc = 0; (lpc < jf->count) && retval; lpc++ )
			{
				if( initJarEntry(jf, je, &name_strings) )
				{
					addJarEntry(jf, je);
					je++;
				}
				else
				{
					retval = 0;
				}
			}
			
			ensure_valid_memory(jf->table);
		}
		else
		{
			jf->error = "Out of memory";
#if !defined(KAFFEH)
			teamLogEntry("jar",
				     LEA_String, "Error", "Out of memory", -1,
				     LEA_String, "jarname", jf->fileName, -1,
				     LEA_TAG_DONE);
#endif
		}
	}
	else if( jf->error )
	{
#if !defined(KAFFEH)
		teamLogEntry("jar",
			     LEA_String, "Error", "Not a jar file", -1,
			     LEA_String, "jarname", jf->fileName, -1,
			     LEA_TAG_DONE);
#endif
	}
	else
	{
		retval = 1;
#if !defined(KAFFEH)
		teamLogEntry("jar",
			     LEA_String, "Warning", "No entries in jar", -1,
			     LEA_String, "jarname", jf->fileName, -1,
			     LEA_TAG_DONE);
#endif
	}
	return( retval );
}

/*
 * Simple function to handle uncompressing the file data.
 */
static uint8 *inflateJarData(jarFile *jf, jarEntry *je,
			     jarLocalHeader *lh, uint8 *buf)
{
	uint8 *retval = 0;

	require(jf != 0);
	require(je != 0);
	require(lh != 0);
	require(buf != 0);
	
	switch( je->compressionMethod )
	{
	case COMPRESSION_STORED:
		retval = buf;
		break;
	case COMPRESSION_DEFLATED:
		if( (retval = gc_malloc(je->uncompressedSize, GC_ALLOC_JAR)) )
		{
			if( inflate_oneshot(buf,
					    je->compressedSize,
					    retval,
					    je->uncompressedSize) == 0 )
			{
				addToCounter(&jarmem, "vmmem-jar files",
					     1, GCSIZEOF(retval));

				ensure_valid_memory(retval);
			}
			else
			{
				jf->error = "Decompression failed";
#if !defined(KAFFEH)
				teamLogEntry(
					"jar",
					LEA_String, "Error",
					"Decompression failed", -1,
					LEA_String, "jarname",
					jf->fileName, -1,
					LEA_String, "filename",
					je->fileName, -1,
					LEA_TAG_DONE);
				gc_free(retval);
#endif
				retval = 0;
			}
		}
		else
		{
			jf->error = "Out of memory";
#if !defined(KAFFEH)
			teamLogEntry("jar",
				     LEA_String, "Error", "Out of memory", -1,
				     LEA_String, "jarname", jf->fileName, -1,
				     LEA_String, "filename", je->fileName, -1,
				     LEA_TAG_DONE);
#endif
		}
		gc_free(buf);
		break;
	default:
		jf->error = "Unsupported compression in JAR file";
#if !defined(KAFFEH)
		teamLogEntry("jar",
			     LEA_String, "Error",
			     "Unsupported compression used in JAR file", -1,
			     LEA_String, "jarname", jf->fileName, -1,
			     LEA_String, "filename", je->fileName, -1,
			     LEA_TAG_DONE);
#endif
		gc_free(buf);
		break;
	}
	return( retval );
}

uint8 *getDataJarFile(struct _jteam *jt, jarFile *jf, jarEntry *je)
{
	uint8 *buf = 0, *retval = 0;
	jteamReservation *jr;
	jarLocalHeader lh;
#if !defined(KAFFEH)
	int iLockRoot;
#endif

	require(jf != 0);
	require(je != 0);

	if( !(jr = pushTeamReservation(kernelTeam)) )
		return 0;
	lockMutex(jf);
	if( switchToTeam(jt, 0) == 0 )
		goto done;
	/* Move to the local header in the file and read it. */
	if( !jf->error &&
	    (jarSeek(jf, je->localHeaderOffset, SEEK_SET) >= 0) &&
	    readJarHeader(jf, LOCAL_HEADER_SIGNATURE, &lh,
			  FILE_SIZEOF_LOCALHEADER) )
	{
		jarInstantiate(jf, (uint8 *)&lh, instantiateLocalHeader);
		/* Skip the local file name and extra fields */
		jarSeek(jf,
			lh.fileNameLength + lh.extraFieldLength,
			SEEK_CUR);
		/* Allocate some memory and read in the file data */
		if( (buf = (uint8 *)gc_malloc(je->compressedSize,
					      GC_ALLOC_JAR)) )
		{
			if( jarRead(jf, buf, je->compressedSize, 0) < 0 )
			{
				gc_free(buf);
				buf = 0;
				jf->error = "I/O error";
			}
			else
			{
				ensure_valid_memory(buf);
			}
		}
		else
		{
			jf->error = "Out of memory";
#if !defined(KAFFEH)
			teamLogEntry("jar",
				     LEA_String, "Error", "Out of memory", -1,
				     LEA_String, "jarname", jf->fileName, -1,
				     LEA_String, "filename", je->fileName, -1,
				     LEA_TAG_DONE);
#endif
		}
	}
	returnFromTeam();
 done:
	unlockMutex(jf);
	popTeamReservation(jr);
	if( buf )
	{
		if( switchToTeam(jt, 0) )
		{
			/* Try to decompress it */
			retval = inflateJarData(jf, je, &lh, buf);
			returnFromTeam();
		}
	}
	return( retval );
}

/*
 * If a jarFile is cached it may be closed before it is used again so we might
 * have to reopen the file.
 */
static jarFile *delayedOpenJarFile(jarFile *jf)
{
	jarFile *retval = 0;
	int fd, rc;

	require(jf != 0);

	/* Open the file and check for a different modified time */
	if( !(rc = KOPEN(jf->fileName, O_RDONLY|O_BINARY, 0, &fd)) )
	{
#if !defined(STATIC_JAR_FILES)
		struct stat jar_stat;

		if( !(rc = KFSTAT(fd, &jar_stat)) )
		{
			if( jar_stat.st_mtime == jf->lastModified )
			{
#endif /* !defined(STATIC_JAR_FILES) */
#if !defined(KAFFEH)
				int iLockRoot;
#endif
				
				/* Only set the fd in the structure here */
				lockMutex(jf);
				if( jf->fd == -1 )
					jf->fd = fd; /* We're first */
				else
					KCLOSE(fd); /* Someone else set it */
				unlockMutex(jf);
				retval = jf;
#if !defined(STATIC_JAR_FILES)
			}
			else
			{
				KCLOSE(fd);
				/* The file was changed, reopen it. */
				removeJarFile(jf);
				retval = openJarFile(jf->fileName);
			}
		}
		else
		{
			KCLOSE(fd);
		}
#endif /* !defined(STATIC_JAR_FILES) */
	}
	/*
	 * The file either disappeared, or a new one was made, get rid of the
	 * old one.
	 */
	if( retval != jf )
	{
		removeJarFile(jf);
		closeJarFile(jf);
	}
	return( retval );
}

jarFile *openJarFile(char *name)
{
	jarFile *retval = 0;

	require(name != 0);
	
	if( switchToTeam(kernelTeam, 0) == 0 )
		return 0;
	/* Look for it in the cache first */
	if( (retval = findJarFile(name)) )
	{
		/* Check if we need to reopen the file */
		if( (retval->fd == -1)
#ifdef HAVE_MMAP
		    && (retval->data == MAP_FAILED)
#endif
		    )
		{
			retval = delayedOpenJarFile(retval);
		}
	}
	/*
	 * If a cached file wasn't found or its broken then try to make a new
	 * one.
	 */
	if( !retval &&
	    (retval = (jarFile *)gc_malloc(sizeof(jarFile) +
					   strlen(name) +
					   1,
					   GC_ALLOC_JAR)) )
	{
		int rc, error = 0;

		retval->fileName = (char *)(retval + 1);
		strcpy(retval->fileName, name);
		retval->users = 1;
		retval->lastModified = 0;
		retval->count = 0;
		retval->error = 0;
		retval->fd = -1;
		retval->table = 0;
		retval->tableSize = 0;
		if( (rc = KOPEN(name, O_RDONLY|O_BINARY, 0, &retval->fd)) )
		{
			/* Error opening the file */
			error = 1;
#if !defined(KAFFEH)
			teamLogEntry("jar",
				     LEA_String, "Error",
				     "Can't open JAR file", -1,
				     LEA_String, "filename", name, -1,
				     LEA_TAG_DONE);
#endif
		}
		else
		{
			struct stat jar_stat;
			
			if( KFSTAT(retval->fd, &jar_stat) == 0 )
 			{
				if( (jar_stat.st_mode & S_IFDIR) )
				{
					/* Its a directory! */
					error = 1;
#if !defined(KAFFEH)
					/* XXX Fix me */
					teamLogEntry("jar",
						     LEA_String, "Error",
						     "It's a directory!", -1,
						     LEA_String, "filename",
						     name, -1,
						     LEA_TAG_DONE);
#endif
				}
				else
				{
					retval->lastModified =
						jar_stat.st_mtime;
#ifdef HAVE_MMAP
					/*
					 * Setup the mmap members in the
					 * jarFile structure
					 */
					retval->size = jar_stat.st_size;
					retval->data = mmap(NULL,
							    retval->size,
							    PROT_READ,
							    MAP_SHARED,
							    retval->fd,
							    0);
					if( retval->data == MAP_FAILED )
					{
						/*
						 * Unsuccessful mmap, fallback
						 * to the FD for now.
						 */
					}
					else
					{
						/*
						 * Successful mmap, close the
						 * FD
						 */
						KCLOSE(retval->fd);
						retval->fd = -1;
						retval->offset = 0;
					}
#endif
				}
			}
			else
			{
				error = 1;
#if !defined(KAFFEH)
				teamLogEntry("jar",
					     LEA_String, "Error",
					     "Can't stat JAR file", -1,
					     LEA_String, "filename", name, -1,
					     LEA_TAG_DONE);
#endif
			}
			/*
			 * If everythings a go, figure out how many directory
			 * entries there are and read them in.
			 */
			if( !error && !readJarEntries(retval) )
			{
				error = 1;
			}
		}
		if( !error )
		{
			addToCounter(&jarmem, "vmmem-jar files",
				     1, GCSIZEOF(retval));
			if( retval->table )
				addToCounter(&jarmem, "vmmem-jar files",
					     1, GCSIZEOF(retval->table));
			/*
			 * No errors, so we cache the file.  If someone else
			 * beat us to the cache we'll use theirs instead.
			 */
			retval = cacheJarFile(retval);
		}
		else
		{
			/* Something went wrong, cleanup our mess */
			retval->users = 0;
			collectJarFile(retval);
			retval = 0;
		}
	}
	returnFromTeam();

	ensure_valid_memory(retval);
	return( retval );
}

void closeJarFile(jarFile *jf)
{
	if( jf )
	{
#if !defined(KAFFEH)
		int iLockRoot;
#endif

		if( !switchToTeam(kernelTeam, 0) )
		{
			/* XXX This is bad... */
			return;
		}
		lockStaticMutex(&jarCache.lock);
		jf->users--;
		if( jf->users == 0 )
		{
			if( jarCache.count <= JAR_FILE_CACHE_MAX )
			{
				/*
				 * The cache isn't full so if its in there
				 * we can leave it.
				 */
#ifdef HAVE_MMAP
				if( jf->data != MAP_FAILED )
				{
					/*
					 * Unmap the object and force FD I/O
					 * from now on.
					 */
					munmap(jf->data, jf->size);
					jf->data = MAP_FAILED;
				}
				else
#endif
				{
					KCLOSE(jf->fd);
				}
				jf->fd = -1;
			}
			else
			{
				/* Too many files in the cache, remove it */
				removeJarFile(jf);
			}
			if( !(jf->flags & JFF_CACHED) )
			{
				/*
				 * Not cached anymore and this was the last
				 * reference so collect the file.
				 */
				collectJarFile(jf);
			}
		}
		unlockStaticMutex(&jarCache.lock);
		returnFromTeam();
	}
}

jarEntry *lookupJarFile(jarFile *jf, char *entry_name)
{
	jarEntry *retval = 0;

	require(jf != 0);
	require(entry_name != 0);
	
	/*
	 * No need to visit the kernel here since we're just walking the
	 * structure.
	 */
	if( jf->tableSize )
	{
		unsigned int hash;
		jarEntry *je;

		hashName(entry_name, &hash);
		hash = hash % jf->tableSize;
		je = jf->table[hash];
		while( je && !retval )
		{
			if( !strcmp(je->fileName, entry_name) )
				retval = je;
			je = je->next;
		}
	}
	return( retval );
}

int snapshotJarFile(jarFile *jf)
{
	int retval = 0;
#if !defined(KAFFEH)
	int iLockRoot;
#endif

	/* Yank it out of the cache first */
	removeJarFile(jf);
	/* Don't let anyone read from it while we're working. */
	lockMutex(jf);
#ifdef HAVE_MMAP
	if( jf->fd == -1 )
	{
		jf->flags |= JFF_SNAPSHOT;
		/* Its in memory, don't have to worry about it... */
		KREMOVE(jf->fileName);
		retval = 1;
	}
	else
#endif
	if( !(jf->flags & JFF_SNAPSHOT) )
	{
		static char *name_template = "temporary-jar.XXXXXX";
		char *temp_name = 0, *buffer = 0;
		int temp_fd = -1;

		/* Create a new JAR file */
#define BUFFER_SIZE 4096
		if( (temp_name = (char *)
		     gc_malloc(strlen(name_template) + 1, GC_ALLOC_JAR)) &&
		    strcpy(temp_name, name_template) &&
		    (buffer = (char *) gc_malloc(BUFFER_SIZE, GC_ALLOC_JAR)) &&
		    ((temp_fd = mkstemp(temp_name)) >= 0) )
		{
			int rc, bytes_in = 0, bytes_out;

			/*
			 * XXX Copy the file, would be nice if we could just
			 * symlink the sucker.
			 */
			jarSeek(jf, 0, SEEK_SET);
			while( !(rc = KREAD(jf->fd, buffer,
					    BUFFER_SIZE, &bytes_in)) &&
			       (bytes_in > 0) &&
			       !(rc = KWRITE(temp_fd, buffer,
					     bytes_in, &bytes_out)) &&
			       (bytes_in == bytes_out) )
			{
			}
			if( rc == 0 )
			{
				/* Success, switch to the new file */
				KREMOVE(jf->fileName);
				jf->fileName = temp_name;
				KCLOSE(jf->fd);
				jf->fd = temp_fd;
				retval = 1;
			}
			else
			{
				/* Failure, back out */
#if !defined(KAFFEH)
				teamLogEntry(
					"jar",
					LEA_String, "Error", "Snapshot failed", -1,
					LEA_String, "source", jf->fileName, -1,
					LEA_String, "destination", temp_name, -1,
					LEA_TAG_DONE);
#endif
				KCLOSE(temp_fd);
				KREMOVE(temp_name);
				gc_free(temp_name);
			}
		}
		else
		{
#if !defined(KAFFEH)
			teamLogEntry("jar",
				     LEA_String, "Error", "Couldn't make snapshot temporary file", -1,
				     LEA_TAG_DONE);
#endif
			gc_free(temp_name);
			gc_free(buffer);
		}
#undef BUFFER_SIZE
	}
	if( retval )
	{
		jf->flags |= JFF_SNAPSHOT;
	}
	unlockMutex(jf);
	return( retval );
}
