/*
 * lockPool.c
 * Maintain a pool of iLock objects.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <assert.h>

#include "lockPool.h"
#include "assert_pp.h"
#include "gtypes.h"
#include "gc.h"
#include "jthread.h"
#include "threadData.h"
#include <janosvm/jteam.h>
#include "module.h"
#include <misc/memoryStack.h>
#include <misc/memoryStack-inline.h>

/** Module declarations **/
int initLockPoolModule(int stage);
int testLockPoolModule(void);
void diagnoseLockPoolModule(void);

module_t lockPoolModule = {
	"lockPool",
	0,
	0,
	0,
	testLockPoolModule,
	diagnoseLockPoolModule
};

iLock metalILock;

static iLock *poolLock;
#define LOCK_POOL_SIZE 16
METAL_MEMORY_STACK(rootLockPool, sizeof(iLock) * LOCK_POOL_SIZE);
static int usedLockCount = 0;
static int freeLockCount = 0;
static int totalLockCount = 0;
static iLock *freeLocks;

/** Module functions **/
int testLockPoolModule(void)
{
	iLock *locks[LOCK_POOL_SIZE / 2];
	int lpc, retval = 1;

	/* Allocate some locks */
	for( lpc = 0; lpc < (LOCK_POOL_SIZE / 2); lpc++ )
	{
		assert(locks[lpc] = obtainLock());
		assert(locks[lpc]->link.owner == 0);
	}
	/* Make sure they're all different */
	for( lpc = 0; lpc < (LOCK_POOL_SIZE / 2); lpc++ )
	{
		int lpc2;

		for( lpc2 = 0; lpc2 < (LOCK_POOL_SIZE / 2); lpc2++ )
		{
			if( lpc2 != lpc )
				assert(locks[lpc] != locks[lpc2]);
		}
	}
	/* Give them back */
	for( lpc = 0; lpc < (LOCK_POOL_SIZE / 2); lpc++ )
	{
		releaseLock(locks[lpc]);
	}
	return( retval );
}

void diagnoseLockPoolModule(void)
{
	if( (usedLockCount + freeLockCount) != totalLockCount )
	{
		fprintf(stderr,
			"lockPool: Inconsistent pool state\n"
			"  used=%d free=%d total=%d\n",
			usedLockCount,
			freeLockCount,
			totalLockCount);
	}
}

/*
 * Allocate a lock in the kernel and add it to the free list.
 *
 * returns - True if the allocation worked, false otherwise.
 */
static
int addFreeLock(void)
{
	int retval = 0;
	iLock *il;

	if( !(il = allocStackSpace_shallow(&rootLockPool.ms, sizeof(iLock))) )
	{
		/*
		 * The kernel manages the lock pool, so we need to switch to
		 * it to allocate more space on the stack.
		 */
		if( switchToTeam(kernelTeam, 0) )
		{
			/* Try to allocate a lock from the stack */
			il = allocStackSpace(&rootLockPool.ms, sizeof(iLock));
			returnFromTeam();
		}
	}
	if( il )
	{
		/* Got it, link it in */
		il->link.next = freeLocks;
		freeLocks = il;
		totalLockCount++;
		freeLockCount++;
		retval = 1;
	}
	return( retval );
}

iLock *obtainLock(void)
{
	iLock *retval = 0;
	int iLockRoot;

	lockStaticMutex(&poolLock);
	if( freeLocks || addFreeLock() )
	{
		/* Theres a lock available, pluck it off the free list */
		retval = freeLocks;
		freeLocks = retval->link.next;
		retval->link.next = 0;
		retval->link.owner = 0;
		retval->holder = 0;
		retval->mux = 0;
		retval->cv = 0;
		usedLockCount++;
		freeLockCount--;
	}
	unlockStaticMutex(&poolLock);
	return( retval );
}

void releaseLock(iLock *il)
{
	int iLockRoot;
	
	require(il != 0);

	/*
	 * XXX Stupidity here, releaseLock is called from the thread thats
	 * releasing its own iLock so if there were contention on poolLock
	 * it would create all sorts of problems.  But there won't be
	 * since obtain/releaseLock end up being called with intsDisabled().
	 * Technically the call to this should be put in destroyThreadData
	 * but then its kinda late...
	 */
	lockStaticMutex(&poolLock);
	usedLockCount--;
	freeLockCount++;
	il->link.next = freeLocks;
	freeLocks = il;
	unlockStaticMutex(&poolLock);
}

void reclaimTeamLocks(struct _jteam *jt)
{
	iLock *reclaimed_locks, **reclaimed_tail = &reclaimed_locks;
	int reclaimed_count = 0;
	memoryStack *curr = 0;
	int iLockRoot;
	iLock *il;
	
	/*
	 * Walk through the pool looking for locks owned by the team.  Since
	 * the team is dead and the part of the stack that holds our locks
	 * won't change we don't need to lock it.
	 */
	il = (iLock *)bottomOfMemoryStack(&rootLockPool.ms);
	while( (il = (iLock *)walkMemoryStack(&rootLockPool.ms,
					      &curr,
					      il,
					      sizeof(iLock))) )
	{
		if( il->link.owner == jt )
		{
			reclaimed_count++;
			*reclaimed_tail = il;
			reclaimed_tail = &il->link.next;
		}
	}
	lockStaticMutex(&poolLock);
	/* Update stats */
	usedLockCount -= reclaimed_count;
	freeLockCount += reclaimed_count;
	/* Put the reclaimed locks on the free list */
	*reclaimed_tail = freeLocks;
	freeLocks = reclaimed_locks;
	unlockStaticMutex(&poolLock);
}
