/*
 * lockPool.h
 * Maintain a pool of iLock objects.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef __lock_pool_h
#define __lock_pool_h

struct _jteam;

/*
 * An iLock structure represents the heavy weight lock used by the locking
 * system when the light weight lock isn't enough.
 *
 * `link.next' - Next pointer in linked list
 * `link.owner' - The team that currently owns this lock.  NOTE: This is only
 *   set when the lock is put in place, otherwise the lock is referenced by
 *   the thread its owned by.
 * `holder' - NULL or a pointer to some place in the stack of the thread that
 *   holds the lock
 * `mux' - Linked list of threads waiting to take the lock
 * `cv' - Linked list of threads waiting to be signalled
 */
typedef struct _iLock {
	union {
		struct _iLock *next;
		struct _jteam *owner;
	} link;
	void *holder;
	struct _jthread *mux;
	struct _jthread *cv;
} iLock;

/*
 * Obtain an iLock object from the global pool and returns it.  NOTE: This
 * function expects the thread to be in an unkillable state (i.e. interrupts
 * disabled, in a visit).
 *
 * returns - A new iLock object.
 */
iLock *obtainLock(void);
/*
 * Releases an iLock object back into the global pool.  NOTE: This function
 * expects the thread to be in an unkillable state (i.e. interrupts disabled,
 * in a visit).
 *
 * `il' - An iLock object as returned by obtainLock.
 */
void releaseLock(iLock *il);
/*
 * Reclaim any iLock objects that might have been in use when a team was
 * killed.
 *
 * `jt' - The dead team that we should reclaim locks from.
 */
void reclaimTeamLocks(struct _jteam *jt);

/*
 * Metal iLock object for use by the metal thread.
 */
extern iLock metalILock;

/*
 * AddModule: lockPoolModule
 */

#endif
