/*
 * locks.c
 * Manage locking system
 *
 * Copyright (c) 1996-1999
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "object.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "thread.h"
#include "locks.h"
#include "ksem.h"
#include "errors.h"
#include "exception.h"
#include "md.h"
#include "jthread.h"
#include "gc.h"
#include "debug.h"
#include "threadData-inline.h"
#include <janosvm/jteam.h>

#ifndef STACKREDZONE
# ifdef REDZONE
#  define STACKREDZONE	REDZONE
# else
#  define STACKREDZONE	1024
# endif
#endif

/*
 * If we don't have an atomic compare and exchange defined then make
 * one out of a simple atomic exchange (using the LOCKINPROGRESS value
 * as the place holder).  If we don't have ATOMIC_EXCHANGE, we'll just
 * fake it.
 */
#if !defined(COMPARE_AND_EXCHANGE)
#if defined(ATOMIC_EXCHANGE)
#define	COMPARE_AND_EXCHANGE(A,O,N) \
	({ \
		iLock* val = LOCKINPROGRESS; \
		ATOMIC_EXCHANGE((A), val); \
		if (val == (O)) { \
			*(A) = (N); \
		} \
		else { \
			*(A) = (O); \
		} \
		(val == (O) ? 1 :  0); \
	})
#else
#error Please define COMPARE_AND_EXCHANGE or ATOMIC_EXCHANGE 
#endif
#endif

/* Count number of backoffs.  XXX move into the STAT infrastructure. */
int backoffcount = 0;

static
void illegalState(void)
{
	throwException(IllegalMonitorStateException);
}

/*
 * Get a heavy lock for the object and hold it.  If the object doesn't
 * have a heavy lock then we create one.
 */
static
iLock*
getHeavyLock(iLock** lkp)
{
	jthread_t jtid;
	iLock* old;
	iLock* lk = 0;
	jlong timeout;

DBG(SLOWLOCKS,
    	dprintf("  getHeavyLock(**lkp=%p, *lk=%p, th=%p)\n",
		lkp, *lkp, jthread_current());
)

	jtid = jthread_current();
	timeout = 1;
	for (;;) {
		/*
		 * Get the current lock and replace it with LOCKINPROGRESS to
		 * indicate changes are in progress.
		 */
		old = *lkp;
		if( old == LOCKINPROGRESS ||
		    !COMPARE_AND_EXCHANGE(lkp, old, LOCKINPROGRESS) )
		{
			/* Someone else put the lock in LOCKINPROGRESS state */
			backoffcount++;
			jthread_sleep(timeout);
			/* Back off */
			timeout = (timeout << 1)|timeout;
			continue;
		}

		/*
		 * If bottom bit is set, strip off and use pointer as pointer
		 * to heavy lock
		 */
		if ((((uintp)old) & 1) == 1) {
DBG(SLOWLOCKS,
    			dprintf("    got cached lock\n");
)
			lk = (iLock*)(((uintp)old) & (uintp)-2);
		}
		else {
			/* Create a heavy lock object for others to find. */
DBG(SLOWLOCKS,
    			dprintf("    got new lock\n");
    )
	
			lk = takeThreadLock(jthread_get_data(jtid));
			lk->holder = (void*)old;
			lk->mux = 0;
			lk->cv = 0;
		}
		return (lk);
	}
}

/*
 * Release the lock - only the one who has claimed it can do this
 * so there's no need for locked instructions.
 */
static
void
putHeavyLock(iLock** lkp, iLock* lk)
{
	assert(*lkp == LOCKINPROGRESS);

DBG(SLOWLOCKS,
	dprintf("  putHeavyLock(**lkp=%p, *lk=%p, th=%p)\n", 
		lkp, lk, jthread_current());
)

	if (lk == LOCKFREE) {
		*lkp = LOCKFREE;
	}
	else {
		*lkp = (iLock*)(1|(uintp)lk);
	}
}

int _lockedMutex(iLock **lkp)
{
	int retval = 0;
	uintp val;

	val = (uintp)*lkp;
	if( val != 0 )
	{
		uintp holder;

		if( val & 1 )
		{
			iLock *lk;

			/* Cached lock */
			lk = getHeavyLock(lkp);
			holder = (uintp)lk->holder;
			if( lk->mux || lk->cv )
			{
				putHeavyLock(lkp, lk);
			}
			else
			{
				giveThreadLock(jthread_get_data(
					jthread_current()), lk);
				*lkp = (iLock *)holder;
			}
		}
		else
		{
			holder = val;
		}
		retval = jthread_on_current_stack((void *)holder);
	}
	return retval;
}

/*
 * Slowly lock a mutex.  We get the heavy lock and lock that instead.
 * If we can't lock it we suspend until we can.
 */
/*static*/ void
slowLockMutex(iLock** lkp, void* where)
{
	jthread_t jtid;
	iLock* lk;

DBG(SLOWLOCKS,
    	dprintf("slowLockMutex(**lkp=%p, where=%p, th=%p)\n",
	       lkp, where, jthread_current());
)
	//jthread_disable_stop(); /* protect the heavy lock, and its queues */

	jtid = jthread_current();
	for (;;) {
		lk = getHeavyLock(lkp);

		/* If I hold the heavy lock then just keep on going */
		if (jthread_on_current_stack(lk->holder)) {
			if( lk->mux || lk->cv )
			{
				putHeavyLock(lkp, lk);
			}
			else
			{
				giveThreadLock(jthread_get_data(jtid), lk);
				*lkp = lk->holder;
			}
			//jthread_enable_stop();
			
			ensure(jthread_get_data(jthread_current())->
			       freeLock != 0);
			return;
		}

		/* If no one holds the heavy lock then claim it */
		if (lk->holder == 0) {
			if( lk->mux || lk->cv )
			{
				if( jthread_get_data(jtid)->freeLock == 0 )
				{
					iLock *stolenLock;
					
					if( lk->mux )
					{
						stolenLock = takeThreadLock(jthread_get_data(lk->mux));
					}
					else
					{
						stolenLock = takeThreadLock(jthread_get_data(lk->cv));
					}
					giveThreadLock(jthread_get_data(jtid),
						       stolenLock);
				}
				lk->holder = where;
				putHeavyLock(lkp, lk);
			}
			else
			{
				giveThreadLock(jthread_get_data(jtid), lk);
				*lkp = where;
			}
			//jthread_enable_stop();
			
			ensure(jthread_get_data(jthread_current())->
			       freeLock != 0);
			return;
		}

		/* Otherwise wait for holder to release it */
		jthread_get_data(jtid)->nextlk = lk->mux;
		lk->mux = jtid;
		putHeavyLock(lkp, lk);
		ksemGet(&jthread_get_data(jtid)->sem, 0);
	}
	
	ensure(jthread_get_data(jthread_current())->freeLock != 0);
}

/*
 * Slowly unlock a mutex.  If there's someone waiting then we wake them up
 * so they can claim the lock.  If no one is waiting we revert the lock to
 * a fast thin lock.
 */
/*static*/ void
slowUnlockMutex(iLock** lkp, void* where)
{
	jthread_t jtid;
	iLock* lk;

DBG(SLOWLOCKS,
    	dprintf("slowUnlockMutex(**lkp=%p, where=%p, th=%p)\n",
	       lkp, where, jthread_current());
)
	//jthread_disable_stop(); /* protect the heavy lock, and its queues */
	lk = getHeavyLock(lkp);

	/* Only the lock holder can be doing an unlock */
	if (!jthread_on_current_stack(lk->holder)) {
		if( lk->mux || lk->cv )
		{
			putHeavyLock(lkp, lk);
		}
		else
		{
			giveThreadLock(jthread_get_data(jthread_current()),
				       lk);
			*lkp = lk->holder;
		}
		//jthread_enable_stop();
		illegalState();
	}

	/* If holder isn't where we are now then this isn't the final unlock */
	if (lk->holder > where) {
		if( lk->mux || lk->cv )
		{
			putHeavyLock(lkp, lk);
		}
		else
		{
			giveThreadLock(jthread_get_data(jthread_current()),
				       lk);
			*lkp = lk->holder;
		}
		//jthread_enable_stop();

		ensure(jthread_get_data(jthread_current())->freeLock != 0);
		return;
	}

	/* Final unlock - if someone is waiting for it now would be a good
	 * time to tell them.
	 */
	if (lk->mux != 0) {
		{
			jtid = lk->mux;
			lk->mux = jthread_get_data(jtid)->nextlk;

			/* No loops please... */
			ensure(lk->mux != jtid);
		}
		jthread_get_data(jtid)->nextlk = 0;
		lk->holder = 0;
		if( jthread_get_data(jtid)->freeLock == 0 )
		{
			if( lk->mux || lk->cv )
			{
				iLock *stolenLock;
				
				if( lk->mux )
				{
					stolenLock = takeThreadLock(jthread_get_data(lk->mux));
				}
				else
				{
					stolenLock = takeThreadLock(jthread_get_data(lk->cv));
				}
				giveThreadLock(jthread_get_data(jtid),
					       stolenLock);
				ensure((lk->cv == jthread_current()) ||
				       jthread_get_data(jthread_current())->
				       freeLock != 0);
				putHeavyLock(lkp, lk);
			}
			else
			{
				giveThreadLock(jthread_get_data(jtid), lk);
				ensure((lk->cv == jthread_current()) ||
				       jthread_get_data(jthread_current())->
				       freeLock != 0);
				putHeavyLock(lkp, LOCKFREE);
			}
		}
		else if( lk->cv || lk->mux )
		{
			ensure((lk->cv == jthread_current()) ||
			       (jthread_get_data(jthread_current())->
				freeLock != 0));
			putHeavyLock(lkp, lk);
		}
		else
		{
			giveThreadLock(jthread_get_data(jthread_current()),
				       lk);
			putHeavyLock(lkp, LOCKFREE);
		}
		ksemPut(&jthread_get_data(jtid)->sem);
	}
	/* If someone's waiting to be signaled keep the heavy in place */
	else if (lk->cv != 0) {
		lk->holder = 0;
		putHeavyLock(lkp, lk);
		//jthread_enable_stop();

		ensure((lk->cv == jthread_current()) ||
		       jthread_get_data(jthread_current())->freeLock != 0);
		return;
	}
	else {
		require(jthread_get_data(jthread_current())->freeLock == 0);

		giveThreadLock(jthread_get_data(jthread_current()), lk);
		putHeavyLock(lkp, LOCKFREE);
	}
	//jthread_enable_stop();
}

void
_slowUnlockMutexIfHeld(iLock** lkp, void* where)
{
	iLock* lk;
	void* holder;

DBG(SLOWLOCKS,
    	dprintf("slowUnlockMutexIfHeld(**lkp=%p, where=%p, th=%p)\n",
	       lkp, where, jthread_current());
)
	lk = getHeavyLock(lkp);
	holder = lk->holder;
	if( (lk->mux == 0) && (lk->cv == 0) )
	{
	    giveThreadLock(jthread_get_data(jthread_current()), lk);
	    *lkp = holder;
	}
	else
	{
	    putHeavyLock(lkp, lk);
	}

	if (jthread_on_current_stack(holder) && (holder <= where))
	{
		slowUnlockMutex(lkp, where);
	}
}

void*
_releaseLock(iLock** lkp)
{
	iLock* lk;
	void* holder;

DBG(SLOWLOCKS,
    	dprintf("_releaseLock(**lkp=%p, th=%p)\n",
	       lkp, jthread_current());
)

	lk = getHeavyLock(lkp);
	holder = lk->holder;

	/* I must be holding the damn thing */
	if (!jthread_on_current_stack(holder)) {
		putHeavyLock(lkp, holder);
		throwException(IllegalMonitorStateException);
	}

	putHeavyLock(lkp, lk);
	slowUnlockMutex(lkp, holder);

	return (holder);
}

void
_acquireLock(iLock** lkp, void* holder)
{
	slowLockMutex(lkp, holder);
}

jboolean
_waitCond(iLock** lkp, jlong timeout)
{
	iLock* lk;
	void* holder;
	struct _jthread *jtid, **ptr;
	memResourceClient *mrc;
	jboolean r;

DBG(SLOWLOCKS,
    	dprintf("_waitCond(**lkp=%p, timeout=%ld, th=%p)\n",
	       lkp, (long)timeout, jthread_current());
)

	lk = getHeavyLock(lkp);
	holder = lk->holder;

	jtid = jthread_current();
	/* I must be holding the damn thing */
	if (!jthread_on_current_stack(holder)) {
		if( jthread_get_data(jtid)->freeLock == 0 )
		{
			giveThreadLock(jthread_get_data(jtid), lk);
			*lkp = holder;
		}
		else if( lk->mux || lk->cv )
		{
			putHeavyLock(lkp, lk);
		}
		illegalState();
	}

	jthread_get_data(jtid)->waiter = lkp;
	jthread_get_data(jtid)->nextlk = lk->cv;
	lk->cv = jtid;
	putHeavyLock(lkp, lk);
	slowUnlockMutex(lkp, holder);
	r = ksemGet(&jthread_get_data(jtid)->sem, timeout);
	jthread_get_data(jtid)->waiter = 0;

	/* Timeout */
	if (r == false) {
		lk = getHeavyLock(lkp);
		/* Remove myself from CV or MUX queue - if I'm not on either
		 * then I should wait on myself to remove any pending signal.
		 */
		for (ptr = &lk->cv;
		     *ptr != 0;
		     ptr = &jthread_get_data(*ptr)->nextlk) {
			if ((*ptr) == jtid) {
				*ptr = jthread_get_data(jtid)->nextlk;
				goto found;
			}
		}
		for (ptr = &lk->mux;
		     *ptr != 0;
		     ptr = &jthread_get_data(*ptr)->nextlk) {
			if ((*ptr) == jtid) {
				*ptr = jthread_get_data(jtid)->nextlk;
				goto found;
			}
		}
		/* Not on list - so must have been signalled after all -
		 * decrease the semaphore to avoid problems.
		 */
		ksemGet(&jthread_get_data(jtid)->sem, 0);

		found:;
		putHeavyLock(lkp, lk);
		
		if( (mrc = gc_get_resource(lkp)) )
		{
			if( mrc->client.owner->state != JTS_ALIVE )
			{
				slowLockMutex(lkp, holder);
				illegalState();
			}
		}
	}

	slowLockMutex(lkp, holder);

	ensure(jthread_get_data(jthread_current())->freeLock != 0);
	
	return (r);
}

void
_signalCond(iLock** lkp)
{
	struct _jthread *jtid;
	iLock* lk;
	
DBG(SLOWLOCKS,
    	dprintf("_signalCond(**lkp=%p, th=%p)\n",
	       lkp, jthread_current());
)

	lk = getHeavyLock(lkp);

	if (!jthread_on_current_stack(lk->holder)) {
		if( lk->mux || lk->cv )
		{
			putHeavyLock(lkp, lk);
		}
		else
		{
			giveThreadLock(jthread_get_data(jthread_current()),
				       lk);
			*lkp = lk->holder;
		}
		illegalState();
	}

	/* Move one CV's onto the MUX */
	jtid = lk->cv;
	if (jtid != 0) {
		lk->cv = jthread_get_data(jtid)->nextlk;
		jthread_get_data(jtid)->nextlk = lk->mux;
		lk->mux = jtid;
	}

	if( lk->mux || lk->cv )
	{
		putHeavyLock(lkp, lk);
	}
	else
	{
		giveThreadLock(jthread_get_data(jthread_current()), lk);
		*lkp = lk->holder;
	}

	ensure(jthread_get_data(jthread_current())->freeLock != 0);
}

void
_broadcastCond(iLock** lkp)
{
	struct _jthread *jtid;
	iLock* lk;

DBG(SLOWLOCKS,
    	dprintf("_broadcastCond(**lkp=%p, th=%p)\n",
	       lkp, jthread_current());
)

	lk = getHeavyLock(lkp);

	if (!jthread_on_current_stack(lk->holder)) {
		if( lk->mux || lk->cv )
		{
			putHeavyLock(lkp, lk);
		}
		else
		{
			giveThreadLock(jthread_get_data(jthread_current()),
				       lk);
			*lkp = lk->holder;
		}
		illegalState();
	}

	/* Move all the CV's onto the MUX */
	while (lk->cv != 0) {
		jtid = lk->cv;
		lk->cv = jthread_get_data(jtid)->nextlk;
		jthread_get_data(jtid)->nextlk = lk->mux;
		lk->mux = jtid;
	}

	if( lk->mux )
	{
		putHeavyLock(lkp, lk);
	}
	else
	{
		giveThreadLock(jthread_get_data(jthread_current()), lk);
		*lkp = lk->holder;
	}

	ensure(jthread_get_data(jthread_current())->freeLock != 0);
}

/*
 * Lock a mutex - try to do this quickly but if we failed because
 * we can't determine if this is a multiple entry lock or we've got
 * contention then fall back on a slow lock.
 */
void
_lockMutex(iLock** lkp, void* where)
{
	uintp val;

	require(jthread_current());
	require(jthread_get_data(jthread_current()));
	require(jthread_get_data(jthread_current())->freeLock != 0);
	
	val = (uintp)*lkp;

	if (val == 0) {
		if (!COMPARE_AND_EXCHANGE(lkp, 0, (iLock*)where)) {
			slowLockMutex(lkp, where);
		}
	}
	else if ((val - (uintp)where) > (STACKREDZONE / 2)) {
		/* XXX count this in the stats area */
		slowLockMutex(lkp, where);
	}
	
	ensure(jthread_current());
	ensure(jthread_get_data(jthread_current()));
	ensure(jthread_get_data(jthread_current())->freeLock != 0);
}

/*
 * Unlock a mutex - try to do this quickly but if we failed then
 * we've got contention so fall back on a slow lock.
 */
void
_unlockMutex(iLock** lkp, void* where)
{
	uintp val;

	val = (uintp)*lkp;

	if ((val & 1) != 0) {
		slowUnlockMutex(lkp, where);
	}
	else if ((val == (uintp)where) /* XXX squirrely bit */
		&& !COMPARE_AND_EXCHANGE(lkp, (iLock*)where, LOCKFREE)) {
		slowUnlockMutex(lkp, where);
	}
	
	ensure(jthread_get_data(jthread_current())->freeLock != 0);
}

void
lockObject(Hjava_lang_Object* obj)
{
	_lockMutex(&obj->lock, &obj);
}

void
unlockObject(Hjava_lang_Object* obj)
{
	_unlockMutex(&obj->lock, &obj);
}

void
slowLockObject(Hjava_lang_Object* obj, void* where)
{
	slowLockMutex(&obj->lock, where);
}

void
slowUnlockObject(Hjava_lang_Object* obj, void* where)
{
	slowUnlockMutex(&obj->lock, where);
}

void
dumpLocks(void)
{
}
