/* gc-incremental.c
 * The garbage collector.
 * The name is misleading.  GC is non-incremental at this point.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#undef	SUPPORT_VERBOSEMEM

#include "config.h"
#include "debug.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "gc.h"
#include "gc-private.h"
#include "locks.h"
#include "thread.h"
#include "jthread.h"
#include "errors.h"
#include "md.h"
#include "stats.h"
#include "exception.h"
#include "stringSupport.h"
#include "baseClasses.h"
#include "support.h"
#include "mungWallProxy.h"
#include <janosvm/logResource.h>
#include <janosvm/leak.h>
#include <janosvm/jteam.h>
#include <janosvm/jteamReservation.h>

#include <jsi/jsi.h>

gc_collector gc_metal_collector;
gc_collector *defined_collectors;
iLock *collectorListLock;

#define lock_collector(C)				\
{							\
	lockMutex(C);					\
	(C)->gcman.owner = &iLockRoot;			\
}

#define unlock_collector(C) 			\
{						\
	(C)->gcman.owner = 0;			\
	unlockMutex(C);				\
}

/* Is this pointer within our managed heap? */ 
#define IS_A_HEAP_POINTER(from) \
	((uintp) (from) >= gc_heap_base && \
	 (uintp) (from) < gc_heap_base + gc_heap_range)

/* Avoid recursively allocating OutOfMemoryError */
#define OOM_ALLOCATING		((void *) -1)

#define GCSTACKSIZE		16384
#define FINALIZERSTACKSIZE	THREADSTACKSIZE

static const int finalised = 5;		/* moved to white at start of gc */
static const int mustfree = 4;		/* temporary list */
static const int white = 3;
static const int grey = 2;
static const int black = 1;
static const int finalise = 0;

#if defined(SUPPORT_VERBOSEMEM)

static void objectStatsChange(gc_unit*, int);
static void objectStatsPrint(void);

#define	OBJECTSTATSADD(M)	objectStatsChange(M, 1)
#define	OBJECTSTATSREMOVE(M)	objectStatsChange(M, -1)
#define	OBJECTSTATSPRINT()	objectStatsPrint()

#else

#define	OBJECTSTATSADD(M)
#define	OBJECTSTATSREMOVE(M)
#define	OBJECTSTATSPRINT()

#endif

static void 
dump_totals(gc_collector *c, char *str, ...)
{
	va_list ap;

	va_start(ap, str);
	fprintf(stderr, "[heap %p ", c);
	vfprintf(stderr, str, ap);
	fprintf(stderr, " total=%dk, limit=%dk"
		" local=%dk, trigger=%dk, objects=%dk (%d)"
		" %d entries, %d finalizable]\n",
		c->client->used/1024, c->client->max/1024,
		c->local/1024, c->trigger/1024,
		c->stats.totalmem/1024, c->stats.totalobj,
		c->n_refs, c->stats.finalobj);
	va_end(ap);
}

static void* gcMalloc1(gc_collector* c, size_t size, int fidx,
		       struct _errorInfo *info);

/* For statistics gathering, record how many objects and how 
 * much memory was marked.
 */
static inline
void record_marked(gc_collector *c, int nr_of_objects, uint32 size)
{       
        c->stats.markedobj += nr_of_objects;
        c->stats.markedmem += size;
} 

static void gcFree(Collector* gcif, void* mem);

/* Standard GC function sets.  We call them "allocation types" now */
#if !defined(DEBUG)
static 
#endif
gcFuncs gcFunctions[GC_ALLOC_MAX_INDEX + 1];

/* Number of registered allocation types */
static int nrTypes;

/*
 * register an allocation type under a certain index
 * NB: we could instead return a pointer to the record describing the
 * allocation type.  This would give us more flexibility, but it wouldn't
 * allow us to use compile-time constants.
 */
static void
registerTypeByIndex(int index, walk_func_t walk, final_func_t final,
	destroy_func_t destroy,
	const char *description)
{
	/* once only, please */
	assert (gcFunctions[index].description == 0);
	/* don't exceed bounds */
	assert (index >= 0 && 
		index < sizeof(gcFunctions)/sizeof(gcFunctions[0]));
	gcFunctions[index].walk = walk;
	gcFunctions[index].final = final;
	gcFunctions[index].destroy = destroy;
	gcFunctions[index].description = description;
	if (index >= nrTypes) {
		nrTypes = index + 1;
	}
}

/*
 * Register a fixed allocation type.  The only reason we tell them apart
 * is for statistical purposes.
 */
static void
gcRegisterFixedTypeByIndex(Collector* gcif, 
	int index, const char *description)
{
	registerTypeByIndex(index, 0, GC_OBJECT_FIXED, 0, description);
}

/*
 * Register a allocation type that is subject to gc.  
 */
static void
gcRegisterGcTypeByIndex(Collector* gcif,
	int index, walk_func_t walk, final_func_t final,
	destroy_func_t destroy,
	const char *description)
{
	registerTypeByIndex(index, walk, final, destroy, description);
}

static void startGC(gc_collector *gcif);
void finishGC(gc_collector *gcif, int *);

/* Return true if gc_unit is pointer to an allocated object */
static inline int
gc_heap_isobject(gc_collector *c, gc_block *info, gc_unit *unit)
{
	uintp p = (uintp) UTOMEM(unit) - gc_heap_base;
	int idx;

	if (!(p & (MEMALIGN - 1)) && p < gc_heap_range && info->firstblock == info) {
		/* Make sure 'unit' refers to the beginning of an
		 * object.  We do this by making sure it is correctly
		 * aligned within the block.
		 */
		idx = GCMEM2IDX(info, unit);
		if (idx < info->nr && GCBLOCK2MEM(info, idx) == unit
		    && (GC_GET_COLOUR(info, idx) & GC_COLOUR_INUSE) ==
		    GC_COLOUR_INUSE) {
			return 1;
		}
	}
	return 0;
}

static inline int
gc_heap_isfixed(gc_collector *c, gc_block *info, gc_unit *unit)
{
	uintp p = (uintp) UTOMEM(unit) - gc_heap_base;
	int idx;

	if (!(p & (MEMALIGN - 1)) && p < gc_heap_range && info->firstblock == info) {
		/* Make sure 'unit' refers to the beginning of an
		 * object.  We do this by making sure it is correctly
		 * aligned within the block.
		 */
		idx = GCMEM2IDX(info, unit);
		if (idx < info->nr && GCBLOCK2MEM(info, idx) == unit
		    && (GC_GET_COLOUR(info, idx) & GC_COLOUR_MASK) ==
		    GC_COLOUR_FIXED) {
			return 1;
		}
	}
	return 0;
}

static void
markObjectDontCheck(gc_collector *c, gc_unit *unit, gc_block *info,
		    int idx, const void *from)
{
	int walking_thread_stack = 0;
	gc_block *blk;
	int myidx;

	if( isLeak(UTOMEM(unit)) )
	{
		printLeak(UTOMEM(unit), (void *)from);
	}
	/* from is null --- if we walk a global ref */
	if (from != 0 && !IS_A_HEAP_POINTER(from)) {
		/*
		 * or from is "base" of main thread's stack (which is outside 
		 * the heap range)
		 */
		walking_thread_stack = 1;
	} else 
	if (from != 0) {
		blk = GCMEM2BLOCK(from);
		myidx = GCMEM2IDX(blk, from);

		/* If it's a pointer from a thread stack, we do not
		 * consider it a heap pointer
		 * XXX: make sure that THREADCTX is only used for stacks!
		 */
		if (GC_GET_FUNCS(blk, myidx) == GC_ALLOC_THREADSTACK) {
			walking_thread_stack = 1;
		} else {
			/* If the from location is not on a stack, it
			 * shouldn't be outside of our heap
			 */
			assert(blk->collector == c);
		}
	}
	
	/* are we asked to mark an object that is not on our heap? */
	if (info->collector != c) {
		/* Is this safe at all?  we can have gcs and
		 * finalizers competing with each other for cpu time,
		 * so we can create new refs in a finalizer thread
		 * stack while a foreign heap is being collected.
		 *
		 * However, it is OK to perform two gcs
		 * simultaneously.
		 */
		/*assert (GC_GET_COLOUR(info, idx) == GC_COLOUR_WHITE);*/

		/*
		 * The write barrier and assignment should be atomic.
		 * This requires changing kaffe.def.  Adding the exit
		 * point is already atomic, so we can handle the
		 * window between exit-point creation and assignment
		 * by keeping exits points from our own stack alive.
		 */

		return;
	}

	/* If object's been traced before, don't do it again */
	if (GC_GET_COLOUR(info, idx) != GC_COLOUR_WHITE) {
		return;
	}

	DBG(GCWALK,	
	    dprintf("  marking @%08p: %s %d bytes\n", UTOMEM(unit),
		    describeObject(UTOMEM(unit)), 
		    GCBLOCKSIZE(GCMEM2BLOCK(unit)));
	    );
	DBG(GCSTAT,
	    switch (GC_GET_FUNCS(info, idx)) {
	    case GC_ALLOC_NORMALOBJECT:
	    case GC_ALLOC_FINALIZEOBJECT:
	    case GC_ALLOC_PRIMARRAY:
	    case GC_ALLOC_REFARRAY: {
		    Hjava_lang_Object *obj;
		    obj = (Hjava_lang_Object *)(unit+1);
		    if (obj->dtable != NULL) {
			    Hjava_lang_Class *c;
			    c = OBJECT_CLASS(obj);
			    if (c)
				    c->live_count++;
		    }
	    }});
	    

	/* If we found a new white object, mark it as grey and
	 * move it into the grey list.
	 */
	DBG(GCHEAPS,
	    if (!from) dprintf("%p: marked %s@%p from stack\n",
			       c, describeObject(UTOMEM(unit)), UTOMEM(unit));
	    );

	DBG(GCFIN,
	    if (GC_GET_FUNCS(info, idx) == GC_ALLOC_FINALIZEOBJECT
		&& GC_GET_STATE(info, idx) == GC_STATE_FINALIZED) {
		    dprintf("%p: finalized object %s@%p",
			    c, describeObject(UTOMEM(unit)),
			    UTOMEM(unit));
		    dprintf(" is reachable from %s@%p\n",
			    describeObject(from), from);
	    });
		
	GC_SET_COLOUR(info, idx, GC_COLOUR_GREY);
	UREMOVELIST(unit);
	UAPPENDLIST(c->lists[grey], unit);
}

/*
 * Mark the memory given by an address if it really is an object.
 */
static void
gcMarkAddress(Collector* gcif, const void* mem, const void *from)
{
	gc_block* info;
	gc_unit* unit;
	gc_collector *c = (void *) gcif;

	/*
	 * First we check to see if the memory 'mem' is in fact the
	 * beginning of an object.  If not we just return.
	 */

	/* Get block info for this memory - if it exists */
	info = GCMEM2BLOCK(mem);
	unit = UTOUNIT(mem);

	if (gc_heap_isobject(c, info, unit)) {
		markObjectDontCheck(c, unit, info,
				    GCMEM2IDX(info, unit), from);
	}
}

/*
 * Mark an object.  Argument is assumed to point to a valid object.
 */
static void
gcMarkObject(Collector* gcif, const void* objp, const void *from)
{
	gc_collector *c = (void *) gcif;
	gc_unit *unit = UTOUNIT(objp);
	gc_block *info = GCMEM2BLOCK(unit);
	
	DBG(GCDIAG, assert(gc_heap_isobject(c, info, unit)));
	markObjectDontCheck(c, unit, info,
			    GCMEM2IDX(info, unit), from);
}

static void
gcWalkConservative(Collector* gcif, const void* base, uint32 size,
		   const void *from_addr)
{
	gc_collector *c = (void *) gcif;
	int8* mem;

	assert(from_addr);
DBG(GCWALK,	
	dprintf("scanning %d bytes conservatively from %p-%p\n", 
		size, base, base+size);
    )

	record_marked(c, 1, size);

	if (size > 0) {
		for (mem = ((int8*)base) + (size & -ALIGNMENTOF_VOIDP) - sizeof(void*); (void*)mem >= base; mem -= ALIGNMENTOF_VOIDP) {
			void *p = *(void **)mem;
			if (p) {
				if( isLeak(p) )
				{
					printLeak(p, (void *)from_addr);
				}
				gcMarkAddress(gcif, p, from_addr);
			}
		}
	}
}

/*
 * Like walkConservative, except that length is computed from the block size
 * of the object.  Must be called with pointer to object allocated by gc.
 */
static
uint32
gcGetObjectSize(Collector* gcif, const void* mem)
{
	return (GCBLOCKSIZE(GCMEM2BLOCK(UTOUNIT(mem))));
}

static
int
gcGetObjectIndex(Collector* gcif, const void* mem)
{
	gc_collector *c = (void *) gcif;
	gc_unit* unit = UTOUNIT(mem);
	gc_block* info = GCMEM2BLOCK(unit);

	/* This function may be called on a non-heap address.  We
	 * can't use gc_heap_isobject though, since it is false for
	 * explicitly managed objects.
	 */
	if (gc_heap_isobject(c, GCMEM2BLOCK(mem), UTOUNIT(mem))
	    || gc_heap_isfixed(c, GCMEM2BLOCK(mem), UTOUNIT(mem))) {
		if( info && (info->collector == c) ) {
			return (GC_GET_FUNCS(info, GCMEM2IDX(info, unit)));
		}
	}
	return -1;
}

/*
 * Given a pointer within an object, find the base of the object.
 * This works for both gcable and fixed object types.
 *
 * This method uses many details of the allocator implementation.
 * Specifically, it relies on the contiguous layout of block infos
 * and the way GCMEM2BLOCK and GCMEM2IDX are implemented.
 *
 * NB: gcif is unused.   It is not checked whether the pointer is actually
 * within an object allocated by this instance.  This should be a "static"
 * method.
 */
static
void*
gcGetObjectBase(Collector *gcif, const void* mem)
{
	int idx;
	gc_block* info;

	/* quickly reject pointers that are not part of this heap */
	if (!IS_A_HEAP_POINTER(mem)) {
		return (0);
	}

	info = GCMEM2BLOCK(mem);
	if (info->firstblock == 0) {
		return (0);
	}
	info = info->firstblock;
	idx = GCMEM2IDX(info, mem);

	/* report fixed objects as well */
	if (idx < info->nr && 
	    ((GC_GET_COLOUR(info, idx) & GC_COLOUR_INUSE) || 
	     (GC_GET_COLOUR(info, idx) & GC_COLOUR_FIXED))) 
	{
		return (UTOMEM(GCBLOCK2MEM(info, idx)));
	}
	return (0);
}

static
const char*
gcGetObjectDescription(Collector* gcif, const void* mem)
{
	int idx = gcGetObjectIndex(gcif, mem);
	
	return (idx == -1 ? "invalid object pointer"
		: gcFunctions[idx].description);
}

/*
 * Walk a bit of memory.
 */
static
void
gcWalkMemory(Collector* gcif, void* mem)
{
	gc_collector *c = (void *) gcif;

	gc_block* info;
	int idx;
	gc_unit* unit;
	uint32 size;
	walk_func_t walkf;

	unit = UTOUNIT(mem);
	info = GCMEM2BLOCK(unit);
	idx = GCMEM2IDX(info, unit);

	if (info->collector == c) {
		UREMOVELIST(unit);
		UAPPENDLIST(c->lists[black], unit);
		GC_SET_COLOUR(info, idx, GC_COLOUR_BLACK);
	}

	assert(GC_GET_FUNCS(info, idx) < 
		sizeof(gcFunctions)/sizeof(gcFunctions[0]));
	size = GCBLOCKSIZE(info);
	record_marked(c, 1, size);
	walkf = gcFunctions[GC_GET_FUNCS(info, idx)].walk;
	if (walkf != 0) {
DBG(GCWALK,	
		dprintf("walking %d bytes @%08p: %s\n", size, mem, 
			describeObject(mem));
    )
		walkf(gcif, mem, size);
	}
}

/*
 * The Garbage Collector sits in a loop starting a collection, waiting
 * until it's finished incrementally, then tidying up before starting
 * another one.
 */
void
gcMan(void* arg)
{
	gc_unit* unit;
	gc_unit* nunit;
	gc_block* info;
	int idx;
	gc_collector *c = (gc_collector*)arg;
	int iLockRoot;

	/* Cancelation is disabled at startup, and remains disabled
	   throughout this thread's life, we poll at safe cancel
	   points and never try to allocate a threaddeath.
	*/
	lockMutex(&c->gcman);
	/* On startup, cancelation is disabled.  after calling
	 * lock_collector, cancelation is disabled twice. */

	/* Wake up anyone waiting for the GC to finish every time we're done */
	for(;; c->gcman.gcRunning = 0, broadcastCond(&c->gcman)) {
		
		while (c->gcman.gcRunning == 0) {
			waitCond(&c->gcman, 0);
			/* This is a good place to respond to
			 * cancelation
			 */
			if (jthread_killed()) {
				/* It is preferable not to let the
				 * collector try to allocate a
				 * ThreadDeath exception on its way
				 * out.
				 */
				/* XXX Of course not.  What does this have to
				 * do with freeing the reserve below?
				 */
				if (c->reserve) {
DBG(GCHEAPS, 
    dprintf("%x: freeing reserve @ %p\n",
	    c, c->reserve); 
)
					gc_primitive_free(c, c->reserve);
				}
				unlockMutex(&c->gcman);
				jthread_exit();
			}
		}

		lock_collector(c);

		startGC(c);
		
		for (unit = c->lists[grey].cnext; unit != &c->lists[grey]; unit = c->lists[grey].cnext) {
			gcWalkMemory(&c->collector, UTOMEM(unit));
		}

		/* Now walk any white objects which will be finalized.  They
		 * may get reattached, so anything they reference must also
		 * be live just in case.
		 */
		for (unit = c->lists[white].cnext; unit != &c->lists[white]; unit = nunit) {
			nunit = unit->cnext;
			info = GCMEM2BLOCK(unit);
			idx = GCMEM2IDX(info, unit);
			if (GC_GET_STATE(info, idx) == GC_STATE_NEEDFINALIZE) {
				/* this assert is somewhat expensive */
				DBG(GCDIAG,
				    assert(gc_heap_isobject(c, info, unit)));
				GC_SET_STATE(info, idx, GC_STATE_INFINALIZE);
				markObjectDontCheck(c, unit, info,
						    idx, UTOMEM(unit));
			}
		}
		/* We may now have more grey objects, so walk them */
		for (unit = c->lists[grey].cnext; unit != &c->lists[grey]; unit = c->lists[grey].cnext) {
			gcWalkMemory(&c->collector, UTOMEM(unit));
		}

		finishGC(c, &iLockRoot);

		unlock_collector(c);
		
		/* If there's stuff to be finalised then we'd better do it */
		if (c->lists[finalise].cnext != &c->lists[finalise]) {
			lockMutex(&c->finman);
			c->finman.finalRunning = true;
			signalCond(&c->finman);
			unlockMutex(&c->finman);
		}
		
		if (Kaffe_JavaVMArgs[0].enableVerboseGC > 0 && 0) {
			/* print out all the info you ever wanted to
			   know */
			printf(
			    "<GC %p: heap %dK, total before %dK,"
			    " after %dK (%d/%d objs)\n %2.1f%% free,"
			    " alloced %dK (#%d), marked %dK,"
			    " swept %dK (#%d)\n"
			    " %d objs (%dK) awaiting finalization>\n",
				c,
				c->client->used/1024, 
				c->stats.totalmem/1024, 
				(c->stats.totalmem-c->stats.freedmem)/1024, 
				c->stats.totalobj,
				c->stats.totalobj-c->stats.freedobj,
				(1.0 - ((c->stats.totalmem-c->stats.freedmem)/
					(double)c->client->used)) * 100.0,
				c->stats.allocmem/1024,
				c->stats.allocobj,
				c->stats.markedmem/1024, 
				c->stats.freedmem/1024,
				c->stats.freedobj,
				c->stats.finalobj,
				c->stats.finalmem/1024);
		}
		if (Kaffe_JavaVMArgs[0].enableVerboseGC > 1) {
			OBJECTSTATSPRINT();
		}
		c->stats.totalmem -= c->stats.freedmem;
		c->stats.totalobj -= c->stats.freedobj;
		c->stats.allocobj = 0;
		c->stats.allocmem = 0;

		/*
		 * Don't gc again until either we run out of memory or
		 * We allocate another 1/3 of the current heap size.
		 *
		 * Ignore the total memory limit, gc_primitive_alloc
		 * will start a gc if it needs to anyway.
		 *
		 * Use totalmem, rather than the absolute local heap
		 * size , so that we don't expand the heap as
		 * fragmentation sets in.  Put another way, we want to
		 * plug the gaps in the pages we have before sparsly
		 * filling new pages.
		 */
		if( c->policy.obj )
		{
			jlong trigger;
			
			trigger = do_execute_java_method(c->policy.obj,
							 "newTrigger",
							 "(JJ)J",
							 0,
							 0,
							 (jlong)c->stats.
							 totalmem,
							 (jlong)0 /* XXX */).j;
			c->trigger = trigger;
			c->client->gcTrigger = trigger;
		}
		else
		{
			c->trigger = c->stats.totalmem + c->stats.totalmem/3;
			c->client->gcTrigger =
				c->client->used +
				(c->client->used / 3);
		}
		if (Kaffe_JavaVMArgs[0].enableVerboseGC) {
			dump_totals(c, "finished gc");
		}
	}
}

/*
 * Start the GC process by scanning the root and thread stack objects.
 */
static
void
startGC(gc_collector *c)
{
	gc_unit* unit;
	gc_unit* nunit;

	c->stats.freedmem = 0;
	c->stats.freedobj = 0;
	c->stats.markedobj = 0;
	c->stats.markedmem = 0;

#if 0
	/* We held the allocator lock since cond_wait returned.  No
	 * need to aquire it again.  We rely on the scheduler to keep
	 * other threads in the process stopped.
	 */
	jmutex_lock(&c->lock);
	/* disable the mutator to protect colour lists */
#endif
	STOPWORLD();
	if (Kaffe_JavaVMArgs[0].enableVerboseGC) {
		dump_totals(c, "starting gc");
	}
	/* measure time */
	startTiming(&c->gc_time, "gctime-scan");

	//clean_self_entries(c);

	/* Walk all objects on the finalizer list.  By definition
	   these are not cross heap refs */
	for (unit = c->lists[finalise].cnext;
	     unit != &c->lists[finalise]; unit = nunit) {
		nunit = unit->cnext;
		GC_markObject(c, UTOMEM(unit), UTOMEM(unit));
	}

	/* Merge already finalised objects into the white list for
	   immediate collection */
	if (c->lists[finalised].cnext != c->lists + finalised) {
		c->lists[finalised].cnext->cprev = c->lists[white].cprev;
		c->lists[white].cprev->cnext = c->lists[finalised].cnext;
		c->lists[finalised].cprev->cnext = &c->lists[white];
		c->lists[white].cprev = c->lists[finalised].cprev;
		c->lists[finalised].cprev = c->lists[finalised].cnext
			= &c->lists[finalised];
	}

	GC_walk_refs(&c->collector);
}

/*
 * Finish off the GC process.  Any unreached (white) objects are moved
 * for finalising and the finaliser woken.
 * The reached (black) objects are moved onto the now empty white list
 * and recoloured white.
 */
void
finishGC(gc_collector *c, int *pILockRoot)
{
	gc_unit* unit;
	gc_block* info;
	int idx;
#if 0
	int totalblacksize = 0;
#endif
	
	/* There shouldn't be any grey objects at this point */
	assert(c->lists[grey].cnext == &c->lists[grey]);

	/* 
	 * Any white objects should now be freed, but we cannot call
	 * gc_heap_free here because we might block in gc_heap_free, 
	 * which would leave the white list unprotected.
	 * So we move them to a 'mustfree' list from where we'll pull them
	 * off later.
	 *
	 * We could just switch the list heads and rename the colors, but
	 * we must also unhash the exit points --- so we have to find them
	 * by iterating thru the list.
	 */
	while (c->lists[white].cnext != &c->lists[white]) {
		unit = c->lists[white].cnext;
		UREMOVELIST(unit);

		info = GCMEM2BLOCK(unit);
		idx = GCMEM2IDX(info, unit);

		assert(GC_GET_COLOUR(info, idx) == GC_COLOUR_WHITE);
		assert(GC_GET_STATE(info, idx) == GC_STATE_NORMAL);
		c->stats.freedmem += GCBLOCKSIZE(info);
		c->stats.freedobj += 1;

		UAPPENDLIST(c->lists[mustfree], unit);
		OBJECTSTATSREMOVE(unit);
	}

	/* 
	 * Now move the black objects back to the white queue for next time.
	 * Note that all objects that were eligible for finalization are now
	 * black - this is so because we marked and then walked them.
	 * We recognize them by their "INFINALIZE" state, however, and put
	 * them on the finalise list.
	 */
	while (c->lists[black].cnext != &c->lists[black]) {
		unit = c->lists[black].cnext;
		UREMOVELIST(unit);

		info = GCMEM2BLOCK(unit);
		idx = GCMEM2IDX(info, unit);
#if 0
		totalblacksize += GCBLOCKSIZE(info);
#endif

		assert(GC_GET_COLOUR(info, idx) == GC_COLOUR_BLACK);

		if (GC_GET_STATE(info, idx) == GC_STATE_INFINALIZE) {
			c->stats.finalmem += GCBLOCKSIZE(info);
			c->stats.finalobj += 1;
			UAPPENDLIST(c->lists[finalise], unit);
			if( isLeak(UTOMEM(unit)) )
			{
				unregisterLeak(UTOMEM(unit));
			}
		}
		else {
			UAPPENDLIST(c->lists[white], unit);
			if( isLeak(UTOMEM(unit)) )
			{
				unregisterLeak(UTOMEM(unit));
			}
		}
		GC_SET_COLOUR(info, idx, GC_COLOUR_WHITE);
	}
#if 0 && defined(DEBUG)
	dprintf("total unfragmented size of black objects in collector %p is %d\n", c, totalblacksize);
#endif /* DEBUG */

	/* this is where we'll stop locking out other threads 
	 * measure gc time until here.  This is not quite accurate, as
	 * it excludes the time to sweep objects, but lacking
	 * per-thread timing it's a reasonable thing to do.
	 */
	stopTiming(&c->gc_time);

	/* 
	 * Now that all lists that the mutator manipulates are in a
	 * consistent state, we can reenable the mutator here 
	 */
#if 1
	RESUMEWORLD();
#endif

	/* 
	 * Now free the objects.  We can block here since we're the only
	 * thread manipulating the "mustfree" list.
	 */
	DBG(GCHEAPS, dprintf("%p: sweeping\n", c));
	startTiming(&c->sweep_time, "gctime-sweep");

	while (c->lists[mustfree].cnext != &c->lists[mustfree]) {
		int bsz;
		destroy_func_t destroy;
		unit = c->lists[mustfree].cnext;

		/* invoke destroy function before freeing the object */
		info = GCMEM2BLOCK(unit);
		idx = GCMEM2IDX(info, unit);
		bsz = GCBLOCKSIZE(info);
		
		destroy = gcFunctions[GC_GET_FUNCS(info,idx)].destroy;
		if (destroy != 0) {
			/* Destroy methods should be uncommon, but
			 * they may call gcFree.
			 */
			_unlockMutex(&c->lock, pILockRoot);
			destroy(&c->collector, UTOMEM(unit));
			_lockMutex(&c->lock, pILockRoot);
		}

		UREMOVELIST(unit);
		addToCounter(&c->gcablemem, "gcmem-gcable objects", 1, 
			-((jlong)GCBLOCKSIZE(info)));

		c->track[GC_GET_FUNCS(info, idx)].freeCount++;
		c->track[GC_GET_FUNCS(info, idx)].usage -= bsz;

#if 0
		printf("%p: freeing %s@%p\n",
		       c,
		       describeObject(UTOMEM(unit)),
		       UTOMEM(unit));
#endif
		if (GC_GET_FUNCS(info,idx) == GC_ALLOC_FINALIZEOBJECT) {
			DBG(GCFIN,
			    dprintf("%p: freeing %s@%p\n", c,
				    describeObject(UTOMEM(unit)),
				    UTOMEM(unit)));
		}
		if( isLeak(UTOMEM(unit)) )
		{
			unregisterLeak(UTOMEM(unit));
		}
		//assert(idx >= 0 && idx <= GC_ALLOC_MAX_INDEX);
		gc_heap_free(c, unit);
	}
	stopTiming(&c->sweep_time);
}

/*
 * The finaliser sits in a loop waiting to finalise objects.  When a
 * new finalised list is available, it is woken by the GC and finalises
 * the objects in turn.  An object is only finalised once after which
 * it is deleted.
 */
void
finaliserMan(void* arg)
{
	gc_block* info;
	gc_unit* unit;
	int idx;
	gc_collector *c = (gc_collector*)arg;
	int iLockRoot;
	
	/* Cancelation is already disabled */
	lockMutex(&c->finman);

	for (;;) {

		c->finman.finalRunning = false;
		while (c->finman.finalRunning == false) {
			if (jthread_killed()) {
				/* Don't allocate anything while heap
				 * is being shut down.
				 */
				unlockMutex(&c->finman);
				jthread_exit();
			}
			waitCond(&c->finman, 0);
		}
		assert(c->finman.finalRunning == true);

		lock_collector(c);
		while (c->lists[finalise].cnext != &c->lists[finalise]) {
			unit = c->lists[finalise].cnext;
			UREMOVELIST(unit);
			UAPPENDLIST(c->lists[finalised], unit);

			info = GCMEM2BLOCK(unit);
			idx = GCMEM2IDX(info, unit);
			c->stats.finalmem -= GCBLOCKSIZE(info);
			c->stats.finalobj -= 1;

			assert(GC_GET_STATE(info,idx) == GC_STATE_INFINALIZE);
			/* Objects are only finalised once */
			GC_SET_STATE(info, idx, GC_STATE_FINALIZED);

			DBG(GCFIN,
			    dprintf("%p: finalizing %s@%p\n",
				    c, describeObject(UTOMEM(unit)),
				    UTOMEM(unit)));

			GC_SET_COLOUR(info, idx, GC_COLOUR_WHITE);
			unlock_collector(c);

			/* Call finaliser */
			unlockMutex(&c->finman);

			(*gcFunctions[GC_GET_FUNCS(info,idx)].final)
				(&c->collector, UTOMEM(unit));

			lockMutex(&c->finman);
			lock_collector(c);
			DBG(GCFIN,
			    dprintf("%p: finalized %s@%p\n",
				    c, describeObject(UTOMEM(unit)),
				    UTOMEM(unit)));
		}
		unlock_collector(c);

		/* Wake up anyone waiting for the finalizer to finish */
		broadcastCond(&c->finman);
	}
}

void gcFinalizeObject(Collector *gcif, void *obj)
{
	gc_collector *c = (void *)gcif;
	gc_block *info;
	gc_unit *unit;
	int idx;

	unit = UTOUNIT(obj);
	info = GCMEM2BLOCK(unit);
	idx = GCMEM2IDX(info, unit);
	(*gcFunctions[GC_GET_FUNCS(info,idx)].final)(&c->collector,
						     UTOMEM(unit));
}

/*
 * Explicity invoke the garbage collector and wait for it to complete.
 */
static
void
gcInvokeGC(Collector* gcif, int mustgc)
{
	gc_collector *c = (void *) gcif;
	int iLockRoot;

	lockMutex(&c->gcman);
	if (c != &gc_metal_collector && 0) {
	DBG(GCHEAPS,
	    /* XXX: Collect kernel heap to catch missing entry points: */
	    dprintf("About to gc in %p, running gc in kernel first\n", c);
	    lockMutex(&gc_metal_collector.gcman);
	    assert(!gc_metal_collector.gcman.gcRunning);
	    gc_metal_collector.gcman.gcRunning = 1;
	    signalCond(&gc_metal_collector.gcman);
	    while (gc_metal_collector.gcman.gcRunning)
	    waitCond(&gc_metal_collector.gcman, 0);
	    unlockMutex(&gc_metal_collector.gcman);

	    dprintf("\ngc in kernel done, commencing gc in %p\n", c);
	    );
	}
	c->gcman.gcRunning = 1;
	signalCond(&c->gcman);
	waitCond(&c->gcman, 0);
	unlockMutex(&c->gcman);
}

/*
 * GC and invoke the finalizer.  Used to run finalizers on exit.
 */
static
void
gcInvokeFinalizer(Collector* gcif)
{
	gc_collector *c = (void *) gcif;
	int iLockRoot;
	
	/* First invoke the GC */
	GC_invoke(gcif, 1);

	/* Run the finalizer (if might already be running as a result of
	 * the GC)
	 */
	lockMutex(&c->finman);
	if (c->finman.finalRunning == false) {
		c->finman.finalRunning = true;
		signalCond(&c->finman);
	}
	waitCond(&c->finman, 0);
	unlockMutex(&c->finman);
}

/*
 * Allocate a new object.  The object is attached to the white queue.
 */

void throwOutOfMemory(void) __NORETURN__;

/*
 * Internal malloc routine, called with allocation lock held */
static void*
gcMalloc1(gc_collector* c, size_t size, int fidx,
	  struct _errorInfo *einfo)
{
	gc_block* info;
	gc_unit* unit;
	void * volatile mem;	/* needed on SGI, see comment below */
	int i;
	size_t bsz;

	assert(fidx < nrTypes && size != 0);
	unit = gc_heap_malloc(c, size + sizeof(gc_unit));

	/* keep pointer to object */
	mem = UTOMEM(unit);
	if (unit == 0) {
		if (Kaffe_JavaVMArgs[0].enableVerboseGC) {
			dump_totals(c, "gcMalloc(%d, %d) failed: error=`%s'",
				    size, fidx, gc_alloc_failure);
			dump_totals(&gc_metal_collector, "system heap stats");
#if defined(DEBUG)
			gc_prim_map();
#endif
		}
		return 0;
	}

	assert(fidx >= 0 && fidx < GC_ALLOC_MAX_INDEX);

	info = GCMEM2BLOCK(mem);
	i = GCMEM2IDX(info, unit);

	bsz = GCBLOCKSIZE(info);
	c->stats.totalmem += bsz;
	c->stats.totalobj += 1;
	c->stats.allocmem += bsz;
	c->stats.allocobj += 1;

	c->track[fidx].allocCount++;
	c->track[fidx].usage += bsz;
	c->track[fidx].peakUsage = MAX(c->track[fidx].peakUsage,
				       c->track[fidx].usage);
	
	GC_SET_FUNCS(info, i, fidx);

	OBJECTSTATSADD(unit);

	/* Determine whether we need to finalise or not */
	if (gcFunctions[fidx].final == GC_OBJECT_NORMAL
	    || gcFunctions[fidx].final == GC_OBJECT_FIXED) {
		GC_SET_STATE(info, i, GC_STATE_NORMAL);
	}
	else {
		GC_SET_STATE(info, i, GC_STATE_NEEDFINALIZE);
	}

	/* If object is fixed, we give it the fixed colour and do not
	 * attach it to any lists.  This object is not part of the GC
	 * regime and must be freed explicitly.
	 */
	if (gcFunctions[fidx].final == GC_OBJECT_FIXED) {
		addToCounter(&c->fixedmem, "gcmem-fixed objects", 1, bsz);
		GC_SET_COLOUR(info, i, GC_COLOUR_FIXED);
		// dprintf("\nFIXED ALLOC %p %d bytes ", mem, bsz);
		// fflush(stderr);
	} else {
		addToCounter(&c->gcablemem, "gcmem-gcable objects", 1, bsz);
		/*
		 * Note that as soon as we put the object on the white list,
		 * the gc might come along and free the object if it can't
		 * find any references to it.  This is why we need to keep
		 * a reference in `mem'.  Note that keeping a reference in
		 * `unit' will not do because markObject performs a UTOUNIT()!
		 * In addition, on some architectures (SGI), we must tell the
		 * compiler to not delay computing mem by defining it volatile.
		 */
		GC_SET_COLOUR(info, i, GC_COLOUR_WHITE);
		UAPPENDLIST(c->lists[white], unit);
	}
	return (mem);
}

static void *
gcMalloc(Collector *gcif, size_t size, int fidx)
{
	void *mem;
	gc_collector *c = (void *) gcif;
	int iLockRoot;
	
	lock_collector(c);
	assert(!c->dying);
	mem = gcMalloc1((gc_collector *) gcif, size, fidx, 0);
	/*
	 * The first thread to notice this pointer become null is
	 * responsible for allocating it.  Before calling newObject,
	 * we must be sure to relinquish our lock and avoid
	 * recursion.  Reserving pages in the primitive freelist is
	 * only slightly less complicated:  gc_primitive_reserve may
	 * trigger a collection, which means giving up the allocator
	 * lock.  We must be the only thread inside
	 * gc_primitive_reserve. 
	 */
	if (!(c->outOfMem || c->reserve) && !c->outOfMem_allocator
	    && c->gcman.thread) {
		c->outOfMem_allocator = jthread_current();
		if (!c->reserve) {
			c->reserve = gc_primitive_reserve(c);
		}
		if (c->outOfMem) {
			c->outOfMem_allocator = 0;
		}
	}
	unlock_collector(c);

	if( !mem )
	{
		teamLogEntry("gc",
			     LEA_String, "Warning", "Out of memory", -1,
			     LEA_Integer, "used", c->client->used,
			     LEA_Integer, "max", c->client->max,
			     LEA_Integer, "request-size", size,
			     LEA_String, "type",
			     gcFunctions[fidx].description, -1,
			     LEA_TAG_DONE);
	}
	
	/* If we threw an outOfMemory exception and recovered, we
	 * should preallocate a new one.
	 *
	 * Don't try to allocate one while bootstrapping.  This should
	 * happen with c->lock held, there should be an internal
	 * version of gcAddRef.
	 */
	if (!c->outOfMem && c->gcman.thread
	    && c->outOfMem_allocator == jthread_current()) {
		c->outOfMem = OOM_ALLOCATING;
		/* XXX Why are we visiting our own team again? */
		if( switchToTeam(c->client->client.owner, 0) )
		{
			/*
			 * NOTE: OutOfMemoryError Calls
			 * execute_java_constructor
			 */
			c->outOfMem = OutOfMemoryError;
			returnFromTeam();
		}
		else
		{
			c->outOfMem = 0;
		}
		
		c->outOfMem_allocator = 0;
		if (c->outOfMem && !GC_add_ref(&c->collector, c->outOfMem)) {
			/* assume it will be gced soon */
			c->outOfMem = 0;
		}
	}
	return (mem);
}

Hjava_lang_Throwable *
gcThrowOOM(Collector *gcif)
{
	gc_collector *c = (void *) gcif;
	Hjava_lang_Throwable *ret = 0;
	void *reffed;
	int iLockRoot;

	/*
	 * Make sure we are the only thread to use this exception
	 * object.
	 */
	lock_collector(c);

	/* get and clear preallocated OOM object */
	ret = c->outOfMem;
	reffed = c->outOfMem;
	c->outOfMem = 0;

	/* We try allocating reserved pages before we allocate the
	 * outOfMemory error.  We can use some or all of the reserved
	 * pages to actually grab an error.
	 */
	if (c->reserve) {
		jthread_t jt = jthread_current();
		memoryStack *ms = jthread_get_data(jt)->journal;
		
		/*
		 * XXX: how can we be sure that we get to use the memory
		 * we just put in the free pool?
		 */
		if( c->primitiveStack )
		{
			jthread_get_data(jt)->journal = c->primitiveStack;
		}
		gc_primitive_free(c, c->reserve);
		jthread_get_data(jt)->journal = ms;
		c->reserve = 0;
		if (!ret || ret == OOM_ALLOCATING) {
			unlock_collector(c);
			ret = OutOfMemoryError;
			lock_collector(c);
		}
	}
	if (Kaffe_JavaVMArgs[0].enableVerboseGC) {
		dump_totals(c, "throwing outOfMem  %p", ret);
	}
	if (ret == OOM_ALLOCATING || ret == 0) {
		unlock_collector(c);
		return 0;
	}
	unlock_collector(c);
	if (reffed) {
		GC_rm_ref(gcif, reffed);
	}
	return (ret);
}

/*
 * Reallocate an object.
 */
static
void*
gcRealloc(Collector* gcif, void* mem, size_t size, int fidx)
{
	gc_collector *c = (void *) gcif;

	gc_block* info;
	int idx;
	void* newmem;
	gc_unit* unit;
	int osize;
	int iLockRoot;
	
	assert(gcFunctions[fidx].final == GC_OBJECT_FIXED);

	/* If nothing to realloc from, just allocate */
	if (mem == NULL) {
		return (gcMalloc(gcif, size, fidx));
	}

	lock_collector(c);
	unit = UTOUNIT(mem);
	info = GCMEM2BLOCK(unit);
	idx = GCMEM2IDX(info, unit);
	/* Size of the old memory block shouldn't include the gc_unit header */
	osize = GCBLOCKSIZE(info) - sizeof(gc_unit);

	/* Can only handled fixed objects at the moment */
	assert(GC_GET_COLOUR(info, idx) == GC_COLOUR_FIXED);
	info = 0;
	unlock_collector(c);

	/* XXX We don't shrink the block. */
	
	/* If we'll fit into the current space, just send it back */
	if (osize >= size) {
		return (mem);
	}

	/* Allocate new memory, copy data, and free the old */
	if( (newmem = gcMalloc(gcif, size, fidx)) )
	{
		memcpy(newmem, mem, osize);
		gcFree(gcif, mem);
	}
	return (newmem);
}

/*
 * Explicitly free an object.
 */
static
void
gcFree(Collector* gcif, void* mem)
{
	gc_collector *c = (void *) gcif;

	gc_block* info;
	int idx;
	gc_unit* unit;

	if (mem != 0) {
		int iLockRoot;
		
		lock_collector(c);
		unit = UTOUNIT(mem);
		info = GCMEM2BLOCK(unit);
		idx = GCMEM2IDX(info, unit);

		if (GC_GET_COLOUR(info, idx) == GC_COLOUR_FIXED) {
			size_t sz = GCBLOCKSIZE(info);

			c->track[GC_GET_FUNCS(info, idx)].freeCount++;
			c->track[GC_GET_FUNCS(info, idx)].usage -= sz;
			
			//assert(idx >= 0 && idx <= GC_ALLOC_MAX_INDEX);
			// dprintf("\nFIXED FREE %p %d bytes ", mem, sz);
			// fflush(stderr);
			OBJECTSTATSREMOVE(unit);

			/* Keep the stats correct */
			c->stats.totalmem -= sz;
			c->stats.totalobj -= 1;
			addToCounter(&c->fixedmem, "gcmem-fixed objects", 1, -(jlong)sz);

			gc_heap_free(c, unit);
		}
		else {
		    printf("Attempt to explicitly free nonfixed object");
		    assert(!!!"Attempt to explicitly free nonfixed object");
		}
		unlock_collector(c);
	}
}

struct Hjava_lang_Object *
edu_utah_janosvm_resources_GCPolicy_currentPolicy(void)
{
	return ((gc_collector *)currentCollector())->policy.obj;
}

jbool
edu_utah_janosvm_resources_GCPolicy_startThreads(struct Hjava_lang_Object *obj)
{
	gc_collector *c = (gc_collector *)currentCollector();
	struct _errorInfo einfo;
	jbool retval = 1;
	char gc_name[19+2];
	char fin_name[19+9];

	if( c->finman.fin_thread != 0 )
		return true;
	
	sprintf(gc_name, "gc%p", c);
	sprintf(fin_name, "finaliser%p", c);
	
	/* allocate gc threads out of my own heap */
	
	/*
	 * Start gc and finalizer threads.  All
	 * processes have the same priority, so the
	 * all gcs run at a higher priority than any
	 * finalizer.  We don't want someone else's
	 * finalizers obtaining new references to our
	 * objects while we are doing a gc.
	 */
	c->finman.fin_thread = createDaemon(&finaliserMan,
					    fin_name, c,
					    THREAD_MAXPRIO-1,
					    FINALIZERSTACKSIZE,
					    &einfo);
	if (c->finman.fin_thread == 0) {
		retval = 1;
		goto bad;
	}
	
	c->gcman.thread = createDaemon(&gcMan, gc_name, c,
				       THREAD_MAXPRIO, 
				       GCSTACKSIZE,
				       &einfo);
	if (c->gcman.thread == 0) {
		retval = 1;
		goto bad;
	}
	
	/* pin down these threads, we don't them
	 * somehow gc'ed before the heap dies.
	 */
	if (!GC_add_ref((Collector *)c, c->finman.fin_thread)) {
		retval = 1;
		goto bad;
	}
	
	if (!GC_add_ref((Collector *)c, c->gcman.thread)) {
		GC_rm_ref((Collector *)c, c->finman.fin_thread);
		retval = 1;
		goto bad;
	}
 bad:
	return (retval);
}

/*
 * Start gc threads, which enable collection
 */
static 
int
/* ARGSUSED */
gcEnable(Collector* gcif)
{
	gc_collector *c = (void *) gcif;
	int retval;

	c->primitiveStack = createMemoryStack(0, 2048);
	
	if( c->policy.name )
	{
		char *old_name = c->policy.name;
		jlong max = 0, trigger;

		max = (jlong)c->client->max;
		c->policy.obj = execute_java_constructor(c->policy.name,
							 0,
							 0,
							 "(J)V",
							 max);
		GC_add_ref(gcif, c->policy.obj);
		retval = do_execute_java_method(c->policy.obj,
						"enable",
						"()Z", 0, 0).j;
		trigger = do_execute_java_method(c->policy.obj,
						 "initialTrigger",
						 "()J", 0, 0).j;
		/* XXX Just to make sure the sucker is jitted */
		(void)do_execute_java_method(c->policy.obj,
					     "newTrigger",
					     "(JJ)J",
					     0,
					     0,
					     (jlong)c->stats.
					     totalmem,
					     (jlong)0 /* XXX */);
		c->trigger = trigger;
		c->client->gcTrigger = trigger;
		/* Use gcFree since we may be proxied */
		gcFree(gcif, old_name);
	}
	else
	{
		retval = edu_utah_janosvm_resources_GCPolicy_startThreads(0);
	}
	return( retval );
}

#if defined(SUPPORT_VERBOSEMEM)

/* --------------------------------------------------------------------- */
/* The following functions are strictly for statistics gathering	 */

static
void
objectStatsChange(gc_unit* unit, int diff)
{
	gc_block* info;
	int idx;

	info = GCMEM2BLOCK(unit);
	idx = GC_GET_FUNCS(info, GCMEM2IDX(info, unit));

	assert(idx >= 0 && idx < nrTypes);
	gcFunctions[idx].nr += diff * 1;
	gcFunctions[idx].mem += diff * GCBLOCKSIZE(info);
}

static void
objectStatsPrint(void)
{
	int cnt = 0;

	dprintf("Memory statistics:\n");
	dprintf("------------------\n");

	while (cnt < nrTypes) {
		dprintf("%14.14s: Nr %6d  Mem %6dK",
			gcFunctions[cnt].description, 
			gcFunctions[cnt].nr, 
			gcFunctions[cnt].mem/1024);
		if (++cnt % 2 != 0) {
			dprintf("   ");
		} else {
			dprintf("\n");
		}
	}

	if (cnt % 2 != 0) {
		dprintf("\n");
	}
}
#endif

/*
 * Add a persistent reference to an object.
 */
static bool
Gc_add_ref(Collector *gcif, const void* mem)
{
	uint32 idx;
	refObject* obj;
	gc_collector *c = (void *) gcif;
	int iLockRoot;

	lock_collector(c);
	assert(IS_A_HEAP_POINTER(mem));
	assert(GCMEM2BLOCK(mem)->collector == c);
	idx = REFOBJHASH(mem);
	for (obj = c->hash[idx]; obj != 0; obj = obj->next) {
		/* Found it - just increase reference */
		if (obj->mem == mem) {
			obj->ref++;
	    		DBG(GCREFS,
			    dprintf("%p: add_ref count %s@%p -> %d\n",
			    c, describeObject(obj->mem), obj->mem, obj->ref);
			)
			goto done;
		}
	}
	/* Not found - create a new one */
	obj = (refObject*)gcMalloc1(c, sizeof(refObject), GC_ALLOC_REF, 0);
	if (!obj) {
		unlock_collector(c);
		return false;
	}
	obj->mem = mem;
	obj->ref = 1;
	obj->next = c->hash[idx];
	c->hash[idx] = obj;
	c->n_refs++;
	DBG(GCREFS,
	    dprintf("%p: creating first reference to %s@%p\n",
		    c, describeObject(obj->mem), obj->mem));
done:
	unlock_collector(c);
	return true;
}

/*
 * Remove a persistent reference to an object.  If the count becomes
 * zero then the reference is removed.
 */
static bool
Gc_rm_ref(Collector *gcif, const void* mem)
{
	uint32 idx;
	refObject** objp;
	refObject* obj = 0;
	gc_collector *c = (void *) gcif;
	int iLockRoot;
	
	lock_collector(c);
	idx = REFOBJHASH(mem);
	for (objp = &c->hash[idx]; *objp != 0; objp = &obj->next) {
		obj = *objp;
		/* Found it - just decrease reference */
		if (obj->mem == mem) {
			obj->ref--;
			if (obj->ref == 0) {
				*objp = obj->next;
				c->n_refs--;
				break;
			}
	    		DBG(GCREFS,
			    dprintf("%p: rm_ref count %s@%p -> %d\n",
			    c, describeObject(obj->mem), obj->mem, obj->ref);
			)
			unlock_collector(c);
			return (true);
		}
	}

	/* either it is not found or we need to free it */
	unlock_collector(c);
	if (obj && obj->mem == mem) {
		DBG(GCREFS,
		    dprintf("%p: last reference to %s@%p gone\n",
			    c, describeObject(obj->mem), obj->mem));
		gcFree(gcif, obj);
		return (true);
	}
	/* Not found!! */
	return (false);
}

static bool
Gc_find_ref(Collector *gcif, const void *mem)
{
	gc_collector *c = (void *) gcif;
	bool retval = false;
	refObject *curr;
	uint32 idx;
	int iLockRoot;
	
	lock_collector(c);
	idx = REFOBJHASH(mem);
	curr = c->hash[idx];
	while( curr && !retval )
	{
		if( curr->mem == mem )
		{
			retval = true;
		}
		curr = curr->next;
	}
	unlock_collector(c);
	return( retval );
}

/*
 * Walk stacks of all threads 
 */
static void
gc_walk_stacks(void *_c, jthread_t jtid, void *mem)
{
	gc_collector *c = _c;

	/* no need to walk the thread of the collector thread */
	/* XXX catch all collector threads here and don't walk */
	if (mem == c->gcman.thread) {
		return;
	}

	/* We only have to walk the finalizer thread when it is
	 * executing java code.
	 */
	if (mem == c->finman.fin_thread && c->finman.finalRunning == false) {
		return;
	}

	if( mem )
	{
		if (c == GCMEM2BLOCK(mem)->collector) {
			GC_markObject(_c, mem, mem);
		}
	}

	if( jtid )
	{
#if defined(STACK_GROWS_UP)
		void *from;
		unsigned len;

		/* Stack == [USED ... FREE] */
		if (jthread_extract_stack(jtid, &from, &len, NULL, NULL)) {
			GC_walkConservative((Collector *)c, from, len,
					    from);
		}
#else
		void *from;
		void *stackBase;
		unsigned len;

		/* Stack == [FREE ... USED] */
		if (jthread_extract_stack(jtid, &from, &len, &stackBase, NULL)) {
			GC_walkConservative((Collector *)c, from, len,
					    stackBase
					    );
		}
#endif

	
	}
}

/*
 * Walk the set of registered root references.  
 * This is invoked at the beginning of each GC cycle. 
 */
static void
Gc_walk_refs(gc_collector *collector)
{
        int i;
	refObject* robj;

DBG(GCWALK,
	dprintf("Walking gc roots...\n");
    )

        /* Walk the referenced objects */
        for (i = 0; i < REFOBJHASHSZ; i++) {
                for (robj = collector->hash[i]; robj != 0; robj = robj->next) {
			assert(GCMEM2BLOCK(robj)->collector == collector);
			assert(IS_A_HEAP_POINTER(robj->mem));
			
                        GC_markObject(&collector->collector,
				      robj->mem,
				      /* from */ 0);
                }
        }
}

/* Walk local roots, kernel and user heaps */
/* Called at the end of 'startGC' */
static void
gc_walk_local_roots(Collector *gcif)
{
	gc_collector *c = (void *) gcif;

	/*
	 * Walk my references and my thread stacks.
	 */
	Gc_walk_refs(c);
	jthread_walkLiveThreads(c, gc_walk_stacks);
}

static void _gc_noop(void) { }
#define gc_noop (void *) _gc_noop

hashtab_t       
gcGetStringTable(Collector *gcif)
{
	gc_collector *c = (gc_collector *) gcif;
	return (c->stringTable);
}

void       
gcSetStringTable(Collector *gcif, hashtab_t h)
{
	gc_collector *c = (gc_collector *) gcif;
	c->stringTable = h;
}

struct Hjava_lang_Object *
gcCopyObject(Collector *gcif, struct Hjava_lang_Object *obj)
{
	/* gc_collector *c = (gc_collector *)gcif; */
	struct Hjava_lang_Object *retval;
	struct Hjava_lang_Class *cl;
	gc_block *info;
	gc_unit *unit;
	char type;

	if( obj == 0 )
		return 0;
	unit = UTOUNIT(obj);
	info = GCMEM2BLOCK(unit);
	type = GC_GET_FUNCS(info, GCMEM2IDX(info, unit));
	switch( type )
	{
	case GC_ALLOC_JAVASTRING:
	case GC_ALLOC_NORMALOBJECT:
	case GC_ALLOC_REFARRAY:
	case GC_ALLOC_FINALIZEOBJECT:
		cl = obj->dtable->class;
		if( (retval = GC_malloc(gcif,
					CLASS_FSIZE(cl),
					cl->alloc_type)) )
		{
			int lpc, curr, nbits;
			int *layout;
			int8 *mem, *cp_mem;
			
			memcpy(retval, obj, CLASS_FSIZE(cl));
			nbits = CLASS_FSIZE(cl)/ALIGNMENTOF_VOIDP;
			layout = cl->gc_layout;
			mem = (int8 *)obj;
			cp_mem = (int8 *)retval;
			while( retval && (nbits > 0) )
			{
				curr = *layout++;
				for( lpc = 0; lpc < BITMAP_BPI; lpc++ )
				{
					if( curr == 0 )
					{
						mem += (BITMAP_BPI - lpc) *
							ALIGNMENTOF_VOIDP;
						break;
					}
					if( curr < 0 )
					{
						struct Hjava_lang_Object *sub,
							*newsub;
						
						sub = (struct Hjava_lang_Object *)
							(*(void **)mem);
						newsub = gcCopyObject(gcif, sub);
						*((void **)cp_mem) = newsub;
						if( sub && !newsub )
						{
							retval = 0;
						}
					}
					curr <<= 1;
					mem += ALIGNMENTOF_VOIDP;
					cp_mem += ALIGNMENTOF_VOIDP;
				}
				nbits -= BITMAP_BPI;
			}
		}
		break;
		/* XXX Handle others ? */
	default:
		if( (retval = GC_malloc(gcif, info->size, type)) )
		{
			memcpy(retval, obj, info->size);
		}
		break;
	}
	return( retval );
}

static void
gcPrintInfo(Collector *gcif)
{
	gc_collector *c = (gc_collector *) gcif;
#if defined(DEBUG)
	dprintf("Information for heap %x\n", c);
#define PS(V, C) do { dprintf(#C": "#V" = %dKB\n", V/1024); } while (0)
#define PB(V, C) do { dprintf(#C": "#V" = 0x%x\n", V); } while (0)
	PS(c->client->max, maximum size to which heap should grow);
	PS(c->client->used, memory allocated to this heap);
	PS(c->client->gcTrigger, initial size to trigger gc);
	PS(c->stats.totalmem, actual bytes in use by the application);
	PS(c->stats.allocmem, bytes in alloced objects since last gc);
	PS(c->stats.freedmem, bytes freed in last gc);
	PS(c->stats.markedmem, bytes marked in last gc);
	PS(c->stats.finalmem, bytes finalized in last gc);
	PS(c->local, memory allocated to this heap);
	PS(c->trigger, gc trigger amount);
	PB(c->n_refs, number of entries);
#undef PS
#undef PB

#else
	fprintf(stderr, 
	    "No information about collector %p available, compile with DEBUG\n",
	    c);
#endif
}

static void
gcInit(Collector *gcif)
{
}
     
/*
 * vtable for object implementing the collector interface.
 */
static struct GarbageCollectorInterface_Ops system_Ops = {
	0,		/* reserved */
	0,		/* reserved */
	0,		/* reserved */
	Gc_add_ref,
	Gc_rm_ref,
	gc_walk_local_roots,
	gcMalloc,
	gcRealloc,
	gcFree,
	gcInvokeGC,
	gcInvokeFinalizer,
	gcInit,
	gcEnable,
	gcMarkAddress,
	gcMarkObject,
	gcFinalizeObject,
	gcGetObjectSize,
	gcGetObjectDescription,
	gcGetObjectIndex,
	gcGetObjectBase,
	gcWalkMemory,
	gcWalkConservative,
	gcRegisterFixedTypeByIndex,
	gcRegisterGcTypeByIndex,
	gcThrowOOM,
	gcGetStringTable,
	gcSetStringTable,
	gcCopyObject,
	gcPrintInfo
};

static struct GarbageCollectorInterface_Ops local_Ops = {
	0,		/* reserved */
	0,		/* reserved */
	0,		/* reserved */
	Gc_add_ref,
	Gc_rm_ref,
	gc_walk_local_roots,
	gcMalloc,
	gcRealloc,
	gcFree,
	gcInvokeGC,
	gcInvokeFinalizer,
	gcInit,
	gcEnable,
	gcMarkAddress,
	gcMarkObject,
	gcFinalizeObject,
	gcGetObjectSize,
	gcGetObjectDescription,
	gcGetObjectIndex,
	gcGetObjectBase,
	gcWalkMemory,
	gcWalkConservative,
	gc_noop,
	gc_noop,
	gcThrowOOM,
	gcGetStringTable,
	gcSetStringTable,
	gcCopyObject,
	gcPrintInfo
};

gc_collector gc_metal_collector = {
	{ &system_Ops }
};

static Collector *
init_collector(gc_collector *c, memResourceClient *mrc, char *policyClass)
{
	int i;
	int iLockRoot;
	
	c->client = mrc;
	mrc->collector = c;
	c->trigger = mrc->gcTrigger;
	gc_mem_init(c);

	for (i = 0; i < sizeof(c->lists)/sizeof(gcList); i++) {
		URESETLIST(c->lists[i]);
	}

	if( policyClass )
	{
		c->policy.name = gcMalloc((Collector *)c,
					  strlen(policyClass) + 1,
					  GC_ALLOC_FIXED);
		strcpy(c->policy.name, policyClass);
	}

	lockStaticMutex(&collectorListLock);
	c->next = defined_collectors;
	defined_collectors = c;
	unlockStaticMutex(&collectorListLock);

	return &c->collector;
}

Collector *
systemCollector(int threshhold, int limit)
{
	memResourceClient *mrc;
	
	gc_heap_initialize();
	gc_prim_initialize();

	mrc = (memResourceClient *)
		getTeamInternalResource(metalTeam, &metalMemResource.link);
	mrc->gcTrigger = threshhold;
	mrc->max = limit;
	if (Kaffe_JavaVMArgs[0].enableVerboseGC) {
		fprintf(stderr, "[heap: system collector is %p]\n", 
			&(&gc_metal_collector)->collector);
	}
	return init_collector(&gc_metal_collector, mrc, 0);
}

Collector *
newCollector(memResourceClient *mrc, char *policyClass)
{
	gc_collector *r;

	r = KMALLOC(sizeof(*r));
	r->collector.ops = &local_Ops;
	if (Kaffe_JavaVMArgs[0].enableVerboseGC) {
		fprintf(stderr,
			"[heap: new collector created at %p]\n", 
			&r->collector);
	}
	return init_collector(r, mrc, policyClass);
}

void gcUnlockMe(gc_collector *me)
{
	me->lock = 0;
}

/* Kill heap 'me', merge result into heap 'where'.
 *
 * 'where' is usually the system collector, but maybe we can code it in a
 * more general way so that maybe in the future we can merge arbitrary heaps
 * should it become useful.
 */
void
gcKillMe(gc_collector *me)
{
	gc_collector **pcol;
	int iLockRoot;
	
	/* don't do any of that if -vmdebug NOGC is given */
	if (DBGEXPR(NOGC, true, false)) {
		return;
	}
	
	if (Kaffe_JavaVMArgs[0].enableVerboseGC) {
			fprintf(stderr, 
				"<kill heap %p (limit=%p): "
				"heap %dK total %dK (%d objs)>\n",
				me,
				me->client,
				me->client->used/1024, 
				me->stats.totalmem/1024, 
				me->stats.totalobj);
	}

#if defined(KAFFE_STATS)
	statsReport();	/* report stats as set on cmdline, dont exit */
#endif
	
	assert((me->gcman.thread && me->finman.fin_thread) 
		|| (!me->gcman.thread && !me->finman.fin_thread));

	/* drop OOM object if necessary */
	if (me->outOfMem != 0) {
		GC_rm_ref(&me->collector, me->outOfMem);
		me->outOfMem = 0;
	}
	me->dying = 1;

	/* unregister counters */
	unregisterCounter(&me->gc_time);
	unregisterCounter(&me->sweep_time);
	unregisterCounter(&me->gcablemem);
	unregisterCounter(&me->fixedmem);

	gc_release_blocks(me);

	/* remove from list */
	lockStaticMutex(&collectorListLock);
	for (pcol = &defined_collectors; *pcol; pcol = &(*pcol)->next) {
		if (*pcol == me) {
			*pcol = (*pcol)->next;
		}
	}
	unlockStaticMutex(&collectorListLock);

	{
		jteam *prev;

		require(currentTeam() == me->client->client.owner);
		
		prev = quickChangeTeamReservation(kernelTeam);
		gc_free(me);
		quickChangeTeamReservation(prev);
		
		ensure(currentTeam() == prev);
		ensure(currentTeam() != kernelTeam);
	}
}

#if 0
#define WB_COUNT	1
#else
#undef WB_COUNT
#endif

#if defined(WB_COUNT)
static long long wbhits;
static long long wbhits_user_kernel;
static long long wbhits_kernel_user;
#endif

void 
reportWriteBarriers(void)
{
#if defined(WB_COUNT)
	fprintf(stderr,
		"wbhits = %qd user_kernel = %qd "
		"kernel_user = %qd kernel_shared = %qd "
		"shared_kernel = %qd user_shared = %qd\n",
		wbhits,
		wbhits_user_kernel,
		wbhits_kernel_user,
		wbhits_kernel_shared,
		wbhits_shared_kernel,
		wbhits_user_shared);
#endif
}

#if defined (__i386__) && defined(STORE_HEAP_WITH_OBJECT)
asm("
	.globl	soft_addreference
soft_addreference:
	movl    8(%esp), %eax
	testl   %eax, %eax
	je      okay
	movl	4(%esp), %ecx
	movl	4(%eax), %eax
	cmpl	%eax, 4(%ecx)
	je	okay
	jmpl 	soft_addreference_slow
okay:
	ret
");
#define SOFT_ADDREFERENCE	soft_addreference_slow
#else
#define SOFT_ADDREFERENCE	soft_addreference
#endif

static
void postSegFault(jteam *destTeam, void *destObject,
		  jteam *refTeam, void *refObject,
		  errorInfo *einfo)
{
	char destDesc[100];
	char refDesc[100];

/* XXX verbose descriptions should be DEBUG only? */
	strncpy(destDesc, describeObject(destObject), sizeof(destDesc) - 1);
	destDesc[sizeof(destDesc) - 1] = '\0';
	strncpy(refDesc, describeObject(refObject), sizeof(refDesc) - 1);
	refDesc[sizeof(refDesc) - 1] = '\0';
	postExceptionMessage(
		einfo,
		"edu.utah.janosvm.sys.SegmentationFaultError",
		"Team(`%s', %u).%s(%p).field = Team(`%s', %u).%s(%p);",
		destTeam->name, destTeam->elemID, destDesc, destObject,
		refTeam->name, refTeam->elemID, refDesc, refObject);
}

/**
 * Record a reference to an object being added.  This is where
 * JanosVM detects cross-process heap writes.
 * 
 * 'dest' is the *object* being written to.
 * 'ref' is the reference being written.
 *
 * That is "dest.<field> = ref" what is happening.
 */
int SOFT_ADDREFERENCE (void *dest, void *ref, errorInfo *ret_error)
{
	int retval = 0;

	if( ref != 0 )
	{
		struct _gc_collector *destCollector, *refCollector;
		jteam *destTeam, *refTeam;

		destCollector = GCMEM2BLOCK(dest)->collector;
		refCollector = GCMEM2BLOCK(ref)->collector;

		destTeam = destCollector->client->client.owner;
		refTeam = refCollector->client->client.owner;

		if( (destTeam == refTeam) ||
		    hasTeamDependency(destTeam, refTeam) ||
		    Gc_find_ref(((memResourceClient *)
				 getTeamInternalResource(refTeam,
					&metalMemResource.link)
					)->collector, ref) )
		{
		}
		else
		{
			errorInfo einfo, *peinfo;

			if( ret_error )
				peinfo = ret_error;
			else
				peinfo = &einfo;
			postSegFault(destTeam, dest, refTeam, ref, peinfo);
			if( !ret_error )
				throwError(peinfo);
			retval = 1;
		}
	}
	return( retval );
}

#if 0
/* XXX is this code complete toast, or what? */
/*
 * Validate an interheap reference.  Return 0 if it is valid.
 * Otherwise, fill in the provided errorInfo and return 1, or throw
 * segv if ret_error is null.
 */
int SOFT_ADDREFERENCE (void *a, void *b, errorInfo *ret_error)
{
	/*
	 * a points to b  (a -> b)
	 */
#if	1
	errorInfo _error;
	errorInfo *error;
	struct _gc_collector *xcoll, *ycoll;

#if defined(WB_COUNT)
	wbhits++;
#endif

	if (!b) return (0);

#if 0
/* No can do since "soft_addreference_slow" also has to work for non-java
 * objects...
 */
	xcoll = ((Hjava_lang_Object*)a)->heap;
	ycoll = ((Hjava_lang_Object*)b)->heap;
#endif

	xcoll = GCMEM2BLOCK(a)->collector;
	ycoll = GCMEM2BLOCK(b)->collector;

	if (xcoll != ycoll) {
		int xk, yk, u2k, k2u;

		error = ret_error ? ret_error : &_error;

		xk = xcoll == &gc_metal_collector;
		yk = ycoll == &gc_metal_collector;
		u2k = xk;
		k2u = yk;

		if (u2k || k2u) {


#if defined(WB_COUNT)
			if (k2s) {
			    wbhits_kernel_shared++;
			} else
			if (s2k) {
			    wbhits_shared_kernel++;
			} else
			if (u2k) {
			    wbhits_user_kernel++;
			} else 
			if (u2s) {
			    wbhits_user_shared++;
			} else
			if (k2u) {
			    wbhits_kernel_user++;
			}
#endif
			return 0;
		} else {
#if 0
			char adesc[100]; /* I love static buffers */

			strncpy(adesc, describeObject(a), sizeof(adesc) - 1);
			adesc[sizeof(adesc) - 1] = '\0';
			fprintf(stderr,
				"segv %s@%p(heap=%p) =-> %s@%p(heap=%p) jthr=%p\n",
				adesc,
				a,
				xcoll,
				describeObject(b),
				b,
				ycoll,
				jthread_current());
#endif
#if 0
			postExceptionMessage
				(error,
				 "kaffeos.sys.SegmentationViolationError",
				 "%s@%p(heap=%p) =-> "
				 "%s@%p(heap=%p) jthr=%p\n",
				 adesc, a, xcoll,
				 describeObject(b), b, ycoll,
				 jthread_current());
			if (!ret_error) {
				throwError(error);
			}
			return 1;
#else
			return 0;
#endif
		}
	}
	return 0;
#endif
}
#endif

Collector *
gc_get_heap(void *a)
{
	if( (((uintp)a) >= gc_heap_base)
	    && ((uintp)a < (gc_heap_base + gc_heap_range)) )
	{
		gc_block *info;

		info = GCMEM2BLOCK(a);
		if( info->firstblock )
			info = info->firstblock;
		/* This functions must work for all objects; so can't rely on
		 * 'heap' field if defined(STORE_HEAP_WITH_OBJECT)
		 */
		return (&info->collector->collector);
	}
	else
	{
		return 0;
	}
}

memResourceClient *
gc_get_resource(void *a)
{
	if( (((uintp)a) >= gc_heap_base)
	    && ((uintp)a < (gc_heap_base + gc_heap_range)) )
	{
		gc_block *info;

		info = GCMEM2BLOCK(a);
		if( info->firstblock )
			info = info->firstblock;
		if( info->collector == 0 )
			return 0;
		/* This functions must work for all objects; so can't rely on
		 * 'heap' field if defined(STORE_HEAP_WITH_OBJECT)
		 */
		return ((memResourceClient *)info->collector->client);
	}
	else
	{
		return 0;
	}
}

int gc_report(Collector *gcif, jsiMessage *jm)
{
	gc_collector *c = (gc_collector *)gcif;
	int lpc, retval = 1;
	char *pos;
	char desc[128];

	for( lpc = 0; retval && (lpc < GC_ALLOC_MAX_INDEX); lpc++ )
	{
		strcpy(desc, gcFunctions[lpc].description);
		pos = desc + strlen(desc);
		*pos = '.';
		pos++;
		strcpy(pos, "allocCount");
		retval = jsiMessageAdd(jm,
				       JMA_Integer,
				       desc,
				       c->track[lpc].allocCount,
				       JMA_TAG_DONE);
		strcpy(pos, "freeCount");
		retval = retval && jsiMessageAdd(jm,
						 JMA_Integer,
						 desc,
						 c->track[lpc].freeCount,
						 JMA_TAG_DONE);
		strcpy(pos, "peakUsage");
		retval = retval && jsiMessageAdd(jm,
						 JMA_Integer,
						 desc,
						 c->track[lpc].peakUsage,
						 JMA_TAG_DONE);
		strcpy(pos, "usage");
		retval = retval && jsiMessageAdd(jm,
						 JMA_Integer,
						 desc,
						 c->track[lpc].usage,
						 JMA_TAG_DONE);
	}
	jsiMessageAdd(jm,
		      JMA_Integer, "used", c->client->used,
		      JMA_Integer, "peak", c->client->peak,
		      JMA_Integer, "max", c->client->max,
		      JMA_TAG_DONE);
	return( retval );
}

void gc_print(Collector *gcif)
{
	gc_collector *c = (gc_collector *)gcif;
	int lpc;

	printf("                   \tallocs    frees     peak    usage\n");
	for( lpc = 0; lpc < GC_ALLOC_MAX_INDEX; lpc++ )
	{
		printf("  %18s\t%8d %8d %8d %8d\n",
		       gcFunctions[lpc].description,
		       c->track[lpc].allocCount,
		       c->track[lpc].freeCount,
		       c->track[lpc].peakUsage,
		       c->track[lpc].usage);
	}
}
