/* gc-incremental.h
 * The incremental garbage collector.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */
/*
 * This file contains all definitions relating to gc and allocation types.
 */
#ifndef __gc_incremental_h
#define __gc_incremental_h

#include "md.h"		/* make sure we pick up a #def of gc_pgbits if any */

typedef struct _gc_block {
#ifdef DEBUG
	uint32			magic;	/* Magic number */
#endif
	struct _gc_collector   *collector; /* collector used for this heap */
	struct _gc_freeobj*	free;	/* Next free sub-block */
	struct _gc_block*	next;	/* Next block in prim/small freelist */
	struct _gc_block*	nfree;	/* Next block on sub-freelist */
	void*			firstblock; 
					/* If 0, it means this block is not
					 * allocated.  Otherwise, points at
					 * the first block of the larger
					 * aggregate region to which this
					 * block belongs
					 */
	uint32			size;	/* Size of objects in this block */
	uint16			nr;	/* Nr of objects in block */
	uint16			avail;	/* Nr of objects available in block */
	uint8*			funcs;	/* Function for objects */
	uint8*			state;	/* Colour & state of objects */
	uint8*			data;	/* Address of first object in */
} gc_block;

#ifndef gc_pgsize
extern size_t gc_pgsize;
extern int gc_pgbits;
#endif

extern size_t gc_heap_base;
extern size_t gc_block_base;
extern uintp  gc_heap_range;	/* last gc-able address - gc_heap_base */

#define	GC_MAGIC		0xD0DECADE
#define GCBLOCK_LIVE		((gc_block *) -1) /* block->next when alloced*/
#define GC_BLOCKS		((gc_block *) gc_block_base)

#define	GCMEM2FREE(M)		((gc_freeobj*)(M))
#define	GCMEM2IDX(B, M)		(((uint8*)(M) - (B)->data) / (B)->size)
#define GCMEM2BLOCK(M)		(GC_BLOCKS + ((((uintp) M) - gc_heap_base) \
					      >> gc_pgbits))
#if 0
extern int gc_pgbitsX;
#define GCMEM2BLOCK1(M)		(GC_BLOCKS + ((((uintp) M) - gc_heap_base) \
					      >> gc_pgbitsX))
#endif
#define GCMEM2BLOCK2(M)		({ \
				    GCMEM2BLOCK(M); \
				 })

/* Since not all objects for which we use the GC_WRITE macro have the
 * heap pointer stored with them (only java objects have), we can't use
 * the fast version.  So use the slow version instead.  
 *
 * XXX clean this up.
 */
#if defined (__i386__) && defined(STORE_HEAP_WITH_OBJECT)
#define	CSOFT_ADDREFERENCE	soft_addreference_slow
#else
#define	CSOFT_ADDREFERENCE	soft_addreference
#endif

#if GC_WRITE_BARRIERS
#define	GC_WRITE(_o, _p)						    \
do {									    \
	if (_p && GCMEM2BLOCK2(_o)->collector != GCMEM2BLOCK2(_p)->collector) \
		CSOFT_ADDREFERENCE(_o, _p, 0);				    \
} while (0)

/* Check that a write is possible.  If a write is not possible, fill
 *  in the errorInfo *, and evaluate to true.
 */
#define GC_CHECK(_o, _p, _e)						\
(_p && GCMEM2BLOCK(_o)->collector != GCMEM2BLOCK(_p)->collector	\
   && CSOFT_ADDREFERENCE(_o, _p, _e))
#else
#define GC_WRITE(_o, _p)
#define GC_CHECK(_o, _p, _e) 0
#endif

#if GC_WRITE_BARRIERS
#define GC_SET(BASE, TARG, EXP)			\
{						\
	void *_val = (EXP);			\
	GC_WRITE(BASE, _val);			\
	TARG = _val;				\
}

#define GC_SET_SAFE(BASE, TARG, EXP, EINFO)		\
({							\
	int _ret = 0;					\
	void *_val = (EXP);				\
							\
	if (GC_CHECK(BASE, _val, EINFO)) _ret = 1;	\
	else TARG = _val;				\
	_ret;						\
})
#else

#define GC_SET(BASE, TARG, EXP)			\
	TARG = (EXP);

#define GC_SET_SAFE(BASE, TARG, EXP, EINFO)	\
	(TARG = (EXP)), 0

#endif	/* GC_WRITE_BARRIERS */

struct _errorInfo;

extern Collector *systemCollector(int threshhold, int limit);
extern int soft_addreference(void *, void*, struct _errorInfo *);

#endif
