/* gc-mem.c
 * The heap manager.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "debug.h"

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "baseClasses.h"
#include "support.h"
#include "stats.h"
#include "locks.h"
#include "jthread.h"
#include "gc.h"
#include "gc-private.h"
#include "jni.h"
#include "md.h"
#include <janosvm/jteam.h>
#include <janosvm/jteamReservation.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#if defined(HAVE_MPROTECT) && defined(DEBUG)
#include <sys/mman.h>
#endif

/* NB: if CALLTARGET_ALIGNMENT is given, we assume that the allocator 
 * returns 8-byte aligned addresses ----
 * On sparc etc., we need 8-byte alignment for double etc.
 *
 * This is only guaranteed if all numbers passed to S() are divisible by 
 * MEMALIGN, which is 8.
 *
 * On x86, if we turn CALLTARGET_ALIGNMENT off, we can set MEMALIGN == 4
 *
 * We do that now to account for the additional 4 bytes in the 'hp'
 * version somewhat.
 */
static gc_freelist initial_freelist[NR_FREELISTS+1]
	= {
#define	S(sz)	{ 0, sz }
#if (MEMALIGN == 4) && (defined(STORE_HEAP_WITH_OBJECT) || defined(FAKE_HEAP_POINTERS))
	S(20),
	S(28),
	S(36),
	S(44),
	S(52),
	S(60),
	S(68),
	S(84),
	S(100),
	S(116),
	S(132),
	S(164),
	S(196),
	S(228),
        S(244),
        S(500),
        S(1004),
        S(2020),
        S(4044),
#else
	S(16),
	S(24),
	S(32),
	S(40),
	S(48),
	S(56),
	S(64),
	S(80),
	S(96),
	S(112),
	S(128),
	S(160),
	S(192),
	S(224),
        S(240),
        S(496),
        S(1000),
        S(2016),
        S(4040),
#endif
	{ (gc_block *)0xffffffff, 0 }
};

static gc_sztable sztable[MAX_SMALL_OBJECT_SIZE+1];

#if defined(DEBUG)
/* bound checked version for debugging */
#define	SIZE2FREELIST(s)	({ 				\
	assert((s) >= 0 && (s) <= MAX_SMALL_OBJECT_SIZE); 	\
	sztable[(s)].list; })
#else
#define	SIZE2FREELIST(s)	(sztable[(s)].list)
#endif

static int max_freelist;

size_t max_small_object_size;

#ifndef gc_pgsize
size_t gc_pgsize;
int gc_pgbits;
#else
#endif
//int gc_pgbitsX = gc_pgbits;

#ifdef DEBUG
int gc_system_alloc_cnt;
#endif

extern struct Hjava_lang_Thread* garbageman;

#ifdef DEBUG
/*
 * analyze the slack incurred by small objects
 */
static int totalslack;
static int totalsmallobjs;

static void 
printslack(void)
{
	dprintf(
		"allocated %d small objects, total slack %d, slack/per "
		"object %8.2f\n", 
		totalsmallobjs, totalslack, totalslack/(double)totalsmallobjs);
}

/* Is this pointer within our managed heap? */ 
#define IS_A_HEAP_POINTER(from) \
	((uintp) (from) >= gc_heap_base && \
	 (uintp) (from) < gc_heap_base + gc_heap_range)

/*
 * check whether the heap is still in a consistent state
 */
void
gc_heap_check(gc_collector *c)
{
	gc_block *rinfo = 0;
	gc_block *b = (void *) gc_block_base;
        int i, end;

	/* check roots */
	for( i = 0; i < REFOBJHASHSZ; i++ )
	{
		refObject *robj;

		robj = c->hash[i];
		while( robj )
		{
			assert(GCMEM2BLOCK(robj)->collector == c);
			assert(IS_A_HEAP_POINTER(robj->mem));
			assert(GCMEM2BLOCK(robj->mem)->collector == c);
			
			robj = robj->next;
		}
	}
	
	/* check freelists */
	for (i = 0; i < NR_FREELISTS; i++) {
		gc_block* blk = c->freelist[i].list;

		if (blk == 0 || blk == (gc_block*)~0) {
			continue;
		} else {
			gc_freeobj* mem = blk->free;
			int count = 0;

			assert(blk->size == c->freelist[i].sz);
			assert(blk->collector == c);
			while (mem) {
				ASSERT_ONBLOCK(mem, blk);
				mem = mem->next;
				count++;
			}
			assert(blk->avail == count);
		}
	}

	/* check block information */
	end = gc_heap_range >> gc_pgbits;
	for (i = 0; i < end; i++) {
		gc_block *info = &b[i];
		uint32 sz = info->size; 

		if (info->firstblock == 0) {
			rinfo = 0;
			continue;
		}

		/* first block of larger block */
		if (info->firstblock == info) {
			rinfo = info;

			DBG(GCDIAG, assert(info->magic == GC_MAGIC));

			/* a block has been allocated, but it's been reserved,
			 * so funcs et al are not initialized
			 */
			if (info->nr == 0) {
				continue;
			}

			/* else it's a small or a large block */
			assert(info->funcs == (uint8*)GCBLOCK2BASE(info));
			assert(info->state == (uint8*)(info->funcs + info->nr));

			if (GC_SMALL_OBJECT(sz)) {
				int lpc;
				
				assert(info->avail <= info->nr);
				for( lpc = 0; lpc < info->nr; lpc++ )
				{
					assert(!(info->state[lpc] & 0x80));
				}
			} else {
				assert(info->nr == 1);
				assert(info->avail <= 1);
				assert(info->free == 0);
			}
		} else {
			/* ensure they all point back to the first block */
			assert(rinfo != 0 && rinfo == info->firstblock);
		}
	}
}

#endif /* DEBUG */

/*
 * Initialise allocator.
 */
void
gc_heap_initialize(void)
{
	int sz = 0;
	uint16 flidx = 0;

#ifndef gc_pgsize
	gc_pgsize = getpagesize();
	for (gc_pgbits = 0;
	     (1 << gc_pgbits) != gc_pgsize && gc_pgbits < 64;
	     gc_pgbits++)
		;
	assert(gc_pgbits < 64);
#endif

	/*
	 * Use the preinitialized initial_freelist table to initialize
	 * the sztable.
	 */
	while (initial_freelist[flidx].list == 0) {
		for (; sz <= initial_freelist[flidx].sz; sz++) {
			sztable[sz].list = flidx;
		}
		flidx++;
	}
	max_small_object_size = sz - 1;
	max_freelist = flidx;

DBG(SLACKANAL,
	atexit(printslack);
    )

#undef	OBJSIZE
}

void gc_mem_init(gc_collector *c)
{
#if 0
	static int once = 0;

	if (!once) {
		gc_heap_initialize();
		gc_prim_initialize();
		once = 1;
	}
#endif
	memcpy(c->freelist, initial_freelist, sizeof(c->freelist));
}

/*
 * Allocate a new block of GC'ed memory.  The block will contain 'nr' objects
 * each of 'sz' bytes.
 */
static
gc_block*
gc_small_block(gc_collector *c, size_t sz)
{
	jthread_t jt = jthread_current();
	memoryStack *ms;
	gc_block* info;
	int i;
	int nr;

	ms = jthread_get_data(jt)->journal;
	if( c->primitiveStack )
		jthread_get_data(jt)->journal = c->primitiveStack;
	info = gc_primitive_alloc(c, gc_pgsize);
	jthread_get_data(jt)->journal = ms;
	if (info == 0) {
		return (0);
	}

	/* Calculate number of objects in this block */
	nr = (gc_pgsize-GCBLOCK_OVH-ROUNDUPALIGN(1))/(sz+2);

	/* Setup the meta-data for the block */
	DBG(GCDIAG, info->magic = GC_MAGIC);

	info->size = sz;
	info->nr = nr;
	info->avail = nr;
	info->funcs = (uint8*)GCBLOCK2BASE(info);
	info->state = (uint8*)(info->funcs + nr);
	info->data = (uint8*)ROUNDUPALIGN(info->state + nr);

	DBG(GCDIAG, memset(info->data, 0, sz * nr));

	/* Build the objects into a free list */
	for (i = nr-1; i >= 0; i--) {
		GCBLOCK2FREE(info, i)->next = GCBLOCK2FREE(info, i+1);
		GC_SET_COLOUR(info, i, GC_COLOUR_FREE);
		GC_SET_STATE(info, i, GC_STATE_NORMAL);
	}
	GCBLOCK2FREE(info, nr-1)->next = 0;
	info->free = GCBLOCK2FREE(info, 0);
DBG(SLACKANAL,
	int slack = ((void *)info) 
		+ gc_pgsize - (void *)(GCBLOCK2MEM(info, nr));
	totalslack += slack;
    )
	return (info);
}

/*
 * Allocate a new block of GC'ed memory.  The block will contain one object
 */
static
gc_block*
gc_large_block(gc_collector *c, size_t sz)
{
	jthread_t jt = jthread_current();
	memoryStack *ms;
	gc_block* info;
	size_t msz;

	/* Add in management overhead */
	msz = sz+GCBLOCK_OVH+ROUNDUPALIGN(2);
	/* Round size up to a number of pages */
	msz = ROUNDUPPAGESIZE(msz);

	ms = jthread_get_data(jt)->journal;
	if( c->primitiveStack )
		jthread_get_data(jt)->journal = c->primitiveStack;
	info = gc_primitive_alloc(c, msz);
	jthread_get_data(jt)->journal = ms;
	if (info == 0) {
		return (0);
	}

	/* Setup the meta-data for the block */
	DBG(GCDIAG, info->magic = GC_MAGIC);

	info->size = sz;
	info->nr = 1;
	info->avail = 1;
	info->funcs = (uint8*)GCBLOCK2BASE(info);
	info->state = (uint8*)(info->funcs + 1);
	info->data = (uint8*)ROUNDUPALIGN(info->state + 1);
	info->free = 0;

	DBG(GCDIAG, memset(info->data, 0, sz));

	GCBLOCK2FREE(info, 0)->next = 0;
	GC_SET_COLOUR(info, 0, GC_COLOUR_FREE);
	GC_SET_STATE(info, 0, GC_STATE_NORMAL);

	return (info);
}

/*
 * gc_heap_malloc
 * Allocate a piece of memory.
 */
void*
gc_heap_malloc(gc_collector *c, size_t sz)
{
	size_t lnr;
	gc_freeobj* mem;
	gc_block** mptr;
	gc_block* blk;
	size_t nsz;

	/* Initialise GC heap first time in - we must assume single threaded
	 * operation here so we can do the lock initialising.
	 */

DBG(SLACKANAL,
	if (GC_SMALL_OBJECT(sz)) {
		totalslack += (c->freelist[SIZE2FREELIST(sz)].sz - sz);
		totalsmallobjs++;
	}
    )

DBG(GCDIAG, 
	gc_heap_check(c);
    )

	if (GC_SMALL_OBJECT(sz)) {

		/* Translate size to object free list */
		lnr = SIZE2FREELIST(sz);
		nsz = c->freelist[lnr].sz;
		assert(nsz >= sz);

		/* No available objects? Allocate some more */
		mptr = &c->freelist[lnr].list;
		if (*mptr != 0) {
			blk = *mptr;
			assert(blk->free != 0);
DBG(GCALLOC,		dprintf("gc_heap_malloc: freelist %d at %p free %p\n", sz, *mptr, blk->free);)
		}
		else {
			blk = gc_small_block(c, nsz);
			if (blk == 0) {
				nsz = gc_pgsize;
				goto nospace;
			}
			blk->next = *mptr;
			*mptr = blk;

DBG(GCALLOC,		dprintf("gc_heap_malloc: small block %d at %p free %p\n", sz, *mptr, blk->free);)
		}

		/* Unlink free one and return it */
		mem = blk->free;

		DBG(GCDIAG,
		    assert(blk->magic == GC_MAGIC);
		    ASSERT_ONBLOCK(mem, blk);
		    if (mem->next) ASSERT_ONBLOCK(mem->next, blk));

		blk->free = mem->next;

		GC_SET_STATE(blk, GCMEM2IDX(blk, mem), GC_STATE_NORMAL);

		/* Once we use all the sub-blocks up, remove the whole block
		 * from the freelist.
		 */
		assert(blk->avail > 0);
		blk->avail--;
		if (blk->avail == 0) {
			*mptr = blk->next;
		}
	}
	else {
		nsz = sz;
		blk = gc_large_block(c, nsz);
		if (blk == 0) {
			nsz = nsz + GCBLOCK_OVH + sizeof(gcFuncs*) + ROUNDUPALIGN(1);
			nsz = ROUNDUPPAGESIZE(nsz);
			goto nospace;
		}
		mem = GCBLOCK2FREE(blk, 0);
		GC_SET_STATE(blk, 0, GC_STATE_NORMAL);
DBG(GCALLOC,	dprintf("gc_heap_malloc: large block %d at %p\n", sz, mem);	)
		blk->avail--;
		assert(blk->avail == 0);
	}

	/* Clear memory */
	bzero(mem, nsz);

	assert(GC_OBJECT_SIZE(mem) >= sz);
	
DBG(GCDIAG, 
	gc_heap_check(c);
    )

	return (mem);

	/* --------------------------------------------------------------- */

DBG(GCDIAG, 
	gc_heap_check(c);
    )
	nospace:;
	return 0;
}

/*
 * Free a piece of memory.
 */
void
gc_heap_free(gc_collector *c, void* mem)
{
	gc_block* info;
	gc_freeobj* obj;
	int lnr;
	int msz;
	int idx;


	info = GCMEM2BLOCK(mem);
	idx = GCMEM2IDX(info, mem);

	if (info->collector != c) {
		fprintf(stderr, "Collector %p is trying to free object "
			"allocated by collector %p (kernel=%p)\n",
			c, info->collector, main_collector);
	}
	assert(info->collector == c);
	DBG(GCDIAG,
	    gc_heap_check(c);
	    assert(info->magic == GC_MAGIC);
	    assert(GC_GET_COLOUR(info, idx) != GC_COLOUR_FREE)
	    );

	GC_SET_COLOUR(info, idx, GC_COLOUR_FREE);

DBG(GCFREE,
	dprintf("gc_heap_free: memory %p size %d\n", mem, info->size);	)

	if (GC_SMALL_OBJECT(info->size)) {
		lnr = SIZE2FREELIST(info->size);
		/* If this block contains no free sub-blocks yet, attach
		 * it to freelist.
		 */
		if (info->avail == 0) {
			info->next = c->freelist[lnr].list;
			c->freelist[lnr].list = info;
		}
		info->avail++;
		DBG(GCDIAG,
		    /* write pattern in memory to see when live objects were
		     * freed - Note that (f4f4f4f4 == -185273100)
		     */
		    memset(mem, 0xf4, info->size));
		obj = GCMEM2FREE(mem);
		obj->next = info->free;
		info->free = obj;

		ASSERT_ONBLOCK(obj, info);

		/* If we free all sub-blocks, free the block */
		assert(info->avail <= info->nr);
		if (info->avail == info->nr) {
			jthread_t jt = jthread_current();
			memoryStack *ms;
			gc_block** finfo = &c->freelist[lnr].list;
			
			ms = jthread_get_data(jt)->journal;
			if( c->primitiveStack )
			{
				jthread_get_data(jt)->journal =
					c->primitiveStack;
			}
			for (;;) {
				if (*finfo == info) {
					(*finfo) = info->next;
					info->size = gc_pgsize;
					gc_primitive_free(c, info);
					break;
				}
				finfo = &(*finfo)->next;
				assert(*finfo != 0);
			}
			jthread_get_data(jt)->journal = ms;
		}
	}
	else {
		jthread_t jt = jthread_current();
		memoryStack *ms;
		
		ms = jthread_get_data(jt)->journal;
		if( c->primitiveStack )
			jthread_get_data(jt)->journal = c->primitiveStack;
		assert(info->nr == 1);
		/* Calculate true size of block */
		msz = info->size + GCBLOCK_OVH + ROUNDUPALIGN(2);
		msz = ROUNDUPPAGESIZE(msz);
		info->size = msz;
		gc_primitive_free(c, info);
		jthread_get_data(jt)->journal = ms;
	}
	DBG(GCDIAG,
	    gc_heap_check(c);
	    );

}

/*
 * Release any blocks owned by `me', which is part of a dead team.
 */
void
gc_release_blocks(gc_collector *me)
{
	gc_block *b = (gc_block*)gc_block_base;
	jteam *prev;
	int i, end;

	prev = quickChangeTeamReservation(kernelTeam);
	end = gc_heap_range >> gc_pgbits;
        for (i = 0; i < end; i++) {
                if (b[i].collector == me) {
			int msz = 0;

			if( GC_SMALL_OBJECT(b[i].size) )
			{
				bzero(b[i].funcs, b[i].nr);
				memset(b[i].state,
				       GC_COLOUR_FREE,
				       b[i].nr);
				b[i].nr = 0;
				b[i].avail = 0;
				b[i].size = gc_pgsize;
			}
			else if( b[i].data && (b[i].size % gc_pgsize) )
			{
#if 0
				if( (b[i].size % gc_pgsize) == 0 )
				{
					fprintf(stderr,
						"warning, pgsize block at "
						"%d %p size %d",
						i, &b[i], b[i].size);
				}
#endif
				msz = b[i].size +
					GCBLOCK_OVH +
					ROUNDUPALIGN(2);
				msz = ROUNDUPPAGESIZE(msz);
				assert(b[i].size <= msz);
				b[i].size = msz;
			}
			gc_primitive_free2(me, &b[i]);
			if( msz )
			{
				i += (msz / gc_pgsize) - 1;
			}
                }
        }
	quickChangeTeamReservation(prev);

	ensure(currentTeam() != kernelTeam);
}
