/*
 * Primitive block management:  Allocating and freeing whole pages.
 *
 * Unused pages may be marked unreadable.  This is only done when
 * compiled with DEBUG.
 */ 

#include "debug.h"
/* undefine this to revert to old tile scheme */
#define	PREDEFINED_NUMBER_OF_TILES

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "baseClasses.h"
#include "support.h"
#include "stats.h"
#include "locks.h"
#include "thread.h"
#include "jthread.h"
#include "gc.h"
#include "gc-private.h"
#include "jni.h"
#include "assert_pp.h"
#include <janosvm/jteam.h>
#include <janosvm/jteamReservation.h>
#include <janosvm/memResource.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_MPROTECT
#include <sys/mman.h>
#endif


/*
 * This blows up if we ever need to realloc the block array.
 */

static gc_block* gc_prim_freelist;
iLock *prim_lock;

size_t gc_heap_base;
size_t gc_block_base;
uintp gc_heap_range;
int	gc_frag_failure;
char   *gc_alloc_failure;

size_t gc_heap_total;		/* current size of the heap */
size_t gc_heap_increment;	/* amount of memory by which to grow heap */
size_t gc_heap_initial_size;	/* amount of memory to initially allocate */
size_t gc_heap_limit;		/* maximum size to which heap should grow */

#if defined(KAFFE_STATS)
static counter gcpages;
#endif

static size_t prim_alloced;
static size_t prim_free;

static void *gc_system_alloc(size_t sz);

#if 0
/* should we be using this?  See MemResource_finalize ...
 *
 */
/* Destroy a memlimit structure on behalf of a MemResource. */
void
gc_free_limit(memlimit *l)
{
	jmutex_lock(&prim_lock);
	if (l->parent) {
		if (l->hard) {
			gc_bump_limit(l->parent, -l->limit);
		} else {
			gc_bump_limit(l->parent, -l->total);
		}
	}
	jmutex_unlock(&prim_lock);
}
#endif

void gc_prim_initialize(void)
{
	gc_heap_increment = Kaffe_JavaVMArgs[0].allocHeapSize;
	gc_heap_initial_size = Kaffe_JavaVMArgs[0].minHeapSize;
	gc_heap_limit = Kaffe_JavaVMArgs[0].maxHeapSize;

	/*
	 * Perform some sanity checks.
	 */
	if (gc_heap_initial_size > gc_heap_limit) {
		fprintf(stderr, 
		    "Initial heap size (%dK) > Maximum heap size (%dK)\n",
		    gc_heap_initial_size/1024, gc_heap_limit/1024);
		EXIT(-1);
	}

	/* Round 'gc_heap_increment' up to pagesize */
	gc_heap_increment = ROUNDUPPAGESIZE(gc_heap_increment);

	/* Round 'gc_heap_initial_size' up to pagesize */
	gc_heap_initial_size = ROUNDUPPAGESIZE(gc_heap_initial_size);

	/* allocate heap of initial size from system */
	gc_system_alloc(gc_heap_initial_size);
}

#if !defined(HAVE_MPROTECT) || !defined(DEBUG) || 1	/* TEMP */
#define mprotect(A,L,P)
#define ALL_PROT
#define NO_PROT
#else
#ifndef PROT_NONE
# define PROT_NONE 0
#endif

/* In a sense, this is backwards. */
#define ALL_PROT PROT_READ|PROT_WRITE|PROT_EXEC
#define NO_PROT  PROT_NONE
#endif

/* Mark this block as in-use */
static inline void 
gc_block_add(gc_block *b)
{
	gc_block *ptr, *nptr;

	assert(!b->firstblock);

	/* set firstblock ptr for all pages belonging to that block */
	ptr = b;
	nptr = GCBLOCKEND(b);
	while (ptr < nptr) {
		ptr->firstblock = b;
		ptr++;
	}
	mprotect(GCBLOCK2BASE(b), b->size, ALL_PROT);
}

/* Mark this block as free */
static inline void 
gc_block_rm(gc_block *b)
{
	gc_block *ptr, *nptr;

	/* clear firstblock ptr for all pages belonging to that block */
	ptr = b;
	nptr = GCBLOCKEND(b);
	while (ptr < nptr) {
		require(ptr->firstblock == b);
		
		ptr->firstblock = 0;
		ptr++;
	}

	mprotect(GCBLOCK2BASE(b), b->size, NO_PROT);
}

/*
 * System memory management:  Obtaining additional memory from the
 * OS.  This looks more complicated than it is, since it does not require
 * sbrk.
 */
/* Get some page-aligned memory from the system. */
static uintp
pagealloc(size_t size)
{
	void* ptr;

#define	CHECK_OUT_OF_MEMORY(P)	if ((P) == 0) return 0;

#if defined(HAVE_SBRK)

	/* Our primary choice for basic memory allocation is sbrk() which
	 * should avoid any unsee space overheads.
	 */
	for (;;) {
		int missed;
		ptr = sbrk(size);
		if (ptr == (void*)-1) {
			ptr = 0;
			break;
		}
		if ((uintp)ptr % gc_pgsize == 0) {
			break;
		}
		missed = gc_pgsize - ((uintp)ptr % gc_pgsize);
		DBG(GCSYSALLOC,
		    dprintf("unaligned sbrk %p, missed %d bytes\n",
			    ptr, missed));
		sbrk(-size + missed);
	}
	CHECK_OUT_OF_MEMORY(ptr);

#elif defined(HAVE_MEMALIGN)

        ptr = memalign(gc_pgsize, size);
	CHECK_OUT_OF_MEMORY(ptr);

#elif defined(HAVE_VALLOC)

        ptr = valloc(size);
	CHECK_OUT_OF_MEMORY(ptr);

#else

	/* Fallback ...
	 * Allocate memory using malloc and align by hand.
	 */
	size += gc_pgsize;

        ptr = malloc(size);
	CHECK_OUT_OF_MEMORY(ptr);
	ptr = (void*)((((uintp)ptr) + gc_pgsize - 1) & -gc_pgsize);

#endif
	addToCounter(&gcpages, "gcmem-system pages", 1, size);
	return ((uintp) ptr);
}

/* Free memory allocated with pagealloc */
static void pagefree(uintp base, size_t size)
{
#ifdef HAVE_SBRK
	sbrk(-size);
#else
	/* it must have been allocated with memalign, valloc or malloc */
	free((void *)base);
#endif
}

/*
 * Allocate size bytes of heap memory, and return the corresponding
 * gc_block *.
 */
static void *
gc_block_alloc(size_t size)
{
	int size_pg = (size>>gc_pgbits);
	static int n_live = 0;	/* number of pages in java heap */
	static int nblocks;	/* number of gc_blocks in array */
	uintp heap_addr;
	static uintp last_addr;

	if (!gc_block_base) {
		nblocks = (gc_heap_limit>>gc_pgbits);

		nblocks += nblocks/4;
		gc_block_base = (uintp) malloc(nblocks * sizeof(gc_block));
		if (!gc_block_base) return 0;
		bzero((void *)gc_block_base, nblocks * sizeof(gc_block));
	}

	heap_addr = pagealloc(size);

	if (!heap_addr) return 0;
	
	if (!gc_heap_base) {
		gc_heap_base = heap_addr;
	}

	if (GCMEM2BLOCK(heap_addr + size)
	    > ((gc_block *)gc_block_base) + nblocks
	    || heap_addr < gc_heap_base) {
		uintp old_blocks = gc_block_base;
		int onb = nblocks;
		int min_nb;	/* minimum size of array to hold heap_addr */
#if defined(KAFFE_STATS)
		static timespent growtime;
#endif

		startTiming(&growtime, "gctime-blockrealloc");
		/* Pick a new size for the gc_block array.  Remember,
		   malloc does not simply grow a memory segment.

		   We can extrapolate how many gc_blocks we need for
		   the entire heap based on how many heap pages
		   currently fit in the gc_block array.  But, we must
		   also make sure to allocate enough blocks to cover
		   the current allocation */
		nblocks = (nblocks * (gc_heap_limit >> gc_pgbits))
			/ n_live;
		if (heap_addr < gc_heap_base) 
			min_nb = nblocks
			  + ((gc_heap_base - heap_addr) >> gc_pgbits);
		else
			min_nb = ((heap_addr + size) - gc_heap_base) >>
			  gc_pgbits;
		nblocks = MAX(nblocks, min_nb);
		DBG(GCSYSALLOC,
		    dprintf("growing block array from %d to %d elements\n",
			    onb, nblocks));

		jthread_spinon(0);
		gc_block_base = (uintp) realloc((void *) old_blocks,
						nblocks * sizeof(gc_block));
		if (!gc_block_base) {
			/* roll back this call */
			pagefree(heap_addr, size);
			gc_block_base = old_blocks;
			nblocks = onb;
			jthread_spinoff(0);
			return 0;
		}

		/* If the array's address has changed, we have to fix
		   up the pointers in the gc_blocks, as well as all
		   external pointers to the gc_blocks.  We can only
		   fix gc_prim_freelist and the size-freelist array.
		   There should be no gc_block *'s on any stack
		   now. */ 
		if (gc_block_base != old_blocks) {
#if 0
			extern gc_block *gc_prim_freelist;
			int i;
			gc_block *b = (void *) gc_block_base;
			uintp delta = gc_block_base - old_blocks;

#define R(X) if (X) ((uintp) (X)) += delta

			DBG(GCSYSALLOC,
			    dprintf("relocating gc_block array\n"));
			for (i = 0; i < onb; i++) R(b[i].next);
			bzero(b + onb, (nblocks - onb) * sizeof(gc_block));

			R(gc_prim_freelist);

			for (i = 0; freelist[i].list != (void *) -1; i++) 
				R(freelist[i].list);
#undef R
#else
			assert(!"can't relocate block pointers in kaffeOS");
#endif
		}

		jthread_spinoff(0);
		stopTiming(&growtime);
	}
	n_live += size_pg;
	last_addr = MAX(last_addr, heap_addr + size);
	gc_heap_range = last_addr - gc_heap_base;
	DBG(GCSYSALLOC, dprintf("%d unused bytes in heap addr range\n",
				gc_heap_range - gc_heap_total));
	mprotect((void *) heap_addr, size, NO_PROT);
	return GCMEM2BLOCK(heap_addr);
}

/*
 * Wake up our gc thread.  Called with both primitive and allocator
 * locks held.
 */
static void
gc_now(gc_collector *c, int *pILockRoot)
{
	int *origOwner;
	int iLockRoot;
	
	if (!c->gcman.thread) {
		return;	/* support NOGC */
	}

	/*
	 * XXX Have to drop our lock on the allocator so that the gc thread
	 * can get it.  Not the most elegant solution...
	 */
	if( pILockRoot )
		_unlockMutex(&prim_lock, pILockRoot);
	origOwner = c->gcman.owner;
	_unlockMutex(&c->lock, c->gcman.owner);
	lockMutex(&c->gcman);
	if (!c->gcman.gcRunning) {
		c->gcman.gcRunning = 1;
		signalCond(&c->gcman);
	}
	while (c->gcman.gcRunning) {
		waitCond(&c->gcman, 0);
	}
	unlockMutex(&c->gcman);
	_lockMutex(&c->lock, origOwner);
	if( pILockRoot )
		_lockMutex(&prim_lock, pILockRoot);
}

/*
 * Try to avoid throwing out of memory in the system heap.  If we are
 * allocating on behalf of some process, gc its heap in the hopes of
 * freeing up some pages.
 */

#if 0
/*
 * XXX: When we throw out of memory in the system heap, we may end up
 * grabbing an OOM object from the user heap.  This seems to be a
 * basic problem with the way errors are propagated.
 *
 * XXX 2: Why doesn't GC_invoke wait for the collector to actually finish?
 */
#define GC_NOW(c, tried) {					\
	if ((c) == &gc_metal_collector && (tried) == 1		\
	    && GET_HOME_HEAP() != (c)) {			\
		jmutex_unlock(&prim_lock);			\
		GC_invoke((Collector *) GET_HOME_HEAP(), 1);	\
		jmutex_lock(&prim_lock);			\
		(tried)++;					\
		gc_prim_map();					\
	}							\
	if (!(tried)) {						\
		gc_now(c);					\
		(tried)++;					\
	}							\
}
#else
#define GC_NOW(c, root, tried) if (!(tried)) { gc_now(c, root); (tried)++; }
#endif

/* Compute the end of a live block:  The size does not account for gc
 * metadata.  XXX: fix max_small_object_size.  Also, we must be aware
 * of sparse memory allocation.
 */
extern size_t max_small_object_size;

static gc_block *nextpage(gc_block *p)
{
	while (!p->size
	       && p < ((gc_block *) gc_block_base) + (gc_heap_range>>gc_pgbits))
		p++;

	/* XXX: this works for blocks that are completely allocated, but
	 * not blocks we just returned to gc-mem.c
	 */
	if (p->size < max_small_object_size) p++;
	else if (!p->collector) p = GCBLOCKEND(p);
	else p += (ROUNDUPPAGESIZE(p->size + 8)) >> gc_pgbits;

	while (!p->size
	       && p < ((gc_block *) gc_block_base) + (gc_heap_range>>gc_pgbits))
		p++;
	return p;
}

#if defined(DEBUG)
static int fill_range(int val, int count, int i)
{
	while (count) {
		if (!(i % 64)) dprintf("\n");
		dprintf("%c", val);
		count--;
		i++;
	}
	return i;
}

static void
dump_block_state(gc_block* b)
{
	int i;

	for (i = 0; i < b->nr; i++) {
		int p = '?', q = '?', r = '?', c = b->state[i] & 0xFF;
		// dprintf("%02x", b->state[i]);
		switch (c & GC_COLOUR_MASK) {
		case GC_COLOUR_FREE: 		p = '_'; continue;
		case GC_COLOUR_FIXED: 		p = 'F'; break;
		case GC_COLOUR_WHITE: 		p = 'W'; break;
		case GC_COLOUR_BLACK: 		p = 'B'; break;
		}
		dprintf("%c", p);
		switch (c & GC_STATE_MASK) {
		case GC_STATE_NEEDFINALIZE: 	q = 'n'; break;
		case GC_STATE_INFINALIZE: 	q = 'i'; break;
		case GC_STATE_NORMAL: 		q = 0; break;
		}
		if (q) {
		    dprintf("%c", q);
		}
		if (r) {
		    dprintf("%c", r);
		}
	}
}

static void
dump_block_state_long(gc_block* b)
{
	int i;

	for (i = 0; i < b->nr; i++) {
		int c = b->state[i] & 0xFF;
		if ((c & GC_COLOUR_MASK) != GC_COLOUR_FREE) {
			dprintf("[%03d]: %s\n", i, describeObject(UTOMEM(GCBLOCK2MEM(b, i))));
		}
	}
}

static int type_totalb[GC_ALLOC_MAX_INDEX + 1];
static int type_totaln[GC_ALLOC_MAX_INDEX + 1];

static void 
init_type_counters(void) 
{
	bzero(type_totalb, sizeof(type_totalb));
	bzero(type_totaln, sizeof(type_totaln));
}

static void
sum_up_block_types(gc_block* b)
{
	int i;

	for (i = 0; i < b->nr; i++) {
		int c = b->state[i] & 0xFF;
		if ((c & GC_COLOUR_MASK) != GC_COLOUR_FREE) {
			int fidx = GC_GET_FUNCS(b, i);
			assert(fidx < GC_ALLOC_MAX_INDEX + 1);
			type_totalb[fidx] += b->size;
			type_totaln[fidx] += 1;
		}
	}
}

static void
print_type_counters(void)
{
	int i;

	for (i = 0; i < GC_ALLOC_MAX_INDEX + 1; i++) {
		extern gcFuncs gcFunctions[];
		dprintf("type %s: %d objects with %d bytes\n",
			gcFunctions[i].description,
			type_totaln[i], type_totalb[i]);
	}
}

			
/* print a fully detailed heap map */
void
gc_detailed_heap_map(void)
{
	gc_block *b = (void *) gc_block_base;
	gc_block *end = b + (gc_heap_range >> gc_pgbits);

	init_type_counters();

	while (b < end) {
		gc_block *ob = b;
		dprintf("%p: ", b);
		if (b->firstblock == b) {	/* begin of block */
			uint32 sz = b->size;
			dprintf("%03d/%03d sz=%5d ", 
				b->nr - b->avail, b->nr, sz);
			if (b->nr == 0) {
				dprintf(" reserved\n");
				b = GCBLOCKEND(b);
			} else {
				if (getenv("LONGDUMP") == 0) {
					dump_block_state(b);
					dprintf("\n");
				} else {
					dprintf("\n");
					dump_block_state_long(b);
					dprintf("\n");
				}
				sum_up_block_types(b);

				if (b->size > max_small_object_size) {
					int sz = ROUNDUPPAGESIZE(b->size + 8);
					b += (sz>>gc_pgbits);
				} else {
					b++;
				}
			}
		} else
		if (b->firstblock == 0) {
			if (b->size == 0) {
				dprintf(" one block unused(!?)\n");
				b++;
			} else {
				dprintf("%d Kbytes free\n", 
				    ((GCBLOCKEND(b)-b) * gc_pgsize)/1024);
				assert(b != GCBLOCKEND(b));
				b = GCBLOCKEND(b);
			}
		}	
		assert(b != ob);	/* or else no progress in this loop */
	}
	print_type_counters();
}

void
gc_prim_map(void)
{
	gc_collector *c;
	gc_block *b = (void *) gc_block_base;
	gc_block *end = b + (gc_heap_range >> gc_pgbits);
	int i = 0;
	int j, n;
	extern gc_collector *defined_collectors;

	dprintf("Key to heap map:\n");
	dprintf("' ' -- no such block\n");
	dprintf("'=' -- continued block\n");
	dprintf("'_' -- free block\n");

	/* count first to ensure consistent ordering */
	for (c = defined_collectors, n = 0; c; c = c->next, n++) {
		continue;
	}
	n--;

	for (c = defined_collectors, j = 0; c; c = c->next, j++) {
		dprintf("'%c' -- owned by heap %p %s\n", 'A' + n - j, c,
		    (c == &gc_metal_collector) ? "system heap" : "user heap");
	}
	init_type_counters();
	while (b < end) {
		if (b->collector) {
			int count = 0;

			if (!(i % 64)) dprintf("\n");

			for (c = defined_collectors, j = 0;
			     c && c != b->collector;
			     c = c->next, j++)
				;

			if (!c) {
				dprintf("block %p collector %p not found\n",
					b, b->collector);
				assert(c || !!!"collector not found");
			}
			dprintf("%c", 'A'+n-j);
			if (b->size > max_small_object_size) {
				int sz = ROUNDUPPAGESIZE(b->size + 8);
				count = (sz>>gc_pgbits) - 1;
			}
			sum_up_block_types(b);
			b += count+1;
			i = fill_range('=', count, ++i);
		}
		else if (!b->size) {
			if (!(i % 64)) dprintf("\n");

			dprintf(" ");
			i++;
			b++;
		}
		else {
			gc_block *end = GCBLOCKEND(b);
			int count = end - b;

			i = fill_range('_', count, i);
			b = end;
		}
	}
	dprintf("\nPrinting totals for allocation types:\n");
	print_type_counters();
	dprintf("\n");

#define PS(V, C) do { dprintf(#C": "#V" = %dKB\n", V/1024); } while (0)
#define PB(V, C) do { dprintf(#C": "#V" = 0x%x\n", V); } while (0)
	PS(gc_heap_total, current size of the heap);
	PS(gc_heap_increment, amount of memory by which to grow heap);
	PS(gc_heap_initial_size, amount of memory to initially allocate);
	PS(gc_heap_limit, maximum size to which heap should grow);
	PB(gc_heap_base, current base of heap);
	PB(gc_block_base, current base of block table);
	PS(gc_heap_range, current size of heap);
	PS(prim_alloced, ?);
	PS(prim_free, ?);

#undef PS
#undef PB

	DBG(GCWALK, gc_detailed_heap_map());
}
#endif
#define NEXTPAGE(x) nextpage(x)

void
gc_prim_validate(void)
{
	gc_block *real = (void *) gc_block_base;
	gc_block *end = real + (gc_heap_range>>gc_pgbits);
	gc_block *list = gc_prim_freelist;

	while (real < end && real->collector) real = NEXTPAGE(real);
	while (real < end) {
		assert(real == list);
		list = list->next;
		real = NEXTPAGE(real);
		while (real < end && real->collector) real = NEXTPAGE(real);
	}
}

/*
 * Called with prim lock held.  Try to allocate sz bytes from prim
 * freelist.
 */
static gc_block *
gc_primitive_alloc_1(gc_collector *c, size_t sz)
{
	gc_block* ptr;
	gc_block** pptr;
	gc_block *r = 0;
	
	for (pptr = &gc_prim_freelist; *pptr != 0; pptr = &ptr->next) {
		ptr = *pptr;
		/* First fit */
		if (sz <= ptr->size) {
			size_t left;

			/* If there's more than a page left, split it */
			left = ptr->size - sz;
			if (left >= gc_pgsize) {
				gc_block* nptr;

				ptr->size = sz;
				nptr = GCBLOCKEND(ptr);
				nptr->size = left;

				DBG(GCDIAG,
				    nptr->magic = GC_MAGIC;
				    assert(!nptr->firstblock);
				    );

				nptr->next = ptr->next;
				ptr->next = nptr;
			}
			*pptr = ptr->next;
DBG(GCPRIM,		dprintf("gc_primitive_alloc: %d bytes from freelist @ %p\n", ptr->size, ptr); )
			gc_block_add(ptr);
			r = ptr;

			prim_alloced += r->size;
			prim_free -= r->size;
			break;
		}
	}
	return r;
}

/*
 * Allocate a block of memory from the free list or, failing that, the
 * system pool.
 */
gc_block*
gc_primitive_alloc(gc_collector *c, size_t sz)
{
	jteamReservation *jr;
	gc_block *r = 0;
	int tried_gc = 0;
	int iLockRoot;
	
	/*
	 * Enforce per-process memory limit, and possibly trigger a
	 * collection.
	 */
	require(sz % gc_pgsize == 0);

	/*
	 * If we reach the gc trigger point, gc now.
	 * However, don't trust the trigger too much.  If we're already
	 * at a point where we'd hit our limit, we must gc as well.
	 *
	 * If, after gcing, we exceed the local limit, return 0.  If
	 * we are exceeding some ancestor limit, and haven't tried to gc
	 * yet, gc now and try to bump the limit again.
	 */
	if ((c->client->used > c->client->gcTrigger) ||
	    (c->client->used + sz) > c->client->max)
	{
		GC_NOW(c, 0, tried_gc);
	}

	if( !chargeMemory(c->client, sz) )
	{
		GC_NOW(c, 0, tried_gc);
		if( !chargeMemory(c->client, sz) )
		{
			gc_alloc_failure = "MemResource exceeded";
			goto done;
		}
	}
	
	if( (jr = pushTeamReservation(kernelTeam)) == 0 )
		return 0;
	lockStaticMutex(&prim_lock);
	/* Try to allocate space again, if a failed allocation
	 * triggered a gc.  Keep the primitive lock and remember that
	 * we already tried a gc
	 */
	/* gc_prim_validate();*/
	r = gc_primitive_alloc_1(c, sz);

	if (!r) {
		/*
		 * We are entitled to more memory, but none is
		 * available, throw another large chunk on the
		 * gc_prim_freelist.
		 *
		 * Maybe gc_system_alloc will fail.  In this case, we
		 * should gc, unless we already tried.
		 *
		 * If this doesn't work, perhaps there should be a way
		 * to force a full gc.
		 */
		if (gc_system_alloc(MAX(sz, gc_heap_increment))) {
			r = gc_primitive_alloc_1(c, sz);
		} else {
			gc_alloc_failure = "gc_system_alloc failed";
			// turn on verbose walking to see the crap that 
			// has accumulated
			// kaffevmDebugMask |= DBG_GCWALK;
		}

		if (!r && !tried_gc) {
			unlockStaticMutex(&prim_lock);
			GC_NOW(c, &iLockRoot, tried_gc);
			lockStaticMutex(&prim_lock);
			r = gc_primitive_alloc_1(c, sz);
			if (!r) {
				gc_alloc_failure = "Could not alloc chunk!?";
			}
		}
	}

	if (r) {
		c->local += r->size;
		assert(r->size == sz);
		r->collector = c;
	} else {
		/* Allocation may fail because the heap is
		 * fragmented.  If this happens, we need to fix up the
		 * limits.
		 *
		 * Aren't we hosed by then now anyway?
		 * This seems like a problem inherent in our approach, btw:
		 * Fragmentation in one heap may cause another heap's
		 * inability to allocate large blocks, right?
		 */
		creditMemory(c->client, sz);
		gc_frag_failure++;
	}
	unlockStaticMutex(&prim_lock);
	popTeamReservation(0);
done:
	return (r);
}

/*
 * Try to reserve some memory for OOM exception handling.  Gc once at
 * the beginning.  We start out looking for an arbitrary number of
 * pages (4), and cut our expectations in half until we are able to
 * meet them.
 */
gc_block *
gc_primitive_reserve(gc_collector *c)
{
	jteamReservation *jr;
	gc_block *r = 0;
	int size = 4 * gc_pgsize;
	int tried_gc = 0;
	int iLockRoot;
	
	if( !chargeMemory(c->client, size) )
	{
		GC_NOW(c, 0, tried_gc);
		while( size >= gc_pgsize && !chargeMemory(c->client, size) )
		{
			size /= 2;
		}
	}
	
	if( (jr = pushTeamReservation(kernelTeam)) == 0 )
		return 0;
	lockStaticMutex(&prim_lock);
	while (size >= gc_pgsize && !(r = gc_primitive_alloc_1(c, size))) {
		if (!tried_gc) {
			GC_NOW(c, &iLockRoot, tried_gc);
			if ((r = gc_primitive_alloc_1(c, size)))
				break;
		}
		if (size == gc_pgsize) {
			creditMemory(c->client, size);
			break;
		}
		size /= 2;
		creditMemory(c->client, size);
	}
	if (r) {
		r->collector = c;
		c->local += size;

		/* make sure this block is not mistaken for an small 
		 * or large allocated block
		 */
		r->nr = r->avail = 0;
		r->data = r->funcs = r->state = 0;
	}
	unlockStaticMutex(&prim_lock);
DBG(GCHEAPS, 
	dprintf("%p: reserved %d bytes of memory for OOM @ block %p\n",
			     c, r ? GCBLOCKSIZE(r) : 0, r);
    )
	popTeamReservation(jr);
	return (r);
}

/*
 * Return a block of memory to the free list.  Called with primitive lock.
 * Precondition is that mem->size is valid and represents the true size of
 * the block to be freed.
 *
 * NB: the gc_prim_freelist is kept sorted by address
 */
void
gc_primitive_free_1(gc_block* mem)
{
	gc_block* lptr;
	gc_block* nptr;

	assert(mem->size % gc_pgsize == 0);

	/* Remove from object hash */
	assert(mem->firstblock);
	gc_block_rm(mem);
	mem->next = 0;

	prim_alloced -= mem->size;
	prim_free += mem->size;
	
	if(mem < gc_prim_freelist || gc_prim_freelist == 0) {
		/* If this block is directly before the first block on the
		 * freelist, merge it into that block.  Otherwise just
		 * attached it to the beginning.
		 */
		if (GCBLOCKEND(mem) == gc_prim_freelist) {
DBG(GCPRIM,	dprintf("gc_primitive_free: Merging (%d,%p) beginning of freelist\n", mem->size, mem); )
			mem->size += gc_prim_freelist->size;
			mem->next = gc_prim_freelist->next;
		}
		else {
DBG(GCPRIM,	dprintf("gc_primitive_free: Prepending (%d,%p) beginning of freelist\n", mem->size, mem); )
			mem->next = gc_prim_freelist;
		}
		gc_prim_freelist = mem;
		return;
	}

	/* Search the freelist for the logical place to put this block */
	lptr = gc_prim_freelist;
	while (lptr->next != 0) {
		nptr = lptr->next;
		assert(lptr->firstblock == 0);
		assert(nptr->firstblock == 0);
		if (mem > lptr && mem < nptr) {
			/* Block goes here in the logical scheme of things.
			 * Work out how to merge it with those which come
			 * before and after.
			 */
			if (GCBLOCKEND(lptr) == mem) {
				if (GCBLOCKEND(mem) == nptr) {
					/* Merge with last and next */
DBG(GCPRIM,				dprintf("gc_primitive_free: Merging (%d,%p) into list\n", mem->size, mem); )
					lptr->size += mem->size + nptr->size;
					lptr->next = nptr->next;
				}
				else {
					/* Merge with last but not next */
DBG(GCPRIM,				dprintf("gc_primitive_free: Merging (%d,%p) with last in list\n", mem->size, mem); )
					lptr->size += mem->size;
				}
			}
			else {
				if (GCBLOCKEND(mem) == nptr) {
					/* Merge with next but not last */
DBG(GCPRIM,				dprintf("gc_primitive_free: Merging (%d,%p) with next in list\n", mem->size, mem); )
					mem->size += nptr->size;
					mem->next = nptr->next;
					lptr->next = mem;
				}
				else {
					/* Wont merge with either */
DBG(GCPRIM,				dprintf("gc_primitive_free: Inserting (%d,%p) into list\n", mem->size, mem); )
					mem->next = nptr;
					lptr->next = mem;
				}
			}
			return;
		}
		lptr = nptr;
	}

	/* If 'mem' goes directly after the last block, merge it in.
	 * Otherwise, just add in onto the list at the end.
	 */
	if (GCBLOCKEND(lptr) == mem) {
DBG(GCPRIM,	dprintf("gc_primitive_free: Merge (%d,%p) onto last in list\n", mem->size, mem); )
		lptr->size += mem->size;
	}
	else {
DBG(GCPRIM,	dprintf("gc_primitive_free: Append (%d,%p) onto last in list\n", mem->size, mem); )
		lptr->next = mem;
	}
}

void
gc_primitive_free(gc_collector *c, gc_block* mem)
{
	jteamReservation *jr;
	
	require(c != 0);
	require(mem != 0);
	require(mem->collector == c);
	require(mem->firstblock != 0);
	
	if( (jr = pushTeamReservation(kernelTeam)) )
	{
		gc_primitive_free2(c, mem);
		popTeamReservation(jr);
	}
}

void
gc_primitive_free2(gc_collector *c, gc_block* mem)
{
	int iLockRoot;
	
	require(c != 0);
	require(mem != 0);
	require(mem->collector == c);
	require(mem->firstblock != 0);
	
	lockStaticMutex(&prim_lock);
	creditMemory(c->client, GCBLOCKSIZE(mem));
	c->local -= GCBLOCKSIZE(mem);
	assert(c->local >= 0);
	gc_primitive_free_1(mem);
	mem->collector = 0;
	unlockStaticMutex(&prim_lock);
}

/*
 * Allocate memory from the system and add it to the freelist
 */
static void*
gc_system_alloc(size_t sz)
{
	gc_block* blk = 0;

	assert(sz % gc_pgsize == 0);

	/* If we will pass the heap boundary, return 0 to indicate that
	 * we're run out.parent
	 */
	if (gc_heap_total + sz > gc_heap_limit) {
		return 0;
	}
#ifdef DEBUG
	gc_system_alloc_cnt++;
#endif

	blk = gc_block_alloc(sz);
	
DBG(GCSYSALLOC,
	dprintf("gc_system_alloc: %d byte at %p\n", sz, blk);		)

	if (blk == 0) {
		return (blk);
	}
	gc_heap_total += sz;
	prim_alloced += sz;

	/* Place block into the freelist for subsequent use */
	DBG(GCDIAG, blk->magic = GC_MAGIC);
	blk->size = sz;

	/* fake this to satisfy assert in primitive_free */
	blk->firstblock = blk;
	{
		gc_block *ptr, *nptr;
		ptr = blk;
		nptr = GCBLOCKEND(blk);
		while (ptr < nptr) {
			ptr->firstblock = blk;
			ptr++;
		}
	}

	/* Free block into the system */
	gc_primitive_free_1(blk);

	return (blk);
}

