/* gc-private.h
 * The heap manager.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */
/*
 * NB: this file contains only types and declaration relating to heap
 * management.  There's nothing relating to gc in here.
 */
#ifndef __gc_private_h
#define	__gc_private_h

#include "jthread.h"
#include "stats.h"
#include "hashtab.h"
#include <janosvm/memResource.h>
#include <janosvm/nativeClass.h>

#include <misc/memoryStack.h>

/* ------------------------------------------------------------------------ */

/*
 * This record describes an allocation type.
 */
typedef struct _gcFuncs {
        walk_func_t             walk;
        final_func_t            final;
        destroy_func_t          destroy;
	const char*		description;
        int                     nr;		/* only used ifdef STATS */
        int                     mem;		/* only used ifdef STATS */
} gcFuncs;

/*
 * Default values for initial and maximum heap size and heap increment.
 */
#define	MIN_HEAPSIZE	(5*1024*1024)
#define	MAX_HEAPSIZE	(64*1024*1024)
#define	ALLOC_HEAPSIZE	(1024*1024)

extern size_t gc_heap_total;
extern size_t gc_heap_allocation_size;
extern size_t gc_heap_limit;
extern char   *gc_alloc_failure;	/* error msg for debugging */

#ifdef DEBUG
extern int gc_system_alloc_cnt;
#endif

/* ------------------------------------------------------------------------ */

typedef struct _gc_freeobj {
	struct _gc_freeobj*	next;
} gc_freeobj;

/* ------------------------------------------------------------------------ */

#define	MIN_OBJECT_SIZE		8
#define	MAX_SMALL_OBJECT_SIZE	8192
#define	NR_FREELISTS		20
#define	GC_SMALL_OBJECT(S)	((S) <= max_small_object_size)

#if defined (__i386__) && (defined(STORE_HEAP_WITH_OBJECT) || defined(FAKE_HEAP_POINTERS))
#define	MEMALIGN		4
#else
#define	MEMALIGN		8
#endif

#define	ROUNDUPALIGN(V)		(((uintp)(V) + MEMALIGN - 1) & -MEMALIGN)
#define	ROUNDDOWNALIGN(V)	((uintp)(V) & -MEMALIGN)
#define	ROUNDUPPAGESIZE(V)	(((uintp)(V) + gc_pgsize - 1) & -gc_pgsize)

/* ------------------------------------------------------------------------ */

#define	GC_OBJECT_SIZE(M)	GCMEM2BLOCK(M)->size

/*
 * Per-heap GC statistics.
 */
typedef struct _gcStats {
        uint32  totalmem;	/* actual bytes in use by the application */
				/* adjusted by freed bytes after gc for gcable
				 * objects; adjusted by obj size for non-gcable
				 * objects when they're freed
				 */
        uint32  totalobj;	/* actual # of objs in use by the application */
        uint32  freedmem;
        uint32  freedobj;
        uint32  markedobj;
        uint32  markedmem;
        uint32  allocobj;	/* # of alloced objects since last gc */
        uint32  allocmem;	/* bytes in alloced objects since last gc */
        uint32  finalobj;
        uint32  finalmem;
} gcStats;

/* ------------------------------------------------------------------------ */

typedef struct _gc_unit {
        struct _gc_unit*        cprev;
        struct _gc_unit*        cnext;
} gc_unit;

typedef	gc_unit			gcList;

#define GC_COLOUR_MASK          0x0F
#define GC_COLOUR_INUSE         0x08
#define GC_COLOUR_FREE          0x00
#define GC_COLOUR_FIXED         0x01
#define GC_COLOUR_WHITE         0x08
#define GC_COLOUR_GREY          0x09
#define GC_COLOUR_BLACK         0x0A

#define GC_STATE_CLEAR          0xF0
#define GC_STATE_MASK           0x30
#define GC_STATE_NORMAL         0x00            /* Has no finalise method */
#define GC_STATE_FINALIZED      0x00            /* Has been finalised */
#define GC_STATE_NEEDFINALIZE   0x10            /* Needs finalising */
#define GC_STATE_INFINALIZE     0x20            /* Starting finalisation */

#define GC_SET_COLOUR(B, I, C) \
                (B)->state[I] = ((B)->state[I] & (~GC_COLOUR_MASK)) | (C)
#define GC_GET_COLOUR(B, I)     ((B)->state[I] & GC_COLOUR_MASK)

#define GC_SET_STATE(B, I, C) \
                (B)->state[I] = ((B)->state[I] & (~GC_STATE_CLEAR)) | (C)
#define GC_GET_STATE(B, I)      ((B)->state[I] & GC_STATE_MASK)

#define GC_SET_FUNCS(B, I, F)   (B)->funcs[I] = (F)
#define GC_GET_FUNCS(B, I)      (B)->funcs[I]

/* A simple macro to find the state byte. */
#define GC_STATE(B, I)          ((B)->state[I])

#define URESETLIST(LIST)	(LIST).cnext = &(LIST); \
				(LIST).cprev = &(LIST)

#define UAPPENDLIST(LIST, OBJ)	(OBJ)->cprev = (LIST).cprev; \
				(OBJ)->cnext = (LIST).cprev->cnext; \
				(LIST).cprev->cnext = (OBJ); \
				(LIST).cprev = (OBJ)

#define UREMOVELIST(OBJ)	(OBJ)->cprev->cnext = (OBJ)->cnext; \
				(OBJ)->cnext->cprev = (OBJ)->cprev; \
				(OBJ)->cprev = 0; \
				(OBJ)->cnext = 0

#define	UTOMEM(OBJ)		((void*)(((gc_unit*)(OBJ))+1))
#define	UTOUNIT(OBJ)		(((gc_unit*)(OBJ))-1)

/* ------------------------------------------------------------------------ */

typedef struct {
	gc_block* list;
	uint16	  sz;
} gc_freelist;

typedef struct {
	uint16	list;
} gc_sztable;

typedef struct _refObject {
	const void*		mem;
	unsigned int		ref;
	struct _refObject*	next;
} refObject;

/* This is a bit homemade.  We need a 7-bit hash from the address here */
#define	REFOBJHASH(V)	((((uintp)(V) >> 2) ^ ((uintp)(V) >> 9))%REFOBJHASHSZ)
#define	REFOBJHASHSZ	128

struct _gc_collector;

typedef struct _typeTracker {
	int allocCount;
	int freeCount;
	int peakUsage;
	int usage;
} typeTracker;

/*
 * This structure is shared by the system and per-process collectors,
 * but the walking code differs.
 */
typedef struct _gc_collector {
	Collector collector;
	struct _gc_collector *next; /* maintain a list of all heaps */

	iLock *lock;

	struct {
		iLock *lock;
		int *owner;
		volatile int gcRunning;	/* gc condition */
		void *thread;		/* gc thread */
	} gcman;
	
	struct {
		iLock *lock;
		volatile bool finalRunning; /* finalizer condition */
		void *fin_thread;	/* finalizer thread */
	} finman;

	memResourceClient *client;
	union {
		char *name;
		struct Hjava_lang_Object *obj;
	} policy;

	void *outOfMem;		/* preallocated java.lang.OutOfMemoryError */
	void *reserve;		/* up to 3 pages of memory reserved
				 * for handling outOfMem conditions
				 */
	void *outOfMem_allocator; /* thread currently grabbing it */
	
	gc_freelist freelist[NR_FREELISTS+1]; /* lists of pages w/free
					       objects */
	
	gcList lists[6];
	timespent gc_time;	/* total time in mark phase */
	timespent sweep_time;	/* total time in sweep phase */
	counter gcablemem;
	counter fixedmem;
	gcStats stats;		/* needed? */

	size_t local;		/* memory allocated local to this heap
				 * local <= limit->total.
				 */
	size_t trigger;		/* start gc when total >= trigger */

	int n_refs;		/* entries in hash */
	refObject *hash[REFOBJHASHSZ]; /* hash table of global refs */

	int dying;		/* debug early heap destroy */

	hashtab_t		stringTable;

	typeTracker track[GC_ALLOC_MAX_INDEX];

	memoryStack *primitiveStack;
} gc_collector;

extern gc_collector gc_metal_collector;
/* ------------------------------------------------------------------------ */

#define	GCBLOCK2STATE(B, N)	(&(B)->state[(N)])
#define	GCBLOCK2MEM(B, N)	((gc_unit*)(&(B)->data[(N)*(B)->size]))
#define	GCBLOCK2FREE(B, N)	((gc_freeobj*)GCBLOCK2MEM(B, N))
#define	GCBLOCKSIZE(B)		(B)->size

/* find first usable address in block */ 
#define GCBLOCK2BASE(B)	(((char *)gc_heap_base)		\
			 + gc_pgsize * ((B) - GC_BLOCKS))


/* This is OK, gc_prim_(alloc|free) never assume GCBLOCKEND is really
   a valid block */
#define GCBLOCKEND(B)		((B) + (((B)->size+gc_pgsize-1)>>gc_pgbits))
#define GCBLOCK_OVH		0

#define ASSERT_ONBLOCK(OBJ, BLK) assert(GCMEM2BLOCK(OBJ) == BLK)

extern void gc_heap_initialize(void);
extern void gc_mem_init(gc_collector *c);
extern void *gc_heap_malloc(gc_collector *c, size_t);    
extern void gc_heap_free(gc_collector *c, void*);

extern void gc_prim_initialize(void);
extern gc_block *gc_primitive_alloc(gc_collector *c, size_t);
extern gc_block *gc_primitive_reserve(gc_collector *c);
extern void gc_primitive_free(gc_collector *c, gc_block*);
extern void gc_primitive_free2(gc_collector *c, gc_block*);
extern void gc_release_blocks(gc_collector *me);

#if defined(DEBUG)
extern void gc_prim_map(void);
#endif

extern Collector *newCollector(memResourceClient *mrc, char *policyClass);
extern void gcKillMe(gc_collector *me);

#endif
