/*
 * module.h
 * A simple system for managing program modules
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _module_h
#define _module_h

/*
 * This is basically a poor man's Knit...  We provide a simple infrastructure
 * for getting initialization done in the proper order without having to
 * manually change the global initializer function.  It also provides a nice
 * path to test the modules after they've been initialized.
 *
 * This is only the runtime for the module code, see also moduleList.awk to
 * see how the static list of modules is built.
 */

/*
 * A module_t structure is used represent a module in the program.  Currently,
 * this just means it stores some function pointers for initializing, testing,
 * and finalizing the module.
 *
 * `name' - The name of the module.
 * `init' - The init function for the module.  If the module needs to be
 *   initialized in stages then the stage number is passed in the argument.
 * `kill' - The finalizer function for the module.
 * `test' - The tester function for the module.
 */
typedef struct _module {
	char *name;
	int staticID;
	int (*init)(int stage);
	void (*kill)(int stage);
	int (*test)(void);
	void (*diagnose)(void);
} module_t;

/*
 * A moduleInitializer_t structure is used to describe an initialization stage
 * for a module.  A null terminated array of these is passed t initModules
 * to initialize a set of modules/stages.
 *
 * `module' - Pointer to the module to initialize.
 * `stage' - The stage number for the module initialization.
 */
typedef struct _moduleInitializer {
	module_t *module;
	int stage;
} moduleInitializer_t;

/*
 * Initialize a group of modules.  If the argument is null, then it initializes
 * the global list of builtin modules.  Otherwise, it should be a NULL
 * terminated array ordered in the same way as the modules should be
 * initialized
 *
 * `mi' - A NULL terminated array of moduleInitializers that should be run.
 * returns - True if the all of the modules initialized correctly.
 */
int initModules(moduleInitializer_t *mi);
/*
 * Finalize the modules array, this basically just calls `module_t.kill' in
 * the reverse order that they were initialized.
 *
 * `mi' - The same NULL terminated array that was passed to initModules.
 */
void killModules(moduleInitializer_t *mi);
/*
 * Test the null terminated array of modules.  If `mods' is NULL then it
 * defaults to the builtin module list.
 *
 * `mods' - A NULL terminated array of module_t pointers.
 */
int testModules(module_t **mods);
/*
 * Run any module diagnostic functions.
 */
void diagnoseModules(module_t **mods);

#endif
