/*
 * mungWallProxy.c
 * A gcProxy that tries to detect overflow/underflow in memory buffers.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <assert_pp.h>

#include "mungWallProxy.h"

/* The primitive type to use when processing wall's */
#define MEMORY_WALL_TYPE long
/* The data to store in the wall */
#define MEMORY_WALL_DATA 0xDEADF00D
/* The number of MEMORY_WALL_TYPE elements that make up a wall */
#define MEMORY_WALL_SIZE 2

/* The primitive type to use when dumping garbage in memory */
#define MEMORY_GARBAGE_TYPE long
/* The data to dump in memory when its garbage */
#define MEMORY_GARBAGE_DATA 0xDEADBEEF

#define ADJUST_SIZE(x) (((x) + sizeof( long ) - 1) & ~(sizeof( long ) - 1))

struct mMemoryHeader {
	size_t mh_Size;
};

/*
 * Fill in the given memory with data used for a wall.
 *
 * `wall' - Pointer to allocated memory that must be atleast
 *   sizeof(MEMORY_WALL_TYPE) * MEMORY_WALL_SIZE bytes in size.
 * returns - The start of memory after the wall.
 */
static
void *mungBuildWall(MEMORY_WALL_TYPE *wall)
{
	int lpc;

	require( wall != 0 );

	for( lpc = 0; lpc < MEMORY_WALL_SIZE; lpc++ )
	{
		*wall = MEMORY_WALL_DATA;
		wall++;
	}
	return( wall );
}

/*
 * Fill in the given memory block with garbage data.
 *
 * `mem' - The memory to be trashed.
 * `size' - The size, in bytes, of `mem'.
 */
static
void mungDumpGarbage(MEMORY_GARBAGE_TYPE *mem, int size)
{
	int lpc;

	size /= sizeof(MEMORY_GARBAGE_TYPE);
	for( lpc = 0; lpc < size; lpc++ )
	{
		mem[lpc] = MEMORY_GARBAGE_DATA;
	}
}

/*
 * Check a single wall to make sure the data that was filled in when the
 * memory was allocated is still intact.
 *
 * `wall' - Pointer to allocated memory that must be atleast
 *   sizeof(MEMORY_WALL_TYPE) * MEMORY_WALL_SIZE bytes in size.
 * returns - True if the wall is still intact, false otherwise.
 */
static
int mungCheckWall(MEMORY_WALL_TYPE *wall)
{
	int retval;
	int lpc;

	require( wall != 0 );
  
	retval = 1;
	for( lpc = 0; retval && (lpc < MEMORY_WALL_SIZE); lpc++ )
	{
		if( *wall != MEMORY_WALL_DATA )
			retval = 0;
		wall++;
	}
	return( retval );
}

void mungValidateMemory(void *storage)
{
	memResourceClient *owner = storage ? gc_get_resource(storage) : 0;
	int type = (storage && owner) ?
		GC_getObjectIndex((Collector *)
				  teamCollector(owner->client.owner),
				  storage) : 0;

	if( storage && owner && (type != -1) )
	{
		int wall_size = sizeof( MEMORY_WALL_TYPE ) * MEMORY_WALL_SIZE;
		struct mMemoryHeader *mh;
		char *ptr;
		
		switch( type )
		{
		case GC_ALLOC_JAVASTRING:
		case GC_ALLOC_NOWALK:
		case GC_ALLOC_NORMALOBJECT:
		case GC_ALLOC_PRIMARRAY:
		case GC_ALLOC_REFARRAY:
		case GC_ALLOC_REF:
		case GC_ALLOC_CLASSOBJECT:
		case GC_ALLOC_FINALIZEOBJECT:
		case GC_ALLOC_JAVALOADER:
		case GC_ALLOC_TRAMPOLINE:
			assert(0);
			break;
		default:
			break;
		}
		ptr = storage;
		mh = (struct mMemoryHeader *)(ptr -
					      wall_size -
					      sizeof(struct mMemoryHeader));
		/* Check the walls on either side... */
		if( !mungCheckWall((MEMORY_WALL_TYPE *)(ptr - wall_size) ) ||
		    !mungCheckWall((MEMORY_WALL_TYPE *)(ptr + mh->mh_Size) ) )
		{
			int total_size =
				wall_size +
				ADJUST_SIZE(sizeof(struct mMemoryHeader) +
					    mh->mh_Size) +
				wall_size;
			
			printf("Broken wall in memory at %p", storage);
			dPrintMemory(mh, total_size);
			abort();
		}
	}
}

/*
 * Wrapper function for malloc, if `type' is a non GC'ed allocation then it
 * will bracket the allocated memory with walls to help detect underflow /
 * overflow.
 *
 * `gc' - The collector to allocate the memory from.
 * `size' - The amount of bytes to allocate, in the case of a non GC'ed
 *   this size will be increased to accomodate the walls.
 * `type' - The type of memory to allocate.
 * returns - The newly allocated memory or NULL if the request failed.
 */
static
void *mungMalloc(Collector *gc, size_t size, int type)
{
	int wall_size = sizeof( MEMORY_WALL_TYPE ) * MEMORY_WALL_SIZE;
	gcProxy *gp = (gcProxy *)gc->ops;
	int real_size;
	void *retval;
	
	switch( type )
	{
	case GC_ALLOC_JAVASTRING:
	case GC_ALLOC_NOWALK:
	case GC_ALLOC_NORMALOBJECT:
	case GC_ALLOC_PRIMARRAY:
	case GC_ALLOC_REFARRAY:
	case GC_ALLOC_REF:
	case GC_ALLOC_CLASSOBJECT:
	case GC_ALLOC_FINALIZEOBJECT:
	case GC_ALLOC_JAVALOADER:
	case GC_ALLOC_TRAMPOLINE:
		retval = gp->originalOps->malloc(gc, size, type);
		break;
	default:
		size = ADJUST_SIZE(size);
		real_size = ADJUST_SIZE(size + sizeof(struct mMemoryHeader)) +
			wall_size * 2;
		if( (retval = gp->originalOps->malloc(gc, real_size, type)) )
		{
			struct mMemoryHeader *mh;
			char *ptr;

			mh = retval;
			mh->mh_Size = size;
			ptr = ((char *)mh) + sizeof(struct mMemoryHeader);
			retval = mungBuildWall((MEMORY_WALL_TYPE *)ptr);
			ptr += wall_size + size;
			mungBuildWall((MEMORY_WALL_TYPE *)ptr);
		}
		break;
	}
	return( retval );
}

/*
 * Wrapper function for realloc, if `type' is a non GC'ed allocation then it
 * will validate the old memory, realloc it, and move the walls accordingly.
 *
 * `gc' - The collector to allocate the memory from.
 * `mem' - The old memory to reallocate.
 * `size' - The amount of bytes to allocate, in the case of a non GC'ed
 *   this size will be increased to accomodate the walls.
 * `type' - The type of memory to allocate.
 * returns - The newly allocated memory or NULL if the request failed.
 */
static
void *mungRealloc(Collector *gc, void *mem, size_t size, int type)
{
	int wall_size = sizeof( MEMORY_WALL_TYPE ) * MEMORY_WALL_SIZE;
	gcProxy *gp = (gcProxy *)gc->ops;
	void *retval = 0;
	
	switch( type )
	{
	case GC_ALLOC_JAVASTRING:
	case GC_ALLOC_NOWALK:
	case GC_ALLOC_NORMALOBJECT:
	case GC_ALLOC_PRIMARRAY:
	case GC_ALLOC_REFARRAY:
	case GC_ALLOC_REF:
	case GC_ALLOC_CLASSOBJECT:
	case GC_ALLOC_FINALIZEOBJECT:
	case GC_ALLOC_JAVALOADER:
	case GC_ALLOC_TRAMPOLINE:
		ensure(0);
		break;
	default:
		if( mem )
		{
			struct mMemoryHeader *mh;
			int old_size, real_size;
			char *ptr;

			mungValidateMemory(mem);
			ptr = mem;
			mh = (struct mMemoryHeader *)
				(ptr -
				 wall_size -
				 sizeof(struct mMemoryHeader));
			old_size = mh->mh_Size;
			size = ADJUST_SIZE(size);
			real_size = ADJUST_SIZE(size +
						sizeof(struct mMemoryHeader)) +
				wall_size * 2;
			if( (retval = gp->originalOps->realloc(gc,
							       mh,
							       real_size,
							       type)) )
			{
				mh = retval;
				mh->mh_Size = size;
				ptr = ((char *)mh) +
					sizeof(struct mMemoryHeader) +
					wall_size +
					old_size;
				memset(ptr, 0, wall_size);
				ptr = ((char *)mh) +
					sizeof(struct mMemoryHeader) +
					wall_size +
					size;
				mungBuildWall((MEMORY_WALL_TYPE *)ptr);
				retval = ((char *)mh) +
					sizeof(struct mMemoryHeader) +
					wall_size;
				mungValidateMemory(retval);
			}
		}
		else
		{
			retval = mungMalloc(gc, size, type);
		}
		break;
	}
	return( retval );
}

/*
 * Wrapper function for free, if the memory being freed isn't GC'ed then it
 * will validate it, dump some garbage, and then actually free the memory.
 *
 * `gc' - The collector that was used to allocate `mem'.
 * `mem' - The memory to be freed.
 */
static
void mungFree(Collector *gc, void *mem)
{
	int wall_size = sizeof( MEMORY_WALL_TYPE ) * MEMORY_WALL_SIZE;
	gcProxy *gp = (gcProxy *)gc->ops;
	struct mMemoryHeader *mh;
	char *ptr;
	int type;

	if( !mem )
		return;
	type = gp->originalOps->getObjectIndex(gc, mem);
	switch( type )
	{
	case GC_ALLOC_JAVASTRING:
	case GC_ALLOC_NOWALK:
	case GC_ALLOC_NORMALOBJECT:
	case GC_ALLOC_PRIMARRAY:
	case GC_ALLOC_REFARRAY:
	case GC_ALLOC_REF:
	case GC_ALLOC_CLASSOBJECT:
	case GC_ALLOC_FINALIZEOBJECT:
	case GC_ALLOC_JAVALOADER:
	case GC_ALLOC_TRAMPOLINE:
		gp->originalOps->free(gc, mem);
		break;
	default:
		mungValidateMemory(mem);
		ptr = mem;
		mh = (struct mMemoryHeader *)(ptr -
					      wall_size -
					      sizeof(struct mMemoryHeader));
		mungDumpGarbage((MEMORY_GARBAGE_TYPE *)mh,
				sizeof(struct mMemoryHeader) +
				mh->mh_Size + wall_size * 2);
		gp->originalOps->free(gc, mh);
		break;
	}
}

/*
 * Wrapper function for GC_getObjectBase which adjusts the base pointer to
 * account for the wall at the beginning of the chunk of memory.
 *
 * `gc' - The collector that allocated `mem'.
 * `mem' - A possibly internal memory pointer that we should get the base for.
 * returns - The base pointer of `mem'.
 */
static
void *mungObjectBase(Collector *gc, void *mem)
{
	int wall_size = sizeof( MEMORY_WALL_TYPE ) * MEMORY_WALL_SIZE;
	gcProxy *gp = (gcProxy *)gc->ops;
	void *retval;
	char *ptr;
	int type;
	
	type = gp->originalOps->getObjectIndex(gc, mem);
	switch( type )
	{
	case GC_ALLOC_JAVASTRING:
	case GC_ALLOC_NOWALK:
	case GC_ALLOC_NORMALOBJECT:
	case GC_ALLOC_PRIMARRAY:
	case GC_ALLOC_REFARRAY:
	case GC_ALLOC_REF:
	case GC_ALLOC_CLASSOBJECT:
	case GC_ALLOC_FINALIZEOBJECT:
	case GC_ALLOC_JAVALOADER:
	case GC_ALLOC_TRAMPOLINE:
		retval = gp->originalOps->getObjectBase(gc, mem);
		break;
	default:
		ptr = ((char *)gp->originalOps->getObjectBase(gc, mem)) +
			sizeof(struct mMemoryHeader) + wall_size;
		retval = ptr;
		break;
	}
	return( retval );
}

/*
 * Wrapper function for GC_getObjectSize which adjusts the size of the memory
 * to account for the walls.
 *
 * `gc' - The collector that was used to allocate `mem'.
 * `mem' - The memory block to get the size of.
 * returns - The size of the memory pointed to by `mem'.
 */
static
int mungObjectSize(Collector *gc, void *mem)
{
	int wall_size = sizeof( MEMORY_WALL_TYPE ) * MEMORY_WALL_SIZE;
	gcProxy *gp = (gcProxy *)gc->ops;
	int type, retval;
	
	type = gp->originalOps->getObjectIndex(gc, mem);
	switch( type )
	{
	case GC_ALLOC_JAVASTRING:
	case GC_ALLOC_NOWALK:
	case GC_ALLOC_NORMALOBJECT:
	case GC_ALLOC_PRIMARRAY:
	case GC_ALLOC_REFARRAY:
	case GC_ALLOC_REF:
	case GC_ALLOC_CLASSOBJECT:
	case GC_ALLOC_FINALIZEOBJECT:
	case GC_ALLOC_JAVALOADER:
		retval = gp->originalOps->getObjectSize(gc, mem);
		break;
	default:
		retval = gp->originalOps->getObjectSize(gc, mem) -
			wall_size * 2 - sizeof(struct mMemoryHeader);
		break;
	}
	return( retval );
}

int addMungWallProxy(Collector *gc)
{
	int retval = 0;

	if( wrapCollector(gc, sizeof(gcProxy),
			  GPA_Malloc, mungMalloc,
			  GPA_Realloc, mungRealloc,
			  GPA_Free, mungFree,
			  GPA_GetObjectBase, mungObjectBase,
			  GPA_GetObjectSize, mungObjectSize,
			  GPA_TAG_DONE) )
	{
		retval = 1;
	}
	return( retval );
}

int addMungWallProxy_mem(Collector *gc, gcProxy *mem)
{
	int retval = 0;

	if( wrapCollector_mem(gc, mem,
			      GPA_Malloc, mungMalloc,
			      GPA_Realloc, mungRealloc,
			      GPA_Free, mungFree,
			      GPA_GetObjectBase, mungObjectBase,
			      GPA_GetObjectSize, mungObjectSize,
			      GPA_TAG_DONE) )
	{
		retval = 1;
	}
	return( retval );
}
