/*
 * stackTrace.c
 * Handle stack trace for the interpreter or translator.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "config-std.h"
#include "config-signal.h"
#include "config-mem.h"
#include "config-setjmp.h"
#include "debug.h"
#include "jtypes.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "classMethod.h"
#include "code.h"
#include "exception.h"
#include "baseClasses.h"
#include "lookup.h"
#include "thread.h"
#include "errors.h"
#include "itypes.h"
#include "external.h"
#include "soft.h"
#include "md.h"
#include "stackTrace.h"
#include "support.h"

#include <janosvm/classGroup.h>
#include <janosvm/nameSpace.h>
#include <janosvm/logResource.h>
#include <janosvm/stdioResource.h>

#include <misc/objectStack.h>

static
Hjava_lang_Object*
_buildStackTrace(struct _exceptionFrame* base, stackTraceInfo *_info)
{
      int cnt;
      struct _stackTrace trace;
      stackTraceInfo* info = _info;
      struct _exceptionFrame orig;

      (void) orig;                    /* avoid compiler warning in intrp */
      DBG(STACKTRACE,
          dprintf("STACKTRACEINIT(trace, base, base, orig);\n"); );
      STACKTRACEINIT(trace, base, base, orig);
      cnt = 0;
      while(!STACKTRACEEND(trace)) {
              STACKTRACESTEP(trace);
              cnt++;
      }

      /* Build an array of stackTraceInfo */
      if (info == 0) {
	      /*
	       * NOTE:  We're allocating a Java object that is expected to be
	       * collected by the GC and not explicitly freed.
	       */
              info = gc_malloc(sizeof(stackTraceInfo) * (cnt+1),
			       GC_ALLOC_NOWALK);
              if (!info)
	      {
		      return 0;
	      }
      }

      cnt = 0;

      DBG(STACKTRACE,
          dprintf("STACKTRACEINIT(trace, &orig, base, orig);\n"); );
      STACKTRACEINIT(trace, &orig, base, orig);

      for(; !STACKTRACEEND(trace); STACKTRACESTEP(trace)) {
              info[cnt].pc = STACKTRACEPC(trace);
              info[cnt].fp = STACKTRACEFP(trace);
              info[cnt].meth = STACKTRACEMETHCREATE(trace);
              cnt++;
      }
      info[cnt].pc = 0;
      info[cnt].meth = ENDOFSTACK;

      return ((Hjava_lang_Object*)info);
}

/*
 * XXX threadDumpStackTrace is unix-jthread specific, so disable it.
 * Its only a debugging hack for debugging deadlocks without GDB.
 */
#if 0
#if defined(__i386__) && defined(linux)
static stackTraceInfo staticTInfo[100];
#endif
#endif /* 0 */

void
threadDumpStackTrace(jthread_t jtid)
{
#if 0
#if defined(__i386__) && defined(linux)
        Hjava_lang_Throwable t;
        struct _exceptionFrame eframe;
        // jthread_t c = currentJThread;
	jthread_t c = jthread_current();
        int self = c == jtid;
	
        if (!self) {
                eframe.retbp = jtid->env[0].__jmpbuf[JB_BP];
                eframe.retpc = jtid->env[0].__jmpbuf[JB_PC];
        }
        currentJThread = jtid;
        unhand(&t)->backtrace = _buildStackTrace(self ? 0 : &eframe,
                                                 staticTInfo);
        printStackTrace(&t, 0, 1);
        currentJThread = c;
#endif
#endif /* 0 */
}

Hjava_lang_Object*
buildStackTrace(struct _exceptionFrame* base)
{
	return _buildStackTrace(base, 0);
}

/*
 * We make these exported functions because we use them in classMethod.c
 */
#if defined(TRANSLATOR)
Method*
stacktraceFindMethod(stackTraceInfo *info)
{
	return (findMethodFromPC(info->pc));
}

#elif defined(INTERPRETER)

Method*
stacktraceFindMethod(stackTraceInfo *info)
{
	return (info->meth);
}
#endif

void
printStackTrace(struct Hjava_lang_Throwable* o,
		struct Hjava_lang_Object* p, int nullOK)
{
	int i;
	stackTraceInfo* info;
	Method* meth;
	uintp pc;
	int32 linenr;
	uintp linepc;
	char *buf;
	int len;
	int j;
	jchar* cptr;
	char* class_dot_name;
	errorInfo einfo;

	info = (stackTraceInfo*)unhand(o)->backtrace;
	if (info == 0) {
		teamLogEntry("stackTrace",
			     LEA_String, "Note", "No backtrace in exception", -1,
			     LEA_TAG_DONE);
		return;
	}
	for (i = 0; info[i].meth != ENDOFSTACK; i++) {
		pc = info[i].pc;
		meth = stacktraceFindMethod(&info[i]);
		if (meth != 0) {
			classView *cv;
			
			linepc = 0;
			linenr = -1;
			if (meth->lines != 0) {
				for (j = 0; j < meth->lines->length; j++) {
					if (pc >= meth->lines->entry[j].start_pc && linepc <= meth->lines->entry[j].start_pc) {
						linenr = meth->lines->entry[j].line_nr;
						linepc = meth->lines->entry[j].start_pc;
					}
				}
			}
			/* Even if we are reporting an out of memory and
			   checkPtr fails, this is ok.  If we can't allocate
			   a new one, the vm will die in an orderly manner.

			   XXX NO! TWINK, its not OK
			*/
			cv = getClassView(meth->class);
			class_dot_name =
			    gc_malloc(strlen(cv->name->data) + 1,
				      GC_ALLOC_STACK_TRACE);
			if( !class_dot_name )
				return;
			pathname2classname(cv->name->data, class_dot_name);
			buf = gc_malloc(strlen(class_dot_name)
					+ strlen(meth->name->data)
					+ strlen(CLASS_SOURCEFILE(meth->class))
					+ 64,
					GC_ALLOC_STACK_TRACE);
			if( !buf )
			{
				gc_free(class_dot_name);
				return;
			}

			if (linenr == -1) {
				if (meth->accflags & ACC_NATIVE) {
					sprintf(buf, "\tat %s.%s(%s:native)",
						class_dot_name,
						meth->name->data, 
						CLASS_SOURCEFILE(meth->class));
				}
				else {
					sprintf(buf, "\tat %s.%s(%s:line unknown, pc %p)",
						class_dot_name,
						meth->name->data, 
						CLASS_SOURCEFILE(meth->class),
						(void*)pc);
				}
			}
			else {
				sprintf(buf, "\tat %s.%s(%s:%d)",
					class_dot_name,
					meth->name->data,
					CLASS_SOURCEFILE(meth->class),
					linenr);
			}
			gc_free(class_dot_name);
			if (p != 0 || !nullOK) {
				Hjava_lang_Object* str;
				
				len = strlen(buf);
				str = newArrayChecked(TYPE_CLASS(TYPE_Char),
						      len,
						      &einfo);
				if( !str )
				{
					gc_free(buf);
					if (nullOK) {
						return;
					} else {
						throwError(&einfo);
					}
				}
				cptr = (jchar*)OBJARRAY_DATA(str);
				for (j = len;  --j >= 0; ) {
					cptr[j] = (unsigned char)buf[j];
				}
				do_execute_java_method(p, "println",
					"([C)V", 0, 0, str);
			} else {
				stdioPrint(STDIO_ERR, "%s\n", buf);
			}
			gc_free(buf);
		}
	}
	if (p != 0 || !nullOK) {
		do_execute_java_method(p, "flush", "()V", 0, 0);
	}
}

static
void getLineNumber(Method *meth, uintp pc, int32 *linenr, uintp *linepc)
{
	*linenr = -1;
	*linepc = 0;
	if( meth->lines != 0 )
	{
		int lpc;
		
		for( lpc = 0; lpc < meth->lines->length; lpc++ )
		{
			if( (pc >= meth->lines->entry[lpc].start_pc) &&
			    (*linepc <= meth->lines->entry[lpc].start_pc) )
			{
				*linenr = meth->lines->entry[lpc].line_nr;
				*linepc = meth->lines->entry[lpc].start_pc;
			}
		}
	}
}

int traceMethod(objectStack *os, uintp pc, Method *meth)
{
	char *class_dot_name;
	int retval = 0;
	classView *cv;
	uintp linepc;
	int32 linenr;

	cv = getClassView(meth->class);
	if( (class_dot_name = gc_malloc(strlen(cv->name->data) + 1,
					GC_ALLOC_STACK_TRACE)) )
	{
		getLineNumber(meth, pc, &linenr, &linepc);
		pathname2classname(cv->name->data, class_dot_name);
		retval = pushStackObjects(os,
					  OSA_String, "\tat ",
					  OSA_String, class_dot_name,
					  OSA_Char, '.',
					  OSA_String, meth->name->data,
					  OSA_Char, '(',
					  OSA_String, CLASS_SOURCEFILE(meth->
								       class),
					  OSA_TAG_DONE);
		if( linenr == -1 )
		{
			if (meth->accflags & ACC_NATIVE)
			{
				retval = retval && pushStackObjects(
					os,
					OSA_String, ":native)\n",
					OSA_TAG_DONE);
			}
			else
			{
				retval = retval && pushStackObjects(
					os,
					OSA_String, ":line unknown, pc ",
					OSA_PointerString, pc,
					OSA_String, ")\n",
					OSA_TAG_DONE);
			}
		}
		else
		{
			retval = retval && pushStackObjects(
				os,
				OSA_Char, ':',
				OSA_IntegerString, linenr,
				OSA_String, ")\n",
				OSA_TAG_DONE);
		}
		gc_free(class_dot_name);
	}
	return( retval );
}

Hjava_lang_Object *
stackTraceString(struct Hjava_lang_Throwable *o)
{
	Hjava_lang_Object *retval = NULL;
	stackTraceInfo *info;
	objectStack *os;
	
	info = (stackTraceInfo *)unhand(o)->backtrace;
	if( info &&
	    (os = createObjectStack(1024)) )
	{
		int len, lpc, error = 0;
		Method *meth;
		jchar *cptr;
		char *str;
		uintp pc;

		for( lpc = 0; !error && (info[lpc].meth != ENDOFSTACK); lpc++ )
		{
			pc = info[lpc].pc;
			if( (meth = stacktraceFindMethod(&info[lpc])) )
			{
				error = !traceMethod(os, pc, meth);
			}
		}
		if( !error && (str = topStackObject(os, &len)) )
		{
			if( (retval = newArray(TYPE_CLASS(TYPE_Char), len)) )
			{
				cptr = (jchar *)OBJARRAY_DATA(retval);
				for( lpc = 0; lpc < len; lpc++ )
				{
					cptr[lpc] = (unsigned char)str[lpc];
				}
			}
			popStackObject(os, str, len);
		}
		deleteObjectStack(os);
	}
	return( retval );
}
