/*
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah.  All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file.
 *
 * Written by the Flux Research Group at the University of Utah.
 * Authors: Godmar Back, Leigh Stoller, Patrick Tullmann
 */

/*
 * jthread creation functions
 */

#include "anjthread_types.h"
#include "anjthread_types_priv.h"
#include "gc.h"
#include "anjthread_globals.h"
#include "anjthread_intr.h"
#include "anjthread_create.h"
#include "debug.h"

#include <janosvm/jteamReservation.h>

#if defined(THREAD_ANALYZER)
#include <janosvm/leak.h>
#include <janosvm/logResource.h>

void threadAnalyzerFunction(struct _logAnalyzerMask *lam,
			    struct _logger *l,
			    struct _logEntry *le,
			    struct _logEntryAttribute **attrs)
{
	jthread_t th = (jthread_t)lam;

	if( !le->classification )
		return;
	if( !strcmp(le->classification, "leak") )
	{
		if( (attrs[1]->value.ptr == th->jlThread) )
		{
			if( !strcmp(attrs[0]->id, "unregister-leak") )
			{
				assert(th->status == JTHREAD_STATUS_INNODEOS);
			}
			le->classification = 0;
		}
		else
		{
			Hjava_lang_Thread *jlt = attrs[1]->value.ptr;
			
			th = (jthread_t)jlt->PrivateInfo;
			if( th->threadData.ownerTeam != currentTeam() )
				le->classification = 0;
		}
	}
}

static
logAnalyzer threadAnalyzer = {
	threadAnalyzerFunction
};
#endif

static void
anjthread_startthread(void *arg) __NORETURN__;

/*
 * Start function for explicitly created threads.  Most threads
 * are "partial" threads, and they come in through the function
 * anjthread_enterkaffe()
 *
 * This function install the cleanup handler, sets jthread-specific 
 * data and calls the actual work function.
 */
static void
anjthread_startthread(void *arg)
{
	jthread_t jtid = (jthread_t)arg;

DBG(JTHREAD, 
	dprintf("anjthread_startthread(jth=%p)\n", jtid); 
    )
	assert(jtid);
	assert(jtid->stackptr);

	an_mutex_lock(&jthread_creationLock);
	an_thread_sethandler(_jthread_intr_handler);
	DBGIF(an_thread_setdata(jtid));
	an_mutex_unlock(&jthread_creationLock);

	/*
	 * Change thread status.  Check to make sure thread
	 * wasn't killed/interrupted before starting.
	 */
	jthread_spinon(jtid);
	assert(jtid->status == JTHREAD_STATUS_NEWBORN);
	jtid->status = JTHREAD_STATUS_RUNNING;
	if ((jtid->flags & JTHREAD_FLAG_INTR_MASK) != 0)
		jthread_handleInterrupt(jtid);
	jthread_spinoff(jtid);

DBG(JTHREAD, 
	dprintf("anjthread_startthread(jth=%p) calling func (%p)\n",
		jtid, jtid->func); 
    )

	/* Invoke the thread's actual start function */
	jtid->func(jtid->jlThread);

	panic("firstStartThread has returned");
}

/*
 * Create a new jthread and explicitly start it running the given
 * callback.
 */
jthread_t
jthread_create(unsigned int pri, void (*func)(void *), int daemon,
	       void *jlThread, size_t threadStackSize)
{
	jthread_t tid;
	void* th;
	int rc;

	/* XXX Check against OSKit minimum stack limit */
	if (threadStackSize < (5 * 0x1000))
		threadStackSize = (5 * 0x1000); 

	/* Allocate the tid */
        tid = gc_malloc(sizeof (*tid), GC_ALLOC_THREADCTX);
	if (tid == NULL)
		return NULL; /* XXX good enough? */

	/* Allocate the thread's stack */
	tid->stackptr = gc_malloc(threadStackSize, GC_ALLOC_THREADSTACK);
	if (tid->stackptr == NULL)
	{
		gc_free(tid);
		return NULL;
	}

	/* Basic tid initialization */
        tid->jlThread = jlThread;
        tid->func = func; /* XXX duplicated at too many levels. */
        tid->status = JTHREAD_STATUS_NEWBORN;

	/* Setup the new thread's Team data */
	rc = inheritThreadData(&tid->threadData,
				&(jthread_current()->threadData),
				tid);
	if (rc != 1)
	{
		gc_free(tid->stackptr);
		gc_free(tid);
		return NULL;
	}

	/* Create the NodeOS thread. */
	th = an_thread_create(&(tid->native_thread),
		tid->stackptr, threadStackSize);
	assert(th == &(tid->native_thread));

#if defined(THREAD_ANALYZER)
	tid->lam.mask = 0;
	tid->lam.analyzer = &threadAnalyzer;
	addTeamLogAnalyzer(&tid->lam);
	registerLeak(jlThread, "jt");
#endif
	
	an_mutex_lock(&jthread_creationLock);
	
        tid->nextlive = jthread_liveThreads;
        jthread_liveThreads = tid;

        jthread_talive++;       
        if ((tid->daemon = daemon) != 0) {
                jthread_tdaemon++;
        }

	/* Start the NodeOS thread. */
	rc = an_thread_start(&(tid->native_thread), an_domain_current(),
			      anjthread_startthread, tid);
	if (rc != 0)
	{
		an_mutex_unlock(&jthread_creationLock);
		gc_free(tid->stackptr);
		gc_free(tid);
		return NULL; /* will be returned after unlocking */
	}

	an_mutex_unlock(&jthread_creationLock);

DBG(JTHREAD,
	dprintf("jthread_create(pri=%d, daemon=%d) -> jth=%p nth=%p\n",
		pri, daemon, tid, &(tid->native_thread));
    )
	
	assert(tid);
        return (tid);
}

static void errorInPartialThread(void *arg)
{
	panic("errorInPartialThread()");
}

/*
 * Create a new jthread, but do not start it.  Partial threads
 * are queued up in the NodeOS, on a threadpool, and will be
 * injected back into the VM through a callback function
 * (usu. on an inChannel).  All callbacks should come
 * anjthread_enterkaffe().
 *
 * XXX should return int error code!
 */
void
anjthread_createpartialthread(jthread_t tid,
			      void *jlThread,
			      size_t threadStackSize)
{
	void* th;
	int err;
	
        assert(tid);
        assert(jlThread);
	/* assert(threadStackSize); */

	/*
	 * Partial threads aren't really alive, so they don't get
	 * added to the liveThread queue.  Partial threads don't get
	 * started even.  This means we don't need to take the
	 * jthread_creationLock.
	 */

        tid->jlThread = jlThread;
        tid->func = errorInPartialThread;

	/* Create the NodeOS thread object. */

	/* XXX Check against OSKit minimum stack limit */
	if (threadStackSize < (5 * 0x1000))
		threadStackSize = (5 * 0x1000); 

	tid->stackptr = gc_malloc(threadStackSize,
		GC_ALLOC_THREADSTACK);
	assert(tid->stackptr); /* XXX handle this! */

        tid->status = JTHREAD_STATUS_INNODEOS;
	th = an_thread_create(&(tid->native_thread),
		tid->stackptr, threadStackSize);
	assert(th == &(tid->native_thread));

	/*
	 * Setup the thread's Team state.
	 */
	err = inheritThreadData(&tid->threadData,
				&(jthread_current()->threadData),
				tid);
	assert(err != 0); /* XXX handle this! */

DBG(JTHREAD,
    dprintf(__FUNCTION__ "(tid=%p, jlThread=%p, threadStackSize=%d) done.\n",
	    tid, jlThread, threadStackSize);
    )
}


/**
 * This function never returns.  All partial threads should enter
 * the VM through this function.
 */
void
anjthread_enterkaffe(void *arg, jmethodID runmethod, ...)
{
	struct Hjava_lang_Object* startArg = arg;
	jthread_t tid = jthread_current();
	va_list argptr;

DBG(JTHREAD,
    	dprintf(__FUNCTION__ "(arg=%p, meth=%p, ...) jtid=%p\n",
		arg, runmethod, tid);
    )
	assert(tid->stackptr);
        tid->daemon = 0;

	an_mutex_lock(&jthread_creationLock);

	an_thread_sethandler(_jthread_intr_handler);

	/* From jthread_create: */
        tid->nextlive = jthread_liveThreads;
        jthread_liveThreads = tid;

	/*
	 * XXX do *not* increment talive here.  Our caller does that.
	 * (For implicit threads, the immediate calling function does
	 * it; for explicit threads, the function that kicks this
	 * thread does it.
	 */

	an_mutex_unlock(&jthread_creationLock);
	
	/*
	 * Change thread status.  Check to make sure thread
	 * wasn't killed/interrupted before starting.
	 */
	jthread_spinon(tid);
	assert(tid->status == JTHREAD_STATUS_INNODEOS);
        tid->status = JTHREAD_STATUS_RUNNING;
	if ((tid->flags & JTHREAD_FLAG_INTR_MASK) != 0)
		jthread_handleInterrupt(tid);
	jthread_spinoff(tid);


	/* XXX assert that thread's team state is in reasonable shape. */

	/* XXX check that team isn't dying? */

	/* XXX function defined in ../../thread.c: */
	allowTeamThread(currentTeam());
	va_start(argptr, runmethod);
	threadEnterVM_valist(jthread_getcookie(jthread_current()),
			     startArg, runmethod, argptr);
	va_end(argptr);

	panic("threadEnterVM_valist should not return");
}

void
anjthread_enterkaffe_c(void)
{
	jthread_t tid = jthread_current();
	int allowed;
	
	assert(tid->stackptr);
        tid->daemon = 0;

	an_mutex_lock(&jthread_creationLock);

	an_thread_sethandler(_jthread_intr_handler);

	/* From jthread_create: */
        tid->nextlive = jthread_liveThreads;
        jthread_liveThreads = tid;

	/*
	 * XXX do *not* increment talive here.  Our caller does that.
	 * (For implicit threads, the immediate calling function does
	 * it; for explicit threads, the function that kicks this
	 * thread does it.
	 */

	an_mutex_unlock(&jthread_creationLock);
	
	/*
	 * Change thread status.  Check to make sure thread
	 * wasn't killed/interrupted before starting.
	 */
	jthread_spinon(tid);
	assert(tid->status == JTHREAD_STATUS_INNODEOS);
        tid->status = JTHREAD_STATUS_RUNNING;
	if ((tid->flags & JTHREAD_FLAG_INTR_MASK) != 0)
		jthread_handleInterrupt(tid);
	jthread_spinoff(tid);

	allowed = allowTeamThread(currentTeam());
	assert(allowed);
	pushTeamReservation(currentTeam());
}

void
anjthread_exitkaffe_c(void)
{
	jthread_t tid = jthread_current();
	
	rollbackReservationStack((void *)~0, (void *)0, (void *)~0);
	unregisterTeamThread(currentTeam(),
			     jthread_current(),
			     tid->daemon);
	jthread_slaughter(tid);
}
