/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah.  All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file.
 *
 * Contributed by the Flux Research Group at the University of Utah.
 * Authors: Godmar Back, Leigh Stoller, Patrick Tullmann
 */
/*
 * Jthread end-of-life functions
 */

#include "anjthread_types.h"
#include "anjthread_types_priv.h"
#include "gc.h"
#include "anjthread_current.h"
#include "anjthread_globals.h"
#include "anjthread_end.h"
#include "debug.h"

#include <janosvm/logResource.h>

/*
 * Have the current thread die by throwing a ThreadDeath
 * exception.
 *
 * Does not return.
 */
void
jthread_die(void)
{
	jthread_t jtid = jthread_currentJthread();

	assert(jtid);

DBG(JTHREAD,
	dprintf("jthread_die() jthd=%p\n", jtid);
    )

	jthread_spinon(jtid);
	jtid->flags |= JTHREAD_FLAG_DYING;
	jthread_spinoff(jtid);

	assert(jthread_onstop);
	(*jthread_onstop)();

	panic("jthread_onstop returned");
}

/*
 * Have the current thread exit.  It will remove its jthread
 * structure from various internal lists.  The underlying
 * thread object will exit (so it can never be scheduled
 * again.)
 *
 * Does not return.
 */
void
jthread_exit(void)
{
	jthread_t currentJThread = jthread_currentJthread();

DBG(JTHREAD,
	dprintf("jthread_exit() jthd=%p\n", currentJThread);
    )

	jthread_exit_internal(currentJThread, true);

	panic("an_thread_exit() returned.");
}

/*
 * Have the current thread do real internal cleanup.  (Called
 * by jthread_exit and the jthread interrupt handlers).
 * Calls underlying an_thread_exit();
 *
 * Does not return.
 */
void
jthread_exit_internal(jthread_t tid, bool doFinalizer)
{
	jthread_t* ntid;
	int found = 0;

	assert(tid == jthread_currentJthread());
	assert(tid->status != JTHREAD_STATUS_INNODEOS);
	assert(tid->status != JTHREAD_STATUS_NEWBORN);
	assert(tid->stackptr != NULL);
	assert(tid->threadData.visits == 0);

	if (jthread_destructor1 && doFinalizer) {
		(*jthread_destructor1)(tid->jlThread);
	}

	an_mutex_lock(&jthread_creationLock);

	jthread_talive--;
	if (tid->daemon) {
		jthread_tdaemon--;
	}

	/* Remove thread from live list so it won't be walked by GC. */
	/* XXX this is bogus, its O(n)!  Use a doubly linked list. */
	for (ntid = &jthread_liveThreads;
	     *ntid != 0; // XXX && !found
	     ntid = &(*ntid)->nextlive) 
	{
		if (tid == (*ntid)) {
			found = 1;
			(*ntid) = tid->nextlive;
			break;
		}
	}

	assert(found);
	
	jthread_spinon(tid);
	tid->status = JTHREAD_STATUS_INNODEOS;
	tid->flags &= ~(JTHREAD_FLAG_INTR_SLAUGHTER|JTHREAD_FLAG_INTR_STOP);
	jthread_spinoff(tid);

	an_mutex_unlock(&jthread_creationLock);  

	/* Wake up the team terminator */
	if( tid->threadData.flags & TDF_ZOMBIE )
	{
		signalTeam(tid->threadData.ownerTeam, SIGNAL_THREAD_DEATH);
	}
	
	if ((jthread_talive == jthread_tdaemon)
	    && (!tid->daemon)) {
		/*
		 * Normal JVMs exit as this point.  We are not a
		 * normal JVM and must get an explicit System.exit() call
		 * to shutdown.
		 */
	}

	an_thread_exit();

	panic("an_thread_exit() returned.");
}

/* 
 * Destroy the given jthread structure.  The given thread
 * must be dead and buried.
 *
 * Invoked by the thread finalizer when a thread has become
 * unreachable by the GC.
 */
void    
jthread_destroy(jthread_t tid)
{
	an_error_t err;
	
	assert(tid);
	assert(tid != jthread_currentJthread());
	assert((tid->status == JTHREAD_STATUS_INNODEOS));
	assert(tid->stackptr != NULL);
	/* Team stuff */
	assert(tid->threadData.visits == 0);
	
DBG(JTHREAD, 
	dprintf("jthread_destroy(jth=%p)\n", tid);
    )

	/*
	 * XXX the root thread wasn't allocated by the GC.
	 * XXX the root thread's stack wasn't allocated by the GC.
	 * XXX the root thread currently isn't collected at all.
	 */
	if ((tid->flags & JTHREAD_FLAG_ROOTTHREAD) == 0) {
		/* Clean up thread's team state */
		killThreadData(&(tid->threadData), tid);
		/* Kill the native thread */
		err = an_thread_destroy(&(tid->native_thread));
		assert(err == 0);
		gc_free(tid->stackptr);
		DBGIF(tid->stackptr = NULL);
		gc_free(tid);
	}

#if defined(THREAD_ANALYZER)
	remTeamLogAnalyzer(&tid->lam);
#endif
	/* XXX anything else??? */
}

