/*
 * Copyright (c) 1998, 1999, 2000, 2001, 2002 The University of Utah. All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * Contributed by the Flux Research Group at the University of Utah.
 * Authors: Godmar Back, Leigh Stoller, Patrick Tullmann
 */
/*
 * Routines for the initialization of the jthread layer and for
 * functions that are for the whole layer (not a specific thread).
 */

#include "anjthread_types.h"
#include "anjthread_types_priv.h"
#include "anjthread_globals.h"
#include "anjthread_intr.h"
#include "anjthread_current.h"
#include "anjthread_init.h"
#include "gc.h"
#include "janosvm/jteam.h"
#include "kaffevm/lockPool.h"
#include "kaffevm/thread.h"
#include "kaffevm/exception.h"
#include "debug.h"
#include <janosvm/jteamReservation.h>

void
anjthread_preinitroot(jthread_t rootthread)
{
	threadData metalData = METAL_THREAD_DATA;
	rootthread->threadData = metalData;
}


/*
 * Initialize the threading system.
 *
 * Assumes that the NodeOS threading layer is already initialized.
 */
jthread_t 
jthread_init(int pre,
	int maxpr, int minpr,
	void *(*_allocator)(size_t), 
	void (*_deallocator)(void*),
	void (*_destructor1)(void*),
	void (*_onstop)(void),
	void (*_ondeadlock)(void))		/* ignored for now */
{
	jthread_t	jtid;
	
/*	XXX these are unused: 
	jthread_max_priority = maxpr;
	jthread_min_priority = minpr;
*/
	jthread_allocator = _allocator; // XXX deprecated
	jthread_deallocator = _deallocator; // XXX deprecated

	jthread_onstop = _onstop;
	jthread_destructor1 = _destructor1;

	an_mutex_create(&jthread_creationLock);

	/*
	 * We already allocated the complete jthread object
	 * as part of creating the thread pool.
	 */

        jtid = jthread_currentJthread();
	DBGIF(an_thread_setdata(jtid));

	initThreadData(&jtid->threadData, jtid);

        jtid->nextlive = jthread_liveThreads;
        jthread_liveThreads  = jtid;
	jtid->status = JTHREAD_STATUS_RUNNING;
	jtid->flags |= JTHREAD_FLAG_ROOTTHREAD;
	{
		size_t ignoreSize;
		void *ignorePtr;
		an_thread_stackinfo(&(jtid->native_thread),
				    &(jtid->stackptr),
				    &ignoreSize, &ignorePtr);
	}
	
        jthread_talive++;
DBG(JTHREAD,
	dprintf("main thread is jth=%p (nth would be %p)\n",
		jtid, &jtid->native_thread);
    )

	return (jtid);
}

/*
 * Create a NodeOS jthread context to go with the initial thread in the system.
 * (The initial thread is created in jthread_init, above.
 */
jthread_t
jthread_createfirst(size_t mainThreadStackSize, unsigned char prio, void* jlThread)
{
        jthread_t jtid;
	int lpc;
	
	jtid = jthread_currentJthread();
	assert(jtid != NULL);
	assert(jtid->status == JTHREAD_STATUS_RUNNING);

	jtid->jlThread = jlThread;
	/*
	 * This thread will return to the pool when it exits so we have to
	 * make sure its java object will still be around if it ever returns.
	 */
	gc_add_ref(jlThread);
	DBGIF(assert(an_thread_getdata() == jtid));

	/* XXX what to do with mainThreadStackSize?? */

	/* XXX NodeOS threads do not support java priorities. */

	an_thread_sethandler(_jthread_intr_handler);

	/*
	 * Make java.lang.Thread peers for the other preallocated threads.
	 * XXX Is this a good place to do this?
	 */
	for( lpc = 0; lpc < ROOT_DOMAIN_THREAD_CT; lpc++ )
	{
		struct Hjava_lang_Thread *tid;

		jtid = an_jthreadlist[lpc];
		/* Ignore the main thread, we just did it. */
		if( jtid == jthread_currentJthread() )
			continue;

		initThreadData(&jtid->threadData, jtid);
		
		jtid->status = JTHREAD_STATUS_INNODEOS;
		jtid->flags = 0;
		
		tid = (Hjava_lang_Thread*)newObject(ThreadClass);
		assert(tid != 0);
		gc_add_ref(tid);
		jtid->jlThread = tid;
		
		unhand(tid)->name = 0;
		unhand(tid)->holder =
			(struct Hjava_lang_ThreadObjectHolder *)
			execute_java_constructor("java.lang.ThreadObjectHolder",
						 0,
						 0,
						 "()V");
		unhand(tid)->sleeper =
			execute_java_constructor("java.lang.Thread$Sleeper",
						 0,
						 0,
						 "()V");
		unhand(tid)->suspendResume =
			execute_java_constructor("java.lang.Thread$Suspender",
						 0,
						 0,
						 "()V");
		unhand(tid)->priority = java_lang_Thread_NORM_PRIORITY;
		unhand(tid)->threadQ = 0;
		unhand(tid)->daemon = 0;
		unhand(tid)->interrupting = 0;
		unhand(tid)->target = 0;
		unhand(tid)->group = standardGroup;
		
		unhand(tid)->PrivateInfo =
			(struct Hkaffe_util_Ptr*)jtid;
		unhand(tid)->stackOverflowError = 
			(Hjava_lang_Throwable*)StackOverflowError;
		unhand(tid)->needOnStack = STACK_HIGH;
		
		/* Attach thread to threadGroup */
		if( unhand(tid)->group )
		{
			do_execute_java_method(unhand(tid)->group,
					       "add",
					       "(Ljava/lang/Thread;)V",
					       0, 0,
					       tid);
		}
	}
	return (jtid);
}

void
jthread_migrate_root(jteam *jt)
{
	threadData *td;
	jthread_t tid;
	int lpc;
	
	for( lpc = 0; lpc < ROOT_DOMAIN_THREAD_CT; lpc++ )
	{
		tid = an_jthreadlist[lpc];
		td = &tid->threadData;
		td->boundary->currentReservation->entry.id = jt;
		td->currentTeam = jt;
		td->ownerTeam = jt;
	}
}

/*
 * set a function to be run when all non-daemon threads have exited
 */
void
jthread_atexit(void (*f)(void))
{
	/*
	 * XXX this never happens on a NodeOS.
	 */
}

