/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah.  All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file.
 *
 * Contributed by the Flux Research Group at the University of Utah.
 * Authors: Godmar Back, Leigh Stoller, Patrick Tullmann
 */
/*
 * jthread interruption-related routines.  Signals, stop, etc.
 */

#include "anjthread_types.h"
#include "anjthread_types_priv.h"
#include "anjthread_globals.h"
#include "anjthread_end.h"
#include "anjthread_intr.h"
#include "debug.h"

static jthread_exchandler_t nullHandler = NULL;
static jthread_exchandler_t floatingHandler = NULL;

#if defined(INTERPRETER)
#define	DEFINEFRAME(f)		/* Does nothing */
#define	EXCEPTIONFRAME(f, c)	/* Does nothing */
#define	EXCEPTIONFRAMEPTR(f)	NULL
#elif defined(TRANSLATOR)
#define	DEFINEFRAME(f)		exceptionFrame f
#define	EXCEPTIONFRAMEPTR(f)	(&(f))
#endif /* TRANSLATOR */


void
jthread_initexceptions(jthread_exchandler_t _nullHandler,
		       jthread_exchandler_t _floatingHandler)
{
	nullHandler = _nullHandler;
	floatingHandler = _floatingHandler;
}

void
jthread_handleInterrupt(jthread_t jtid)
{
	/* XXX jthread_spinon(jtid) must have been invoked by caller. */

	if (jtid->flags & JTHREAD_FLAG_INTR_SLAUGHTER)
	{
		jtid->flags &= ~JTHREAD_FLAG_INTR_SLAUGHTER;
		jthread_spinoff(jtid);
		
		/* XXX race condition here? */
		
		/* Exit immediately, without running finalizer. */
		jthread_exit_internal(jtid, false);
		
		panic("jthread_exit_internal() returned!");
	}
	
	if (jtid->flags & JTHREAD_FLAG_INTR_STOP)
	{
		/* Just clear the flag and perform the stop. */
		jtid->flags &= ~JTHREAD_FLAG_INTR_STOP;
		jthread_spinoff(jtid);
		
		/* XXX race condition here? */
		jthread_die();
		
		panic("jthread_die() returned!");
	}
}

void
_jthread_intr_handler(int type, void *arg)
{
	struct sigcontext* ctx = arg; /* XXX THIS IS MAGICAL. */

	DEFINEFRAME(frame);

	assert(nullHandler);
	assert(floatingHandler);
	assert(an_thread_current());

	if (type == AN_THREAD_TRAP_SEGV)
	{
		an_thread_unblockintr();
		EXCEPTIONFRAME(frame, ctx);
		nullHandler(EXCEPTIONFRAMEPTR(frame));
	}
	else if (type == AN_THREAD_TRAP_FPU)
	{
		an_thread_unblockintr();
		EXCEPTIONFRAME(frame, ctx);
		floatingHandler(EXCEPTIONFRAMEPTR(frame));
	}
	else if (type == AN_THREAD_INTR)
	{
		jthread_t jtid = jthread_currentJthread();

		jthread_spinon(jtid);
		jthread_handleInterrupt(jtid);
		jthread_spinoff(jtid);

	}
	else
		panic("Unknown thread interrupt type %d\n", type);
}

void
jthread_interrupt(jthread_t jtid)
{
DBG(JTHREAD,
	dprintf("jthread_interrupt(jtid=%p)\n", jtid);
    )

	if (&jtid->native_thread != an_thread_current())
	{
		int doInterrupt = 0;

		jthread_spinon(jtid);
		doInterrupt = (jtid->status == JTHREAD_STATUS_RUNNING);
		jthread_spinoff(jtid);

		/* XXX interrupt on newborn thread is simply ignored. */

		if (doInterrupt)
			/* XXX implicitly causes lock/cvar wait to interrupt? */
			an_thread_intr(&jtid->native_thread);
	}
}


void
jthread_stop(jthread_t jtid)
{
DBG(JTHREAD,
	dprintf("jthread_stop(jtid=%p)\n", jtid);
    )

	if (&jtid->native_thread != an_thread_current())
	{
		int doInterrupt = 0;

		jthread_spinon(jtid);
		jtid->flags |= JTHREAD_FLAG_INTR_STOP;
		doInterrupt = (jtid->status == JTHREAD_STATUS_RUNNING);
		jthread_spinoff(jtid);
		if (doInterrupt)
			an_thread_intr(&jtid->native_thread);
	}
	else 
	{
		/* Stop self */
		jthread_die();
	}
}

/*
 * Slaughter a thread.
 *
 * Return 0 if thread was slaughtered.  Return 1 if postponed.
 * Does not return if slaughtering the current thread.
 */
int
jthread_slaughter(jthread_t jtid)
{
DBG(JTHREAD,
	dprintf("jthread_slaughter(jtid=%p)\n", jtid);
    )
	
	if (&jtid->native_thread != an_thread_current())
	{
		int doInterrupt = 0;

		jthread_spinon(jtid);
		jtid->flags |= JTHREAD_FLAG_INTR_SLAUGHTER;
		jtid->threadData.flags |= TDF_ZOMBIE;
		doInterrupt = (jtid->status == JTHREAD_STATUS_RUNNING);
		jthread_spinoff(jtid);

		if (doInterrupt)
			an_thread_intr(&jtid->native_thread);

		return 1;
	}
	else 
	{
		/*
		 * For the current thread, just do a jthread_exit().
		 * (without running the finalizer).
		 *
		 * Note that this function does not return.
		 */
		jthread_exit_internal(jtid, false);

		return 0;
	}
}

/*
 * The NodeOS provides a simple facility for suspending all (other)
 * threads in a flow.  Which is what this stuff is used for.
 *
 * XXX This might not be right.
 */
void 
jthread_suspendall(void)
{
	an_thread_setprio(an_thread_current(), AN_THREAD_PRIO_EXCLUSIVE);
}

void
jthread_unsuspendall(void)
{
	an_thread_setprio(an_thread_current(), AN_THREAD_PRIO_NORMAL);
}


void
blockAsyncSignals(void)
{
	/* XXX shouldn't really be needed on Moab. */
}

void
unblockAsyncSignals(void)
{
	/* XXX shouldn't really be needed on Moab. */
}

#ifdef SMP
/*
 * On the SMP OSKit, spinlocks will be spinlocks and Leigh will invent
 * a way to stop all processors
 */
#error SMP support is not available for Moab threads in Kaffe
#endif

