/*
 * Copyright (c) 1999, 2000, 2001 The University of Utah.  All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file.
 *
 * Contributed by the Flux Research Group at the University of Utah.
 * Authors: Godmar Back, Leigh Stoller, Patrick Tullmann
 */

/*
 * jthread status functions.  Liveness, stack info, etc.
 */

#include "config.h"
#include "gtypes.h"
#include "jtypes.h"
#include "anjthread_types.h"
#include "anjthread_status.h"
#include "anjthread_current.h"

/* 
 * Check whether a thread is alive.
 *
 * Note that threads executing their cleanup function are not (jthread-) alive
 * (they're set to JTHREAD_STATUS_DEAD).
 */
int
jthread_alive(jthread_t tid)
{
	return tid && ((tid->status == JTHREAD_STATUS_NEWBORN) || 
		       (tid->status == JTHREAD_STATUS_RUNNING));
}

bool
jthread_killed(void)
{
	jthread_t tid = jthread_currentJthread();
	
	assert(tid);
	
	return (tid->status == JTHREAD_STATUS_DEAD);
}

/*
 * Determine the interesting stack range for a conservative gc.
 *
 * Returns 0 if there was a problem, 1 if there was not.
 */
int
jthread_extract_stack(jthread_t jtid,
		      void **used_from, unsigned *used_len,
		      void **free_from, unsigned *free_len)
{
	void *stackbase;
	int stacksize;
	void *stackptr;
	an_error_t err;

	assert(jtid);

	err = an_thread_stackinfo(&(jtid->native_thread),
				  &stackbase,
				  &stacksize,
				  &stackptr);
	if (err != 0)
	{
		if (used_from)
			*used_from = (void*)-1;
		if (free_from)
			*free_from =  (void*)-1;
		if (used_len)
			*used_len = 0;
		if (free_len)
			*free_len = 0;
		return 0;
	}

#if defined(STACK_GROWS_UP)
#error Cannot handle stacks that grow up
#else
	assert(stacksize > 0);
	assert(stackbase < stackptr);
	assert(stackbase == jtid->stackptr);

	if (used_from)
		*used_from = stackptr;
	if (used_len)
		*used_len = stackbase + stacksize - stackptr;
	if (free_from)
		*free_from = stackbase;
	if (free_len)
		*free_len = stackptr - stackbase;

#ifdef DEBUG
	if ((stackptr < 0) ||
	    ((stackbase+stacksize-stackptr) > (256*1024))) {
	    panic("(%p) an_thread_stackinfo(%p) reported obscene numbers: "
		  "base = %p, sp = %p\n", 
		  jtid,
		  jtid->native_thread,
		  stackbase, stackptr);
	    exit(-1);
	}
#endif /* DEBUG */
#endif /* STACK_GROWS_UP */

DBG(JTHREAD,
	dprintf("extract_stack(jth=%p): base=%p size=%d sp=%p;",
		jtid,
		stackbase, stacksize, stackptr);
        if (used_from)
		dprintf(" used_from=%p;", *used_from);
        if (used_len)
		dprintf(" used_len=%d;", *used_len);
        if (free_from)
		dprintf(" free_from=%p;", *free_from);
        if (free_len)
		dprintf(" free_len=%d;", *free_len);
    )

	return 1;
}



/*
 * See if there is enough room on the stack.
 *
 * XXX change signature to return amount available.
 */
int
jthread_stackcheck(int need)
{
	int room;
	void *stackbase;
	int stacksize;
	void *stackptr;
	an_thread_t *ntid = an_thread_current();
	an_error_t err;

	err = an_thread_stackinfo(ntid,
				  &stackbase,
				  &stacksize,
				  &stackptr);
	assert(err == 0); /* XXX */
	assert(stackbase == jthread_currentJthread()->stackptr);

#if defined(STACK_GROWS_UP)
#	error FIXME
#else
	room = (stackptr - stackbase);
#endif
	
DBG(JTHREAD,
	dprintf("stackcheck(jth=%p, need=%d): base=%p size=%d sp=%p room=%d\n",
		jthread_currentJthread(),
		need, stackbase, stacksize, stackptr, room);
    )
	return (room >= need);
}

/* 
 * XXX this is supposed to count the number of stack frames 
 */
int
jthread_frames(jthread_t thrd)
{
	panic("jthread_frames() called.\n");
        return 0;
}

