/*
 * Copyright (c) 1996, 1997:
 *      Transvirtual Technologies, Inc.  All rights reserved.
 *
 * Copyright (c) 1998, 1999, 2000:
 *      The University of Utah. All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file.
 *
 * Contributed by the Flux Research Group at the University of Utah.
 * Author: Patrick Tullmann, Godmar Back
 */
/*
 * Syscall definitions for NodeOS threads port.
 */

#include "config.h"
#include "debug.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-net.h"
#include "config-io.h"
#include "config-signal.h"
#include "jsyscall.h"
#include "gtypes.h"
#include "jtypes.h"
#include "support.h"
#include "anjthread_lock.h"
#include "anjthread_current.h"
#include "anjthread_sleep.h"

/* return error code or 0 */   
#define ERR(EXP) (((EXP) < 0) ? errno : 0)

/*
 * XXX
 */
#define MAX_KAFFE_FDS 1024
static int next_free_fd;
static int fd_list[MAX_KAFFE_FDS];
static an_file_t fd_to_an_file[MAX_KAFFE_FDS];

static void nodeos_thread_fd_table_init(void)
{
        if( next_free_fd == 0 )
	{
		int lpc;
		
		next_free_fd = 3;

		fd_list[0] = -1; // stdin is allocated
		fd_list[1] = -1; // stdout is allocated
		fd_list[2] = -1; // stderr is allocated
		for( lpc = 3; lpc < MAX_KAFFE_FDS; lpc++ )
		{
			fd_list[lpc] = lpc + 1;
		}
		fd_list[MAX_KAFFE_FDS - 1] = -1;
	}
}

static int nodeos_thread_alloc_fd(void)
{
	jthread_t th = jthread_current();
	int retval;
	
	jthread_spinon(th);
	nodeos_thread_fd_table_init();
	retval = next_free_fd;
	if( retval != -1 )
	{
		next_free_fd = fd_list[retval];
		fd_list[retval] = -1;
	}
	jthread_spinoff(th);
	return( retval );
}

static void nodeos_thread_free_fd(int fd)
{
	jthread_t th = jthread_current();
	
	jthread_spinon(th);
	fd_list[fd] = next_free_fd;
	next_free_fd = fd;
	jthread_spinoff(th);
}

static inline int check_fd(int fd)
{
	int retval = EINVAL;

	if( fd_list[fd] == -1 )
		retval = 0;
	return( retval );
}

static int
nodeos_thread_open(const char *path, int f, int m, int *outfd)
{
	an_error_t error;
	
	if( (*outfd = nodeos_thread_alloc_fd()) != -1 )
	{
		an_acd_t acd;
		an_file_t *retval;
		
		acd.st_mode = m;
		retval = &fd_to_an_file[*outfd];
		if( (error = an_file_open(path,
					  f,
					  &acd,
					  &retval)) != 0 )
		{
			nodeos_thread_free_fd(*outfd);
			*outfd = -1;
		}
	}
	else
	{
		error = ENFILE;
	}
	return (*outfd == -1) ? error : 0;
}

static int
nodeos_thread_read(int fd, void* b, size_t l, ssize_t *out)
{
	an_error_t error;
	
	*out = -1;
	if( (error = check_fd(fd)) == 0 )
	{
		error = an_file_read(&fd_to_an_file[fd], l, b, out);
	}
	return (*out == -1) ? error : 0;
}

static int
nodeos_thread_read_and_wait(int fd, void* b, size_t l, int sample_time,
			    ssize_t *out)
{
	an_error_t error;
	
	*out = -1;
	if( (error = check_fd(fd)) == 0 )
	{
		*out = 0;
		while( (*out) == 0 )
		{
			error = an_file_read(&fd_to_an_file[fd], l, b, out);
			if( ((*out) == 0) && !error )
			{
				jthread_sleep(sample_time);
			}
		}
	}
	return (*out == -1) ? error : 0;
}

static int
nodeos_thread_write(int fd, const void* b, size_t l, ssize_t *out)
{
	an_error_t error;
	
	*out = -1;
	if( (error = check_fd(fd)) == 0 )
	{
		error = an_file_write(&fd_to_an_file[fd], l, (void *)b, out);
	}
	return (*out == -1) ? error : 0;
}

static int
nodeos_thread_lseek(int fd, off_t o, int w, off_t *out)
{
	an_error_t error;
	
	*out = -1;
	if( (error = check_fd(fd)) == 0 )
	{
		error = an_file_lseek(&fd_to_an_file[fd], o, w, out);
	}
	return (*out == -1) ? error : 0;
}

static int
nodeos_thread_close(int f)
{
	an_error_t error;
	int retval = -1;
	
	if( (error = check_fd(f)) == 0 )
	{
		if( (error = an_file_close(&fd_to_an_file[f])) == 0 )
		{
			nodeos_thread_free_fd(f);
			retval = 0;
		}
	}
	return (retval == -1) ? error : 0;
}

#define VALUE_NOT_AVAILABLE 0

static void
nodeos_thread_translate_stat(struct stat *st, an_stat_t *an_st)
{
	st->st_dev = VALUE_NOT_AVAILABLE;
	st->st_ino = VALUE_NOT_AVAILABLE;
	st->st_mode = an_st->st_acd.st_mode;
	switch( an_st->st_type )
	{
	case ANFILE_UNKNOWN:
		/* XXX Leave it as 0? */
		break;
	case ANFILE_FILE:
		st->st_mode |= S_IFREG;
		break;
	case ANFILE_DIR:
		st->st_mode |= S_IFDIR;
		break;
	}
	st->st_nlink = VALUE_NOT_AVAILABLE;
	st->st_uid = VALUE_NOT_AVAILABLE;
	st->st_gid = VALUE_NOT_AVAILABLE;
	st->st_rdev = VALUE_NOT_AVAILABLE;
	st->st_atimespec.tv_sec = an_st->st_atimespec.tv_sec;
	st->st_atimespec.tv_nsec = an_st->st_atimespec.tv_nsec;
	st->st_mtimespec.tv_sec = an_st->st_mtimespec.tv_sec;
	st->st_mtimespec.tv_nsec = an_st->st_mtimespec.tv_nsec;
	st->st_ctimespec.tv_sec = an_st->st_ctimespec.tv_sec;
	st->st_ctimespec.tv_nsec = an_st->st_ctimespec.tv_nsec;
	st->st_size = an_st->st_size;
	st->st_blocks = VALUE_NOT_AVAILABLE;
	st->st_blksize = VALUE_NOT_AVAILABLE;
}

static int
nodeos_thread_fstat(int f, struct stat *st)
{
	an_error_t error;
	int retval = -1;
	
	if( (error = check_fd(f)) == 0 )
	{
		an_stat_t an_st;
		
		if( (error = an_file_fstat(&fd_to_an_file[f], &an_st)) == 0 )
		{
			nodeos_thread_translate_stat(st, &an_st);
			retval = 0;
		}
	}
	return (retval == -1) ? error : 0;
}

static int
nodeos_thread_stat(const char *p, struct stat *st)
{
	an_error_t error;
	int retval = -1;
	an_stat_t an_st;
	
	if( (error = an_file_stat(p, &an_st)) == 0 )
	{
		nodeos_thread_translate_stat(st, &an_st);
		retval = 0;
	}
	return (retval == -1) ? error : 0;
}

static int
nodeos_thread_mkdir(const char *p, int m)
{
	an_error_t error;
	int retval = -1;
	an_acd_t acd;

	acd.st_mode = m;
	if( (error = an_file_mkdir(p, &acd)) == 0 )
	{
		retval = 0;
	}
	return (retval == -1) ? error : 0;
}

static int
nodeos_thread_rmdir(const char *p)
{
	an_error_t error;
	int retval = -1;

	if( (error = an_file_unlink(p)) == 0 )
	{
		retval = 0;
	}
	return (retval == -1) ? error : 0;
}

static int
nodeos_thread_rename(const char *p, const char *q)
{
	an_error_t error;
	int retval = -1;

	if( (error = an_file_rename(p, q)) )
	{
		retval = 0;
	}
	return (retval == -1) ? error : 0;
}

static int
nodeos_thread_remove(const char *p)
{
	an_error_t error;
	int retval = -1;

	if( (error = an_file_unlink(p)) == 0 )
	{
		retval = 0;
	}
	return (retval == -1) ? error : 0;
}

static int
nodeos_thread_socket(int a, int b, int c, int *outfd)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_connect(int fd, struct sockaddr* addr, int len, int timeout)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_bind(int fd, struct sockaddr* addr, int len)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_listen(int fd, int n)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_accept(int fd, struct sockaddr* a, int *l, int timeout, 
	int* outfd)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_sock_read(int f, void* b, size_t l, int timeout, int *out)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_recvfrom(int a, void* b, size_t c, int d, struct sockaddr* e, 
	int* f, int timeout, int *out)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_sendto(int a, const void* b, size_t c, int d, 
	const struct sockaddr* e, int f, ssize_t *out)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_setsockopt(int a, int b, int c, const void* d, int e)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_getsockopt(int a, int b, int c, void* d, int* e)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_getsockname(int a, struct sockaddr* b, int* c)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_getpeername(int a, struct sockaddr* b, int* c)
{
	unimp(__FUNCTION__ ": Should not be using BSD network interfaces on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_gethostbyname(const char*host, struct hostent**out)
{
	int rc;
	
	rc = 0;
	jthread_spinon(0);
	/* NB: this will block the whole process while we're looking up
	 * a name.  However, gethostbyname is extremely async-signal-unsafe,
	 * and hence we have no other choice.  Now I understand why old 
	 * Netscapes blocked when looking up a name.  
	 *
	 * In a UNIXy system, a possible solution is to spawn a process
	 * for lookups, which I think is what NS eventually did.
	 */
	*out = gethostbyname(host);
	if (*out == 0) {
		rc = h_errno;
	}
	jthread_spinoff(0);
	return (rc);
}

static int
nodeos_thread_gethostbyaddr(const char*host, int l, int t, struct hostent**out)
{
	int rc = 0;

	jthread_spinon(0);
	/* NB: same comment as for jthreadedGetHostByName applies here */
	*out = gethostbyaddr(host, l, t);
	if (*out == 0) {
		rc = h_errno;
	}
	jthread_spinoff(0);
	return (rc);
}

static int
nodeos_thread_select(int a, fd_set* b, fd_set* c, fd_set* d, struct timeval* e,
	int* out)
{
	panic(": Uh oh.  Don't have select on NodeOS.");
}

static int
nodeos_thread_forkexec(char *argv[], char *env[], int fd[4], int *outpid, const char* unknown)
{
	unimp(__FUNCTION__ ": forkexec() not implemented on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_waitpid(int a, int* b, int c, int* out)
{
	unimp(__FUNCTION__ ": waitpid() not implemented on NodeOS");
	return (EOPNOTSUPP);
}

static int
nodeos_thread_kill(int a, int b)
{
	unimp(__FUNCTION__ ": kill() not implemented on NodeOS");
	return (EOPNOTSUPP);
}

SystemCallInterface Kaffe_SystemCallInterface = {
	nodeos_thread_open,
	nodeos_thread_read,
	nodeos_thread_read_and_wait,
	nodeos_thread_write, 
	nodeos_thread_lseek,
	nodeos_thread_close,
	nodeos_thread_fstat,
	nodeos_thread_stat,
	nodeos_thread_mkdir,
	nodeos_thread_rmdir,
	nodeos_thread_rename,
	nodeos_thread_remove,
	nodeos_thread_socket,
	nodeos_thread_connect,
	nodeos_thread_bind,
	nodeos_thread_listen,
	nodeos_thread_accept, 
	nodeos_thread_sock_read,
	nodeos_thread_recvfrom,
	nodeos_thread_write, 
	nodeos_thread_sendto,	
	nodeos_thread_setsockopt,
	nodeos_thread_getsockopt,
	nodeos_thread_getsockname, 
	nodeos_thread_getpeername,
	nodeos_thread_close,
	nodeos_thread_gethostbyname,
	nodeos_thread_gethostbyaddr,
	nodeos_thread_select,	
	nodeos_thread_forkexec,
	nodeos_thread_waitpid,
	nodeos_thread_kill,
};
