/*
 * thread.c
 * Thread support.
 *
 * Copyright (c) 1996, 1997
 *	Transvirtual Technologies, Inc.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution 
 * of this file. 
 */

#include "config.h"
#include "debug.h"
#include "config-std.h"
#include "config-mem.h"
#include "config-io.h"
#include "config-signal.h"
#include "jtypes.h"
#include "access.h"
#include "object.h"
#include "constants.h"
#include "errors.h"
#include "classMethod.h"
#include "baseClasses.h"
#include "stringSupport.h"
#include "lookup.h"
#include "thread.h"
#include "jthread.h"
#include "locks.h"
#include "ksem.h"
#include "exception.h"
#include "support.h"
#include "external.h"
#include "gc.h"
#include "jni.h"
#include "md.h"
#include <janosvm/jteam.h>
#include <janosvm/jteamReservation.h>
#include <janosvm/stdioResource.h>
#include "module.h"
#include "moduleList.h"

/* If not otherwise specified, assume at least 1MB for main thread */
#ifndef MAINSTACKSIZE
#define MAINSTACKSIZE (1024*1024)
#endif

/* 
 * store the native thread for the main thread here 
 * between init and createFirst 
 */
static jthread_t	mainthread;
int threadStackSize;	/* native stack size */

/* referenced by native/Runtime.c */
jbool runFinalizerOnExit;	/* should we run finalizers? */

jbool deadlockDetection = 1;	/* abort if we detect deadlock */

Hjava_lang_Class* ThreadClass;
Hjava_lang_Class* ThreadGroupClass;
Hjava_lang_Class* ThreadDeathClass;
Hjava_lang_ThreadGroup* standardGroup;

void threadEnterVM(void*) __NORETURN__;
void threadEnterVM_arglist(struct Hjava_lang_Thread *tid,
			   struct Hjava_lang_Object* obj,
			   jmethodID runmethod, ...) __NORETURN__;
void threadEnterVM_valist(struct Hjava_lang_Thread* tid,
			  struct Hjava_lang_Object* obj,
			  jmethodID runmethod, va_list argptr) __NORETURN__;
void threadSpecialEnterVM(void* arg);
static void createInitialThread(const char*);
static void runfinalizer(void);
static iLock* thread_start_lock;

static int initThreadModule(int stage);

module_t threadModule = {
	"thread",
	0,
	initThreadModule,
	0,
	0
};

static int initThreadModule(int stage)
{
	switch( stage )
	{
	case SID_THREADMODULE_JTHREAD:
		{
			/* Set default thread stack size if not set */
			extern JavaVMInitArgs Kaffe_JavaVMArgs[];
			int stackSize;
			
			stackSize = Kaffe_JavaVMArgs[0].nativeStackSize;
			if( stackSize == 0 )
				stackSize = THREADSTACKSIZE;
			initNativeThreads(stackSize);
		}
		break;
	case SID_THREADMODULE:
		initThreads();
		break;
	}
	return( 1 );
}

/*
 * How do I get memory?
 */
static
void *
thread_malloc(size_t s)
{
	// return gc_malloc(s, GC_ALLOC_THREADCTX);
	DBGIF(dprintf("thread.c:thread_malloc() invoked; deprecated;\n"));
	return NULL;
}

static
void
thread_free(void *p)
{
	//gc_free(p);
	DBGIF(dprintf("thread.c:thread_free() invoked; deprecated;\n"));
}

/*
 * Initialise threads.
 */
void
initThreads(void)
{
	errorInfo info;

	/* Get a handle on the thread and thread group classes */
	ThreadClass = lookupClass(THREADCLASS, NULL, &info);
	assert(ThreadClass != 0);
	ThreadGroupClass = lookupClass(THREADGROUPCLASS, NULL, &info);
	assert(ThreadGroupClass != 0);
	ThreadDeathClass = lookupClass(THREADDEATHCLASS, NULL, &info);
	assert(ThreadDeathClass != 0);
	
	/* Create base group */
	standardGroup = (Hjava_lang_ThreadGroup*)newObject(ThreadGroupClass);
	
	assert(standardGroup != 0);
	unhand(standardGroup)->parent = 0;
	unhand(standardGroup)->name = stringC2Java("main");
	assert(unhand(standardGroup)->name != NULL);
	unhand(standardGroup)->maxPriority = java_lang_Thread_MAX_PRIORITY;
	unhand(standardGroup)->destroyed = 0;
	unhand(standardGroup)->daemon = 0;
	unhand(standardGroup)->nthreads = 0;
	unhand(standardGroup)->threads = (HArrayOfObject*)
		newArray(ThreadClass, 0);
	unhand(standardGroup)->ngroups = 0;
	unhand(standardGroup)->groups = (HArrayOfObject*)
		newArray(ThreadGroupClass, 0);

	/* Allocate a thread to be the main thread */
	createInitialThread("main");
}

/*
 * This guy can fail because of allocation failure, or because of a
 * write barrier.
 */
static int
createThread(Hjava_lang_Thread* tid, void* func, size_t stacksize,
	     struct _errorInfo *einfo)
{
	struct Hkaffe_util_Ptr* nativethread;

	nativethread = (struct Hkaffe_util_Ptr*)jthread_create(
		unhand(tid)->priority,
		func,
		unhand(tid)->daemon,
		tid,
		stacksize);

	if (!nativethread) {
		postOutOfMemory(einfo);
		return 0;
	}

	unhand(tid)->PrivateInfo = nativethread;
	unhand(tid)->needOnStack = STACK_HIGH;
	return 1;
}

/*
 * Start a new thread running.
 */
void
startThread(Hjava_lang_Thread* tid)
{
	int success = 0;
	struct _errorInfo info;
	jteamReservation *jr;
	jteam *owner;
	int iLockRoot;

#if 0
	if (aliveThread(tid) == true) {
		throwException(IllegalThreadStateException);
	}
#endif

	/* Hold the start lock while the thread is created.
	 * This lock prevents the new thread from running until we're
	 * finished in create.
	 * See also threadEnterVM.
	 */
	owner = currentTeam();
	if( (jr = pushTeamReservation(kernelTeam)) )
	{
		lockStaticMutex(&thread_start_lock);
		success = createThread(tid, &threadEnterVM,
				       threadStackSize, &info);
		unlockStaticMutex(&thread_start_lock);
		popTeamReservation(jr);
	}
	if (!success) {
		throwError(&info);
	}
}

/*
 * Interrupt a thread
 */
void
interruptThread(Hjava_lang_Thread* tid)
{
	if ((jthread_t)unhand(tid)->PrivateInfo) {
		jthread_interrupt((jthread_t)unhand(tid)->PrivateInfo);
	}
}

/*
 * Stop a thread from running and terminate it.
 */
void
stopThread(Hjava_lang_Thread* tid, Hjava_lang_Object* obj)
{
	if (getCurrentThread() == tid) {
		throwException((Hjava_lang_Throwable*)obj);
	}
	else {
		/*
		 * Note that we deviate from the spec here in that the target 
		 * thread won't throw the exception `obj', but it will 
		 * construct a new ThreadDeath exception when it dies.
		 */
		if ((jthread_t)unhand(tid)->PrivateInfo)
			jthread_stop((jthread_t)unhand(tid)->PrivateInfo);
	}
}

/*
 * Destroy a thread hard. 
 */
void
destroyThread(Hjava_lang_Thread* tid)
{
	jteamReservation *jr;
	jthread_t jtid = 0;
	int iLockRoot;

	if( (jr = pushTeamReservation(kernelTeam)) )
	{
		lockStaticMutex(&thread_start_lock);
		jtid = (jthread_t) unhand(tid)->PrivateInfo;
		unlockStaticMutex(&thread_start_lock);
		popTeamReservation(jr);
	}
	
	/*
	 * By synchronizing with thread_start_lock, we make sure that
	 * the jthread was created and we can actually submit an
	 * effective jthread_stop()
	 */

	if( jtid == jthread_current() )
	{
		threadData *td = jthread_get_data(jthread_current());

		if( td->visits == 0 )
			jthread_exit();
		else
			crashIntoTeamBoundary(0);
	}
	else if( jtid )
	{
		jthread_stop(jtid);
	}
}

/*
 * Create the initial thread with a given name.
 *  We should only ever call this once.
 */
static
void
createInitialThread(const char* nm)
{
	Hjava_lang_Thread* tid;

	/* Allocate a thread to be the main thread */
	tid = (Hjava_lang_Thread*)newObject(ThreadClass);
	assert(tid != 0);

	unhand(tid)->name = stringC2CharArray(nm);
	assert(unhand(tid)->name);
	unhand(tid)->holder =
		(struct Hjava_lang_ThreadObjectHolder *)
		execute_java_constructor("java.lang.ThreadObjectHolder",
					 0,
					 0,
					 "()V");
	unhand(tid)->sleeper =
		execute_java_constructor("java.lang.Thread$Sleeper",
					 0,
					 0,
					 "()V");
	unhand(tid)->suspendResume =
		execute_java_constructor("java.lang.Thread$Suspender",
					 0,
					 0,
					 "()V");
	unhand(tid)->priority = java_lang_Thread_NORM_PRIORITY;
	unhand(tid)->threadQ = 0;
	unhand(tid)->daemon = 0;
	unhand(tid)->interrupting = 0;
	unhand(tid)->target = 0;
	unhand(tid)->group = standardGroup;

	jthread_atexit(runfinalizer);
	/* set Java thread associated with main thread */
	mainthread = jthread_createfirst(MAINSTACKSIZE, java_lang_Thread_NORM_PRIORITY, tid);
	unhand(tid)->PrivateInfo = (struct Hkaffe_util_Ptr*)mainthread;
	unhand(tid)->stackOverflowError = 
		(Hjava_lang_Throwable*)StackOverflowError;
	unhand(tid)->needOnStack = STACK_HIGH;

	/* Attach thread to threadGroup */
	if( unhand(tid)->group )
		do_execute_java_method(unhand(tid)->group,
				       "add",
				       "(Ljava/lang/Thread;)V",
				       0, 0,
				       tid);
}

/*
 * helper function to start daemon threads
 */
void
threadSpecialEnterVM(void* arg)
{
	Hjava_lang_Thread* tid;
	void (*func)(void *);
	jteamReservation *ja;
	void *argument;

	tid  = (Hjava_lang_Thread*)arg;
	assert(tid);
	assert(tid == getCurrentThread());

	registerTeamThread(currentTeam(),
			   jthread_current(),
			   unhand(tid)->daemon);
	ja = pushTeamReservation(currentTeam());
	
	jthread_enable_stop();
	
	/* target and group are overloaded by createDaemon */
	func = (void*)tid->sFunc;
	argument = (void*)tid->sArg;
	assert(func);
	tid->target = 0;
	tid->group = 0;

	func(argument);
	unregisterTeamThread(currentTeam(),
			     jthread_current(),
			     unhand(tid)->daemon);
}

/*
 * Start a daemon thread, such as a gc or finalizer thread.
 * We give these threads a java incarnation for consistency.
 * It might not be strictly necessary for the gc thread.
 *
 * This guy can fail because of allocation failure, or because
 * createThread failed.
 * XXX
 */
Hjava_lang_Thread*
createDaemon(void* func, const char* nm, void *arg, int prio,
	     size_t stacksize, struct _errorInfo *einfo)
{
	Hjava_lang_Thread* tid;
	jteamReservation *jr;

DBG(VMTHREAD,	dprintf("createDaemon %s\n", nm);	)

	/* Keep daemon threads as root objects */
	tid = (Hjava_lang_Thread*)newObject(ThreadClass);
	assert(tid != 0);

DBG(VMTHREAD,	dprintf("createDaemon tid %p\n", tid);	)

	if (GC_SET_SAFE(tid, unhand(tid)->name, stringC2CharArray(nm), einfo)) {
		return (0);
	}
	if (!unhand(tid)->name) {
		postOutOfMemory(einfo);
		return 0;
	}
	unhand(tid)->priority = prio;
	unhand(tid)->threadQ = 0;
	unhand(tid)->daemon = 1;
	unhand(tid)->interrupting = 0;
	/* preallocate a stack overflow error for this thread in case it 
	 * runs out 
	 */
	if (GC_SET_SAFE(tid,  unhand(tid)->stackOverflowError,
			(Hjava_lang_Throwable*)StackOverflowError,
			einfo)) {
		return 0;
	}
	unhand(tid)->holder =
		(struct Hjava_lang_ThreadObjectHolder *)
		execute_java_constructor("java.lang.ThreadObjectHolder",
					 0,
					 0,
					 "()V");
	unhand(tid)->sleeper =
		execute_java_constructor("java.lang.Thread$Sleeper",
					 0,
					 0,
					 "()V");
	unhand(tid)->suspendResume =
		execute_java_constructor("java.lang.Thread$Suspender",
					 0,
					 0,
					 "()V");
	
	unhand(tid)->sFunc = (void*)func;
	unhand(tid)->sArg = (void*)arg;

	if( (jr = pushTeamReservation(kernelTeam)) )
	{
		if( !createThread(tid,
				  threadSpecialEnterVM,
				  stacksize,
				  einfo) )
		{
			tid = 0;
		}
		popTeamReservation(jr);
	}
	else
	{
		tid = 0;
	}
	return (tid);
}

/*
 * By default, all threads enter the KaffeVM here.  The void* argument
 * is currently ignored by threadEnterVM.
 *
 * threadEnterVM just invokes the "void run()" on the current
 * thread's java.lang.Thread object.
 *
 * This function never returns.
 */
void
threadEnterVM(void* arg)
{
	extern JNIEnv Kaffe_JNIEnv;
	JNIEnv *env = &Kaffe_JNIEnv;
	Hjava_lang_Thread* tid;
	jmethodID runmethod;

	tid = (Hjava_lang_Thread*)arg;
	assert(tid == getCurrentThread());

	/* Find the run()V method */
	runmethod = (*env)->GetMethodID(env,
					(*env)->GetObjectClass(env, tid),
					"run",
					"()V");
	threadEnterVM_arglist(tid, (struct Hjava_lang_Object*)tid, runmethod);
}

void
threadEnterVM_arglist(struct Hjava_lang_Thread* tid,
		      struct Hjava_lang_Object* obj,
		      jmethodID runmethod,
		      ...)
{
	va_list argptr;
	va_start(argptr, runmethod);
	threadEnterVM_valist(tid, obj, runmethod, argptr);
	va_end(argptr);
}

void
printStackTrace(struct Hjava_lang_Throwable* o,
		struct Hjava_lang_Object* p, int nullOK);

/*
 * An alternative enterypoint for threads to begin their lives in the
 * KaffeVM.  This entrypoint makes no assumptions about the
 * initial java method to execute (and the args to pass to it).
 *
 * This function never returns.
 */
void
threadEnterVM_valist(struct Hjava_lang_Thread* tid,
		     struct Hjava_lang_Object* obj,
		     jmethodID runmethod,
		     va_list argptr)
{
	extern JNIEnv Kaffe_JNIEnv;
	JNIEnv *env = &Kaffe_JNIEnv;
#if !defined(NO_PRECONDITIONS)
	threadData *td = jthread_get_data(jthread_current());
#endif
	jthrowable eobj;
	int iLockRoot;

	require(td->visits == 0);
	require(td->currentTeam == currentTeam());
	
	registerTeamThread(currentTeam(),
			   jthread_current(),
			   unhand(tid)->daemon);
	/* 
	 * Make sure the thread who created us returned from
	 * startThread.  This ensures that privateInfo
	 * is set when we run.
	 */
	lockStaticMutex(&thread_start_lock);
	unlockStaticMutex(&thread_start_lock);
	tid  = getCurrentThread();

DBG(VMTHREAD,	
	dprintf("threadEnterVM(jltid=%p, obj=%p::%p)\n",
		tid, obj, runmethod);
    )

	jthread_enable_stop();
	
	/*
	 * We use JNI here to make sure the stack is unrolled when we get
	 * into the uncaughtException handler.  Otherwise, we wouldn't be
	 * able to handle StackOverflowError.
	 */

	if (runmethod != 0) {
		(*env)->CallVoidMethodV(env, obj, runmethod, argptr);

		/* note that since exception.c does not allow JNI to catch
		 * ThreadDeath (for now!), we won't see ThreadDeath here.
		 * That is, we must invoke the uncaughtException paper
		 * if we see an exception here.
		 */
		eobj = (*env)->ExceptionOccurred(env);
		(*env)->ExceptionClear(env);
	} else {
		/* eobj will usually be NoSuchMethodError */
		eobj = (*env)->ExceptionOccurred(env);
		(*env)->ExceptionClear(env);
	}

	/* If all else fails we call the the uncaught exception method
	 * on this thread's group.  Note we must set a flag so we 
	 * don't do this again while in the handler.
	 *
	 * Maybe this thread is so fucked up, or memory may be so tight,
	 * that we can't even do that. 
	 */
	if (eobj != 0 && unhand(tid)->dying == false && unhand(tid)->group)
	{
		jobject group = unhand(tid)->group;
		jclass groupclass = (*env)->GetObjectClass(env, group);
		jmethodID uncaughtmeth;

		if (!groupclass)
			goto done;
		
		uncaughtmeth =
			(*env)->GetMethodID(env, groupclass,
				"uncaughtException",
				"(Ljava/lang/Thread;Ljava/lang/Throwable;)V");
		if (!uncaughtmeth)
			goto done;
		
		unhand(tid)->dying = true;
		
		DBG(INVOKE,
		    /* this is too annoying for words */
		    dprintf("BEGIN DEBUG----uncaughtException, describing exception  -------------------------\n");                
		    do_execute_java_method(eobj, "printStackTrace", "()V", 0, 0, eobj);
		    dprintf("END DEBUG-----------------------------------\n");
		    );

		(*env)->CallVoidMethod(env, group, uncaughtmeth, tid, eobj);
		/* exceptions thrown in `uncaughtException' are 
		 * silently ignored says the JLS.
		 */
		(*env)->ExceptionClear(env);
	}
	else if( (eobj != 0) &&
		 (unhand(tid)->dying == false) &&
		 !unhand(tid)->group &&
		 (currentTeam()->state == JTS_ALIVE) &&
		 !(*env)->IsInstanceOf(env, eobj, ThreadDeathClass) )
	{
		Hjava_lang_Throwable *th = (Hjava_lang_Throwable *)eobj;
		char *cmsg = 0;
		
		if( th->message )
		{
			cmsg = stringJava2C(th->message);
		}
		stdioPrint(STDIO_ERR,
			   "%s:%s\n",
			   CLASS_CNAME(OBJECT_CLASS((Hjava_lang_Object *)th)),
			   cmsg ? cmsg : "");
		printStackTrace(th, 0, 1);
		gc_free(cmsg);
	}
 done:
	ensure(td->visits == 0);
	unregisterTeamThread(currentTeam(),
			     jthread_current(),
			     unhand(tid)->daemon);
	exitThread();
}

/*
 * Yield process to another thread of equal priority.
 */
void
yieldThread(void)
{
	jthread_yield();
}

/*
 * Change thread priority.
 */
void
setDaemonThread(Hjava_lang_Thread* tid, jbool on)
{
	/* no native thread yet */
	if (unhand(tid)->PrivateInfo == 0)
		return;

	reregisterTeamThread((jthread_t)unhand(tid)->PrivateInfo, on);
}

/*
 * Change thread priority.
 */
void
setPriorityThread(Hjava_lang_Thread* tid, int prio)
{
	unhand(tid)->priority = prio;

	/* no native thread yet */
	if (unhand(tid)->PrivateInfo == 0)
		return;

	jthread_setpriority((jthread_t)unhand(tid)->PrivateInfo, prio);
}

/*
 * Terminate a thread.  This function never returns.
 */
void
exitThread(void)
{
	Hjava_lang_Thread* tid;

	tid = getCurrentThread();
        do_execute_java_method(getCurrentThread(), "finish", "()V", 0, 0);

	rollbackReservationStack((void *)~0, (void *)0, (void *)~0);
	assert(tid);
	/* This never returns */
	jthread_exit();
}

#if 0
/*
 * Put a thread to sleep.
 */
void
sleepThread(jlong time)
{
	if (time > 0) {
		jthread_sleep(time);
	}
}
#endif

/*
 * Is this thread alive?
 */
bool
aliveThread(Hjava_lang_Thread* tid)
{
	bool status;

DBG(VMTHREAD,	dprintf("aliveThread: tid %p\n", tid);		)

	status = jthread_alive((jthread_t)unhand(tid)->PrivateInfo);

	return (status);
}

/*
 * How many stack frames have I invoked?
 */
jint
framesThread(Hjava_lang_Thread* tid)
{
	return (jthread_frames((jthread_t)unhand(tid)->PrivateInfo));
}

/*
 * Get the current Java thread.
 */
Hjava_lang_Thread*
getCurrentThread(void)
{
	Hjava_lang_Thread* tid;
	
	tid = jthread_getcookie(jthread_current());
#ifdef JIT3
	assert(tid);
#endif
	return tid;
}

/*
 * Finalize a thread.
 *  This is to free the native thread context.
 */
void
finalizeThread(Hjava_lang_Thread* tid)
{
	if (unhand(tid)->PrivateInfo != 0) {
		jthread_t jtid = (jthread_t)unhand(tid)->PrivateInfo;
		unhand(tid)->PrivateInfo = 0;
		jthread_destroy(jtid);
	}
}

/* 
 * extract name of a thread as a C string.
 * Returns static buffer.
 */
char *
nameThread(Hjava_lang_Thread *tid)
{
	static char buf[80];
	int i = 0;
	HArrayOfChar* name = unhand(tid)->name;
	while (i < sizeof buf - 1 && i < ARRAY_SIZE(name)) {
		buf[i] = ((short*)ARRAY_DATA(name))[i];
		i++;  
	}
	buf[i] = 0;
	return buf;
}

static void 
broadcastDeath(void *jlThread)
{
        Hjava_lang_Thread *tid = jlThread;
	int iLockRoot;

	/* Notify on the object just in case anyone is waiting */
	lockMutex(&tid->base);
	broadcastCond(&tid->base);
	unlockMutex(&tid->base);
}

static void 
throwDeath(void)
{
	Hjava_lang_Thread *cur = getCurrentThread();
	Hjava_lang_Throwable *death = cur->death;

	cur->death = NULL;
	throwException(death ? death : ThreadDeath);
}

static
void
runfinalizer(void)
{
	if (runFinalizerOnExit) {
		invokeFinalizer();
	}
}

/*
 * print info about a java thread.
 */
static void
dumpJavaThread(void *arg, jthread_t jtid, void *jlThread)
{
        Hjava_lang_Thread *tid = jlThread;
	printf("`%s' ", nameThread(tid));
	jthread_dumpthreadinfo((jthread_t)unhand(tid)->PrivateInfo);
	printf("\n");
}

void
dumpThreads(void)
{
	printf("Dumping live threads: (current = %p)\n",
	       jthread_current());
	jthread_walkLiveThreads(0, dumpJavaThread);
}

/*
 * Return the name of a java thread, given its native thread pointer.
 */
char*
nameNativeThread(void* native)
{
	return nameThread((Hjava_lang_Thread*)
		jthread_getcookie((jthread_t)native));
}

/*
 * Invoked when threading system detects a deadlock.
 */
static void
onDeadlock(void)
{
	if (!deadlockDetection) {
		return;
	}
#if defined(JTHREAD_RESTORE_FD)
        jthreadRestoreFD(2);
#endif
	dumpLocks();
	dumpThreads();
	printf("Deadlock: all threads blocked on internal events\n");
	fflush(stdout);
	ABORT();
}

void
initNativeThreads(int nativestacksize)
{
	/* Even though the underlying operating or threading system could
	 * probably extend the main thread's stack, we must impose this 
	 * artificial boundary, because otherwise we wouldn't be able to 
	 * catch stack overflow exceptions thrown by the main thread.
	 */
	threadStackSize = nativestacksize;
	jthread_init(
		DBGEXPR(JTHREADNOPREEMPT, false, true),
		java_lang_Thread_MAX_PRIORITY+1,
		java_lang_Thread_MIN_PRIORITY,
		thread_malloc,
		thread_free,
		broadcastDeath,
		throwDeath,
		onDeadlock);
}
