/*
 * threadData.h
 * Manage any thread specific data.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "assert_pp.h"
#include "gtypes.h"
#include "jthread.h"
#include "lockPool.h"
#include "gc.h"
#include <janosvm/jteam.h>
#include <janosvm/jteamReservation.h>
#include "threadData.h"
#include "threadData-inline.h"

METAL_MEMORY_STACK(metalJournalStack, JOURNAL_STACK_SIZE);

void initThreadData(threadData *td, jthread_t jtid)
{
	require(td != 0);

	td->nextlk = 0;
	td->freeLock = obtainLock();
	ksemInit(&td->sem);
	td->flags = 0;
	td->visits = 0;
	/* Start up in the metal team, we'll get migrated later on... */
	td->ownerTeam = metalTeam;
	td->currentTeam = td->ownerTeam;
	td->boundary = 0;
	td->journal = createMemoryStack(0, JOURNAL_STACK_SIZE);
	initRootReservation(td);
}

void killThreadData(threadData *td, jthread_t jtid)
{
	require(td != 0);
	require(td->visits == 0);
	require(jtid != 0);

	if( td->freeLock )
		releaseLock(td->freeLock);
	td->freeLock = 0;
	td->flags = 0;
	td->visits = 0;
	if( kernelTeam )
	{
		/* Theres a kernelTeam return to it. */
		td->ownerTeam = kernelTeam;
		td->currentTeam = td->ownerTeam;
	}
	else
	{
		/* We're in bare metal mode, return to the metalTeam */
		td->ownerTeam = metalTeam;
		td->currentTeam = td->ownerTeam;
	}
}

void destroyThreadData(threadData *td)
{
	require(td != 0);
	
	//ksemDestroy(&td->sem);
	deleteMemoryStack(td->journal);
}

int inheritThreadData(threadData *child, threadData *parent, jthread_t jtid)
{
	memoryStack *ms = 0;
	int retval = 0;
	iLock *il;
	
	require(child != 0);
	require(parent != 0);
	require(jtid != 0);
	require(jthread_get_data(jtid) == child);

	/*
	 * Make sure the team is still alive before proceeding.  This
	 * should be called within the lock used to track threads so we
	 * shouldn't have to worry about any race conditions with
	 * killing team threads.
	 *
	 * XXX This is prolly too dependent on unix-jthreads
	 */
	if( (il = obtainLock()) &&
	    (ms = createMemoryStack(0, JOURNAL_STACK_SIZE)) &&
	    allowTeamThread(parent->currentTeam) )
	{
		child->nextlk = 0;
		child->freeLock = il;
		ksemInit(&child->sem);
		child->flags = 0;
		child->visits = 0;
		/*
		 * This thread is owned by the team that the caller is
		 * executing within.
		 */
		child->currentTeam = parent->currentTeam;
		child->ownerTeam = parent->currentTeam;
		child->boundary = 0;
		child->journal = ms;
		initRootReservation(child);
		retval = 1;
	}
	else if( il )
	{
		deleteMemoryStack(ms);
		releaseLock(il);
	}
	return( retval );
}

void printThreadData(FILE *file, threadData *td)
{
	require(file != 0);
	require(td != 0);
	
	fprintf(file,
		"threadData[nextlk=%p; flags=%lx; visits=%d; "
		"currentTeam=%p; ownerTeam=%p]",
		td->nextlk,
		td->flags,
		td->visits,
		td->currentTeam,
		td->ownerTeam);
}
