/*
 * threadData.h
 * Manage any thread specific data.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _threaddata_h
#define _threaddata_h

#include "ksem.h"
#include "lockPool.h"
#include <misc/memoryStack.h>

struct _jteam;
struct _jteamBoundary;
struct _jthread;
struct _iLock;

/*
 * The threadData structure is used to track thread local data for the VM.
 * Each jthread should contain a threadData structure in one way or another and
 * is accessible through the jthread_get_data() function call.
 *
 * `nextlk' - Link used by the java locking system.
 * `freeLock' - The lock object used for grabbing a heavy lock.
 * `waiter' - The iLock the thread is waiting on.
 * `sem' - Semaphore used to block threads on java locks.
 * `flags' - Bitmask of flags for this thread.
 * `visits' - The number of teams we are visiting.
 * `currentTeam' - The current team we are executing within.
 * `ownerTeam' - Our owner team.
 * `boundary' - The current boundary for the thread.
 * `journal' - The journal for this thread.  Used for tracking interesting
 *   events that might need to be backed out of when "Bad Things" happen.
 */

enum {
	TDB_ZOMBIE	/*
			 * Someone attempted to slaughter us while
			 * stop was disabled.
			 */
};

enum {
	TDF_ZOMBIE = (1L << TDB_ZOMBIE)
};

typedef struct _threadData {
	struct _jthread *teamlk;
	struct _jthread *nextlk;
	struct _iLock *freeLock;
	struct _iLock **waiter;
	Ksem sem;
	unsigned long flags;
	int visits;
	struct _jteam *currentTeam;
	struct _jteam *ownerTeam;
	struct _jteamBoundary *boundary;
	memoryStack *journal;
	void *userData;
} threadData;

#define JOURNAL_STACK_SIZE 2048

DECLARE_METAL_MEMORY_STACK(metalJournalStack, JOURNAL_STACK_SIZE);

/* Define the thread data for the metal thread */
#define METAL_THREAD_DATA \
{ \
	0, \
	0, \
	&metalILock, \
	0, \
	KSEM_STATIC_INITIALIZER, \
	0, \
	0, \
	&metalTeamAndResources.team, \
	&metalTeamAndResources.team, \
	0, \
	&metalJournalStack.ms, \
        NULL \
}

/*
 * Initialize the threadData structure for the first jthread.  NOTE: This
 * function expects the calling thread to be unkillable (i.e. interrupts
 * disabled, in a visit).
 *
 * `td' - The threadData object to initialize.
 * `jtid' - Unused at the moment...
 */
void initThreadData(threadData *td, struct _jthread *jtid);
/*
 * Finalize the threadData structure for any thread.  NOTE: This function
 * expects the calling thread to be unkillable (i.e. interrupts disabled, in a
 * visit).
 *
 * `td' - The threadData object to finalize.
 * `jtid' - Unused at the moment...
 */
void killThreadData(threadData *td, struct _jthread *jtid);
/*
 * Release any memory held by the thread data structure.  This is separate
 * from killThreadData because it requires a real thread in a less restrictive
 * calling context, whereas killThreadData will prolly be called in the dying
 * threads context.
 *
 * `td' - The threadData object that still has allocated memory.
 */
void destroyThreadData(threadData *td);
/*
 * Inherit thread data from the creating thread.  Basically, this is like
 * initThreadData except that the new thread will be owned by whatever
 * the team the current thread is in.
 *
 * `child' - The threadData object for the new thread.
 * `parent' - The threadData object for the current thread.
 * `jtid' - Unused at the moment...
 */
int inheritThreadData(threadData *child, threadData *parent,
		      struct _jthread *jtid);
/*
 * Take the thread lock from the given threadData object and clears the
 * freeLock field.
 *
 * `td' - The threadData object to take the lock from, not necessarily the
 *   current thread.
 * returns - An iLock object.
 */
iLock *takeThreadLock(threadData *td);
/*
 * Give a threadData object the iLock object that was retrieved with
 * takeThreadLock.
 *
 * `td' - The threadData object to give the lock to.
 * `il' - The iLock object to be given.
 */
void giveThreadLock(threadData *td, iLock *il);
/*
 * Print out the state of the threadData structure.
 * (For debugging purposes only)
 */
void printThreadData(FILE *file, threadData *td);

#endif
