/*
 * feedback.c
 * Routines for generating information that can be fed back into kaffe for
 * future runs.
 *
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Contributed by the Flux Research Group, Department of Computer Science,
 * University of Utah, http://www.cs.utah.edu/flux/
 */

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "gtypes.h"
#include "access.h"
#include "classMethod.h"
#include "lookup.h"
#include "support.h"
#include "stringSupport.h"
#include "methodCache.h"
#include "thread.h"
#include "external.h"
#include "module.h"

#include "feedback.h"
#include "feedbackSections.h"

/* XXX Temporary for now until we define interface header file that
 * declares "translate"
 */
#if defined(TRANSLATOR) 
#if defined(JIT3)
#include "jit3/machine.h"
#else
#include "jit/machine.h"
#endif
#endif

static int initFeedbackModule(int stage);

module_t feedbackModule = {
	"feedback",
	0,
	initFeedbackModule,
	0,
	0
};

sectionFile *kaffe_feedback_file = 0;
char *feedback_filename = 0;

static int initFeedbackModule(int stage)
{
	if( feedback_filename )
	{
		if( feedbackFile(feedback_filename) )
		{
			syncFeedback();
			processFeedback();
		}
	}
	return( 1 );
}

int feedbackFile(char *filename)
{
	int retval = 0;

	/* Create a section file object to store the feedback information */
	if( (kaffe_feedback_file = createSectionFile()) )
	{
		setSectionFileName(kaffe_feedback_file, filename);
		/*
		 * Add an atexit call to syncFeedback so we know the
		 * information got out
		 */
		if( !atexit((void (*)(void))syncFeedback) )
		{
			retval = 1;
		}
		else
		{
			deleteSectionFile(kaffe_feedback_file);
			kaffe_feedback_file = 0;
		}
	}
	else
	{
		fprintf(stderr,
			"Warning: Can't allocate section file for "
			"feedback.\n");
	}
	return( retval );
}

int syncFeedback(void)
{
	int retval = 0;

	if( kaffe_feedback_file )
	{
		if( syncSectionFile(kaffe_feedback_file, 1) )
		{
			retval = 1;
		}
		else
		{
			fprintf(stderr,
				"Warning: Unable to sync with feedback "
				"file.\n");
		}
	}
	return( retval );
}

static void feedbackRunnable(void *arg)
{
	/* Walk over the sections */
	walkFileSections(kaffe_feedback_file);
}

int processFeedback(void)
{
	int retval = 1;

	if( kaffe_feedback_file )
	{
		feedbackRunnable(0);
		retval = 1;
	}
	return( retval );
}

int feedbackJITMethod(char *method, char *address, int size, int precompile)
{
	sectionFileData *sfd;
	jitSectionData *jsd;
	int retval = 0, iLockRoot;

	if( !kaffe_feedback_file )
		return( 0 );
	lockMutex(kaffe_feedback_file);
	if( !(sfd = findSectionInFile(kaffe_feedback_file,
				      &jitSectionType, method)) )
	{
		/*
		 * If the section doesn't exist we need to create and add it.
		 * We only set precompile here since the user might've changed
		 * the file to specify otherwise.
		 */
		if( (sfd = createFileSection(jitSectionType.name, method)) )
		{
			jsd = LOCAL_SECTION_DATA(jitSectionData *,
						 &jitSectionType, sfd);
			jsd->flags = precompile ? JSDF_PRECOMPILE : 0;
			addSectionToFile(kaffe_feedback_file, sfd);
		}
	}
	if( sfd )
	{
		/* Set whatever attributes are interesting */
		jsd = LOCAL_SECTION_DATA(jitSectionData *,
					 &jitSectionType, sfd);
		jsd->address = address;
		jsd->size = size;
		retval = 1;
	}
	unlockMutex(kaffe_feedback_file);
	return( retval );
}

int feedbackLibrary(char *name, int preload)
{
	sectionFileData *sfd;
	int retval = 0, iLockRoot;

	if( !kaffe_feedback_file )
		return( 0 );
	lockMutex(kaffe_feedback_file);
	if( !(sfd = findSectionInFile(kaffe_feedback_file,
				      &libSectionType, name)) )
	{
		/*
		 * If the section doesn't exist we need to create and add it.
		 * We only set preload here since the user might've changed
		 * the file to specify otherwise.
		 */
		if( (sfd = createFileSection(libSectionType.name, name)) )
		{
			libSectionData *lsd;

			lsd = LOCAL_SECTION_DATA(libSectionData *,
						 &libSectionType,
						 sfd);
			lsd->flags = preload ? LSDF_PRELOAD : 0;
			addSectionToFile(kaffe_feedback_file, sfd);
			retval = 1;
		}
	}
	else
	{
		retval = 1;
	}
	unlockMutex(kaffe_feedback_file);
	return( retval );
}
