/*
 * feedbackSections.c
 * Definitions for feedback file sections
 *
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Contributed by the Flux Research Group, Department of Computer Science,
 * University of Utah, http://www.cs.utah.edu/flux/
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "module.h"
#include "jmalloc.h"
#include "gtypes.h"
#include "access.h"
#include "classMethod.h"
#include "lookup.h"
#include "external.h"
#include "stringSupport.h"

#include "sectionFile.h"
#include "feedbackSections.h"

/* XXX Temporary for now until we define interface header file that
 * declares "translate"
 */
#if defined(TRANSLATOR) 
#if defined(JIT3)
#include "jit3/machine.h"
#else
#include "jit/machine.h"
#endif
#endif

/* Internal functions */

static int initFeedbackSections(int stage);

/* jitSectionType functions */
static int jitInit(struct _sectionType *st, struct _sectionFileData *sfd);
static int jitCache(struct _sectionType *st, struct _sectionFileData *sfd,
		    void *ps, char *line, int len);
static int jitFlush(struct _sectionType *st, struct _sectionFileData *sfd,
		    void *ps, char *line, int len, FILE *out_file);
static int jitSectionWalker(struct _sectionType *st,
			    struct _sectionFileData *sfd,
			    struct _sectionFile *sf);

/* libSectionType functions */
static int libInit(struct _sectionType *st, struct _sectionFileData *sfd);
static int libCache(struct _sectionType *st, struct _sectionFileData *sfd,
		    void *ps, char *line, int len);
static int libFlush(struct _sectionType *st, struct _sectionFileData *sfd,
		    void *ps, char *line, int len, FILE *out_file);
static int libSectionWalker(struct _sectionType *st,
			    struct _sectionFileData *sfd,
			    struct _sectionFile *sf);

static sectionTypeFunctions jitSectionFunctions = {
	0,
	0,
	jitInit,
	0,
	jitCache,
	jitFlush,
	jitSectionWalker
};

static sectionTypeFunctions libSectionFunctions = {
	0,
	0,
	libInit,
	0,
	libCache,
	libFlush,
	libSectionWalker
};

module_t feedbackSectionsModule = {
	"feedbackSections",
	0,
	initFeedbackSections,
	0,
	0
};

sectionType jitSectionType;
sectionType libSectionType;

static
int initFeedbackSections(int stage)
{
	/* Setup and install the handlers */
	installSectionType(&jitSectionType,
			   STA_Name, "jit-code",
			   STA_Functions, &jitSectionFunctions,
			   STA_TAG_DONE);
	installSectionType(&libSectionType,
			   STA_Name, "library",
			   STA_Functions, &libSectionFunctions,
			   STA_TAG_DONE);
	return( 1 );
}

/* Set a value in the jit section structure by name */
static int setJITSectionValue(jitSectionData *jsd,
			      char *tag, char *value)
{
	int retval = 1;
	
	if( !strcmp(tag, "precompile") )
	{
		jsd->flags = parseFlagString(value,
					     jsd->flags,
					     JSDF_PRECOMPILE);
	}
	else if( !strcmp(tag, "address") )
	{
		void *addr;
		
		if( sscanf(value, "%p", &addr) == 1 )
			jsd->address = addr;
		else
			retval = 0;
	}
	else if( !strcmp(tag, "size") )
	{
		if( sscanf(value, "%lu", &jsd->size) != 1 )
			retval = 0;
	}
	return( retval );
}

/* Write out the current value of a jit attribute */
static int writeJITSectionValue(jitSectionData *jsd,
				void *ps,
				char *tag, char *value,
				FILE *out_file)
{
	char scratch[128];
	int retval = 1;

	/*
	 * Figure out which attribute we're processing and construct the new
	 * value
	 */
	if( !tag )
	{
		tag = "";
		value = "";
	}
	else if( !strcmp(tag, "precompile") )
	{
		value = makeFlagString(jsd->flags, JSDF_PRECOMPILE, value);
	}
	else if( !strcmp(tag, "size") )
	{
		sprintf(scratch, "%lu", jsd->size);
		value = scratch;
	}
	else if( !strcmp(tag, "address") )
	{
		sprintf(scratch, "%p", jsd->address);
		value = scratch;
	}
	/*
	 * parseSectionLine will do the actual writing so its formatted
	 * correctly
	 */
	retval = parseSectionLine(ps, &tag, &value, out_file);
	return( retval );
}

int jitInit(struct _sectionType *st, struct _sectionFileData *sfd)
{
	jitSectionData *jsd = LOCAL_SECTION_DATA(jitSectionData *, st, sfd);
	int retval = 1;
	
	/* Initialize out section */
	jsd->flags = 0;
	jsd->size = 0;
	jsd->address = 0;
	return( retval );
}

int jitCache(struct _sectionType *st, struct _sectionFileData *sfd,
	     void *ps, char *line, int len)
{
	jitSectionData *jsd = LOCAL_SECTION_DATA(jitSectionData *, st, sfd);
	char *tag = 0, *value = 0;
	int retval = 1;

	/* Read the line in and set the value */
	parseSectionLine(ps, &tag, &value, 0);
	if( tag )
		retval = setJITSectionValue(jsd, tag, value);
	return( retval );
}

int jitFlush(struct _sectionType *st, struct _sectionFileData *sfd,
	     void *ps, char *line, int len, FILE *out_file)
{
	jitSectionData *jsd = LOCAL_SECTION_DATA(jitSectionData *, st, sfd);
	char *tag, *value;
	int retval = 1;
	
	if( line )
	{
		/* Parse the line and write out the current value */
		retval = parseSectionLine(ps, &tag, &value, 0);
		retval = writeJITSectionValue(jsd, ps,
					      tag, value,
					      out_file);
	}
	else if( sfd->flags & SFDF_DIRTY )
	{
		/*
		 * The section wasn't in the file before, write the whole thing
		 * out
		 */
		fprintf(out_file,
			"\tprecompile %s\n",
			jsd->flags & JSDF_PRECOMPILE ? "true" : "false");
		fprintf(out_file, "\tsize %lu\n", jsd->size);
		fprintf(out_file, "\taddress %p\n", jsd->address);
	}
	else
	{
	}
	return( retval );
}

int jitSectionWalker(struct _sectionType *st, struct _sectionFileData *sfd,
		     struct _sectionFile *sf)
{
	int retval = 1;
	
#if defined(TRANSLATOR)
	jitSectionData *jsd = LOCAL_SECTION_DATA(jitSectionData *, st, sfd);
	
	if( jsd->flags & JSDF_PRECOMPILE )
	{
		int len, lpc, sig_start = -1, meth_start = -1;
		Utf8Const *u8cname, *u8mname, *u8sig;
		Hjava_lang_Class *cls;
		char *full_name;
		
		/*
		 * Parse the name of the section to get the class, method, and
		 * signature
		 */
		full_name = sfd->name;
		len = strlen(full_name);
		for( lpc = len - 1; (lpc >= 0) && (meth_start < 0); lpc-- )
		{
			switch( full_name[lpc] )
			{
			case '(':
				sig_start = lpc;
				break;
			case '/':
				if( sig_start > 0 )
					meth_start = lpc + 1;
				break;
			}
		}
		if( (sig_start > 0) && (meth_start > 0) )
		{
			jobject loader = 0;
			errorInfo info;
			
			/* Get the right strings and find the class */
			u8cname = utf8ConstNew(full_name, meth_start - 1);
			u8mname = utf8ConstNew(&full_name[meth_start],
					       sig_start - meth_start);
			u8sig = utf8ConstNew(&full_name[sig_start],
					     len - sig_start);
			if( u8cname && u8mname && u8sig &&
			    (cls = loadClass(u8cname, loader, &info)) )
			{
				Method *meth;
				
				if( (meth = findMethodLocal(cls,
							    u8mname,
							    u8sig)) &&
				    !(meth->accflags & ACC_NATIVE) )
				{
					if( translate(meth, &info) )
					{
					}
					else
					{
						fprintf(stderr,
							"Feedback: Precompile "
							"failed for %s!\n",
							full_name);
					}
				}
				else if( !meth )
				{
					fprintf(stderr,
						"Feedback: Didn't find method"
						" %s\n",
						full_name);
				}
			}
			else
			{
				fprintf(stderr,
					"Feedback: Couldn't load class %s\n",
					u8cname->data);
			}
			utf8ConstRelease(u8cname);
			utf8ConstRelease(u8mname);
			utf8ConstRelease(u8sig);
		}
		else
		{
			fprintf(stderr,
				"Feedback: Malformed method `%s'\n",
				full_name);
		}
	}
#else
	{
		static int precompile_msg = 0;
		
		if( !precompile_msg )
		{
			precompile_msg = 1;
			fprintf(stderr,
				"Feedback: Cannot precompile java for "
				"the interpreter\n");
		}
	}
#endif
	return( retval );
}

/* Set a value in the lib section structure by name */
static int setLibSectionValue(libSectionData *lsd,
			      char *tag, char *value)
{
	int retval = 1;

	if( !strcmp(tag, "preload") )
	{
		lsd->flags = parseFlagString(value,
					     lsd->flags,
					     LSDF_PRELOAD);
	}
	return( retval );
}

/* Write out the current value of a jit attribute */
static int writeLibSectionValue(libSectionData *lsd,
				void *ps,
				char *tag, char *value,
				FILE *out_file)
{
	int retval = 1;

	/*
	 * Figure out which attribute we're processing and construct the new
	 * value
	 */
	if( !tag )
	{
		tag = "";
		value = "";
	}
	else if( !strcmp(tag, "preload") )
	{
		value = makeFlagString(lsd->flags, LSDF_PRELOAD, value);
	}
	/*
	 * parseSectionLine will do the actual writing so its formatted
	 * correctly
	 */
	retval = parseSectionLine(ps, &tag, &value, out_file);
	return( retval );
}

int libInit(struct _sectionType *st, struct _sectionFileData *sfd)
{
	libSectionData *lsd = LOCAL_SECTION_DATA(libSectionData *, st, sfd);
	int retval = 1;

	lsd->flags = 0;
	return( retval );
}

int libCache(struct _sectionType *st, struct _sectionFileData *sfd,
	     void *ps, char *line, int len)
{
	libSectionData *lsd = LOCAL_SECTION_DATA(libSectionData *, st, sfd);
	char *tag = 0, *value = 0;
	int retval = 1;

	/* Read the line in and set the value */
	parseSectionLine(ps, &tag, &value, 0);
	if( tag )
		retval = setLibSectionValue(lsd, tag, value);
	return( retval );
}

int libFlush(struct _sectionType *st, struct _sectionFileData *sfd,
	     void *ps, char *line, int len, FILE *out_file)
{
	libSectionData *lsd = LOCAL_SECTION_DATA(libSectionData *, st, sfd);
	char *tag, *value;
	int retval = 1;

	if( line )
	{
		/* Parse the line and write out the current value */
		retval = parseSectionLine(ps, &tag, &value, 0);
		retval = writeLibSectionValue(lsd, ps,
					      tag, value,
					      out_file);
	}
	else if( sfd->flags & SFDF_DIRTY )
	{
		/*
		 * The section wasn't in the file before, write the whole thing
		 * out
		 */
		fprintf(out_file,
			"\tpreload %s\n",
			lsd->flags & LSDF_PRELOAD ?
			"true" : "false");
	}
	else
	{
	}
	return( retval );
}

int libSectionWalker(struct _sectionType *st, struct _sectionFileData *sfd,
		     struct _sectionFile *sf)
{
	libSectionData *lsd = LOCAL_SECTION_DATA(libSectionData *, st, sfd);
	int retval = 1;

	if( lsd->flags & LSDF_PRELOAD )
		loadNativeLibrary2(sfd->name, 0, 0, 0);
	return( retval );
}
