/*
 * logger.c
 * A generic log implementation
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <assert.h>
#include <string.h>

#include "assert_pp.h"

#include "module.h"
#include "memoryStack-inline.h"
#include "logger.h"

#define LOG_STACK_DEFAULT_SIZE 2048

/** Module declarations **/
static int testLoggerModule(void);

module_t loggerModule = {
	"logger",
	0,
	0,
	0,
	testLoggerModule
};

/** Module functions **/

static
int dummyTestFlushArg;

static
void testFlusher(void *arg,
		 struct _logger *l,
		 struct _logEntry *le,
		 struct _logEntryAttribute **attrs)
{
	static char *testValues[] = {
		"Buffy",
		"Xander",
		"Willow",
		"A slightly longer test case",
		"true",
		"b",
		"cafef00d",
		"3.14",
		"0x12345678"
	};
	char *inString, buf[32];
	int lpc, index;
	
	assert(arg == &dummyTestFlushArg);
	for( lpc = 0; lpc < le->attributeCount; lpc++ )
	{
		inString = buf;
		
		sscanf(attrs[lpc]->id, "%d", &index);
		switch( attrs[lpc]->type )
		{
		case LEA_Boolean:
			sprintf(inString, "%s",
				attrs[lpc]->value.b ? "true" : "false");
			break;
		case LEA_Character:
			sprintf(inString, "%c", attrs[lpc]->value.c);
			break;
		case LEA_Integer:
			sprintf(inString, "%x", attrs[lpc]->value.i);
			break;
		case LEA_Float:
			sprintf(inString, "%3.2f", attrs[lpc]->value.f);
			break;
		case LEA_String:
			inString = attrs[lpc]->value.str;
			assert(inString == (char *)(attrs[lpc] + 1));
			break;
		case LEA_Pointer:
			sprintf(inString, "%p", attrs[lpc]->value.ptr);
			break;
		default:
			assert(0);
			break;
		}
		assert(!strcmp(inString, testValues[index]));
	}
}

static
void testFlusher2(void *arg,
		  struct _logger *l,
		  struct _logEntry *le,
		  struct _logEntryAttribute **attrs)
{
	int *i_ptr = (int *)arg;
	
	(*i_ptr)++;
}

static
void nullFlusher(void *arg,
		 struct _logger *l,
		 struct _logEntry *le,
		 struct _logEntryAttribute **attrs)
{
}

static
int testNoIDMask(struct _logAnalyzerMask *lam,
		 struct _logger *l,
		 struct _logEntry *le,
		 struct _logEntryAttribute **attrs)
{
	int retval = 0;
	
	if( attrs[0]->id[0] == '\0' )
		retval = 1;
	return( retval );
}

static
int testAnalyzerCount;

static
void testAnalyzer(struct _logAnalyzerMask *lam,
		  struct _logger *l,
		  struct _logEntry *le,
		  struct _logEntryAttribute **attrs)
{
	testAnalyzerCount++;
}

int testLoggerModule(void)
{
	int count, retval = 1;
	logger l;

	if( initLogger(&l,
		       LA_Name, "test-log",
		       LA_FlushFunction, testFlusher,
		       LA_FlushFunctionArg, &dummyTestFlushArg,
		       LA_TAG_DONE) )
	{
		assert(!strcmp(l.name, "test-log"));
		addLogEntry(&l, "test",
			    LEA_String, "0", "Buffy", -1,
			    LEA_TAG_DONE);
		addLogEntry(&l, "test",
			    LEA_String, "1", "Xander", -1,
			    LEA_String, "2", "Willow", -1,
			    LEA_String, "3", "A slightly longer test case", -1,
			    LEA_TAG_DONE);
		addLogEntry(&l, "test",
			    LEA_Boolean, "4", 1,
			    LEA_Character, "5", 'b',
			    LEA_Integer, "6", 0xcafef00d,
			    LEA_Float, "7", 3.14159,
			    LEA_Pointer, "8", (void *)0x12345678,
			    LEA_TAG_DONE);
		flushLogger(&l);
		killLogger(&l);
	}
	count = 0;
	if( initLogger(&l,
		       LA_Name, "test-log2",
		       LA_FlushThreshold, 512,
		       LA_FlushFunction, testFlusher2,
		       LA_FlushFunctionArg, &count,
		       LA_TAG_DONE) )
	{
		int lpc;

		for( lpc = 0; lpc < 16; lpc++ )
		{
			addLogEntry(&l, "test",
				    LEA_String, "a", "Play netrek", -1,
				    LEA_TAG_DONE);
		}
		assert(count > 0);
		assert(count < 16);
		killLogger(&l);
	}
	if( initLogger(&l,
		       LA_Name, "test-log3",
		       LA_FlushFunction, nullFlusher,
		       LA_TAG_DONE) )
	{
		logAnalyzerMask lam;
		logAnalyzer la;

		lam.mask = testNoIDMask;
		lam.analyzer = &la;
		la.func = testAnalyzer;
		addLogAnalyzer(&l, &lam);
		addLogEntry(&l, "test",
			    LEA_String, "a", "", -1,
			    LEA_TAG_DONE);
		addLogEntry(&l, "test",
			    LEA_String, "", "", -1,
			    LEA_TAG_DONE);
		addLogEntry(&l, "test",
			    LEA_String, "a", "", -1,
			    LEA_TAG_DONE);
		addLogEntry(&l, "test",
			    LEA_String, "", "", -1,
			    LEA_TAG_DONE);
		addLogEntry(&l, "test",
			    LEA_String, "a", "", -1,
			    LEA_TAG_DONE);
		addLogEntry(&l, "test",
			    LEA_String, "", "", -1,
			    LEA_TAG_DONE);
		addLogEntry(&l, "test",
			    LEA_String, "a", "", -1,
			    LEA_TAG_DONE);
		addLogEntry(&l, "test",
			    LEA_String, "", "", -1,
			    LEA_TAG_DONE);
		killLogger(&l);
		assert(testAnalyzerCount == 4);
	}
	return( retval );
}

/** logger functions **/
static
char *attrTypes[] = {
	"none",
	"bool",
	"char",
	"int",
	"float",
	"string",
	"ptr",
};

static
void defaultLogFlushFunction(void *arg,
			     struct _logger *l,
			     struct _logEntry *le,
			     struct _logEntryAttribute **lea)
{
	if( le->classification )
	{
		int lpc;
		
		kaffe_dprintf("%s:%s:%d:%s[%ld.%ld] -",
			      l->name,
			      le->sourceFile,
			      le->sourceLine,
			      le->classification,
			      (long)(le->recordTime.tv_sec),
			      (long)(le->recordTime.tv_usec));
		for( lpc = 0; lpc < le->attributeCount; lpc++ )
		{
			if( lpc > 0 )
				kaffe_dprintf(",");
			kaffe_dprintf(" %s:%s=",
				      lea[lpc]->id,
				      attrTypes[lea[lpc]->type]);
			switch( lea[lpc]->type )
			{
			case LEA_Boolean:
				kaffe_dprintf("%s",
					      lea[lpc]->value.b ?
					      "true" : "false");
				break;
			case LEA_Character:
				kaffe_dprintf("%c", lea[lpc]->value.c);
				break;
			case LEA_Integer:
				kaffe_dprintf("%d", lea[lpc]->value.i);
				break;
			case LEA_Float:
				kaffe_dprintf("%f", lea[lpc]->value.f);
				break;
			case LEA_String:
				kaffe_dprintf("%s", lea[lpc]->value.str);
				break;
			case LEA_Pointer:
				kaffe_dprintf("%p", lea[lpc]->value.ptr);
				break;
			default:
				assert(0);
				break;
			}
		}
		kaffe_dprintf("\n");
	}
}

#define DEFAULT_ATTRIBUTE_ARRAY_LENGTH 10

int initLoggerV(logger *l, va_list args)
{
	int tag, log_size = LOG_STACK_DEFAULT_SIZE, retval = 0;
 
	require(l != 0);

	memset(l, 0, sizeof(logger));
	l->name = "";
	l->flush = defaultLogFlushFunction;
	tag = va_arg(args, int);
	while( tag != LA_TAG_DONE )
	{
		switch( tag )
		{
		case LA_Name:
			l->name = va_arg(args, char *);
			break;
		case LA_LogSize:
			log_size = va_arg(args, int);
			break;
		case LA_FlushThreshold:
			l->flushThreshold = va_arg(args, int);
			break;
		case LA_FlushFunction:
			l->flush = va_arg(args, void *);
			break;
		case LA_FlushFunctionArg:
			l->flushArg = va_arg(args, void *);
			break;
		default:
			break;
		}
		tag = va_arg(args, int);
	}
	if( l->flushThreshold <= 0 )
	{
		l->flushThreshold = (log_size * 3) / 4;
	}
	if( (l->stack = createMemoryStack(0, log_size)) &&
	    (l->attributeArray.data = gc_malloc(sizeof(logEntryAttribute *) *
						DEFAULT_ATTRIBUTE_ARRAY_LENGTH,
						GC_ALLOC_FIXED)) )
	{
		l->attributeArray.length = DEFAULT_ATTRIBUTE_ARRAY_LENGTH;
		retval = 1;
	}
	else if( l->stack )
	{
		deleteMemoryStack(l->stack);
	}
	return( retval );
}

int initLogger(logger *l, ...)
{
	va_list args;
	int retval;

	require(l != 0);
	
	va_start(args, l);
	retval = initLoggerV(l, args);
	va_end(args);
	return( retval );
}

void killLogger(logger *l)
{
	if( l )
	{
		flushLogger(l);
		deleteMemoryStack(l->stack);
	}
}

void addLogEntryV(logger *l, char *srcFile, int srcLine, char *cl,
		  va_list args)
{
	memoryStack *ms;
	logEntry *le;

	ms = l->stack;
	if( (le = allocStackSpace(ms, sizeof(logEntry))) )
	{
		int entry_size, count = 0, tag, error = 0;

		entry_size = sizeof(logEntry);
		gettimeofday(&le->recordTime, 0);
		le->sourceFile = srcFile;
		le->sourceLine = srcLine;
		le->classification = cl;
		tag = va_arg(args, int);
		while( !error && (tag != LEA_TAG_DONE) )
		{
			int attr_size = sizeof(logEntryAttribute);
			logEntryAttribute *lea;
			char *id, *string = 0;
			int str_size = 0;

			id = va_arg(args, char *);
			switch( tag )
			{
			case LEA_String:
				string = va_arg(args, char *);
				str_size = va_arg(args, int);
				if( str_size == -1 )
				{
					if( string )
						str_size = strlen(string);
					else
						str_size = 0;
				}
				attr_size += str_size + 1;
				break;
			default:
				break;
			}
			if( (lea = allocStackSpace(ms, attr_size)) )
			{
				entry_size += attr_size;
				lea->type = tag;
				lea->id = id;
				switch( tag )
				{
				case LEA_Boolean:
					lea->value.b = va_arg(args, int);
					break;
				case LEA_Character:
					lea->value.c = va_arg(args, int);
					break;
				case LEA_Integer:
					lea->value.i = va_arg(args, int);
					break;
				case LEA_Float:
					lea->value.f = va_arg(args, double);
					break;
				case LEA_String:
					lea->value.str = (char *)(lea + 1);
					strncpy(lea->value.str,
						string,
						str_size);
					lea->value.str[str_size] = '\0';
					break;
				case LEA_Pointer:
					lea->value.ptr = va_arg(args, void *);
					break;
				default:
					assert(0);
					break;
				}
			}
			else
			{
				error = 1;
			}
			tag = va_arg(args, int);
			count++;
		}
		if( !error && (count > l->attributeArray.length) )
		{
			void *new_data;
			
			if( (new_data =
			     gc_realloc(l->attributeArray.data,
					sizeof(logEntryAttribute) *
					(count +
					 (DEFAULT_ATTRIBUTE_ARRAY_LENGTH / 2)),
					GC_ALLOC_FIXED)) )
			{
				l->attributeArray.data = new_data;
				l->attributeArray.length =
					(count +
					 (DEFAULT_ATTRIBUTE_ARRAY_LENGTH / 2));
			}
			else
			{
				error = 1;
			}
		}
		if( error )
		{
			cutMemoryStack(ms, le);
			l->lostEntryCount++;
		}
		else
		{
			int lpc = 0, prev_size = 0;
			logEntryAttribute *lea;
			memoryStack *curr = 0;
			logAnalyzerMask *lam;
			void *prev;
			
			le->attributeCount = count;
			l->currentSize += entry_size;
			prev = le;
			prev = walkMemoryStack(ms, &curr,
					       prev, sizeof(logEntry));
			prev_size = sizeof(logEntry);
			while( (lea = walkMemoryStack(ms, &curr,
						      prev, prev_size)) )
			{
				prev = lea;
				prev_size = sizeof(logEntryAttribute);
				switch( lea->type )
				{
				case LEA_String:
					prev_size +=
						strlen(lea->value.str)
						+ 1;
					break;
				default:
					break;
				}
				l->attributeArray.data[lpc] = lea;
				lpc++;
			}
			lam = l->analyzers;
			while( lam )
			{
				if( !lam->mask ||
				    lam->mask(lam,
					      l,
					      le,
					      l->attributeArray.data) )
				{
					lam->analyzer->func(
						lam,
						l,
						le,
						l->attributeArray.data);
				}
				lam = lam->next;
			}
			if( l->currentSize > l->flushThreshold )
			{
				flushLogger(l);
			}
		}
	}
	else
	{
		l->lostEntryCount++;
		fprintf(stderr,
			"WARNING: Memory is low, logging is non-functional\n");
	}
}

void addLogEntry_real(logger *l, char *srcFile, int srcLine, char *cl, ...)
{
	va_list args;

	va_start(args, cl);
	addLogEntryV(l, srcFile, srcLine, cl, args);
	va_end(args);
}

void addLogMessage_real(logger *l, char *srcFile, int srcLine,
			char *cl, char *msg)
{
	addLogEntry_real(l, srcFile, srcLine, cl,
			 LEA_String, "", msg, -1,
			 LEA_TAG_DONE);
}

void addLogAnalyzer(logger *l, logAnalyzerMask *lam)
{
	require(l != 0);
	require(lam != 0);
	require(lam->analyzer != 0);
	require(lam->analyzer->func != 0);
	
	lam->next = l->analyzers;
	l->analyzers = lam;
}

void remLogAnalyzer(logger *l, logAnalyzerMask *lam)
{
	logAnalyzerMask *curr, **prev;

	prev = &l->analyzers;
	curr = l->analyzers;
	while( curr != lam )
	{
		prev = &curr->next;
		curr = curr->next;
	}
	*prev = curr->next;
	lam->next = 0;
}

void flushLogger(logger *l)
{
	int done = 0, lpc, prev_size = 0;
	memoryStack *ms, *curr = 0;
	logEntryAttribute *lea;
	logEntry *le;
	void *prev;
	
	require(l != 0);
	
	ms = l->stack;
	prev = bottomOfMemoryStack(ms);
	while( !done )
	{
		if( (le = walkMemoryStack(ms, &curr, prev, prev_size)) )
		{
			prev = le;
			prev_size = sizeof(logEntry);
			for( lpc = 0; lpc < le->attributeCount; lpc++ )
			{
				lea = walkMemoryStack(ms, &curr,
						      prev, prev_size);
				prev = lea;
				prev_size = sizeof(logEntryAttribute);
				switch( lea->type )
				{
				case LEA_String:
					prev_size +=
						strlen(lea->value.str)
						+ 1;
					break;
				default:
					break;
				}
				l->attributeArray.data[lpc] = lea;
			}
			l->flush(l->flushArg,
				 l,
				 le,
				 l->attributeArray.data);
			l->flushedEntryCount++;
		}
		else
		{
			done = 1;
		}
	}
	cutMemoryStack(l->stack, bottomOfMemoryStack(l->stack));
	l->currentSize = 0;
}
