/*
 * logger.h
 * A generic log implementation
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _logger_h
#define _logger_h

#include <stdarg.h>
#include <sys/time.h>

#include "memoryStack.h"

/*
 * I was recently inspired to care more about logging so I built up this
 * infrastructure and figured I'd experiment with something a little more
 * sophisticated that `fprintf(stderr, ...)'.  Basically, I wanted something
 * that would make it easy for internal code to analyze and act upon its own
 * log entries.
 */

struct _logger;
struct _logEntry;
struct _logEntryAttribute;
struct _logAnalyzer;
struct _logAnalyzerMask;

/*
 * Callback function used when log entries need to be flushed from buffer
 * memory.
 *
 * `arg' - User defined arg as defined in l.flushArg.
 * `l' - The logger that is being flushed.
 * `le' - The current log entry that is being flushed.
 * `attrs' - An array of pointers to attributes that are paired with this
 *   log entry.  The attributeCount field in the logEntry will define how
 *   many elements are valid in this array.
 */
typedef void (*logFlushFunction)(void *arg,
				 struct _logger *l,
				 struct _logEntry *le,
				 struct _logEntryAttribute **attrs);

/*
 * Callback function used to test whether an analyzer function cares to be
 * told about a new log entry.
 *
 * `lam' - The logAnalyzerMask object that the function pointer was retrieved
 *   from.
 * `l' - The logger that the new log entry was added too.
 * `le' - The new log entry.
 * `attrs' - An array of pointers to attributes that are paired with this
 *   log entry.  The attributeCount field in the logEntry will define how
 *   many elements are valid in this array.
 */
typedef int (*logAnalyzerMaskFunction)(struct _logAnalyzerMask *lam,
				       struct _logger *l,
				       struct _logEntry *le,
				       struct _logEntryAttribute **attrs);

/*
 * Callback function used to analyze a log entry.
 *
 * `lam' - The logAnalyzerMask object that the function pointer was retrieved
 *   from.
 * `l' - The logger that the new log entry was added too.
 * `le' - The new log entry.
 * `attrs' - An array of pointers to attributes that are paired with this
 *   log entry.  The attributeCount field in the logEntry will define how
 *   many elements are valid in this array.
 */
typedef void (*logAnalyzerFunction)(struct _logAnalyzerMask *lam,
				    struct _logger *l,
				    struct _logEntry *le,
				    struct _logEntryAttribute **attrs);

/*
 * Function tags/type for logEntryAttributes.  When used in the addLogEntry
 * function the argument after the tag is the attributes unique identifier
 * followed by the attribute's value.
 */
typedef enum {
	LEA_TAG_DONE,	/* Terminate the tag list. */
	LEA_Boolean,	/* (char *, int) - A boolean attribute value */
	LEA_Character,	/* (char *, char) - A character attribute value */
	LEA_Integer,	/* (char *, int) - An integer attribute value */
	LEA_Float,	/* (char *, float) - A float attribute value */
	LEA_String,	/*
			 * (char *, char *, int) - A string attribute value.
			 * NOTE: The third argument is the length of the
			 * string, or -1 if the string is null terminated.
			 */
	LEA_Pointer,	/* (char *, void *) - A pointer attribute value */
} logger_type_t;

/*
 * A logEntryAttribute is used to store data associated with a logEntry.
 * More than one of these can be attached to an entry and we maintain type
 * information with the data so that its easier for the analyzers to do
 * their job.
 *
 * `id' - The identifier for this attribute.  This string isn't copied so its
 *   lifetime must be greater than the log.
 * `type' - The type of data held by this attribute.
 * `value.*' - The various values that can be held by an attribute.  NOTE:
 *   A string value stored in here is copied and stored immediately after
 *   the attribute structure.
 */
typedef struct _logEntryAttribute {
	char *id;
	logger_type_t type;
	union {
		char b;
		char c;
		int i;
		double f;
		char *str;
		void *ptr;
	} value;
} logEntryAttribute;

/*
 * A logEntry structure encodes a single entry in the log and acts as a
 * header for its attributes list.
 *
 * `recordTime' - The time that this entry was added to the log.
 * `classification' - The class of this entry.
 * `attributeCount' - The number of attributes associated with this entry.
 */
typedef struct _logEntry {
	char *sourceFile;
	int sourceLine;
	struct timeval recordTime;
	char *classification;
	int attributeCount;
} logEntry;

/*
 * The logger structure is the root for managing logs.  It records entries and
 * their attributes in a buffer until its full and then uses the flush function
 * to push the entries out to more permanent storage.  You can also attach
 * `analyzer' functions to the logger which can be used to analyze log entries
 * and take some action to resolve reocurring problems, or something.
 *
 * `name' - The name associated with this logger.  This is just a copy of the
 *   pointer, it is not duplicated internally.
 * `stack' - The stack that buffers our log entries and their attributes.
 * `currentSize' - The total size of the stack.
 * `flushThreshold' - The threshold for determining when to flush the log
 *   entries in the stack.
 * `lostEntryCount' - The number of log entries that were lost because of
 *   lack of memory or other errors.
 * `flushedEntryCount' - The number of log entries that we've flushed.
 * `flush' - The flush function, generally, this should write the entries in
 *   `stack' to a more permanent location.
 * `flushArg' - User defined arg that is passed to `flush'.
 * `analyzers' - The list of analyzers to apply to new log entries.
 * `attributeArray.data' - Array for storing pointers to the attributes
 *   for a log entry.  This array is reallocated during an addLogEntry so that
 *   it is atleast large enough to hold the maximum number of attributes for
 *   all log entries.
 * `attributeArray.length' - The length of the `attributeArray.data' array.
 */
typedef struct _logger {
	char *name;
	memoryStack *stack;
	int currentSize;
	int flushThreshold;
	int lostEntryCount;
	int flushedEntryCount;
	logFlushFunction flush;
	void *flushArg;
	struct _logAnalyzerMask *analyzers;
	struct {
		struct _logEntryAttribute **data;
		int length;
	} attributeArray;
} logger;

/*
 * The logAnalyzer structure is meant to be a base for more application
 * specific functionality.  For example, the analyzer function might want
 * to track how often some warning/error/whatever gets logged and take a
 * more forceful action if they are reocurring too often.
 *
 * `func' - The analyzer callback.
 */
typedef struct _logAnalyzer {
	logAnalyzerFunction func;
} logAnalyzer;

/*
 * The logAnalyzerMask is a layer of indirection for the logAnalyzer object.
 * It acts as the link in the logger's list and can be used to mask certain
 * entries from the actual analyzer function.  Also, it can be used to store
 * information specific to a single logger object if there are multiple
 * logs in the system.
 *
 * `next' - Link to the next logAnalyzerMask in the list.
 * `mask' - Callback function that will determine if the analyzer function
 *   should even see the log entry.
 * `analyzer' - The logAnalyzer that will do the actual work.
 */
typedef struct _logAnalyzerMask {
	struct _logAnalyzerMask *next;
	logAnalyzerMaskFunction mask;
	logAnalyzer *analyzer;
} logAnalyzerMask;

/*
 * Function tags for initLogger.
 */
enum {
	LA_TAG_DONE,		/* Terminate the tag list. */
	LA_Name,		/* (char *) - The name of the logger */
	LA_LogSize,		/*
				 * (int) - The `puddle' size for the internal
				 * log memory.
				 */
	LA_FlushThreshold,	/*
				 * (int) - The threshold for flushing the
				 * log.
				 */
	LA_FlushFunction,	/*
				 * (logFlushFunction) - The flush function
				 * for the log.
				 */
	LA_FlushFunctionArg,	/*
				 * (void *) - The user defined argument for
				 * the flush function.
				 */
};

/*
 * Initialize a preallocated logger object.
 *
 * `l' - An unitialized logger object.
 * `...' - A function tag list using one of the above `LA_*' tags.
 * returns - Non-zero if the initialization succeeded, zero otherwise.
 */
int initLogger(logger *l, ...);
/*
 * Finalize an initialized logger object.
 *
 * `l' - An initialized logger object.
 */
void killLogger(logger *l);
/*
 * Add a message to the log.
 *
 * `l' - The logger object to add the message too.
 * `cl' - The class of the message.
 * `msg' - The message to add to the log.
 */
void addLogMessage_real(logger *l, char *srcFile, int srcLine,
			char *cl, char *msg);
#define addLogMessage(l, cl, msg) \
	addLogMessage_real(l, __FILE__, __LINE__, cl, msg)
/*
 * A more generic way to add to the log.  Basically, this will add an entry
 * the log with whatever attributes you specify using the function tags.
 *
 * Example:
 *   addLogEntry(&myLog, "myModule",
 *               LEA_String, "Warning", "Something is going wrong", -1,
 *               LEA_Integer, "value", 2,
 *               LEA_TAG_DONE);
 *
 * `l' - The logger object to add the message too.
 * `cl' - The class of the message.
 * `...' - A function tag list using of the above `LEA_*' tags.  
 */
void addLogEntry_real(logger *l, char *srcFile, int srcLine, char *cl, ...);
#define addLogEntry(l, cl, args...) \
	addLogEntry_real(l, __FILE__, __LINE__, cl, ##args)
/* The va_list version of the above function */
void addLogEntryV(logger *l, char *srcFile, int srcLine, char *cl,
		  va_list args);
/*
 * Add a logAnalyzerMask to a logger.
 *
 * `l' - The log to add the analyzer too.
 * `lam' - The mask which refers to the analyzer to add.
 */
void addLogAnalyzer(logger *l, logAnalyzerMask *lam);
/*
 * Remove a logAnalyzerMask from a log.
 *
 * `l' - The log to remove the analyzer from.
 * `lam' - The mask which refers to the analyzer to remove.
 */
void remLogAnalyzer(logger *l, logAnalyzerMask *lam);
/*
 * Forcefully flush the log buffer.
 *
 * `l' - The log to flush.
 */
void flushLogger(logger *l);

/*
 * AddModule: loggerModule
 *   Comment: Just used to test the module
 */

#endif
