/*
 * memoryStack-inline.h
 * A generic stack implementation
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "memoryStack.h"
#include "mungWallProxy.h"

#define ADJUST_SIZE(x) \
	(((x) + (ALIGNMENTOF_VOIDP - 1)) & ~(ALIGNMENTOF_VOIDP - 1))

KAFFE_INLINE
void *allocStackSpace(memoryStack *root_ms, size_t size)
{
	memoryStack *ms;
	void *retval;

	require(root_ms != 0);
	require_valid_memory(root_ms);

	ms = root_ms->prev;
	size = ADJUST_SIZE(size);
	if( (ms->used + size) <= ms->max )
	{
		/* Fast path: There's enough memory in the current node. */
		retval = &ms->data.start[ms->used];
		ms->used += size;
	}
	else
	{
		/* Bah, not enough memory, need to allocate another node. */
		retval = allocStackSpace_slow(root_ms, size);
	}
	return( retval );
}

KAFFE_INLINE
void *allocStackSpace_shallow(memoryStack *root_ms, size_t size)
{
	void *retval = 0;
	memoryStack *ms;

	require(root_ms != 0);
	require_valid_memory(root_ms);

	ms = root_ms->prev;
	size = ADJUST_SIZE(size);
	if( (ms->used + size) <= ms->max )
	{
		/* Fast path: There's enough memory in the current node. */
		retval = &ms->data.start[ms->used];
		ms->used += size;
	}
	return( retval );
}

KAFFE_INLINE
void freeStackSpace(memoryStack *root_ms, size_t size)
{
	memoryStack *ms = root_ms->prev;
	
	require(root_ms != 0);
	require_valid_memory(root_ms);
	
	size = ADJUST_SIZE(size);
	if( size < ms->used )
	{
		/* Fast path: Just decrement the used amount */
		ms->used -= size;
#if defined(DEBUG)
		/* Poison the memory */
		memset(&ms->data.start[ms->used], 0xDEADBEEF, size);
#else
		/* Don't care */
#endif
	}
	else
	{
		/* size >= ms->used, we'll need to free some nodes. */
		freeStackSpace_slow(root_ms, size);
	}
}

KAFFE_INLINE
void *bottomOfMemoryStack(memoryStack *root_ms)
{
	require(root_ms != 0);
	require_valid_memory(root_ms);
	
	return (void *)root_ms->data.start;
}

KAFFE_INLINE
void *topOfMemoryStack(memoryStack *root_ms, size_t offset)
{
	memoryStack *ms;

	require(root_ms != 0);
	require_valid_memory(root_ms);
	
	ms = root_ms->prev;
	return &ms->data.start[ms->used - ADJUST_SIZE(offset)];
}

#undef ADJUST_SIZE
