/*
 * memoryStack.h
 * A generic stack implementation
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <config.h>
#include <assert_pp.h>

#include "gtypes.h"
#include "support.h"
#include "gc.h"
#include "mungWallProxy.h"
#include "module.h"
#include "memoryStack.h"
#include "memoryStack-inline.h"

/** Module declarations **/
static int testMemoryStackModule(void);

module_t memoryStackModule = {
	"memory-stack",
	0,
	0,
	0,
	testMemoryStackModule
};

/** Module functions **/

int testMemoryStackModule(void)
{
	memoryStack *ms;
	int retval = 1;

#define TEST_STACK_SIZE 1024
	if( (ms = createMemoryStack(0, TEST_STACK_SIZE)) )
	{
		int lpc, *first_i_ptr = 0, *i_ptr;
		memoryStack *curr = 0;
		void *mem;

		/* Make sure the allocation worked properly */
		assert(ms->prev == ms);
		assert(ms->used == 0);
		assert(ms->max == TEST_STACK_SIZE);
		{
			/*
			 * Trample the memory to make sure its all ours,
			 * hopefully something else blows up if its not working
			 * right.
			 */
			memset(ms->data.start, 0xdeadbeef, ms->max);

			ensure_valid_memory(ms);
		}
		/* Try some allocations/deallocations */
		assert(mem = allocStackSpace(ms, 32));
		freeStackSpace(ms, 32);
		assert(mem == allocStackSpace(ms, 32));
		freeStackSpace(ms, 32);
		/* Make sure a single node handles a full size alloc */
		assert(mem = allocStackSpace(ms, TEST_STACK_SIZE));
		assert(ms->prev == ms);
		freeStackSpace(ms, TEST_STACK_SIZE);
		/* Try a cut */
		mem = allocStackSpace(ms, 32);
		allocStackSpace(ms, 32);
		cutMemoryStack(ms, mem);
		assert(mem == allocStackSpace(ms, 32));
		freeStackSpace(ms, 32);
		/* Test bottom/topOfMemoryStack */
		assert(mem == bottomOfMemoryStack(ms));
		assert(mem == topOfMemoryStack(ms, 0));
		allocStackSpace(ms, 32);
		assert(mem == topOfMemoryStack(ms, 32));
		freeStackSpace(ms, 32);
		/* Walk an empty stack */
		assert(walkMemoryStack(ms, &curr,
				       bottomOfMemoryStack(ms),
				       sizeof(int)) == 0);
		curr = 0;
		/* Allocate some memory for walking */
		for( lpc = 0; lpc < 10; lpc++ )
		{
			i_ptr = allocStackSpace(ms, sizeof(int));
			if( lpc == 0 )
				first_i_ptr = i_ptr;
			i_ptr[0] = lpc;
		}
		i_ptr = bottomOfMemoryStack(ms);
		for( lpc = 0; lpc < 10; lpc++ )
		{
			assert(i_ptr = walkMemoryStack(ms,
						       &curr,
						       i_ptr,
						       sizeof(int)));
			assert(i_ptr[0] == lpc);
		}
		cutMemoryStack(ms, mem);
		assert(mem == allocStackSpace(ms,
					      TEST_STACK_SIZE -
					      sizeof(int) -
					      ALIGNMENTOF_VOIDP));
		/* Force an allocation for a new node */
		if( allocStackSpace(ms, TEST_STACK_SIZE / 2) )
		{
			assert(ms->prev != ms);
			freeStackSpace(ms, TEST_STACK_SIZE);
			/* Walk across nodes */
			for( lpc = 0; lpc < 10; lpc++ )
			{
				i_ptr = allocStackSpace(ms, sizeof(int));
				if( lpc == 0 )
					first_i_ptr = i_ptr;
				i_ptr[0] = lpc;
			}
			i_ptr = first_i_ptr;
			curr = 0;
			for( lpc = 0; lpc < 10; lpc++ )
			{
				assert(i_ptr = walkMemoryStack(ms,
							       &curr,
							       i_ptr,
							       sizeof(int)));
				assert(i_ptr[0] == lpc);
			}
		}
		deleteMemoryStack(ms);
	}
#undef TEST_STACK_SIZE
	return( retval );
}

/** memoryStack functions **/

#define ADJUST_SIZE(x) \
	(((x) + (ALIGNMENTOF_VOIDP - 1)) & ~(ALIGNMENTOF_VOIDP - 1))

memoryStack *createMemoryStack(memoryStack *root, size_t max)
{
	memoryStack *retval = 0;

	require(max > 0);
	
	/* Allocate a new node */
	if( (retval = gc_malloc(ADJUST_SIZE(sizeof(memoryStack) + max),
				GC_ALLOC_FIXED)) )
	{
		if( root )
		{
			/* Push the new node on the top of the stack */
			retval->prev = root->prev;
			root->prev = retval;
		}
		else
		{
			/* We're the root, link back to ourself */
			retval->prev = retval;
		}
		/* Setup the rest of the struct */
		retval->used = 0;
		retval->max = max;
#if defined(DEBUG)
		/* Fill the memory with garbage to help debugging */
		memset(retval->data.start, 0xDEADBEEF, max);
#else
		/* Don't care whats in data.start */
#endif
	}
	return( retval );
}

void deleteMemoryStack(memoryStack *ms)
{
	if( ms )
	{
		memoryStack *prev;

		require_valid_memory(ms);
		
		/* Start at the top of the stack */
		prev = ms->prev;
		/* Break the circular link */
		ms->prev = 0;
		/* Walk the list and free all the nodes */
		ms = prev;
		while( ms )
		{
			require_valid_memory(ms);
			
			prev = ms->prev;
			gc_free(ms);
			ms = prev;
		}
	}
}

#define STACK_PAD 128

void *allocStackSpace_slow(memoryStack *root_ms, size_t size)
{
	void *retval = 0;
	memoryStack *ms;

	require(root_ms != 0);
	require_valid_memory(root_ms);

	ms = root_ms->prev;
	size = ADJUST_SIZE(size);
	if( (ms->used + size) <= ms->max )
	{
		/*
		 * Shouldn't really get here since the inline should take
		 * care of this case...
		 */
		retval = &ms->data.start[ms->used];
		ms->used += size;
	}
	else
	{
		memoryStack *newStack;
		int allocSize;

		/*
		 * Need a new node, allocate a node with the same max as the
		 * current top or bigger if necessary to hold the request.
		 */
		allocSize = MAX(ms->max, size + STACK_PAD);
		if( (newStack = createMemoryStack(root_ms, allocSize)) )
		{
			newStack->used += size;
			retval = newStack->data.start;

			ensure(root_ms->prev == newStack);
		}
	}

	ensure_valid_memory(ms);
	
	return( retval );
}

void freeStackSpace_slow(memoryStack *root_ms, size_t size)
{
	memoryStack *ms, *prev;
	
	require(root_ms != 0);
	require_valid_memory(root_ms);
	
	ms = root_ms->prev;
	size = ADJUST_SIZE(size);
	while( size && (size >= ms->used) )
	{
		/*
		 * Make sure we're not trying to free more than was ever
		 * allocated.
		 */
		require(ms->used > 0);
		
		/*
		 * We're gonna be able to free this node, assuming its not
		 * the root.
		 */
		size -= ms->used;
		ms->used = 0;
		prev = ms->prev;
		if( ms != root_ms )
			gc_free(ms);
		ms = prev;

		ensure_valid_memory(ms);
	}
	root_ms->prev = ms;
	ms->used -= size;
#if defined(DEBUG)
	/* Poison the memory */
	memset(&ms->data.start[ms->used], 0xDEADBEEF, size);
#else
	/* Don't care */
#endif
}

void cutMemoryStack(memoryStack *root_ms, void *addr)
{
	memoryStack *ms, *prev;

	require(root_ms != 0);
	require_valid_memory(root_ms);

	ms = root_ms->prev;
	while( ((char *)addr < ms->data.start) ||
	       ((ms->data.start + ms->used) < (char *)addr) )
	{
		/*
		 * If we're at the root node and the address is outside of its
		 * range than its completely outside the stack.
		 */
		require(ms != root_ms);
		
		/* The cut address is outside this node, drop it */
		ms->used = 0;
		prev = ms->prev;
		if( ms != root_ms )
			gc_free(ms);
		ms = prev;

		ensure_valid_memory(ms);
	}
	root_ms->prev = ms;
	ms->used -= ((ms->data.start + ms->used) - (char *)addr);
}

void *walkMemoryStack(memoryStack *root_ms, memoryStack **ms_ptr,
		      void *prev, size_t size)
{
	void *retval = 0;
	memoryStack *ms;

	require(root_ms != 0);
	require_valid_memory(root_ms);
	require(ms_ptr != 0);
	require(prev != 0);
	
	if( *ms_ptr == 0 )
	{
		ms = root_ms->prev;
		while( ((char *)prev < ms->data.start) ||
		       ((ms->data.start + ms->used) < (char *)prev) )
		{
			/*
			 * If the address is outside the range of the root node
			 * then its not within the range of the stack.
			 */
			require(ms != root_ms);
			
			ms = ms->prev;
		}
		/* Found it, make sure theres something there. */
		if( (ms->data.start + ms->used) > (char *)prev )
		{
			*ms_ptr = ms;
			retval = prev;
		}
	}
	else
	{
		void *new_obj;

		require(*ms_ptr != 0);
		require_valid_memory(*ms_ptr);
		
		ms = *ms_ptr;
		size = ADJUST_SIZE(size);
		new_obj = ((char *)prev) + size;
		if( ((char *)new_obj > ms->data.start) &&
		    ((char *)new_obj < (ms->data.start + ms->used)) )
		{
			/* Still within range of this node */
			retval = new_obj;
		}
		else
		{
			memoryStack *curr;
			
			curr = ms;
			/*
			 * Loop around to the node immediately after the
			 * current one...
			 */
			ms = curr->prev;
			while( ms->prev != curr )
			{
				ms = ms->prev;
			}
			if( ms == root_ms )
			{
				/* At the top of the stack...  Just return 0 */
			}
			else
			{
				*ms_ptr = ms;
				/*
				 * The next object in the stack will be at the
				 * start of memory in this node so there's
				 * no need to seek into it.
				 */
				retval = ms->data.start;
				
				ensure((*ms_ptr)->prev == curr);
			}
		}
	}
	return( retval );
}
