/*
 * memoryStack.h
 * A generic stack implementation
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#ifndef _memory_stack_h
#define _memory_stack_h

/*
 * The `memory stack' module provides functionality for managing dynamic memory
 * using a stack interface.  Its primarily positioned as infrastructure for
 * pre/pool allocators since it provides a generic interface thats easy to use
 * and its operating model will amortize the cost of some operations, like
 * initialization (big deal...).  For example, if there was a group of objects
 * that were frequently allocated/deallocated you might want to preallocate a
 * pool of them and store them in a linked list.  Unfortunately, this means
 * writing some code thats very similar to the other pool allocators in the
 * system and some allocations might get hit with a delay when the free list
 * runs dry and you have to allocate/initialize a new pool. This is where the
 * stack allocator comes in, basically it relieves you of the need to write
 * most of the pool allocator code.  And, the stack model avoids the need to
 * initialize all the nodes at once, since you basically just push a new node
 * on the stack if you're free list becomes empty.  Propaganda ends here...
 */

#include "config.h"

/*
 * The memoryStack structure is the building block used to track the
 * memory contained in the stack.  So, one or more of these structures linked
 * together in a circular list make up the stack.  Here's an illustration:
 *
 *  +--> root (bottom of stack) --> top -+
 *  |                                    |
 *  +---------- ... middle ... <---------+
 *
 * The root memoryStack node is the bottom of the stack and is the handle that
 * is passed to most of the accompanying functions.  This is convenient since
 * the user has a single canonical pointer for passing to functions and the
 * code can get the top of the stack by just referencing the previous pointer
 * in the node.
 *
 * `prev' - Link to the previous memoryStack node.
 * `used' - The amount of memory currently used in this node.
 * `max' - The maximum amount of memory this node contains.
 * `data.start' - The start of the memory contained in this node.
 */
typedef struct _memoryStack {
	struct _memoryStack *prev;
	size_t used;
	size_t max;
	union {
		char start[1];
		double align_me[1]; /* XXX Will this always work? */
	} data;
} memoryStack;

/*
 * Create a declaration for a statically allocated memoryStack.
 *
 * `NAME' - The name of the variable to declare.
 * `SIZE' - The amount of memory to allocate for the root node.
 */
#define DECLARE_METAL_MEMORY_STACK(NAME, SIZE) \
extern struct _NAME { \
	memoryStack ms; \
	char data[SIZE - 1]; \
} NAME;

/*
 * Actually define a memoryStack and initialize the root node with its
 * memory.
 */
#define METAL_MEMORY_STACK(NAME, SIZE) \
struct _NAME NAME = { \
	{ &NAME.ms, 0, SIZE, { { 0 } } } \
}

/*
 * Create a memoryStack node.
 *
 * `root_ms' - The root node for the stack.  Should be NULL if you want to
 * allocate a root node, otherwise its just for internal use.
 * `max' - The maximum memory size for each node in the stack.
 * returns - The newly allocated memoryStack node.
 */
memoryStack *createMemoryStack(memoryStack *root_ms, size_t max);
/*
 * Delete the memory stack.  Basically, all of the nodes in the list are
 * freed.
 */
void deleteMemoryStack(memoryStack *root_ms);
/*
 * Allocate some stack space.
 *
 * `root_ms' - The root memoryStack node.
 * `size' - The amount of memory, in bytes, to allocate.
 * returns - An aligned pointer to the memory that was allocated.
 */
void *allocStackSpace(memoryStack *root_ms, size_t size);
/* Helper function for the above */
void *allocStackSpace_slow(memoryStack *root_ms, size_t size);
/*
 * Similar to the above functions, except it won't allocate a new node if
 * the current can't satisfy the request.  Handy, if you want to avoid doing
 * a visit whenever possible.
 */
void *allocStackSpace_shallow(memoryStack *root_ms, size_t size);
/*
 * Free the given amount of stack space.
 *
 * `root_ms' - The root memoryStack node.
 * `size' - The amount of memory, in bytes, to pop off the stack.
 */
void freeStackSpace(memoryStack *root_ms, size_t size);
/* Helper function for the above */
void freeStackSpace_slow(memoryStack *root_ms, size_t size);
/*
 * Cut the stack to the given address.  This is handy for freeing large chunks
 * of stack space in one fell swoop.
 *
 * `root_ms' - The root memoryStack node.
 * `addr' - An address returned by a previous allocStackSpace call.
 */
void cutMemoryStack(memoryStack *root_ms, void *addr);
/*
 * Get a pointer to the bottom of the stack.
 *
 * `root_ms' - The root memoryStack node.
 * returns - A pointer to the bottom of the memory stack.  Basically, the
 * pointer returned by the first allocStackSpace call on this stack.
 */
void *bottomOfMemoryStack(memoryStack *root_ms);
/*
 * Return a pointer `offset' bytes from the top of the stack.
 *
 * `root_ms' - The root memoryStack node.
 * `offset' - The number of bytes to count down from the top of the stack.
 * returns - A pointer that is `offset' bytes from the top of the stack.
 */
void *topOfMemoryStack(memoryStack *root_ms, size_t offset);
/*
 * Walk the stack to examine the objects it holds.  For example, if a stack was
 * made up of only integers you could walk the whole stack with this code:
 *
 *   memoryStack *curr = 0;
 *   int *int_ptr;
 *   
 *   int_ptr = (int *)bottomOfMemoryStack(ms);
 *   while( (int_ptr = (int *)walkMemoryStack(ms, &curr,
 *                                            int_ptr, sizeof(int))) )
 *   {
 *      printf("i[%d] = %d\n", lpc, *int_ptr);
 *   }
 *
 * `root_ms' - The root memoryStack node.
 * `ms_ptr' - A pointer to a pointer to the current memoryStack node.  This
 *   should point to a NULL pointer on the first invocation so that the
 *   function can tell you're starting a new walk.  NOTE: This value will
 *   change as you progress through the nodes in the stack as you walk it.
 * `prev' - Pointer to the previously walked object on the stack.
 * `size' - The size of the object pointed to by `prev'.
 * returns - `prev' on the first invocation and the sequence of objects after
 *   `prev' in future invocations.
 */
void *walkMemoryStack(memoryStack *root_ms, memoryStack **ms_ptr,
		      void *prev, size_t size);

/*
 * AddModule: memoryStackModule
 *   Comment: Just used to test the module
 */

#endif
