/*
 * objectStack.h
 * Infrastructure for building potentially large objects.
 *
 * Copyright (c) 2000, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdarg.h>

#include <assert_pp.h>

#include "gtypes.h"
#include "support.h"
#include "gc.h"
#include "mungWallProxy.h"
#include "objectStack.h"

static
objectSlice *appendSlice(objectStack *os)
{
	objectSlice *retval;

	require(os != 0);

	/* Check for already allocated slices */
	if( os->current.end->link.ln_Succ->ln_Succ )
	{
		/* Free slice, clear it and advance the pointer up */
		retval = (objectSlice *)os->current.end->link.ln_Succ;
		retval->size = 0;
		os->current.end = retval;
	}
	else if( (retval = gc_malloc(sizeof(objectSlice) +
				     (os->puddleSize +
				      ((os->puddleSize * 2) / 3)),
				     GC_ALLOC_FIXED)) )
	{
		os->puddleSize += (os->puddleSize * 2) / 3;
		/* Add it to the list of allocated slices */
		lnAddTail(&os->slices, &retval->link);
		retval->buffer = (char *)(retval + 1);
		retval->size = 0;
		retval->max = os->puddleSize;
		os->current.end = retval;
	}

	ensure( !retval || (retval == os->current.end) );
	
	return( retval );
}

static
void collectSlices(objectSlice *slice)
{
	objectSlice *succ;
	
	while( slice->link.ln_Succ )
	{
		succ = (objectSlice *)slice->link.ln_Succ;
		lnRemove(&slice->link);
		gc_free(slice);
		slice = succ;
	}
}

objectStack *createObjectStack(int puddleSize)
{
	objectStack *retval;

	if( (retval = gc_malloc(sizeof(objectStack) +
				sizeof(objectSlice) +
				puddleSize,
				GC_ALLOC_FIXED)) )
	{
		objectSlice *slice = (objectSlice *)(retval + 1);
		char *buffer = (char *)(slice + 1);

		lnNewList(&retval->slices);
		retval->current.begin = slice;
		retval->current.end = slice;
		retval->current.offset = 0;
		retval->current.total = 0;
		retval->puddleSize = puddleSize;
		lnAddTail(&retval->slices, &slice->link);
		slice->buffer = buffer;
		slice->size = 0;
		slice->max = puddleSize;
	}
	return( retval );
}

void deleteObjectStack(objectStack *os)
{
	if( os )
	{
		collectSlices((objectSlice *)os->slices.lh_Head->ln_Succ);
		gc_free(os);
	}
}

void *objectStackFrame(objectStack *os)
{
	void *retval;

	retval = &os->current.end->buffer[os->current.end->size];
	return( retval );
}

void cutObjectStack(objectStack *os, void *frame)
{
	objectSlice *slice;

	slice = os->current.end;
	while( !(((char *)frame >= slice->buffer) &&
		 ((char *)frame < (slice->buffer + slice->max))) )
	{
		slice = (objectSlice *)slice->link.ln_Pred;
	}
	os->current.begin = slice;
	os->current.end = slice;
	slice->size = (char *)frame - slice->buffer;
	os->current.offset = slice->size;
	os->current.total = 0;
}

int pushStackObject(objectStack *os, void *obj, int length)
{
	objectSlice *slice;
	int retval = 1;

	require(os != 0);
	require(obj != 0);
	require(length >= 0);

	if( (os->current.end->size + length) < os->current.end->max )
	{
		slice = os->current.end;
		memcpy(&slice->buffer[slice->size], obj, length);
		slice->size += length;
		os->current.total += length;
	}
	else
	{
		int offset = 0, segmentSize;

		slice = os->current.end;
		while( length && retval )
		{
			segmentSize = MIN(length, slice->max - slice->size);
			memcpy(&slice->buffer[slice->size],
			       &((char *)obj)[offset],
			       segmentSize);
			offset += segmentSize;
			slice->size += segmentSize;
			length -= segmentSize;
			if( length && !(slice = appendSlice(os)) )
			{
				retval = 0;
			}
		}
	}
	return( retval );
}

int pushStackObjects(objectStack *os, int tag, ...)
{
	char tempBuffer[32];
	int len, retval = 1;
	void *p_value;
	char *str_ptr;
	char c_value;
	int i_value;
	va_list args;

	va_start(args, tag);
	while( (tag != OSA_TAG_DONE) && retval )
	{
		switch( tag )
		{
		case OSA_Char:
			c_value = va_arg(args, int);
			pushStackObject(os, &c_value, 1);
			break;
		case OSA_Bytes:
			str_ptr = va_arg(args, char *);
			len = va_arg(args, int);
			pushStackObject(os, str_ptr, len);
			break;
		case OSA_String:
			str_ptr = va_arg(args, char *);
			len = strlen(str_ptr);
			pushStackObject(os, str_ptr, len);
			break;
		case OSA_IntegerString:
			i_value = va_arg(args, int);
			sprintf(tempBuffer, "%d", i_value);
			pushStackObject(os, tempBuffer, strlen(tempBuffer));
			break;
		case OSA_PointerString:
			p_value = va_arg(args, void *);
			sprintf(tempBuffer, "%p", p_value);
			pushStackObject(os, tempBuffer, strlen(tempBuffer));
			break;
		default:
			assert(0);
			break;
		}
		tag = va_arg(args, int);
	}
	va_end(args);
	return( retval );
}

void *topStackObject(objectStack *os, int *length)
{
	objectSlice *slice;
	void *retval = 0;

	*length = os->current.total;
	os->current.total = 0;
	if( os->current.begin == os->current.end )
	{
		slice = os->current.begin;
		retval = &slice->buffer[os->current.offset];
		os->current.offset += *length;
	}
	else if( (retval = gc_malloc(*length, GC_ALLOC_FIXED)) )
	{
		int offset = 0, bytesLeft = *length;
		
		slice = os->current.begin;
		while( bytesLeft > 0 )
		{
			memcpy(&((char *)retval)[offset],
			       slice->buffer,
			       MIN(slice->size, slice->max));
			bytesLeft -= slice->max;
			offset += slice->max;
			slice = (objectSlice *)slice->link.ln_Succ;
		}
		os->current.end = os->current.begin;
		os->current.begin->size = os->current.offset;

		ensure_valid_memory(retval);
	}
	return( retval );
}

void popStackObject(objectStack *os, void *obj, int length)
{
	objectSlice *slice;

	slice = os->current.end;
	while( slice->link.ln_Pred &&
	       !(((char *)obj >= slice->buffer) &&
		 ((char *)obj < (slice->buffer + slice->max))) )
	{
		slice = (objectSlice *)slice->link.ln_Pred;
	}
	if( slice->link.ln_Pred )
	{
		os->current.begin = slice;
		os->current.end = slice;
		slice->size = (char *)obj - slice->buffer;
	}
	else
	{
		gc_free(obj);
	}
}
