/*
 * sectionFile.c
 * Routines for reading/writing a formatted text file
 *
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Contributed by the Flux Research Group, Department of Computer Science,
 * University of Utah, http://www.cs.utah.edu/flux/
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <assert_pp.h>

#include "jmalloc.h"
#include "stringSupport.h"
#include "mungWallProxy.h"
#include "module.h"

#include "sectionFile.h"
#include "sectionFile-inline.h"

static int testSectionFileModule(void);

module_t sectionFileModule = {
	"sectionFile",
	0,
	0,
	0,
	testSectionFileModule
};

/*
 * Produce a hash value within table_size, for string `name'
 */
static int hashName(const char *name, int table_size)
{
	int h;

	require(name != 0);
	require(table_size > 0);
	
	for( h = 0; *name; name++ )
		h = (64 * h + tolower(*name)) % table_size;

	ensure(h >= 0);
	ensure(h < table_size);
	
	return( h );
}

sectionFile *createSectionFile()
{
	sectionFile *retval;

	/* Allocate the structure and do any initialization */
	if( (retval = (sectionFile *)KMALLOC(sizeof(sectionFile))) )
	{
		int lpc;

		retval->lock = 0;
		retval->users = 1;
		retval->readers = 0;
		retval->filename = 0;
		retval->time = 0;
		retval->ordered_sections = 0;
		retval->last_section = &retval->ordered_sections;
		retval->delayed_writes = 0;
		retval->last_delayed = &retval->delayed_writes;
		for( lpc = 0; lpc < SECTION_FILE_HASH_SIZE; lpc++ )
		{
			retval->sections[lpc] = 0;
		}
	}
	return( retval );
}

sectionFile *grabSectionFile(sectionFile *sf)
{
	sectionFile *retval = sf;
	int iLockRoot;

	lockMutex(sf);
	{
		require(sf->users > 0);
		
		sf->users++;
	}
	unlockMutex(sf);
	return( retval );
}

void deleteSectionFile(sectionFile *sf)
{
	if( sf )
	{
		int iLockRoot;
		int done = 0;
		
		lockMutex(sf);
		{
			require(sf->users > 0);
			
			sf->users--;
			done = (sf->users == 0);
		}
		unlockMutex(sf);
		if( done )
		{
			sectionFileData *sfd, *next;
			int lpc;
			
			/*
			 * Walk through the section and have their handlers
			 * delete the section data objects.
			 */
			for( lpc = 0; lpc < SECTION_FILE_HASH_SIZE; lpc++ )
			{
				sfd = sf->sections[lpc];
				while( sfd )
				{
					next = sfd->next;
					deleteFileSection(sfd);
					sfd = next;
				}
			}
			KFREE(sf);
		}
	}
}

void setSectionFileName(sectionFile *sf, char *name)
{
	sf->filename = name;
}

/*
 * This is used to actually add a section to a file, we don't immediately
 * use this because the write might have to be delayed until a walk is
 * completed.
 */
static
void realAddSectionToFile(sectionFile *sf, sectionFileData *sfd)
{
	int hash;
	
	hash = hashName(sfd->name, SECTION_FILE_HASH_SIZE);
	sfd->next = sf->sections[hash];
	/* The section is new to this file so we set the dirty bit */
	sfd->flags |= SFDF_DIRTY;
	sf->sections[hash] = sfd;
	*sf->last_section = sfd;
	sf->last_section = &sfd->order;
	sfd->order = 0;
}

void addSectionToFile(sectionFile *sf, sectionFileData *sfd)
{
	int iLockRoot;

	lockMutex(sf);
	if( sf->readers )
	{
		/*
		 * Someone is walking over the files, delay this addition until
		 * they are finished
		 */
		*sf->last_delayed = sfd;
		sf->last_delayed = &sfd->order;
		sfd->order = 0;
	}
	else
	{
		/* Add only if there are no readers */
		realAddSectionToFile(sf, sfd);
	}
	unlockMutex(sf);
}

void remSectionFromFile(sectionFile *sf, sectionFileData *sfd)
{
	/*
	 * We don't remove the section from the lists until we sync with the
	 * file.
	 */
	sfd->flags |= SFDF_REMOVED;
}

sectionFileData *findSectionInFile(sectionFile *sf,
				   sectionType *type,
				   char *name)
{
	sectionFileData *retval = 0, *sfd;
	int iLockRoot, hash;

	lockMutex(sf);
	hash = hashName(name, SECTION_FILE_HASH_SIZE);
	sfd = sf->sections[hash];
	while( sfd && !retval )
	{
		/* Match on the type if one is given */
		if( (!type || (type == sfd->type)) &&
		    !strcmp(name, sfd->name) )
		{
			retval = sfd;
		}
		sfd = sfd->next;
	}
	/*
	 * If it wasn't found, it might've been added while someone was walking
	 */
	if( !retval && sf->delayed_writes )
	{
		sfd = sf->delayed_writes;
		while( sfd && !retval )
		{
			if( (!type || (type == sfd->type)) &&
			    !strcmp(name, sfd->name) )
			{
				retval = sfd;
			}
			sfd = sfd->order;
		}
	}
	unlockMutex(sf);
	return( retval );
}

void printSectionFile(sectionFile *sf)
{
	sectionFileData *sfd;
	int lpc;
	
	printf("Ordered first %p\n", sf);
	sfd = sf->ordered_sections;
	while( sfd )
	{
		printf("section %s %s\n", sfd->type->name, sfd->name);
		sfd = sfd->order;
	}
	printf("hash\n");
	for( lpc = 0; lpc < SECTION_FILE_HASH_SIZE; lpc++ )
	{
		sfd = sf->sections[lpc];
		while( sfd )
		{
			printf("%d section %s %s\n", lpc, sfd->type->name, sfd->name);
			sfd = sfd->next;
		}
	}
}

sectionFileData *getSectionParent(sectionFile *sf, sectionFileData *sfd)
{
	sectionFileData *retval = 0;
	char *parentName;
	int nameLen;
	int iLockRoot;

	if( sfd )
	{
		lockMutex(sf);
		nameLen = strlen(sfd->name);
		/*
		 * Compute the section's parent by stripping off the last
		 * `:' followed by a name.
		 */
		if( (parentName = alloca(nameLen + 1)) )
		{
			char *pos;
			
			for( pos = sfd->name + nameLen - 1;
			     (pos >= sfd->name) && (*pos != ':');
			     pos-- );
			if( pos >= sfd->name )
			{
				strncpy(parentName, sfd->name,
					pos - sfd->name);
				parentName[pos - sfd->name] = 0;
				retval = findSectionInFile(sf, 0, parentName);
			}
		}
		unlockMutex(sf);
	}
	return( retval );
}

char *childSectionName(sectionFileData *sfd)
{
	char *retval;

	retval = sfd->name + strlen(sfd->name);
	while( (retval > sfd->name) && (*retval != ':') )
		retval--;
	if( *retval == ':' )
		retval++;
	return( retval );
}

int walkFileSections(sectionFile *sf)
{
	sectionFileData *sfd;
	int retval = 1;
	int iLockRoot;

	if( !sf )
		return( 0 );

	require_valid_memory(sf);
	
	/* Increment readers to make sure others don't modify the lists */
	lockMutex(sf);
	sf->readers++;
	unlockMutex(sf);
	/* Walk the sections in order */
	sfd = sf->ordered_sections;
	while( sfd && retval )
	{
		require_valid_memory(sfd);
		require_valid_memory(sfd->name);

		if( !(sfd->flags & SFDF_WALKED) )
		{
			retval = sfWalk(sfd, sf);
			sfd->flags |= SFDF_WALKED;
		}
		sfd = sfd->order;

		ensure_valid_memory(sfd);
	}
	/*
	 * Decrement readers count, and add any delayed sections if noone
	 * else is walking.
	 */
	lockMutex(sf);
	sf->readers--;
	if( !sf->readers && sf->delayed_writes )
	{
		sectionFileData *next;

		sfd = sf->delayed_writes;
		while( sfd )
		{
			next = sfd->order;
			realAddSectionToFile(sf, sfd);
			sfd = next;
		}
		sf->delayed_writes = 0;
		sf->last_delayed = &sf->delayed_writes;
	}
	unlockMutex(sf);
	return( retval );
}

int walkFileSections2(sectionFile *sf,
		      sectionType *st,
		      int (*handler)(void *arg,
				     sectionFile *sf,
				     sectionFileData *sfd),
		      void *arg)
{
	sectionFileData *sfd;
	int iLockRoot, retval = 1;

	if( !sf )
		return( 0 );
	lockMutex(sf);
	sf->readers++;
	unlockMutex(sf);
	/* Walk over the ordered links */
	sfd = sf->ordered_sections;
	while( sfd && retval )
	{
		if( (!st || (st == sfd->type)) && !(sfd->flags & SFDF_WALKED) )
		{
			retval = handler(arg, sf, sfd);
			sfd->flags |= SFDF_WALKED;
		}
		sfd = sfd->order;
	}
	lockMutex(sf);
	sf->readers--;
	if( !sf->readers && sf->delayed_writes )
	{
		sectionFileData *next;

		sfd = sf->delayed_writes;
		while( sfd )
		{
			next = sfd->order;
			realAddSectionToFile(sf, sfd);
			sfd = next;
		}
		sf->delayed_writes = 0;
		sf->last_delayed = &sf->delayed_writes;
	}
	unlockMutex(sf);
	return( retval );
}

sectionFileData *createFileSection(char *section_type, char *section_name)
{
	sectionFileData *retval = 0;
	char *new_name = 0;
	sectionType *st;
	
	/* Get the section type and ask it to create a section data object */
	if( (st = findSectionType(section_type)) &&
	    (new_name = (char *)KMALLOC(strlen(section_name) + 1)) &&
	    (retval = (sectionFileData *)KMALLOC(st->dataSize)) )
	{
		retval->name = new_name;
		strcpy(new_name, section_name);
		retval->type = st;
		retval->flags = 0;
		/* Ask the type to do any initialization */
		if( !sfInit(retval) )
		{
			deleteFileSection(retval);
			retval = 0;
		}
		else
		{
			ensure_valid_memory(retval);
			ensure_valid_memory(new_name);
		}
	}
	if( !retval )
	{
		KFREE(new_name);
	}
	return( retval );
}

void deleteFileSection(sectionFileData *sfd)
{
	if( sfd )
	{
		/* Ask the handler to clean up the object */
		sfKill(sfd);
		KFREE(sfd->name);
		KFREE(sfd);
	}
}

int testSectionFileModule(void)
{
#if 0
	aclSectionType teamACL;
	tagSectionType tst;
	sectionFile *sf;
	int retval = 0;

	installSectionType(&tst.type,
			   STA_Name, "test-type",
			   STA_Handler, tagArraySectionHandler,
			   STA_Size, sizeof(tagSectionData),
			   TSA_DefaultArray, 0,
			   STA_TAG_DONE);
	installSectionType(&teamACL.type.type,
			   STA_Parent, &tagSectionParent.type,
			   STA_Name, "team-acl",
			   STA_Handler, aclSectionHandler,
			   STA_Walker, aclSectionWalker,
			   STA_Size, sizeof(aclSectionData),
			   TSA_DefaultArray, 0,
			   ASA_Permissions, teamPerms,
			   STA_TAG_DONE);
	if( (sf = createSectionFile()) )
	{
		setSectionFileName(sf, "test.config");
		if( syncSectionFile(sf) )
		{
			sectionFileData *sfd;

			walkFileSections(sf);
			if( (sfd = findSectionInFile(sf, &tst.type,
						     "test-type")) )
			{
				tagSectionData *tsd =
					LOCAL_SECTION_DATA(tagSectionData *,
							   &tst,
							   sfd);
				int lpc;

				for( lpc = 0;
				     lpc < tsd->taglist->tags.count;
				     lpc++ )
				{
					printTagItem(&tsd->taglist->tags.data[lpc]);
					printf("\n");
				}
			}
			if( (sfd = findSectionInFile(sf, &tst.type,
						     "test-type:sub")) )
			{
				tagSectionData *tsd =
					LOCAL_SECTION_DATA(tagSectionData *,
							   &tst,
							   sfd);
				int lpc;

				for( lpc = 0;
				     lpc < tsd->taglist->tags.count;
				     lpc++ )
				{
					printTagItem(&tsd->taglist->tags.data[lpc]);
					printf("\n");
				}
			}
			if( (sfd = findSectionInFile(sf,
						     &teamACL.type.type,
						     "kernel-team-acl")) )
			{
				aclSectionData *asd =
					LOCAL_SECTION_DATA(aclSectionData *,
							   &teamACL,
							   sfd);

				aclPrint(asd->acl);
				fprintf(stderr, "\n");
			}
			if( aclFindPrincipal(APA_Formatted, "stack@anaka", -1,
					     APA_TAG_DONE) )
			{
				printf("found principal\n");
			}
		}
		deleteSectionFile(sf);
	}
	return( retval );
#else
	return( 1 );
#endif
}
