/*
 * sectionFile.h
 * Routines for reading/writing a formatted text file
 *
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Contributed by the Flux Research Group, Department of Computer Science,
 * University of Utah, http://www.cs.utah.edu/flux/
 */

#ifndef __sectionfile_h
#define __sectionfile_h

/*
 * The section file code reads and writes files that can be used for storing
 * interesting information.  The files themselves are just plain text thats
 * formatted with a number of directives to drive the interpretation of the
 * rest of the data.  The current set of recognized directives are:
 *
 *   %include "<filename>" - Include another file
 *
 *   %begin <type> <name> - Begin a section with the given type and name
 *     <attribute1 name> <attribute1 value> - Specify a value for an attribute
 *     ...
 *     <attributeN name> <attributeN value>
 *   %end - End a section
 *   # - Starts a comment that goes until the end of the line
 *
 *   %define <variable name> <variable value> - Define a variable with the
 *     given value
 *   %(<variable name>) - Replace with the value of the given variable
 *
 *   %if <...> (just `defined', `==', and `!=' for operators)
 *   %else
 *   %endif
 *
 * The code tries to keep the file pristine so the user can edit it without
 * things moving around and disappearing in future edits.
 *
 * The internals of the system works by having the program define a set of
 * types and handlers that can translate between the text names and values in
 * the file and the internal data structures.  The section file takes care
 * of the majority of the work by parsing the outer structure of the file
 * as well as providing functions for handling tasks common to all handlers.
 */

#include <time.h>
#include <stdio.h>
#include <stdarg.h>

#include "gtypes.h"
#include "locks.h"

/* Forward decls */
struct _sectionType;
struct _sectionFile;
struct _sectionFileData;

/*
 * The sectionTypeFunctions structure is used to store the set of functions
 * belonging to a specific section type.
 *
 * `install' - This is called when the type is first installed.
 * `uninstall' - This is called when the type is uninstalled.
 * `init' - This is called to initialize a new sectionFileData structure.
 * `kill' - This is called when a sectionFileData structure is being deleted.
 * `cache' - This is called to read and internalize data from the section file.
 * `flush' - This is called to write data to the section file.
 * `walk' - This is called by the walkSections function to act on the data
 *   in the section.
 */
typedef struct sectionTypeFunctions {
	int (*install)(struct _sectionType *st, int tag, va_list *args);
	void (*uninstall)(struct _sectionType *st);
	int (*init)(struct _sectionType *st, struct _sectionFileData *sfd);
	void (*kill)(struct _sectionType *st, struct _sectionFileData *sfd);
	int (*cache)(struct _sectionType *st, struct _sectionFileData *sfd,
		     void *ps, char *line, int len);
	int (*flush)(struct _sectionType *st, struct _sectionFileData *sfd,
		     void *ps, char *line, int len, FILE *out_file);
	int (*walk)(struct _sectionType *st, struct _sectionFileData *sfd,
		    struct _sectionFile *sf);
} sectionTypeFunctions;

/*
 * A sectionType structure is used to describe a specific type of section
 * that can exist in a file.
 *
 * `next' - A link to the next type in the hash table.
 * `parent' - The parent type.
 * `stf' - The set of section type functions.
 * `offset' - The offset in the sectionFileData structure that has our types
 *   local data.
 * `dataSize' - The amount of memory to allocate for a sectionFileData object
 *   of this type.
 * `name' - The name of this type.
 */
typedef struct _sectionType {
	struct _sectionType *next;
	struct _sectionType *parent;
	sectionTypeFunctions *stf;
	int offset;
	int dataSize;
	char *name;
} sectionType;

/*
 * The sectionFileData structure is used to represent sections present in
 * the file.  This is actually just a header, the real contents of the section
 * immediately follow this structure, the actual size of the structure and
 * the offsets for the local data items is determined by the sectionType
 * structure.
 *
 * `next' - The link to the next sectionFileData in the hash table.
 * `order' - An ordered link so that sections can be walked in the order
 *   that they appeared in the file.
 * `type' - The type of this sectionFileData object.
 * `flags' - Bitmask for storing flags.
 * `name' - The name of this sections.
 */

/* Bit numbers for sectionFileData flags */
enum {
	SFDB_DIRTY,	/* Theres new data in the section */
	SFDB_REMOVED,	/*
			 * The section needs to be removed from the
			 * file
			 */
	SFDB_WALKED,	/* This section has been walked */
};

/* Flag bits for sectionFileData flags */
enum {
	SFDF_DIRTY = (1L << SFDB_DIRTY),
	SFDF_REMOVED = (1L << SFDB_REMOVED),
	SFDF_WALKED = (1L << SFDB_WALKED),
};

typedef struct _sectionFileData {
	struct _sectionFileData *next;
	struct _sectionFileData *order;
	sectionType *type;
	unsigned long flags;
	char *name;
} sectionFileData;

/* A macro for getting the local section data from a sectionFileData object */
#define LOCAL_SECTION_DATA(cast_to, st, sfd) \
	((cast_to)(((uintp)sfd) + ((sectionType *)(st))->offset))

/*
 * The sectionFile structure is the root for accessing the contents of a
 * section file.
 *
 * `lock' - Lock for accessing the object, this is only held while modifying
 *   the section lists.  If someone is just walking over the contents then
 *   the readers count is incremented and the lock is dropped.  This should
 *   help prevent deadlocks...
 * `users' - Reference count.
 * `readers' - The number of people walking over the file.
 * `filename' - The full path to the file.
 * `time' - The last time the file was sync'ed.  Used to determine whether to
 *   read the file into the internal form, or to write out the internal form.
 * `ordered_sections' - An ordered list of sections.
 * `last_section' - Pointer to the last order pointer in the ordered_sections
 *   list.
 * `delayed_writes' - Any new sections added to the file while there are
 *   readers are stored here in order.
 * `last_delayed' - Same as last_section, except its for delayed_writes.
 * `sections' - The hash table used to locate sections by their name.
 */
typedef struct _sectionFile {
	iLock *lock;
	int users;
	int readers;
	char *filename;
	time_t time;
	sectionFileData *ordered_sections;
	sectionFileData **last_section;
	sectionFileData *delayed_writes;
	sectionFileData **last_delayed;
#define SECTION_FILE_HASH_SIZE 31
	sectionFileData *sections[SECTION_FILE_HASH_SIZE];
} sectionFile;

/* Function tags for installSectionType */
enum {
	STA_TAG_DONE,
	STA_Parent,		/* (sectionType *) - The parent type */
	STA_Template,		/*
				 * (sectionType *) - A template to model this
				 * type after.
				 */
	STA_Name,		/* (char *) - The name of the section */
	STA_Size,		/* (int) - The size of the local data */
	STA_Functions,		/* (sectionTypeFile *) - Section functions */
	
	STA_Install,		/* (void *) - The install function */
	STA_Uninstall,		/* (void *) - The uninstall function */
	STA_Init,		/* (void *) - The init function */
	STA_Kill,		/* (void *) - The kill function */
	STA_Cache,		/* (void *) - The cache function */
	STA_Flush,		/* (void *) - The flush function */
	STA_Walk,		/* (void *) - The walk function */
	
	STA_USER = 0x80000000
};

/*
 * Create a section file structure
 */
sectionFile *createSectionFile(void);
/*
 * Add a reference to the given section file.
 */
sectionFile *grabSectionFile(sectionFile *sf);
/*
 * Delete the section file structure
 */
void deleteSectionFile(sectionFile *sf);
/*
 * Set the file name of the section file
 */
void setSectionFileName(sectionFile *sf, char *name);
/*
 * Synchronize the section_file structure with the actual file.
 */
int syncSectionFile(sectionFile *sf, int create);
/*
 * Synchronize the section_file structure with the given file.
 */
int syncSectionFileWith(sectionFile *sf, char *filename, int create);
/*
 * Add a section type to the system, the tags are of STA_ form.
 */
void installSectionType(sectionType *st, int tag, ...);
/*
 * Remove a section type from the system.
 */
void uninstallSectionType(sectionType *st);
/*
 * Find a section type
 */
sectionType *findSectionType(char *name);
/*
 * Add a section to a file
 */
void addSectionToFile(sectionFile *sf, sectionFileData *sfd);
/*
 * Schedule a section to be removed from the file
 */
void remSectionFromFile(sectionFile *sf,
			sectionFileData *sfd);
/*
 * Find a section in a file
 */
sectionFileData *findSectionInFile(sectionFile *sf,
				   sectionType *type,
				   char *name);
/*
 * Get the enclosing sectionFileData object for the given object.  This is
 * done based on the naming convention that enclosed section objects have
 * a `:' between their name and the enclosing section's name.
 */
sectionFileData *getSectionParent(sectionFile *sf, sectionFileData *child);
char *childSectionName(sectionFileData *sfd);
/*
 * Create a file section of the given type, with the given name
 */
sectionFileData *createFileSection(char *section_type, char *section_name);
/*
 * Delete a file section data object
 */
void deleteFileSection(sectionFileData *sfd);
/*
 * Call all of the walk functions for each section in the file.
 */
int walkFileSections(sectionFile *sf);
/*
 * Call the given function for each section in the file.
 */
int walkFileSections2(sectionFile *sf,
		      sectionType *st,
		      int (*handler)(void *arg,
				     sectionFile *sf,
				     sectionFileData *sfd),
		      void *arg);
/*
 * Helper function for file_section handlers to parse a line into an identifier
 * and value
 */
int parseSectionLine(void *ps, char **tag, char **value,
		     FILE *out_file);
/*
 * Helper function for file_section handlers to parse a flag string
 * (e.g. "true", "false", etc...) and return the `flags' bitmask with the
 * `flag_bit' set/cleared.
 */
unsigned long parseFlagString(char *value,
			      unsigned long flags,
			      unsigned long flag_bit);
/*
 * Helper function for file_section handlers to get a string which that
 * indicates whether `flag_bit' is set/cleared in `flags'.  The `value'
 * string should be the previous value of the flag, and the string
 * returned tries mimic the value.  For example, if value is "true", but
 * the bit is cleared in the flags, the returned string is "false", instead
 * of "0, "f", etc...
 */
char *makeFlagString(unsigned long flags, unsigned long flag_bit, char *value);

extern char *trueStrings[];
extern char *falseStrings[];

/*
 * AddModule: sectionFileModule
 *   Comment: Make a target for templates to install themselves
 * Implementation: templates
 *   Comment: A target for types to install themselves.
 * Implementation: types
 */

#endif /* __sectionfile_h */
