/*
 * sectionFileParser.c
 * Functions for parsing a section file
 *
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Contributed by the Flux Research Group, Department of Computer Science,
 * University of Utah, http://www.cs.utah.edu/flux/
 */

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/stat.h>

#include "mungWallProxy.h"
#include "stringSupport.h"
#include "sectionFile.h"
#include "sectionFile-inline.h"

/*
 * Internal state for the file parser
 */
typedef struct _sectionParseState {
	struct _sectionParseState *parent;	/* Keep track of nesting */
	char *filename;				/* Current file name */
	sectionFile *section_file;		/* Root section file */
	FILE *in_file;				/* Input File handle */
	FILE *out_file;				/* Output File handle */
	sectionFileData *section;		/* Current section */
	int line;				/* Current line number */
	int column;				/* Current column number */
	char *data;				/* Current line string */
	int len;				/* Length of line */
	char save[2];				/* Storage for damage */
} sectionParseState;

static void restoreLine(sectionParseState *ps);
static int syncFile(sectionParseState *parent, sectionFile *sf,
		    char *filename, int create);

/*
 * Produce a hash value within table_size, for string `name'
 */
static int hashName(const char *name, int table_size)
{
	int h;

	require(name != 0);
	require(table_size > 0);
	
	for( h = 0; *name; name++ )
		h = (64 * h + tolower(*name)) % table_size;
	
	ensure(h >= 0);
	ensure(h < table_size);
	
	return( h );
}

/*
 * Check for valid characters in a string from the file
 */
static inline
int validStringChar(int c)
{
	int retval = 0;
	
	if( isprint(c) &&
	    (c != '#') &&
	    (c != '%') &&
	    (c != '\\') )
		retval = 1;
	return( retval );
}

/*
 * Write out a utf string with the proper escapes, if needed.
 */
static int writeUtfString(sectionParseState *ps, char *ustr)
{
	int ch, len, error = 0;
	char *curr, *end;

	len = strlen(ustr);
	end = ustr + len;
	curr = ustr;
	while( (curr < end) && !error )
	{
		ch = UTF8_GET(curr, end);
		if( ch < 0 )
		{
			error = 1;
		}
		else if( validStringChar(ch) )
		{
			/* Plain character, just write it out */
			fprintf(ps->out_file, "%c", (char)ch);
		}
		else
		{
			/* Write out a UTF8 character with escape sequence */
			fprintf(ps->out_file, "\\u%04x", ch);
		}
	}
	return( !error );
}

enum {
	SFS_INVALID,
	SFS_PLAIN,
	SFS_UTF8,
};

/*
 * Figure out what kind of string has come from a file.  If it has UTF8
 * escapes we'll need to translate those.
 */
static int stringType(char *str)
{
	int retval = SFS_PLAIN, utf8char = 0;

	while( *str && (retval != SFS_INVALID) )
	{
		if( *str == '\\' )
		{
			/* Do a thorough check of the escape */
			if( (*(str + 1) == 'u') &&
			    (sscanf(str + 2, "%x", &utf8char) == 1) )
				retval = SFS_UTF8;
			else
				retval = SFS_INVALID;
		}
		str++;
	}
	return( retval );
}

/*
 * Convert a UTF8 escaped string from the file into an internal format.
 */
static char *convertUtfString(sectionParseState *ps, char *pstr, char *cstr)
{
	char *retval = 0;
	int plen, len;

	plen = pstr ? strlen(pstr) : 0;
	len = strlen(cstr);
	/* We're allocating more than is really needed... */
	if( (retval = (char *)KMALLOC(plen + len + 1)) )
	{
		int lpc, utf8char;
		char *curr;

		if( pstr )
		{
			strcpy(retval, pstr);
			curr = retval + plen;
		}
		else
		{
			curr = retval;
		}
		for( lpc = 0; lpc < len; lpc++ )
		{
			switch( cstr[lpc] )
			{
			case '\\':
				utf8char = 0;
				lpc += 2; /* Skip the 'u' */
				if( sscanf(&cstr[lpc],
					   "%4x",
					   &utf8char) == 1 )
				{
					char plain_char;

					/*
					 * XXX This isn't a proper check for a
					 * UTF char.  We just check if its can
					 * be held within 8 bits and write it
					 * directly to the string, otherwise we
					 * store it as 16 bits.
					 */
					plain_char = utf8char;
					if( plain_char == utf8char )
					{
						*curr = plain_char;
					}
					else
					{
						*curr = utf8char >> 8;
						curr++;
						*curr = utf8char & 0xff;
					}
					curr++;
				}
				lpc += 3; /* Skip the value */
				break;
			default:
				/* Plain character, just copy it over */
				*curr = cstr[lpc];
				curr++;
				break;
			}
		}
		*curr = 0;
	}

	ensure_valid_memory(retval);
	
	return( retval );
}

/*
 * Do the actual removal of a section from a file, this is called after
 * write sync when a section was scheduled for removal.
 */
static void reallyRemoveSectionFromFile(sectionFile *sf,
					sectionFileData *sfd)
{
	sectionFileData **prev, *curr;
	int hash;

	/* Remove it from the hash table */
	hash = hashName(sfd->name, SECTION_FILE_HASH_SIZE);
	curr = sf->sections[hash];
	prev = &sf->sections[hash];
	while( curr && (curr != sfd) )
	{
		prev = &curr->next;
		curr = curr->next;
	}
	if( curr )
		*prev = sfd->next;
	/* Remove it from the ordered list */
	curr = sf->ordered_sections;
	prev = &sf->ordered_sections;
	while( curr && (curr != sfd) )
	{
		prev = &curr->order;
		curr = curr->order;
	}
	if( curr )
		*prev = sfd->order;
}

/* Parse a directive line */
static int parseDirective(sectionParseState *ps, int curr)
{
	int retval = 1, lpc, directive_len, args_len, line_len;
	char *directive, *args, *save, *line;
	sectionFile *sf;
	FILE *out_file;

	sf = ps->section_file;
	out_file = ps->out_file;
	line_len = ps->len;
	line = ps->data;
	save = ps->save;
	
	/* The line passed in points to the beginning of the directive */
	directive = &line[curr];
	/* Move to the end of the directive */
	for( lpc = curr; (lpc < line_len) && !isspace(line[lpc]); lpc++ );
	directive_len = lpc - curr;
	/* The arguments to the directive directly follow it */
	args = &line[lpc + 1];
	args_len = line_len - (lpc + 1);
	if( !strncmp(directive, "%include", directive_len) )
	{
		int name_start, name_end;

		/* Try and get a quoted filename from the args */
		for( name_start = 0;
		     (name_start < args_len) && (args[name_start] != '\"');
		     name_start++ );
		for( name_end = name_start + 1;
		     (name_end < args_len) && (args[name_end] != '\"');
		     name_end++ );
		if( (name_start < args_len) && (name_end < args_len) )
		{
			args[name_end] = 0;
			/* Try and sync the include file */
			if( !syncFile(ps, ps->section_file,
				      args + name_start + 1, 0) )
			{
				fprintf(stderr,
					"Error:%s:%d:%d - Unable to process "
					"file %s.\n",
					ps->filename,
					ps->line,
					ps->column,
					args + name_start + 1);
				retval = 0;
			}
			args[name_end] = '\"';
		}
		else
		{
			fprintf(stderr,
				"Error:%s:%d:%d - %%include directive needs "
				"a quoted filename.\n",
				ps->filename,
				ps->line,
				ps->column);
		}
		if( out_file )
		{
			fwrite(line, 1, line_len, out_file);
		}
	}
	else if( !strncmp(directive, "%begin", directive_len) )
	{
		char *section_type = 0, *section_name = 0;
		char *parent_name = 0;
		sectionType *type;
		int stype;
		
		/* This is the beginning of a section */

		if( ps->section )
			parent_name = ps->section->name;
		/* Parse the args into type and name */
		ps->data = args;
		ps->len = args_len;
		parseSectionLine(ps, &section_type, &section_name, 0);
		ps->data = line;
		ps->len = line_len;
		if( section_name[0] == '\0' )
			section_name = section_type;
		type = findSectionType(section_type);
		switch( (stype = stringType(section_name)) )
		{
		case SFS_INVALID:
			retval = 0;
			fprintf(stderr,
				"Error:%s:%d - Invalid characters in section "
				"name\n",
				ps->filename,
				ps->line);
			break;
		case SFS_PLAIN:
			if( parent_name )
			{
				char *new_name;
				
				if( (new_name = KMALLOC(strlen(parent_name) +
							1 +
							strlen(section_name) +
							1)) )
				{
					sprintf(new_name,
						"%s:%s",
						parent_name,
						section_name);
					section_name = new_name;

					ensure_valid_memory(new_name);
				}
				else
				{
					retval = 0;
				}
			}
			break;
		case SFS_UTF8:
			if( !(section_name = convertUtfString(ps,
							      parent_name,
							      section_name)) )
			{
				retval = 0;
			}
			break;
		}
		if( !retval )
		{
		}
		else if( (ps->section = findSectionInFile(sf,
							  type,
							  section_name)) )
		{
			/* The section already exists */
			if( out_file )
			{
				/*
				 * We'll be writing out the data in a bit
				 * so clear the dirty bit.
				 */
				ps->section->flags &= ~SFDF_DIRTY;
			}
		}
		else if( out_file )
		{
			/*
			 * If we're writing and there's no section then
			 * something is screwy and we should probably cache
			 * the unseen data...  or not
			 */
		}
		else
		{
			/*
			 * The section hasn't been seen before, make a new
			 * one and add it to the structure.
			 */
			if( (ps->section = createFileSection(section_type,
							     section_name)) )
			{
				addSectionToFile(sf, ps->section);
			}
			else
			{
				retval = 0;
			}
		}
		if( (stype == SFS_UTF8) || parent_name )
			KFREE(section_name);
		/* Fix the damage parseSectionLine did */
		restoreLine(ps);
		if( out_file && ps->section &&
		    !(ps->section->flags & SFDF_REMOVED) )
		{
			fwrite(line, 1, line_len, out_file);
		}
	}
	else if( !strncmp(directive, "%end", directive_len) )
	{
		/* This is the end of a section */
		if( out_file && ps->section )
		{
			if( ps->section->flags & SFDF_REMOVED )
			{
				/* The section was pending removal */
				reallyRemoveSectionFromFile(sf,
							    ps->section);
				deleteFileSection(ps->section);
			}
			else
			{
				/* Ask the handler to write any new data out */
				retval = sfFlush(ps->section,
						 ps,
						 0,
						 0,
						 out_file);
				fwrite(line, 1, line_len, out_file);
			}
		}
		/* We're out of the section so clear the current section */
		ps->section = getSectionParent(sf, ps->section);
	}
	else
	{
		fprintf(stderr,
			"Error:%s:%d - Directive `%s' is not valid\n",
			ps->filename,
			ps->line,
			directive);
	}
	if( out_file && ferror(out_file) )
		retval = 0;
	return( retval );
}

/* XXX bleh */
#define MAX_LINE_SIZE 1024

/*
 * Parse the input file and send any output to `out_file' if there is one
 */
static int parseFile(sectionParseState *ps)
{
	char *line, buffer[MAX_LINE_SIZE];
	int retval = 1, line_len;
	FILE *in_file, *out_file;
	sectionFile *sf;

	sf = ps->section_file;
	in_file = ps->in_file;
	out_file = ps->out_file;
	ps->data = buffer;
	
	/* Walk over the input file line by line */
	while( retval && (line = fgets(buffer, MAX_LINE_SIZE, in_file)) )
	{
		int curr;

		ps->line++;
		ps->column = 0;
		ps->len = line_len = strlen(line);
		/* Skip white space */
		for( curr = 0;
		     (curr < line_len) && isspace(line[curr]);
		     curr++ )
			ps->column++;
		if( curr == line_len )
		{
			/* Empty line, dump to output */
			if( out_file )
				fwrite(line, 1, line_len, out_file);
			continue;
		}
		/* Found something, parse it */
		switch( line[curr] )
		{
		case '%':
			/* Directive's start with %, handle it */
			retval = parseDirective(ps, curr);
			curr = line_len;
			break;
		case '#':
			/* A comment, dump the rest of the line to output */
			if( out_file )
			{
				fwrite(line, 1, line_len, out_file);
			}
			curr = line_len;
			break;
		default:
			if( ps->section )
			{
				if( ps->section->flags & SFDF_REMOVED )
					continue;
				/*
				 * We're in a section, ask the handler to
				 * process it
				 */
				if( !line_len || (line[line_len - 1] != '\n') )
				{
					retval = 0;
				}
				else if( out_file )
				{
					/* Bring the file up to date */
					retval = sfFlush(ps->section,
							 ps,
							 line,
							 line_len,
							 out_file);
				}
				else
				{
					/* Read new data in */
					retval = sfCache(ps->section,
							 ps,
							 line,
							 line_len);
				}
			}
			else
			{
				fprintf(stderr,
					"Error:%s:%d:%d - Text outside of "
					"section\n",
					ps->filename,
					ps->line,
					ps->column);
				if( out_file )
				{
					/*
					 * Be nice and output it, but put it in
					 * a comment
					 */
					fprintf(out_file, "# %s", line);
				}
			}
			curr = line_len;
			break;
		}
	}
	/* Check for errors */
	if( ferror(in_file) || (out_file && ferror(out_file)) )
		retval = 0;
	return( retval );
}

/*
 * Check for any new sections added to the section_file and output them to
 * the file.
 */
static int writeNewSections(sectionParseState *ps)
{
	sectionFileData *sfd;
	sectionFile *sf;
	int retval = 1;
	FILE *out_file;

	sf = ps->section_file;
	out_file = ps->out_file;
	
	/* Walk over the hash table links */
	sfd = sf->ordered_sections;
	while( sfd && retval )
	{
		if( sfd->flags & SFDF_DIRTY )
		{
			/*
			 * This section wasn't touched while sync'ing
			 * it must be new to the file.
			 */
			if( sfd->name[0] )
			{
				fprintf(out_file,
					"\n%%begin %s ",
					sfd->type->name);
				retval = writeUtfString(ps, sfd->name);
				fprintf(out_file, "\n");
			}
			else
			{
				fprintf(out_file,
					"\n%%begin %s\n",
					sfd->type->name);
			}
			retval = sfFlush(sfd, ps, 0, 0, out_file);
			fprintf(out_file, "%%end\n");
			sfd->flags &= ~SFDF_DIRTY;
		}
		sfd = sfd->order;
	}
	if( ferror(out_file) )
		retval = 0;
	return( retval );
}

/*
 * Write out the internal data to the file
 */
static int writeFile(sectionParseState *ps)
{
	static char *name_template = "sf.XXXXXX";
	int retval = 0, temp_fd = -1;
	FILE *in_file, *out_file;
	sectionFile *sf;
	char *temp_name = 0;

	sf = ps->section_file;
	in_file = ps->in_file;
	
	/* Create a temporary file for the output */
	if( (temp_name = (char *)KMALLOC(strlen(name_template) + 1)) &&
	    strcpy(temp_name, name_template) &&
	    ((temp_fd = mkstemp(temp_name)) >= 0) &&
	    (out_file = fdopen(temp_fd, "w+")) )
	{
		ps->out_file = out_file;
		/* Parse the old file and update with current data */
		if( (retval = parseFile(ps)) )
		{
			if( !ps->parent )
			{
				/*
				 * Write out any sections that don't have peers
				 * in the input file
				 */
				retval = writeNewSections(ps);
			}
		}
		else
		{
			fprintf(stderr,
				"Error: Unable to sync file %s\n",
				ps->filename);
		}
		fclose(out_file);
		/*
		 * If there weren't any errors we rename the temporary file
		 * so that it overwrites the input file.  Otherwise, we remove
		 * the bad output file, and leave the input file intact.
		 */
		if( retval )
		{
			rename(temp_name, ps->filename);
		}
		else
		{
			remove(temp_name);
		}
	}
	else
	{
		fprintf(stderr,
			"Error: Unable to create temporary file for "
			"rewriting %s.\n",
			ps->filename);
		if( temp_fd >= 0 )
		{
			remove(temp_name);
			close(temp_fd);
		}
	}
	KFREE(temp_name);
	return( retval );
}

/*
 * Internal file sync function.  This allows for included files to be processed
 * and their data contained in the root section file.
 */
static int syncFile(sectionParseState *parent, sectionFile *sf,
		    char *filename, int create)
{
	struct stat file_stat;
	time_t file_time = 0;
	int retval = 0;
	FILE *file;

	/* Make sure theres a file to sync with */
	if( !filename || !filename[0] )
		return( 0 );
	/* Get the file's time stamp so we know whether its new or old */
	if( !stat(filename, &file_stat) )
	{
		file_time = file_stat.st_mtime;
	}
	else if( create )
	{
		/* Make an empty file if it doesn't exist */
		if( (file = fopen(filename, "w+")) )
		{
			fclose(file);
		}
	}
	if( (file = fopen(filename, "r")) )
	{
		sectionParseState ps;

		/* Initialize the parser state */
		ps.parent = parent;
		ps.filename = filename;
		ps.section_file = sf;
		ps.in_file = file;
		ps.out_file = 0;
		ps.section = 0;
		ps.line = 0;
		ps.column = 0;
		ps.data = 0;
		ps.len = 0;
		if( sf->time <= file_time )
		{
			/* The file is newer, read it in */
			if( (retval = parseFile(&ps)) )
			{
				if( !parent )
					sf->time = time(NULL);
			}
		}
		else
		{
			/* The file is older, write out our data */
			if( (retval = writeFile(&ps)) )
			{
				if( !parent )
					sf->time = time(NULL);
			}
		}
		fclose(file);
	}
	return( retval );
}

/*
 * States for parsing a section line made up of an identifier followed by
 * white space, and a data string
 */
enum {
	SS_TAG,		/* Getting the tag */
	SS_GUTTER,	/* White space between tag and value */
	SS_VALUE,	/* Value to the end of the line, minus ws, comments */
};

int parseSectionLine(void *ps, char **tag, char **value, FILE *out_file)
{
	int state, tag_start = -1, tag_end = -1;
	int value_start = -1, value_end = -1;
	int retval = 1, curr, line_len;
	sectionParseState *sps = ps;
	char *line, *save;

	line = sps->data;
	line_len = sps->len;
	save = sps->save;
	
	/* We start out looking for the tag */
	state = SS_TAG;
	/* Walk through the line */
	for( curr = 0; curr < line_len; curr++ )
	{
		switch(line[curr])
		{
		case '#':
			/* Its a comment */
			if( out_file )
			{
				/* Write out value */
				fwrite(*value, 1, strlen(*value), out_file);
				if( value_end > 0 )
				{
					/*
					 * Reset the trampled character at the
					 * end of value
					 */
					line[value_end + 1] = save[1];
					/* Write out the comment */
					fwrite(&line[value_end + 1],
					       1,
					       line_len - (value_end + 1),
					       out_file);
				}
				else
				{
					fwrite(&line[curr], 1, line_len - curr,
					       out_file);
				}
			}
			/* Nothing left to parse, jump to the end */
			curr = line_len;
			break;
		case 0: /* Trampled white space */
		case ' ':
		case '\t':
			switch(state)
			{
			case SS_TAG:
				if( tag_start != -1 )
				{
					/* Found the gutter */
					if( out_file )
					{
						/* Write the tag out */
						fwrite(*tag,
						       1,
						       strlen(*tag),
						       out_file);
						/*
						 * Write out the trampled char
						 */
						fwrite(&save[0],
						       1, 1, out_file);
					}
					state = SS_GUTTER;
				}
				else if( out_file )
				{
					/* Write out the space */
					fwrite(&line[curr], 1, 1, out_file);
				}
				break;
			case SS_GUTTER:
				if( out_file )
				{
					/* Write out the space */
					fwrite(&line[curr], 1, 1, out_file);
				}
				break;
			case SS_VALUE:
				if( out_file )
				{
					/*
					 * If the trampled char was a linefeed
					 * then we need to write the value and
					 * the linefeed.
					 */
					if( (line[curr] == 0) &&
					    (curr == line_len - 1) )
					{
						fwrite(*value, 1,
						       strlen(*value),
						       out_file);
						fwrite("\n", 1, 1, out_file);
					}
				}
				break;
			default:
				break;
			}
			break;
		case '\n':
			if( out_file )
			{
				/* Write out value and the linefeed */
				fwrite(*value, 1, strlen(*value), out_file);
				fwrite(&line[curr], 1, 1, out_file);
			}
			break;
		default:
			switch(state)
			{
			case SS_TAG:
				if( tag_start == -1 )
					tag_start = curr;
				tag_end = curr;
				break;
			case SS_GUTTER:
				state = SS_VALUE;
				/* Fallthrough */
			case SS_VALUE:
				if( value_start == -1 )
					value_start = curr;
				value_end = curr;
				break;
			}
			break;
		}
		sps->column++;
	}
	if( (tag_start != -1) && (tag_end != -1) )
	{
		/* Save the char we're gonna trample */
		save[0] = line[tag_end + 1];
		line[tag_end + 1] = 0;
		*tag = &line[tag_start];
	}
	else
	{
		*tag = ""; /* better than NULL... */
	}
	if( (value_start != -1) && (value_end != -1) )
	{
		/* Save the char we're gonna trample */
		save[1] = line[value_end + 1];
		line[value_end + 1] = 0;
		*value = &line[value_start];
	}
	else
	{
		*value = "";
	}
	/* Check for I/O error */
	if( out_file && ferror(out_file) )
		retval = 0;
	return( retval );
}

/*
 * Silly function to restore the characters that parseLine tramples
 */
static void restoreLine(sectionParseState *sps)
{
	int save_idx = 0, lpc;

	for( lpc = 0; lpc < sps->len; lpc++ )
	{
		/* Find any nulls and replace with saved chars */
		if( !sps->data[lpc] )
		{
			sps->data[lpc] = sps->save[save_idx];
			save_idx++;
		}
	}
}

/* Strings that represent `true' */
char *trueStrings[] = {
	"1",
	"true",
	"yes",
	"y",
	"t",
	0
};

/* Strings that represent `false' */
char *falseStrings[] = {
	"0",
	"false",
	"no",
	"n",
	"f",
	0
};
/*
 * Note:  The order of the values in trueStrings and falseStrings is important,
 * and should be of consistent style.
 */

unsigned long parseFlagString(char *value,
			      unsigned long flags,
			      unsigned long flag_bit)
{
	int retval = flags;
	int lpc;

	for( lpc = 0; trueStrings[lpc]; lpc++ )
	{
		/* If `value' is true, then set the bit */
		if( !strcasecmp(value, trueStrings[lpc]) )
		{
			retval |= flag_bit;
			break;
		}
		/* If `value' is false, then clear the bit */
		if( !strcasecmp(value, falseStrings[lpc]) )
		{
			retval &= ~flag_bit;
			break;
		}
	}
	return( retval );
}

char *makeFlagString(unsigned long flags, unsigned long flag_bit, char *value)
{
	/* Default to the old value */
	char *retval = value;

	/* Check for a change from the old value */
	if( parseFlagString(value, flags, flag_bit) != flags )
	{
		int lpc;

		/* Loop through all the strings */
		for( lpc = 0; trueStrings[lpc]; lpc++ )
		{
			/*
			 * If this true matches, then return the corresponding
			 * false string
			 */
			if( !strcasecmp(value, trueStrings[lpc]) )
			{
				retval = falseStrings[lpc];
				break;
			}
			/*
			 * If this false matches, then return the corresponding
			 * true string
			 */
			if( !strcasecmp(value, falseStrings[lpc]) )
			{
				retval = trueStrings[lpc];
				break;
			}
		}
	}
	return( retval );
}

int syncSectionFile(sectionFile *sf, int create)
{
	int iLockRoot, retval;

	lockMutex(sf);
	retval = syncFile(0, sf, sf->filename, create);
	unlockMutex(sf);
	return( retval ); 
}

int syncSectionFileWith(sectionFile *sf, char *filename, int create)
{
	int iLockRoot, retval;

	lockMutex(sf);
	retval = syncFile(0, sf, filename, create);
	sf->time = 0; // XXX
	unlockMutex(sf);
	return( retval );
}
