/*
 * sectionType.c
 * Functions for managing the sectionType structure
 *
 * Copyright (c) 2000 University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Contributed by the Flux Research Group, Department of Computer Science,
 * University of Utah, http://www.cs.utah.edu/flux/
 */

#include <ctype.h>

#include "sectionFile.h"
#include "sectionFile-inline.h"

/* Number of buckets for the file_section hash table */
#define FILE_SECTION_TYPES_SIZE 31

/* Hash table for file_section's */
static sectionType *fileSectionTypes[FILE_SECTION_TYPES_SIZE];

/* nullSectionType functions */
static int nullInstall(struct _sectionType *st, int tag, va_list *args);
static void nullUninstall(struct _sectionType *st);
static int nullInit(struct _sectionType *st, struct _sectionFileData *sfd);
static void nullKill(struct _sectionType *st, struct _sectionFileData *sfd);
static int nullCache(struct _sectionType *st, struct _sectionFileData *sfd,
		     void *ps, char *line, int len);
static int nullFlush(struct _sectionType *st, struct _sectionFileData *sfd,
		     void *ps, char *line, int len, FILE *out_file);

static sectionTypeFunctions nullSectionFunctions = {
	nullInstall,
	nullUninstall,
	nullInit,
	nullKill,
	nullCache,
	nullFlush,
	0,
};

static sectionType nullSectionType = {
	0,
	0,
	&nullSectionFunctions,
	0,
	sizeof(sectionFileData),
	"null"
};

/*
 * Produce a hash value within table_size, for string `name'
 */
static int hashName(const char *name, int table_size)
{
	int h;
	
	for( h = 0; *name; name++ )
		h = (64 * h + tolower(*name)) % table_size;
	return( h );
}

static
void copyFunctions(sectionTypeFunctions *dest, sectionTypeFunctions *src)
{
	if( !dest->install )
		dest->install = src->install;
	if( !dest->uninstall )
		dest->uninstall = src->uninstall;
	if( !dest->init )
		dest->init = src->init;
	if( !dest->kill )
		dest->kill = src->kill;
	if( !dest->cache )
		dest->cache = src->cache;
	if( !dest->flush )
		dest->flush = src->flush;
	if( !dest->walk )
		dest->walk = src->walk;
}

void installSectionType(sectionType *st, int tag, ...)
{
	sectionType *temp = &nullSectionType;
	int hash;
	va_list args;

	va_start(args, tag);
	/* Null everything out... */
	st->parent = 0;
	st->name = 0;
	st->offset = 0;
	st->dataSize = 0;
	st->stf = 0;
	/* Fill in with the tag data */
	while( tag != STA_TAG_DONE )
	{
		switch( tag )
		{
		case STA_Template:
			temp = va_arg(args, sectionType *);
			break;
		case STA_Parent:
			st->parent = va_arg(args, sectionType *);
			break;
		case STA_Name:
			st->name = va_arg(args, char *);
			break;
		case STA_Size:
			st->dataSize = va_arg(args, int);
			break;
		case STA_Functions:
			st->stf = va_arg(args, sectionTypeFunctions *);
			copyFunctions(st->stf, temp->stf);
			break;

		case STA_Install:
			st->stf->install = va_arg(args, void *);
			break;
		case STA_Uninstall:
			st->stf->uninstall = va_arg(args, void *);
			break;
		case STA_Init:
			st->stf->init = va_arg(args, void *);
			break;
		case STA_Kill:
			st->stf->kill = va_arg(args, void *);
			break;
		case STA_Cache:
			st->stf->cache = va_arg(args, void *);
			break;
		case STA_Flush:
			st->stf->flush = va_arg(args, void *);
			break;
		case STA_Walk:
			st->stf->walk = va_arg(args, void *);
			break;
			
		default:
			sfInstall(st, tag, &args);
			break;
		}
		tag = va_arg(args, int);
	}
	/* Fill in any functions that are missing with the template's */
	if( !st->stf )
		st->stf = temp->stf;
	st->dataSize += temp->dataSize;
	if( !st->parent )
		st->parent = temp->parent;
	/* Align the size */
	st->dataSize = (st->dataSize + (ALIGNMENTOF_VOIDP - 1)) &
		~(ALIGNMENTOF_VOIDP - 1);
	if( st->parent )
	{
		/* Compute the offset for the type's local data */
		st->offset = st->parent->dataSize;
		/* Compute the size of the whole object */
		st->dataSize += st->parent->dataSize;
	}
	else
	{
		st->offset = sizeof(sectionFileData);
		st->dataSize += sizeof(sectionFileData);
	}
	/* Add the type to the hash table if it has a name */
	if( st->name )
	{
		hash = hashName(st->name, FILE_SECTION_TYPES_SIZE);
		st->next = fileSectionTypes[hash];
		fileSectionTypes[hash] = st;
	}
	va_end(args);
}

void uninstallSectionType(sectionType *st)
{
	sectionType *curr, **prev;
	int hash;

	hash = hashName(st->name, FILE_SECTION_TYPES_SIZE);
	prev = &fileSectionTypes[hash];
	curr = fileSectionTypes[hash];
	while( curr != st )
	{
		prev = &curr->next;
		curr = curr->next;
	}
	*prev = curr->next;
}

sectionType *findSectionType(char *name)
{
	sectionType *retval = 0, *st;
	int hash;

	hash = hashName(name, FILE_SECTION_TYPES_SIZE);
	st = fileSectionTypes[hash];
	while( st && !retval )
	{
		if( !strcmp(name, st->name) )
		{
			retval = st;
		}
		st = st->next;
	}
	return( retval );
}

/* Dummy functions if a handler doesn't implement them */
int nullInstall(struct _sectionType *st, int tag, va_list *args)
{
	if( st->parent )
		return( sfSuperInstall(st, tag, args) );
	else
		return( 1 );
}

void nullUninstall(struct _sectionType *st)
{
	if( st->parent )
		sfSuperUninstall(st);
}

int nullInit(struct _sectionType *st, struct _sectionFileData *sfd)
{
	if( st->parent )
		return( sfSuperInit(st, sfd) );
	else
		return( 1 );
}

void nullKill(struct _sectionType *st, struct _sectionFileData *sfd)
{
	if( st->parent )
		sfSuperKill(st, sfd);
}

int nullCache(struct _sectionType *st, struct _sectionFileData *sfd,
	      void *ps, char *line, int len)
{
	if( st->parent )
		return( sfSuperCache(st, sfd, ps, line, len) );
	else
		return( 1 );
}

int nullFlush(struct _sectionType *st, struct _sectionFileData *sfd,
	      void *ps, char *line, int len, FILE *out_file)
{
	if( st->parent )
		return( sfSuperFlush(st, sfd, ps, line, len, out_file) );
	else
		return( 1 );
}
