/*
 * PortAuthority.java
 * Emulates the services provided by the real class.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

public class PortAuthority
{
    private Hashtable exports = new Hashtable();

    private Vector anonymousExports = new Vector();
    
    synchronized void revokeAllExports()
    {
	Hashtable table;
	Enumeration e;
	int lpc, size;
	Vector anon;
	
	synchronized(this)
	{
	    table = this.exports;
	    this.exports = null;

	    anon = this.anonymousExports;
	    this.anonymousExports = null;
	}
	if( table != null )
	{
	    e = table.elements();
	    size = table.size();
	    for( lpc = 0; lpc < size; lpc++ )
	    {
		ExportManager em = (ExportManager)e.nextElement();

		em.cancelImports();
	    }
	    size = anon.size();
	    for( lpc = 0; lpc < size; lpc++ )
	    {
		((ExportManager)anon.elementAt(lpc)).cancelImports();
	    }
	}
    }

    /**
     * Register the object internally so that we're revoked when the
     * team dies.
     */
    static boolean register(Importable im)
    {
	return true;
    }
    
    /**
     * Unregister the object internally.
     */
    static void unregister(Importable im)
    {
    }

    void revokeAllImports()
    {
    }

    public synchronized ExportManager reserveExportID(Object id)
	throws DuplicateExportException
    {
	ExportManager retval;
	
	if( this.exports.get(id) == null )
	{
	    retval = new ExportManager();
	    this.exports.put(id, retval);
	}
	else
	{
	    throw new DuplicateExportException(id.toString());
	}
	return retval;
    }
    
    public synchronized ExportManager exportObject(Object id, Exportable ex)
	throws DuplicateExportException
    {
	ExportManager retval;
	
	if( this.exports.get(id) == null )
	{
	    retval = new ExportManager(ex);
	    this.exports.put(id, retval);
	}
	else
	{
	    throw new DuplicateExportException(id.toString());
	}
	return retval;
    }

    public void revokeObject(Object id)
	throws NoSuchExportException
    {
	ExportManager em;
	
	synchronized(this)
	{
	    em = (ExportManager)this.exports.remove(id);
	}
	if( em != null )
	{
	    em.cancelImports();
	}
	else
	{
	    throw new NoSuchExportException(id.toString());
	}
    }

    public synchronized ExportManager exportAnonymousObject(Exportable ex)
    {
	ExportManager retval = new ExportManager(ex);

	this.anonymousExports.addElement(retval);
	return retval;
    }
    
    private void revokeAnonymousObject(ExportManager em)
    {
	synchronized(this)
	{
	    this.anonymousExports.removeElement(em);
	}
	em.cancelImports();
    }

    public void revokeAnonymousObject(Exportable ex)
	throws NoSuchExportException
    {
	boolean done = false;
	int lpc, size;

	size = this.anonymousExports.size();
	for( lpc = 0; (lpc < size) && !done; lpc++ )
	{
	    ExportManager em = (ExportManager)this.anonymousExports.
		elementAt(lpc);

	    if( em.getExport() == ex )
	    {
		this.anonymousExports.removeElementAt(lpc);
		this.revokeAnonymousObject(em);
		done = true;
	    }
	}
	if( !done )
	{
	    throw new NoSuchExportException(ex.toString());
	}
    }
    
    private synchronized ExportManager findExport(Object id)
    {
	ExportManager retval;

	retval = (ExportManager)this.exports.get(id);
	return retval;
    }

    public void importObject(Importable to, TeamHandle from, Object id)
	throws NoSuchExportException
    {
	try
	{
	    TeamReservation ta = from.pushReservation();
	    ExportManager em;
	    
	    ta.visit();
	    {
		PortAuthority pa = Team.current().getPortAuthority();

		em = pa.findExport(id);
	    }
	    ta.leave();
	    try
	    {
		em.addImporter(to);
	    }
	    catch(NullPointerException e)
	    {
		throw new NoSuchExportException(id.toString());
	    }
	    ta.pop();
	}
	catch(TeamEjectionException e)
	{
	    throw new NoSuchExportException(id.toString());
	}
	catch(DeadTeamException e)
	{
	    throw new NoSuchExportException(id.toString());
	}
    }

    public void cancelImport(Importable im)
    {
	try
	{
	    TeamBackEnd be = null;
	    ExportManager em;
	    
	    synchronized(im)
	    {
		em = im.getExportManager();
		if( em != null )
		{
		    be = em.getOwner();
		}
	    }
	    if( em != null )
	    {
		TeamReservation ta;

		try
		{
		    ta = be.pushReservation();
		    if( em.remImporter(im) )
		    {
			ta.visit();
			{
			    PortAuthority pa = Team.current().
				getPortAuthority();

			    pa.revokeAnonymousObject(em);
			}
			ta.leave();
		    }
		    ta.pop();
		}
		catch(DeadTeamException e)
		{
		}
	    }
	}
	catch(TeamEjectionException e)
	{
	}
    }
}
