/*
 * domainResource.c
 * Domain resource for teams, wraps an_domain object from Moab.
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include "assert_pp.h"
#include "locks.h"
#include "thread.h"
#include "janosvm/resource.h"
#include "janosvm/memResource.h"
#include "janosvm/jteam.h"
#include "janosvm/nativeClass.h"
#include "janosvm/nativeClass-inline.h"
#include "module.h"

#include "domainResource.h"

#include "edu_utah_janos_nodeos_moab_DomainSpec.h"
#define JAVA_DOMAIN_SPEC "edu/utah/janos/nodeos/moab_DomainSpec"

/** Module declarations **/
int initDomainResourceModule(int stage);
void killDomainResourceModule(int stage);

module_t domainResourceModule = {
	"domain",
	0,
	initDomainResourceModule,
	killDomainResourceModule,
	0,
};

/** Resource struct functions **/
static int domainSubscribe(struct _resource *res,
			struct _resourceClient *rc,
			struct _resourceSpec *rs);
static int domainEnable(struct _resource *res,
		     struct _resourceClient *rc);
static void domainUnsubscribe(struct _resource *res,
			   struct _resourceClient *rc);

static resourceFunctions domainResourceFunctions = {
	0,		/* install */
	0,		/* uninstall */
	0,		/* configure */
	0,		/* fillSpec */
	0,		/* createJSISpec */
	0,		/* deleteJSISpec */
	0,		/* defaultSpec */
	domainSubscribe,
	domainEnable,
	domainUnsubscribe
};

/** domainResource variables **/

static domainResource metalDomainResource;

/*
 * We statically allocate a client for the metal team so that we can get it set
 * up early on.
 */
domainResourceClient metalDomainResourceClient = {
	{
		0,
		0,
		&metalDomainResource.link,
		0
	},
	NULL
};

/** Module functions **/
int initDomainResourceModule(int stage)
{
	metalDomainResourceClient.domain = an_domain_current();

	installResource(&metalDomainResource.link,
		RA_Name, "domain",
		/* Not a permanent resource: RA_Flags, RF_PERMANENT, */
		RA_Functions, &domainResourceFunctions,
		RA_ClientSize, sizeof(domainResourceClient),
		RA_SpecSize, sizeof(domainResourceSpec),
		RA_SpecNativeClass, JAVA_DOMAIN_SPEC,
		RA_TAG_DONE);

	/* Fixup the metal team's domainResource. */
	metalDomainResourceClient.client.owner = metalTeam;

	return 1;
}

void killDomainResourceModule(int stage)
{
	uninstallResource(&metalDomainResource.link);
}

/** Resource functions **/

int domainSubscribe(struct _resource *res,
		  struct _resourceClient *rc,
		  struct _resourceSpec *rs)
{
	domainResourceClient *frc = (domainResourceClient *)rc;

	require(rs != NULL);

	/*
	 * XXX
	 *
	 * At the momement, I don't want to do the actual domain
	 * creation here (as I'd have to pass in all the domain
	 * create parameters via a domainResourceSpec, etc.).
	 * So, we fill in the actual domain pointer with the
	 * hack function domainResourceFixup(), below.
	 */
	rc->id = 0;
	frc->domain = (an_domain_t*)-1;

	return 1;
}

int domainResourceFixupCurrent(an_domain_t *domain)
{

	domainResourceClient* frc;

	require(domain != NULL);

	frc = (domainResourceClient*)
		getTeamExternalResource(NULL, &metalDomainResource.link, 0);

	require(frc != NULL);
	require(frc->domain == (an_domain_t*)(-1));

	frc->domain = domain;

	return 1;
}

int domainEnable(struct _resource *res, struct _resourceClient *rc)
{
	require(currentTeam() == rc->owner);

	return 1;
}

void domainUnsubscribe(struct _resource *res, struct _resourceClient *rc)
{
	domainResourceClient *frc = (domainResourceClient *)rc;
	jthread_t th = jthread_current();
	an_error_t err;

	require(jthread_get_data(th)->currentTeam == rc->owner);
	require(frc->domain != NULL);
	require(frc->domain != an_domain_current());
	
	/*
	 * Kill the domain.
	 */
	err = an_domain_destroy(frc->domain);
	assert(err == 0); /* XXX XXX */

	ensure(jthread_get_data(th)->currentTeam == rc->owner);
}

/* XXX move into moab_DomainSpec.c */

void edu_utah_janos_nodeos_moab_DomainSpec_nativeConstructor(
	struct Hedu_utah_janos_nodeos_moab_DomainSpec *mfs)
{
	domainResourceSpec *frs;

	frs = OBJECT2NATIVE(&metalDomainResource.link.specNativeClass,
		(Hjava_lang_Object*)mfs);

	frs->spec.handler = &metalDomainResource.link;
}

void edu_utah_janos_nodeos_moab_DomainSpec_nativeFinalizer(
	struct Hedu_utah_janos_nodeos_moab_DomainSpec *mfs)
{
	/* nothing to clean up... */
}
