/*
 * jsi.h
 * JanosVM Scripting Interface functions
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

#include <stdio.h>

#define DO_JSI_INLINE

#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>

#include "assert_pp.h"
#include <jsi/jsiImpl.h>

#include <jsi/jsi.h>
#include <jsi/jsiPlexLink.h>
#include <jsi/jsiPlexChannel.h>
#include "jsiPlexUDP.h"

#include "kaffevm/locks.h"

#include <misc/stringParsing.h>

#include <nodeos/lib/arplib.h>

static
void obtainUDPLink(jsiPlexLink *pl, int *iLockRoot)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;

	_lockMutex(&pu->lock, iLockRoot);
}

static
void releaseUDPLink(jsiPlexLink *pl, int *iLockRoot)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;

	_unlockMutex(&pu->lock, iLockRoot);
}

static
int waitForUDPLink(jsiPlexLink *pl, int ms_timeout)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;
	int retval;

	retval = _waitCond(&pu->lock, ms_timeout);
	return( retval );
}

static
void signalUDPLink(jsiPlexLink *pl)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;

	_signalCond(&pu->lock);
}

struct ethHeader {
#define ETHER_ADDRESS_SIZE 6
	char daddr[ETHER_ADDRESS_SIZE];
	char saddr[ETHER_ADDRESS_SIZE];
#define ETHER_TYPE_IP 0x800
	unsigned short type;
};

struct recvContext {
	jsiPlexLinkContext plc;
	jsiPlexUDPIn *pui;
	unsigned int ipAddr;
	unsigned short udpPort;
	unsigned short srcPort;
	char physAddr[ETHER_ADDRESS_SIZE];
};

static
void jsiCallBack(an_pbuf_t *buffer, void *arg, int payloadStart)
{
	jsiPlexInLink *pil = (jsiPlexInLink *)arg;
	jsiPlexUDPIn *pui = (jsiPlexUDPIn *)arg;
	char *data = an_pbuf_data(buffer);
	jsiPlexLinkPacket *plp;
	struct recvContext rc;
	int iplen, iLockRoot;
	struct ethHeader *eh;
	struct udphdr *udph;
	jsiChannel *jc;
	struct ip *iph;

	anjthread_enterkaffe_c();
	eh = (struct ethHeader *)data;
	iph = (struct ip *)(eh + 1);
	iplen = (iph->ip_hl << 2);
	udph = (struct udphdr *)(((char *)iph) + iplen);
	rc.pui = pui;
	rc.ipAddr = iph->ip_src.s_addr;
	rc.udpPort = udph->uh_sport;
	rc.srcPort = udph->uh_dport;
	rc.plc.hashValue = rc.ipAddr;
	memcpy(rc.physAddr, eh->saddr, ETHER_ADDRESS_SIZE);
	plp = (jsiPlexLinkPacket *)(&data[sizeof(struct ethHeader) +
					 iplen +
					 sizeof(struct udphdr) -
					 sizeof(short) -
					 sizeof(void *)]);
	plp->packetSize = ntohs(udph->uh_ulen) - sizeof(struct udphdr);
	plp->header.fragOffset = ntohs(plp->header.fragOffset);
	plp->header.totalLength = ntohs(plp->header.totalLength);
	plp->header.messageID = ntohs(plp->header.messageID);
	plp->header.channelID = ntohs(plp->header.channelID);
	jsiObtainPlexLink(pil->link);
	jc = jsiPlexWindowRecvPacket(pil, 0, &rc.plc, plp);
	jsiReleasePlexLink(pil->link);
	if( jc )
		pil->provider->handleConnection->func(pil->provider, jc);
	an_pbuf_return(buffer);
	anjthread_exitkaffe_c();

	/* Shouldn't return */
	ensure(0);
}

unsigned short ipchksum(char *addr, int len)
{
	long retval = 0;
	
	while( len > 1 )
	{
		retval += *(unsigned short *)addr;
		addr += 2;
		len -= 2;
	}
	if( len > 0 )
		retval += *(unsigned char *)addr;
	while( retval >> 16 )
		retval = (retval & 0xffff) + (retval >> 16);
	retval = ~retval;
	return( retval );
}

static int ip_magic_id = 0x23;

static
int writeUDPLink(jsiPlexOutLink *pol, void *buffer, int len)
{
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;
	int total_size, retval = 0;
	struct udphdr *udph;
	struct ip *iph;
	void *data;
	int bcnt = 0;

	while (an_pbuf_isbusy(&puo->pbuf))
	{
		bcnt++;
		if( bcnt > 100000 )
			printf("bcnt is huge %d\n", bcnt);
	}
	total_size = sizeof(struct ip) + sizeof(struct udphdr) + len;
	data = an_pbuf_data(&puo->pbuf);
	an_pbuf_setsize(&puo->pbuf, total_size, 0);
	iph = (struct ip *)data;
	iph->ip_len = htons(total_size);
	iph->ip_id = ++ip_magic_id;
	iph->ip_sum = 0;
	iph->ip_sum = ipchksum((char *)iph, sizeof(struct ip));
	assert(ipchksum((char *)iph, sizeof(struct ip)) == 0);
	udph = (struct udphdr *)(iph + 1);
	udph->uh_ulen = htons(len + sizeof(struct udphdr));
	memcpy(((char *)data) + sizeof(struct ip) + sizeof(struct udphdr),
	       buffer,
	       len);
	if( an_outchan_send(&puo->outChannel, &puo->pbuf) == 0 )
	{
		retval = 1;
	}
	else
	{
		fprintf(stderr, "Outchan send failed?\n");
	}
	return( retval );
}

static
jsiChannel *handleUDPLink(jsiPlexInLink *pil)
{
	return NULL;
}

static
jsiPlexInLink *getUDPIn(jsiPlexLink *pl, char *format)
{
	parsedString interface, protocol, host;
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;
	jsiPlexInLink *retval = 0;
	int port;

	if( parseString(format,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_NonEmptyString, &interface,
			SPO_Expect, "/",
			SPO_NonEmptyString, &host,
			SPO_Expect, "/",
			SPO_Integer, &port,
			SPO_End) )
	{
		struct in_addr addr;
		jsiPlexUDPIn *pui;

		host.data[host.len] = '\0';
		if( inet_aton(host.data, &addr) &&
		    (pui = jsiMalloc(sizeof(jsiPlexUDPIn))) )
		{
			struct {
				an_chanproto_t proto;
				an_chanaddr_t addr;
				an_netspec_t net;
				an_pbuf_t *bufs[JSI_PLEX_UDP_RECV_QUEUE_SIZE];
			} attr;
			an_domain_t *curdomain = an_domain_current();
			an_error_t err;
			void *ptr;
			int lpc;
			
			pu->ipAddr = addr.s_addr;
			pu->inPort = htons(port);
			pu->outPort = htons(1025); // XXX
			pui->users += 1;
			pui->link.link = pl;
			pui->link.provider = 0;
			ptr = an_demuxkey_create(&pui->demux, curdomain);
			assert(ptr != 0);
			err = an_demuxkey_addeth(&pui->demux,
						 AN_DEMUXKEY_ETH_ANYSADDR |
						 AN_DEMUXKEY_ETH_ANYDADDR,
						 0,
						 0,
						 htons(ETHER_TYPE_IP));
			assert(err == 0);
			err = an_demuxkey_addipv4(&pui->demux,
						  AN_DEMUXKEY_IPV4_ANYSADDR,
						  0, 0,
						  pu->ipAddr, 0xffffffff,
						  IPPROTO_UDP);
			err = an_demuxkey_addudp(&pui->demux,
						 AN_DEMUXKEY_UDP_ANYSPORT,
						 0,
						 pu->inPort);
			assert(err == 0);
			memset(&attr, 0, sizeof(attr));
			strncpy(attr.proto.data,
				interface.data,
				interface.len);
			attr.proto.len = strlen(attr.proto.data);
			sscanf(attr.proto.data, "eth%d", &pu->ifNumber);
			attr.net.maxthreads = 1;
			attr.net.bandwidth = 100;
			attr.net.buffers.npbufs = JSI_PLEX_UDP_RECV_QUEUE_SIZE;
			attr.net.buffers.pbufs = attr.bufs;
			for( lpc = 0;
			     lpc < JSI_PLEX_UDP_RECV_QUEUE_SIZE;
			     lpc++ )
			{
				attr.bufs[lpc] = &pui->pbufs[lpc];
				ptr = an_pbuf_create(&pui->pbufs[lpc],
						     AN_PBUF_ETH,
						     0,
						     0);
				assert(ptr != 0);
			}
			ptr = an_inchan_create(&pui->inChannel,
					       curdomain,
					       &pui->demux,
					       &attr.proto,
					       &attr.addr,
					       &attr.net,
					       jsiCallBack,
					       pui);
			assert(ptr != 0);
			{
				/* Setup ARP for the interface */
				arp_init();
				arp_config_iface(pu->ifNumber, &addr);
			}
			retval = &pui->link;
		}
		else
		{
			fprintf(stderr,
				"Unable to parse `%s' for a IP address\n",
				host.data);
		}
	}
	else
	{
		fprintf(stderr,
			"Provider format is plex:udp:<ethernet interface>/"
			"<ip address>/<port number>\n");
	}
	return( retval );
}

static
void dropUDPIn(jsiPlexInLink *pil)
{
	jsiPlexUDPIn *pui = (jsiPlexUDPIn *)pil;

	pui->users -= 1;
	if( pui->users <= 0 )
	{
		int lpc;
		
		an_chan_destroy(&pui->inChannel);
		for( lpc = 0; lpc < JSI_PLEX_UDP_RECV_QUEUE_SIZE; lpc++ )
		{
			an_pbuf_destroy(&pui->pbufs[lpc]);
		}
		jsiFree(pil);
	}
}

static
jsiPlexOutLink *getUDPOutFromContext(jsiPlexLink *pl, jsiPlexLinkContext *plc)
{
	struct recvContext *rc = (struct recvContext *)plc;
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;
	jsiPlexUDPOut *puo, *match = 0;
	jsiPlexOutLink *retval = 0;

	puo = pu->outLinks;
	while( puo && !match )
	{
		if( (puo->ipAddr == rc->ipAddr) &&
		    (puo->udpPort == rc->udpPort) )
		{
			match = puo;
		}
		puo = puo->next;
	}
	if( match )
	{
		match->link.users++;
		retval = &match->link;
	}
	else if( (puo = (jsiPlexUDPOut *)jsiMalloc(sizeof(jsiPlexUDPOut))) )
	{
		struct {
			an_chanproto_t proto;
			an_chanaddr_t addr;
			an_netspec_t net;
		} attr;
		struct udphdr *udph;
		void *data, *ptr;
		struct ip *iph;
		
		puo->link.link = pl;
		puo->link.hashValue = rc->ipAddr;
		puo->link.flags = POLF_UP;
		puo->link.users = 1;
		puo->pui = rc->pui;
		puo->pui->users += 1;
		puo->ipAddr = rc->ipAddr;
		puo->udpPort = rc->udpPort;
		puo->srcPort = rc->srcPort;
		memset(&attr, 0, sizeof(attr));
		puo->next = pu->outLinks;
		pu->outLinks = puo;
		sprintf(attr.proto.data, "if%d", pu->ifNumber);
		attr.proto.len = strlen(attr.proto.data);
		memcpy(attr.addr.eth.daddr,
		       rc->physAddr,
		       ETHER_ADDRESS_SIZE);
		attr.addr.eth.type = htons(ETHER_TYPE_IP);
		an_pbuf_create(&puo->pbuf, AN_PBUF_ETH, 0, 0);
		data = an_pbuf_data(&puo->pbuf);
		iph = (struct ip *)data;
		iph->ip_v = 4;
		iph->ip_hl = 20 >> 2;
		iph->ip_tos = 0;
		iph->ip_len = 0;
		iph->ip_off = 0;
		iph->ip_ttl = 64;
		iph->ip_p = IPPROTO_UDP;
		iph->ip_sum = 0;
		iph->ip_src.s_addr = pu->ipAddr;
		iph->ip_dst.s_addr = puo->ipAddr;
		udph = (struct udphdr *)(iph + 1);
		udph->uh_sport = puo->srcPort;
		udph->uh_dport = puo->udpPort;
		udph->uh_ulen = 0;
		udph->uh_sum = 0;
		attr.net.bandwidth = 100;
		// XXX Check return value...
		ptr = an_outchan_create(&puo->outChannel,
					an_domain_current(),
					&attr.proto,
					&attr.addr,
					&attr.net);
		assert(ptr != 0);
		retval = &puo->link;
	}
	return( retval );
}

static
jsiPlexOutLink *getUDPOutFromFormat(jsiPlexLink *pl, char *format)
{
	jsiPlexOutLink *retval = 0;

#if 0
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;
	parsedString protocol, host;
	int port;

	if( parseString(format,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_NonEmptyString, &host,
			SPO_Expect, "/",
			SPO_Integer, &port,
			SPO_End) )
	{
		struct recvContext rc;
		int valid = 1;
		char save;

		memset(&rc, 0, sizeof(struct recvContext));
		save = host.data[host.len];
		host.data[host.len] = '\0';
		if( (rc.ipAddr = inet_addr(host.data)) == -1 )
		{
			/* Its not a number address either */
			valid = 0;
		}
		rc.udpPort = htons(port);
		rc.srcPort = pu->outPort;
		host.data[host.len] = save;
		arp_arpfor(pu->ifNumber,
			   (struct in_addr *)&rc.ipAddr, // XXX
			   rc.physAddr);
		if( valid )
			retval = getUDPOutFromContext(pl, &rc.plc);
	}
#endif
	return( retval );
}

static
void dropUDPOut(jsiPlexOutLink *pol)
{
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;

	require(puo != 0);
	require(puo->link.users > 0);

	puo->link.users--;
	if( puo->link.users == 0 )
	{
		jsiPlexLinkPacket *plp, *plp_next;
		jsiPlexUDPOut *curr, **prev;

		dropUDPIn(&puo->pui->link);
		curr = ((jsiPlexUDP *)pol->link)->outLinks;
		prev = &((jsiPlexUDP *)pol->link)->outLinks;
		while( curr != puo )
		{
			prev = &curr->next;
			curr = curr->next;
		}
		*prev = curr->next;
		while (an_pbuf_isbusy(&puo->pbuf));
		plp = pol->packetPool;
		while( plp )
		{
			plp_next = plp->next;
			jsiFree(plp);
			plp = plp_next;
		}
		an_chan_destroy(&puo->outChannel);
		an_pbuf_destroy(&puo->pbuf);
		jsiFree(puo);
	}
}

int compareUDPOut(jsiPlexOutLink *pol, jsiPlexLinkContext *plc)
{
	struct recvContext *rc = (struct recvContext *)plc;
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;
	int retval = 0;

	if( (puo->ipAddr == rc->ipAddr) && (puo->udpPort == rc->udpPort) )
		retval = 1;
	return( retval );
}

char *encodeUDPOut(jsiPlexOutLink *pol)
{
	jsiPlexUDPOut *puo = (jsiPlexUDPOut *)pol;
	char *retval = 0;

	if( (retval = jsiMalloc(3 + 1 +
				15 + 1 +
				5 +
				1)) )
	{
		sprintf(retval,
			"udp:%d.%d.%d.%d/%d/%d",
			((char *)&puo->ipAddr)[0] & 0xff,
			((char *)&puo->ipAddr)[1] & 0xff,
			((char *)&puo->ipAddr)[2] & 0xff,
			((char *)&puo->ipAddr)[3] & 0xff,
			ntohs(puo->udpPort),
			ntohs(puo->srcPort));
	}
	return( retval );
}

static
jsiPlexOutLink *getUDPOutFromEncoding(jsiPlexLink *pl, char *enc)
{
	jsiPlexUDP *pu = (jsiPlexUDP *)pl;
	parsedString protocol, host;
	jsiPlexOutLink *retval = 0;
	int srcPort,port;

	if( parseString(enc,
			SPO_NonEmptyString, &protocol,
			SPO_Expect, ":",
			SPO_NonEmptyString, &host,
			SPO_Expect, "/",
			SPO_Integer, &port,
			SPO_Expect, "/",
			SPO_Integer, &srcPort,
			SPO_End) )
	{
		struct recvContext rc;
		int valid = 1;
		char save;
		
		memset(&rc, 0, sizeof(struct recvContext));
		save = host.data[host.len];
		host.data[host.len] = '\0';
		if( (rc.ipAddr = inet_addr(host.data)) == -1 )
		{
			/* Its not a number address either */
			valid = 0;
		}
		rc.udpPort = htons(port);
		rc.srcPort = htons(srcPort);
		host.data[host.len] = save;
		arp_arpfor(pu->ifNumber,
			   (struct in_addr *)&rc.ipAddr, // XXX
			   rc.physAddr);
		if( valid )
			retval = getUDPOutFromContext(pl, &rc.plc);
	}
	return( retval );
}

jsiPlexLinkFunctions jsiPlexUDPFunctions = {
	0,
	0,
	0,
	obtainUDPLink,
	releaseUDPLink,
	waitForUDPLink,
	signalUDPLink,
	handleUDPLink,
	writeUDPLink,
	getUDPIn,
	dropUDPIn,
	getUDPOutFromFormat,
	getUDPOutFromContext,
	dropUDPOut,
	compareUDPOut,
	encodeUDPOut,
	getUDPOutFromEncoding,
};

static
char *plexUDPNames[] = {
	"udp",
	0
};

jsiPlexLink *jsiCreatePlexUDP()
{
	jsiPlexLink *retval = 0;
	jsiPlexUDP *pu;

	/* fprintf(stderr, "Creating moab jsiPlexUDP\n"); */
	if( (pu = (jsiPlexUDP *)jsiMalloc(sizeof(jsiPlexUDP))) )
	{
		pu->link.next = 0;
		pu->link.names = plexUDPNames;
		pu->link.ops = &jsiPlexUDPFunctions;
		pu->link.maxPacketSize = JSI_PLEX_UDP_MAX_BODY_SIZE;
		pu->lock = 0;
		pu->outLinks = 0;
		retval = &pu->link;
	}
	return( retval );
}
