/*
 * ClassPathResource.java
 * Provides access methods to the current team's class path.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.resources;

import java.io.IOException;

import edu.utah.janosvm.sys.DeadTeamException;

/**
 * Provides access methods to the current team's class path.
 * <p>
 * The class path in a typical VM is used to specify where to find the class
 * files on the local file system, however, in the JanosVM we're trying to
 * simulate multiple JVMs so a single class path for all of the teams might be
 * insufficient.  For example, a user might want to have access to a standard
 * and debugging build of their classes or a more paranoid user might want a
 * team to have no class path to make sure it can't gain access to anything
 * outside of its sandbox.  Whatever the case may be, per-team class paths help
 * maintain the illusion of multiple JVMs running in a single instance of the
 * JanosVM.
 * <p>
 * The current implementation of the resource simply takes the class path as
 * determined at startup and uses it as the global path from which each team
 * can select a subset of paths.  Then when a team is created you can create a
 * <b>ClassPathSpec</b> with an array of strings containing the paths that the
 * team should have access too or just use the default of everything in the
 * global path.
 * <p>
 * <h4>To Do</h4>
 * The internals provide many functions for adding and removing paths from the
 * per-team paths, however, these haven't been
 * exported to the java level.
 *
 * @see ClassPathSpec
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a> */
public final class ClassPathResource
{
    /**
     * Load the file with the given name from the class path.  The file
     * may be anything (not just a .class file).
     *
     * @return The contents of the file as a byte array or null if it wasn't
     * found.
     * @exception Throws an IOException if the file couldn't be found or there
     * was a problem while loading.
     * @exception Throws a DeadTeamException if the current team was unable
     * to visit the kernel to add the path.
     */
    public static native byte[] loadFile(String name)
	    throws IOException, DeadTeamException;

    /**
     * Add a class path to the globally available set.
     *
     * @param path The path string to add to the global class path.
     * @exception Throws an IOException if the resource is unable to open the
     * path.
     * @exception Throws a DeadTeamException if the current team was unable
     * to visit the kernel to add the path.
     */
    public static native void addPath(String path)
	throws IOException, DeadTeamException;

    /**
     * Hide a path from the global class path.
     *
     * @param path The path string to hide from the global class path.
     * @exception Throws an IOException if the resource is unable to open the
     * path.
     * @exception Throws a DeadTeamException if the current team was unable
     * to visit the kernel to hide the path.
     */
    public static native void hidePath(String path)
	throws IOException, DeadTeamException;
    
    /**
     * Remove a path from the global class path.
     *
     * @param path The path string to remove from the global class path.
     * @exception Throws a DeadTeamException if the current team was unable
     * to visit the kernel to remove the path.
     */
    public static native void removePath(String path)
	throws DeadTeamException;

    /**
     * @return An array containing the current set of global class paths.
     */
    public static native String[] getPaths();

    /**
     * Its a static only class.
     */
    private ClassPathResource()
    {
    }
}
