/*
 * NameSpaceGroup.java
 * Helper class for NameSpaceSpec
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.resources;

import edu.utah.janosvm.sys.TeamHandle;

/**
 * Helper class for NameSpaceSpec that specifies which class groups to
 * import from a team.  In addition, you can specify a view group that
 * will map the class names in the class group to something more appropriate
 * for the new team.
 *
 * @see NameSpaceSpec
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public final class NameSpaceGroup
{
    /**
     * The team to import the classes from.
     */
    private final TeamHandle team;
    /**
     * The class group type to import.  Currently, the set of group types is
     * defined in the config file as `classgroup' sections.
     */
    private final String classGroup;
    /**
     * The view group type to import.  Currently, the set of group types is
     * defined in the config file as `viewgroup' sections within the
     * classgroup section that they are mapping.
     */
    private final String viewGroup;

    /**
     * Construct a NameSpaceGroup object with the given parameters.
     *
     * @param team The team to import classes from.
     * @param cg The name of the class group type.
     * @param vg The name of the view group type.
     */
    public NameSpaceGroup(TeamHandle team, String cg, String vg)
    {
	if( team == null )
	    throw new IllegalArgumentException("Team can't be null");
	else
	    this.team = team;
	if( cg == null )
	    throw new IllegalArgumentException("Class group can't be null");
	else
	    this.classGroup = cg;
	if( vg == null )
	    throw new IllegalArgumentException("View group can't be null");
	else
	    this.viewGroup = vg;
    }

    /**
     * Construct a NameSpaceGroup object with the given parameters and
     * the "identity" view group.
     *
     * @param team The team to import classes from.
     * @param cg The name of the class group type.
     */
    public NameSpaceGroup(TeamHandle team, String cg)
    {
	this(team, cg, "identity");
    }

    /**
     * @return The team the classes will be imported from.
     */
    public TeamHandle getTeam()
    {
	return this.team;
    }
    
    /**
     * @return The name of the class group type that will be imported.
     */
    public String getClassGroup()
    {
	return this.classGroup;
    }
    
    /**
     * @return The name of the view group type that will be imported.
     */
    public String getViewGroup()
    {
	return this.viewGroup;
    }

    public String toString()
    {
	return "NameSpaceGroup[team="
	    + this.team
	    + "; classGroup="
	    + this.classGroup
	    + "; viewGroup="
	    + this.viewGroup
	    + "]";
    }
}
