/*
 * StdioSpec.java
 * Resource specification for standard I/O.
 *
 * Copyright (c) 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.resources;

/**
 * Resource specification for the standard I/O channels for a Team.
 * The standard I/O streams (stdin, stderr, stdout) for a new Team
 * are specified here.  Streams are specified in a flexible URI-like
 * format.
 *
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 * @see StdioSpec#StdioSpec(String,String,String)
 */
public final class StdioSpec
    extends ResourceSpec
    implements Cloneable
{

    /**
     * Construct the native part of this class.
     */
    private native void nativeConstructor(String inURI,
					  String outURI,
					  String errURI);

    /**
     * Constructs a standard I/O resource specification with the given
     * stream identifiers.  The stream identifier is a URI like string
     * that describes the type of stream and any additional options.
     * The format is `<var>protocol</var><b>://</b><var>path</var>' where
     * protocol is one of:
     * <menu>
     *   <li>builtin - The VM's "real" stdio streams.  Possible paths are:
     *     <menu>
     *        <li>`none' which specifies no stream</li>
     *        <li>`stdin' for standard in</li>
     *        <li>`stdout' for standard out</li>
     *        <li>`stderr' for standard error.</li>
     *     </menu></li>
     *   <li>jsi - A connected jsiChannel, currently, just the one that
     *     executed run of team x.</li>
     *   <li>file - A regular file where path should be the path of the file to
     *     create.</li>
     *   <li>fd - A file descriptor in the VM, the fd given in the path is
     *     dup'd and used to build a FILE object.</li>
     * </menu>
     *
     * Examples:
     *
     * <menu>
     *   <li>"builtin://none"</li>
     *   <li>"builtin://stdin"</li>
     *   <li>"jsi://"</li>
     *   <li>"file://foo.out"</li>
     * </menu>
     *
     * @param inURI URI for the standard input.
     * @param outURI URI for the standard output.
     * @param errURI URI for the standard error.
     */
    public StdioSpec(String inURI, String outURI, String errURI)
    {
	this.nativeConstructor(inURI, outURI, errURI);
    }

    /**
     * Finalize the native part of the class.
     */
    private native void nativeFinalize();

    public void finalize()
    {
	this.nativeFinalize();
    }

    /**
     * @return The StdioSpec object for the current team.
     */
    public static native StdioSpec current();

    /**
     * @return The StdioSpec object for the current team that has file
     * descriptor URI's.  This is useful for sharing input/output streams
     * with child teams.
     */
    public static native StdioSpec currentFD();

    /**
     * @return The standard input URI.
     */
    public native String getInput();

    /**
     * @return The standard output URI.
     */
    public native String getOutput();

    /**
     * @return The standard error URI.
     */
    public native String getError();

    public String toString()
    {
	return "StdioSpec[stdin="
	    + this.getInput()
	    + "; stdout="
	    + this.getOutput()
	    + "; stderr="
	    + this.getError()
	    + "]";
    }

    public native Object clone();
}
