/*
 * ExportManager.java
 * Class that manages the set of importers for an Exportable.
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

import java.util.Enumeration;
import java.util.NoSuchElementException;

/**
 * Tracks the importables associated with a single exportable.  Each
 * ExportManager is associated with exactly one Exportable.  An
 * exportable may have multiple ExportManagers (if it is exported
 * under mutliple names, for example).
 * <p>
 * An ExportManager tracks the links between an exported object and its
 * imports.  This is required so that we can revoke any references when a team
 * dies or the export is no longer available to other teams.
 *
 * @see Importable
 * @see Exportable
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public final class ExportManager
{
    /**
     * The team that owns the objects managed by this ExportManager.
     */
    private final TeamBackEnd owner = TeamBackEnd.getCurrentTeam();

    /**
     * The object being exported.
     */
    private Exportable export;

    /**
     * List of objects importing our managed object.
     */
    private final List importers = new List();
    
    /**
     * Construct a manager that is exporting the given object.
     *
     * @param export The exported object this object should manage.
     */
    ExportManager(Exportable export)
    {
	this.export = export;
    }

    /**
     * Construct an empty manager.
     */
    ExportManager()
    {
    }

    /**
     * @return The owner of the exported object.
     */
    TeamBackEnd getOwner()
    {
	return this.owner;
    }

    /**
     * If this manager is empty you can fill in its export with this function.
     *
     * @param ex The exported object to be managed.
     * @exception Throws a NullPointerException if ex is null.
     */
    public synchronized void setExport(Exportable ex)
    {
	if( ex == null )
	    throw new NullPointerException();
	if( this.export == null )
	{
	    this.export = ex;
	}
    }

    /**
     * @return The object that is being exported.
     */
    Exportable getExport()
    {
	return this.export;
    }

    /**
     * Add an Importable to the list of objects importing our export.
     *
     * @param im The Importable that wants a reference to our export.
     */
    public synchronized void addImporter(Importable im)
    {
	if( this.export != null )
	{
	    synchronized(im)
	    {
		try
		{
		    /*
		     * Do the bindTo first so that we don't mess up the state
		     * of the world it fails.
		     */
		    im.bindTo(this.export);
		    im.setExportManager(this);
		}
		catch(RuntimeException th)
		{
		    im.setExportManager(null);
		    im.unbindFrom();
		    throw th;
		}
		catch(Error th)
		{
		    im.setExportManager(null);
		    im.unbindFrom();
		    throw th;
		}
	    }
	    this.importers.addTail(im);
	}
	else
	{
	    throw new NullPointerException();
	}
    }

    /**
     * Remove an importer from the list of objects importing our export.
     *
     * @param im The Importable to remove.
     * @return True if this was the last importer of a TemporaryExport, false
     * otherwise.
     */
    synchronized boolean remImporter(Importable im)
    {
	boolean retval = false;

	synchronized(im)
	{
	    im.setExportManager(null);
	    try
	    {
		im.unbindFrom();
	    }
	    catch(Throwable th)
	    {
		VMInstance.panic("Caught exception in unbindFrom:", th);
	    }
	}
	im.remove();
	if( this.importers.isEmpty() &&
	    (this.export instanceof TemporaryExport) &&
	    ((TemporaryExport)this.export).expired() )
	{
	    retval = true;
	}
	return retval;
    }

    /**
     * Disconnect all of the importers of our export.
     */
    synchronized void cancelImports()
    {
	ListNode node;
	
	/* Disallow further importers */
	this.export = null;
	while( (node = this.importers.remHead()) != null )
	{
	    if( node instanceof Importable )
	    {
		Importable im = (Importable)node;
		
		synchronized(im)
		{
		    try
		    {
			im.setExportManager(null);
			im.unbindFrom();
		    }
		    catch(Throwable th)
		    {
			th.printStackTrace();
		    }
		}
	    }
	    else if( node instanceof ImportEnumerator )
	    {
	    }
	    else
	    {
		throw new InternalError();
	    }
	}
    }

    /**
     * The ImportEnumerator is used to traverse the list of importers of
     * an export.  This is primarily useful if an export wants to do callbacks
     * on its imports.  This object is actually inserted into the
     * ExportManager.importer's list and then moved along as it traverses
     * each Importable node.
     */
    public final class ImportEnumerator
	extends ListNode
    {
	/**
	 * Pointer to a TeamReservation that refers to the owner of the current
	 * Importable.  Or, null if there isn't a valid Importable.
	 */
	private TeamReservation tr;

	/**
	 * Pointer to the current Importable we're traversing.
	 */
	private Importable im;

	/**
	 * @return this.tr
	 */
	public TeamReservation getReservation()
	{
	    return this.tr;
	}

	/**
	 * @return this.im
	 */
	public Importable getImportable()
	{
	    return this.im;
	}

	/**
	 * Traverse the next Importable in the list.
	 *
	 * @return True if the next Importable was successfully traversed,
	 * false if its at the end of the list.
	 */
	public boolean nextImport()
	{
	    boolean retval = false;

	    if( this.tr != null )
	    {
		/*
		 * The reservation from the previous importable is still active
		 * pop it off the journal.
		 */
		this.im = null;
		this.tr.pop();
		this.tr = null;
	    }
	    synchronized(ExportManager.this)
	    {
		ListNode curr = null;

		/* Start at the node immediately following the enumerator. */
		curr = this.succ;
		/* Remove ourself so we can be reinserted later. */
		this.remove();
		while( (curr != null) && !retval )
		{
		    if( curr instanceof Importable )
		    {
			TeamBackEnd be;

			/* Get the owner of the Importable */
			be = TeamBackEnd.getObjectOwner(curr);
			if( be != null )
			{
			    try
			    {
				this.tr = be.pushReservation(1);
				/*
				 * The team is reserved so its safe to keep a
				 * pointer to the Importable for awhile.
				 */
				this.im = (Importable)curr;
				retval = true;
			    }
			    catch(DeadTeamException e)
			    {
			    }
			}
		    }
		    else
		    {
			/*
			 * Possibly another ImportEnumerator or the tail
			 * ListNode
			 */
		    }
		    if( !retval )
			curr = curr.succ;
		}
		if( retval &&
		    (curr.succ != null) &&
		    (curr.succ.succ != null) )
		{
		    /*
		     * Still more nodes in the list, add ourselves back in
		     * so we can traverse them.
		     */
		    curr.append(this);
		}
	    }
	    return retval;
	}
    }

    /**
     * @return An ImportEnumerator for this ExportManager.
     */
    public ImportEnumerator getEnumerator()
    {
	return new ImportEnumerator();
    }

    /**
     * Setup the given ImportEnumerator to traverse the importers list.
     *
     * @param ie The ImportEnumerator that will be inserted into the importers
     * list.
     * @return True if this manager is still valid and its importers can
     * be traversed, false otherwise.
     */
    public synchronized boolean imports(ImportEnumerator ie)
    {
	boolean retval = false;

	if( !TeamBackEnd.getCurrentTeam().owns(this) )
	{
	    throw new InternalError();
	}
	if( this.export != null )
	{
	    this.importers.addHead(ie);
	    retval = true;
	}
	return retval;
    }

    /**
     * Similar to the above except that it constructs a new ImportEnumerator
     * and adds that to the list.
     *
     * @return A new ImportEnumerator ready to traverse the list or NULL if
     * this manager is invalid.
     */
    public synchronized ImportEnumerator imports()
    {
	ImportEnumerator retval = null;

	if( this.export != null )
	{
	    retval = new ImportEnumerator();
	    this.imports(retval);
	}
	return retval;
    }

    public String toString()
    {
	return "ExportManager[owner="
	    + this.owner
	    + "; export="
	    + this.export
	    + "; importers="
	    + this.importers
	    + "]";
    }
}
