/*
 * Exportable.java
 * Base class for objects exported from teams.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */
package edu.utah.janosvm.sys;

/**
 * Base class for objects exported from teams.
 */
public abstract class Exportable
{
    /**
     * The owner team of this Exportable, cached here so that importers can
     * easily visit the team to do some work.
     */
    private final TeamBackEnd owner;

    /**
     * Our ExportManager if we're exported anonymously.
     */
    private ExportManager anonManager;

    /**
     * Special constructor used by the kernel constructor since we can't call
     * TeamBackEnd.getCurrentTeam() this early.
     *
     * @param dummy Parameter for overloading in the kernel.
     */
    Exportable(int dummy)
    {
	this.owner = (TeamBackEnd)this;
    }

    /**
     * Standard Exportable constructor, currently, just caches the owner team
     * pointer.
     */
    protected Exportable()
    {
	this.owner = TeamBackEnd.getCurrentTeam();
    }

    /**
     * Set the ExportManager created when this object was exported anonymously.
     *
     * @param em ExportManager thats managing this object.
     */
    void setAnonymousManager(ExportManager em)
    {
	this.anonManager = em;
    }

    /**
     * @return The ExportManager thats managing this anonymously exported
     * object.
     */
    public ExportManager getAnonymousManager()
    {
	return this.anonManager;
    }

    /**
     * Switch the current thread to the this objects owner team.
     *
     * @exception DeadTeamException if one of the teams involved is dead.
     * @exception TeamEjectionException if this thread was ejected from the
     * team while it was running.
     */
    public void switchToOwner()
	throws DeadTeamException, TeamEjectionException
    {
	if( !this.owner.switchTo(1) )
	    throw new DeadTeamException();
    }

    /**
     * Switch the current thread to the previous team context.
     */
    public void returnFromOwner()
    {
	TeamBackEnd.returnFrom();
    }

    /**
     * Push a TeamReservation for this objects owner team.
     *
     * @return A TeamReservation referring to the owner.
     * @exception A DeadTeamException if one of the teams involved is dead.
     */
    public TeamReservation pushOwnerReservation()
	throws DeadTeamException
    {
	return this.owner.pushReservation(1);
    }

    public String toString()
    {
	return "Exportable[owner="
	    + this.owner
	    + "; anonManager="
	    + this.anonManager
	    + "]";
    }
}
