/*
 * Importable.java
 * Base class for objects imported into teams.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */
package edu.utah.janosvm.sys;

/**
 * An Importable is a safe, cross-process reference to a remote
 * "Exportable" object.
 * <p>
 * Methods on Importable and subclasses of Importable may store
 * pointers in a remote process.  Beware that getting these operations
 * wrong can create bogus pointers into remote heaps.  Such problems
 * generally manifest as the VM crashing.
 * <p>
 * Example of a 'Foo' and 'FooRef'.
 * <pre>
 * // The "server" object that other Teams will have references to
 * class FooBackEnd extends Exportable {
 *   ...
 * }
 *
 * // The "client" object that Teams have locally as a reference to the
 * // "server"s Foo.
 * class FooHandle extends Importable {
 *   ...
 *   private FooBackEnd be;
 *
 *   private FooHandle() { } 
 *
 *   protected void bindTo(Exportable ex)
 *   {
 *     this.be = (FooBackEnd)ex;
 *   }
 *
 *   protected void unbindFrom()
 *   {
 *     this.be = null;
 *   }
 *
 *   // Lookup a FooBackEnd in the given team.
 *   public static FooHandle lookupFoo(Team team, Object id) {
 *     FooBackEnd be;
 *     FooHandle fh = new FooHandle();
 *     
 *     team.importObject(fh, id);
 *     return fh;
 *   }
 *
 *   // Create an anonymous FooBackEnd and import it into the current team.
 *   public static FooHandle newFoo(Team team) {
 *     FooHandle retval = new FooHandle();
 *
 *     team.switchTo();
 *     {
 *       FooBackEnd be = new FooBackEnd();
 *       ExportManager em;
 *
 *       em = Team.current().exportAnonymousObject(be);
 *       em.addImporter(retval);
 *     }
 *     team.returnFrom();
 *     return retval;
 *   }
 *
 *   // perform an operation via the foo handle on the foo back end
 *   public void performOp(...) {
 *     synchronized(this) {
 *       // do some stuff in the context of current team
 *       be.doSomethingHere(...);
 *
 *       // do some stuff in the context of the owner
 *       be.switchToOwner();
 *       be.doSomethingThere(...);
 *       be.returnFromOwner();
 *     }
 *   }
 *   ...
 * }
 * </pre>
 *
 * In the above example, a <code>FooHandle</code> is bound to a particular
 * Team's <code>FooBackEnd</code> object by looking the object up with some
 * sort of key.  Operations on the remote foo are perfomed by invoking
 * <code>performOp()</code> on the FooRef.  Internally this function visits the
 * owner of the actual Foo, and performs the operation.
 * <p>
 * Note that all of the ops are done in a synchronized block on the importable.
 * This will prevent the remote team from revoking the FooHandle while the
 * operation is being performed on it (unless the operation in question is to
 * destroy the Foo... be careful).
 *
 * @see TeamHandle#switchTo()
 * @see Exportable
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public abstract class Importable
    extends ListNode
{
    /**
     * The manager of the export we're importing.
     */
    private ExportManager em;
    
    /**
     * Construct an empty Importable.
     */
    public Importable()
    {
	if( !PortAuthority.register(this) )
	    throw new InstantiationError();
    }

    /**
     * Store the ExportManager for the object that is being imported.
     *
     * @param ex The ExportManager that manages the object being imported.
     */
    void setExportManager(ExportManager em)
    {
	this.em = em;
    }

    /*
     * @return The ExportManager that manages the object being imported, or
     * NULL if this object is importing anything.
     */
    ExportManager getExportManager()
    {
	return this.em;
    }

    /**
     * Bind this Importable to the given Exportable.  The most common
     * implementation of this method would be to cast the Exportable to the
     * type that the Importable subclass handles and store it in the object as
     * well as any other handy pointers/values.
     *
     * @param ex The Exportable that is being imported.
     */
    protected abstract void bindTo(Exportable ex);

    /**
     * Unbind this Importable from any objects its referencing.  The most
     * common implementation of this method would be null out any object
     * references in this object.
     */
    protected abstract void unbindFrom();

    /**
     * Copy the reference of the given Importable.
     *
     * @param im The Importable to copy the reference from.
     */
    protected void copy(Importable im)
    {
	synchronized(im)
	{
	    if( im.em != null )
		im.em.addImporter(this);
	}
    }

    public void revoke()
    {
	/* We're still holding a reference, cancel it. */
	PortAuthority pa = Team.current().getPortAuthority();
	
	pa.cancelImport(this);
    }
    
    protected void finalize()
    {
	if( this.em != null )
	{
	    /* We're still holding a reference, cancel it. */
	    PortAuthority pa = Team.current().getPortAuthority();

	    pa.cancelImport(this);
	}
	PortAuthority.unregister(this);
    }

    public String toString()
    {
	return "Importable["
	    + super.toString()
	    + "; em="
	    + this.em
	    + "]";
    }
}
