/*
 * PortAuthority.java
 * Class for managing the imports and exports of a team.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * Manages the imports and exports of a team.
 */
public final class PortAuthority
{
    /**
     * The table of named exports for a team.
     */
    private Hashtable exports = new Hashtable();

    /**
     * The list of anonymous exports for a team.
     */
    private Vector anonymousExports = new Vector();

    /**
     * Revoke all of a teams exports.
     */
    synchronized void revokeAllExports()
    {
	Hashtable table;
	Enumeration e;
	int lpc, size;
	Vector anon;
	
	synchronized(this)
	{
	    /*
	     * Get references to the tables and null them out so they can't
	     * be used anymore.
	     */
	    table = this.exports;
	    this.exports = null;

	    anon = this.anonymousExports;
	    this.anonymousExports = null;
	}
	if( table != null )
	{
	    /*
	     * Just walk the list of Exports and revoke them like you would
	     * do them individually.
	     */
	    e = table.elements();
	    size = table.size();
	    for( lpc = 0; lpc < size; lpc++ )
	    {
		ExportManager em = (ExportManager)e.nextElement();

		try
		{
		    em.cancelImports();
		}
		catch(InternalError ie)
		{
		    VMInstance.panic("Internal Error in cancelImports", ie);
		}
		catch(Throwable th)
		{
		    try
		    {
			th.printStackTrace();
		    }
		    catch(Throwable th2)
		    {
			/* Bah, can't do anything about it... */
		    }
		}
	    }
	    size = anon.size();
	    for( lpc = 0; lpc < size; lpc++ )
	    {
		((ExportManager)anon.elementAt(lpc)).cancelImports();
	    }
	}
    }

    /**
     * Register the Importable internally so that we're revoked when the
     * team dies.
     */
    static native boolean register(Importable im);
    
    /**
     * Unregister the Importable internally.
     */
    static native void unregister(Importable im);

    /**
     * Revoke all of a team's imports.
     */
    static native void revokeAllImports();

    /**
     * Attempt to reserve an exported identifier for future use.  This is
     * useful if you want to delay the construction of the Exportable object
     * until you're sure that the identifier is available.
     *
     * @param id The identifier to reserve.
     * @return An empty ExportManager that should eventually be filled in with
     * the export.
     * @exception DuplicateExportException if the `id' is already used.
     */
    public synchronized ExportManager reserveExportID(Object id)
	throws DuplicateExportException
    {
	ExportManager retval;
	
	if( this.exports.get(id) == null )
	{
	    retval = new ExportManager();
	    this.exports.put(id, retval);
	}
	else
	{
	    throw new DuplicateExportException(id.toString());
	}
	return retval;
    }

    /**
     * Export an object with the given identifier.
     *
     * @param id The identifier to export the object under.
     * @param ex The object to export.
     * @return The ExportManager for this mapping of `id' to `ex'.
     * @exception DuplicateExportException if the `id' is already used.
     */
    public synchronized ExportManager exportObject(Object id, Exportable ex)
	throws DuplicateExportException
    {
	ExportManager retval;
	
	if( this.exports.get(id) == null )
	{
	    retval = new ExportManager(ex);
	    this.exports.put(id, retval);
	}
	else
	{
	    throw new DuplicateExportException(id.toString());
	}
	return retval;
    }

    /**
     * Revoke access to the export with given identifier.
     * <p>
     * XXX rename to revokeExport()
     *
     * @param id The identifier of the export to revoke.
     * @exception NoSuchExportException if there isn't an export with the given
     * identifier in the table.
     */
    public void revokeObject(Object id)
	throws NoSuchExportException
    {
	ExportManager em;
	
	synchronized(this)
	{
	    em = (ExportManager)this.exports.remove(id);
	}
	if( em != null )
	{
	    em.cancelImports();
	}
	else
	{
	    throw new NoSuchExportException(id.toString());
	}
    }

    /**
     * Export an object without a public identifier.
     *
     * @param ex The object to export.
     * @return The ExportManager for the newly exported object.
     */
    public synchronized ExportManager exportAnonymousObject(Exportable ex)
    {
	ExportManager retval = new ExportManager(ex);

	this.anonymousExports.add(retval);
	ex.setAnonymousManager(retval);
	return retval;
    }

    private void revokeAnonymousObject(Exportable ex, ExportManager em)
    {
	synchronized(this)
	{
	    ex.setAnonymousManager(null);
	    this.anonymousExports.remove(em);
	}
	em.cancelImports();
    }

    /**
     * Revoke an anonymously exported object.
     *
     * @param ex The object to revoke access to.
     * @exception NoSuchExportException if the given object isn't in the export
     * table.
     */
    public void revokeAnonymousObject(Exportable ex)
	throws NoSuchExportException
    {
	boolean done = false;
	int lpc, size;

	size = this.anonymousExports.size();
	for( lpc = 0; (lpc < size) && !done; lpc++ )
	{
	    ExportManager em = (ExportManager)this.anonymousExports.
		elementAt(lpc);

	    if( em.getExport() == ex )
	    {
		this.anonymousExports.remove(lpc);
		this.revokeAnonymousObject(ex, em);
		done = true;
	    }
	}
	if( !done )
	{
	    throw new NoSuchExportException(ex.toString());
	}
    }

    /**
     * Find an export with the given identifier.
     *
     * @param id The identifier associated with the ExportManager to look for.
     * @return The ExportManager associated with the given id or NULL if its
     * not in the table.
     */
    private synchronized ExportManager findExport(Object id)
    {
	ExportManager retval;

	retval = (ExportManager)this.exports.get(id);
	return retval;
    }

    /**
     * Import a publicly available object from another team.
     *
     * @param to The Importable that should be set to reference the export.
     * @param from Reference to the team to import the object from.
     * @param id The identifier of the object to import from the given team.
     * @exception DeadTeamException if the team couldn't be reached.
     * @exception TeamEjectionException if the thread was thrown out of the
     * team during processing.
     * @exception NoSuchExportException if the object doesn't exist in the
     * other team.
     */
    public void importObject(Importable to, TeamHandle from, Object id)
	throws DeadTeamException,
	       TeamEjectionException,
	       NoSuchExportException
    {
	if( to.getExportManager() != null )
	{
	    this.cancelImport(to);
	}
	{
	    TeamReservation tr = from.pushReservation();
	    ExportManager em;

	    /* Ask the other team's PortAuthority for the object. */
	    tr.visit();
	    {
		PortAuthority pa = Team.current().getPortAuthority();

		em = pa.findExport(id);
	    }
	    tr.leave();
	    try
	    {
		em.addImporter(to);
	    }
	    catch(NullPointerException e)
	    {
		/* Didn't find it in the table. */
		throw new NoSuchExportException(id.toString());
	    }
	    tr.pop();
	}
    }

    /**
     * Break the reference of the given Importable.
     *
     * @param im The Importable that should have its reference cleared.
     */
    public void cancelImport(Importable im)
    {
	ExportManager em;
	
	synchronized(im)
	{
	    /* Get a pointer to the manager object */
	    em = im.getExportManager();
	}
	if( em != null )
	{
	    try
	    {
		TeamReservation tr;
		TeamBackEnd be;
		Exportable ex;
		
		be = em.getOwner();
		tr = be.pushReservation();
		{
		    ex = em.getExport();
		    if( em.remImporter(im) )
		    {
			tr.visit();
			{
			    PortAuthority pa = Team.current().
				getPortAuthority();
			    
			    pa.revokeAnonymousObject(ex, em);
			}
			tr.leave();
		    }
		}
		tr.pop();
	    }
	    catch(DeadTeamException e)
	    {
		/*
		 * Can't reserve the team, oh well, they should've revoked us.
		 */
		if( im.getExportManager() != null )
		    VMInstance.panic("We weren't revoked by the other guy...");
	    }
	    catch(TeamEjectionException e)
	    {
		/* XXX Don't know what to do here... */
		VMInstance.panic("Caught exception in cancelImport:", e);
	    }
	    catch(Throwable th)
	    {
		VMInstance.panic("Caught exception in cancelImport:", th);
	    }
	}
    }

    public String toString()
    {
	return "PortAuthority[exports="
	    + this.exports
	    + "; anonymousExports="
	    + this.anonymousExports
	    + "]";
    }
}
