/*
 * Team.java
 * Peer object for a team.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

import java.util.Vector;

import edu.utah.janosvm.resources.ResourceSpec;
import edu.utah.janosvm.resources.ResourceSpecException;
import edu.utah.janosvm.resources.DuplicateResourceSpecException;

/**
 * Peer object for a TeamBackEnd.  An instance of a Team
 * is created inside each team to contain that team's local
 * data.  Specifically, this class keeps track of the cross-team
 * references
 * coming from its team.  Team objects are never created
 * directly, they are created implicitly as part of 'Team.create()'.
 * <p>
 * The association with the TeamBackEnd in the kernel team is implicit
 * (through TeamBackEnd.getCurrentTeam()).
 * <p>
 * Because this class extends TeamHandle, it is importable.
 *
 * @see TeamBackEnd
 * @see TeamHandle
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public final class Team
    extends TeamHandle
{
    private static TeamHandle create(String name,
				     ResourceSpec[] specs,
				     kaffe.util.Ptr nativeSpecs)
	throws DuplicateExportException,
	       ResourceSpecException,
	       DeadTeamException
    {
	TeamBackEnd kernel = TeamBackEnd.getKernelTeam();
	TeamHandle retval = null;
	TeamReservation tr;

	try
	{
	    ExportManager em = null;
	    
	    tr = kernel.pushReservation();
	    tr.visit();
	    {
		PortAuthority pa = kernel.getPeer().getPortAuthority();
		
		if( (name != null) && !name.equals("") )
		{
		    name = CloneFactory.clone(name);
		    em = pa.reserveExportID(name);
		}

		TeamBackEnd be;

		try
		{
		    be = new TeamBackEnd(name, specs, nativeSpecs);
		}
		catch(Throwable th)
		{
		    if( em != null )
		    {
			try
			{
			    pa.revokeObject(name);
			}
			catch(NoSuchExportException nsee)
			{
			}
		    }
		    if( th instanceof IndexedResourceSpecException )
		    {
			throw (IndexedResourceSpecException)th;
		    }
		    else if( th instanceof DuplicateResourceSpecException )
		    {
			throw (DuplicateResourceSpecException)th;
		    }
		    else if( th instanceof DeadTeamException )
		    {
			throw (DeadTeamException)th;
		    }
		    else if( th instanceof Error )
		    {
			throw (Error)th;
		    }
		    else if( th instanceof RuntimeException )
		    {
			throw (RuntimeException)th;
		    }
		    else
		    {
			th.printStackTrace();
			throw new InternalError();
		    }
		}
		
		Team t = be.getPeer();
		
		if( em == null )
		    em = pa.exportAnonymousObject(be);
		em.setExport(be);
		em.addImporter(t);
		TeamBackEnd.teamList.event(t, TeamEvent.CREATED);
	    }
	    tr.leave();
	    retval = new TeamHandle();
	    em.addImporter(retval);
	    tr.pop();
	}
	catch(TeamEjectionException e)
	{
	}
	catch(DeadTeamException e)
	{
	    retval = null;
	}
	catch(IndexedResourceSpecException e)
	{
	    int index;

	    index = e.getIndex();
	    if( (specs == null) || index >= specs.length )
		throw new ResourceSpecException();
	    else
		throw new ResourceSpecException(specs[e.getIndex()]);
	}
	return retval;
    }

    public static TeamHandle create(String name, ResourceSpec[] specs)
	throws DuplicateExportException,
	       ResourceSpecException,
	       DeadTeamException
    {
	return Team.create(name, specs, null);
    }

    public static TeamHandle create(String name)
	throws DuplicateExportException,
	       ResourceSpecException,
	       DeadTeamException
    {
	return Team.create(name, null);
    }

    public static TeamHandle create(ResourceSpec[] specs)
	throws ResourceSpecException,
	       DeadTeamException
    {
	try
	{
	    return Team.create(null, specs);
	}
	catch(DuplicateExportException e)
	{
	    VMInstance.panic("Duplicate export for an anonymous object?", e);
	    return null;
	}
    }
    
    public static TeamHandle create()
	throws ResourceSpecException,
	       DeadTeamException
    {
	try
	{
	    return Team.create(null, null);
	}
	catch(DuplicateExportException e)
	{
	    VMInstance.panic("Duplicate export for an anonymous object?", e);
	    return null;
	}
    }
    
    /**
     * The manager of any cross team references.
     */
    private final PortAuthority pa = new PortAuthority();

    /**
     * The vector of data that is local to this team.  This
     * vector is not used by the JanosVM.  The protocol for
     * using this vector is up to the system hosted by the JanosVM.
     */
    private final Vector localData = new Vector();

    /**
     * The cached kernel team handle.
     */
    private final TeamHandle kernel = new TeamHandle();

    /**
     * A description of the contents of the team.
     */
    private String description;
    
    /**
     * @return The Team object for the current team.
     */
    public static Team current()
    {
	return TeamBackEnd.getCurrentTeam().getPeer();
    }

    /**
     * Get a handle on the Team with the given name.  TeamHandles with the null
     * name cannot be found this way.
     *
     * @param name The name of the TeamHandle to locate.
     * @return A TeamHandle object that imports the TeamBackEnd with
     * the given name.
     */
    public static TeamHandle find(String name)
	throws DeadTeamException,
	       TeamEjectionException,
	       NoSuchExportException
    {
	TeamHandle retval = null;

	if( (name != null) && !name.equals("") )
	{
	    Team t = Team.current();
	    PortAuthority pa = t.getPortAuthority();
	    TeamHandle th = new TeamHandle();

	    pa.importObject(th, t.getKernelTeam(), name);
	    retval = th;
	}
	else
	{
	    throw new NoSuchExportException("null or empty name");
	}
	return( retval );
    }

    Team()
    {
    }
    
    /**
     * Construct a Team object.
     * <p>
     * NOTE: For kernel use only; there is only one of these objects
     * per team.
     */
    Team(ExportManager kernelManager)
    {
	kernelManager.addImporter(this.kernel);
    }

    void setKernelHandle(ExportManager kernelManager)
    {
	kernelManager.addImporter(this.kernel);
    }

    /**
     * @param str A string that describes this team.
     */
    public synchronized void setDescription(String str)
    {
	this.description = CloneFactory.clone(str);
	this.be.setDescription(this.description);
    }

    /**
     * @return A string that describes this team.
     */
    public synchronized String getDescription()
    {
	return this.description;
    }
    
    /**
     * Get a handle on the kernel team.
     *
     * @return The cached kernel Team handle.
     */
    public TeamHandle getKernelTeam()
    {
	return this.kernel;
    }

    /**
     * Get the ExportTable for the current team.
     *
     * @return The ExportTable for the current team.
     */
    public PortAuthority getPortAuthority()
    {
	return this.pa;
    }

    public ExportManager reserveExportID(Object id)
	throws DuplicateExportException
    {
	return this.pa.reserveExportID(id);
    }
    
    public ExportManager exportObject(Object id, Exportable ex)
	throws DuplicateExportException
    {
	return this.pa.exportObject(id, ex);
    }

    /**
     * XXX rename to revokeExport
     */
    public void revokeObject(Object id)
	throws NoSuchExportException
    {
	this.pa.revokeObject(id);
    }
    
    public ExportManager exportAnonymousObject(Exportable ex)
    {
	return this.pa.exportAnonymousObject(ex);
    }
    
    public void revokeAnonymousObject(Exportable ex)
	throws NoSuchExportException
    {
	this.pa.revokeAnonymousObject(ex);
    }
    
    /**
     * @return The local data vector for the current team.
     */
    public Vector getData()
    {
	return this.localData;
    }

    /**
     * Store a per-team class local object.
     *
     * @param cl The class the object should be stored with.
     * @param obj The object to store with the class.
     */
    public void setClassLocal(Class cl, Object obj)
    {
	if( this.be != null )
	    this.be.setClassLocal(cl, obj);
	else
	    throw new NullPointerException();
    }

    public static void setCurrentClassLocal(Class cl, Object obj)
    {
	TeamBackEnd.setCurrentClassLocal(cl, obj);
    }
    
    /**
     * @param cl The class to retrieve the local object for.
     * @return The Object associated with the class or null.
     */
    public Object getClassLocal(Class cl)
    {
	if( this.be != null )
	    return this.be.getClassLocal(cl);
	else
	    throw new NullPointerException();
    }

    public static Object getCurrentClassLocal(Class cl)
    {
	return TeamBackEnd.getCurrentClassLocal(cl);
    }

    /**
     * Add a dependency to this team.
     *
     * @param th The team this one should depend on.
     * @return True if the addition succeeded, false otherwise.
     */
    public synchronized boolean addDependency(TeamHandle th)
    {
	boolean retval = false;

	if( this.be != null )
	{
	    retval = this.be.addDependency(th.be);
	}
	return retval;
    }

    /**
     * Set the flag on the team object that indicates that the last non-daemon
     * thread in the team should send out a TeamEvent.FINISHED event.
     */
    public synchronized void notifyOnFinish()
    {
	if( this.be != null )
	{
	    this.be.notifyOnFinish();
	}
    }

    public String toString()
    {
	return "Team["
	    + super.toString()
	    + "; pa="
	    + this.pa
	    + "; localData="
	    + this.localData
	    + "; kernel="
	    + this.kernel
	    + "; description="
	    + this.description
	    + "]";
    }
}
