/*
 * TeamHandle.java
 * A handle for accessing TeamBackEnd objects
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

import java.util.Vector;
import edu.utah.janosvm.resources.ResourceSpec;
import edu.utah.janosvm.resources.ResourceSpecException;

/**
 * Provides a safe handle to a Team.  Technically, this points to the
 * kernel team's TeamBackEnd object that represents the team in
 * question. 
 *
 * @see TeamBackEnd
 * @see Team
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class TeamHandle
    extends Importable
    implements Cloneable
{
    protected TeamBackEnd be;

    public TeamHandle()
    {
    }
    
    protected void bindTo(Exportable ex)
    {
	this.be = (TeamBackEnd)ex;
    }

    protected void unbindFrom()
    {
	this.be = null;
    }

    public boolean equals(Object obj)
    {
	boolean retval = false;
	
	if( obj instanceof TeamHandle )
	{
	    TeamHandle th = (TeamHandle)obj;

	    if( th.be == this.be )
		retval = true;
	}
	return retval;
    }
    
    /**
     * Get the name of the TeamBackEnd associated with this
     * TeamHandle handle.
     *
     * @return A copy of the TeamHandle's name.
     */
    public synchronized String getName()
    {
	String retval = null;

	if( this.be != null )
	{
	    String name;
	    
	    if( (name = this.be.getName()) != null )
		retval = CloneFactory.clone(name);
	    else
		retval = "(none)";
	}
	return retval;
    }

    /**
     * Get the Team's description string.
     *
     * @return A copy of the Team's description.
     */
    public synchronized String getDescription()
    {
	String retval = null;

	if( this.be != null )
	{
	    String desc;

	    if( (desc = this.be.getPeer().getDescription()) != null )
		retval = CloneFactory.clone(desc);
	}
	return retval;
    }

    /**
     * @return The unique integer identifier for this TeamHandle.
     */
    public synchronized int getID()
    {
	int retval = 0;

	if( this.be != null )
	{
	    retval = this.be.getID();
	}
	return retval;
    }

    /**
     * Test if the given object is owned by this TeamHandle.
     *
     * @param obj The object to test for ownership.
     * @return True if this TeamHandle owns the object.
     */
    public boolean owns(Object obj)
    {
	if( this.be != null )
	    return this.be.owns(obj);
	else
	    return false;
    }

    /**
     * Terminate the team referenced by this handle.  
     * This will clean up all
     * the importables and exportables owned by the
     * team and schedule all of its threads for
     * destruction.
     * <p>
     * If the calling thread is owned by the team being
     * terminated, this function will not return.
     * <p>
     * Active visits to the team being terminated will
     * delay termination until all the visits have been completed.
     *
     * @see TeamBackEnd#terminate()
     */
    public synchronized void terminate()
    {
	if( this.be != null )
	{
	    this.be.terminate();
	    this.be = null;
	}
    }

    public synchronized boolean isTerminated()
    {
	boolean retval;

	if( this.be == null )
	    retval = true;
	else
	    retval = false;
	return retval;
    }
    
    public void importObject(Importable to, Object id)
	throws DeadTeamException,
	       TeamEjectionException,
	       NoSuchExportException
    {
	Team team = Team.current();

	team.getPortAuthority().importObject(to, this, id);
    }
    
    public void cancelImport(Importable im)
    {
	Team team = Team.current();

	team.getPortAuthority().cancelImport(im);
    }
    
    /**
     * Create a new reservation that reserves the TeamHandle this object
     * points to.  The reservation remains "active" until
     * <code>popReservation()</code> is invoked by the current thread.
     * Unlike a switchTo(), this call does not implicitly switch context
     * to the other TeamHandle.  That is done by invoking use on the
     * TeamReservation instance returned by this call.
     * 
     * @return A TeamReservation object that reserves this TeamHandle.
     * @see #switchTo()
     * @see TeamBackEnd#pushReservation(int)
     * @see TeamReservation#visit()
     * @see TeamReservation#leave()
     */
    public synchronized TeamReservation pushReservation()
	throws DeadTeamException
    {
	TeamReservation retval = null;

	if( this.be != null )
	    retval = this.be.pushReservation(1);
	if( retval == null )
	    throw new DeadTeamException();
	return retval;
    }

    /**
     * Start a visit to the TeamHandle this object points to.  While the
     * visit is active, neither TeamHandle can be terminated, so object
     * references in either TeamHandle can be manipulated.  Write-barriers
     * are also disabled during this sequence.  The visit ends
     * when <code>returnFromTeam()</code> is invoked.
     *
     * @see #returnFrom()
     * @see TeamBackEnd#switchTo(int)
     */
    public void switchTo()
	throws DeadTeamException, TeamEjectionException
    {
	if( this.be == null )
	    throw new DeadTeamException();
	if( !this.be.switchTo(1) )
	    throw new DeadTeamException();
    }

    /**
     * End a visit to this TeamHandle.
     *
     * @see #switchTo()
     * @see TeamBackEnd#returnFrom()
     */
    public static void returnFrom()
    {
	TeamBackEnd.returnFrom();
    }
    
    /**
     * Pop the last TeamReservation off the stack.
     */
    public static void popReservation()
    {
	TeamBackEnd.popReservation();
    }

    public Object clone()
	throws CloneNotSupportedException
    {
	TeamHandle retval = new TeamHandle();

	retval.copy(this);
	return retval;
    }
    
    public String toString()
    {
	return "TeamHandle[name="
	    + this.getName()
	    + "; id="
	    + this.getID()
	    + "]";
    }
}
