/*
 * TeamReservation.java
 * Manages team reservation/migration.
 *
 * Copyright (c) 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.sys;

/**
 * Record a reservation with another team.  A TeamReservation is a link between
 * two teams that prevents either from being terminated.  A TeamReservation is
 * more flexible than TeamHandle.switchTo, as the termination-delaying time can
 * be expanded to include more than the context switch to the other team.
 * Here's an example:
 * <p>
 * NOTE: TeamReservation objects are a bit odd since they are
 * never explicitly created.  They are only implicitly created by the JanosVM
 * due to a call to <code>team.pushReservation()</code>.
 * 
 * <pre>
 * {
 *      TeamReservation act;
 *      ...
 *      act = otherTeam.pushReservation();
 *      if (act != null)
 *      {
 *		Object obj;
 *
 *		act.visit();
 *		// Do things in the context of 'otherTeam'
 *		obj = lookupSimpleFoo();
 *		act.leave();
 *
 *		// Do things in original context, but guarantee that
 *		// otherTeam will not go away ... Reservation still active. 
 *  
 *		Object localObj = new Foo(obj);
 *
 *		// End the Reservation, as we're done with all pointers into otherTeam
 *		Team.popReservation();
 *      }
 * }
 * </pre>
 *
 *
 * @see Team#pushReservation()
 * @see Team#popReservation()
 * @see TeamHandle#switchTo()
 * @author Tim Stack, <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a> */
public class TeamReservation
{
    /**
     * Hide the constructor.  No one can allocate a TeamReservation
     * the JVM manufactures them in special memory in response to
     * a <code>Team.pushReservation()</code>.
     *
     * @see Team#pushReservation()
     */
    private TeamReservation()
    {
    }

    /**
     * Switch the current thread to the team encoded in this Reservation.
     */
    public native void visit() throws TeamEjectionException;

    /**
     * Return to the previous team.
     */
    public static native void leave();

    /**
     * Pop the last TeamReservation off the stack.
     */
    public static void pop()
    {
	TeamBackEnd.popReservation();
    }

    public String toString()
    {
	return "TeamReservation["
	    + "]";
    }
}
