/*
 * Facsimile.java
 * Base class for simple cloneable messages that can be sent between teams.
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.amiga;

import edu.utah.janosvm.sys.ListNode;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.CrossTeamCloneable;
import edu.utah.janosvm.sys.TeamEjectionException;

/**
 * NOTE: Facsimile is used since the message will be copied into the other team
 * instead of just sharing like a Message.
 */
public class Facsimile
    extends ListNode
    implements CrossTeamCloneable, Cloneable
{
    /**
     * The FaxPort that is waiting to receive a reply.
     */
    private FaxPortHandle replyPort;
    
    /**
     * Indicates whether or not this is a reply message.
     */
    private boolean reply = false;
    
    public Facsimile()
    {
    }

    /**
     * Set the reply port for this fax.
     *
     * @param mp The port to expect replies on.
     */
    public void setReplyPort(FaxPortHandle mp)
    {
	this.replyPort = mp;
    }

    /**
     * Set the reply port for this fax.  Package access since we don't
     * want people to be able to change the reply port on a message they've
     * received.  XXX Eh? Do we really care?
     *
     * @param mp The port to expect replies on.
     */
    void setReplyPort(FaxPort mp)
    {
	this.replyPort = mp;
    }

    /**
     * @return The reply port for this fax.
     */
    public FaxPortHandle getReplyPort()
    {
	return this.replyPort;
    }

    /**
     * @return Whether or not this fax is a reply to another.
     */
    public boolean isReply()
    {
	return this.reply;
    }

    /**
     * Send this fax back as a reply.
     *
     * @exception Throws DeadTeamException if the destination port's owner team
     * is dead.
     */
    protected void reply()
	throws DeadTeamException, TeamEjectionException
    {
	this.reply = true;
	if( this.replyPort != null )
	{
	    FaxPortHandle port = this.replyPort;
	    
	    this.replyPort = null;
	    port.putFacsimile(this);
	}
	else
	{
	    throw new NoReplyPortError();
	}
    }
    
    public Object clone()
    {
	try
	{
	    Object retval = super.clone();
	    
	    if( this.replyPort != null )
	    {
		Facsimile fax = (Facsimile)retval;
		
		fax.replyPort = new FaxPortHandle(this.replyPort);
	    }
	    return retval;
	}
	catch(CloneNotSupportedException e)
	{
	    return null;
	}
    }

    public String toString()
    {
	return "Facsimile[replyPort="
	    + this.replyPort
	    + "; reply="
	    + this.reply
	    + "]";
    }
}
