/*
 * FaxPort.java
 * Local handle on a FaxPortBackEnd port.
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.amiga;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.Importable;
import edu.utah.janosvm.sys.PortAuthority;
import edu.utah.janosvm.sys.ExportManager;
import edu.utah.janosvm.sys.DuplicateExportException;

/**
 * Local handle on a FaxPortBackEnd port.
 */
public final class FaxPort
    extends FaxPortHandle
{
    /**
     * Construct a FaxPort and a FaxPortBackEnd with the given values.
     *
     * @param id The identifier to export the FaxPortBackEnd under.
     * @param ss The SignalSet that the FaxPortBackEnd should signal when it
     * receives a fax.
     * @param sn The SignalNumber that the FaxPortBackEnd should signal when
     * it receives a fax.
     * @exception Throws DuplicateExportException if there is already an
     * export with the given identifier.
     */
    public FaxPort(Object id, SignalSet ss, int sn)
	throws DuplicateExportException
    {
	PortAuthority pa = Team.current().getPortAuthority();
	FaxPortBackEnd be = new FaxPortBackEnd(ss, sn);
	ExportManager em = pa.exportObject(id, be);
	
	em.addImporter(this);
    }

    /**
     * Construct a FaxPort and a FaxPortBackEnd with the given values.
     *
     * @param id The identifier to export the FaxPortBackEnd under.
     * @param ss The SignalSet that the FaxPortBackEnd should signal when it
     * receives a fax.
     * @exception Throws DuplicateExportException if there is already an
     * export with the given identifier.
     */
    public FaxPort(Object id, SignalSet ss)
	throws DuplicateExportException
    {
	PortAuthority pa = Team.current().getPortAuthority();
	FaxPortBackEnd be = new FaxPortBackEnd(ss);
	ExportManager em = pa.exportObject(id, be);

	em.addImporter(this);
    }

    /**
     * Construct a FaxPort and a FaxPortBackEnd with the given values.
     *
     * @param ss The SignalSet that the FaxPortBackEnd should signal when it
     * receives a fax.
     * @param sn The SignalNumber that the FaxPortBackEnd should signal when
     * it receives a fax.
     */
    public FaxPort(SignalSet ss, int sn)
    {
	PortAuthority pa = Team.current().getPortAuthority();
	FaxPortBackEnd be = new FaxPortBackEnd(ss, sn);
	ExportManager em = pa.exportAnonymousObject(be);

	em.addImporter(this);
    }

    /**
     * Construct a FaxPort and a FaxPortBackEnd with the given values.
     *
     * @param ss The SignalSet that the FaxPortBackEnd should signal when it
     * receives a fax.
     */
    public FaxPort(SignalSet ss)
    {
	PortAuthority pa = Team.current().getPortAuthority();
	FaxPortBackEnd be = new FaxPortBackEnd(ss);
	ExportManager em = pa.exportAnonymousObject(be);
	
	em.addImporter(this);
    }

    /**
     * @return The signal number that will get signalled when a fax is
     * received.
     */
    public int getSignalNumber()
    {
	if( this.be != null )
	    return this.be.getSignalNumber();
	else
	    return -1;
    }

    /**
     * @return The Facsimile at the head of the queue or null if the queue
     * is empty.
     */
    public Facsimile getFacsimile()
    {
	if( this.be != null )
	    return this.be.getFacsimile();
	else
	    return null;
    }

    /**
     * Wait for and return the first Facsimile received.
     *
     * @return The first facsimile.
     */
    public Facsimile waitForFacsimile()
    {
	if( this.be != null )
	    return this.be.waitForFacsimile(0);
	else
	    return null;
    }

    /**
     * Wait for and return the first Facsimile received.
     *
     * @param timeout Milliseconds to wait before returning.
     * @return The first facsimile.
     */
    public Facsimile waitForFacsimile(long timeout)
    {
	if( this.be != null )
	    return this.be.waitForFacsimile(timeout);
	else
	    return null;
    }

    /**
     * Close this fax port and revoke any references to it.
     */
    public synchronized void close()
    {
	PortAuthority pa = Team.current().getPortAuthority();

	pa.cancelImport(this);
    }

    public String toString()
    {
	return "FaxPort["
	    + super.toString()
	    + "]";
    }
}
