/*
 * FaxPortBackEnd.java
 * An Exportable object that can receive Facsimile objects.
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.amiga;

import edu.utah.janosvm.sys.List;
import edu.utah.janosvm.sys.ListNode;
import edu.utah.janosvm.sys.Exportable;
import edu.utah.janosvm.sys.DuplicateExportException;

/**
 * An Exportable object that can receive Facsimile objects.
 */
final class FaxPortBackEnd
    extends Exportable
{
    /**
     * The queue of faxes received by this port.
     */
    private final List faxQueue = new List();

    /**
     * The SignalSet that should be signalled when a message is received.
     */
    private final SignalSet listener;

    /**
     * The SignalNumber that should be signalled when a message is received.
     */
    private final int signalNumber;

    /**
     * Construct a FaxPortBackEnd with the given values.
     *
     * @param id The identifier to export the FaxPortBackEnd under.
     * @param listener The SignalSet that the FaxPortBackEnd should signal when
     * it receives a fax.
     * @param signalNumber The SignalNumber that the FaxPortBackEnd should
     * signal when it receives a fax.
     * @exception Throws DuplicateExportException if there is already an
     * export with the given identifier.
     */
    public FaxPortBackEnd(SignalSet listener, int signalNumber)
    {
	this.listener = listener;
	this.signalNumber = signalNumber;
    }
	
    /**
     * Construct a FaxPortBackEnd with the given values.
     *
     * @param id The identifier to export the FaxPortBackEnd under.
     * @param listener The SignalSet that the FaxPortBackEnd should signal when
     * it receives a fax.
     * @exception Throws DuplicateExportException if there is already an
     * export with the given identifier.
     */
    public FaxPortBackEnd(SignalSet listener)
    {
	this(listener, listener.allocate());
    }

    /**
     * @return The signal number that will get signalled when a fax is
     * received.
     */
    public int getSignalNumber()
    {
	return this.signalNumber;
    }

    /**
     * @param fax The Facsimile to add to the queue.
     */
    public void putFacsimile(Facsimile fax)
    {
	synchronized( this.faxQueue )
	{
	    this.faxQueue.addTail(fax);
	}
	this.listener.signal(this.signalNumber);
    }

    /**
     * @return The Facsimile at the head of the queue or null if the queue
     * is empty.
     */
    public Facsimile getFacsimile()
    {
	Facsimile retval;
	
	synchronized( this.faxQueue )
	{
	    retval = (Facsimile)this.faxQueue.remHead();
	}
	return retval;
    }

    /**
     * Wait for and return the first Facsimile received.
     *
     * @return The first facsimile.
     */
    public Facsimile waitForFacsimile(long timeout)
    {
	boolean done = false;
	Facsimile retval;
	
	while( ((retval = this.getFacsimile()) == null) && !done )
	{
	    if( this.listener.waitFor(1L << this.signalNumber, timeout) == 0 )
		done = true;
	    /*
	     * This will loop around if we immediately picked up the message
	     * instead of waiting for a signal on a previous call.
	     */
	}
	return retval;
    }

    protected void finalize()
    {
	this.listener.release(this.signalNumber);
    }

    public String toString()
    {
	return "FaxPortBackEnd[faxQueue={"
	    + this.faxQueue
	    + "}; listener="
	    + this.listener
	    + "; signalNumber="
	    + this.signalNumber
	    + "]";
    }
}
