/*
 * FaxPortHandle.java
 * A handle for sending faxes to other teams.
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.amiga;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.ListNode;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.Importable;
import edu.utah.janosvm.sys.Exportable;
import edu.utah.janosvm.sys.PortAuthority;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;
import edu.utah.janosvm.sys.NoSuchExportException;

/**
 * A handle for sending faxes to other teams.
 */
public class FaxPortHandle
    extends Importable
    implements Cloneable
{
    /**
     * The reference to the back end object.
     */
    protected FaxPortBackEnd be;

    /**
     * Construct an empty FaxPortHandle.
     */
    public FaxPortHandle()
    {
    }

    /**
     * Construct a copy of the given FaxPortHandle.
     *
     * @param fph The FaxPortHandle to copy.
     */
    FaxPortHandle(FaxPortHandle fph)
    {
	this.copy(fph);
    }

    protected void bindTo(Exportable ex)
    {
	this.be = (FaxPortBackEnd)ex;
    }

    protected void unbindFrom()
    {
	this.be = null;
    }
    
    /**
     * Construct a FaxPortHandle handle that refers to a specific export in a
     * team.
     *
     * @param team The team to import the FaxPortBackEnd from.
     * @param id The identifier the FaxPortBackEnd is exported under.
     * @exception Throws NoSuchExportException if the FaxPortBackEnd doesn't
     * exist.
     */
    public FaxPortHandle(TeamHandle team, Object id)
	throws DeadTeamException,
	       TeamEjectionException,
	       NoSuchExportException
    {
	PortAuthority pa = Team.current().getPortAuthority();
	
	pa.importObject(this, team, id);
    }

    /**
     * Put a fax on the FaxPortBackEnd's queue.  This is done by visiting
     * the team, cloning the Facsimile and then adding that to the queue.
     *
     * @param fax The facsimile to send to the other team.
     * @exception Throws DeadTeamException if the visit fails.
     */
    public synchronized void putFacsimile(final Facsimile fax)
	throws DeadTeamException, TeamEjectionException
    {
	if( this.be != null )
	{
	    this.be.switchToOwner();
	    {
		this.be.putFacsimile((Facsimile)fax.clone());
	    }
	    this.be.returnFromOwner();
	}
	else
	{
	    throw new DeadTeamException();
	}
    }

    /**
     * @return A clone of the current handle.
     */
    public final synchronized FaxPortHandle cloneHandle()
    {
	return new FaxPortHandle(this);
    }

    public String toString()
    {
	return "FaxPortHandle[be="
	    + this.be
	    + "]";
    }
}
