/*
 * MsgPort.java
 * Message ports for threads within the same team.
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.amiga;

import edu.utah.janosvm.sys.List;
import edu.utah.janosvm.sys.ListNode;

/**
 * A MsgPort is a simple communication mechanism for threads within the same
 * team.  
 */
public final class MsgPort
{
    /**
     * The queue of messages received by this port.
     */
    private final List msgQueue = new List();

    /**
     * The SignalSet that should be signalled when a message is received.
     */
    private final SignalSet listener;

    /**
     * The SignalNumber that should be signalled when a message is received.
     */
    private final int signalNumber;

    /**
     * Construct a MsgPort with the given values.
     *
     * @param listener The SignalSet that the RemoteFaxPort should signal when
     * it receives a fax.
     * @param signalNumber The SignalNumber that the RemoteFaxPort should
     * signal when it receives a fax.
     */
    public MsgPort(SignalSet listener, int signalNumber)
    {
	this.listener = listener;
	this.signalNumber = signalNumber;
    }

    /**
     * Construct a MsgPort with the given values.
     *
     * @param listener The SignalSet that the RemoteFaxPort should signal when
     * it receives a fax.
     */
    public MsgPort(SignalSet listener)
    {
	this(listener, listener.allocate());
    }

    /**
     * @return The signal number that will get signalled when a fax is
     * received.
     */
    public int getSignalNumber()
    {
	return this.signalNumber;
    }

    /**
     * @param msg The Message to add to the queue.  Note: The message isn't
     * copied so you mustn't change any values until you receive the reply.
     */
    public void putMessage(Message msg)
    {
	synchronized( this.msgQueue )
	{
	    this.msgQueue.addTail(msg);
	}
	this.listener.signal(this.signalNumber);
    }

    /**
     * @return The Message at the head of the queue or null if the queue
     * is empty.
     */
    public Message getMessage()
    {
	Message retval;
	
	synchronized( this.msgQueue )
	{
	    retval = (Message)this.msgQueue.remHead();
	}
	return retval;
    }

    /**
     * Wait for and return the first Message received.
     *
     * @return The first message.
     */
    public Message waitForMessage()
    {
	Message retval;

	while( (retval = this.getMessage()) == null )
	{
	    this.listener.waitFor(1L << this.signalNumber);
	    /*
	     * This will loop around if we immediately picked up the message
	     * instead of waiting for a signal.
	     */
	}
	return retval;
    }

    public String toString()
    {
	return "MsgPort[msgQueue="
	    + this.msgQueue
	    + "; listener="
	    + this.listener
	    + "; signalNumber="
	    + this.signalNumber
	    + "]";
    }
    
    protected void finalize()
	throws Throwable
    {
	this.listener.release(this.signalNumber);
	super.finalize();
    }
}
