/*
 * SignalDispatcher.java
 * Dispatches any signals received by a SignalSet.
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.amiga;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.List;
import edu.utah.janosvm.sys.ListNode;

/**
 * A SignalDispatcher provides some basic infrastructure for dispatching
 * signals received by a SignalSet.
 */
public class SignalDispatcher
    implements Runnable
{
    private static Class thisClass = SignalDispatcher.class;

    /**
     * Reserved signal used to cause the dispatcher to exit the run() method.
     */
    public static final int SIG_EXIT = 0;

    /**
     * Reserved signal used to indicate that there was a change in the signal
     * mask.
     */
    private static final int SIG_NEW_MASK = 1;

    /**
     * The SignalSet to handle signals for.
     */
    private final SignalSet ss = new SignalSet();

    /**
     * The list of objects that can handle certain signals.
     */
    private final List handlers = new List();

    /**
     * The set of signals to wait for.
     */
    private long waitMask;

    /**
     * Flag indicating whether the run method is currently running or not.
     */
    private boolean active = false;

    /**
     * Construct an empty SignalDispatcher.
     */
    public SignalDispatcher()
    {
	this.ss.allocate(SIG_EXIT);
	this.waitMask |= (1L << SIG_EXIT);
	this.ss.allocate(SIG_NEW_MASK);
	this.waitMask |= (1L << SIG_NEW_MASK);
    }

    /**
     * @return The SignalSet that this dispatcher listens on.
     */
    public SignalSet getSignalSet()
    {
	return this.ss;
    }

    /**
     * Add a SignalHandler to this dispatcher.
     *
     * @param sh An initialized SignalHandler that will handle a given signal.
     */
    public synchronized void addHandler(SignalHandler sh)
    {
	this.handlers.addHead(sh);
	this.waitMask |= (1L << sh.getSignalNumber());
	this.ss.signal(SIG_NEW_MASK);
    }

    /**
     * Remove a SignalHandler from the dispatcher.
     *
     * @param sh The SignalHandler to remove.
     */
    public synchronized void remHandler(SignalHandler sh)
    {
	sh.remove();
	this.waitMask &= ~(1L << sh.getSignalNumber());
    }

    /**
     * @return True if the run() method is running, false otherwise.
     */
    public boolean isActive()
    {
	return this.active;
    }
    
    public void run()
    {
	synchronized(this)
	{
	    if( !this.active )
		this.active = true;
	    else
		return;
	}
	try
	{
	    boolean done = false;

	    while( !done )
	    {
		long recvMask;

		recvMask = this.ss.waitFor(this.waitMask);
		if( (recvMask & (1L << SIG_EXIT)) != 0 )
		{
		    done = true;
		}
		else
		{
		    SignalHandler sh;
		    ListNode ln;
		    
		    synchronized(this)
		    {
			ln = handlers.head.succ;
			while( ln.succ != null )
			{
			    sh = (SignalHandler)ln;
			    if( (recvMask & (1L << sh.getSignalNumber()))
				!= 0 )
			    {
				try
				{
				    sh.run();
				}
				catch(Exception e)
				{
				    System.err.println("Caught exception in "
						       + "SignalHandler:"
						       + sh);
				    e.printStackTrace(System.err);
				}
			    }
			    ln = ln.succ;
			}
		    }
		}
	    }
	    this.active = false;
	}
	catch(Exception e)
	{
	    e.printStackTrace();
	}
    }

    /**
     * Send the SIG_EXIT signal to the dispatcher so that it will exit the
     * handler loop.
     */
    public void exit()
    {
	this.ss.signal(SIG_EXIT);
    }

    public String toString()
    {
	String retval = "SignalDispatcher[ss="
	    + this.ss
	    + "; waitMask="
	    + Long.toBinaryString(this.waitMask)
	    + "; handlers[]={";
	SignalHandler sh;
	ListNode ln;

	ln = this.handlers.head.succ;
	while( ln.succ != null )
	{
	    sh = (SignalHandler)ln;
	    retval += " " + sh + " ";
	    ln = ln.succ;
	}
	retval += "}]";
	return retval;
    }

    /**
     * @return The SignalDispatcher for this team.
     */
    public static final SignalDispatcher getTeamDispatcher()
    {
	Team team = Team.current();
	SignalDispatcher retval;
	
	synchronized(team)
	{
	    if( (retval = (SignalDispatcher)team.
		 getClassLocal(thisClass)) == null )
	    {
		retval = new SignalDispatcher();
		team.setClassLocal(thisClass, retval);
	    }
	}
	return retval;
    }
}
