/*
 * SignalSet.java
 * Maintains a set of signal bits for threads to communicate with.
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.amiga;

/**
 * A SignalSet is a JanosVM implementation of the Amiga's signal system.
 * Basically, it provides a simple, low level, low bandwidth mechanism for
 * threads to communicate with each other.
 */
public final class SignalSet
{
    public static final int MAXIMUM_SIGNAL_COUNT = 64;
    
    /**
     * Mask of the signals that are currently allocated.
     */
    private long allocatedSignals = 0;

    /**
     * Mask of signals that have been received so far.
     */
    private long receivedSignals = 0;

    /**
     * Mask of signals that the thread is waiting for at the moment.
     */
    private long waitSignals = 0;

    public SignalSet()
    {
    }

    /**
     * @return A newly allocated signal number.
     * @exception Throws an OutOfSignalsError if there are no more signal bits
     * available.
     */
    public synchronized int allocate()
    {
	int lpc, retval = -1;

	for( lpc = 0; (retval == -1) && (lpc < 64); lpc++ )
	{
	    if( ((1L << lpc) & this.allocatedSignals) == 0 )
	    {
		this.allocatedSignals |= (1L << lpc);
		retval = lpc;
	    }
	}
	if( retval == -1 )
	{
	    throw new OutOfSignalsError();
	}
	return retval;
    }

    /**
     * @param bit Allocate the given signal bit.
     */
    public synchronized void allocate(int bit)
    {
	if( (this.allocatedSignals & (1L << bit)) == 0 )
	{
	    this.allocatedSignals |= (1L << bit);
	}
	else
	{
	    /*
	     * What to do here:
	     *   1. throw an OutOfSignalsError
	     *   2. throw a different Exception/Error
	     *   3. return false
	     */
	    throw new OutOfSignalsError();
	}
    }

    /**
     * @param bit Release the allocated signal bit.
     */
    public synchronized void release(int bit)
    {
	this.allocatedSignals &= ~(1L << bit);
    }

    /**
     * Wait for one or more of the signals in the mask to be received, or
     * for a timeout.
     *
     * @param mask The mask of signals to wait on.
     * @param timeout The number of milliseconds to wait for signals.
     * @return The mask of signals received, which will be a subset of the
     * `mask' parameter.
     */
    public synchronized long waitFor(long mask, long timeout)
    {
	long retval = 0;

	if( (this.receivedSignals & mask) == 0 )
	{
	    try
	    {
		this.waitSignals = mask;
		this.wait(timeout);
		this.waitSignals = 0;
	    }
	    catch(InterruptedException e)
	    {
		/* XXX Miscommunication */
	    }
	}
	retval = this.receivedSignals & mask;
	this.receivedSignals &= ~mask;
	return retval;
    }

    /**
     * Wait for one or more of the signals in the mask to be received.
     *
     * @param mask The mask of signals to wait on.
     * @return The mask of signals received, which will be a subset of the
     * `mask' parameter.
     */
    public long waitFor(long mask)
    {
	return this.waitFor(mask, 0);
    }

    /**
     * @param bit The signal bit number to send.
     */
    public synchronized void signal(int bit)
    {
	long bitmask = (1L << bit);

	if( (this.allocatedSignals & bitmask) != 0 )
	{
	    this.receivedSignals |= bitmask;
	    if( (this.receivedSignals & this.waitSignals) != 0 )
		this.notify();
	}
	else
	{
	    throw new Error(); // XXX Better exception here please.
	}
    }

    public synchronized long setSignal(long newMask)
    {
	long retval = this.receivedSignals;

	this.receivedSignals = newMask;
	return retval;
    }

    public String toString()
    {
	return "SignalSet[allocatedSignals="
	    + Long.toBinaryString(this.allocatedSignals)
	    + "; receivedSignals="
	    + Long.toBinaryString(this.receivedSignals)
	    + "; waitSignals="
	    + Long.toBinaryString(this.waitSignals)
	    + "]";
    }
}
