/*
 * CommSpace.java
 *
 * Copyright (c) 1999, 2000, 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */
package edu.utah.janosvm.kit.comm.commspace;

import java.lang.ArrayIndexOutOfBoundsException;

import edu.utah.janosvm.sys.Exportable;
import edu.utah.janosvm.sys.TemporaryExport;

/**
 * Variation on an array, with some tuple-space like properties.
 * Objects are not shared between flows, they are copied on read.
 * <p>
 * XXX This still includes some jnodeos debug-isms such as the TRACE macros,
 * but these have been commented out.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
class CommSpace
    extends Exportable
    implements TemporaryExport
{
    /**
     * The table mapping indicies to elements.
     */
    private final CommSpaceElement[] table;

    /**
     * @param size The size of CommSpace to construct.
     */
    CommSpace(int size)
    {
	this.table = new CommSpaceElement[size];
    }

    /**
     * @return The size of the CommSpace.
     */
    final int size()
    {
	return this.table.length;
    }

    /**
     * @param index Index to verify that it is within the bounds of the
     * space.
     * @exception ArrayIndexOutOfBoundsException when the index is not
     * valid.
     */
    private final void checkIndex(int index)
    {
	if ((index >= this.table.length)
	    || (index < 0))
	    throw new ArrayIndexOutOfBoundsException("CommSpace table only has "
						     +size()+ " elements");
    }

    /**
     * Write the given element at the given index.
     */
    void write(int index, CommSpaceElement elem)
	throws ArrayIndexOutOfBoundsException
    {
	checkIndex(index);
	CommSpaceElement elemCopy = CommSpaceElement.copy(elem);

	// TRACE(Trace.NODEOS_COMMSPACE, "write(idx="+index+",elem="+elem+"; table="+this+")");
	synchronized(this.table)
	{
	    this.table[index] = elemCopy;
	    notifyOnIndex(index);
	}
    }

    CommSpaceElement take(int index)
	throws ArrayIndexOutOfBoundsException
    {
	CommSpaceElement e;

	// TRACE(Trace.NODEOS_COMMSPACE, "take(idx="+index+"; table="+this+")");
	checkIndex(index);
	synchronized(this.table)
	{
	    e = this.table[index];
	    this.table[index] = null;
	}
	return e;
    }

    CommSpaceElement waitToTake(int index)
	throws ArrayIndexOutOfBoundsException, InterruptedException
    {
	CommSpaceElement e;

	// TRACE(Trace.NODEOS_COMMSPACE, "waitToTake(idx="+index+"; table="+this+")");

	checkIndex(index);
	synchronized(this.table)
	{
	    e = this.table[index];
	    while (e == null)
	    {
		waitOnIndex(index);
		e = this.table[index];
	    }
	    this.table[index] = null;
	}
	return e;
    }

    CommSpaceElement read(int index)
	throws ArrayIndexOutOfBoundsException
    {
	CommSpaceElement e;

	// TRACE(Trace.NODEOS_COMMSPACE, "read(idx="+index+"; table="+this+")");

	checkIndex(index);
	synchronized(this.table)
	{
	    e = this.table[index];
	}
	return e;
    }

    CommSpaceElement swap(int index, CommSpaceElement elem)
    {
	CommSpaceElement e;
		
	// TRACE(Trace.NODEOS_COMMSPACE, "swap(idx="+index+",elem="+elem+"; table="+this+")");

	checkIndex(index);
	synchronized(this.table)
	{
	    e = this.table[index];
	    this.table[index] = CommSpaceElement.copy(elem);
	    notifyOnIndex(index);
	}
	return e;
    }

    // Assume there won't be a lot of contention/waiting in a
    // single CommSpace, and just wait/signal on the whole table,
    // not individual elements.

    /**
     * Notify potential interested parties about a change at the
     * given index.
     */
    private void notifyOnIndex(int index)
    {
	// table.assertSynchronizeLocked()
	// TRACE(Trace.NODEOS_COMMSPACE, " notifyOnIndex(idx="+index+"; table="+this+")");
	this.table.notifyAll();
    }

    /**
     * Wait for something to change in the slot at the
     * given index.
     */
    private void waitOnIndex(int index)
	throws InterruptedException
    {
	// table.assertSynchronizeLocked()
	// TRACE(Trace.NODEOS_COMMSPACE, " waitOnIndex(idx="+index+"; table="+this+")");
	this.table.wait();
	// TRACE(Trace.NODEOS_COMMSPACE, " WOKEN OnIndex(idx="+index+"; table="+this+")");
    }

    public boolean expired()
    {
	return true;
    }

    public String toString()
    {
	return "CommSpace[t=@"+Integer.toHexString(this.table.hashCode())+"]";
    }
}
