/*
 * CommSpaceElement.java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */
package edu.utah.janosvm.kit.comm.commspace;

import kaffe.util.Assert;

/**
 * Element in the CommSpace for communication.
 * This is what gets written to the CommSpace
 * <p>
 * XXX does this need to exist or can I just stick
 * Objects in the CommSpace?
 * <p>
 * XXX because the val field is final, its impossible to
 * re-use a CommSpaceElement (i.e., for a different object...)
 * Could work around this by making the val field read-only
 * while the object was lodged in the CommSpace...
 * <p>
 * XXX these are shared between processes.... i.e., they leak (~2 words).
 * <p>
 * "Big Deal." -- Chris H.
 * <p>
 * XXX Should the get methods be copying their contents?
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 */
public final class CommSpaceElement
	implements Cloneable
{
	private final Object val;
	private final int type;

	private static final int INT_TYPE = 1;
	private static final int CHAR_ARRAY_TYPE = 2;
	private static final int BYTE_ARRAY_TYPE = 3;
	private static final int COMM_SPACE_HANDLE_TYPE = 4;
	private static final int COMM_QUEUE_TYPE = 5;
	private static final int COMM_HASHTABLE_TYPE = 6;
	private static final int BOOLEAN_TYPE = 7;
	private static final int STRING_TYPE = 8;
	private static final int COMMSPACEABLE_TYPE = 9;
	private static final int LONG_TYPE = 10;

	private CommSpaceElement(Object val, int type)
	{
		if (val == null)
			throw new NullPointerException(); // XXX

		this.val = val;
		this.type = type;
	}

	CommSpaceElement(CommSpaceElement from)
	{
		this(from.val, from.type);
	}

	public CommSpaceElement(boolean val)
	{
		this(new Boolean(val), BOOLEAN_TYPE);
	}

	public CommSpaceElement(Boolean val)
	{
		this(val, BOOLEAN_TYPE);
	}

	public CommSpaceElement(int val)
	{
		this(new Integer(val), INT_TYPE);
	}

	public CommSpaceElement(Long val)
	{
		this(val, LONG_TYPE);
	}

	public CommSpaceElement(long val)
	{
		this(new Long(val), LONG_TYPE);
	}

	public CommSpaceElement(Integer val)
	{
		this(val, INT_TYPE);
	}

	public CommSpaceElement(char[] val)
	{
		this(val, CHAR_ARRAY_TYPE);
	}

	public CommSpaceElement(byte[] val)
	{
		this(val, BYTE_ARRAY_TYPE);
	}

	public CommSpaceElement(String val)
	{ 
		this(val, STRING_TYPE);
	}

//  	public CommSpaceElement(InChannelHandle val)
//  	{
//  		this(val, IN_CHANNEL_TYPE);
//  	}

//  	public CommSpaceElement(Address val)
//  	{
//  		this(val, ADDRESS_TYPE);
//  	}

//  	public CommSpaceElement(BufferHandle val)
//  	{
//  		this(val, BUFFER_HANDLE_TYPE);
//  	}

	public CommSpaceElement(CommSpaceHandle val)
	{
		this(val, COMM_SPACE_HANDLE_TYPE);
	}

	public CommSpaceElement(CommQueuePusher val)
	{
		this(val, COMM_QUEUE_TYPE);
	}

//  	public CommSpaceElement(CommMonitor val)
//  	{
//  		this(val, COMM_MONITOR_TYPE);
//  	}
	
	public CommSpaceElement(CommHashtableHandle val)
	{
		this(val, COMM_HASHTABLE_TYPE);
	}

	public CommSpaceElement(CommSpaceable val)
	{
		this(val, COMMSPACEABLE_TYPE);
	}

	/* XXX Why is this public but the ID's above are private? */
	public int type()
	{
		return type;
	}

	public int getInt()
	{
		return ((Integer) val).intValue();
	}

	public Integer getInteger()
	{
		// XXX is not copying this?!!
		return (Integer)val;
	}

	public long getLong()
	{
		return ((Long) val).longValue();
	}

	public Long getLongObject()
	{
		// XXX is not copying this?!!
		return (Long)val;
	}

	public boolean getBool()
	{
		return ((Boolean) val).booleanValue();
	}

	public Boolean getBoolean()
	{
		return ((Boolean) val);
	}

	public String getString()
	{
		//return CloneFactory.clone((String)val);
		return new String(((String)val).toCharArray());
	}

	public char[] getCharArray()
	{
		char[] from = (char[]) val;
		char[] to = new char[from.length];
		System.arraycopy(from, 0, to, 0, from.length);
		return to;
	}

	public void getCharArray(char[] outbuf, int index)
	{
		char[] from = (char[]) val;
		System.arraycopy(from, 0, outbuf, index, from.length);
	}

	public byte[] getByteArray()
	{
		byte[] from = (byte[]) val;
		byte[] to = new byte[from.length];
		System.arraycopy(from, 0, to, 0, from.length);
		return to;
	}

	public void getByteArray(byte[] outbuf, int index)
	{
		byte[] from = (byte[]) val;
		System.arraycopy(from, 0, outbuf, index, from.length);
	}

//  	public InChannelHandle getInChannelHandle()
//  	{
//  		InChannelHandle ich = new InChannelHandle();
//  		ich.copy((InChannelHandle) val);
//  		return ich;
//  	}

//  	public Address getAddress()
//  	{
//  		Address addr = (Address)val;
		
//  		return (Address)addr.clone();
//  	}

//  	public BufferHandle getBufferHandle()
//  	{
//  		return BufferHandle.create((BufferHandle) val);
//  	}

	public CommSpaceHandle getCommSpaceHandle()
	{
		return new CommSpaceHandle((CommSpaceHandle)val);
	}

	public CommQueuePusher getCommQueue()
	{
		return new CommQueuePusher((CommQueuePusher)val);
	}

//  	public CommMonitor getCommMonitor()
//  	{
//  		return (CommMonitor)val;
//  	}

	public CommHashtableHandle getCommHashtableHandle()
	{
		return new CommHashtableHandle((CommHashtableHandle)val);
	}

	public CommSpaceable getCommSpaceable()
	{
		CommSpaceable csa = (CommSpaceable)val;

		return (CommSpaceable)csa.deepCopy();
	}

	public String toString()
	{
		if (val==null)
			return new String("CommSpaceElement[null]");
		return "CommSpaceElement["+val.toString()+"]";
	}

	private static boolean compareByteArrays(byte b1[], byte b2[])
	{
		boolean retval = false;

		if( b1.length == b2.length )
		{
			int lpc;

			retval = true;
			for( lpc = 0; (lpc < b1.length) && retval; lpc++ )
			{
				if( b1[lpc] != b2[lpc] )
					retval = false;
			}
		}
		return retval;
	}

	private static boolean compareCharArrays(char c1[], char c2[])
	{
		boolean retval = false;

		if( c1.length == c2.length )
		{
			int lpc;

			retval = true;
			for( lpc = 0; (lpc < c1.length) && retval; lpc++ )
			{
				if( c1[lpc] != c2[lpc] )
					retval = false;
			}
		}
		return retval;
	}

	public boolean equals(Object obj)
	{
		boolean retval = false;

		if( obj instanceof CommSpaceElement )
		{
			CommSpaceElement cse = (CommSpaceElement)obj;

			if( cse.type == this.type )
			{
				switch( this.type )
				{
				case BYTE_ARRAY_TYPE:
					retval = compareByteArrays(
						(byte[])this.val,
						(byte[])cse.val);
					break;
				case CHAR_ARRAY_TYPE:
					retval = compareCharArrays(
						(char[])this.val,
						(char[])cse.val);
					break;
				default:
					retval = this.val.equals(cse.val);
					break;
				}
			}
		}
		return retval;
	}

	public int hashCode()
	{
		int retval;
		
		switch( this.type )
		{
		case BYTE_ARRAY_TYPE:
			byte barray[] = (byte[])this.val;
			
			switch( barray.length )
			{
			case 0:
				retval = 0;
				break;
			case 1:
				retval = (int)barray[0];
				break;
			case 2:
				retval = (int)(barray[0] | (barray[1] << 8));
				break;
			case 3:
				retval = (int)(barray[0]
					       | (barray[1] << 8)
					       | (barray[2] << 16));
				break;
			default:
				retval = (int)(barray[0]
					       | (barray[1] << 8)
					       | (barray[2] << 16)
					       | (barray[3] << 24));
				break;
			}
			break;
		case CHAR_ARRAY_TYPE:
			char carray[] = (char[])this.val;
			
			switch( carray.length )
			{
			case 0:
				retval = 0;
				break;
			case 1:
				retval = (int)carray[0];
				break;
			default:
				retval = (int)(carray[0] | (carray[1] << 16));
				break;
			}
			break;
		default:
			retval = this.val.hashCode();
			break;
		}
		return retval;
	}

	static CommSpaceElement copy(CommSpaceElement cse)
	{
		CommSpaceElement retval = null;

		if( cse != null )
			retval = (CommSpaceElement)cse.clone();
		return retval;
	}
	
	public Object clone()
	{
		CommSpaceElement retval = null;
		
		switch( this.type )
		{
		case INT_TYPE:
			retval = new CommSpaceElement(this.getInt());
			break;
		case BOOLEAN_TYPE:
			retval = new CommSpaceElement(this.getBoolean());
			break;
		case STRING_TYPE:
			retval = new CommSpaceElement(this.getString());
			break;
		case CHAR_ARRAY_TYPE:
			retval = new CommSpaceElement(this.getCharArray());
			break;
		case BYTE_ARRAY_TYPE:
			retval = new CommSpaceElement(this.getByteArray());
			break;
//  		case IN_CHANNEL_TYPE:
//  			retval = new CommSpaceElement(
//  				this.getInChannelHandle());
//  			break;
//  		case ADDRESS_TYPE:
//  			retval = new CommSpaceElement(this.getAddress());
//  			break;
//  		case BUFFER_HANDLE_TYPE:
//  			retval = new CommSpaceElement(this.getBufferHandle());
//  			break;
		case COMM_SPACE_HANDLE_TYPE:
			retval = new CommSpaceElement(
				this.getCommSpaceHandle());
			break;
		case COMM_QUEUE_TYPE:
			retval = new CommSpaceElement(this.getCommQueue());
			break;
//  		case COMM_MONITOR_TYPE:
//  			retval = new CommSpaceElement(this.getCommMonitor());
//  			break;
		case COMM_HASHTABLE_TYPE:
			retval = new CommSpaceElement(
				this.getCommHashtableHandle());
			break;
		case COMMSPACEABLE_TYPE:
			retval = new CommSpaceElement(
				this.getCommSpaceable());
			break;
		case LONG_TYPE:
			retval = new CommSpaceElement(this.getLong());
			break;
		default:
			// panic("Unknown type " +this.type+ " in CommSpaceElement.clone()");
			Assert.that(false,"Unknown type " +this.type+ " in CommSpaceElement.clone()");
			//break;
		}
		return retval;
	}
}
