/*
 * CommSpaceHandle.java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */
package edu.utah.janosvm.kit.comm.commspace;

import java.lang.ArrayIndexOutOfBoundsException;

import kaffe.util.Assert;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.Importable;
import edu.utah.janosvm.sys.Exportable;
import edu.utah.janosvm.sys.TeamReservation;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

/**
 * A handle on a CommSpace. A CommSpace is an inter-process communication
 * facility based lightly on Linda's "tuples".  The space consists of
 * an array of CommSpaceElements.  Elements are copied out on read.  A
 * thread may write, read, swap, "wait" or "take" any element.
 * <p>
 * XXX Semantics during a failed take() are probably wrong.
 * 
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Janos Ministry of Development</a>
 * @see CommSpace
 */
public final class CommSpaceHandle
	extends Importable
{
	private CommSpace cs;
	
	/**
	 * Create a new CommSpace of the given size and return a handle to
	 * it.
	 *
	 * @param size The size of the CommSpace.
	 * @return The handle to the CommSpace.
	 */
	public static CommSpaceHandle newCommSpace(int size)
	{
		CommSpaceHandle retval = new CommSpaceHandle();
		
		Team.current().
			exportAnonymousObject(new CommSpace(size)).
			addImporter(retval);
		return retval;
	}

	/**
	 * Create a CommSpace and attach it to the given handle.
	 *
	 * @param csh The CommSpaceHandle to attach to the newly created
	 * CommSpace.
	 * @param size The size of the CommSpace.
	 * @return The handle that was passed in.
	 */
	public static CommSpaceHandle newCommSpace(CommSpaceHandle csh,
						   int size)
	{
		Team.current().
			exportAnonymousObject(new CommSpace(size)).
			addImporter(csh);
		return csh;
	}

	/**
	 * Create an empty handle.
	 *
	 * @see #newCommSpace(CommSpaceHandle,int)
	 */
	public CommSpaceHandle()
	{
	}

	/**
	 * Copy constructor.  Only used when passing a
	 * csh to a child flow.
	 *
	 * @param other The handle to copy.
	 */
	public CommSpaceHandle(CommSpaceHandle other)
	{
		this.copy(other);
	}

	protected void bindTo(Exportable ex)
	{
		this.cs = (CommSpace)ex;
	}

	protected void unbindFrom()
	{
		this.cs = null;
	}

	/**
	 * XXX Without this being public there is no way to tell if this
	 * thing actually points to a commSpace.
	 */
	boolean empty()
	{
		return this.cs == null;
	}

	/**
	 * Return the number of slots in the CommSpace pointed to by
	 * this handle.
	 */
	public int size()
	{
		/* XXX this.get() could be null */
		return this.cs.size();
	}

	/**
	 * Resize the CommSpace pointed to by this handle to contain
	 * the given number of slots.
	 * <p>
	 * XXX what happens when resizing to smaller than current and
	 * dropping items?
	 */
	public int resize(int newsize)
	{
		//throw new InternalError( "resize()" ) ;
		Assert.that(false,"CommSpaceHandle.resize() is not implemented");

		// XXX only doable by allocator of underlying CommSpace object

		// watch out for shrinkage.
		// Can't really blame the compiler for not realizing the
		// assert won't return
		return 0;
	}

	/**
	 * Write the given element to the given index.  Overwrites
	 * whatever is already there.
	 *
	 * @param index The index in the CommSpace to write into.
	 * @param elem The element to copy into the CommSpace.
	 */
	public void write(final int index, final CommSpaceElement elem)
		throws ArrayIndexOutOfBoundsException
	{
		try
		{
			this.cs.switchToOwner();
			{
				this.cs.write(index, elem);
			}
			this.cs.returnFromOwner();
		}
		catch(TeamEjectionException dte)
		{
		}
		catch(DeadTeamException e)
		{
		}
	}
	
	/**
	 * Take an element from the given index and put a null in its place.
	 * The read and replace are atomic with respect to other take()s at the
	 * given index. XXX semantically equivalent to
	 * swap(<var>index</var>, <strong>null</strong>)
	 *
	 * @param index The index in the CommSpace that we should take.
	 * @see #swap(int,CommSpaceElement)
	 * @return null or the CommSpaceElement that was in the given index.
	 */
	public CommSpaceElement take(final int index)
		throws ArrayIndexOutOfBoundsException
	{
		CommSpaceElement retval = null;
		TeamReservation ta;

		try
		{
			ta = this.cs.pushOwnerReservation();
			ta.visit();
			{
				retval = this.cs.take(index);
			}
			ta.leave();
			if (retval != null)
				retval = (CommSpaceElement)retval.clone();
			ta.pop();
		}
		catch(TeamEjectionException dte)
		{
		}
		catch(DeadTeamException e)
		{
		}
		return retval;
	}
	
	/**
	 * Wait for an element to be placed into the given index.  Then
	 * atomically take the element and replace with null.
	 *
	 * @param index The index in the CommSpace that we should take.
	 * @return the CommSpaceElement that was in the given index.
	 */
	public CommSpaceElement waitToTake(final int index)
		throws InterruptedException, ArrayIndexOutOfBoundsException
	{
		CommSpaceElement retval = null;
		TeamReservation ta;
		
		try
		{
			ta = this.cs.pushOwnerReservation();
			ta.visit();
			{
				retval = this.cs.waitToTake(index);
			}
			ta.leave();
			if (retval != null)
				retval = (CommSpaceElement)retval.clone();
			ta.pop();
		}
		catch(TeamEjectionException dte)
		{
		}
		catch(DeadTeamException e)
		{
		}
		return retval;
	}
	
	/**
	 * Read the element at the given index.
	 *
	 * @param index The index in the CommSpace that we should take.
	 * @return Null or the CommSpaceElement that was in the given index.
	 */
	public CommSpaceElement read(final int index)
		throws ArrayIndexOutOfBoundsException
	{
		CommSpaceElement retval = null;
		TeamReservation ta;
		
		try
		{
			ta = this.cs.pushOwnerReservation();
			ta.visit();
			{
				retval = this.cs.read(index);
			}
			ta.leave();
			if (retval != null)
				retval = (CommSpaceElement)retval.clone();
			ta.pop();
		}
		catch(TeamEjectionException dte)
		{
		}
		catch(DeadTeamException e)
		{
		}
		return retval;
	}
	
	/**
	 * Swap an element with the one currently in the given index.
	 *
	 * @param index The index in the CommSpace that we should take.
	 * @param elem The element to swap into the index.
	 * @return Null or the CommSpaceElement that was in the given index.
	 */
	public CommSpaceElement swap(final int index,
				     final CommSpaceElement elem)
		throws ArrayIndexOutOfBoundsException
	{
		CommSpaceElement retval = null;
		TeamReservation ta;
		
		try
		{
			ta = this.cs.pushOwnerReservation();
			ta.visit();
			{
				retval = this.cs.swap(index, elem);
			}
			ta.leave();
			if (retval != null)
				retval = (CommSpaceElement)retval.clone();
			ta.pop();
		}
		catch(TeamEjectionException dte)
		{
		}
		catch(DeadTeamException e)
		{
		}
		return retval;
	}
}
