/*
 * JSIHandlerContext.java
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.jsi;

/**
 * A context used for processing regular and interim messages.  An interim
 * message is one that is sent by the handler in the course of processing a
 * request.
 */
public final class JSIHandlerContext
{
    /**
     * The default set of JSIHandlers.
     */
    private static final JSIHandlerRoot DEFAULT_ROOT_HANDLERS =
	new JSIHandlerRoot();

    /**
     * @return The default set of root handlers.
     */
    public static JSIHandlerRoot getDefaultRootHandlers()
    {
	return DEFAULT_ROOT_HANDLERS;
    }
    
    /**
     * Add a JSIHandler to the set of default roots.
     *
     * @param name The name of the root handler.
     * @param handler The handler itself.
     */
    public static void addDefaultRootHandler(String name, JSIHandler handler)
    {
	DEFAULT_ROOT_HANDLERS.addHandler(name, handler);
    }

    /**
     * The set of root handlers used to process messages.
     */
    private final JSIHandlerRoot rootHandlers;

    /**
     * The channel used to send interim messages.
     */
    private final JSIChannel jc;

    /**
     * Construct a context with the given values.
     *
     * @param rootHandlers The set of root handlers.
     * @param jc A connected channel.
     */
    public JSIHandlerContext(JSIHandlerRoot rootHandlers, JSIChannel jc)
    {
	this.rootHandlers = rootHandlers;
	this.jc = jc;
    }
    
    /**
     * Construct a context with the given values and the default set of root
     * handlers.
     *
     * @param jc A connected channel.
     */
    public JSIHandlerContext(JSIChannel jc)
    {
	this(DEFAULT_ROOT_HANDLERS, jc);
    }

    /**
     * @return The set of root handlers.
     */
    public JSIHandlerRoot getRootHandlers()
    {
	return this.rootHandlers;
    }

    /**
     * @return The channel used to send interim messages.
     */
    public JSIChannel getChannel()
    {
	return this.jc;
    }
    
    public String toString()
    {
	return "JSIHandlerContext[rootHandlers="
	    + this.rootHandlers
	    + "; jc="
	    + this.jc
	    + "]";
    }
}
