/*
 * JSIHandlerRoot.java
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.jsi;

import java.util.Hashtable;
import java.util.Enumeration;

/**
 * A class that manages the set of root JSI handlers.
 */
public class JSIHandlerRoot
    implements JSIHandler
{
    private static final boolean verbose = true;
    
    /**
     * A mapping from property name to JSIHandler.
     */
    private final Hashtable rootHandlers = new Hashtable();

    /**
     * Construct an empty root handler set.
     */
    public JSIHandlerRoot()
    {
    }
    
    /**
     * Add a handler with the given name.
     *
     * @param name The property name for the handler.
     * @param jh The handler.
     */
    public synchronized void addHandler(String name, JSIHandler jh)
    {
	if( name.length() == 0 )
	{
	    throw new IllegalArgumentException(
		"The handler name can't be an empty string.");
	}
	if( this.rootHandlers.containsKey(name) )
	{
	    throw new IllegalArgumentException("The handler name, "
					       + name
					       + ", is already in use.");
	}
	if( jh == null )
	    throw new IllegalArgumentException("JSIHandler can't be null.");
	this.rootHandlers.put(name, jh);
    }

    public synchronized JSIHandler getHandler(String name)
    {
	if( name.length() == 0 )
	{
	    throw new IllegalArgumentException(
		"The handler name can't be an empty string.");
	}
	return (JSIHandler)this.rootHandlers.get(name);
    }

    /**
     * Remove a handler with the given name.
     *
     * @param name The property name for the handler.
     */
    public synchronized JSIHandler remHandler(String name)
    {
	JSIHandler retval;
	
	if( name.length() == 0 )
	{
	    throw new IllegalArgumentException(
		"The handler name can't be an empty string.");
	}
	retval = (JSIHandler)this.rootHandlers.remove(name);
	return retval;
    }

    /**
     * Route a message to the appropriate destination.
     *
     * @param jhc The context to locate/execute the handler within.
     * @param jm The message to dispatch and contains the reply on return.
     */
    public void dispatchMessage(JSIHandlerContext jhc, JSIMessage jm)
    {
	try
	{
	    JSIPair jp;
	    
	    if( (jp = jm.popSpecifier()) != null )
	    {
		JSIHandler jh;

		if( (jh = this.resolveSpecifier(jm, jp)) != null )
		{
		    jh.handleMessage(jhc, jm);
		}
		else if( jm.getWhat() != JSIMessage.ERROR )
		{
		    jm.error("Unknown handler: " + jp.getTag());
		}
		else
		{
		    /* They filled out the error message. */
		}
	    }
	    else
	    {
		this.handleMessage(jhc, jm);
	    }
	}
	catch(Throwable th)
	{
	    if( verbose )
		th.printStackTrace();
	    if( th.getMessage() != null )
		jm.error("Exception: " + th.getMessage());
	    else
		jm.error("Exception: " + th.getClass().getName());
	}
    }

    public JSIHandler resolveSpecifier(JSIMessage jm, JSIPair jp)
	throws Throwable
    {
	JSIHandler jh, retval = null;

	synchronized( this )
	{
	    jh = (JSIHandler)this.rootHandlers.get(jp.getTag());
	}
	if( jh != null )
	{
	    retval = jh.resolveSpecifier(jm, jp);
	}
	return retval;
    }

    public void handleMessage(JSIHandlerContext jhc, JSIMessage jm)
    {
	switch( jm.getWhat() )
	{
	case JSIMessage.LIST_PROPERTIES:
	    if( jm.popSpecifier() == null )
	    {
		Enumeration keys;

		jm.clear();
		jm.setWhat(JSIMessage.REPLY);
		synchronized( this )
		{
		    keys = this.rootHandlers.keys();
		    while( keys.hasMoreElements() )
		    {
			String name = (String)keys.nextElement();
			
			jm.add("result", name);
		    }
		}
	    }
	    else
	    {
		jm.error("Too many specifiers");
	    }
	    break;
	case JSIMessage.GET_SUITES:
	    if( jm.popSpecifier() == null )
	    {
		jm.clear();
		jm.setWhat(JSIMessage.REPLY);
		this.getSuites(jm);
	    }
	    else
	    {
		jm.error("Too many specifiers");
	    }
	    break;
	default:
	    jm.error("Operation not supported on the root property.");
	    break;
	}
    }

    public void getSuites(JSIMessage jm)
    {
	Enumeration elements;

	synchronized( this )
	{
	    elements = this.rootHandlers.elements();
	    while( elements.hasMoreElements() )
	    {
		JSIHandler jh = (JSIHandler)elements.nextElement();
		
		jh.getSuites(jm);
	    }
	}
    }

    public String toString()
    {
	return "JSIHandlerRoot[rootHandlers="
	    + this.rootHandlers
	    + "]";
    }
}
