/*
 * JSIHandlerThread.java
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.jsi;

import edu.utah.janosvm.kit.comm.jsi.JSIChannel;
import edu.utah.janosvm.kit.comm.jsi.JSIMessage;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerRoot;
import edu.utah.janosvm.kit.comm.jsi.JSIHandlerContext;
import edu.utah.janosvm.kit.comm.jsi.JSIEncodingException;

/**
 * Runnable used to dispatch and reply to messages received over a given
 * JSIChannel.
 */
public abstract class JSIHandlerThread
    implements Runnable
{
    /**
     * The JSIChannel that should be used for all communication.
     */
    private final JSIChannel jc;

    /**
     * The JSIHandlerRoot that should be the root for dispatching messages.
     */
    private final JSIHandlerRoot jhr;

    /**
     * Construct a JSIHandlerThread that will process the messages received
     * from the given JSIChannel.
     *
     * @param jhr The JSIHandlerRoot that should be the root for dispatching
     * messages.
     * @param jc A connected JSIChannel.
     */
    JSIHandlerThread(JSIHandlerRoot jhr, JSIChannel jc)
    {
	this.jhr = jhr;
	this.jc = jc;
    }

    /**
     * Construct a JSIHandlerThread that will process the messages received
     * from the given JSIChannel.
     *
     * @param jc A connected JSIChannel.
     */
    JSIHandlerThread(JSIChannel jc)
    {
	this(JSIHandlerContext.getDefaultRootHandlers(), jc);
    }

    /**
     * @return The JSIHandlerRoot that should be the root for dispatching
     * messages.
     */
    JSIHandlerRoot getHandlerRoot()
    {
	return this.jhr;
    }

    /**
     * @return The JSIChannel that should be used for all communication.
     */
    JSIChannel getChannel()
    {
	return this.jc;
    }

    /**
     * Call back method used when a SHUTDOWN message is received.
     */
    public abstract void shutdown();

    public void run()
    {
	JSIHandlerContext jhc;
	byte msg[];

	jhc = new JSIHandlerContext(this.jhr, this.jc);
	try
	{
	    while( (msg = this.jc.recvMessage()) != null )
	    {
		JSIMessage jm;
		
		jm = JSIMessage.instantiate(msg);
		/* Act on the request... */
		switch( jm.getWhat() )
		{
		case JSIMessage.SHUTDOWN:
		    this.shutdown();
		    throw new ThreadDeath();
		default:
		    this.jhr.dispatchMessage(jhc, jm);
		    break;
		}
		/* Verify there is a reply */
		switch( jm.getWhat() )
		{
		case JSIMessage.REPLY:
		case JSIMessage.ERROR:
		    break;
		default:
		    jm.error("Internal error: Handler didn't create a reply.");
		    break;
		}
		msg = jm.flatten();
		this.jc.replyMessage(msg);
	    }
	}
	catch(ThreadDeath td)
	{
	    throw td;
	}
	catch(Throwable th)
	{
	    System.err.println("JSIHandlerThread died with exception:");
	    th.printStackTrace(System.err);
	}
	finally
	{
	    this.jc.disconnect();
	}
    }

    public String toString()
    {
	return "JSIHandlerThread[jhr="
	    + this.jhr
	    + "; jc="
	    + this.jc
	    + "]";
    }
}
