/*
 * JSIMessage.java
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.jsi;

import java.util.Vector;

/**
 * The JSIMessage 
 */
public class JSIMessage
{
    /*
     * Message types.
     */
    public static final int WHAT_NONE = 0;
    public static final int ERROR = 1;
    public static final int SHUTDOWN = 2;
    public static final int CARRIER = 3;
    public static final int REPLY = 4;
    public static final int GET_SUITES = 5;
    public static final int LIST_PROPERTIES = 6;
    public static final int GET_PROPERTY = 7;
    public static final int SET_PROPERTY = 8;
    public static final int CREATE_PROPERTY = 9;
    public static final int DELETE_PROPERTY = 10;
    public static final int EXECUTE_PROPERTY = 11;
    public static final int WHAT_MAX = 12;

    /**
     * The name of the attribute that contains the data to be used in a set.
     */
    public static final String SET_DATA = "data";

    /**
     * The size of an integer on the wire.
     */
    static final int SIZEOF_INT = 4;

    /**
     * @param size The size to align.
     * @return The wire aligned size.
     */
    static int alignedSize(int size)
    {
	int retval;

	retval = (size + (SIZEOF_INT - 1)) & ~(SIZEOF_INT -1);
	return retval;
    }

    /**
     * The buffer used when marshalling/unmarshalling a flattened message.
     */
    private byte buffer[];

    /**
     * The current position in the buffer.
     */
    private int cursor = 0;

    /**
     * The size of the flattened message.
     */
    private int messageSize;

    /**
     * The type of message.
     */
    private int what;

    /**
     * The specifier stack.
     */
    private final Vector specifiers = new Vector();

    /**
     * The current position in the specifier stack.
     */
    private int specifierIndex;

    /**
     * The set of identifier/value pairs in the message.
     */
    private final Vector identifiers = new Vector();

    /**
     * Construct an empty message.
     */
    public JSIMessage()
    {
	this.clear();
    }

    /**
     * @param what The message type.
     */
    public void setWhat(int what)
    {
	this.what = what;
    }

    /**
     * @return The message type.
     */
    public int getWhat()
    {
	return this.what;
    }

    /**
     * Push a specifier onto the stack.
     *
     * @param jp The specifier to add.
     */
    public synchronized void pushSpecifier(JSIPair jp)
    {
	this.specifiers.addElement(jp);
	this.specifierIndex += 1;
	this.messageSize += jp.length() + SIZEOF_INT;
    }

    /**
     * Push a specifier onto the stack.
     * 
     * @param tag The specifier name.
     * @param value The specifier value.
     */
    public void pushSpecifier(String tag, String value)
    {
	this.pushSpecifier(new JSIPair(tag, value));
    }

    /**
     * Push a specifier onto the stack.
     * 
     * @param tag The specifier name.
     */
    public void pushSpecifier(String tag)
    {
	this.pushSpecifier(new JSIPair(tag, ""));
    }

    /**
     * Pop a specifier off the stack.
     *
     * @return The specifier at the top of the stack.
     */
    public synchronized JSIPair popSpecifier()
    {
	JSIPair retval = null;

	if( this.specifierIndex > 0 )
	{
	    this.specifierIndex -= 1;
	    retval = (JSIPair)this.specifiers.elementAt(this.specifierIndex);
	    this.messageSize -= (retval.length() + SIZEOF_INT);
	}
	return retval;
    }

    /**
     * Restore the last specifier popped off the stack.
     */
    public synchronized void restoreSpecifier()
    {
	this.messageSize += ((JSIPair)this.specifiers.
			     elementAt(this.specifierIndex)).length() +
	    SIZEOF_INT;
	this.specifierIndex += 1;
    }

    /**
     * @return The number of specifiers on the stack.
     */
    public int getSpecifierCount()
    {
	return this.specifiers.size();
    }

    /**
     * @return The number of specifiers left on the stack.
     */
    public int getRemainingSpecifierCount()
    {
	return this.specifierIndex;
    }

    /**
     * @return An enumeration containing the identifier/value pairs.
     */
    public Vector getIdentifiers()
    {
	return this.identifiers;
    }

    /**
     * Add an identifier/value pair to the message.
     *
     * @param jp The pair to add.
     * @return this
     */
    public synchronized JSIMessage add(JSIPair jp)
    {
	this.identifiers.addElement(jp);
	this.messageSize += jp.length();
	return this;
    }

    /**
     * Add an identifier/value pair to the message.
     *
     * @param tag The identifier
     * @param value The value
     * @return this
     */
    public JSIMessage add(String tag, String value)
    {
	this.add(new JSIPair(tag, value));
	return this;
    }
    
    /**
     * Add an identifier/value pair to the message.
     *
     * @param tag The identifier
     * @param value The value
     * @return this
     */
    public JSIMessage add(String tag, JSIPropertyInfo value)
    {
	this.add(new JSIPair(tag, value));
	return this;
    }

    /**
     * Add an identifier/value pair to the message.
     *
     * @param tag The identifier
     * @param value An array of JSIPropertyInfo's to add to the message.
     * @return this
     */
    public JSIMessage add(String tag, JSIPropertyInfo value[])
    {
	int lpc;

	for( lpc = 0; lpc < value.length; lpc++ )
	{
	    this.add(new JSIPair(tag, value[lpc]));
	}
	return this;
    }
    
    /**
     * Find an identifier instance in the message.
     *
     * @param tag The identifier to find.
     * @param instance The instance of the identifier to find.
     * @return The index of the identifier in the Vector or -1 if it couldn't
     * @return be found.
     */
    private int findIdentifier(String tag, int instance)
    {
	int lpc, size, retval = -1;
	
	size = this.identifiers.size();
	if( instance >= 0 )
	{
	    for( lpc = 0; (retval == -1) && (lpc < size); lpc++ )
	    {
		JSIPair jp = (JSIPair)this.identifiers.elementAt(lpc);
		
		if( jp.getTag().equals(tag) )
		{
		    if( instance == 0 )
		    {
			retval = lpc;
		    }
		    else
		    {
			instance -= 1;
		    }
		}
	    }
	}
	else
	{
	    for( lpc = (size - 1); (retval == -1) && (lpc >= 0); lpc-- )
	    {
		JSIPair jp = (JSIPair)this.identifiers.elementAt(lpc);
		
		if( jp.getTag().equals(tag) )
		{
		    if( instance == -1 )
		    {
			retval = lpc;
		    }
		    else
		    {
			instance += 1;
		    }
		}
	    }
	}
	return retval;
    }

    /**
     * Remove an identifier value pair from the message.
     *
     * @param tag The identifier to remove.
     * @param instance The instance of the identifier to remove.
     */
    public synchronized void remove(String tag, int instance)
    {
	int slot;
	
	if( (slot = this.findIdentifier(tag, instance)) > -1 )
	{
	    JSIPair jp;

	    jp = (JSIPair)this.specifiers.remove(slot);
	    this.messageSize -= jp.length();
	}
    }

    /**
     * Count the number of identifier/value pairs in the message.
     *
     * @param tag The identifier to count.
     * @return The number of occurrences of the given identifier in the message
     */
    public synchronized int count(String tag)
    {
	int lpc, size, retval = 0;
	
	size = this.identifiers.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    JSIPair jp = (JSIPair)this.identifiers.elementAt(lpc);

	    if( jp.getTag().equals(tag) )
		retval += 1;
	}
	return retval;
    }

    /**
     * Get an identifier/value pair from the message.
     *
     * @param tag The identifier to get.
     * @param instance The instance of the identifier to get.
     * @param defaultValue The value to return if the identifier isn't found.
     * @return The string value of the identifier or the defaultValue if it
     * @return wasn't found.
     */
    public synchronized String get(String tag,
				   int instance,
				   String defaultValue)
    {
	String retval = defaultValue;
	int slot;

	if( (slot = this.findIdentifier(tag, instance)) > -1 )
	{
	    JSIPair jp = (JSIPair)this.identifiers.elementAt(slot);

	    retval = jp.getStringValue();
	}
	return retval;
    }

    /**
     * Get an identifier/value pair from the message.
     *
     * @param tag The identifier to get.
     * @return The string value of the identifier or null if it wasn't found.
     */
    public String get(String tag)
    {
	return this.get(tag, 0, null);
    }

    /**
     * Get an identifier/value pair from the message.
     *
     * @param tag The identifier to get.
     * @param defaultValue The value to return if the identifier isn't found.
     * @return The string value of the identifier or the defaultValue if it
     * @return wasn't found.
     */
    public String get(String tag, String defaultValue)
    {
	return this.get(tag, 0, defaultValue);
    }

    /**
     * Get an identifier/value pair from the message.
     *
     * @param tag The identifier to get.
     * @param instance The instance of the identifier to get.
     * @return The pair associated with the given identifier or null if it
     * @return wasn't found.
     */
    public synchronized JSIPair getPair(String tag, int instance)
    {
	JSIPair retval = null;
	int slot;

	if( (slot = this.findIdentifier(tag, instance)) > -1 )
	{
	    retval = (JSIPair)this.identifiers.elementAt(slot);
	}
	return retval;
    }
    
    /**
     * Get an identifier/value pair from the message.
     *
     * @param tag The identifier to get.
     * @return The pair associated with the given identifier or null if it
     * @return wasn't found.
     */
    public JSIPair getPair(String tag)
    {
	return this.getPair(tag, 0);
    }

    /**
     * Expect an identifier/value pair in the message.
     *
     * @param tag The identifier to get.
     * @return The string value of the pair.
     * @exception Throws a MissingPairException if the tag couldn't be found.
     */
    public String expect(String tag)
	throws MissingPairException
    {
	String retval;

	if( (retval = this.get(tag)) == null )
	{
	    throw new MissingPairException(tag);
	}
	return retval;
    }

    /**
     * Expect an identifier/value pair in the message.
     *
     * @param tag The identifier to get.
     * @return The pair associated with the given identifier.
     * @exception Throws a MissingPairException if the tag couldn't be found.
     */
    public JSIPair expectPair(String tag)
	throws MissingPairException
    {
	JSIPair retval;

	if( (retval = this.getPair(tag)) == null )
	{
	    throw new MissingPairException(tag);
	}
	return retval;
    }

    /**
     * Clear the contents of a message so it can be reused.
     */
    public synchronized void clear()
    {
	this.messageSize = (SIZEOF_INT
			    + SIZEOF_INT
			    + SIZEOF_INT
			    + SIZEOF_INT);
	this.what = WHAT_NONE;
	this.specifiers.removeAllElements();
	this.specifierIndex = 0;
	this.identifiers.removeAllElements();
    }

    /**
     * Clear the message and set it to be a valid reply with the given data.
     *
     * @param result The string value to send back as the reply.
     */
    public void reply(String result)
    {
	this.clear();
	this.what = REPLY;
	this.add("result", result);
    }

    /**
     * Clear the message and set it to be a valid reply with the given data.
     *
     * @param result The property info array to send back as the reply.
     */
    public void reply(JSIPropertyInfo result[])
    {
	this.clear();
	this.what = REPLY;
	this.add("result", result);
    }

    /**
     * Clear the message and set it to be a valid reply with the given data.
     *
     * @param result The integer value to send back as the reply.
     */
    public void reply(long result)
    {
	this.clear();
	this.what = REPLY;
	this.add(new JSIPair("result", result));
    }

    /**
     * Clear the message and set it to be error reply with the given data.
     *
     * @param why A brief explanation of the error.
     */
    public void error(String why)
    {
	this.clear();
	this.what = ERROR;
	this.add("why", why);
    }

    /**
     * @return A flattened version of this message.
     */
    public synchronized byte[] flatten()
    {
	int lpc, specCount, idCount;
	int specifierOffsets[];
	byte retval[];

	// specCount = this.specifiers.size();
	specCount = this.specifierIndex;
	specifierOffsets = new int[specCount];
	idCount = this.identifiers.size();
	retval = this.buffer = new byte[this.messageSize];
	this.writeInt(this.messageSize);
	if( (this.what <= WHAT_NONE) || (this.what >= WHAT_MAX) )
	{
	    throw new IllegalArgumentException("No what value specified.");
	}
	this.writeInt(this.what);
	this.writeInt(specCount);
	this.writeInt(idCount);
	for( lpc = 0; lpc < specCount; lpc++ )
	{
	    JSIPair jp = (JSIPair)this.specifiers.elementAt(lpc);

	    specifierOffsets[lpc] = this.cursor + SIZEOF_INT;
	    jp.marshal(this);
	}
	for( lpc = 0; lpc < idCount; lpc++ )
	{
	    JSIPair jp = (JSIPair)this.identifiers.elementAt(lpc);

	    jp.marshal(this);
	}
	for( lpc = 0; lpc < specCount; lpc++ )
	{
	    this.writeInt(specifierOffsets[lpc]);
	}
	if( this.cursor != this.buffer.length )
	    throw new InternalError();
	this.buffer = null;
	this.cursor = 0;
	return retval;
    }

    /**
     * Unmarshal the message in this.buffer.
     *
     * @return True if the process worked.
     */
    public void unflatten(byte buf[])
	throws JSIEncodingException
    {
	int msgSize, specCount, idCount;

	this.buffer = buf;
	msgSize = this.readInt();
	this.what = this.readInt();
	specCount = this.readInt();
	idCount = this.readInt();
	if( (this.buffer.length == msgSize) &&
	    (specCount >= 0) &&
	    (idCount >= 0) &&
	    (this.what > WHAT_NONE) &&
	    (this.what < WHAT_MAX) )
	{
	    int specifierOffsets[] = new int[specCount];
	    int currSpec = 0, nextSpec = 0, lpc;

	    this.cursor = msgSize - (SIZEOF_INT * specCount);
	    for( lpc = 0; lpc < specCount; lpc++ )
	    {
		specifierOffsets[lpc] = this.readInt();
	    }
	    if( specCount > 0 )
		nextSpec = specifierOffsets[0];
	    this.cursor = SIZEOF_INT * 4;
	    for( lpc = 0; lpc < (specCount + idCount); lpc++ )
	    {
		int tagLength, valueLength;

		tagLength = this.readInt() - 1;
		if( (tagLength <= 0) ||
		    ((this.cursor + tagLength) > this.buffer.length) )
		{
		    throw new JSIEncodingException("Bogus tag at offset: "
						   + this.cursor);
		}
		else if( this.cursor == nextSpec )
		{
		    JSIPair jp;
		    String tag;

		    tag = new String(this.readBytes(tagLength));
		    valueLength = this.readInt() - 1;
		    if( valueLength > 0 )
		    {
			jp = new JSIPair(tag, this.readBytes(valueLength));
		    }
		    else
		    {
			jp = new JSIPair(tag, "");
			this.cursor += 4;
		    }
		    this.pushSpecifier(jp);
		    currSpec += 1;
		    if( currSpec < specCount )
		    {
			nextSpec = specifierOffsets[currSpec];
		    }
		    else
		    {
			nextSpec = 0;
		    }
		}
		else if( this.identifiers.size() < idCount )
		{
		    String tag;

		    tag = new String(this.readBytes(tagLength));
		    valueLength = this.readInt() - 1;
		    this.add(new JSIPair(tag, this.readBytes(valueLength)));
		}
		else
		{
		    throw new JSIEncodingException("Unknown tag data");
		}
	    }
	    if( (this.cursor + SIZEOF_INT * specCount) != this.buffer.length )
		throw new JSIEncodingException("Surplus data at offset: "
					       + (this.cursor +
						  SIZEOF_INT * specCount));
	}
	else
	{
	    throw new JSIEncodingException("Bad header: size="
					   + msgSize
					   + "; specCount="
					   + specCount
					   + "; idCount="
					   + idCount
					   + "; what="
					   + this.what);
	}
	this.buffer = null;
	this.cursor = 0;
    }

    /**
     * @param i The integer to write to the internal buffer.
     */
    void writeInt(int i)
    {
	this.buffer[this.cursor++] = (byte)(i >> 24);
	this.buffer[this.cursor++] = (byte)(i >> 16);
	this.buffer[this.cursor++] = (byte)(i >>  8);
	this.buffer[this.cursor++] = (byte)(i      );
    }
    
    /**
     * @param b The byte array to write to the internal buffer.
     */
    void writeBytes(byte b[])
    {
	if( b.length > 0 )
	{
	    System.arraycopy(b, 0, this.buffer, this.cursor, b.length);
	}
	this.cursor += alignedSize(b.length + 1);
    }

    /**
     * @return An integer decoded from the internal buffer.
     */
    int readInt()
    {
	int retval;

	retval = (((this.buffer[this.cursor++] & 0xFF) << 24) |
		  ((this.buffer[this.cursor++] & 0xFF) << 16) |
		  ((this.buffer[this.cursor++] & 0xFF) <<  8) |
		  ((this.buffer[this.cursor++] & 0xFF)      ));
	return retval;
    }

    /**
     * @param length The length of the byte array to unmarshal from the
     * @param length internal buffer.
     * @return The unmarshalled byte array from the internal buffer.
     */
    byte[] readBytes(int length)
    {
	byte retval[] = new byte[length];

	if( length == 0 )
	{
	    this.cursor += SIZEOF_INT;
	}
	else
	{
	    System.arraycopy(this.buffer, this.cursor, retval, 0, length);
	    this.cursor += alignedSize(length + 1);
	}
	return retval;
    }

    public String toString()
    {
	return "JSIMessage[messageSize="
	    + this.messageSize
	    + "; what="
	    + this.what
	    + "; specifiers="
	    + this.specifiers
	    + "; specifierIndex="
	    + this.specifierIndex
	    + "; identifiers="
	    + this.identifiers
	    + "]";
    }

    /**
     * Instantiate a flattened message.
     *
     * @param msg The byte array containing the flattened message.
     * @return The unflattened JSIMessage or null if the process failed.
     */
    public final static JSIMessage instantiate(byte msg[])
	throws JSIEncodingException
    {
	JSIMessage retval;

	retval = new JSIMessage();
	retval.unflatten(msg);
	return retval;
    }
}
