/*
 * JSIPair.java
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.jsi;

/**
 * A simple class used to couple an identifier string and a byte array
 * value (Usually a string).
 */
public class JSIPair
{
    private static final byte emptyArray[] = new byte[0];

    /**
     * The identifier.
     */
    private String tag;

    /**
     * The value.
     */
    private byte value[];

    /**
     * Construct an identifier/value pair with the given values.
     *
     * @param tag The identifer
     * @param value The byte array value.
     */
    public JSIPair(String tag, byte value[])
    {
	this.tag = tag;
	this.value = value;
    }
    
    /**
     * Construct an identifier/value pair with the given values.
     *
     * @param tag The identifer
     * @param value The string value.
     */
    public JSIPair(String tag, String value)
    {
	this(tag, (value == null) ? emptyArray : value.getBytes());
    }

    /**
     * Construct an identifier/value pair with the given values.
     *
     * @param tag The identifer
     * @param value The character value that is downcasted to a single byte.
     */
    public JSIPair(String tag, char value)
    {
	this(tag, new byte[] { (byte)value });
    }

    /**
     * Construct an identifier/value pair with the given values.
     *
     * @param tag The identifer
     * @param value The long value that is stored as a decimal string.
     */
    public JSIPair(String tag, long value)
    {
	this(tag, Long.toString(value));
    }

    /**
     * Construct an identifier/value pair with the given values.
     *
     * @param tag The identifer
     * @param value The float value that is stored as a decimal string.
     */
    public JSIPair(String tag, float value)
    {
	this(tag, Float.toString(value));
    }

    /**
     * Construct an identifier/value pair with the given values.
     *
     * @param tag The identifer
     * @param value The property info value to be stored in its own custom
     * @param value format.
     */
    public JSIPair(String tag, JSIPropertyInfo pi)
    {
	this(tag, pi.encode());
    }

    /**
     * @param tag The identifier
     */
    public void setTag(String tag)
    {
	this.tag = tag;
    }

    /**
     * @return The identifier
     */
    public String getTag()
    {
	return this.tag;
    }

    /**
     * @param value The value
     */
    public void setValue(byte value[])
    {
	this.value = value;
    }

    /**
     * @param value The value
     */
    public byte[] getValue()
    {
	return this.value;
    }

    /**
     * @return The first byte in the value.
     */
    public char getCharValue()
    {
	return (char)this.value[0];
    }

    /**
     * @return The long as parsed from the string value.
     * @exception Throws NumberFormatException if the string in the value
     * @exception isn't a valid integer.
     */
    public long getIntValue()
	throws NumberFormatException
    {
	return Long.decode(new String(this.value)).longValue();
    }

    /**
     * @return The float as parsed from the string value.
     * @exception Throws NumberFormatException if the string in the value
     * @exception isn't a valid flaot.
     */
    public float getFloatValue()
	throws NumberFormatException
    {
	return Float.parseFloat(new String(this.value));
    }

    /**
     * @return The value as a string.
     */
    public String getStringValue()
    {
	return new String(this.value);
    }

    /**
     * @return The value as a byte array decoded from a hex string.
     */
    public byte[] getByteArrayFromHexValue()
	throws NumberFormatException
    {
	char cvalue[] = new String(this.value).toCharArray();
	byte retval[];
	int lpc;

	if( (cvalue.length % 2) != 0 )
	    throw new NumberFormatException("Not a valid hex string.");
	retval = new byte[cvalue.length / 2];
	for( lpc = 0; lpc < cvalue.length; lpc += 2 )
	{
	    int hi, lo;

	    hi = Character.digit(cvalue[lpc], 16);
	    lo = Character.digit(cvalue[lpc + 1], 16);
	    retval[lpc / 2] = (byte)
		(((hi << 4) & 0xf0) |
		 ((lo     ) & 0x0f));
	}
	return retval;
    }

    /**
     * @return The value as a property info.
     */
    public JSIPropertyInfo getPropertyInfoValue()
    {
	return JSIPropertyInfo.instantiate(new String(this.value));
    }

    /**
     * @return The aligned length of the marshalled pair.
     */
    int length()
    {
	return JSIMessage.SIZEOF_INT
	    + JSIMessage.alignedSize(this.tag.length() + 1)
	    + JSIMessage.SIZEOF_INT
	    + JSIMessage.alignedSize(this.value.length + 1);
    }

    /**
     * Marshal this pair into the given message.
     *
     * @param dst The message this pair should be added to.
     */
    void marshal(JSIMessage dst)
    {
	byte tagBytes[];
	    
	tagBytes = this.tag.getBytes();
	dst.writeInt(tagBytes.length + 1);
	dst.writeBytes(tagBytes);
	dst.writeInt(this.value.length + 1);
	dst.writeBytes(this.value);
    }

    public int hashCode()
    {
	return this.tag.hashCode();
    }

    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( obj instanceof JSIPair )
	{
	    JSIPair js = (JSIPair)obj;

	    if( this.tag.equals(js.tag) &&
		(this.value.length == js.value.length) )
	    {
		int lpc;

		retval = true;
		for( lpc = 0; retval && (lpc < this.value.length); lpc++ )
		{
		    retval = (this.value[lpc] == js.value[lpc]);
		}
	    }
	}
	return retval;
    }

    public String toString()
    {
	return "JSIPair[tag="
	    + this.tag
	    + "; value="
	    + new String(this.value)
	    + "]";
    }
}
