/*
 * JSIPropertyInfo.java
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.jsi;

/**
 * A simple class used to describe a JSI property.
 */
public final class JSIPropertyInfo
{
    /**
     * The name of the property.
     */
    private String name;

    /**
     * A bitmask of the message types that the property understands.
     */
    private int commands;

    /**
     * The specifier value needed to address the property.
     */
    private String specifiers;

    /**
     * A description of the property and its uses.
     */
    private String usage;

    /**
     * Construct a property info with the given values.
     *
     * @param name The name of the property.
     * @param commands A bitmask of the message types that the property
     * @param commands understands.
     * @param specifiers The specifier value needed to address the property.
     * @param usage A description of the property and its uses.
     */
    public JSIPropertyInfo(String name,
			   int commands,
			   String specifiers,
			   String usage)
    {
	this.name = name;
	this.commands = commands;
	this.specifiers = specifiers;
	this.usage = usage;
    }

    /**
     * Construct a property info with the given values and no specifier value.
     *
     * @param name The name of the property.
     * @param commands A bitmask of the message types that the property
     * @param commands understands.
     * @param usage A description of the property and its uses.
     */
    public JSIPropertyInfo(String name, int commands, String usage)
    {
	this(name, commands, "", usage);
    }

    /**
     * @return The name of the property.
     */
    public String getName()
    {
	return this.name;
    }

    /**
     * @return The bitmask of message types the property understands.
     */
    public int getCommands()
    {
	return this.commands;
    }

    /**
     * @return The specifier values needed to address the property.
     */
    public String getSpecifiers()
    {
	return this.specifiers;
    }

    /**
     * @return A description of the property and its uses.
     */
    public String usage()
    {
	return this.usage;
    }

    /**
     * @return An encoded version of the property info structure.
     */
    String encode()
    {
	return "<:"
	    + this.name
	    + ":><:0x"
	    + Integer.toHexString(this.commands)
	    + ":><:"
	    + this.specifiers
	    + ":><:"
	    + this.usage
	    + ":>";
    }

    public int hashCode()
    {
	return this.name.hashCode();
    }

    public boolean equals(Object obj)
    {
	boolean retval = false;

	if( obj instanceof JSIPropertyInfo )
	{
	    JSIPropertyInfo jpi = (JSIPropertyInfo)obj;

	    if( this.name.equals(jpi.name) &&
		(this.commands == jpi.commands) &&
		this.specifiers.equals(jpi.specifiers) &&
		this.usage.equals(jpi.usage) )
	    {
		retval = true;
	    }
	}
	return retval;
    }

    public String toString()
    {
	return "JSIPropertyInfo[name="
	    + this.name
	    + "; commands=0x"
	    + Integer.toHexString(this.commands)
	    + "; specifiers="
	    + this.specifiers
	    + "; usage="
	    + this.usage
	    + "]";
    }

    /**
     * Instantiate a JSIPropertyInfo object from an encoded string.
     *
     * @param str The string containing the encoded version of a
     * @param str JSIPropertyInfo
     * @return The decoded JSIPropertyInfo or null if it failed.
     */
    static JSIPropertyInfo instantiate(String str)
    {
	String values[] = new String[4];
	JSIPropertyInfo retval = null;
	int pos = 0, sIdx = 0, eIdx;
	
	while( (pos < 4) &&
	       ((sIdx = str.indexOf("<:", sIdx)) != -1) &&
	       ((eIdx = str.indexOf(":>", sIdx)) != -1) )
	{
	    values[pos] = str.substring(sIdx + 2, eIdx);
	    sIdx = eIdx + 2;
	    pos += 1;
	}
	if( pos == 4 )
	{
	    try
	    {
		retval = new JSIPropertyInfo(values[0],
					     Integer.decode(values[1]).
					     intValue(),
					     values[2],
					     values[3]);
	    }
	    catch(NumberFormatException e)
	    {
	    }
	}
	return retval;
    }
}
