/*
 * TCPChannel.java
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.comm.jsi;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import java.net.Socket;
import java.net.UnknownHostException;

/**
 * A straight forward TCP implementation of a JSIChannel.
 */
public final class TCPChannel
    implements JSIChannel
{
    private Socket socket;
    private InputStream is;
    private OutputStream os;
    private final byte msgSize[] = new byte[JSIMessage.SIZEOF_INT];

    /**
     * Construct a TCPChannel that runs over the given socket.
     *
     * @param socket A connected socket.
     */
    TCPChannel(Socket socket)
	throws IOException
    {
	this.socket = socket;
	this.is = this.socket.getInputStream();
	this.os = this.socket.getOutputStream();
    }

    /**
     * Construct a TCPChannel thats connected to the given host/port.
     *
     * @param host The host to connect to.
     * @param port The port the provider is listening on.
     */
    public TCPChannel(String host, int port)
	throws UnknownHostException,
	       IOException
    {
	this(new Socket(host, port));
    }
    
    public void disconnect()
    {
	try
	{
	    this.socket.close();
	}
	catch(IOException e)
	{
	}
	this.socket = null;
    }

    public void sendMessage(byte msg[])
	throws IOException
    {
	this.os.write(msg);
	this.os.flush();
    }

    public void replyMessage(byte msg[])
	throws IOException
    {
	this.sendMessage(msg);
    }

    private int read(byte buffer[], int offset, int length)
	throws IOException
    {
	int readLength, retval = 0;

	do
	{
	    readLength = this.is.read(buffer, offset, length);
	    offset += readLength;
	    length -= readLength;
	    retval += readLength;
	}
	while( (length > 0) && (readLength > 0) );
	return retval;
    }

    private int read(byte buffer[])
	throws IOException
    {
	return this.read(buffer, 0, buffer.length);
    }

    public synchronized byte[] recvMessage()
	throws IOException
    {
	byte retval[] = null;
	int count;
	
	if( (count = this.read(this.msgSize)) == this.msgSize.length )
	{
	    int size, rest;
	    byte msg[];
	    
	    size = (((this.msgSize[0] & 0xFF) << 24) |
		    ((this.msgSize[1] & 0xFF) << 16) |
		    ((this.msgSize[2] & 0xFF) <<  8) |
		    ((this.msgSize[3] & 0xFF)      ));
	    msg = new byte[size];
	    System.arraycopy(this.msgSize, 0,
			     msg, 0,
			     this.msgSize.length);
	    rest = msg.length - this.msgSize.length;
	    if( this.read(msg, this.msgSize.length, rest) == rest )
	    {
		retval = msg;
	    }
	}
	else if( count > 0 )
	{
	    throw new IOException("Short message");
	}
	return retval;
    }

    public String toString()
    {
	return "TCPChannel[socket="
	    + this.socket
	    + "; is="
	    + this.is
	    + "; os="
	    + this.os
	    + "]";
    }
}
