/*
 * Isolate.java
 * An implementation of the JSR-121 straw man classes.
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.isolate;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;

import java.util.Vector;
import java.util.Hashtable;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamEvent;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.TeamReservation;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

import edu.utah.janosvm.resources.StdioSpec;
import edu.utah.janosvm.resources.ResourceSpec;
import edu.utah.janosvm.resources.ClassPathSpec;
import edu.utah.janosvm.resources.NameSpaceSpec;
import edu.utah.janosvm.resources.NameSpaceResource;
import edu.utah.janosvm.resources.ResourceSpecException;
import edu.utah.janosvm.resources.InvalidClassGroupException;
import edu.utah.janosvm.resources.ClassGroupNotFoundException;

/**
 * A basic implementation of the JSR-121 Isolate API.
 */
public final class Isolate
    implements Cloneable
{
    /**
     * Reference to the java Class object to be used with
     * Team.setClassLocal/getClassLocal.
     */
    private static Class thisClass = Isolate.class;

    /**
     * The name of the class group in the config file that has all the
     * Isolate classes.
     */
    private static String classGroupName = "kit-jsr121-isolate";

    static {
	try
	{
	    Team curr = Team.current();
	    IsolateManager im;
	    
	    NameSpaceResource.exportClassGroup(classGroupName);
	    
	    /* Create a primordial Isolate object. */
	    Isolate iso = new Isolate(curr, null, null, "root");
	    /* Store it local to this team and class */
	    curr.setClassLocal(thisClass, iso);
	    /* Create the IsolateManager instance for this team. */
	    im = IsolateManager.initIsolateManager(curr);
	    /* Put the isolate in the unique ID to team mapping */
	    im.addMapping(iso.getUniqueId(), curr);
	}
	catch(ClassGroupNotFoundException e)
	{
	    throw new ExceptionInInitializerError("Can't find class group, `"
						  + classGroupName
						  + "', used for Isolates.");
	}
	catch(ClassNotFoundException e)
	{
	    throw new ExceptionInInitializerError(e);
	}
	catch(InvalidClassGroupException e)
	{
	    throw new ExceptionInInitializerError(e);
	}
	catch(Exception e)
	{
	    throw new ExceptionInInitializerError(e);
	}
    }

    /**
     * The possible state's for the Isolate object.
     */
    static final int STATE_INVALID = 0;
    static final int STATE_NEWBORN = 1;
    static final int STATE_RUNNING = 2;
    static final int STATE_SUSPENDED = 3;
    static final int STATE_FINISHED = 4;
    static final int STATE_DYING = 5;
    static final int STATE_DEAD = 6;

    /**
     * Reference to the team, local or remote.
     */
    private TeamHandle th;

    /**
     * The Isolate program name since the team name is anonymous.
     */
    private String name;

    /**
     * The approximate state of the isolate.
     */
    private int state;

    /**
     * A TeamList listener for this Isolate object.  This is used to listen
     * for changes in the Isolate's team state.
     */
    private IsolateTeamList itl;

    /**
     * The name of the class whose `main' method should be run.
     */
    private String mainClassName;

    /**
     * The list of args for mainClassName's main method.
     */
    private String args[];

    /**
     * The unique team identifier.
     */
    private int id;

    /**
     * The Isolate parent.
     */
    private Isolate parent;

    /**
     * The list of children of this Isolate.
     */
    private Vector children;

    private Isolate()
    {
    }

    /**
     * Construct the Isolate object that is local to a team.
     *
     * @param current The team this Isolate corresponds too.
     * @param mainClassName The class to run.
     * @param args The arguments for the `main' method of the class.
     * @param name The name to give the Isolate program.
     */
    private Isolate(Team current,
		    String mainClassName,
		    String args[],
		    String name)
    {
	int lpc;

	this.id = current.getID();
	this.state = STATE_NEWBORN;
	this.th = current;
	/* Copy the name so its local to this team */
	if( mainClassName != null )
	    this.mainClassName = (String)mainClassName.clone();
	if( args != null )
	{
	    /* Copy the arguments over to this team. */
	    this.args = new String[args.length];
	    for( lpc = 0; lpc < args.length; lpc++ )
	    {
		this.args[lpc] = (String)args[lpc].clone();
	    }
	}
	/* Copy the program name */
	this.name = (String)name.clone();
	this.itl = new IsolateTeamList(this);
	this.children = new Vector();
    }
    
    /**
     * Construct the Isolate object that refers to another team.
     *
     * @param current A handle on the other team.
     * @param name The name of the Isolate.
     */
    private Isolate(TeamHandle current, String name)
    {
	this.state = STATE_NEWBORN;
	this.th = current;
	this.name = (String)name.clone();
	this.itl = new IsolateTeamList(this);
    }

    public Isolate(String mainClassName,
		   String[] args,
		   String[] classpath,
		   String name)
	throws SecurityException,
	       UnavailableIsolateFeatureException
    {
	try
	{
	    IsolateManager im = IsolateManager.getCurrentIsolateManager();
	    Isolate parentIsolate = Isolate.getCurrentIsolate();
	    Team parentTeam = Team.current();
	    TeamReservation tr;

	    /* Resource specs for the child. */
	    ResourceSpec specs[] = {
		/* Use default class path if none is specified */
		(classpath != null) ? new ClassPathSpec(classpath) : null,
		/*
		 * Import the same classes as the parent.  Since a
		 * NameSpaceResource.exportClassGroup() does an implicit
		 * import we'll automatically get the isolate classes from
		 * the primordial isolate.
		 */
		NameSpaceSpec.current(),
		/* Use the parents stdio as well. */
		StdioSpec.currentFD()
	    };

	    this.state = STATE_INVALID;
	    /*
	     * Create an IsolateTeamList to listen for team events related
	     * to the Isolate we're about to create.
	     */
	    this.itl = new IsolateTeamList(this);
	    
	    /* Check if we need to use the default name */
	    if( (name == null) || name.equals("") )
		name = "Isolate";

	    /* Save any parameters needed for future use. */
	    this.name = name;

	    /* Reserve the team that is managing Isolates */
	    tr = im.getTeamHandle().pushReservation();

	    /* Create the team */
	    this.th = Team.create(specs);

	    /* Get its unique team identifier */
	    this.id = this.th.getID();

	    try
	    {
		/*
		 * Switch to the manager, this will keep us from dying so
		 * we can add the parent/child dependency.
		 */
		tr.visit();
		{
		    /* Clone the child's team handle for the manager */
		    TeamHandle th_clone = (TeamHandle)this.th.clone();
		    Hashtable uniqueIDtoTeam = im.getMappings();
		    
		    synchronized(uniqueIDtoTeam)
		    {
			/* Switch to the child to set up its state */
			this.th.switchTo();
			{
			    Team child = Team.current();
			    /* Create the child's local Isolate object */
			    Isolate iso = new Isolate(child,
						      mainClassName,
						      args,
						      name);

			    IsolateManager.initIsolateManager(
				(TeamHandle)im.getTeamHandle().clone());
			    /* Add the parent/child dependency */
			    child.addDependency(parentTeam);
			    child.notifyOnFinish();
			    iso.parent = parentIsolate;
			    /*
			     * Hook the local Isolate into the team's global
			     * state.
			     */
			    child.setClassLocal(thisClass, iso);
			}
			this.th.returnFrom();
			/*
			 * Put the mapping from unique ID to TeamHandle into
			 * the table.
			 *
			 * XXX So when exactly is this mapping supposed to
			 * be removed?  Same for the children array...
			 */
			uniqueIDtoTeam.put(name + "-" + th_clone.getID(),
					   th_clone);
		    }
		}
		tr.leave();
		/* We're good to go, switch to the newborn state. */
		this.state = STATE_NEWBORN;
	    }
	    catch(CloneNotSupportedException e)
	    {
		throw new InternalError("TeamHandle doesn't support clone?");
	    }
	    catch(DeadTeamException e)
	    {
		/* We/they are toast */
		this.th.terminate();
		this.th = null;
		
		throw new UnavailableIsolateFeatureException(); // XXX
	    }
	    catch(TeamEjectionException e)
	    {
		/* Ack... kill em */
		this.th.terminate();
		this.th = null;
		
		throw new UnavailableIsolateFeatureException(); // XXX
	    }
	    tr.pop();
	    Isolate.getCurrentIsolate().children.add(this);
	}
	catch(DeadTeamException e)
	{
	    /* We're toast */
	    throw new UnavailableIsolateFeatureException(); // XXX
	}
	catch(ResourceSpecException e)
	{
	    throw new SecurityException();
	}
    }

    public String getName()
    {
	return this.name;
    }

    /**
     * @return The Isolate's team identifier.
     */
    int getID()
    {
	return this.id;
    }

    public String getUniqueId()
    {
	/* Use the form <Isolate program name>-<team id> */
	return (this.name + "-" + this.th.getID());
    }

    /**
     * @return The state of the Isolate.
     */
    synchronized int getState()
    {
	return this.state;
    }

    /**
     * @param state The state of the Isolate.  If the state is valid it will
     * signal any objects that might be waiting for a state change.
     */
    synchronized void setState(int state)
    {
	if( state > this.state )
	{
	    this.state = state;
	    this.notifyAll();
	}
    }

    /**
     * Wait for the Isolate to progress to a given state or time out.
     *
     * @param state The state to wait for.
     * @param millis The number of milliseconds to wait.
     * @param nanos The number of nanoseconds to wait.
     */
    synchronized void waitForState(int state, long millis, int nanos)
    {
	long start, curr, end;

	start = curr = System.currentTimeMillis();
	end = start + millis;
	while( ((millis == 0) || (curr < end)) && (this.state < state) )
	{
	    try
	    {
		/* Wait for a signal or timeout */
		this.wait(millis, nanos);
		curr = System.currentTimeMillis();
	    }
	    catch(InterruptedException ie)
	    {
	    }
	    catch(IllegalMonitorStateException e)
	    {
		this.state = STATE_DEAD;
	    }
	}
    }

    void removeMapping()
    {
	IsolateManager im = IsolateManager.getCurrentIsolateManager();
	
	try
	{
	    TeamHandle th = im.getTeamHandle();
	    
	    th.switchTo();
	    {
		im.getMappings().remove(this.getUniqueId());
	    }
	    th.returnFrom();
	}
	catch(DeadTeamException e)
	{
	}
	catch(TeamEjectionException e)
	{
	}
    }
    
    public static Isolate getIsolate(String uniqueId)
	throws InaccessibleIsolateException
    {
	TeamReservation tr;
	Isolate retval;

	try
	{
	    IsolateManager im = IsolateManager.getCurrentIsolateManager();
	    Hashtable uniqueIDtoTeam = im.getMappings();
	    
	    /* Reserve the manager so we can search the hash table */
	    tr = im.getTeamHandle().pushReservation();
	    synchronized(uniqueIDtoTeam)
	    {
		TeamHandle th = (TeamHandle)uniqueIDtoTeam.get(uniqueId);
		
		if( th != null )
		{
		    /* Found it */
		    try
		    {
			String name;

			/* Get the program name from `<program name>-<tid>' */
			name = uniqueId.substring(0,
						  uniqueId.lastIndexOf('-'));
			retval = new Isolate((TeamHandle)th.clone(), name);
		    }
		    catch(CloneNotSupportedException e)
		    {
			throw new InternalError(
			    "TeamHandle doesn't support clone?");
		    }
		}
		else
		{
		    throw new InaccessibleIsolateException(
			"Couldn't find Isolate with unique ID: " + uniqueId);
		}
	    }
	    tr.pop();
	}
	catch(DeadTeamException e)
	{
	    /* We/they are dead */
	    throw new InaccessibleIsolateException(
		"Couldn't find Isolate with unique ID: " + uniqueId);
	}
	return retval;
    }

    public void start()
	throws IllegalIsolateStateException,
	       SecurityException,
	       InaccessibleIsolateException
    {
	try
	{
	    /* Try and switch to the child. */
	    th.switchTo();
	    {
		Isolate local;

		/* Get its Isolate object */
		if( (local = (Isolate)Team.current().
		     getClassLocal(thisClass)) != null )
		{
		    synchronized(local)
		    {
			if( local.state != STATE_NEWBORN )
			{
			    throw new IllegalIsolateStateException(
				"Isolate already running");
			}
			local.state = STATE_RUNNING;
		    }
		    try
		    {
			/* Setup parameters */
			Class paramTypes[] = { local.args.getClass() };
			final Object params[] = { local.args };
			final Method m;
			Class cl;

			/* Load the main class within this team's context */
			cl = NameSpaceResource.
			    loadTeamClass(local.mainClassName);
			/* Get its main method */
			m = cl.getMethod("main", paramTypes);
			if( !Modifier.isStatic(m.getModifiers()) )
			    throw new IllegalIsolateStateException();
			/* Start a new thread in the team to run the method */
			new Thread(new Runnable() {
			    public void run()
			    {
				Isolate iso = Isolate.getCurrentIsolate();
				TeamHandle th;
				
				try
				{
				    m.invoke(null, params);
				}
				catch(InvocationTargetException e)
				{
				    e.getTargetException().printStackTrace();
				}
				catch(Throwable e)
				{
				    e.printStackTrace();
				}
				/*
				 * Become a daemon and wait for any other
				 * non-daemons to finish.
				 */
				Thread.currentThread().setDaemon(true);
				iso.waitForState(STATE_FINISHED, 0, 0);

				/*
				 * We're dead, remove ourself from the global
				 * mapping table.
				 */
				iso.removeMapping();
				Team.current().terminate();
			    }
			}).start();
		    }
		    catch(NoSuchMethodException e)
		    {
			throw new IllegalIsolateStateException(
			    "No main method in " + local.mainClassName);
		    }
		    catch(ClassNotFoundException e)
		    {
			throw new IllegalIsolateStateException(
			    "No class " + local.mainClassName);
		    }
		}
		else
		{
		    throw new IllegalIsolateStateException();
		}
	    }
	    th.returnFrom();
	}
	catch(DeadTeamException e)
	{
	    throw new InaccessibleIsolateException();
	}
	catch(TeamEjectionException e)
	{
	    this.stop();
	    throw new InaccessibleIsolateException();
	}
    }

    public void stop()
	throws SecurityException,
	       InaccessibleIsolateException
    {
	Isolate curr = Isolate.getCurrentIsolate();
	
	this.th.terminate();
	this.removeMapping();
	if( this.parent == curr )
	{
	    curr.children.removeElement(this);
	}
    }

    public void suspend()
	throws IllegalIsolateStateException,
	       SecurityException,
	       InaccessibleIsolateException
    {
	/* Unsupported... */
    }

    public void resume()
	throws IllegalIsolateStateException,
	       SecurityException,
	       InaccessibleIsolateException
    {
	/* Unsupported... */
    }

    public void join()
	throws IllegalIsolateStateException,
	       InaccessibleIsolateException
    {
	this.join(0, 0);
    }
    
    public void join(long millis)
	throws IllegalIsolateStateException,
	       InaccessibleIsolateException
    {
	this.join(millis, 0);
    }
    
    public void join(long millis, int nanos)
	throws IllegalIsolateStateException,
	       InaccessibleIsolateException
    {
	/* Wait for the finished state */
	this.waitForState(STATE_FINISHED, millis, nanos);
	if( !this.th.isTerminated() )
	{
	    this.waitForState(STATE_DEAD, millis, nanos);
	}
    }

    public boolean isStarted()
	throws InaccessibleIsolateException
    {
	boolean retval = false;

	try
	{
	    /* Try and switch to the child. */
	    th.switchTo();
	    {
		Isolate local;

		/* Get its Isolate object */
		if( (local = (Isolate)Team.current().
		     getClassLocal(thisClass)) != null )
		{
		    if( local.state >= STATE_RUNNING )
			retval = true;
		}
	    }
	    th.returnFrom();
	}
	catch(DeadTeamException e)
	{
	}
	catch(TeamEjectionException e)
	{
	}
	return retval;
    }
    
    public boolean isSuspended()
	throws InaccessibleIsolateException
    {
	return false;
    }

    public boolean isStopped()
	throws InaccessibleIsolateException
    {
	return this.th.isTerminated();
    }

    public Isolate getParent()
    {
	Isolate retval = null;

	if( this.parent != null )
	{
	    retval = (Isolate)this.parent.clone();
	}
	else
	{
	    try
	    {
		TeamReservation tr;
		Isolate iso;
		
		tr = this.th.pushReservation();
		tr.visit();
		{
		    iso = Isolate.getCurrentIsolate().parent;
		}
		tr.leave();
		if( iso != null )
		    retval = (Isolate)iso.clone();
		tr.pop();
	    }
	    catch(DeadTeamException e)
	    {
	    }
	    catch(TeamEjectionException e)
	    {
	    }
	}
	return retval;
    }

    public Isolate[] getChildren()
    {
	/*
	 * XXX Uh, I don't fully understand this function.  Are we supposed to
	 * return the immediate set of children or the entire subtree?
	 */
	Isolate retval[] = null;
	
	if( this.children == null )
	{
	    /*
	     * This is just a handle, switch to the real thing and ask it
	     * for the children.
	     */
	    try
	    {
		TeamReservation tr;
		Isolate iso;
		
		tr = this.th.pushReservation();
		tr.visit();
		{
		    iso = Isolate.getCurrentIsolate();
		}
		tr.leave();
		retval = iso.getChildren();
		tr.pop();
	    }
	    catch(DeadTeamException e)
	    {
	    }
	    catch(TeamEjectionException e)
	    {
	    }
	}
	else
	{
	    synchronized(this.children)
	    {
		int lpc, size, liveIndex = 0, liveCount = 0;
		
		size = this.children.size();
		for( lpc = 0; lpc < size; lpc++ )
		{
		    Isolate child = (Isolate)this.children.elementAt(lpc);
		    
		    /* Make sure its still alive */
		    if( !child.th.isTerminated() )
		    {
			liveCount++;
		    }
		}
		/* One more time to fill out the array */
		retval = new Isolate[liveCount];
		for( lpc = 0; lpc < size; lpc++ )
		{
		    Isolate child = (Isolate)this.children.elementAt(lpc);
		    
		    if( !child.th.isTerminated() )
		    {
			/* XXX Copy to avoid cross team problems. */
			retval[liveIndex] = (Isolate)child.clone();
			liveIndex++;
		    }
		}
	    }
	}
	return retval;
    }
    
    public static Isolate getCurrentIsolate()
    {
	return (Isolate)Team.current().getClassLocal(thisClass);
    }

    public boolean equals(Isolate isolate)
    {
	return this.th.equals(isolate.th);
    }

    public Object clone()
    {
	try
	{
	    Object retval = new Isolate((TeamHandle)this.th.clone(),
					this.name);

	    return retval;
	}
	catch(CloneNotSupportedException e)
	{
	    throw new InternalError("Clone not supported by Isolate?");
	}
    }
}
