/*
 * MemoryBlock.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.server.memfs;

import java.io.IOException;
import java.io.FileNotFoundException;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.Exportable;
import edu.utah.janosvm.sys.Importable;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.ExportManager;
import edu.utah.janosvm.sys.TeamReservation;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.NoSuchExportException;
import edu.utah.janosvm.sys.TeamEjectionException;
import edu.utah.janosvm.sys.DuplicateExportException;

/**
 * A fault tolerant handle for a MemoryBlockBackEnd.
 */
final class MemoryBlock
    extends Importable
{
    /**
     * Pointer to this class, used for set/getClassLocal.
     */
    private static final Class THIS_CLASS = MemoryBlock.class;
    
    /**
     * The number of milliseconds to wait before retrying to connect to the
     * server.
     */
    private static final long CONNECT_DELAY = 10;

    /**
     * The number of times to try connecting to the server.
     */
    private static final int CONNECT_RETRIES = 10;

    /**
     * The name of the block to import.
     */
    private final String fileName;

    /**
     * The exportable.
     */
    private MemoryBlockBackEnd be;

    /**
     * Construct a MemoryBlock handle that will import the MemoryBlockBackEnd
     * with the given name.
     *
     * @param fileName The name of the MemoryBlockBackEnd to import.
     */
    MemoryBlock(String fileName)
    {
	if( fileName == null  )
	    throw new NullPointerException();
	this.fileName = fileName;
    }

    protected void bindTo(Exportable ex)
    {
	/*
	 * Store a link to the export.
	 *
	 * The write barrier doesn't apply here because Importables are
	 * exempted from its effects.
	 */
	this.be = (MemoryBlockBackEnd)ex;
    }

    protected void unbindFrom()
    {
	/* Clear any cross-team pointers and cached values */
	this.be = null;
    }

    /**
     * @return True if the handle is connected to an export.
     */
    synchronized boolean isConnected()
    {
	return this.be != null;
    }

    /**
     * A fault tolerant method for connecting to the server and the exported
     * object.
     *
     * @param create Create the object if it doesn't already exist.
     * @exception Throws FileNotFoundException if the file couldn't be found
     * and create is false or if the method was unable to connect to the
     * server.
     */
    void connect(boolean create)
	throws FileNotFoundException
    {
	Team current = Team.current();
	TeamHandle kernel, server;
	int retries;

	/* Get a TeamHandle object for the server team. */
	if( (server = (TeamHandle)Team.
	     getCurrentClassLocal(THIS_CLASS)) == null )
	{
	    /* No need to synchronize since they should be the same. */
	    server = new TeamHandle();
	    Team.setCurrentClassLocal(THIS_CLASS, server);
	}
	kernel = current.getKernelTeam();
	/* Loop until connection or we run out of retries. */
	for( retries = CONNECT_RETRIES;
	     (retries > 0) && !this.isConnected();
	     retries-- )
	{
	    try
	    {
		/* Try to import a handle to the server. */
		kernel.importObject(server, MemoryFileSystem.TEAM_NAME);
		try
		{
		    server.switchTo();
		    {
			MemoryFileSystem mfs;
			ExportManager em;

			/* We're in the server, get the file system */
			mfs = MemoryFileSystem.current();
			mfs.waitForLoad();
			if( create )
			{
			    String fileNameClone;

			    /*
			     * Need a clone since the file name was allocated
			     * in the other team and the Hashtable will want
			     * to store a reference.
			     */
			    fileNameClone = CloneFactory.clone(this.fileName);
			    em = mfs.put(fileNameClone,
					 new MemoryBlockBackEnd());
			}
			else
			{
			    em = mfs.get(this.fileName);
			}
			if( em != null )
			    em.addImporter(this); /* Success */
			else
			    retries = 0; /* FileNotFound */
		    }
		    server.returnFrom();
		}
		catch(DeadTeamException e)
		{
		    /* Can't reach the server, retry. */
		}
		catch(TeamEjectionException e)
		{
		    /* Can't reach the server, retry. */
		}
	    }
	    catch(DeadTeamException e)
	    {
		/* Can't reach the kernel, we're toast. */
		retries = 0;
	    }
	    catch(TeamEjectionException e)
	    {
		/* Can't reach the kernel, we're toast. */
		retries = 0;
	    }
	    catch(NoSuchExportException e)
	    {
		/* No server team, try to start it ourselves. */
		MemoryFileSystem.start();
	    }
	    if( retries > 1 )
	    {
		try
		{
		    /* Sleep a bit before trying again. */
		    Thread.sleep(CONNECT_DELAY);
		}
		catch(InterruptedException e)
		{
		}
	    }
	}
	if( retries <= 0 )
	{
	    /* No more retries, we couldn't reach the team or find the file. */
	    throw new FileNotFoundException(this.fileName);
	}
    }

    /**
     * @param cursor The offset to read from the file.
     * @param buffer The buffer to read the data into.
     * @param offset The offset into buffer to start reading.
     * @param length The number of bytes to read into buffer.
     * @return The number of bytes read.
     * @exception Throws a DeadTeamException if the import has been revoked.
     */
    synchronized int read(int cursor, byte buffer[], int offset, int length)
	throws DeadTeamException
    {
	int retval = 0;
	
	if( this.be != null )
	{
	    /*
	     * No need to switch teams here since this team is unkillable
	     * as long as we hold the Importable's lock.
	     */
	    retval = this.be.read(cursor, buffer, offset, length);
	}
	else
	{
	    throw new DeadTeamException();
	}
	return retval;
    }
    
    /**
     * @param cursor The offset to start writing into the file.
     * @param buffer The buffer to read the data to write.
     * @param offset The offset into buffer to start reading.
     * @param length The number of bytes to write into buffer.
     * @exception Throws a DeadTeamException if the import has been revoked or
     * the server team is unreachable.
     * @exception Throws a TeamEjectionException if the thread was ejected
     * during processing.
     */
    synchronized void write(int cursor, byte buffer[], int offset, int length)
	throws DeadTeamException, TeamEjectionException
    {
	if( this.be != null )
	{
	    this.be.write(cursor, buffer, offset, length);
	}
	else
	{
	    throw new DeadTeamException();
	}
    }

    /**
     * @return True if the file was deleted from the file system.
     */
    synchronized boolean delete()
	throws DeadTeamException, TeamEjectionException
    {
	boolean retval;
	
	if( this.be != null )
	{
	    /* Switch to the server since the remove() could do anything. */
	    this.be.switchToOwner();
	    {
		MemoryFileSystem mfs;

		mfs = MemoryFileSystem.current();
		retval = mfs.remove(this.fileName);
	    }
	    this.be.returnFromOwner();
	}
	else
	{
	    throw new DeadTeamException();
	}
	return retval;
    }

    /**
     * Force a write to disk of the whole MemoryFileSystem.
     */
    synchronized void sync()
	throws DeadTeamException, TeamEjectionException, IOException
    {
	if( this.be != null )
	{
	    this.be.switchToOwner();
	    {
		MemoryFileSystem mfs;

		mfs = MemoryFileSystem.current();
		mfs.store(MemoryFileSystem.SYNC_FILE);
	    }
	    this.be.returnFromOwner();
	}
	else
	{
	    throw new DeadTeamException();
	}
    }
    
    public String toString()
    {
	return "MemoryBlock[be="
	    + this.be
	    + "]";
    }
}
