/*
 * MemoryBlockBackEnd.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.server.memfs;

import edu.utah.janosvm.sys.Exportable;
import edu.utah.janosvm.sys.TemporaryExport;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

/**
 * The object representing a file in the memory file system.
 */
final class MemoryBlockBackEnd
    extends Exportable
    implements TemporaryExport
{
    /**
     * The size of blocks allocated by the object.
     */
    static final int DEFAULT_BLOCK_SIZE = 1024;

    /**
     * @param size The number of bytes that needs to be contained in blocks.
     * @return The number of blocks needed to contain `size'.
     */
    static int blockCount(int size)
    {
	return (size / DEFAULT_BLOCK_SIZE) + 1;
    }

    /**
     * Indicates whether or not the block has been deleted.  Used to answer
     * the expired() question asked by the ExportManager when the last
     * import is removed.
     */
    private boolean expired;

    /**
     * The file data.
     */
    private byte data[];

    /**
     * The actual size of the file.
     */
    private int actual;

    /**
     * Construct an empty block.
     */
    MemoryBlockBackEnd()
    {
	this.data = new byte[DEFAULT_BLOCK_SIZE];
    }

    /**
     * Construct a block with the given values.
     *
     * @param data A block aligned byte array containing the file data.
     * @param actual The actual size of the data.
     */
    MemoryBlockBackEnd(byte data[], int actual)
    {
	this.data = data;
	this.actual = actual;
    }

    /**
     * @return A block aligned byte array containing the file data.
     */
    byte[] getData()
    {
	return this.data;
    }

    /**
     * @return The actual size of the data.
     */
    int getActual()
    {
	return this.actual;
    }

    /**
     * @param cursor The offset to read from the file.
     * @param buffer The buffer to read the data into.
     * @param offset The offset into buffer to start reading.
     * @param length The number of bytes to read into buffer.
     * @return The number of bytes read.
     */
    synchronized int read(int cursor, byte buffer[], int offset, int length)
    {
	int retval;

	if( cursor < this.actual )
	{
	    if( (this.actual - cursor) < length )
		retval = this.actual - cursor;
	    else
		retval = length;
	    System.arraycopy(this.data, cursor,
			     buffer, offset,
			     retval);
	}
	else
	{
	    retval = 0;
	}
	return retval;
    }

    /**
     * @param cursor The offset to start writing into the file.
     * @param buffer The buffer to read the data to write.
     * @param offset The offset into buffer to start reading.
     * @param length The number of bytes to write into buffer.
     */
    synchronized void write(int cursor, byte buffer[], int offset, int length)
	throws DeadTeamException, TeamEjectionException
    {
	if( cursor == -1 )
	    cursor = this.actual;
	if( (cursor + length) > this.data.length )
	{
	    byte newData[];

	    /* We need to do the allocation in the server team. */
	    this.switchToOwner();
	    {
		newData = new byte[blockCount(cursor + length) *
				  DEFAULT_BLOCK_SIZE];
		System.arraycopy(this.data, 0,
				 newData, 0,
				 this.actual);
		this.data = newData;
	    }
	    this.returnFromOwner();
	}
	if( (cursor + length) > this.actual )
	    this.actual = cursor + length;
	System.arraycopy(buffer, offset,
			 this.data, cursor,
			 length);
    }

    /**
     * @param expired The new value of the expired field.
     */
    void setExpired(boolean expired)
    {
	this.expired = expired;
    }

    public boolean expired()
    {
	return this.expired;
    }
    
    public String toString()
    {
	return "MemoryBlockBackEnd[data="
	    + this.data
	    + "; actual="
	    + this.actual
	    + "; expired="
	    + this.expired
	    + "]";
    }
}
