/*
 * MemoryFileOutputStream.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.server.memfs;

import java.io.IOException;
import java.io.OutputStream;
import java.io.FileNotFoundException;

import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;

/**
 * An OutputStream for a MemoryFileSystem file.
 */
public class MemoryFileOutputStream
    extends OutputStream
{
    /**
     * The handle on the memory block in the server team.
     */
    private final MemoryBlock mb;
    
    /**
     * The current location in the file.
     */
    private int cursor;
    
    /**
     * Construct a MemoryFileOutputStream that writes to the given file.
     *
     * @param fileName The name of the file in the MemoryFileSystem.
     * @param append If true, the anything written to the file will be
     * appended.
     * @exception Throws a FileNotFoundException if the file couldn't be
     * found or the server team was unreachable.
     */
    public MemoryFileOutputStream(String fileName, boolean append)
	throws FileNotFoundException
    {
	/* Create and connect the import */
	this.mb = new MemoryBlock(fileName);
	this.mb.connect(true);
	if( append )
	    this.cursor = -1;
    }

    /**
     * Construct a MemoryFileOutputStream that writes to the given file.
     *
     * @param fileName The name of the file in the MemoryFileSystem.
     * @exception Throws a FileNotFoundException if the file couldn't be
     * found or the server team was unreachable.
     */
    public MemoryFileOutputStream(String fileName)
	throws FileNotFoundException
    {
	this(fileName, false);
    }
    
    public synchronized void write(byte buffer[], int offset, int length)
	throws IOException
    {
	do {
	    try
	    {
		this.mb.write(this.cursor, buffer, offset, length);
		if( this.cursor != -1 )
		    this.cursor += length;
		break;
	    }
	    catch(TeamEjectionException e)
	    {
	    }
	    catch(DeadTeamException e)
	    {
	    }
	    /* The server is unreachable, try to reconnect */
	    try
	    {
		this.mb.connect(false);
	    }
	    catch(FileNotFoundException fnfe)
	    {
		throw new IOException();
	    }
	} while( true );
    }

    public synchronized void write(int i)
	throws IOException
    {
	this.write(new byte[] { (byte)i }); /* XXX */
    }

    public void flush()
	throws IOException
    {
	do {
	    try
	    {
		this.mb.sync();
		break;
	    }
	    catch(TeamEjectionException e)
	    {
	    }
	    catch(DeadTeamException e)
	    {
	    }
	    /* The server is unreachable, try to reconnect */
	    try
	    {
		this.mb.connect(false);
	    }
	    catch(FileNotFoundException fnfe)
	    {
		throw new IOException();
	    }
	} while( true );
    }

    public void close()
	throws IOException
    {
	this.flush();
    }
    
    public String toString()
    {
	return "MemoryFileOutputStream[mb="
	    + this.mb
	    + "; cursor="
	    + this.cursor
	    + "]";
    }
}
