/*
 * MemoryFileSystem.java
 *
 * Copyright (c) 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.server.memfs;

import java.util.Hashtable;
import java.util.Enumeration;

import java.io.IOException;
import java.io.EOFException;
import java.io.FileNotFoundException;

import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.DataOutputStream;
import java.io.FileOutputStream;

import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.ExportManager;
import edu.utah.janosvm.sys.PortAuthority;
import edu.utah.janosvm.sys.TeamReservation;
import edu.utah.janosvm.sys.DeadTeamException;
import edu.utah.janosvm.sys.TeamEjectionException;
import edu.utah.janosvm.sys.DuplicateExportException;

import edu.utah.janosvm.resources.ResourceSpec;
import edu.utah.janosvm.resources.NameSpaceSpec;
import edu.utah.janosvm.resources.NameSpaceGroup;
import edu.utah.janosvm.resources.ResourceSpecException;

/**
 * The core class for the MemoryFileSystem.  This class provides a few static
 * variables for public use, otherwise, it is for internal use only.
 */
public final class MemoryFileSystem
{
    /**
     * Pointer to this class, used for set/getClassLocal.
     */
    private static final Class THIS_CLASS = MemoryFileSystem.class;

    /**
     * The expected name of the class group in the config file.
     */
    public static final String CLASS_GROUP_NAME = "kit-server-memfs";

    /**
     * The name of the team, needed when others are trying to access files.
     */
    public static final String TEAM_NAME = "MemoryFileSystem Server";

    /**
     * The name of the file we'll back up the store too.
     */
    public static final String SYNC_FILE = "memfs.store";

    /**
     * Startup the MemoryFileSystem Server.
     */
    static void start()
    {
	Team current = Team.current();
	TeamHandle kernel;

	kernel = current.getKernelTeam();
	try
	{
	    boolean done = false;

	    /* Switch to the kernel so we don't get slaughtered. */
	    kernel.switchTo();
	    do {
		ResourceSpec rs[];
		Team kernelTeam;
		TeamHandle th;

		/*
		 * The kernel team reference from the other team would cause
		 * a segfault.
		 */
		kernelTeam = Team.current();
		/*
		 * Create a resource spec that imports the class group from the
		 * kernel.
		 *
		 * XXX This contaminates the kernel name space.  Oh well...
		 */
		rs = new ResourceSpec[] {
		    new NameSpaceSpec(new NameSpaceGroup[] {
			new NameSpaceGroup(kernelTeam, "kernel"),
			new NameSpaceGroup(kernelTeam, CLASS_GROUP_NAME),
		    }),
		};
		/* Create the server team. */
		th = Team.create(TEAM_NAME, rs);
		/* Successfully created */
		try
		{
		    th.switchTo();
		    {
			MemoryFileSystem mfs;

			/* Load the file into mfs */
			mfs = current();
			mfs.load(SYNC_FILE);
		    }
		    th.returnFrom();
		    done = true;
		}
		catch(DeadTeamException e)
		{
		    /* It was killed before we could get to it. */
		}
		catch(TeamEjectionException e)
		{
		    th.terminate();
		}
	    } while( !done );
	    kernel.returnFrom();
	}
	catch(DuplicateExportException e)
	{
	    /* Somebody got there before us */
	}
	catch(ResourceSpecException e)
	{
	    /* XXX Hmm... */
	    e.printStackTrace();
	}
	catch(TeamEjectionException e)
	{
	    /* We're probably dead... */
	}
	catch(DeadTeamException e)
	{
	    /* We're probably dead... */
	}
    }

    /**
     * There is only one MemoryFileSystem object maintained in the data team
     * and this will return it.
     *
     * @return The MemoryFileSystem object.
     */
    static MemoryFileSystem current()
    {
	MemoryFileSystem retval;

	retval = (MemoryFileSystem)Team.getCurrentClassLocal(THIS_CLASS);
	if( retval == null )
	{
	    Team current;

	    current = Team.current();
	    synchronized( current )
	    {
		if( (retval = (MemoryFileSystem)current.
		     getClassLocal(THIS_CLASS)) == null )
		{
		    retval = new MemoryFileSystem();
		    current.setClassLocal(THIS_CLASS, retval);
		}
	    }
	}
	return retval;
    }

    /**
     * The MemoryFileSystem object has just been created and hasn't loaded
     * the store from disk.
     */
    private static final int STATE_NEW_BORN = 0;
    
    /**
     * The MemoryFileSystem object is ready to accept new objects.
     */
    private static final int STATE_READY = 1;

    /**
     * The team's PortAuthority, used to export new MemoryBlockBackEnds.
     */
    private final PortAuthority pa = Team.current().getPortAuthority();

    /**
     * Our "custom" name space for the exported objects.
     */
    private final Hashtable table = new Hashtable();

    /**
     * The state of the object.
     */
    private int state = STATE_NEW_BORN;

    /**
     * Construct an empty MemoryFileSystem object.
     */
    private MemoryFileSystem()
    {
    }

    /**
     * Wait for the object to finish loading the store from disk.
     */
    synchronized void waitForLoad()
    {
	try
	{
	    while( this.state == STATE_NEW_BORN )
	    {
		this.wait();
	    }
	}
	catch(InterruptedException e)
	{
	}
	catch(IllegalMonitorStateException e)
	{
	}
    }

    /**
     * Load the object with the data from the given file name.
     *
     * @param fileName The name of the file containing the flattened version
     * of the MemoryFileSystem.
     */
    synchronized void load(String fileName)
    {
	try
	{
	    DataInputStream dis;
	    
	    dis = new DataInputStream(new FileInputStream(fileName));
	    try
	    {
		while( true )
		{
		    String name;
		    byte data[];
		    int actual;

		    name = dis.readUTF();
		    actual = dis.readInt();
		    data = new byte[MemoryBlockBackEnd.blockCount(actual) *
				   MemoryBlockBackEnd.DEFAULT_BLOCK_SIZE];
		    dis.read(data, 0, actual);
		    this.put(name, new MemoryBlockBackEnd(data, actual));
		}
	    }
	    catch(EOFException e)
	    {
	    }
	    finally
	    {
		dis.close();
	    }
	}
	catch(FileNotFoundException e)
	{
	}
	catch(IOException e)
	{
	}
	finally
	{
	    this.state = STATE_READY;
	    this.notifyAll();
	}
    }

    /**
     * Save the file system to disk in the given file name.
     *
     * @param fileName The file name to save the data to.
     */
    synchronized void store(String fileName)
	throws IOException
    {
	DataOutputStream dos;
	
	dos = new DataOutputStream(new FileOutputStream(fileName));
	try
	{
	    Enumeration enum;
	    
	    enum = this.table.keys();
	    while( enum.hasMoreElements() )
	    {
		MemoryBlockBackEnd be;
		byte data[];
		String name;
		int actual;
		
		name = (String)enum.nextElement();
		be = (MemoryBlockBackEnd)this.table.get(name);
		synchronized( be )
		{
		    data = be.getData();
		    actual = be.getActual();
		}
		dos.writeUTF(name);
		dos.writeInt(actual);
		dos.write(data, 0, actual);
	    }
	}
	finally
	{
	    dos.close();
	}
    }

    /**
     * Put a memory block in the file system.
     *
     * @param name The name that maps to the given export.
     * @param newbe The MemoryBlockBackEnd to put in the file system.
     * @return The ExportManager for the newly exported MemoryBlockBackEnd or
     * the existing back end.
     */
    synchronized ExportManager put(String name, MemoryBlockBackEnd newbe)
    {
	MemoryBlockBackEnd be;
	ExportManager retval;
	
	if( (be = (MemoryBlockBackEnd)this.table.get(name)) == null )
	{
	    this.pa.exportAnonymousObject(newbe);
	    this.table.put(name, newbe);
	    be = newbe;
	}
	retval = be.getAnonymousManager();
	return retval;
    }

    /**
     * Get a memory block from the file system.
     *
     * @param name The name that maps to the given block.
     * @return The ExportManager for the block or null if it couldn't be found.
     */
    synchronized ExportManager get(String name)
    {
	ExportManager retval = null;
	MemoryBlockBackEnd be;
	
	if( (be = (MemoryBlockBackEnd)this.table.get(name)) != null )
	{
	    retval = be.getAnonymousManager();
	}
	return retval;
    }

    /**
     * @param name The name of the block to remove from the file system.
     * @return True if the name was found and removed.
     */
    synchronized boolean remove(String name)
    {
	boolean retval = false;
	MemoryBlockBackEnd be;

	if( (be = (MemoryBlockBackEnd)this.table.remove(name)) != null )
	{
	    be.setExpired(true);
	    retval = true;
	}
	return retval;
    }

    public String toString()
    {
	return "MemoryFileSystem[table="
	    + this.table
	    + "; state="
	    + this.state
	    + "]";
    }
}
