/*
 * SharedTable.java
 * Internal class that maintains a cross-team shared hash table.
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.util;

/**
 * The SharedTable class is the core implementation of a hash table that
 * can be shared between multiple teams.
 */
class SharedTable
{
    /**
     * The maximum number of SharedTable object's to stack together.
     */
    private static final int DEPTH_THRESHOLD = 3;

    /**
     * Default primary array size for a new shared table.
     */
    private static final int INITIAL_PRIMARY_SIZE = 15;

    /**
     * Default overflow array size for a new shared table.
     */
    private static final int INITIAL_OVERFLOW_SIZE = 5;

    /**
     * Link to the next table in the stack.
     */
    private SharedTable lastGeneration;

    /**
     * The tables position in the stack.
     * 0 - The bottom of the stack.
     * 1+ - The upper elements in the stack.
     */
    private int depth = 0;

    /**
     * The array of references to Associations for this table.
     */
    private Association entries[];

    /**
     * The size of the primary hash table.
     */
    private final int primaryTableSize;

    /**
     * Array of indexes that link the collision list together.
     */
    private int overflowLink[];

    /**
     * The number of slots used in the overflow array.
     */
    private int overflowTableUsed;

    /**
     * The size of the overflow array.
     */
    private final int overflowTableSize;

    /**
     * The number of entries in this table.
     */
    private int entryCount;

    /**
     * Construct a SharedTable with the default primary and overflow array
     * sizes.
     */
    public SharedTable()
    {
	this(INITIAL_PRIMARY_SIZE, INITIAL_OVERFLOW_SIZE);
    }

    /**
     * Construct a SharedTable with the given primary and overflow array sizes.
     *
     * @param primarySize The size of the primary hash table array.
     * @param overflowSize The size of the overflow array for the table.
     */
    private SharedTable(int primarySize, int overflowSize)
    {
	if( primarySize <= 0 )
	{
	    throw new IllegalArgumentException(
		"The given primarySize "
		+ primarySize
		+ " is too small");
	}
	if( overflowSize <= 0 )
	{
	    throw new IllegalArgumentException(
		"The size of the overflow table is too small");
	}
	
	this.entries = new Association[primarySize + overflowSize];
	this.primaryTableSize = primarySize;
	this.overflowLink = new int[primarySize + overflowSize];
	this.overflowTableUsed = 0;
	this.overflowTableSize = overflowSize;
    }

    /**
     * @return The number of entries in this table.
     */
    public int getEntryCount()
    {
	return this.entryCount;
    }

    /**
     * @return The number of entries in the stack of tables.
     */
    public int getTotalEntryCount()
    {
	SharedTable st = this;
	int retval = 0;
	
	while( st != null )
	{
	    retval += st.entryCount;
	    st = st.lastGeneration;
	}
	return retval;
    }

    /**
     * Compute the hash for the given identitifer object.
     *
     * @param name The identifier to hash.
     * @return A hash index that is between zero and this.primaryTableSize,
     * inclusive.
     */
    private int hashName(Object name)
    {
	int retval;

	retval = name.hashCode();
	retval = retval >= 0 ? retval : -retval;
	retval = retval % this.primaryTableSize;
	return retval;
    }

    /**
     * Put an Association into the table.
     *
     * @param assoc The Association object to put in the table.
     * @exception Throws TableFullException if the overflow array is full and
     * can't accomodate anymore elements.
     */
    void put(Association assoc)
	throws TableFullException
    {
	int hash = this.hashName(assoc.getName());

	if( this.entries[hash] != null )
	{
	    Association curr;

	    curr = this.entries[hash];
	    while( this.overflowLink[hash] != 0 )
	    {
		hash = this.overflowLink[hash];
	    }
	    if( this.overflowTableUsed < this.overflowTableSize )
	    {
		int newSlot = this.primaryTableSize + this.overflowTableUsed++;

		/* There is a free slot, link it in */
		this.overflowLink[hash] = newSlot;
		hash = newSlot;
	    }
	    else
	    {
		throw new TableFullException();
	    }
	}
	/* Put the Association in the table */
	this.entries[hash] = assoc;
	this.entryCount++;
    }

    /**
     * Get an Association from the table using the given identifier object.
     *
     * @param name The identifier to lookup in the table.
     * @return The object associated with `name'.
     */
    Object get(Object name)
    {
	SharedTable st = this;
	boolean found = false;
	Object retval = null;
	Association assoc;
	int hash;

	while( !found && (st != null) )
	{
	    /* Search each SharedTable individually */
	    hash = st.hashName(name);
	    assoc = st.entries[hash];
	    while( !found && (assoc != null) )
	    {
		if( assoc.getName().equals(name) )
		{
		    /* Found it */
		    retval = assoc.getObject();
		    found = true;
		}
		else
		{
		    /* Start walking the collision list. */
		    hash = st.overflowLink[hash];
		    if( hash != 0 )
			assoc = st.entries[hash];
		    else
			assoc = null;
		}
	    }
	    st = st.lastGeneration;
	}
	return retval;
    }

    /**
     * Remove an Association with the given identifier from the table.  This
     * is private since it can only be used by merge() on an unshared table.
     *
     * @param name The identifier to lookup and remove from the table.
     */
    private void remove(Object name)
    {
	int lastHash = 0, hash = this.hashName(name);
	boolean done = false;
	Association assoc;

	assoc = this.entries[hash];
	while( (assoc != null) && !done )
	{
	    if( assoc.getName().equals(name) )
	    {
		if( lastHash == 0 )
		{
		    if( this.overflowLink[hash] != 0 )
		    {
			this.entries[hash] =
			    this.entries[this.overflowLink[hash]];
			this.overflowLink[hash] =
			    this.overflowLink[this.overflowLink[hash]];
		    }
		    else
		    {
			this.entries[hash] = null;
		    }
		}
		else
		{
		    this.overflowLink[lastHash] = this.overflowLink[hash];
		}
		done = true;
	    }
	    else
	    {
		lastHash = hash;
		hash = this.overflowLink[hash];
		if( hash != 0 )
		    assoc = this.entries[hash];
		else
		    assoc = null;
	    }
	}
    }

    public void walk(SharedTableWriter stw, SharedTableWalker walker)
    {
	int lpc;

	for( lpc = 0; lpc < this.entries.length; lpc++ )
	{
	    if( this.entries[lpc] != null )
	    {
		if( walker.walk(this.entries[lpc]) )
		    stw.remove(this.entries[lpc].getName());
	    }
	}
	if( this.lastGeneration != null )
	    this.lastGeneration.walk(stw, walker);
    }
    
    /**
     * Merge the current table with the differences introduced by another
     * table.
     *
     * @param diff A table containing changes to the current table.
     * @param collapse Force the stack of tables to be coalesced into a
     * @param collapse single table.
     * @return Either the `diff' table thats been connected to the current
     * @return table.  Or, a new table that contains all the elements from the
     * @return current stack of tables and any changes introduced by `diff'.
     */
    SharedTable merge(SharedTable diff, boolean collapse)
    {
	SharedTable retval = null;

	if( (this.depth < DEPTH_THRESHOLD) && !collapse )
	{
	    /*
	     * We're still under the threshold, push `diff' onto the top of
	     * the stack.
	     */
	    diff.lastGeneration = this;
	    diff.depth = this.depth + 1;
	    retval = diff;
	}
	else
	{
	    int totalEntries = this.getTotalEntryCount() + diff.entryCount;
	    boolean done = false;

	    /* Need to try this until we don't get a TableFullException... */
	    while( !done )
	    {
		try
		{
		    SharedTable st = this;
		    int lpc;

		    /* XXX Just a guess about the table size */
		    retval = new SharedTable((totalEntries * 3 / 4) + 1,
					     (totalEntries / 2) + 1);
		    /*
		     * Walk the current stack of tables and add their entries
		     * to the new table.
		     */
		    while( st != null )
		    {
			for( lpc = 0; lpc < st.entries.length; lpc++ )
			{
			    if( st.entries[lpc] != null )
			    {
				retval.put(st.entries[lpc]);
			    }
			}
			st = st.lastGeneration;
		    }
		    /* Now walk the diff and apply its changes */
		    for( lpc = 0; lpc < diff.entries.length; lpc++ )
		    {
			if( diff.entries[lpc] != null )
			{
			    if( diff.entries[lpc] instanceof AntiAssociation )
			    {
				/*
				 * Association + AntiAssociation = nothing
				 */
				retval.remove(diff.entries[lpc].getName());
			    }
			    else
			    {
				retval.put(diff.entries[lpc]);
			    }
			}
		    }
		    /* We made it, break out of the loop */
		    done = true;
		}
		catch(TableFullException e)
		{
		    /**
		     * Bah, ran out of space.
		     * XXX Need a more reasonable way to compute the new table
		     * sizes.  Making a perfect hash would be cool, but time
		     * consuming and might not even be possible...
		     */
		    totalEntries += 10;
		}
	    }
	}
	return retval;
    }
}
