/*
 * SharedTableWriter.java
 * Provides read/write access to a shared table.
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.util;

/**
 * Provides read/write access to a shared table.
 */
public class SharedTableWriter
{
    /**
     * The newGeneration table contains any changes to the currently available
     * shared data.
     */
    private SharedTable newGeneration = new SharedTable();

    /**
     * The top of the stack of published SharedTables.
     */
    private SharedTable lastGeneration = new SharedTable();

    /**
     * The reader object that clients can use to lookup data in the current
     * table.  Also, the pointer it holds is updated when the writer does
     * a sync operation.
     */
    private final SharedTableReader tableReader = new SharedTableReader();

    private static final int READ_MISS_THRESHOLD = 10;

    /**
     * The number of times that a read only succeeded in the newest table.
     * This is used to force a sync so that newly created and frequently used
     * data will get published into the shared table.
     */
    private int readMisses;

    public SharedTableWriter()
    {
	this.tableReader.setTable(this.lastGeneration);
    }
    
    /**
     * Put an association into the table.
     *
     * @param name An identifier to associate with the given object.
     * @param obj The object to store in the table.
     */
    public synchronized void put(Object name, Object obj)
    {
	Association assoc = new Association(name, obj);
	
	try
	{
	    this.newGeneration.put(assoc);
	}
	catch(TableFullException e)
	{
	    this.sync();
	    try
	    {
		this.newGeneration.put(assoc);
	    }
	    catch(TableFullException e2)
	    {
		/* er, can't/shouldn't happen */
		throw new InternalError();
	    }
	}
    }

    /**
     * Get an object out of the table.
     *
     * @param name The identifier associated with the object.
     * @return The object associated with the given identifier.
     */
    public synchronized Object get(Object name)
    {
	Object retval;
	
	if( (retval = this.newGeneration.get(name)) != null )
	{
	    this.readMisses++;
	}
	if( this.readMisses >= READ_MISS_THRESHOLD )
	{
	    this.sync();
	}
	return retval;
    }

    /**
     * Remove the given identifier and its associated object from the table.
     *
     * @param name The identifier to remove.
     */
    public synchronized void remove(Object name)
    {
	AntiAssociation aa = new AntiAssociation(name);

	try
	{
	    this.newGeneration.put(aa);
	    this.sync(true);
	}
	catch(TableFullException e)
	{
	    this.sync();
	    try
	    {
		this.newGeneration.put(aa);
	    }
	    catch(TableFullException e2)
	    {
		/* er, can't/shouldn't happen */
		throw new InternalError();
	    }
	}
    }

    /**
     * Synchronize the local table with the globally shared one.  Basically,
     * apply any differences from the last sync.
     *
     * @param collapse Force the collapse of the table stack.
     */
    public synchronized void sync(boolean collapse)
    {
	if( this.newGeneration.getEntryCount() != 0 )
	{
	    this.lastGeneration = this.lastGeneration.
		merge(this.newGeneration, collapse);
	    this.newGeneration = new SharedTable();
	    this.tableReader.setTable(this.lastGeneration);
	    this.readMisses = 0;
	}
    }

    /**
     * Synchronize the local table with the globally shared one.  Basically,
     * apply any differences from the last sync.
     */
    public synchronized void sync()
    {
	this.sync(false);
    }

    /**
     * @return The SharedTableReader object connected to this writer.
     */
    public SharedTableReader getReader()
    {
	return this.tableReader;
    }

    public void walk(SharedTableWalker walker)
    {
	this.lastGeneration.walk(this, walker);
    }

    public String toString()
    {
	return "TableWriter[newGeneration="
	    + this.newGeneration
	    + "; this.lastGeneration="
	    + this.lastGeneration
	    + "; readMisses="
	    + this.readMisses
	    + "]";
    }
}
