/*
 * SparseArray.java
 *
 * Copyright (c) 2001, 2002 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.util;

import java.util.Vector;
import java.util.Enumeration;

/**
 * A simple class for generating/looking up unique identifiers.
 */
public final class SparseArray
{
    /**
     * The default X size of the array.
     */
    public static final int DEFAULT_X_SIZE = 64;

    /**
     * The default Y size of the array.
     */
    public static final int DEFAULT_Y_SIZE = 32;

    /**
     * The number of bits that make up a unique identifier.
     */
    public static final int DEFAULT_ID_BITS = 64;

    /**
     * A guaranteed invalid ID number.
     */
    public static final int INVALID_ID = -1;

    /**
     * Count the number of bits set in the given integer.
     * XXX This sucks.
     *
     * @param bits A value with a few bits set.
     * @return The number of bits in `bits' that are 1.
     */
    private static int bitCount(int bits)
    {
	int lpc, retval = 0;

	for( lpc = 0; lpc < DEFAULT_ID_BITS; lpc++ )
	{
	    if( ((1L << lpc) & bits) != 0 )
		retval++;
	}
	return retval;
    }

    /**
     * The class of elements that should be used in the array.
     */
    private Class elementClass;

    /**
     * The size of the X dimension.
     */
    private int xSize;

    /**
     * The number of bits used to index the X dimension.
     */
    private int xBitCount;

    /**
     * The number of slots actually used in the X dimension.
     */
    private int xUsed;

    /**
     * The size of the Y dimension.
     */
    private int ySize;

    /**
     * The number of bits used to index the Y dimension.
     */
    private int yBitCount;

    /**
     * The number of bits in the delay mask.
     */
    private int timeBitCount;
    
    /**
     * The bitmask that makes up the delay.
     */
    private long timeMask;

    /**
     * The X dimension.
     */
    private final SparseArrayDimension xDimension[];

    /**
     * Construct an array with the given values.
     *
     * @param elementClass The class of objects to use as elements.
     * @param xSize The size of the X dimension.
     * @param ySize The size of the Y dimension.
     */
    public SparseArray(Class elementClass,
		       int xSize,
		       int ySize,
		       int timeBitCount)
    {
	this.elementClass = elementClass;
	this.xSize = xSize;
	this.xBitCount = bitCount(xSize - 1);
	this.ySize = ySize;
	this.yBitCount = bitCount(ySize - 1);
	this.timeBitCount = timeBitCount;
	this.timeMask = ~0;
	this.timeMask <<= (DEFAULT_ID_BITS - this.timeBitCount);
	this.timeMask >>= (DEFAULT_ID_BITS - this.timeBitCount);
	this.timeMask >>= this.xBitCount + this.yBitCount;
	this.timeMask <<= this.xBitCount + this.yBitCount;
	this.xDimension = new SparseArrayDimension[this.xSize];
    }

    /**
     * Construct an array with the given values and the default element class.
     *
     * @param xSize The size of the X dimension.
     * @param ySize The size of the Y dimension.
     */
    public SparseArray(int xSize, int ySize)
    {
	this(ObjectSparseArrayElement.class,
	     xSize,
	     ySize,
	     DEFAULT_ID_BITS - xSize - ySize);
    }

    /**
     * Construct an array with the given values and the default sizes.
     *
     * @param elementClass The class of objects to use as elements.
     */
    public SparseArray(Class cl)
    {
	this(cl,
	     DEFAULT_X_SIZE,
	     DEFAULT_Y_SIZE,
	     DEFAULT_ID_BITS - DEFAULT_X_SIZE - DEFAULT_Y_SIZE);
    }
    
    /**
     * Construct an array with the default values.
     */
    public SparseArray()
    {
	this(DEFAULT_X_SIZE, DEFAULT_Y_SIZE);
    }

    /**
     * @return The class of elements in the array.
     */
    public Class getElementClass()
    {
	return this.elementClass;
    }

    /**
     * @return The size of the X dimension.
     */
    public int getXSize()
    {
	return this.xSize;
    }

    /**
     * @return The size of the Y dimension.
     */
    public int getYSize()
    {
	return this.ySize;
    }

    /**
     * @return The number of bits in the delay mask.
     */
    public int getTimeCount()
    {
	return this.timeBitCount;
    }

    /**
     * @return The start of the delay bits.
     */
    public int getTimeOffset()
    {
	return (this.xBitCount + this.yBitCount);
    }

    /**
     * @return The bitmask that makes up the delay.
     */
    public long getTimeMask()
    {
	return this.timeMask;
    }

    /**
     * @return An allocated SparseArrayElement.
     */
    public synchronized SparseArrayElement getElement()
    {
	SparseArrayDimension sad = null;
	SparseArrayElement retval;
	int lpc;

	for( lpc = this.xUsed - 1; (sad == null) && (lpc >= 0); lpc-- )
	{
	    if( this.xDimension[lpc].hasFreeElements() )
		sad = this.xDimension[lpc];
	}
	if( sad == null )
	{
	    try
	    {
		sad = this.xDimension[this.xUsed] =
		    new SparseArrayDimension(this.xUsed << this.yBitCount,
					     this);
	    }
	    catch(IllegalAccessException e)
	    {
		throw new InternalError();
	    }
	    catch(InstantiationException e)
	    {
		throw new InternalError();
	    }
	    this.xUsed += 1;
	}
	retval = sad.getElement();
	return retval;
    }

    /**
     * @param sae A SparseArrayElement to free.
     */
    public synchronized void dropElement(SparseArrayElement sae)
    {
	SparseArrayDimension sad;
	int xIndex;
		
	xIndex = (int)((sae.getID() >> this.yBitCount) & (this.xSize - 1));
	if( (sad = this.xDimension[xIndex]) != null )
	{
	    sad.dropElement(sae);
	}
    }

    /**
     * @param id The element to find.
     * @return The element with the given identifier or null if it wasn't found
     */
    public synchronized SparseArrayElement findElement(long id)
    {
	SparseArrayElement retval = null;
	SparseArrayDimension sad;
	int xIndex;

	xIndex = (int)((id >> this.yBitCount) & (this.xSize - 1));
	if( (sad = this.xDimension[xIndex]) != null )
	{
	    retval = sad.findElement((int)(id & (this.ySize - 1)), id);
	}
	return retval;
    }

    /**
     * @return An enumeration containing the elements in the array.  NOTE: You
     * @return must synchronize on the SparseArray so that the elements cannot
     * @return be dropped.
     */
    public synchronized Enumeration elements()
    {
	Vector retval = new Vector();
	int lpc;

	for( lpc = 0; lpc < this.xSize; lpc++ )
	{
	    if( this.xDimension[lpc] != null )
		this.xDimension[lpc].elements(retval);
	}
	return retval.elements();
    }

    public String toString()
    {
	return "SparseArray[elementClass="
	    + this.elementClass
	    + "; xSize="
	    + this.xSize
	    + "; xBitCount="
	    + this.xBitCount
	    + "; xUsed="
	    + this.xUsed
	    + "; ySize="
	    + this.ySize
	    + "; yBitCount="
	    + this.yBitCount
	    + "; timeMask=0x"
	    + Long.toHexString(this.timeMask)
	    + "]";
    }
}
