/*
 * SparseArrayDimension.java
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Lesser 
 * General Public License.
 * 
 * See the file "license-lesser.terms" for information on 
 * usage and redistribution of this file.
 */

package edu.utah.janosvm.kit.util;

import java.util.Vector;

/**
 * The objects that make up the SparseArray's X dimension.
 */
final class SparseArrayDimension
{
    /**
     * The SparseArray this dimension is a part of.
     */
    private SparseArray sa;
    
    /**
     * The list of free elements in this dimension.
     */
    private SparseArrayElement freeElements;

    /**
     * The last element in the free list.
     */
    private SparseArrayElement lastElement;

    /**
     * The number of elements in use.
     */
    private int inUse = 0;

    /**
     * The present delay value.
     */
    private long timeID = 0;

    /**
     * The elements themselves.
     */
    private final SparseArrayElement elements[];

    /**
     * Construct a SparseArrayDimension with the given values.
     *
     * @param xIndex The X index in the SparseArray.
     * @param sa The SparseArray this dimension belongs too.
     */
    SparseArrayDimension(int xIndex, SparseArray sa)
	throws InstantiationException,
	       IllegalAccessException
    {
	int elementCount, lpc;
	Class elementClass;

	this.sa = sa;
	elementClass = sa.getElementClass();
	elementCount = sa.getYSize();
	this.elements = new SparseArrayElement[elementCount];
	for( lpc = 0; lpc < elementCount; lpc++ )
	{
	    this.elements[lpc] = (SparseArrayElement)
		elementClass.newInstance();
	    this.elements[lpc].setID(xIndex | lpc);
	    if( lpc > 0 )
	    {
		this.elements[lpc - 1].
		    setNext(this.elements[lpc]);
	    }
	}
	this.freeElements = this.elements[0];
	this.lastElement = this.elements[this.elements.length - 1];
    }

    /**
     * @return True if there are any free elements in this dimension.
     */
    boolean hasFreeElements()
    {
	return (this.freeElements != null);
    }

    /**
     * @return An allocated element from this dimension.
     */
    SparseArrayElement getElement()
    {
	SparseArrayElement retval;
	long newTime;

	retval = this.freeElements;
	this.freeElements = retval.getNext();
	retval.obtain();
	if( this.freeElements == null )
	    this.lastElement = null;
	newTime = retval.getID() & this.sa.getTimeMask();
	if( newTime != this.timeID )
	    this.timeID = newTime;
	this.inUse++;
	return retval;
    }

    /**
     * @param sae The element to be freed.
     */
    void dropElement(SparseArrayElement sae)
    {
	this.inUse--;
	if( this.lastElement == null )
	    this.freeElements = sae;
	else
	    this.lastElement.setNext(sae);
	this.lastElement = sae;
	sae.release(this.sa.getTimeOffset());
    }

    /**
     * Find an element in this dimension.
     *
     * @param yIndex The Y index of the element.
     * @param id The whole id of the element.
     * @return The SparseArrayElement corresponding to the given ID or null.
     */
    SparseArrayElement findElement(int yIndex, long id)
    {
	SparseArrayElement retval = null;

	if( this.elements[yIndex].isAllocated() &&
	    (this.elements[yIndex].getID() == id) )
	{
	    retval = this.elements[yIndex];
	}
	return retval;
    }

    /**
     * Add any allocated elements to the given vector.
     *
     * @param v A vector to add allocated elements to.
     */
    void elements(Vector v)
    {
	int lpc;

	for( lpc = 0; lpc < this.elements.length; lpc++ )
	{
	    if( this.elements[lpc].isAllocated() )
		v.addElement(this.elements[lpc]);
	}
    }
    
    public String toString()
    {
	return "SparseArrayDimension[sa="
	    + "; freeElements="
	    + this.freeElements
	    + "; lastElement="
	    + this.lastElement
	    + "; inUse="
	    + this.inUse
	    + "; timeID="
	    + this.timeID
	    + "; elements="
	    + this.elements
	    + "]";
    }
}
