
import edu.utah.janosvm.kit.comm.commspace.CommQueuePuller;
import edu.utah.janosvm.kit.comm.commspace.CommQueuePusher;
import edu.utah.janosvm.kit.comm.commspace.CommSpaceElement;

public class CommQueueTester
{
	private TestFailedException consumerException = null;
	private final int exitValue = 10;

	/**
	 * Runnable class that consumes data
	 */
	private class Consumer implements Runnable
	{
		private CommQueuePuller cq;
		
		public Consumer(CommQueuePuller cq)
		{
			this.cq = cq;
		}
		
		public void run()
		{
			CommSpaceElement elem;
			
			try
			{
				boolean done = false;
				
				while( !done )
				{
					elem = cq.waitToTake();
					if( elem.getInt() == exitValue )
						done = true;
				}
			}
			catch(InterruptedException e)
			{
				CommQueueTester cqt =
					CommQueueTester.this;
				
				cqt.consumerException =
					new TestFailedException();
			}
		}
	}

	/**
	 * A simple watch dog thread.  It waits for a given length of time and
	 * then interrupts the target thread, unless it gets interrupted
	 * beforehand.
	 */
	private class WatchDog extends Thread
	{
		/**
		 * The amount of time to sleep before interrupting thread
		 */
		private long time;
		
		/**
		 * The thread to watch, and interrupt if time expires before
		 * we're interrupted
		 */
		private Thread thread;

		/**
		 * Construct a watch dog thread that watches a thread for ti
		 * time
		 *
		 * @param ti The length of time to sleep before interrupting
		 * the thread
		 * @param th The thread to watch
		 */
		public WatchDog(long ti, Thread th)
		{
			this.time = ti;
			this.thread = th;
		}
		
		/**
		 * Sleep for time milliseconds and then call interrupt on
		 * thread if we aren't interrupted before hand.  An interrupt
		 * called on our thread will cause make the thread quit
		 * quietly.
		 */
		public void run()
		{
			try
			{
				Thread.sleep(this.time);
				System.err.println("Watch dog timed out!");
				this.thread.interrupt();
				System.err.println("Interrupted thread "
						   + this.thread);
			}
			catch(InterruptedException e)
			{
			}
		}
	}

	
	/**
	 * Exception encapsulating a test failure
	 */
	private class TestFailedException
		extends Exception
	{
		public TestFailedException(String msg)
		{
			super(msg);
		}

		public TestFailedException()
		{
			super("Test failed");
		}
	}

	public void test(boolean val)
		throws TestFailedException
	{
		if (!val)
			throw new TestFailedException();
	}

	public void testUnit() throws TestFailedException
	{
		final CommQueuePuller pull = new CommQueuePuller();
		final CommQueuePusher push = new CommQueuePusher(pull);
		java.lang.Thread watchDog;
		CommSpaceElement cse;
		int lpc;

		System.out.println("Testing queue operations");
		//this.enterSection(ENV_NONE, "Queue operations");

		/* enqueue some elements to be taken */
		int elements[] =
		{
			1,
			2,
			3,
			4
		};
		for( lpc = 0; lpc < elements.length; lpc++ )
		{
			cse = new CommSpaceElement(elements[lpc]);
			push.enqueue(cse);
		}

		/* take the elements, use watch dog so we don't stall */
		watchDog = new WatchDog(10000,
					java.lang.Thread.currentThread());
		watchDog.start();
		try
		{
			for( lpc = 0; lpc < elements.length; lpc++ )
			{
				cse = pull.waitToTake();
				this.test(cse.getInt() == elements[lpc]);
			}
		}
		catch(InterruptedException e)
		{
			throw new TestFailedException();
		}
		watchDog.interrupt();

		java.lang.Thread consumer;

		/* start a consumer thread and enqueue some elements for it */
		consumer = new java.lang.Thread(new Consumer(pull));
		try
		{
			consumer.start();
			for( lpc = 0; lpc <= exitValue; lpc++ )
			{
				/* sleep so the consumer can do something */
				java.lang.Thread.sleep(50);
				push.enqueue(new CommSpaceElement(lpc));
			}
			consumer.join(1000);
			if( consumer.isAlive() )
			{
				consumer.interrupt();
				throw new TestFailedException();
			}
		}
		catch(InterruptedException e)
		{
			consumer.interrupt();
			throw new TestFailedException();
		}
		if( this.consumerException != null )
			throw new TestFailedException();
	}

	public static void main(String args[])
	{
		try
		{
			(new CommQueueTester()).testUnit();
			System.out.println("The test succeeded!");
		}
		catch(TestFailedException e)
		{
			System.out.println("The test failed: "
					 + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		catch(Exception e)
		{
			System.out.println("There was an exception "
					 + "in the testing code:"
					 + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
	}
}

/* Expected output:
Testing queue operations
The test succeeded!
*/
