import edu.utah.janosvm.kit.comm.commspace.CommSpaceHandle;
import edu.utah.janosvm.kit.comm.commspace.CommSpaceElement;

public class CommSpaceHandleTester
{
	private final int startValue = 0;
	private final int exitValue = 10;

	private TestFailedException consumerException = null;

	/**
	 * Runnable thread that consumes data from a space
	 */
	private class Consumer implements Runnable
	{
		private CommSpaceHandle csh;
		
		public Consumer(CommSpaceHandle csh)
		{
			this.csh = csh;
		}
		
		public void run()
		{
			CommSpaceElement cse;
			boolean done = false;
			
			try
			{
				while( !done )
				{
					java.lang.Thread.sleep(10);
					cse = this.csh.waitToTake(0);
					if( cse.getInt() == exitValue )
						done = true;
				}
			}
			catch(Exception e)
			{
				CommSpaceHandleTester csht =
					CommSpaceHandleTester.this;
				
				csht.consumerException =
					new TestFailedException();
			}
		}
	}

	/**
	 * Exception encapsulating a test failure
	 */
	private class TestFailedException
		extends Exception
	{
		public TestFailedException(String msg)
		{
			super(msg);
		}

		public TestFailedException()
		{
			super("Test failed");
		}
	}

	public void test(boolean val)
		throws TestFailedException
	{
		if (!val)
			throw new TestFailedException();
	}
	
	public void testUnit()
		throws TestFailedException
	{
		final int commSpaceSize = 16;
		CommSpaceHandle csh;
		int lpc;

		System.out.println("Testing constructor");
		//this.enterSection(ENV_NONE, "Constructor");
		csh = CommSpaceHandle.newCommSpace(commSpaceSize);
		this.test(csh.size() == commSpaceSize);
		csh = new CommSpaceHandle();
		csh = CommSpaceHandle.newCommSpace(csh, commSpaceSize);
		this.test(csh.size() == commSpaceSize);

		System.out.println("Testing reading/writing");
		//this.enterSection(ENV_NONE, "reading/writing");
		try
		{
			for( lpc = 0; lpc < commSpaceSize; lpc++ )
			{
				this.test(csh.swap(lpc, null) == null);
				this.test(csh.read(lpc) == null);
				this.test(csh.take(lpc) == null);
			}
		}
		catch(ArrayIndexOutOfBoundsException e)
		{
			throw new TestFailedException();
		}

		/* Make sure we can't index something bogus */
		int badIndexes[] =
		{
			Integer.MIN_VALUE,
			-1,
			commSpaceSize,
			commSpaceSize + 1,
			Integer.MAX_VALUE /* check for rollover */
		};
		for( lpc = 0; lpc < badIndexes.length; lpc++ )
		{
			try
			{
				csh.read(badIndexes[lpc]);
				throw new TestFailedException();
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
			}
			try
			{
				csh.write(badIndexes[lpc], null);
				throw new TestFailedException();
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
			}
			try
			{
				csh.take(badIndexes[lpc]);
				throw new TestFailedException();
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
			}
			try
			{
				csh.waitToTake(badIndexes[lpc]);
				throw new TestFailedException();
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
			}
			catch(InterruptedException e)
			{
			}
			try
			{
				csh.swap(badIndexes[lpc], null);
				throw new TestFailedException();
			}
			catch(ArrayIndexOutOfBoundsException e)
			{
			}
		}
		/* Write/read/swap/take/read stuff in the comm space */
		try
		{
			CommSpaceElement cse;

			for( lpc = 0; lpc < commSpaceSize; lpc++ )
			{
				csh.write(lpc,
					  new CommSpaceElement(lpc + 1));
			}
			for( lpc = 0; lpc < commSpaceSize; lpc++ )
			{
				this.test(csh.read(lpc).getInt() == lpc + 1);
				cse = new CommSpaceElement(lpc + 2);
				this.test(csh.swap(lpc, cse).getInt() ==
					  lpc + 1);
				this.test(csh.take(lpc).getInt() == lpc + 2);
				this.test(csh.read(lpc) == null);
			}
		}
		catch(ArrayIndexOutOfBoundsException e)
		{
			throw new TestFailedException();
		}

		java.lang.Thread consumer;

		/* Check for synchronization */
		consumer = new java.lang.Thread(new Consumer(csh));
		try
		{
			consumer.start();
			for( lpc = startValue; lpc < exitValue; lpc++ )
			{
				java.lang.Thread.sleep(50);
				csh.write(0, new CommSpaceElement(lpc));
			}
			csh.write(0, null);
			for( lpc = startValue; lpc <= exitValue; lpc++ )
			{
				java.lang.Thread.sleep(50);
				csh.swap(0, new CommSpaceElement(lpc));
			}
			consumer.join(1000);
			if( consumer.isAlive() )
			{
				consumer.interrupt();
				throw new TestFailedException();
			}
		}
		catch(Exception e)
		{
			consumer.interrupt();
			throw new TestFailedException();
		}
		if( this.consumerException != null )
			throw this.consumerException;
	}

	public static void main(String args[])
		throws Exception
	{
		try
		{
			(new CommSpaceHandleTester()).testUnit();
			System.out.println("The test succeeded!");
		}
		catch(TestFailedException e)
		{
			System.out.println("The test failed: "
					 + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
		catch(Exception e)
		{
			System.out.println("There was an exception "
					 + "in the testing code:"
					 + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}
	}
}

/* Expected output:
Testing constructor
Testing reading/writing
The test succeeded!
*/
