
import edu.utah.janosvm.kit.comm.amiga.MsgPort;
import edu.utah.janosvm.kit.comm.amiga.Message;
import edu.utah.janosvm.kit.comm.amiga.SignalSet;

/**
 * Message test exercises the message port part of the ipc system.  First
 * it sends some messages between threads in the same team, then it will
 * create a new team and exchange messages with it.
 */
public class MessageTest
    extends Thread
{
    /**
     * Simple message that contains a string that the server thread should
     * print out.
     */
    public static class EchoMessage
	extends Message
    {
	private String msg;

	public EchoMessage(String msg)
	{
	    this.msg = msg;
	}

	public String getString()
	{
	    return this.msg;
	}
    }

    /**
     * Another simple message type, this one will ask the server thread for
     * some statistics.  Currently, its just the number of echo messages its
     * processed.
     */
    public static class QueryMessage
	extends Message
    {
	private int echoCount;

	public QueryMessage()
	{
	}

	public void setEchoCount(int count)
	{
	    this.echoCount = count;
	}
	
	public int getEchoCount()
	{
	    return this.echoCount;
	}
    }

    /**
     * Signal allocated by the server thread that indicates that it should
     * exit.
     */
    private static final int SIG_EXIT = 0;

    /**
     * Signal allocated by the root thread that the server thread uses to
     * indicate that it is setup and ready to run.
     */
    private static final int SIG_CONTINUE = 1;

    /**
     * Reference to the root threads signal set, used when sending the
     * SIG_CONTINUE signal.
     */
    private SignalSet parent;

    /**
     * Our signal set, only allocated bits are SIG_EXIT and one for the
     * message port.
     */
    private SignalSet ss;

    private MsgPort mp;

    public MessageTest(SignalSet parent, SignalSet ss)
    {
	this.parent = parent;
	this.ss = ss;
	this.ss.allocate(SIG_EXIT);
	/* Create/Export a message port */
	this.mp = new MsgPort(this.ss);
    }

    public MsgPort getMsgPort()
    {
	return this.mp;
    }
    
    public void run()
    {
	/* Get the signal number our message port is bound too */
	int mpSig = this.mp.getSignalNumber();
	/* Compute the mask of signals that we care to wait for */
	long waitMask = (1L << SIG_EXIT) | (1L << mpSig);
	/* The signals we actually received */
	long recvMask;
	/* The number of EchoMessage's we've processed */
	int echoCount = 0;
	boolean done = false;

	/*
	     * Our message port is exported, tell the root thread to continue
	     */
	if( this.parent != null )
	    this.parent.signal(SIG_CONTINUE);
	while( !done )
	{
	    try
	    {
		/* Wait for an event */
		recvMask = this.ss.waitFor(waitMask);
		if( (recvMask & (1L << mpSig)) != 0 )
		{
		    /* There are some new messages... */
		    
		    Message msg = this.mp.getMessage();

		    while( msg != null )
		    {
			/* Figure out what kind of message it is */
			if( msg instanceof EchoMessage )
			{
			    EchoMessage em = (EchoMessage)msg;
			    
			    echoCount++;
			    System.out.println("Message #"
					       + echoCount
					       + ":");
			    System.out.println(em.getString());
			}
			else if( msg instanceof QueryMessage )
			{
			    QueryMessage qm = (QueryMessage)msg;

			    qm.setEchoCount(echoCount);
			    qm.reply();
			}
			else
			{
			    System.out.println("Can't handle message:  "
					       + msg);
			}
			/*
			 * Keep dequeuing messages until there are no more
			 */
			msg = this.mp.getMessage();
		    }
		}
		if( (recvMask & (1L << SIG_EXIT)) != 0 )
		{
		    /* We got an exit signal */
		    done = true;
		}
	    }
	    catch(Throwable th)
	    {
		/* doh! */
		System.out.println("Got exception: " + th);
		th.printStackTrace();
		done = true;
	    }
	}
	this.parent.signal(SIG_CONTINUE);
    }
    
    public static void main(String args[])
	throws Exception
    {
	/* Test messages */
	String echoMessages[] = {
	    "\'Things involving the computer fill me with a childlike terror."
	    + "  Now, if it were a nice ogre or some such, I'd be more in my"
	    + " element.\' -- Giles",
	    "\'You were looking at my neck.\'\n"
	    + "\'What?\'\n"
	    + "\'You were checking out my neck, I saw that.\'\n"
	    + "\'No, I wasn't.\'\n"
	    + "\'Just keep your distance, pal.\'\n"
	    + "\'I wasn't looking at your neck.\'\n"
	    + "\'I told you to eat before we left.\' -- Xander and Angel",
	    "\'We like to talk big, vampires do. \'I'm going to destroy the "
	    + "world.\' That's just tough guy talk. Strutting around with "
	    + "your friends over a pint of blood. The truth is, I like "
	    + "this world. You've got dog racing. Manchester United. And "
	    + "you've got people. Billions of people walking around like "
	    + "Happy Meals with legs. It's all right here.\' -- Spike",
	    "'How can I thank you, you mysterious black-clad hunk of a "
	    + "night-thing.' 'No need, little lady.  Your tears of gratitude "
	    + "are enough for me. You see, I was once a bad-ass vampire. But "
	    + "love and a pesky curse defanged me and now I'm just a big "
	    + "fluffy puppy with bad teeth. No, not the hair. Never the hair.'"
	    + "'But there must be some way I can show my appreciation.' 'No. "
	    + "Helping those in need is my job and working up a load of "
	    + "sexual tension and prancing away like a magnificent poof is "
	    + "truly thanks enough.' 'I understand, I have a nephew who's "
	    + "gay.' 'Say no more, evil's afoot and I'm almost out of that "
	    + "nancy-boy hair gel I like so much. Quickly! To the "
	    + "Angel-mobile.  Away!' - Spike"
	};
	
	SignalSet me = new SignalSet();
	SignalSet ss = new SignalSet();
	MsgPort replyPort;
	QueryMessage qm;
	MessageTest mt;
	MsgPort mp;
	int lpc;
	
	me.allocate(SIG_CONTINUE);
	/* Start the server thread */
	mt = new MessageTest(me, ss);
	mt.start();
	/* Wait for the continue signal */
	me.waitFor(1L << SIG_CONTINUE);
	/* Get a handle on the server threads message port */
	mp = mt.getMsgPort();
  	/* Make a reply port for QueryMessage's */
	replyPort = new MsgPort(me);
	qm = new QueryMessage();
	qm.setReplyPort(replyPort);
	for( lpc = 0; lpc < echoMessages.length; lpc++ )
	{
	    /* Send an EchoMessage */
	    mp.putMessage(new EchoMessage(echoMessages[lpc]));
	    /* Send a query Message */
	    mp.putMessage(qm);
	    replyPort.waitForMessage();
	    System.out.println("Echo count = "
			       + qm.getEchoCount());
	}

	/* Tell the server thread to exit */
	ss.signal(SIG_EXIT);
	me.waitFor(1L << SIG_CONTINUE);
    }
}

/* Expected Output:
Message #1:
'Things involving the computer fill me with a childlike terror.  Now, if it were a nice ogre or some such, I'd be more in my element.' -- Giles
Echo count = 1
Message #2:
'You were looking at my neck.'
'What?'
'You were checking out my neck, I saw that.'
'No, I wasn't.'
'Just keep your distance, pal.'
'I wasn't looking at your neck.'
'I told you to eat before we left.' -- Xander and Angel
Echo count = 2
Message #3:
'We like to talk big, vampires do. 'I'm going to destroy the world.' That's just tough guy talk. Strutting around with your friends over a pint of blood. The truth is, I like this world. You've got dog racing. Manchester United. And you've got people. Billions of people walking around like Happy Meals with legs. It's all right here.' -- Spike
Echo count = 3
Message #4:
'How can I thank you, you mysterious black-clad hunk of a night-thing.' 'No need, little lady.  Your tears of gratitude are enough for me. You see, I was once a bad-ass vampire. But love and a pesky curse defanged me and now I'm just a big fluffy puppy with bad teeth. No, not the hair. Never the hair.''But there must be some way I can show my appreciation.' 'No. Helping those in need is my job and working up a load of sexual tension and prancing away like a magnificent poof is truly thanks enough.' 'I understand, I have a nephew who's gay.' 'Say no more, evil's afoot and I'm almost out of that nancy-boy hair gel I like so much. Quickly! To the Angel-mobile.  Away!' - Spike
Echo count = 4
*/
