
import edu.utah.janosvm.sys.Team;
import edu.utah.janosvm.sys.TeamHandle;
import edu.utah.janosvm.sys.Importable;
import edu.utah.janosvm.sys.Exportable;
import edu.utah.janosvm.sys.ExportManager;
import edu.utah.janosvm.sys.TemporaryExport;
import edu.utah.janosvm.sys.NoSuchExportException;

import edu.utah.janosvm.resources.StdioSpec;
import edu.utah.janosvm.resources.ResourceSpec;
import edu.utah.janosvm.resources.NameSpaceSpec;
import edu.utah.janosvm.resources.NameSpaceGroup;
import edu.utah.janosvm.resources.NameSpaceResource;

public class RevokeImport
{

    private static class GenericImport
	extends Importable
    {
	public Object obj;

	protected void bindTo(Exportable ex)
	{
	    this.obj = ex;
	}

	protected void unbindFrom()
	{
	    this.obj = null;
	}
    }

    private static class GenericExport
	extends Exportable
    {
	public Object obj;

	public GenericExport(Object val)
	{
	    this.obj = val;
	}
    }

    private static Team lt = Team.current();

    /**
     * A temporary export which will be revoked when the last import referring
     * to it is gone.
     */
    public static class TempTest
	extends Exportable
	implements TemporaryExport
    {
	public int value = 1;

	public boolean expired()
	{
	    return true;
	}

	public void finalize()
	{
	    System.out.println("Temp export is finalized");
	}
    }

    /**
     * Importable for our termporary import above.
     */
    public static class TempTestImport
	extends Importable
    {
	public Object obj;

	protected void bindTo(Exportable ex)
	{
	    this.obj = ex;
	}

	protected void unbindFrom()
	{
	    this.obj = null;
	}
    }

    private static void doTempExport(Importable im)
    {
	try
	{
	    Team team = Team.current();
	    /* Load the class locally and create a new instance */
	    Class cl = NameSpaceResource.loadTeamClass("RevokeImport$TempTest");
	    Object tt = cl.newInstance();
	    Exportable ex = (Exportable)tt;
	    ExportManager em;

	    /* Export the object and subscribe the given Importable to it */
	    em = team.exportObject("tempExport".clone(), ex);
	    em.addImporter(im);
	    System.out.println("Exported temp");
	    System.gc();
	}
	catch(Exception e)
	{
	    System.out.println("Exception while doing export?");
	    e.printStackTrace();
	}
    }
    
    public static void tempExport()
	throws Exception
    {
	Team team = Team.current();
	final TempTestImport im = new TempTestImport();
	
	/* Create a Team.create to export an object */
	TeamHandle child = Team.create("Child 1 of "
				       + lt.getName()
				       + lt.getID(),
				       new ResourceSpec[] {
					       StdioSpec.currentFD(),
					       NameSpaceSpec.current()
				       });

	child.switchTo();
	{
	    /*
	     * Call another function to make sure there aren't any stray
	     * pointers on the stack the gc will pick up.
	     */
	    doTempExport(im);
	}
	child.returnFrom();

	synchronized(im)
	{
	    /* Make sure its there */
	    if( im.obj != null )
	    {
		System.out.println("Export is still alive");
	    }
	    else
	    {
		System.out.println("Export vanished?");
	    }
	}
	/*
	 * Revoke the import which should be the only reference to the export
	 * and therefore the export should go away.
	 */
	team.cancelImport(im);

	/*
	 * Visit the child and GC a few times to make sure the exportable is
	 * gone.
	 */
	child.switchTo();
	{
	    System.gc();
	    System.gc();
	    System.gc();
	    System.gc();
	    System.gc();
	    System.gc();
	}
	child.returnFrom();
	try
	{
	    /* Make sure we can't import it anymore */
	    child.importObject(im, "tempExport");
	}
	catch(NoSuchExportException e)
	{
	    System.out.println("Temp export can't be imported");
	}
	child.terminate();
    }
    
    public static void crossTeamHandle()
	throws Exception
    {
	TeamHandle kernel = Team.current().getKernelTeam();
	Team parent = Team.current();
	
	/* Export a class group for our child teams */
	NameSpaceResource.exportClassGroup("RevokeImport");
	
	NameSpaceSpec nss = new NameSpaceSpec(new NameSpaceGroup[] {
		new NameSpaceGroup(kernel, "kernel", "identity"),
		new NameSpaceGroup(kernel, "janosvm", "identity"),
		new NameSpaceGroup(parent, "RevokeImport", "identity")
	});
	
	/* Pretty simple, just import an object from another team */
	TeamHandle child = Team.create("Child 2 of "
				       + lt.getName()
				       + lt.getID(),
				       new ResourceSpec[] {
					       StdioSpec.currentFD(),
					       nss
				       });
	GenericImport gi = new GenericImport();

	child.switchTo();
	{
	    Team child_team = Team.current();
	    GenericExport ge = new GenericExport(new Integer(1));

	    child_team.exportObject("i".clone(), ge);
	    System.gc();
	}
	child.returnFrom();
	child.importObject(gi, "i");
	synchronized(gi)
	{
	    if( ((Integer)((GenericExport)gi.obj).obj).intValue() == 1 )
	    {
		System.out.println("Got cross team export");
	    }
	}
	child.terminate();
	synchronized(gi)
	{
	    if( gi.obj == null )
		System.out.println("Cross team export was revoked");
	}
	System.gc();
	System.out.println("Still here");
    }

    public static void local()
	throws Exception
    {
	Team team = Team.current();
	/* Export an object and import it in the same team */
	Integer i = new Integer(1);
	Exportable ge = new GenericExport(i);
	GenericImport gi = new GenericImport();

	team.exportObject("i", ge);
	team.importObject(gi, "i");
	if( gi.obj == ge )
	{
	    System.out.println("Got export");
	}
	else
	{
	    System.out.println("Didn't get export");
	}
	team.revokeObject("i");
	if( gi.obj == null )
	{
	    System.out.println("Import revoked");
	}
	else
	{
	    System.out.println("Import still available");
	}
    }
    
    public static void main(String args[])
	throws Exception
    {
	local();
	crossTeamHandle();
	tempExport();
    }
    
}

/* Expected Output:
Got export
Import revoked
Got cross team export
Cross team export was revoked
Still here
Exported temp
Export is still alive
Temp export is finalized
Temp export can't be imported
*/
